# -*- coding: utf-8 -*-
# pylint: disable=line-too-long, unused-wildcard-import
"""
Version 4 of *HydPy-L-Land* is still under development and likely to change in the
future.  Its goal is to extend |lland_v3| with processes dealing with the interception
of snow in canopies, similar to the `SCHNEEINTERZEPTION` option of `LARSIM`_.  The
publically available `LARSIM`_ documentation does not provide all information required
for implementing a perfect emulation, but detailed comparisons suggest that the
essential process equations are defined similarly.

.. _`LARSIM`: http://www.larsim.de/en/the-model/

The following list summarises the current status of |lland_v4| by discussing its
main differences to |lland_v3|:

 * Snow interception occurs only in tree canopies, which is why the results of
   |lland_v4| differ from those of |lland_v3| only for the land-use types |LAUBW|,
   |MISCHW|, and |NADELW|.
 * There is are two additional storages, |SInz| and |STInz|, which handle the total
   and the frozen amount of intercepted snowfall, placed between the "normal"
   interception and the "normal" snow module.
 * If not stated otherwise, |SInz| and |STInz| correspond to |WAeS| and |WATS| and
   are modified by the same or functionally identical methods.
 * If the `LARSIM`_ documentation does not suggest otherwise, we reuse the names
   of the "normal" snow module and suffix "Inz" to it or replace its "Snow" suffix
   (for example, |WaDa| becomes |WaDaInz| and |NetRadiationSnow| becomes
   |NetRadiationInz|).
 * Only a fraction of the water leaving the "normal" interception module reaches the
   snow interception module (see |Calc_NBesInz_V1| and |Calc_SnowIntRate_V1|).
 * The surface temperature of the intercepted snow is always identical with its bulk
   temperature (|TempSInz|), except for the calculation of the latent heat flux
   (see |Return_WSurfInz_V1|).
 * The energy balance of the intercepted snow does not (need to) consider the soil
   heat flux.
 * The radiation exchange with the atmosphere depends on how much the intercepted
   snow shields the ground (see |Calc_NetShortwaveRadiationInz_V1| and
   |Return_NetLongwaveRadiationInz_V1|).
 * During periods with intercepted snow, water evaporates (or sublimates) from the snow
   interception module (|EvSInz|) but not from the "normal" interception module (|EvI|).

:cite:t:`ref-LARSIM` discusses many of these points in more detail.

In our experience, |lland_v4| tends to calculate thicker snowpacks on the ground
than |lland_v3| (the highest values of |WAeS| are usually larger when applying
|lland_v4|).  This counterintuitive behaviour seems to be related to disabling the
"normal" interception evaporation (|EvI|) during snow interception periods and is
still under discussion.

We will add further explanations and references as soon as we can consolidate
|lland_v4|.

The following integration tests are identical to the ones of |lland_v3|, except for
the additional snow interception processes.

Integration tests
=================

.. how_to_understand_integration_tests::

daily simulation
________________

>>> from hydpy import pub, Timegrid
>>> pub.timegrids = "1997-08-01", "1997-09-01", "1d"
>>> from hydpy.models.lland_v4 import *
>>> parameterstep("1h")
>>> from hydpy import Node, Element
>>> outlet = Node("outlet")
>>> inlet = Node("inlet")
>>> land = Element("land", inlets=inlet, outlets=outlet)
>>> land.model = model
>>> nhru(1)
>>> ft(1.0)
>>> fhru(1.0)
>>> from hydpy import IntegrationTest
>>> IntegrationTest.plotting_options.axis1 = inputs.nied
>>> IntegrationTest.plotting_options.axis2 = fluxes.qah
>>> test = IntegrationTest(land)
>>> test.dateformat = "%Y-%d-%m"

.. _lland_v4_acker_summer_daily:

acre (summer)
-------------

>>> lnk(ACKER)
>>> kg(0.94)
>>> kt(0.0)
>>> hinz(0.2)
>>> lai.acker_aug = 3.5
>>> tgr(0.0)
>>> trefn(0.0)
>>> tsp(4.0)
>>> pwmax(1.43)
>>> wmax(309.0)
>>> fk(199.0)
>>> pwp(119.4)
>>> kapgrenz(option="0_WMax/10")
>>> kapmax(0.08)
>>> beta(0.001/24)
>>> fbeta(1.0)
>>> rbeta(False)
>>> dmax(r_dmax=0.55)
>>> dmin(r_dmin=10.0)
>>> bsf(0.3)
>>> volbmax(inf)
>>> gsbmax(1.0)
>>> gsbgrad1(inf)
>>> gsbgrad2(inf)
>>> a1(0.0)
>>> a2(0.5)
>>> tind(0.191056)
>>> eqb(35000.0)
>>> eqi1(400.0)
>>> eqi2(400.0)
>>> eqd1(200.0)
>>> eqd2(50.0)

>>> negq(True)

>>> measuringheightwindspeed(10.0)
>>> emissivity(0.95)
>>> py(119.4)
>>> cropheight.acker_aug = 0.4
>>> albedo.acker_aug = 0.24
>>> wg2z.aug = -2.9
>>> surfaceresistance.acker = 40.0

>>> p1strahl(0.5)
>>> p2strahl(1.0/35.0)
>>> p1wind(0.6)
>>> p2wind(1.0/70.0)
>>> turb0(2.0)
>>> turb1(1.6)
>>> albedo0snow(0.9)
>>> snowagingfactor(0.35)
>>> refreezeflag(True)
>>> ktschnee(inf)

>>> fvf(0.3)
>>> bsff(2.0)

>>> p1simax(8.0)
>>> p2simax(1.5)
>>> p1sirate(0.2)
>>> p2sirate(0.02)
>>> p3sirate(0.003)

>>> test.inits = (
...     (states.inzp, 0.0),
...     (states.stinz, 0.0),
...     (states.sinz, 0.0),
...     (states.esnowinz, 0.0),
...     (states.asinz, 0.0),
...     (states.wats, 0.0),
...     (states.waes, 0.0),
...     (states.esnow, 0.0),
...     (states.taus, 0.0),
...     (states.ebdn, 350.0),
...     (states.bowa, 72.0),
...     (states.qdgz1, 0.0),
...     (states.qdgz2, 0.0),
...     (states.qigz1, 0.0),
...     (states.qigz2, 0.0),
...     (states.qbgz, 0.0),
...     (states.qdga1, 0.0),
...     (states.qdga2, 0.0),
...     (states.qiga1, 0.0),
...     (states.qiga2, 0.0),
...     (states.qbga, 0.0))

>>> inputs.nied.series = (
...     13.3, 0.1, 0.0, 0.1, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.1, 0.0, 0.0, 2.7,
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 2.0, 0.3, 0.0, 0.1, 10.5, 6.5, 15.0, 9.2, 0.0,
...     0.0)
>>> inputs.teml.series = (
...     16.1, 17.3, 18.9, 17.8, 18.5, 19.9, 20.8, 20.9, 22.3, 23.0, 23.1, 23.2, 23.2,
...     21.7, 20.8, 19.5, 21.0, 21.6, 21.1, 22.4, 22.5, 19.6, 20.3, 24.1, 23.7, 21.1,
...     20.9, 20.2, 16.1, 17.0, 18.2)
>>> inputs.sunshineduration.series = (
...     6.3, 1.7, 4.5, 12.4, 13.9, 13.0, 13.8, 12.3, 13.1, 12.8, 13.1, 13.3, 12.7,
...     10.2, 9.4, 10.3, 11.1, 11.0, 8.5, 11.3, 12.4, 0.1, 6.7, 10.4, 6.5, 4.9, 6.6,
...     0.3, 0.1, 5.0, 3.8)
>>> inputs.windspeed.series = (
...     2.4, 1.6, 1.5, 2.8, 3.0, 3.5, 3.8, 3.3, 1.8, 1.9, 3.0, 2.5, 1.6, 2.7, 3.2, 1.3,
...     1.9, 2.0, 2.6, 3.6, 2.5, 2.1, 1.5, 2.2, 1.9, 2.3, 2.1, 2.6, 2.6, 2.1, 2.1)
>>> inputs.relativehumidity.series = (
...     86.2, 85.8, 82.4, 77.6, 74.9, 77.9, 73.5, 80.1, 72.9, 68.5, 68.6, 66.0, 69.0,
...     75.6, 81.4, 79.0, 75.9, 70.0, 70.3, 69.0, 62.1, 84.5, 83.6, 76.5, 76.6, 84.2,
...     85.8, 86.5, 89.6, 78.3, 78.5)
>>> inputs.atmosphericpressure.series = (
...     1007.0, 1013.0, 1016.0, 1017.0, 1018.0, 1018.0, 1019.0, 1019.0, 1019.0, 1020.0,
...     1021.0, 1022.0, 1019.0, 1015.0, 1016.0, 1019.0, 1019.0, 1018.0, 1019.0, 1019.0,
...     1017.0, 1015.0, 1015.0, 1011.0, 1008.0, 1006.0, 1005.0, 1000.0, 1004.0, 1011.0,
...     1014.0)
>>> inlet.sequences.sim.series = 0.0

>>> inputs.possiblesunshineduration.series = (
...     15.69078473, 15.62882066, 15.56613383, 15.50275437, 15.43871163, 15.37403418,
...     15.3087498, 15.24288548, 15.17646742, 15.10952106, 15.04207104, 14.97414122,
...     14.90575471, 14.83693386, 14.76770029, 14.69807488, 14.62807779, 14.5577285,
...     14.48704578, 14.41604776, 14.34475191, 14.27317506, 14.20133345, 14.12924271,
...     14.05691791, 13.98437357, 13.91162366, 13.83868165, 13.76556053, 13.69227282,
...     13.61883057)
>>> inputs.globalradiation.series = (
...     190.2514903, 127.6607822, 164.4012978, 269.0205016, 287.8486672, 274.6742631,
...     284.0478917, 262.9712633, 272.2564135, 267.0192178, 269.6265466, 270.8782909,
...     261.6971638, 227.9621155, 216.3563213, 226.6379036, 235.5285025, 232.8494284,
...     199.7134896, 233.7719021, 246.1212861, 90.58551, 171.9449932, 216.5498126,
...     166.9168296, 145.9856109, 165.5438903, 87.6933456, 84.356885, 142.3092025,
...     126.7220785)

.. integration-test::

    >>> parameters.update()
    >>> test.reset_inits()
    >>> conditions = sequences.conditions
    >>> conditions_acker_summer = test(
    ...     "lland_v4_acker_summer_daily",
    ...     axis1=(inputs.nied, fluxes.qah), axis2=states.bowa,
    ...     get_conditions="1997-08-03")
    |       date | nied | teml | sunshineduration | possiblesunshineduration | globalradiation | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration | dailypossiblesunshineduration | dailyglobalradiation |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair |    g |        tz |         wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |   nbes | sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |     evpo |   evi |      evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatinz | wlatsnow | wsensinz | wsenssnow | wsurfinz | wsurf | sff | fvg |   wada |      qdb | qib1 | qib2 | qbb |     qkap |     qdgz |       qah |        qa | inzp | stinz | sinz | esnowinz | asinz | wats | waes | esnow | taus |       ebdn |      bowa |    qdgz1 | qdgz2 | qigz1 | qigz2 |      qbgz |    qdga1 | qdga2 | qiga1 | qiga2 |      qbga | inlet |    outlet |
    --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |                15.690785 |       190.25149 |             86.2 |       2.4 |              1007.0 | 0.0 | 0.0 |    16.1 |                  86.2 |                   6.3 |                     15.690785 |            190.25149 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                18.326791 |                     18.326791 |                         0.0 |                          0.0 |                      1.170311 |                           1.170311 |            15.797694 |                 15.797694 |     991.202306 |   1.205601 | -2.9 | 11.733333 |      -26.2 |            144.591133 |                      0.0 |                144.591133 |                 144.591133 |                 21.732733 |                     0.0 |                      0.0 |     122.8584 |             0.0 |              0.0 |          122.8584 |             58.847345 |                 100.0 |               111.309423 |               99.889466 | 11.802 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.623832 |   0.7 | 1.720128 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 11.802 | 0.761343 |  0.0 |  0.0 | 0.0 |      0.0 | 0.761343 |  0.196005 |  0.002269 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |      373.3 | 81.320529 | 0.761343 |   0.0 |   0.0 |   0.0 |       0.0 | 0.196005 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.002269 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |                15.628821 |      127.660782 |             85.8 |       1.6 |              1013.0 | 0.0 | 0.0 |    17.3 |                  85.8 |                   1.7 |                     15.628821 |           127.660782 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                19.779139 |                     19.779139 |                         0.0 |                          0.0 |                      1.251029 |                           1.251029 |            16.970501 |                 16.970501 |     996.029499 |   1.207284 | -2.9 | 18.443733 |     6.8624 |             97.022194 |                      0.0 |                 97.022194 |                  97.022194 |                 11.219797 |                     0.0 |                      0.0 |    85.802398 |             0.0 |              0.0 |         85.802398 |             88.271017 |                 100.0 |                 98.30202 |               94.620613 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.555581 | 0.094 | 1.697173 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 |  0.263667 |  0.003052 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |   363.5376 | 79.623355 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.263667 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.003052 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |                15.566134 |      164.401298 |             82.4 |       1.5 |              1016.0 | 0.0 | 0.0 |    18.9 |                  82.4 |                   4.5 |                     15.566134 |           164.401298 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                21.871576 |                     21.871576 |                         0.0 |                          0.0 |                      1.365939 |                           1.365939 |            18.022179 |                 18.022179 |     997.977821 |   1.203775 | -2.9 | 15.632162 | -19.607027 |            124.944986 |                      0.0 |                124.944986 |                 124.944986 |                 16.044181 |                     0.0 |                      0.0 |   108.900806 |             0.0 |              0.0 |        108.900806 |             94.155751 |                 100.0 |               100.628883 |               95.574208 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.286385 |   0.0 | 2.329453 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 |  0.140696 |  0.001628 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 380.244627 | 77.293902 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.140696 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.001628 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |                15.502754 |      269.020502 |             77.6 |       2.8 |              1017.0 | 0.0 | 0.0 |    17.8 |                  77.6 |                  12.4 |                     15.502754 |           269.020502 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                20.413369 |                     20.413369 |                         0.0 |                          0.0 |                      1.286025 |                           1.286025 |            15.840774 |                 15.840774 |    1001.159226 |    1.21051 | -2.9 | 20.443786 |  15.862716 |            204.455581 |                      0.0 |                204.455581 |                 204.455581 |                 35.980407 |                     0.0 |                      0.0 |   168.475175 |             0.0 |              0.0 |        168.475175 |             50.440581 |                 100.0 |                103.85044 |               96.853188 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.847296 | 0.094 | 3.314337 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 |  0.075077 |  0.000869 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 361.481911 | 73.979565 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.075077 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.000869 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |                15.438712 |      287.848667 |             74.9 |       3.0 |              1018.0 | 0.0 | 0.0 |    18.5 |                  74.9 |                  13.9 |                     15.438712 |           287.848667 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                21.331125 |                     21.331125 |                         0.0 |                          0.0 |                      1.336407 |                           1.336407 |            15.977013 |                 15.977013 |    1002.022987 |   1.208738 | -2.9 | 15.040124 | -20.759257 |            218.764987 |                      0.0 |                218.764987 |                 218.764987 |                 39.523937 |                     0.0 |                      0.0 |   179.241051 |             0.0 |              0.0 |        179.241051 |             47.077876 |                 100.0 |               108.495906 |                98.61702 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 6.638473 |   0.0 |  3.73228 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 |  0.040062 |  0.000464 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 379.341169 | 70.247285 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.040062 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.000464 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |                15.374034 |      274.674263 |             77.9 |       3.5 |              1018.0 | 0.0 | 0.0 |    19.9 |                  77.9 |                  13.0 |                     15.374034 |           274.674263 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                23.275348 |                     23.275348 |                         0.0 |                          0.0 |                      1.442189 |                           1.442189 |            18.131496 |                 18.131496 |     999.868504 |   1.201995 | -2.9 |  20.18359 |   1.701539 |             208.75244 |                      0.0 |                 208.75244 |                  208.75244 |                 32.936813 |                     0.0 |                      0.0 |   175.815627 |             0.0 |              0.0 |        175.815627 |             40.352465 |                 100.0 |               113.827642 |              100.519454 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 6.828968 |   0.0 | 3.652387 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 |  0.021377 |  0.000247 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 374.739629 | 66.594898 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.021377 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.000247 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |                 15.30875 |      284.047892 |             73.5 |       3.8 |              1019.0 | 0.0 | 0.0 |    20.8 |                  73.5 |                  13.8 |                      15.30875 |           284.047892 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                24.605348 |                     24.605348 |                         0.0 |                          0.0 |                      1.513854 |                           1.513854 |            18.084931 |                 18.084931 |    1000.915069 |   1.199521 | -2.9 | 18.858347 | -11.649921 |            215.876398 |                      0.0 |                215.876398 |                 215.876398 |                  35.31351 |                     0.0 |                      0.0 |   180.562888 |             0.0 |              0.0 |        180.562888 |             37.166744 |                 100.0 |               119.167697 |              102.297207 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 7.945133 |   0.0 | 4.128687 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 |  0.011407 |  0.000132 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  383.48955 |  62.46621 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.011407 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.000132 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |                15.242885 |      262.971263 |             80.1 |       3.3 |              1019.0 | 0.0 | 0.0 |    20.9 |                  80.1 |                  12.3 |                     15.242885 |           262.971263 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 24.75714 |                      24.75714 |                         0.0 |                          0.0 |                      1.521999 |                           1.521999 |            19.830469 |                 19.830469 |     999.169531 |   1.198331 | -2.9 | 21.378324 |   2.869942 |             199.85816 |                      0.0 |                 199.85816 |                  199.85816 |                 28.392598 |                     0.0 |                      0.0 |   171.465562 |             0.0 |              0.0 |        171.465562 |             42.798069 |                 100.0 |               125.381825 |              104.230888 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 6.427975 |   0.0 | 3.518399 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 |  0.006087 |   0.00007 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 377.719608 | 58.947811 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.006087 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |   0.00007 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |                15.176467 |      272.256414 |             72.9 |       1.8 |              1019.0 | 0.0 | 0.0 |    22.3 |                  72.9 |                  13.1 |                     15.176467 |           272.256414 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 26.96965 |                      26.96965 |                         0.0 |                          0.0 |                      1.639967 |                           1.639967 |            19.660875 |                 19.660875 |     999.339125 |   1.192729 | -2.9 |  19.71658 | -15.500518 |            206.914874 |                      0.0 |                206.914874 |                 206.914874 |                 31.114889 |                     0.0 |                      0.0 |   175.799986 |             0.0 |              0.0 |        175.799986 |             78.463126 |                 100.0 |               130.860307 |              105.804393 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 6.036589 |   0.0 | 4.091259 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 |  0.003248 |  0.000038 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 390.320126 | 54.856552 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.003248 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.000038 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |                15.109521 |      267.019218 |             68.5 |       1.9 |              1020.0 | 0.0 | 0.0 |    23.0 |                  68.5 |                  12.8 |                     15.109521 |           267.019218 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                28.139149 |                     28.139149 |                         0.0 |                          0.0 |                      1.701779 |                           1.701779 |            19.275317 |                 19.275317 |    1000.724683 |   1.191257 | -2.9 |  23.34553 |   2.073177 |            202.934606 |                      0.0 |                202.934606 |                 202.934606 |                 32.010643 |                     0.0 |                      0.0 |   170.923963 |             0.0 |              0.0 |        170.923963 |             74.333488 |                 100.0 |               137.497212 |               107.59145 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 6.493896 |   0.0 | 4.345753 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 |  0.001733 |   0.00002 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 385.346949 | 50.510799 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.001733 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |   0.00002 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |                15.042071 |      269.626547 |             68.6 |       3.0 |              1021.0 | 0.0 | 0.0 |    23.1 |                  68.6 |                  13.1 |                     15.042071 |           269.626547 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                28.309776 |                     28.309776 |                         0.0 |                          0.0 |                      1.710768 |                           1.710768 |            19.420506 |                 19.420506 |    1001.579494 |   1.191966 | -2.9 | 21.913255 |  -7.120473 |            204.916175 |                      0.0 |                204.916175 |                 204.916175 |                 32.400014 |                     0.0 |                      0.0 |   172.516162 |             0.0 |              0.0 |        172.516162 |             47.077876 |                 100.0 |               144.951913 |              109.449121 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 7.914607 |   0.0 | 4.547949 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 |  0.000925 |  0.000011 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 389.567421 | 45.962849 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000925 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.000011 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |                14.974141 |      270.878291 |             66.0 |       2.5 |              1022.0 | 0.0 | 0.0 |    23.2 |                  66.0 |                  13.3 |                     14.974141 |           270.878291 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                28.481304 |                     28.481304 |                         0.0 |                          0.0 |                      1.719796 |                           1.719796 |             18.79766 |                  18.79766 |     1003.20234 |   1.193016 | -2.9 | 23.128751 |  -0.427496 |            205.867501 |                      0.0 |                205.867501 |                 205.867501 |                 34.565783 |                     0.0 |                      0.0 |   171.301718 |             0.0 |              0.0 |        171.301718 |             56.493451 |                 100.0 |               153.358051 |              111.375228 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 7.566564 | 0.094 | 4.554398 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 |  0.000494 |  0.000006 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 387.094917 | 41.408451 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000494 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.000006 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |                14.905755 |      261.697164 |             69.0 |       1.6 |              1019.0 | 0.0 | 0.0 |    23.2 |                  69.0 |                  12.7 |                     14.905755 |           261.697164 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                28.481304 |                     28.481304 |                         0.0 |                          0.0 |                      1.719796 |                           1.719796 |              19.6521 |                   19.6521 |       999.3479 |    1.18911 | -2.9 | 22.416669 |  -4.699983 |            198.889844 |                      0.0 |                198.889844 |                 198.889844 |                  31.34697 |                     0.0 |                      0.0 |   167.542875 |             0.0 |              0.0 |        167.542875 |             88.271017 |                 100.0 |                162.65311 |               113.31833 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.996216 |   0.0 | 4.137046 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 |  0.000263 |  0.000003 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |   388.8949 | 37.271405 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000263 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.000003 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |                14.836934 |      227.962116 |             75.6 |       2.7 |              1015.0 | 0.0 | 0.0 |    21.7 |                  75.6 |                  10.2 |                     14.836934 |           227.962116 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 26.00118 |                      26.00118 |                         0.0 |                          0.0 |                        1.5885 |                             1.5885 |            19.656892 |                 19.656892 |     995.343108 |   1.190432 | -2.9 | 22.935065 |   7.410388 |            173.251208 |                      0.0 |                173.251208 |                 173.251208 |                 25.918437 |                     0.0 |                      0.0 |   147.332771 |             0.0 |              0.0 |        147.332771 |             52.308751 |                 100.0 |                172.21097 |              115.125428 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.983775 |   0.0 | 3.428735 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 |  0.000141 |  0.000002 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 378.584513 |  33.84267 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000141 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.000002 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |                  14.7677 |      216.356321 |             81.4 |       3.2 |              1016.0 | 0.0 | 0.0 |    20.8 |                  81.4 |                   9.4 |                       14.7677 |           216.356321 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                24.605348 |                     24.605348 |                         0.0 |                          0.0 |                      1.513854 |                           1.513854 |            20.028753 |                 20.028753 |     995.971247 |   1.195095 | -2.9 | 19.965673 |  -5.005962 |            164.430804 |                      0.0 |                164.430804 |                 164.430804 |                 23.397902 |                     0.0 |                      0.0 |   141.032902 |             0.0 |              0.0 |        141.032902 |             44.135508 |                 100.0 |               181.322396 |              116.674937 |  1.838 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.470835 |   0.7 | 2.505678 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  1.838 | 0.049893 |  0.0 |  0.0 | 0.0 |      0.0 | 0.049893 |   0.01292 |   0.00015 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 380.690475 | 33.125099 | 0.049893 |   0.0 |   0.0 |   0.0 |       0.0 |  0.01292 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |   0.00015 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |                14.698075 |      226.637904 |             79.0 |       1.3 |              1019.0 | 0.0 | 0.0 |    19.5 |                  79.0 |                  10.3 |                     14.698075 |           226.637904 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                22.704676 |                     22.704676 |                         0.0 |                          0.0 |                      1.411269 |                           1.411269 |            17.936694 |                 17.936694 |    1001.063306 |   1.204916 | -2.9 |  20.57219 |   6.433141 |            172.244807 |                      0.0 |                172.244807 |                 172.244807 |                 28.794151 |                     0.0 |                      0.0 |   143.450656 |             0.0 |              0.0 |        143.450656 |            108.641252 |                 100.0 |               183.411459 |               116.86708 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 4.081206 |   0.0 | 2.836506 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 |  0.017319 |    0.0002 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 371.357334 | 30.288593 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.017319 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |    0.0002 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |                14.628078 |      235.528503 |             75.9 |       1.9 |              1019.0 | 0.0 | 0.0 |    21.0 |                  75.9 |                  11.1 |                     14.628078 |           235.528503 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                24.909749 |                     24.909749 |                         0.0 |                          0.0 |                      1.530181 |                           1.530181 |            18.906499 |                 18.906499 |    1000.093501 |   1.198338 | -2.9 | 17.884246 | -18.694526 |            179.001662 |                      0.0 |                179.001662 |                 179.001662 |                 29.177702 |                     0.0 |                      0.0 |    149.82396 |             0.0 |              0.0 |         149.82396 |             74.333488 |                 100.0 |               192.482064 |              118.222742 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.175279 |   0.0 | 3.276924 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.03799 |      0.0 |  0.009173 |  0.000106 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 387.151861 | 27.049659 |      0.0 |   0.0 |   0.0 |   0.0 |  -0.03799 | 0.009242 |   0.0 |   0.0 |   0.0 | -0.000068 |   0.0 |  0.000106 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |                14.557728 |      232.849428 |             70.0 |       2.0 |              1018.0 | 0.0 | 0.0 |    21.6 |                  70.0 |                  11.0 |                     14.557728 |           232.849428 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                25.842752 |                     25.842752 |                         0.0 |                          0.0 |                      1.580055 |                           1.580055 |            18.089927 |                 18.089927 |     999.910073 |   1.195081 | -2.9 | 22.433069 |   4.998415 |            176.965566 |                      0.0 |                176.965566 |                 176.965566 |                 31.281288 |                     0.0 |                      0.0 |   145.684278 |             0.0 |              0.0 |        145.684278 |             70.616813 |                 100.0 |               204.991875 |              119.966911 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.737858 |   0.0 | 3.581415 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.239244 |      0.0 |  0.004367 |  0.000051 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 379.253445 | 23.707488 |      0.0 |   0.0 |   0.0 |   0.0 | -0.239244 | 0.004931 |   0.0 |   0.0 |   0.0 | -0.000565 |   0.0 |  0.000051 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |                14.487046 |       199.71349 |             70.3 |       2.6 |              1019.0 | 0.0 | 0.0 |    21.1 |                  70.3 |                   8.5 |                     14.487046 |            199.71349 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                25.063177 |                     25.063177 |                         0.0 |                          0.0 |                        1.5384 |                             1.5384 |            17.619414 |                 17.619414 |    1001.380586 |   1.198506 | -2.9 | 20.158326 |  -5.650046 |            151.782252 |                      0.0 |                151.782252 |                 151.782252 |                 26.711296 |                     0.0 |                      0.0 |   125.070956 |             0.0 |              0.0 |        125.070956 |             54.320626 |                 100.0 |               221.727637 |              122.069797 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.866935 |   0.0 | 3.296835 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.446913 |      0.0 |  0.000839 |   0.00001 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 382.003492 | 20.857567 |      0.0 |   0.0 |   0.0 |   0.0 | -0.446913 | 0.002631 |   0.0 |   0.0 |   0.0 | -0.001792 |   0.0 |   0.00001 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |                14.416048 |      233.771902 |             69.0 |       3.6 |              1019.0 | 0.0 | 0.0 |    22.4 |                  69.0 |                  11.3 |                     14.416048 |           233.771902 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                27.134082 |                     27.134082 |                         0.0 |                          0.0 |                      1.648679 |                           1.648679 |            18.722517 |                 18.722517 |    1000.277483 |   1.192743 | -2.9 | 20.950339 |  -8.697966 |            177.666646 |                      0.0 |                177.666646 |                 177.666646 |                 31.116622 |                     0.0 |                      0.0 |   146.550024 |             0.0 |              0.0 |        146.550024 |             39.231563 |                 100.0 |               241.230014 |              124.199013 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  7.82923 |   0.0 | 3.891994 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.623996 |      0.0 |   -0.0023 | -0.000027 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 387.801458 | 17.589568 |      0.0 |   0.0 |   0.0 |   0.0 | -0.623996 | 0.001404 |   0.0 |   0.0 |   0.0 | -0.003704 |   0.0 | -0.000027 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |                14.344752 |      246.121286 |             62.1 |       2.5 |              1017.0 | 0.0 | 0.0 |    22.5 |                  62.1 |                  12.4 |                     14.344752 |           246.121286 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                27.299387 |                     27.299387 |                         0.0 |                          0.0 |                      1.657431 |                           1.657431 |             16.95292 |                  16.95292 |     1000.04708 |   1.190771 | -2.9 | 22.620153 |    0.72092 |            187.052177 |                      0.0 |                187.052177 |                 187.052177 |                 38.297604 |                     0.0 |                      0.0 |   148.754573 |             0.0 |              0.0 |        148.754573 |             56.493451 |                 100.0 |               274.853414 |              127.355551 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 7.322689 |   0.0 | 4.183176 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.827056 |      0.0 | -0.005541 | -0.000064 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 384.180539 | 14.233449 |      0.0 |   0.0 |   0.0 |   0.0 | -0.827056 | 0.000749 |   0.0 |   0.0 |   0.0 |  -0.00629 |   0.0 | -0.000064 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |                14.273175 |        90.58551 |             84.5 |       2.1 |              1015.0 | 0.0 | 0.0 |    19.6 |                  84.5 |                   0.1 |                     14.273175 |             90.58551 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                22.846186 |                     22.846186 |                         0.0 |                          0.0 |                      1.418946 |                           1.418946 |            19.305027 |                 19.305027 |     995.694973 |   1.199129 | -2.9 | 21.577328 |  11.863971 |             68.844988 |                      0.0 |                 68.844988 |                  68.844988 |                   7.02118 |                     0.0 |                      0.0 |    61.823807 |             0.0 |              0.0 |         61.823807 |             67.254108 |                 100.0 |               337.448215 |              131.864309 |   1.18 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.567328 |   0.7 | 1.075698 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |   1.18 | 0.013305 |  0.0 |  0.0 | 0.0 | 1.035592 | 0.013305 | -0.005779 | -0.000067 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 369.416568 | 15.360037 | 0.013305 |   0.0 |   0.0 |   0.0 | -1.035592 | 0.003825 |   0.0 |   0.0 |   0.0 | -0.009605 |   0.0 | -0.000067 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |                14.201333 |      171.944993 |             83.6 |       1.5 |              1015.0 | 0.0 | 0.0 |    20.3 |                  83.6 |                   6.7 |                     14.201333 |           171.944993 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                23.858503 |                     23.858503 |                         0.0 |                          0.0 |                      1.473678 |                           1.473678 |            19.945708 |                 19.945708 |     995.054292 |   1.195981 | -2.9 | 17.325305 | -17.848171 |            130.678195 |                      0.0 |                130.678195 |                 130.678195 |                 18.994889 |                     0.0 |                      0.0 |   111.683306 |             0.0 |              0.0 |        111.683306 |             94.155751 |                 100.0 |                311.33348 |              129.741425 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.374147 | 0.282 | 2.014865 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.96559 |      0.0 | -0.008334 | -0.000096 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 384.364739 | 14.310761 |      0.0 |   0.0 |   0.0 |   0.0 |  -0.96559 | 0.004821 |   0.0 |   0.0 |   0.0 | -0.013155 |   0.0 | -0.000096 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |                14.129243 |      216.549813 |             76.5 |       2.2 |              1011.0 | 0.0 | 0.0 |    24.1 |                  76.5 |                  10.4 |                     14.129243 |           216.549813 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                30.066256 |                     30.066256 |                         0.0 |                          0.0 |                       1.80287 |                            1.80287 |            23.000686 |                 23.000686 |     987.999314 |   1.174651 | -2.9 | 21.630378 | -14.817732 |            164.577858 |                      0.0 |                164.577858 |                 164.577858 |                 21.629737 |                     0.0 |                      0.0 |   142.948121 |             0.0 |              0.0 |        142.948121 |             64.197103 |                 100.0 |                335.42542 |              131.069494 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.573024 |   0.0 | 3.362119 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 1.030788 |      0.0 | -0.014111 | -0.000163 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  396.28247 |  11.97943 |      0.0 |   0.0 |   0.0 |   0.0 | -1.030788 | 0.002573 |   0.0 |   0.0 |   0.0 | -0.016684 |   0.0 | -0.000163 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |                14.056918 |       166.91683 |             76.6 |       1.9 |              1008.0 | 0.0 | 0.0 |    23.7 |                  76.6 |                   6.5 |                     14.056918 |            166.91683 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                29.352596 |                     29.352596 |                         0.0 |                          0.0 |                      1.765541 |                           1.765541 |            22.484088 |                 22.484088 |     985.515912 |   1.172942 | -2.9 | 25.062685 |   8.176109 |             126.85679 |                      0.0 |                 126.85679 |                  126.85679 |                 16.222254 |                     0.0 |                      0.0 |   110.634536 |             0.0 |              0.0 |        110.634536 |             74.333488 |                 100.0 |               419.576726 |               135.15103 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 4.467668 | 0.094 | 2.719352 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 1.175647 |      0.0 | -0.019204 | -0.000222 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 385.206362 | 10.435725 |      0.0 |   0.0 |   0.0 |   0.0 | -1.175647 | 0.001373 |   0.0 |   0.0 |   0.0 | -0.020577 |   0.0 | -0.000222 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |                13.984374 |      145.985611 |             84.2 |       2.3 |              1006.0 | 0.0 | 0.0 |    21.1 |                  84.2 |                   4.9 |                     13.984374 |           145.985611 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                25.063177 |                     25.063177 |                         0.0 |                          0.0 |                        1.5384 |                             1.5384 |            21.103195 |                 21.103195 |     984.896805 |   1.181557 | -2.9 | 21.872765 |   4.636593 |            110.949064 |                      0.0 |                110.949064 |                 110.949064 |                 14.602883 |                     0.0 |                      0.0 |    96.346181 |             0.0 |              0.0 |         96.346181 |             61.405925 |                 100.0 |               522.087742 |              138.401561 |   9.17 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.569826 |   0.7 | 1.597373 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |   9.17 | 0.104578 |  0.0 |  0.0 | 0.0 | 1.271567 | 0.104578 |  0.002769 |  0.000032 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 377.669769 | 19.175341 | 0.104578 |   0.0 |   0.0 |   0.0 | -1.271567 | 0.027656 |   0.0 |   0.0 |   0.0 | -0.024887 |   0.0 |  0.000032 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |                13.911624 |       165.54389 |             85.8 |       2.1 |              1005.0 | 0.0 | 0.0 |    20.9 |                  85.8 |                   6.6 |                     13.911624 |            165.54389 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 24.75714 |                      24.75714 |                         0.0 |                          0.0 |                      1.521999 |                           1.521999 |            21.241626 |                 21.241626 |     983.758374 |   1.181114 | -2.9 | 19.702227 |   -7.18664 |            125.813357 |                      0.0 |                125.813357 |                 125.813357 |                 17.349121 |                     0.0 |                      0.0 |   108.464235 |             0.0 |              0.0 |        108.464235 |             67.254108 |                 100.0 |               256.482293 |              123.963828 |   5.41 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  3.56693 |   0.7 | 1.722258 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |   5.41 | 0.090745 |  0.0 |  0.0 | 0.0 | 0.728523 | 0.090745 |   0.03159 |  0.000366 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 381.956408 | 23.500861 | 0.090745 |   0.0 |   0.0 |   0.0 | -0.728523 |  0.05997 |   0.0 |   0.0 |   0.0 |  -0.02838 |   0.0 |  0.000366 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |                13.838682 |       87.693346 |             86.5 |       2.6 |              1000.0 | 0.0 | 0.0 |    20.2 |                  86.5 |                   0.3 |                     13.838682 |            87.693346 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                23.711531 |                     23.711531 |                         0.0 |                          0.0 |                      1.465752 |                           1.465752 |            20.510475 |                 20.510475 |     979.489525 |   1.178323 | -2.9 | 20.936779 |   4.420674 |             66.646943 |                      0.0 |                 66.646943 |                  66.646943 |                  6.818024 |                     0.0 |                      0.0 |    59.828919 |             0.0 |              0.0 |         59.828919 |             54.320626 |                 100.0 |                222.94421 |               120.11308 |   13.4 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.621228 |   0.7 | 1.071953 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |   13.4 | 0.312945 |  0.0 |  0.0 | 0.0 | 0.459752 | 0.312945 |  0.101121 |   0.00117 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 374.635735 | 35.975715 | 0.312945 |   0.0 |   0.0 |   0.0 | -0.459752 | 0.131528 |   0.0 |   0.0 |   0.0 | -0.030407 |   0.0 |   0.00117 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |                13.765561 |       84.356885 |             89.6 |       2.6 |              1004.0 | 0.0 | 0.0 |    16.1 |                  89.6 |                   0.1 |                     13.765561 |            84.356885 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                18.326791 |                     18.326791 |                         0.0 |                          0.0 |                      1.170311 |                           1.170311 |            16.420805 |                 16.420805 |     987.579195 |   1.201704 | -2.9 | 18.828425 |   16.37055 |             64.111233 |                      0.0 |                 64.111233 |                  64.111233 |                  8.199212 |                     0.0 |                      0.0 |    55.912021 |             0.0 |              0.0 |         55.912021 |             54.320626 |                 100.0 |               175.502292 |              113.152806 |  7.948 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.027587 |   0.7 | 0.717164 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  7.948 | 0.249164 |  0.0 |  0.0 | 0.0 |      0.0 | 0.249164 |  0.168598 |  0.001951 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 355.365185 | 42.957387 | 0.249164 |   0.0 |   0.0 |   0.0 |       0.0 | 0.199719 |   0.0 |   0.0 |   0.0 | -0.031121 |   0.0 |  0.001951 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |                13.692273 |      142.309202 |             78.3 |       2.1 |              1011.0 | 0.0 | 0.0 |    17.0 |                  78.3 |                   5.0 |                     13.692273 |           142.309202 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                19.406929 |                     19.406929 |                         0.0 |                          0.0 |                      1.230421 |                           1.230421 |            15.195625 |                 15.195625 |     995.804375 |   1.206937 | -2.9 | 13.278507 |  -22.32896 |            108.154994 |                      0.0 |                108.154994 |                 108.154994 |                 21.775852 |                     0.0 |                      0.0 |    86.379141 |             0.0 |              0.0 |         86.379141 |             67.254108 |                 100.0 |               159.371564 |              110.061251 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.393474 |   0.0 | 2.035891 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 |  0.127624 |  0.001477 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 374.794145 | 40.921496 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.158633 |   0.0 |   0.0 |   0.0 | -0.031009 |   0.0 |  0.001477 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |                13.618831 |      126.722078 |             78.5 |       2.1 |              1014.0 | 0.0 | 0.0 |    18.2 |                  78.5 |                   3.8 |                     13.618831 |           126.722078 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                20.933479 |                     20.933479 |                         0.0 |                          0.0 |                      1.314615 |                           1.314615 |            16.432781 |                 16.432781 |     997.567219 |   1.204994 | -2.9 | 18.874047 |   4.044283 |              96.30878 |                      0.0 |                  96.30878 |                   96.30878 |                 17.491881 |                     0.0 |                      0.0 |    78.816899 |             0.0 |              0.0 |         78.816899 |             67.254108 |                 100.0 |               163.714409 |               110.73073 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.321963 |   0.0 | 2.018218 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 |   0.05375 |  0.000622 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 367.849862 | 38.903278 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.084648 |   0.0 |   0.0 |   0.0 | -0.030898 |   0.0 |  0.000622 |

>>> from hydpy import round_
>>> round_(model.check_waterbalance(conditions))
0.0

.. _lland_v4_acker_routing_daily:

:ref:`acre (routing) <lland_v1_acker_routing>`
----------------------------------------------

.. integration-test::

    >>> inlet.sequences.sim.series = 0.02
    >>> control.negq(False)
    >>> test("lland_v4_acker_routing_daily",
    ...      axis1=(inputs.nied, fluxes.qah), axis2=states.bowa)
    |       date | nied | teml | sunshineduration | possiblesunshineduration | globalradiation | relativehumidity | windspeed | atmosphericpressure |   qz |   qzh | temltag | dailyrelativehumidity | dailysunshineduration | dailypossiblesunshineduration | dailyglobalradiation |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair |    g |        tz |         wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |   nbes | sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |     evpo |   evi |      evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatinz | wlatsnow | wsensinz | wsenssnow | wsurfinz | wsurf | sff | fvg |   wada |      qdb | qib1 | qib2 | qbb |     qkap |     qdgz |      qah |       qa | inzp | stinz | sinz | esnowinz | asinz | wats | waes | esnow | taus |       ebdn |      bowa |    qdgz1 | qdgz2 | qigz1 | qigz2 |      qbgz |    qdga1 | qdga2 | qiga1 | qiga2 |      qbga | inlet |   outlet |
    --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |                15.690785 |       190.25149 |             86.2 |       2.4 |              1007.0 | 0.02 | 1.728 |    16.1 |                  86.2 |                   6.3 |                     15.690785 |            190.25149 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                18.326791 |                     18.326791 |                         0.0 |                          0.0 |                      1.170311 |                           1.170311 |            15.797694 |                 15.797694 |     991.202306 |   1.205601 | -2.9 | 11.733333 |      -26.2 |            144.591133 |                      0.0 |                144.591133 |                 144.591133 |                 21.732733 |                     0.0 |                      0.0 |     122.8584 |             0.0 |              0.0 |          122.8584 |             58.847345 |                 100.0 |               111.309423 |               99.889466 | 11.802 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.623832 |   0.7 | 1.720128 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 11.802 | 0.761343 |  0.0 |  0.0 | 0.0 |      0.0 | 0.761343 | 1.924005 | 0.022269 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |      373.3 | 81.320529 | 0.761343 |   0.0 |   0.0 |   0.0 |       0.0 | 0.196005 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.022269 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |                15.628821 |      127.660782 |             85.8 |       1.6 |              1013.0 | 0.02 | 1.728 |    17.3 |                  85.8 |                   1.7 |                     15.628821 |           127.660782 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                19.779139 |                     19.779139 |                         0.0 |                          0.0 |                      1.251029 |                           1.251029 |            16.970501 |                 16.970501 |     996.029499 |   1.207284 | -2.9 | 18.443733 |     6.8624 |             97.022194 |                      0.0 |                 97.022194 |                  97.022194 |                 11.219797 |                     0.0 |                      0.0 |    85.802398 |             0.0 |              0.0 |         85.802398 |             88.271017 |                 100.0 |                 98.30202 |               94.620613 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.555581 | 0.094 | 1.697173 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.991667 | 0.023052 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |   363.5376 | 79.623355 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.263667 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.023052 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |                15.566134 |      164.401298 |             82.4 |       1.5 |              1016.0 | 0.02 | 1.728 |    18.9 |                  82.4 |                   4.5 |                     15.566134 |           164.401298 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                21.871576 |                     21.871576 |                         0.0 |                          0.0 |                      1.365939 |                           1.365939 |            18.022179 |                 18.022179 |     997.977821 |   1.203775 | -2.9 | 15.632162 | -19.607027 |            124.944986 |                      0.0 |                124.944986 |                 124.944986 |                 16.044181 |                     0.0 |                      0.0 |   108.900806 |             0.0 |              0.0 |        108.900806 |             94.155751 |                 100.0 |               100.628883 |               95.574208 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.286385 |   0.0 | 2.329453 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.868696 | 0.021628 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 380.244627 | 77.293902 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.140696 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.021628 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |                15.502754 |      269.020502 |             77.6 |       2.8 |              1017.0 | 0.02 | 1.728 |    17.8 |                  77.6 |                  12.4 |                     15.502754 |           269.020502 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                20.413369 |                     20.413369 |                         0.0 |                          0.0 |                      1.286025 |                           1.286025 |            15.840774 |                 15.840774 |    1001.159226 |    1.21051 | -2.9 | 20.443786 |  15.862716 |            204.455581 |                      0.0 |                204.455581 |                 204.455581 |                 35.980407 |                     0.0 |                      0.0 |   168.475175 |             0.0 |              0.0 |        168.475175 |             50.440581 |                 100.0 |                103.85044 |               96.853188 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.847296 | 0.094 | 3.314337 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.803077 | 0.020869 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 361.481911 | 73.979565 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.075077 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020869 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |                15.438712 |      287.848667 |             74.9 |       3.0 |              1018.0 | 0.02 | 1.728 |    18.5 |                  74.9 |                  13.9 |                     15.438712 |           287.848667 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                21.331125 |                     21.331125 |                         0.0 |                          0.0 |                      1.336407 |                           1.336407 |            15.977013 |                 15.977013 |    1002.022987 |   1.208738 | -2.9 | 15.040124 | -20.759257 |            218.764987 |                      0.0 |                218.764987 |                 218.764987 |                 39.523937 |                     0.0 |                      0.0 |   179.241051 |             0.0 |              0.0 |        179.241051 |             47.077876 |                 100.0 |               108.495906 |                98.61702 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 6.638473 |   0.0 |  3.73228 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.768062 | 0.020464 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 379.341169 | 70.247285 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.040062 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020464 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |                15.374034 |      274.674263 |             77.9 |       3.5 |              1018.0 | 0.02 | 1.728 |    19.9 |                  77.9 |                  13.0 |                     15.374034 |           274.674263 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                23.275348 |                     23.275348 |                         0.0 |                          0.0 |                      1.442189 |                           1.442189 |            18.131496 |                 18.131496 |     999.868504 |   1.201995 | -2.9 |  20.18359 |   1.701539 |             208.75244 |                      0.0 |                 208.75244 |                  208.75244 |                 32.936813 |                     0.0 |                      0.0 |   175.815627 |             0.0 |              0.0 |        175.815627 |             40.352465 |                 100.0 |               113.827642 |              100.519454 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 6.828968 |   0.0 | 3.652387 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.749377 | 0.020247 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 374.739629 | 66.594898 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.021377 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020247 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |                 15.30875 |      284.047892 |             73.5 |       3.8 |              1019.0 | 0.02 | 1.728 |    20.8 |                  73.5 |                  13.8 |                      15.30875 |           284.047892 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                24.605348 |                     24.605348 |                         0.0 |                          0.0 |                      1.513854 |                           1.513854 |            18.084931 |                 18.084931 |    1000.915069 |   1.199521 | -2.9 | 18.858347 | -11.649921 |            215.876398 |                      0.0 |                215.876398 |                 215.876398 |                  35.31351 |                     0.0 |                      0.0 |   180.562888 |             0.0 |              0.0 |        180.562888 |             37.166744 |                 100.0 |               119.167697 |              102.297207 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 7.945133 |   0.0 | 4.128687 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.739407 | 0.020132 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  383.48955 |  62.46621 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.011407 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020132 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |                15.242885 |      262.971263 |             80.1 |       3.3 |              1019.0 | 0.02 | 1.728 |    20.9 |                  80.1 |                  12.3 |                     15.242885 |           262.971263 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 24.75714 |                      24.75714 |                         0.0 |                          0.0 |                      1.521999 |                           1.521999 |            19.830469 |                 19.830469 |     999.169531 |   1.198331 | -2.9 | 21.378324 |   2.869942 |             199.85816 |                      0.0 |                 199.85816 |                  199.85816 |                 28.392598 |                     0.0 |                      0.0 |   171.465562 |             0.0 |              0.0 |        171.465562 |             42.798069 |                 100.0 |               125.381825 |              104.230888 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 6.427975 |   0.0 | 3.518399 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.734087 |  0.02007 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 377.719608 | 58.947811 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.006087 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 |  0.02007 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |                15.176467 |      272.256414 |             72.9 |       1.8 |              1019.0 | 0.02 | 1.728 |    22.3 |                  72.9 |                  13.1 |                     15.176467 |           272.256414 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 26.96965 |                      26.96965 |                         0.0 |                          0.0 |                      1.639967 |                           1.639967 |            19.660875 |                 19.660875 |     999.339125 |   1.192729 | -2.9 |  19.71658 | -15.500518 |            206.914874 |                      0.0 |                206.914874 |                 206.914874 |                 31.114889 |                     0.0 |                      0.0 |   175.799986 |             0.0 |              0.0 |        175.799986 |             78.463126 |                 100.0 |               130.860307 |              105.804393 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 6.036589 |   0.0 | 4.091259 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.731248 | 0.020038 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 390.320126 | 54.856552 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.003248 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020038 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |                15.109521 |      267.019218 |             68.5 |       1.9 |              1020.0 | 0.02 | 1.728 |    23.0 |                  68.5 |                  12.8 |                     15.109521 |           267.019218 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                28.139149 |                     28.139149 |                         0.0 |                          0.0 |                      1.701779 |                           1.701779 |            19.275317 |                 19.275317 |    1000.724683 |   1.191257 | -2.9 |  23.34553 |   2.073177 |            202.934606 |                      0.0 |                202.934606 |                 202.934606 |                 32.010643 |                     0.0 |                      0.0 |   170.923963 |             0.0 |              0.0 |        170.923963 |             74.333488 |                 100.0 |               137.497212 |               107.59145 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 6.493896 |   0.0 | 4.345753 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.729733 |  0.02002 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 385.346949 | 50.510799 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.001733 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 |  0.02002 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |                15.042071 |      269.626547 |             68.6 |       3.0 |              1021.0 | 0.02 | 1.728 |    23.1 |                  68.6 |                  13.1 |                     15.042071 |           269.626547 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                28.309776 |                     28.309776 |                         0.0 |                          0.0 |                      1.710768 |                           1.710768 |            19.420506 |                 19.420506 |    1001.579494 |   1.191966 | -2.9 | 21.913255 |  -7.120473 |            204.916175 |                      0.0 |                204.916175 |                 204.916175 |                 32.400014 |                     0.0 |                      0.0 |   172.516162 |             0.0 |              0.0 |        172.516162 |             47.077876 |                 100.0 |               144.951913 |              109.449121 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 7.914607 |   0.0 | 4.547949 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.728925 | 0.020011 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 389.567421 | 45.962849 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000925 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020011 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |                14.974141 |      270.878291 |             66.0 |       2.5 |              1022.0 | 0.02 | 1.728 |    23.2 |                  66.0 |                  13.3 |                     14.974141 |           270.878291 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                28.481304 |                     28.481304 |                         0.0 |                          0.0 |                      1.719796 |                           1.719796 |             18.79766 |                  18.79766 |     1003.20234 |   1.193016 | -2.9 | 23.128751 |  -0.427496 |            205.867501 |                      0.0 |                205.867501 |                 205.867501 |                 34.565783 |                     0.0 |                      0.0 |   171.301718 |             0.0 |              0.0 |        171.301718 |             56.493451 |                 100.0 |               153.358051 |              111.375228 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 7.566564 | 0.094 | 4.554398 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.728494 | 0.020006 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 387.094917 | 41.408451 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000494 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020006 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |                14.905755 |      261.697164 |             69.0 |       1.6 |              1019.0 | 0.02 | 1.728 |    23.2 |                  69.0 |                  12.7 |                     14.905755 |           261.697164 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                28.481304 |                     28.481304 |                         0.0 |                          0.0 |                      1.719796 |                           1.719796 |              19.6521 |                   19.6521 |       999.3479 |    1.18911 | -2.9 | 22.416669 |  -4.699983 |            198.889844 |                      0.0 |                198.889844 |                 198.889844 |                  31.34697 |                     0.0 |                      0.0 |   167.542875 |             0.0 |              0.0 |        167.542875 |             88.271017 |                 100.0 |                162.65311 |               113.31833 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.996216 |   0.0 | 4.137046 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.728263 | 0.020003 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |   388.8949 | 37.271405 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000263 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020003 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |                14.836934 |      227.962116 |             75.6 |       2.7 |              1015.0 | 0.02 | 1.728 |    21.7 |                  75.6 |                  10.2 |                     14.836934 |           227.962116 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 26.00118 |                      26.00118 |                         0.0 |                          0.0 |                        1.5885 |                             1.5885 |            19.656892 |                 19.656892 |     995.343108 |   1.190432 | -2.9 | 22.935065 |   7.410388 |            173.251208 |                      0.0 |                173.251208 |                 173.251208 |                 25.918437 |                     0.0 |                      0.0 |   147.332771 |             0.0 |              0.0 |        147.332771 |             52.308751 |                 100.0 |                172.21097 |              115.125428 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.983775 |   0.0 | 3.428735 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.728141 | 0.020002 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 378.584513 |  33.84267 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000141 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 | 0.020002 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |                  14.7677 |      216.356321 |             81.4 |       3.2 |              1016.0 | 0.02 | 1.728 |    20.8 |                  81.4 |                   9.4 |                       14.7677 |           216.356321 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                24.605348 |                     24.605348 |                         0.0 |                          0.0 |                      1.513854 |                           1.513854 |            20.028753 |                 20.028753 |     995.971247 |   1.195095 | -2.9 | 19.965673 |  -5.005962 |            164.430804 |                      0.0 |                164.430804 |                 164.430804 |                 23.397902 |                     0.0 |                      0.0 |   141.032902 |             0.0 |              0.0 |        141.032902 |             44.135508 |                 100.0 |               181.322396 |              116.674937 |  1.838 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.470835 |   0.7 | 2.505678 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  1.838 | 0.049893 |  0.0 |  0.0 | 0.0 |      0.0 | 0.049893 |  1.74092 |  0.02015 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 380.690475 | 33.125099 | 0.049893 |   0.0 |   0.0 |   0.0 |       0.0 |  0.01292 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 |  0.02015 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |                14.698075 |      226.637904 |             79.0 |       1.3 |              1019.0 | 0.02 | 1.728 |    19.5 |                  79.0 |                  10.3 |                     14.698075 |           226.637904 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                22.704676 |                     22.704676 |                         0.0 |                          0.0 |                      1.411269 |                           1.411269 |            17.936694 |                 17.936694 |    1001.063306 |   1.204916 | -2.9 |  20.57219 |   6.433141 |            172.244807 |                      0.0 |                172.244807 |                 172.244807 |                 28.794151 |                     0.0 |                      0.0 |   143.450656 |             0.0 |              0.0 |        143.450656 |            108.641252 |                 100.0 |               183.411459 |               116.86708 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 4.081206 |   0.0 | 2.836506 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.745319 |   0.0202 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 371.357334 | 30.288593 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.017319 |   0.0 |   0.0 |   0.0 |       0.0 |  0.02 |   0.0202 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |                14.628078 |      235.528503 |             75.9 |       1.9 |              1019.0 | 0.02 | 1.728 |    21.0 |                  75.9 |                  11.1 |                     14.628078 |           235.528503 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                24.909749 |                     24.909749 |                         0.0 |                          0.0 |                      1.530181 |                           1.530181 |            18.906499 |                 18.906499 |    1000.093501 |   1.198338 | -2.9 | 17.884246 | -18.694526 |            179.001662 |                      0.0 |                179.001662 |                 179.001662 |                 29.177702 |                     0.0 |                      0.0 |    149.82396 |             0.0 |              0.0 |         149.82396 |             74.333488 |                 100.0 |               192.482064 |              118.222742 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.175279 |   0.0 | 3.276924 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.03799 |      0.0 | 1.737173 | 0.020106 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 387.151861 | 27.049659 |      0.0 |   0.0 |   0.0 |   0.0 |  -0.03799 | 0.009242 |   0.0 |   0.0 |   0.0 | -0.000068 |  0.02 | 0.020106 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |                14.557728 |      232.849428 |             70.0 |       2.0 |              1018.0 | 0.02 | 1.728 |    21.6 |                  70.0 |                  11.0 |                     14.557728 |           232.849428 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                25.842752 |                     25.842752 |                         0.0 |                          0.0 |                      1.580055 |                           1.580055 |            18.089927 |                 18.089927 |     999.910073 |   1.195081 | -2.9 | 22.433069 |   4.998415 |            176.965566 |                      0.0 |                176.965566 |                 176.965566 |                 31.281288 |                     0.0 |                      0.0 |   145.684278 |             0.0 |              0.0 |        145.684278 |             70.616813 |                 100.0 |               204.991875 |              119.966911 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.737858 |   0.0 | 3.581415 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.239244 |      0.0 | 1.732367 | 0.020051 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 379.253445 | 23.707488 |      0.0 |   0.0 |   0.0 |   0.0 | -0.239244 | 0.004931 |   0.0 |   0.0 |   0.0 | -0.000565 |  0.02 | 0.020051 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |                14.487046 |       199.71349 |             70.3 |       2.6 |              1019.0 | 0.02 | 1.728 |    21.1 |                  70.3 |                   8.5 |                     14.487046 |            199.71349 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                25.063177 |                     25.063177 |                         0.0 |                          0.0 |                        1.5384 |                             1.5384 |            17.619414 |                 17.619414 |    1001.380586 |   1.198506 | -2.9 | 20.158326 |  -5.650046 |            151.782252 |                      0.0 |                151.782252 |                 151.782252 |                 26.711296 |                     0.0 |                      0.0 |   125.070956 |             0.0 |              0.0 |        125.070956 |             54.320626 |                 100.0 |               221.727637 |              122.069797 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.866935 |   0.0 | 3.296835 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.446913 |      0.0 | 1.728839 |  0.02001 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 382.003492 | 20.857567 |      0.0 |   0.0 |   0.0 |   0.0 | -0.446913 | 0.002631 |   0.0 |   0.0 |   0.0 | -0.001792 |  0.02 |  0.02001 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |                14.416048 |      233.771902 |             69.0 |       3.6 |              1019.0 | 0.02 | 1.728 |    22.4 |                  69.0 |                  11.3 |                     14.416048 |           233.771902 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                27.134082 |                     27.134082 |                         0.0 |                          0.0 |                      1.648679 |                           1.648679 |            18.722517 |                 18.722517 |    1000.277483 |   1.192743 | -2.9 | 20.950339 |  -8.697966 |            177.666646 |                      0.0 |                177.666646 |                 177.666646 |                 31.116622 |                     0.0 |                      0.0 |   146.550024 |             0.0 |              0.0 |        146.550024 |             39.231563 |                 100.0 |               241.230014 |              124.199013 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  7.82923 |   0.0 | 3.891994 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.623996 |      0.0 |   1.7257 | 0.019973 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 387.801458 | 17.589568 |      0.0 |   0.0 |   0.0 |   0.0 | -0.623996 | 0.001404 |   0.0 |   0.0 |   0.0 | -0.003704 |  0.02 | 0.019973 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |                14.344752 |      246.121286 |             62.1 |       2.5 |              1017.0 | 0.02 | 1.728 |    22.5 |                  62.1 |                  12.4 |                     14.344752 |           246.121286 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                27.299387 |                     27.299387 |                         0.0 |                          0.0 |                      1.657431 |                           1.657431 |             16.95292 |                  16.95292 |     1000.04708 |   1.190771 | -2.9 | 22.620153 |    0.72092 |            187.052177 |                      0.0 |                187.052177 |                 187.052177 |                 38.297604 |                     0.0 |                      0.0 |   148.754573 |             0.0 |              0.0 |        148.754573 |             56.493451 |                 100.0 |               274.853414 |              127.355551 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 7.322689 |   0.0 | 4.183176 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.827056 |      0.0 | 1.722459 | 0.019936 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 384.180539 | 14.233449 |      0.0 |   0.0 |   0.0 |   0.0 | -0.827056 | 0.000749 |   0.0 |   0.0 |   0.0 |  -0.00629 |  0.02 | 0.019936 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |                14.273175 |        90.58551 |             84.5 |       2.1 |              1015.0 | 0.02 | 1.728 |    19.6 |                  84.5 |                   0.1 |                     14.273175 |             90.58551 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                22.846186 |                     22.846186 |                         0.0 |                          0.0 |                      1.418946 |                           1.418946 |            19.305027 |                 19.305027 |     995.694973 |   1.199129 | -2.9 | 21.577328 |  11.863971 |             68.844988 |                      0.0 |                 68.844988 |                  68.844988 |                   7.02118 |                     0.0 |                      0.0 |    61.823807 |             0.0 |              0.0 |         61.823807 |             67.254108 |                 100.0 |               337.448215 |              131.864309 |   1.18 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.567328 |   0.7 | 1.075698 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |   1.18 | 0.013305 |  0.0 |  0.0 | 0.0 | 1.035592 | 0.013305 | 1.722221 | 0.019933 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 369.416568 | 15.360037 | 0.013305 |   0.0 |   0.0 |   0.0 | -1.035592 | 0.003825 |   0.0 |   0.0 |   0.0 | -0.009605 |  0.02 | 0.019933 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |                14.201333 |      171.944993 |             83.6 |       1.5 |              1015.0 | 0.02 | 1.728 |    20.3 |                  83.6 |                   6.7 |                     14.201333 |           171.944993 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                23.858503 |                     23.858503 |                         0.0 |                          0.0 |                      1.473678 |                           1.473678 |            19.945708 |                 19.945708 |     995.054292 |   1.195981 | -2.9 | 17.325305 | -17.848171 |            130.678195 |                      0.0 |                130.678195 |                 130.678195 |                 18.994889 |                     0.0 |                      0.0 |   111.683306 |             0.0 |              0.0 |        111.683306 |             94.155751 |                 100.0 |                311.33348 |              129.741425 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.374147 | 0.282 | 2.014865 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.96559 |      0.0 | 1.719666 | 0.019904 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 384.364739 | 14.310761 |      0.0 |   0.0 |   0.0 |   0.0 |  -0.96559 | 0.004821 |   0.0 |   0.0 |   0.0 | -0.013155 |  0.02 | 0.019904 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |                14.129243 |      216.549813 |             76.5 |       2.2 |              1011.0 | 0.02 | 1.728 |    24.1 |                  76.5 |                  10.4 |                     14.129243 |           216.549813 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                30.066256 |                     30.066256 |                         0.0 |                          0.0 |                       1.80287 |                            1.80287 |            23.000686 |                 23.000686 |     987.999314 |   1.174651 | -2.9 | 21.630378 | -14.817732 |            164.577858 |                      0.0 |                164.577858 |                 164.577858 |                 21.629737 |                     0.0 |                      0.0 |   142.948121 |             0.0 |              0.0 |        142.948121 |             64.197103 |                 100.0 |                335.42542 |              131.069494 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.573024 |   0.0 | 3.362119 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 1.030788 |      0.0 | 1.713889 | 0.019837 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  396.28247 |  11.97943 |      0.0 |   0.0 |   0.0 |   0.0 | -1.030788 | 0.002573 |   0.0 |   0.0 |   0.0 | -0.016684 |  0.02 | 0.019837 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |                14.056918 |       166.91683 |             76.6 |       1.9 |              1008.0 | 0.02 | 1.728 |    23.7 |                  76.6 |                   6.5 |                     14.056918 |            166.91683 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                29.352596 |                     29.352596 |                         0.0 |                          0.0 |                      1.765541 |                           1.765541 |            22.484088 |                 22.484088 |     985.515912 |   1.172942 | -2.9 | 25.062685 |   8.176109 |             126.85679 |                      0.0 |                 126.85679 |                  126.85679 |                 16.222254 |                     0.0 |                      0.0 |   110.634536 |             0.0 |              0.0 |        110.634536 |             74.333488 |                 100.0 |               419.576726 |               135.15103 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 4.467668 | 0.094 | 2.719352 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 1.175647 |      0.0 | 1.708796 | 0.019778 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 385.206362 | 10.435725 |      0.0 |   0.0 |   0.0 |   0.0 | -1.175647 | 0.001373 |   0.0 |   0.0 |   0.0 | -0.020577 |  0.02 | 0.019778 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |                13.984374 |      145.985611 |             84.2 |       2.3 |              1006.0 | 0.02 | 1.728 |    21.1 |                  84.2 |                   4.9 |                     13.984374 |           145.985611 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                25.063177 |                     25.063177 |                         0.0 |                          0.0 |                        1.5384 |                             1.5384 |            21.103195 |                 21.103195 |     984.896805 |   1.181557 | -2.9 | 21.872765 |   4.636593 |            110.949064 |                      0.0 |                110.949064 |                 110.949064 |                 14.602883 |                     0.0 |                      0.0 |    96.346181 |             0.0 |              0.0 |         96.346181 |             61.405925 |                 100.0 |               522.087742 |              138.401561 |   9.17 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.569826 |   0.7 | 1.597373 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |   9.17 | 0.104578 |  0.0 |  0.0 | 0.0 | 1.271567 | 0.104578 | 1.730769 | 0.020032 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 377.669769 | 19.175341 | 0.104578 |   0.0 |   0.0 |   0.0 | -1.271567 | 0.027656 |   0.0 |   0.0 |   0.0 | -0.024887 |  0.02 | 0.020032 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |                13.911624 |       165.54389 |             85.8 |       2.1 |              1005.0 | 0.02 | 1.728 |    20.9 |                  85.8 |                   6.6 |                     13.911624 |            165.54389 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 24.75714 |                      24.75714 |                         0.0 |                          0.0 |                      1.521999 |                           1.521999 |            21.241626 |                 21.241626 |     983.758374 |   1.181114 | -2.9 | 19.702227 |   -7.18664 |            125.813357 |                      0.0 |                125.813357 |                 125.813357 |                 17.349121 |                     0.0 |                      0.0 |   108.464235 |             0.0 |              0.0 |        108.464235 |             67.254108 |                 100.0 |               256.482293 |              123.963828 |   5.41 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  3.56693 |   0.7 | 1.722258 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |   5.41 | 0.090745 |  0.0 |  0.0 | 0.0 | 0.728523 | 0.090745 |  1.75959 | 0.020366 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 381.956408 | 23.500861 | 0.090745 |   0.0 |   0.0 |   0.0 | -0.728523 |  0.05997 |   0.0 |   0.0 |   0.0 |  -0.02838 |  0.02 | 0.020366 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |                13.838682 |       87.693346 |             86.5 |       2.6 |              1000.0 | 0.02 | 1.728 |    20.2 |                  86.5 |                   0.3 |                     13.838682 |            87.693346 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                23.711531 |                     23.711531 |                         0.0 |                          0.0 |                      1.465752 |                           1.465752 |            20.510475 |                 20.510475 |     979.489525 |   1.178323 | -2.9 | 20.936779 |   4.420674 |             66.646943 |                      0.0 |                 66.646943 |                  66.646943 |                  6.818024 |                     0.0 |                      0.0 |    59.828919 |             0.0 |              0.0 |         59.828919 |             54.320626 |                 100.0 |                222.94421 |               120.11308 |   13.4 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.621228 |   0.7 | 1.071953 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |   13.4 | 0.312945 |  0.0 |  0.0 | 0.0 | 0.459752 | 0.312945 | 1.829121 |  0.02117 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 374.635735 | 35.975715 | 0.312945 |   0.0 |   0.0 |   0.0 | -0.459752 | 0.131528 |   0.0 |   0.0 |   0.0 | -0.030407 |  0.02 |  0.02117 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |                13.765561 |       84.356885 |             89.6 |       2.6 |              1004.0 | 0.02 | 1.728 |    16.1 |                  89.6 |                   0.1 |                     13.765561 |            84.356885 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                18.326791 |                     18.326791 |                         0.0 |                          0.0 |                      1.170311 |                           1.170311 |            16.420805 |                 16.420805 |     987.579195 |   1.201704 | -2.9 | 18.828425 |   16.37055 |             64.111233 |                      0.0 |                 64.111233 |                  64.111233 |                  8.199212 |                     0.0 |                      0.0 |    55.912021 |             0.0 |              0.0 |         55.912021 |             54.320626 |                 100.0 |               175.502292 |              113.152806 |  7.948 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.027587 |   0.7 | 0.717164 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  7.948 | 0.249164 |  0.0 |  0.0 | 0.0 |      0.0 | 0.249164 | 1.896598 | 0.021951 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 355.365185 | 42.957387 | 0.249164 |   0.0 |   0.0 |   0.0 |       0.0 | 0.199719 |   0.0 |   0.0 |   0.0 | -0.031121 |  0.02 | 0.021951 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |                13.692273 |      142.309202 |             78.3 |       2.1 |              1011.0 | 0.02 | 1.728 |    17.0 |                  78.3 |                   5.0 |                     13.692273 |           142.309202 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                19.406929 |                     19.406929 |                         0.0 |                          0.0 |                      1.230421 |                           1.230421 |            15.195625 |                 15.195625 |     995.804375 |   1.206937 | -2.9 | 13.278507 |  -22.32896 |            108.154994 |                      0.0 |                108.154994 |                 108.154994 |                 21.775852 |                     0.0 |                      0.0 |    86.379141 |             0.0 |              0.0 |         86.379141 |             67.254108 |                 100.0 |               159.371564 |              110.061251 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.393474 |   0.0 | 2.035891 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 | 1.855624 | 0.021477 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 374.794145 | 40.921496 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.158633 |   0.0 |   0.0 |   0.0 | -0.031009 |  0.02 | 0.021477 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |                13.618831 |      126.722078 |             78.5 |       2.1 |              1014.0 | 0.02 | 1.728 |    18.2 |                  78.5 |                   3.8 |                     13.618831 |           126.722078 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                20.933479 |                     20.933479 |                         0.0 |                          0.0 |                      1.314615 |                           1.314615 |            16.432781 |                 16.432781 |     997.567219 |   1.204994 | -2.9 | 18.874047 |   4.044283 |              96.30878 |                      0.0 |                  96.30878 |                   96.30878 |                 17.491881 |                     0.0 |                      0.0 |    78.816899 |             0.0 |              0.0 |         78.816899 |             67.254108 |                 100.0 |               163.714409 |               110.73073 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.321963 |   0.0 | 2.018218 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 |  1.78175 | 0.020622 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 367.849862 | 38.903278 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.084648 |   0.0 |   0.0 |   0.0 | -0.030898 |  0.02 | 0.020622 |

>>> round_(model.check_waterbalance(conditions))
0.0

.. _lland_v4_acker_heavy_rain_daily:

acre (heavy rain)
-----------------

.. integration-test::

    >>> lnk(ACKER)
    >>> inlet.sequences.sim.series = 0.0
    >>> nied = inputs.nied.series.copy()
    >>> inputs.nied.series = 20.0
    >>> conditions_acker_heavy_rain = test(
    ...     "lland_v4_acker_heavy_rain_daily",
    ...     axis1=(inputs.nied, fluxes.qah), axis2=states.bowa,
    ...     get_conditions="1997-08-03")
    |       date | nied | teml | sunshineduration | possiblesunshineduration | globalradiation | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration | dailypossiblesunshineduration | dailyglobalradiation | nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair |    g |        tz |         wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |     evpo | evi |      evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatinz | wlatsnow | wsensinz | wsenssnow | wsurfinz | wsurf | sff | fvg | wada |       qdb |     qib1 |     qib2 |      qbb | qkap |      qdgz |       qah |       qa | inzp | stinz | sinz | esnowinz | asinz | wats | waes | esnow | taus |       ebdn |       bowa |     qdgz1 |    qdgz2 |    qigz1 |    qigz2 |     qbgz |     qdga1 |    qdga2 |    qiga1 |    qiga2 |     qbga | inlet |   outlet |
    ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 20.0 | 16.1 |              6.3 |                15.690785 |       190.25149 |             86.2 |       2.4 |              1007.0 | 0.0 | 0.0 |    16.1 |                  86.2 |                   6.3 |                     15.690785 |            190.25149 | 18.8 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                18.326791 |                     18.326791 |                         0.0 |                          0.0 |                      1.170311 |                           1.170311 |            15.797694 |                 15.797694 |     991.202306 |   1.205601 | -2.9 | 11.733333 |      -26.2 |            144.591133 |                      0.0 |                144.591133 |                 144.591133 |                 21.732733 |                     0.0 |                      0.0 |     122.8584 |             0.0 |              0.0 |          122.8584 |             58.847345 |                 100.0 |               111.309423 |               99.889466 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.623832 | 0.7 | 1.720128 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  1.217805 |      0.0 |      0.0 |      0.0 |  0.0 |  1.217805 |   0.31352 | 0.003629 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |      373.3 |  87.162066 |  1.217805 |      0.0 |      0.0 |      0.0 |      0.0 |   0.31352 |      0.0 |      0.0 |      0.0 |      0.0 |   0.0 | 0.003629 |
    | 1997-02-08 | 20.0 | 17.3 |              1.7 |                15.628821 |      127.660782 |             85.8 |       1.6 |              1013.0 | 0.0 | 0.0 |    17.3 |                  85.8 |                   1.7 |                     15.628821 |           127.660782 | 18.8 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                19.779139 |                     19.779139 |                         0.0 |                          0.0 |                      1.251029 |                           1.251029 |            16.970501 |                 16.970501 |     996.029499 |   1.207284 | -2.9 | 18.443733 |     6.8624 |             97.022194 |                      0.0 |                 97.022194 |                  97.022194 |                 11.219797 |                     0.0 |                      0.0 |    85.802398 |             0.0 |              0.0 |         85.802398 |             88.271017 |                 100.0 |                90.407062 |               91.066356 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.555581 | 0.7 | 1.294457 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |   1.48087 |      0.0 |      0.0 |      0.0 |  0.0 |   1.48087 |  0.802994 | 0.009294 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |   363.5376 | 102.486739 |   1.48087 |      0.0 |      0.0 |      0.0 |      0.0 |  0.802994 |      0.0 |      0.0 |      0.0 |      0.0 |   0.0 | 0.009294 |
    | 1997-03-08 | 20.0 | 18.9 |              4.5 |                15.566134 |      164.401298 |             82.4 |       1.5 |              1016.0 | 0.0 | 0.0 |    18.9 |                  82.4 |                   4.5 |                     15.566134 |           164.401298 | 18.8 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                21.871576 |                     21.871576 |                         0.0 |                          0.0 |                      1.365939 |                           1.365939 |            18.022179 |                 18.022179 |     997.977821 |   1.203775 | -2.9 | 15.632162 | -19.607027 |            124.944986 |                      0.0 |                124.944986 |                 124.944986 |                 16.044181 |                     0.0 |                      0.0 |   108.900806 |             0.0 |              0.0 |        108.900806 |             94.155751 |                 100.0 |                70.326844 |               80.364888 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.286385 | 0.7 | 1.922357 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  1.761525 |      0.0 |      0.0 |      0.0 |  0.0 |  1.761525 |  1.191401 | 0.013789 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 380.244627 | 116.902858 |  1.761525 |      0.0 |      0.0 |      0.0 |      0.0 |  1.191401 |      0.0 |      0.0 |      0.0 |      0.0 |   0.0 | 0.013789 |
    | 1997-04-08 | 20.0 | 17.8 |             12.4 |                15.502754 |      269.020502 |             77.6 |       2.8 |              1017.0 | 0.0 | 0.0 |    17.8 |                  77.6 |                  12.4 |                     15.502754 |           269.020502 | 18.8 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                20.413369 |                     20.413369 |                         0.0 |                          0.0 |                      1.286025 |                           1.286025 |            15.840774 |                 15.840774 |    1001.159226 |    1.21051 | -2.9 | 20.443786 |  15.862716 |            204.455581 |                      0.0 |                204.455581 |                 204.455581 |                 35.980407 |                     0.0 |                      0.0 |   168.475175 |             0.0 |              0.0 |        168.475175 |             50.440581 |                 100.0 |                51.993251 |               67.804391 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.847296 | 0.7 | 3.397171 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  2.040918 |      0.0 |      0.0 |      0.0 |  0.0 |  2.040918 |   1.52923 | 0.017699 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 361.481911 | 129.564768 |  2.040918 |      0.0 |      0.0 |      0.0 |      0.0 |   1.52923 |      0.0 |      0.0 |      0.0 |      0.0 |   0.0 | 0.017699 |
    | 1997-05-08 | 20.0 | 18.5 |             13.9 |                15.438712 |      287.848667 |             74.9 |       3.0 |              1018.0 | 0.0 | 0.0 |    18.5 |                  74.9 |                  13.9 |                     15.438712 |           287.848667 | 18.8 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                21.331125 |                     21.331125 |                         0.0 |                          0.0 |                      1.336407 |                           1.336407 |            15.977013 |                 15.977013 |    1002.022987 |   1.208738 | -2.9 | 15.040124 | -20.759257 |            218.764987 |                      0.0 |                218.764987 |                 218.764987 |                 39.523937 |                     0.0 |                      0.0 |   179.241051 |             0.0 |              0.0 |        179.241051 |             47.077876 |                 100.0 |                 48.85611 |               65.355461 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 6.638473 | 0.7 | 3.917105 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  2.300286 | 0.101438 |      0.0 | 0.010165 |  0.0 |  2.300286 |  1.849051 | 0.021401 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 379.341169 | 141.335775 |  2.300286 |      0.0 | 0.101438 |      0.0 | 0.010165 |  1.834649 |      0.0 | 0.014384 |      0.0 | 0.000018 |   0.0 | 0.021401 |
    | 1997-06-08 | 20.0 | 19.9 |             13.0 |                15.374034 |      274.674263 |             77.9 |       3.5 |              1018.0 | 0.0 | 0.0 |    19.9 |                  77.9 |                  13.0 |                     15.374034 |           274.674263 | 18.8 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                23.275348 |                     23.275348 |                         0.0 |                          0.0 |                      1.442189 |                           1.442189 |            18.131496 |                 18.131496 |     999.868504 |   1.201995 | -2.9 |  20.18359 |   1.701539 |             208.75244 |                      0.0 |                 208.75244 |                  208.75244 |                 32.936813 |                     0.0 |                      0.0 |   175.815627 |             0.0 |              0.0 |        175.815627 |             40.352465 |                 100.0 |                 48.85611 |               65.425338 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 6.828968 | 0.7 | 3.913569 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  2.554694 | 0.110654 |      0.0 | 0.021936 |  0.0 |  2.554694 |  2.156542 |  0.02496 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 374.739629 | 152.834923 |  2.554694 |      0.0 | 0.110654 |      0.0 | 0.021936 |  2.117313 |      0.0 | 0.039153 |      0.0 | 0.000076 |   0.0 |  0.02496 |
    | 1997-07-08 | 20.0 | 20.8 |             13.8 |                 15.30875 |      284.047892 |             73.5 |       3.8 |              1019.0 | 0.0 | 0.0 |    20.8 |                  73.5 |                  13.8 |                      15.30875 |           284.047892 | 18.8 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                24.605348 |                     24.605348 |                         0.0 |                          0.0 |                      1.513854 |                           1.513854 |            18.084931 |                 18.084931 |    1000.915069 |   1.199521 | -2.9 | 18.858347 | -11.649921 |            215.876398 |                      0.0 |                215.876398 |                 215.876398 |                  35.31351 |                     0.0 |                      0.0 |   180.562888 |             0.0 |              0.0 |        180.562888 |             37.166744 |                 100.0 |                 48.85611 |               65.496024 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 7.945133 | 0.7 | 4.551444 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  2.817235 | 0.119656 |      0.0 | 0.033435 |  0.0 |  2.817235 |  2.448767 | 0.028342 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  383.48955 | 163.413153 |  2.817235 |      0.0 | 0.119656 |      0.0 | 0.033435 |  2.388893 |      0.0 |   0.0597 |      0.0 | 0.000175 |   0.0 | 0.028342 |
    | 1997-08-08 | 20.0 | 20.9 |             12.3 |                15.242885 |      262.971263 |             80.1 |       3.3 |              1019.0 | 0.0 | 0.0 |    20.9 |                  80.1 |                  12.3 |                     15.242885 |           262.971263 | 18.8 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 24.75714 |                      24.75714 |                         0.0 |                          0.0 |                      1.521999 |                           1.521999 |            19.830469 |                 19.830469 |     999.169531 |   1.198331 | -2.9 | 21.378324 |   2.869942 |             199.85816 |                      0.0 |                 199.85816 |                  199.85816 |                 28.392598 |                     0.0 |                      0.0 |   171.465562 |             0.0 |              0.0 |        171.465562 |             42.798069 |                 100.0 |                 48.85611 |               65.567492 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 6.427975 | 0.7 |  3.76789 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  3.072622 | 0.127938 |      0.0 | 0.044013 |  0.0 |  3.072622 |  2.731761 | 0.031618 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 377.719608 | 174.500689 |  3.072622 |      0.0 | 0.127938 |      0.0 | 0.044013 |  2.654416 |      0.0 | 0.077033 |      0.0 | 0.000313 |   0.0 | 0.031618 |
    | 1997-09-08 | 20.0 | 22.3 |             13.1 |                15.176467 |      272.256414 |             72.9 |       1.8 |              1019.0 | 0.0 | 0.0 |    22.3 |                  72.9 |                  13.1 |                     15.176467 |           272.256414 | 18.8 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 26.96965 |                      26.96965 |                         0.0 |                          0.0 |                      1.639967 |                           1.639967 |            19.660875 |                 19.660875 |     999.339125 |   1.192729 | -2.9 |  19.71658 | -15.500518 |            206.914874 |                      0.0 |                206.914874 |                 206.914874 |                 31.114889 |                     0.0 |                      0.0 |   175.799986 |             0.0 |              0.0 |        175.799986 |             78.463126 |                 100.0 |                 48.85611 |               65.639719 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 6.036589 | 0.7 | 4.120968 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  3.356609 | 0.136619 |      0.0 | 0.055101 |  0.0 |  3.356609 |  3.015047 | 0.034896 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 390.320126 | 184.931393 |  3.356609 |      0.0 | 0.136619 |      0.0 | 0.055101 |  2.922574 |      0.0 | 0.091984 |      0.0 | 0.000489 |   0.0 | 0.034896 |
    | 1997-10-08 | 20.0 | 23.0 |             12.8 |                15.109521 |      267.019218 |             68.5 |       1.9 |              1020.0 | 0.0 | 0.0 |    23.0 |                  68.5 |                  12.8 |                     15.109521 |           267.019218 | 18.8 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                28.139149 |                     28.139149 |                         0.0 |                          0.0 |                      1.701779 |                           1.701779 |            19.275317 |                 19.275317 |    1000.724683 |   1.191257 | -2.9 |  23.34553 |   2.073177 |            202.934606 |                      0.0 |                202.934606 |                 202.934606 |                 32.010643 |                     0.0 |                      0.0 |   170.923963 |             0.0 |              0.0 |        170.923963 |             74.333488 |                 100.0 |                 48.85611 |               65.712681 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 6.493896 | 0.7 | 4.450324 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  3.641281 | 0.144785 |      0.0 | 0.065531 |  0.0 |  3.641281 |  3.304166 | 0.038243 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 385.346949 | 194.729472 |  3.641281 |      0.0 | 0.144785 |      0.0 | 0.065531 |  3.198291 |      0.0 | 0.105171 |      0.0 | 0.000704 |   0.0 | 0.038243 |
    | 1997-11-08 | 20.0 | 23.1 |             13.1 |                15.042071 |      269.626547 |             68.6 |       3.0 |              1021.0 | 0.0 | 0.0 |    23.1 |                  68.6 |                  13.1 |                     15.042071 |           269.626547 | 18.8 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                28.309776 |                     28.309776 |                         0.0 |                          0.0 |                      1.710768 |                           1.710768 |            19.420506 |                 19.420506 |    1001.579494 |   1.191966 | -2.9 | 21.913255 |  -7.120473 |            204.916175 |                      0.0 |                204.916175 |                 204.916175 |                 32.400014 |                     0.0 |                      0.0 |   172.516162 |             0.0 |              0.0 |        172.516162 |             47.077876 |                 100.0 |                 48.85611 |               65.786357 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 7.914607 | 0.7 | 4.992996 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  3.926597 | 0.152456 |      0.0 | 0.075329 |  0.0 |  3.926597 |  3.596239 | 0.041623 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 389.567421 | 203.682094 |  3.926597 |      0.0 | 0.152456 |      0.0 | 0.075329 |   3.47835 |      0.0 | 0.116936 |      0.0 | 0.000953 |   0.0 | 0.041623 |
    | 1997-12-08 | 20.0 | 23.2 |             13.3 |                14.974141 |      270.878291 |             66.0 |       2.5 |              1022.0 | 0.0 | 0.0 |    23.2 |                  66.0 |                  13.3 |                     14.974141 |           270.878291 | 18.8 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                28.481304 |                     28.481304 |                         0.0 |                          0.0 |                      1.719796 |                           1.719796 |             18.79766 |                  18.79766 |     1003.20234 |   1.193016 | -2.9 | 23.128751 |  -0.427496 |            205.867501 |                      0.0 |                205.867501 |                 205.867501 |                 34.565783 |                     0.0 |                      0.0 |   171.301718 |             0.0 |              0.0 |        171.301718 |             56.493451 |                 100.0 |                 48.85611 |               65.860724 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 7.566564 | 0.7 | 4.979765 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  4.204896 | 0.159465 |  0.00956 | 0.084282 |  0.0 |  4.204896 |  3.889149 | 0.045013 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 387.094917 | 212.344126 |  4.204896 |      0.0 | 0.159465 |  0.00956 | 0.084282 |  3.759055 |      0.0 | 0.127503 | 0.001356 | 0.001236 |   0.0 | 0.045013 |
    | 1997-13-08 | 20.0 | 23.2 |             12.7 |                14.905755 |      261.697164 |             69.0 |       1.6 |              1019.0 | 0.0 | 0.0 |    23.2 |                  69.0 |                  12.7 |                     14.905755 |           261.697164 | 18.8 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                28.481304 |                     28.481304 |                         0.0 |                          0.0 |                      1.719796 |                           1.719796 |              19.6521 |                   19.6521 |       999.3479 |    1.18911 | -2.9 | 22.416669 |  -4.699983 |            198.889844 |                      0.0 |                198.889844 |                 198.889844 |                  31.34697 |                     0.0 |                      0.0 |   167.542875 |             0.0 |              0.0 |        167.542875 |             88.271017 |                 100.0 |                 48.85611 |               65.935761 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.996216 | 0.7 | 4.198392 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  4.492791 | 0.166247 | 0.045997 | 0.092944 |  0.0 |  4.492791 |  4.188469 | 0.048478 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |   388.8949 | 221.447755 |  4.492791 |      0.0 | 0.166247 | 0.045997 | 0.092944 |  4.041108 |      0.0 | 0.137079 | 0.008733 | 0.001549 |   0.0 | 0.048478 |
    | 1997-14-08 | 20.0 | 21.7 |             10.2 |                14.836934 |      227.962116 |             75.6 |       2.7 |              1015.0 | 0.0 | 0.0 |    21.7 |                  75.6 |                  10.2 |                     14.836934 |           227.962116 | 18.8 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 26.00118 |                      26.00118 |                         0.0 |                          0.0 |                        1.5885 |                             1.5885 |            19.656892 |                 19.656892 |     995.343108 |   1.190432 | -2.9 | 22.935065 |   7.410388 |            173.251208 |                      0.0 |                173.251208 |                 173.251208 |                 25.918437 |                     0.0 |                      0.0 |   147.332771 |             0.0 |              0.0 |        147.332771 |             52.308751 |                 100.0 |                 48.85611 |               66.011448 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.983775 | 0.7 | 3.701991 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  4.818593 | 0.173374 | 0.100359 | 0.102048 |  0.0 |  4.818593 |  4.509972 | 0.052199 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 378.584513 |  230.65139 |  4.818593 |      0.0 | 0.173374 | 0.100359 | 0.102048 |  4.335644 |      0.0 | 0.145951 | 0.026485 | 0.001893 |   0.0 | 0.052199 |
    | 1997-15-08 | 20.0 | 20.8 |              9.4 |                  14.7677 |      216.356321 |             81.4 |       3.2 |              1016.0 | 0.0 | 0.0 |    20.8 |                  81.4 |                   9.4 |                       14.7677 |           216.356321 | 18.8 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                24.605348 |                     24.605348 |                         0.0 |                          0.0 |                      1.513854 |                           1.513854 |            20.028753 |                 20.028753 |     995.971247 |   1.195095 | -2.9 | 19.965673 |  -5.005962 |            164.430804 |                      0.0 |                164.430804 |                 164.430804 |                 23.397902 |                     0.0 |                      0.0 |   141.032902 |             0.0 |              0.0 |        141.032902 |             44.135508 |                 100.0 |                 48.85611 |               66.087764 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.470835 | 0.7 | 3.155199 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  5.176976 |  0.18058 | 0.168029 | 0.111251 |  0.0 |  5.176976 |  4.865773 | 0.056317 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 380.690475 | 239.959354 |  5.176976 |      0.0 |  0.18058 | 0.168029 | 0.111251 |  4.653151 |      0.0 | 0.154363 |  0.05599 | 0.002268 |   0.0 | 0.056317 |
    | 1997-16-08 | 20.0 | 19.5 |             10.3 |                14.698075 |      226.637904 |             79.0 |       1.3 |              1019.0 | 0.0 | 0.0 |    19.5 |                  79.0 |                  10.3 |                     14.698075 |           226.637904 | 18.8 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                22.704676 |                     22.704676 |                         0.0 |                          0.0 |                      1.411269 |                           1.411269 |            17.936694 |                 17.936694 |    1001.063306 |   1.204916 | -2.9 |  20.57219 |   6.433141 |            172.244807 |                      0.0 |                172.244807 |                 172.244807 |                 28.794151 |                     0.0 |                      0.0 |   143.450656 |             0.0 |              0.0 |        143.450656 |            108.641252 |                 100.0 |                 48.85611 |                66.16469 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 4.081206 | 0.7 | 2.708353 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  5.575883 | 0.187867 | 0.247357 | 0.120559 |  0.0 |  5.575883 |  5.262727 | 0.060911 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 371.357334 | 249.219334 |  5.575883 |      0.0 | 0.187867 | 0.247357 | 0.120559 |  5.000154 |      0.0 | 0.162462 | 0.097435 | 0.002675 |   0.0 | 0.060911 |
    | 1997-17-08 | 20.0 | 21.0 |             11.1 |                14.628078 |      235.528503 |             75.9 |       1.9 |              1019.0 | 0.0 | 0.0 |    21.0 |                  75.9 |                  11.1 |                     14.628078 |           235.528503 | 18.8 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                24.909749 |                     24.909749 |                         0.0 |                          0.0 |                      1.530181 |                           1.530181 |            18.906499 |                 18.906499 |    1000.093501 |   1.198338 | -2.9 | 17.884246 | -18.694526 |            179.001662 |                      0.0 |                179.001662 |                 179.001662 |                 29.177702 |                     0.0 |                      0.0 |    149.82396 |             0.0 |              0.0 |         149.82396 |             74.333488 |                 100.0 |                 48.85611 |               66.242207 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.175279 | 0.7 | 3.378596 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  6.018762 | 0.195117 | 0.335816 | 0.129819 |  0.0 |  6.018762 |  5.706521 | 0.066048 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 387.151861 | 257.261224 |  6.018762 |      0.0 | 0.195117 | 0.335816 | 0.129819 |  5.382685 |      0.0 | 0.170337 | 0.150384 | 0.003114 |   0.0 | 0.066048 |
    | 1997-18-08 | 20.0 | 21.6 |             11.0 |                14.557728 |      232.849428 |             70.0 |       2.0 |              1018.0 | 0.0 | 0.0 |    21.6 |                  70.0 |                  11.0 |                     14.557728 |           232.849428 | 18.8 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                25.842752 |                     25.842752 |                         0.0 |                          0.0 |                      1.580055 |                           1.580055 |            18.089927 |                 18.089927 |     999.910073 |   1.195081 | -2.9 | 22.433069 |   4.998415 |            176.965566 |                      0.0 |                176.965566 |                 176.965566 |                 31.281288 |                     0.0 |                      0.0 |   145.684278 |             0.0 |              0.0 |        145.684278 |             70.616813 |                 100.0 |                 48.85611 |               66.320298 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.737858 | 0.7 | 3.779721 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  6.451575 | 0.201413 | 0.419628 | 0.137861 |  0.0 |  6.451575 |  6.184507 |  0.07158 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 379.253445 | 264.371025 |  6.451575 |      0.0 | 0.201413 | 0.419628 | 0.137861 |  5.790771 |      0.0 | 0.177909 | 0.212245 | 0.003582 |   0.0 |  0.07158 |
    | 1997-19-08 | 20.0 | 21.1 |              8.5 |                14.487046 |       199.71349 |             70.3 |       2.6 |              1019.0 | 0.0 | 0.0 |    21.1 |                  70.3 |                   8.5 |                     14.487046 |            199.71349 | 18.8 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                25.063177 |                     25.063177 |                         0.0 |                          0.0 |                        1.5384 |                             1.5384 |            17.619414 |                 17.619414 |    1001.380586 |   1.198506 | -2.9 | 20.158326 |  -5.650046 |            151.782252 |                      0.0 |                151.782252 |                 151.782252 |                 26.711296 |                     0.0 |                      0.0 |   125.070956 |             0.0 |              0.0 |        125.070956 |             54.320626 |                 100.0 |                 48.85611 |               66.398945 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.866935 | 0.7 | 3.628366 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  6.883257 | 0.206979 | 0.498739 | 0.144971 |  0.0 |  6.883257 |  6.678562 | 0.077298 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 382.003492 | 271.108713 |  6.883257 |      0.0 | 0.206979 | 0.498739 | 0.144971 |  6.210098 |      0.0 | 0.185033 | 0.279355 | 0.004076 |   0.0 | 0.077298 |
    | 1997-20-08 | 20.0 | 22.4 |             11.3 |                14.416048 |      233.771902 |             69.0 |       3.6 |              1019.0 | 0.0 | 0.0 |    22.4 |                  69.0 |                  11.3 |                     14.416048 |           233.771902 | 18.8 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                27.134082 |                     27.134082 |                         0.0 |                          0.0 |                      1.648679 |                           1.648679 |            18.722517 |                 18.722517 |    1000.277483 |   1.192743 | -2.9 | 20.950339 |  -8.697966 |            177.666646 |                      0.0 |                177.666646 |                 177.666646 |                 31.116622 |                     0.0 |                      0.0 |   146.550024 |             0.0 |              0.0 |        146.550024 |             39.231563 |                 100.0 |                 48.85611 |                66.47813 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  7.82923 | 0.7 | 4.624935 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  7.348961 | 0.212254 | 0.577799 | 0.151709 |  0.0 |  7.348961 |  7.189927 | 0.083217 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 387.801458 | 276.293054 |  7.348961 |      0.0 | 0.212254 | 0.577799 | 0.151709 |  6.643946 |      0.0 | 0.191696 | 0.349693 | 0.004593 |   0.0 | 0.083217 |
    | 1997-21-08 | 20.0 | 22.5 |             12.4 |                14.344752 |      246.121286 |             62.1 |       2.5 |              1017.0 | 0.0 | 0.0 |    22.5 |                  62.1 |                  12.4 |                     14.344752 |           246.121286 | 18.8 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                27.299387 |                     27.299387 |                         0.0 |                          0.0 |                      1.657431 |                           1.657431 |             16.95292 |                  16.95292 |     1000.04708 |   1.190771 | -2.9 | 22.620153 |    0.72092 |            187.052177 |                      0.0 |                187.052177 |                 187.052177 |                 38.297604 |                     0.0 |                      0.0 |   148.754573 |             0.0 |              0.0 |        148.754573 |             56.493451 |                 100.0 |                 48.85611 |               66.557838 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 7.322689 | 0.7 | 4.756903 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  7.757566 | 0.216313 | 0.641218 | 0.156893 |  0.0 |  7.757566 |  7.701056 | 0.089133 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 384.180539 | 280.864161 |  7.757566 |      0.0 | 0.216313 | 0.641218 | 0.156893 |  7.077951 |      0.0 | 0.197812 | 0.420163 | 0.005129 |   0.0 | 0.089133 |
    | 1997-22-08 | 20.0 | 19.6 |              0.1 |                14.273175 |        90.58551 |             84.5 |       2.1 |              1015.0 | 0.0 | 0.0 |    19.6 |                  84.5 |                   0.1 |                     14.273175 |             90.58551 | 18.8 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                22.846186 |                     22.846186 |                         0.0 |                          0.0 |                      1.418946 |                           1.418946 |            19.305027 |                 19.305027 |     995.694973 |   1.199129 | -2.9 | 21.577328 |  11.863971 |             68.844988 |                      0.0 |                 68.844988 |                  68.844988 |                   7.02118 |                     0.0 |                      0.0 |    61.823807 |             0.0 |              0.0 |         61.823807 |             67.254108 |                 100.0 |                 48.85611 |               66.638052 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.567328 | 0.7 | 1.361124 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  8.166722 | 0.219892 | 0.698934 | 0.161464 |  0.0 |  8.166722 |  8.197163 | 0.094875 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 369.416568 | 288.356024 |  8.166722 |      0.0 | 0.219892 | 0.698934 | 0.161464 |  7.500252 |      0.0 | 0.203306 | 0.487924 | 0.005681 |   0.0 | 0.094875 |
    | 1997-23-08 | 20.0 | 20.3 |              6.7 |                14.201333 |      171.944993 |             83.6 |       1.5 |              1015.0 | 0.0 | 0.0 |    20.3 |                  83.6 |                   6.7 |                     14.201333 |           171.944993 | 18.8 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                23.858503 |                     23.858503 |                         0.0 |                          0.0 |                      1.473678 |                           1.473678 |            19.945708 |                 19.945708 |     995.054292 |   1.195981 | -2.9 | 17.325305 | -17.848171 |            130.678195 |                      0.0 |                130.678195 |                 130.678195 |                 18.994889 |                     0.0 |                      0.0 |   111.683306 |             0.0 |              0.0 |        111.683306 |             94.155751 |                 100.0 |                 48.85611 |               66.718758 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.374147 | 0.7 | 2.097454 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  8.984347 | 0.225758 | 0.797042 | 0.168956 |  0.0 |  8.984347 |  8.795147 | 0.101796 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 384.364739 | 294.182468 |  8.984347 |      0.0 | 0.225758 | 0.797042 | 0.168956 |  8.021581 |      0.0 | 0.208608 | 0.558706 | 0.006253 |   0.0 | 0.101796 |
    | 1997-24-08 | 20.0 | 24.1 |             10.4 |                14.129243 |      216.549813 |             76.5 |       2.2 |              1011.0 | 0.0 | 0.0 |    24.1 |                  76.5 |                  10.4 |                     14.129243 |           216.549813 | 18.8 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                30.066256 |                     30.066256 |                         0.0 |                          0.0 |                       1.80287 |                            1.80287 |            23.000686 |                 23.000686 |     987.999314 |   1.174651 | -2.9 | 21.630378 | -14.817732 |            164.577858 |                      0.0 |                164.577858 |                 164.577858 |                 21.629737 |                     0.0 |                      0.0 |   142.948121 |             0.0 |              0.0 |        142.948121 |             64.197103 |                 100.0 |                 48.85611 |                66.79994 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 5.573024 | 0.7 | 3.649811 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 |  9.835059 | 0.230319 | 0.876255 | 0.174782 |  0.0 |  9.835059 |  9.544513 | 0.110469 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  396.28247 | 297.516241 |  9.835059 |      0.0 | 0.230319 | 0.876255 | 0.174782 |  8.689617 |      0.0 | 0.213877 | 0.634173 | 0.006846 |   0.0 | 0.110469 |
    | 1997-25-08 | 20.0 | 23.7 |              6.5 |                14.056918 |       166.91683 |             76.6 |       1.9 |              1008.0 | 0.0 | 0.0 |    23.7 |                  76.6 |                   6.5 |                     14.056918 |            166.91683 | 18.8 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                29.352596 |                     29.352596 |                         0.0 |                          0.0 |                      1.765541 |                           1.765541 |            22.484088 |                 22.484088 |     985.515912 |   1.172942 | -2.9 | 25.062685 |   8.176109 |             126.85679 |                      0.0 |                 126.85679 |                  126.85679 |                 16.222254 |                     0.0 |                      0.0 |   110.634536 |             0.0 |              0.0 |        110.634536 |             74.333488 |                 100.0 |                 48.85611 |               66.881585 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 4.467668 | 0.7 | 2.895865 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 | 10.483053 | 0.232929 | 0.922693 | 0.178116 |  0.0 | 10.483053 | 10.322796 | 0.119477 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 385.206362 | 300.903585 | 10.483053 |      0.0 | 0.232929 | 0.922693 | 0.178116 |  9.390662 |      0.0 | 0.218678 | 0.706002 | 0.007454 |   0.0 | 0.119477 |
    | 1997-26-08 | 20.0 | 21.1 |              4.9 |                13.984374 |      145.985611 |             84.2 |       2.3 |              1006.0 | 0.0 | 0.0 |    21.1 |                  84.2 |                   4.9 |                     13.984374 |           145.985611 | 18.8 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                25.063177 |                     25.063177 |                         0.0 |                          0.0 |                        1.5384 |                             1.5384 |            21.103195 |                 21.103195 |     984.896805 |   1.181557 | -2.9 | 21.872765 |   4.636593 |            110.949064 |                      0.0 |                110.949064 |                 110.949064 |                 14.602883 |                     0.0 |                      0.0 |    96.346181 |             0.0 |              0.0 |         96.346181 |             61.405925 |                 100.0 |                 48.85611 |               66.963678 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.569826 | 0.7 | 2.071449 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 | 11.394349 | 0.235581 | 0.970688 | 0.181504 |  0.0 | 11.394349 | 11.136926 |   0.1289 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 377.669769 | 304.150015 | 11.394349 |      0.0 | 0.235581 | 0.970688 | 0.181504 | 10.134751 |      0.0 | 0.222895 | 0.771208 | 0.008071 |   0.0 |   0.1289 |
    | 1997-27-08 | 20.0 | 20.9 |              6.6 |                13.911624 |       165.54389 |             85.8 |       2.1 |              1005.0 | 0.0 | 0.0 |    20.9 |                  85.8 |                   6.6 |                     13.911624 |            165.54389 | 18.8 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 24.75714 |                      24.75714 |                         0.0 |                          0.0 |                      1.521999 |                           1.521999 |            21.241626 |                 21.241626 |     983.758374 |   1.181114 | -2.9 | 19.702227 |   -7.18664 |            125.813357 |                      0.0 |                125.813357 |                 125.813357 |                 17.349121 |                     0.0 |                      0.0 |   108.464235 |             0.0 |              0.0 |        108.464235 |             67.254108 |                 100.0 |                 48.85611 |               67.046207 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  3.56693 | 0.7 | 2.108859 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 | 13.250015 | 0.238123 | 1.017441 |  0.18475 |  0.0 | 13.250015 | 12.737917 |  0.14743 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 381.956408 | 305.450827 |      12.0 | 1.250015 | 0.238123 | 1.017441 |  0.18475 | 10.878137 | 0.792807 | 0.226675 |   0.8316 | 0.008698 |   0.0 |  0.14743 |
    | 1997-28-08 | 20.0 | 20.2 |              0.3 |                13.838682 |       87.693346 |             86.5 |       2.6 |              1000.0 | 0.0 | 0.0 |    20.2 |                  86.5 |                   0.3 |                     13.838682 |            87.693346 | 18.8 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                23.711531 |                     23.711531 |                         0.0 |                          0.0 |                      1.465752 |                           1.465752 |            20.510475 |                 20.510475 |     979.489525 |   1.178323 | -2.9 | 20.936779 |   4.420674 |             66.646943 |                      0.0 |                 66.646943 |                  66.646943 |                  6.818024 |                     0.0 |                      0.0 |    59.828919 |             0.0 |              0.0 |         59.828919 |             54.320626 |                 100.0 |                 48.85611 |               67.129158 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.621228 | 0.7 |  1.33161 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 | 14.550827 | 0.239141 |  1.03638 | 0.186051 |  0.0 | 14.550827 |  14.56294 | 0.168553 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 374.635735 | 306.206818 |      12.0 | 2.550827 | 0.239141 |  1.03638 | 0.186051 | 11.401361 | 2.037971 | 0.229904 | 0.884372 | 0.009332 |   0.0 | 0.168553 |
    | 1997-29-08 | 20.0 | 16.1 |              0.1 |                13.765561 |       84.356885 |             89.6 |       2.6 |              1004.0 | 0.0 | 0.0 |    16.1 |                  89.6 |                   0.1 |                     13.765561 |            84.356885 | 18.8 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                18.326791 |                     18.326791 |                         0.0 |                          0.0 |                      1.170311 |                           1.170311 |            16.420805 |                 16.420805 |     987.579195 |   1.201704 | -2.9 | 18.828425 |   16.37055 |             64.111233 |                      0.0 |                 64.111233 |                  64.111233 |                  8.199212 |                     0.0 |                      0.0 |    55.912021 |             0.0 |              0.0 |         55.912021 |             54.320626 |                 100.0 |                 48.85611 |               67.212519 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.027587 | 0.7 | 0.881773 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 | 15.306818 | 0.239733 | 1.047439 | 0.186807 |  0.0 | 15.306818 | 15.838636 | 0.183318 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 355.365185 | 306.644248 |      12.0 | 3.306818 | 0.239733 | 1.047439 | 0.186807 |  11.68056 | 2.988724 | 0.232478 | 0.926908 | 0.009966 |   0.0 | 0.183318 |
    | 1997-30-08 | 20.0 | 17.0 |              5.0 |                13.692273 |      142.309202 |             78.3 |       2.1 |              1011.0 | 0.0 | 0.0 |    17.0 |                  78.3 |                   5.0 |                     13.692273 |           142.309202 | 18.8 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                19.406929 |                     19.406929 |                         0.0 |                          0.0 |                      1.230421 |                           1.230421 |            15.195625 |                 15.195625 |     995.804375 |   1.206937 | -2.9 | 13.278507 |  -22.32896 |            108.154994 |                      0.0 |                108.154994 |                 108.154994 |                 21.775852 |                     0.0 |                      0.0 |    86.379141 |             0.0 |              0.0 |         86.379141 |             67.254108 |                 100.0 |                 48.85611 |               67.296277 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.393474 | 0.7 | 1.913351 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 | 15.744248 | 0.240076 | 1.053857 | 0.187244 |  0.0 | 15.744248 | 16.593391 | 0.192053 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 374.794145 | 305.605473 |      12.0 | 3.744248 | 0.240076 | 1.053857 | 0.187244 | 11.829543 | 3.558462 | 0.234482 | 0.960303 |   0.0106 |   0.0 | 0.192053 |
    | 1997-31-08 | 20.0 | 18.2 |              3.8 |                13.618831 |      126.722078 |             78.5 |       2.1 |              1014.0 | 0.0 | 0.0 |    18.2 |                  78.5 |                   3.8 |                     13.618831 |           126.722078 | 18.8 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                20.933479 |                     20.933479 |                         0.0 |                          0.0 |                      1.314615 |                           1.314615 |            16.432781 |                 16.432781 |     997.567219 |   1.204994 | -2.9 | 18.874047 |   4.044283 |              96.30878 |                      0.0 |                  96.30878 |                   96.30878 |                 17.491881 |                     0.0 |                      0.0 |    78.816899 |             0.0 |              0.0 |         78.816899 |             67.254108 |                 100.0 |                 48.85611 |               67.380422 | 18.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 3.321963 | 0.7 | 1.882123 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 18.1 | 14.705473 | 0.239262 | 1.038639 | 0.186205 |  0.0 | 14.705473 | 16.209861 | 0.187614 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 367.849862 | 305.653771 |      12.0 | 2.705473 | 0.239262 | 1.038639 | 0.186205 | 11.909042 | 3.070354 | 0.235874 | 0.983359 | 0.011231 |   0.0 | 0.187614 |

>>> round_(model.check_waterbalance(conditions))
0.0

>>> inputs.nied.series = nied

.. _lland_v4_wasser_daily:

water
-----

.. integration-test::

    >>> lnk(WASSER)
    >>> negq(False)
    >>> surfaceresistance.wasser_aug = 0.0
    >>> cropheight.wasser_aug = 0.05
    >>> albedo.wasser_aug = 0.7
    >>> conditions_wasser = test(
    ...     "lland_v4_wasser_daily",
    ...     axis1=(fluxes.nkor, fluxes.evi, fluxes.qah),
    ...     get_conditions="1997-08-03")
    |       date | nied | teml | sunshineduration | possiblesunshineduration | globalradiation | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration | dailypossiblesunshineduration | dailyglobalradiation |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair |   g |  tz |  wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |     evpo |      evi | evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatinz | wlatsnow | wsensinz | wsenssnow | wsurfinz | wsurf | sff | fvg | wada | qdb | qib1 | qib2 | qbb | qkap | qdgz |       qah |       qa | inzp | stinz | sinz | esnowinz | asinz | wats | waes | esnow | taus | ebdn | bowa | qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz | qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |                15.690785 |       190.25149 |             86.2 |       2.4 |              1007.0 | 0.0 | 0.0 |    16.1 |                  86.2 |                   6.3 |                     15.690785 |            190.25149 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                18.326791 |                     18.326791 |                         0.0 |                          0.0 |                      1.170311 |                           1.170311 |            15.797694 |                 15.797694 |     991.202306 |   1.205601 | 0.0 | nan | 0.0 |             57.075447 |                      0.0 |                 57.075447 |                  57.075447 |                 21.732733 |                     0.0 |                      0.0 |    35.342714 |             0.0 |              0.0 |         35.342714 |             88.752361 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.077197 | 1.077197 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 11.424803 | 0.132232 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.132232 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |                15.628821 |      127.660782 |             85.8 |       1.6 |              1013.0 | 0.0 | 0.0 |    17.3 |                  85.8 |                   1.7 |                     15.628821 |           127.660782 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                19.779139 |                     19.779139 |                         0.0 |                          0.0 |                      1.251029 |                           1.251029 |            16.970501 |                 16.970501 |     996.029499 |   1.207284 | 0.0 | nan | 0.0 |             38.298235 |                      0.0 |                 38.298235 |                  38.298235 |                 11.219797 |                     0.0 |                      0.0 |    27.078438 |             0.0 |              0.0 |         27.078438 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.86545 |    0.094 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |                15.566134 |      164.401298 |             82.4 |       1.5 |              1016.0 | 0.0 | 0.0 |    18.9 |                  82.4 |                   4.5 |                     15.566134 |           164.401298 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                21.871576 |                     21.871576 |                         0.0 |                          0.0 |                      1.365939 |                           1.365939 |            18.022179 |                 18.022179 |     997.977821 |   1.203775 | 0.0 | nan | 0.0 |             49.320389 |                      0.0 |                 49.320389 |                  49.320389 |                 16.044181 |                     0.0 |                      0.0 |    33.276209 |             0.0 |              0.0 |         33.276209 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.092365 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |                15.502754 |      269.020502 |             77.6 |       2.8 |              1017.0 | 0.0 | 0.0 |    17.8 |                  77.6 |                  12.4 |                     15.502754 |           269.020502 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                20.413369 |                     20.413369 |                         0.0 |                          0.0 |                      1.286025 |                           1.286025 |            15.840774 |                 15.840774 |    1001.159226 |    1.21051 | 0.0 | nan | 0.0 |              80.70615 |                      0.0 |                  80.70615 |                   80.70615 |                 35.980407 |                     0.0 |                      0.0 |    44.725744 |             0.0 |              0.0 |         44.725744 |             76.073452 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.565894 |    0.094 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |                15.438712 |      287.848667 |             74.9 |       3.0 |              1018.0 | 0.0 | 0.0 |    18.5 |                  74.9 |                  13.9 |                     15.438712 |           287.848667 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                21.331125 |                     21.331125 |                         0.0 |                          0.0 |                      1.336407 |                           1.336407 |            15.977013 |                 15.977013 |    1002.022987 |   1.208738 | 0.0 | nan | 0.0 |               86.3546 |                      0.0 |                   86.3546 |                    86.3546 |                 39.523937 |                     0.0 |                      0.0 |    46.830664 |             0.0 |              0.0 |         46.830664 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.729733 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |                15.374034 |      274.674263 |             77.9 |       3.5 |              1018.0 | 0.0 | 0.0 |    19.9 |                  77.9 |                  13.0 |                     15.374034 |           274.674263 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                23.275348 |                     23.275348 |                         0.0 |                          0.0 |                      1.442189 |                           1.442189 |            18.131496 |                 18.131496 |     999.868504 |   1.201995 | 0.0 | nan | 0.0 |             82.402279 |                      0.0 |                 82.402279 |                  82.402279 |                 32.936813 |                     0.0 |                      0.0 |    49.465466 |             0.0 |              0.0 |         49.465466 |             60.858762 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.826087 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |                 15.30875 |      284.047892 |             73.5 |       3.8 |              1019.0 | 0.0 | 0.0 |    20.8 |                  73.5 |                  13.8 |                      15.30875 |           284.047892 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                24.605348 |                     24.605348 |                         0.0 |                          0.0 |                      1.513854 |                           1.513854 |            18.084931 |                 18.084931 |    1000.915069 |   1.199521 | 0.0 | nan | 0.0 |             85.214368 |                      0.0 |                 85.214368 |                  85.214368 |                  35.31351 |                     0.0 |                      0.0 |    49.900858 |             0.0 |              0.0 |         49.900858 |             56.054123 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.041956 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |                15.242885 |      262.971263 |             80.1 |       3.3 |              1019.0 | 0.0 | 0.0 |    20.9 |                  80.1 |                  12.3 |                     15.242885 |           262.971263 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 24.75714 |                      24.75714 |                         0.0 |                          0.0 |                      1.521999 |                           1.521999 |            19.830469 |                 19.830469 |     999.169531 |   1.198331 | 0.0 | nan | 0.0 |             78.891379 |                      0.0 |                 78.891379 |                  78.891379 |                 28.392598 |                     0.0 |                      0.0 |    50.498781 |             0.0 |              0.0 |         50.498781 |             64.547172 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.800053 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |                15.176467 |      272.256414 |             72.9 |       1.8 |              1019.0 | 0.0 | 0.0 |    22.3 |                  72.9 |                  13.1 |                     15.176467 |           272.256414 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 26.96965 |                      26.96965 |                         0.0 |                          0.0 |                      1.639967 |                           1.639967 |            19.660875 |                 19.660875 |     999.339125 |   1.192729 | 0.0 | nan | 0.0 |             81.676924 |                      0.0 |                 81.676924 |                  81.676924 |                 31.114889 |                     0.0 |                      0.0 |    50.562035 |             0.0 |              0.0 |         50.562035 |            118.336481 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.82186 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |                15.109521 |      267.019218 |             68.5 |       1.9 |              1020.0 | 0.0 | 0.0 |    23.0 |                  68.5 |                  12.8 |                     15.109521 |           267.019218 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                28.139149 |                     28.139149 |                         0.0 |                          0.0 |                      1.701779 |                           1.701779 |            19.275317 |                 19.275317 |    1000.724683 |   1.191257 | 0.0 | nan | 0.0 |             80.105765 |                      0.0 |                 80.105765 |                  80.105765 |                 32.010643 |                     0.0 |                      0.0 |    48.095123 |             0.0 |              0.0 |         48.095123 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.891469 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |                15.042071 |      269.626547 |             68.6 |       3.0 |              1021.0 | 0.0 | 0.0 |    23.1 |                  68.6 |                  13.1 |                     15.042071 |           269.626547 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                28.309776 |                     28.309776 |                         0.0 |                          0.0 |                      1.710768 |                           1.710768 |            19.420506 |                 19.420506 |    1001.579494 |   1.191966 | 0.0 | nan | 0.0 |             80.887964 |                      0.0 |                 80.887964 |                  80.887964 |                 32.400014 |                     0.0 |                      0.0 |     48.48795 |             0.0 |              0.0 |          48.48795 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.106575 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |                14.974141 |      270.878291 |             66.0 |       2.5 |              1022.0 | 0.0 | 0.0 |    23.2 |                  66.0 |                  13.3 |                     14.974141 |           270.878291 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                28.481304 |                     28.481304 |                         0.0 |                          0.0 |                      1.719796 |                           1.719796 |             18.79766 |                  18.79766 |     1003.20234 |   1.193016 | 0.0 | nan | 0.0 |             81.263487 |                      0.0 |                 81.263487 |                  81.263487 |                 34.565783 |                     0.0 |                      0.0 |    46.697704 |             0.0 |              0.0 |         46.697704 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.036929 |    0.094 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |                14.905755 |      261.697164 |             69.0 |       1.6 |              1019.0 | 0.0 | 0.0 |    23.2 |                  69.0 |                  12.7 |                     14.905755 |           261.697164 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                28.481304 |                     28.481304 |                         0.0 |                          0.0 |                      1.719796 |                           1.719796 |              19.6521 |                   19.6521 |       999.3479 |    1.18911 | 0.0 | nan | 0.0 |             78.509149 |                      0.0 |                 78.509149 |                  78.509149 |                  31.34697 |                     0.0 |                      0.0 |    47.162179 |             0.0 |              0.0 |         47.162179 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.808559 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |                14.836934 |      227.962116 |             75.6 |       2.7 |              1015.0 | 0.0 | 0.0 |    21.7 |                  75.6 |                  10.2 |                     14.836934 |           227.962116 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 26.00118 |                      26.00118 |                         0.0 |                          0.0 |                        1.5885 |                             1.5885 |            19.656892 |                 19.656892 |     995.343108 |   1.190432 | 0.0 | nan | 0.0 |             68.388635 |                      0.0 |                 68.388635 |                  68.388635 |                 25.918437 |                     0.0 |                      0.0 |    42.470198 |             0.0 |              0.0 |         42.470198 |             78.890988 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.672736 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |                  14.7677 |      216.356321 |             81.4 |       3.2 |              1016.0 | 0.0 | 0.0 |    20.8 |                  81.4 |                   9.4 |                       14.7677 |           216.356321 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                24.605348 |                     24.605348 |                         0.0 |                          0.0 |                      1.513854 |                           1.513854 |            20.028753 |                 20.028753 |     995.971247 |   1.195095 | 0.0 | nan | 0.0 |             64.906896 |                      0.0 |                 64.906896 |                  64.906896 |                 23.397902 |                     0.0 |                      0.0 |    41.508994 |             0.0 |              0.0 |         41.508994 |             66.564271 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.529542 | 1.529542 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  1.008458 | 0.011672 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.011672 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |                14.698075 |      226.637904 |             79.0 |       1.3 |              1019.0 | 0.0 | 0.0 |    19.5 |                  79.0 |                  10.3 |                     14.698075 |           226.637904 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                22.704676 |                     22.704676 |                         0.0 |                          0.0 |                      1.411269 |                           1.411269 |            17.936694 |                 17.936694 |    1001.063306 |   1.204916 | 0.0 | nan | 0.0 |             67.991371 |                      0.0 |                 67.991371 |                  67.991371 |                 28.794151 |                     0.0 |                      0.0 |    39.197221 |             0.0 |              0.0 |         39.197221 |            163.850513 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.283901 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |                14.628078 |      235.528503 |             75.9 |       1.9 |              1019.0 | 0.0 | 0.0 |    21.0 |                  75.9 |                  11.1 |                     14.628078 |           235.528503 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                24.909749 |                     24.909749 |                         0.0 |                          0.0 |                      1.530181 |                           1.530181 |            18.906499 |                 18.906499 |    1000.093501 |   1.198338 | 0.0 | nan | 0.0 |             70.658551 |                      0.0 |                 70.658551 |                  70.658551 |                 29.177702 |                     0.0 |                      0.0 |    41.480849 |             0.0 |              0.0 |         41.480849 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.510743 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |                14.557728 |      232.849428 |             70.0 |       2.0 |              1018.0 | 0.0 | 0.0 |    21.6 |                  70.0 |                  11.0 |                     14.557728 |           232.849428 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                25.842752 |                     25.842752 |                         0.0 |                          0.0 |                      1.580055 |                           1.580055 |            18.089927 |                 18.089927 |     999.910073 |   1.195081 | 0.0 | nan | 0.0 |             69.854829 |                      0.0 |                 69.854829 |                  69.854829 |                 31.281288 |                     0.0 |                      0.0 |    38.573541 |             0.0 |              0.0 |         38.573541 |            106.502833 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.594572 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |                14.487046 |       199.71349 |             70.3 |       2.6 |              1019.0 | 0.0 | 0.0 |    21.1 |                  70.3 |                   8.5 |                     14.487046 |            199.71349 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                25.063177 |                     25.063177 |                         0.0 |                          0.0 |                        1.5384 |                             1.5384 |            17.619414 |                 17.619414 |    1001.380586 |   1.198506 | 0.0 | nan | 0.0 |             59.914047 |                      0.0 |                 59.914047 |                  59.914047 |                 26.711296 |                     0.0 |                      0.0 |    33.202751 |             0.0 |              0.0 |         33.202751 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.541567 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |                14.416048 |      233.771902 |             69.0 |       3.6 |              1019.0 | 0.0 | 0.0 |    22.4 |                  69.0 |                  11.3 |                     14.416048 |           233.771902 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                27.134082 |                     27.134082 |                         0.0 |                          0.0 |                      1.648679 |                           1.648679 |            18.722517 |                 18.722517 |    1000.277483 |   1.192743 | 0.0 | nan | 0.0 |             70.131571 |                      0.0 |                 70.131571 |                  70.131571 |                 31.116622 |                     0.0 |                      0.0 |    39.014949 |             0.0 |              0.0 |         39.014949 |             59.168241 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.939383 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |                14.344752 |      246.121286 |             62.1 |       2.5 |              1017.0 | 0.0 | 0.0 |    22.5 |                  62.1 |                  12.4 |                     14.344752 |           246.121286 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                27.299387 |                     27.299387 |                         0.0 |                          0.0 |                      1.657431 |                           1.657431 |             16.95292 |                  16.95292 |     1000.04708 |   1.190771 | 0.0 | nan | 0.0 |             73.836386 |                      0.0 |                 73.836386 |                  73.836386 |                 38.297604 |                     0.0 |                      0.0 |    35.538782 |             0.0 |              0.0 |         35.538782 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.826811 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |                14.273175 |        90.58551 |             84.5 |       2.1 |              1015.0 | 0.0 | 0.0 |    19.6 |                  84.5 |                   0.1 |                     14.273175 |             90.58551 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                22.846186 |                     22.846186 |                         0.0 |                          0.0 |                      1.418946 |                           1.418946 |            19.305027 |                 19.305027 |     995.694973 |   1.199129 | 0.0 | nan | 0.0 |             27.175653 |                      0.0 |                 27.175653 |                  27.175653 |                   7.02118 |                     0.0 |                      0.0 |    20.154473 |             0.0 |              0.0 |         20.154473 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.806188 | 0.806188 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  1.073812 | 0.012428 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.012428 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |                14.201333 |      171.944993 |             83.6 |       1.5 |              1015.0 | 0.0 | 0.0 |    20.3 |                  83.6 |                   6.7 |                     14.201333 |           171.944993 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                23.858503 |                     23.858503 |                         0.0 |                          0.0 |                      1.473678 |                           1.473678 |            19.945708 |                 19.945708 |     995.054292 |   1.195981 | 0.0 | nan | 0.0 |             51.583498 |                      0.0 |                 51.583498 |                  51.583498 |                 18.994889 |                     0.0 |                      0.0 |    32.588609 |             0.0 |              0.0 |         32.588609 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.08421 |    0.282 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |                14.129243 |      216.549813 |             76.5 |       2.2 |              1011.0 | 0.0 | 0.0 |    24.1 |                  76.5 |                  10.4 |                     14.129243 |           216.549813 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                30.066256 |                     30.066256 |                         0.0 |                          0.0 |                       1.80287 |                            1.80287 |            23.000686 |                 23.000686 |     987.999314 |   1.174651 | 0.0 | nan | 0.0 |             64.964944 |                      0.0 |                 64.964944 |                  64.964944 |                 21.629737 |                     0.0 |                      0.0 |    43.335207 |             0.0 |              0.0 |         43.335207 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.672232 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |                14.056918 |       166.91683 |             76.6 |       1.9 |              1008.0 | 0.0 | 0.0 |    23.7 |                  76.6 |                   6.5 |                     14.056918 |            166.91683 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                29.352596 |                     29.352596 |                         0.0 |                          0.0 |                      1.765541 |                           1.765541 |            22.484088 |                 22.484088 |     985.515912 |   1.172942 | 0.0 | nan | 0.0 |             50.075049 |                      0.0 |                 50.075049 |                  50.075049 |                 16.222254 |                     0.0 |                      0.0 |    33.852795 |             0.0 |              0.0 |         33.852795 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.374102 |    0.094 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |                13.984374 |      145.985611 |             84.2 |       2.3 |              1006.0 | 0.0 | 0.0 |    21.1 |                  84.2 |                   4.9 |                     13.984374 |           145.985611 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                25.063177 |                     25.063177 |                         0.0 |                          0.0 |                        1.5384 |                             1.5384 |            21.103195 |                 21.103195 |     984.896805 |   1.181557 | 0.0 | nan | 0.0 |             43.795683 |                      0.0 |                 43.795683 |                  43.795683 |                 14.602883 |                     0.0 |                      0.0 |      29.1928 |             0.0 |              0.0 |           29.1928 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.077096 | 1.077096 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  8.792904 |  0.10177 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.10177 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |                13.911624 |       165.54389 |             85.8 |       2.1 |              1005.0 | 0.0 | 0.0 |    20.9 |                  85.8 |                   6.6 |                     13.911624 |            165.54389 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 24.75714 |                      24.75714 |                         0.0 |                          0.0 |                      1.521999 |                           1.521999 |            21.241626 |                 21.241626 |     983.758374 |   1.181114 | 0.0 | nan | 0.0 |             49.663167 |                      0.0 |                 49.663167 |                  49.663167 |                 17.349121 |                     0.0 |                      0.0 |    32.314046 |             0.0 |              0.0 |         32.314046 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.09756 |  1.09756 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |   5.01244 | 0.058014 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.058014 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |                13.838682 |       87.693346 |             86.5 |       2.6 |              1000.0 | 0.0 | 0.0 |    20.2 |                  86.5 |                   0.3 |                     13.838682 |            87.693346 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                23.711531 |                     23.711531 |                         0.0 |                          0.0 |                      1.465752 |                           1.465752 |            20.510475 |                 20.510475 |     979.489525 |   1.178323 | 0.0 | nan | 0.0 |             26.308004 |                      0.0 |                 26.308004 |                  26.308004 |                  6.818024 |                     0.0 |                      0.0 |     19.48998 |             0.0 |              0.0 |          19.48998 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.794859 | 0.794859 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 13.305141 | 0.153995 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.153995 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |                13.765561 |       84.356885 |             89.6 |       2.6 |              1004.0 | 0.0 | 0.0 |    16.1 |                  89.6 |                   0.1 |                     13.765561 |            84.356885 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                18.326791 |                     18.326791 |                         0.0 |                          0.0 |                      1.170311 |                           1.170311 |            16.420805 |                 16.420805 |     987.579195 |   1.201704 | 0.0 | nan | 0.0 |             25.307066 |                      0.0 |                 25.307066 |                  25.307066 |                  8.199212 |                     0.0 |                      0.0 |    17.107854 |             0.0 |              0.0 |         17.107854 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.607776 | 0.607776 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  8.040224 | 0.093058 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.093058 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |                13.692273 |      142.309202 |             78.3 |       2.1 |              1011.0 | 0.0 | 0.0 |    17.0 |                  78.3 |                   5.0 |                     13.692273 |           142.309202 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                19.406929 |                     19.406929 |                         0.0 |                          0.0 |                      1.230421 |                           1.230421 |            15.195625 |                 15.195625 |     995.804375 |   1.206937 | 0.0 | nan | 0.0 |             42.692761 |                      0.0 |                 42.692761 |                  42.692761 |                 21.775852 |                     0.0 |                      0.0 |    20.916908 |             0.0 |              0.0 |         20.916908 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.90065 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |                13.618831 |      126.722078 |             78.5 |       2.1 |              1014.0 | 0.0 | 0.0 |    18.2 |                  78.5 |                   3.8 |                     13.618831 |           126.722078 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                20.933479 |                     20.933479 |                         0.0 |                          0.0 |                      1.314615 |                           1.314615 |            16.432781 |                 16.432781 |     997.567219 |   1.204994 | 0.0 | nan | 0.0 |             38.016624 |                      0.0 |                 38.016624 |                  38.016624 |                 17.491881 |                     0.0 |                      0.0 |    20.524743 |             0.0 |              0.0 |         20.524743 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.912078 |      0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |

>>> round_(model.check_waterbalance(conditions))
0.0

.. _lland_v4_wasser_routing_daily:

water (routing)
---------------

.. integration-test::

    >>> inlet.sequences.sim.series = 0.02
    >>> test("lland_v4_wasser_routing_daily",
    ...      axis1=(fluxes.nkor, fluxes.evi, fluxes.qah))
    |       date | nied | teml | sunshineduration | possiblesunshineduration | globalradiation | relativehumidity | windspeed | atmosphericpressure |   qz |   qzh | temltag | dailyrelativehumidity | dailysunshineduration | dailypossiblesunshineduration | dailyglobalradiation |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair |   g |  tz |  wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |     evpo |      evi | evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatinz | wlatsnow | wsensinz | wsenssnow | wsurfinz | wsurf | sff | fvg | wada | qdb | qib1 | qib2 | qbb | qkap | qdgz |       qah |       qa | inzp | stinz | sinz | esnowinz | asinz | wats | waes | esnow | taus | ebdn | bowa | qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz | qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |                15.690785 |       190.25149 |             86.2 |       2.4 |              1007.0 | 0.02 | 1.728 |    16.1 |                  86.2 |                   6.3 |                     15.690785 |            190.25149 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                18.326791 |                     18.326791 |                         0.0 |                          0.0 |                      1.170311 |                           1.170311 |            15.797694 |                 15.797694 |     991.202306 |   1.205601 | 0.0 | nan | 0.0 |             57.075447 |                      0.0 |                 57.075447 |                  57.075447 |                 21.732733 |                     0.0 |                      0.0 |    35.342714 |             0.0 |              0.0 |         35.342714 |             88.752361 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.077197 | 1.077197 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 13.152803 | 0.152232 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.152232 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |                15.628821 |      127.660782 |             85.8 |       1.6 |              1013.0 | 0.02 | 1.728 |    17.3 |                  85.8 |                   1.7 |                     15.628821 |           127.660782 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                19.779139 |                     19.779139 |                         0.0 |                          0.0 |                      1.251029 |                           1.251029 |            16.970501 |                 16.970501 |     996.029499 |   1.207284 | 0.0 | nan | 0.0 |             38.298235 |                      0.0 |                 38.298235 |                  38.298235 |                 11.219797 |                     0.0 |                      0.0 |    27.078438 |             0.0 |              0.0 |         27.078438 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.86545 |  0.86545 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |   0.95655 | 0.011071 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.011071 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |                15.566134 |      164.401298 |             82.4 |       1.5 |              1016.0 | 0.02 | 1.728 |    18.9 |                  82.4 |                   4.5 |                     15.566134 |           164.401298 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                21.871576 |                     21.871576 |                         0.0 |                          0.0 |                      1.365939 |                           1.365939 |            18.022179 |                 18.022179 |     997.977821 |   1.203775 | 0.0 | nan | 0.0 |             49.320389 |                      0.0 |                 49.320389 |                  49.320389 |                 16.044181 |                     0.0 |                      0.0 |    33.276209 |             0.0 |              0.0 |         33.276209 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.092365 | 1.092365 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.635635 | 0.007357 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.007357 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |                15.502754 |      269.020502 |             77.6 |       2.8 |              1017.0 | 0.02 | 1.728 |    17.8 |                  77.6 |                  12.4 |                     15.502754 |           269.020502 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                20.413369 |                     20.413369 |                         0.0 |                          0.0 |                      1.286025 |                           1.286025 |            15.840774 |                 15.840774 |    1001.159226 |    1.21051 | 0.0 | nan | 0.0 |              80.70615 |                      0.0 |                  80.70615 |                   80.70615 |                 35.980407 |                     0.0 |                      0.0 |    44.725744 |             0.0 |              0.0 |         44.725744 |             76.073452 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.565894 | 1.565894 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.256106 | 0.002964 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.002964 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |                15.438712 |      287.848667 |             74.9 |       3.0 |              1018.0 | 0.02 | 1.728 |    18.5 |                  74.9 |                  13.9 |                     15.438712 |           287.848667 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                21.331125 |                     21.331125 |                         0.0 |                          0.0 |                      1.336407 |                           1.336407 |            15.977013 |                 15.977013 |    1002.022987 |   1.208738 | 0.0 | nan | 0.0 |               86.3546 |                      0.0 |                   86.3546 |                    86.3546 |                 39.523937 |                     0.0 |                      0.0 |    46.830664 |             0.0 |              0.0 |         46.830664 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.729733 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |                15.374034 |      274.674263 |             77.9 |       3.5 |              1018.0 | 0.02 | 1.728 |    19.9 |                  77.9 |                  13.0 |                     15.374034 |           274.674263 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                23.275348 |                     23.275348 |                         0.0 |                          0.0 |                      1.442189 |                           1.442189 |            18.131496 |                 18.131496 |     999.868504 |   1.201995 | 0.0 | nan | 0.0 |             82.402279 |                      0.0 |                 82.402279 |                  82.402279 |                 32.936813 |                     0.0 |                      0.0 |    49.465466 |             0.0 |              0.0 |         49.465466 |             60.858762 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.826087 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |                 15.30875 |      284.047892 |             73.5 |       3.8 |              1019.0 | 0.02 | 1.728 |    20.8 |                  73.5 |                  13.8 |                      15.30875 |           284.047892 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                24.605348 |                     24.605348 |                         0.0 |                          0.0 |                      1.513854 |                           1.513854 |            18.084931 |                 18.084931 |    1000.915069 |   1.199521 | 0.0 | nan | 0.0 |             85.214368 |                      0.0 |                 85.214368 |                  85.214368 |                  35.31351 |                     0.0 |                      0.0 |    49.900858 |             0.0 |              0.0 |         49.900858 |             56.054123 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.041956 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |                15.242885 |      262.971263 |             80.1 |       3.3 |              1019.0 | 0.02 | 1.728 |    20.9 |                  80.1 |                  12.3 |                     15.242885 |           262.971263 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 24.75714 |                      24.75714 |                         0.0 |                          0.0 |                      1.521999 |                           1.521999 |            19.830469 |                 19.830469 |     999.169531 |   1.198331 | 0.0 | nan | 0.0 |             78.891379 |                      0.0 |                 78.891379 |                  78.891379 |                 28.392598 |                     0.0 |                      0.0 |    50.498781 |             0.0 |              0.0 |         50.498781 |             64.547172 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.800053 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |                15.176467 |      272.256414 |             72.9 |       1.8 |              1019.0 | 0.02 | 1.728 |    22.3 |                  72.9 |                  13.1 |                     15.176467 |           272.256414 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 26.96965 |                      26.96965 |                         0.0 |                          0.0 |                      1.639967 |                           1.639967 |            19.660875 |                 19.660875 |     999.339125 |   1.192729 | 0.0 | nan | 0.0 |             81.676924 |                      0.0 |                 81.676924 |                  81.676924 |                 31.114889 |                     0.0 |                      0.0 |    50.562035 |             0.0 |              0.0 |         50.562035 |            118.336481 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.82186 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |                15.109521 |      267.019218 |             68.5 |       1.9 |              1020.0 | 0.02 | 1.728 |    23.0 |                  68.5 |                  12.8 |                     15.109521 |           267.019218 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                28.139149 |                     28.139149 |                         0.0 |                          0.0 |                      1.701779 |                           1.701779 |            19.275317 |                 19.275317 |    1000.724683 |   1.191257 | 0.0 | nan | 0.0 |             80.105765 |                      0.0 |                 80.105765 |                  80.105765 |                 32.010643 |                     0.0 |                      0.0 |    48.095123 |             0.0 |              0.0 |         48.095123 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.891469 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |                15.042071 |      269.626547 |             68.6 |       3.0 |              1021.0 | 0.02 | 1.728 |    23.1 |                  68.6 |                  13.1 |                     15.042071 |           269.626547 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                28.309776 |                     28.309776 |                         0.0 |                          0.0 |                      1.710768 |                           1.710768 |            19.420506 |                 19.420506 |    1001.579494 |   1.191966 | 0.0 | nan | 0.0 |             80.887964 |                      0.0 |                 80.887964 |                  80.887964 |                 32.400014 |                     0.0 |                      0.0 |     48.48795 |             0.0 |              0.0 |          48.48795 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.106575 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |                14.974141 |      270.878291 |             66.0 |       2.5 |              1022.0 | 0.02 | 1.728 |    23.2 |                  66.0 |                  13.3 |                     14.974141 |           270.878291 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                28.481304 |                     28.481304 |                         0.0 |                          0.0 |                      1.719796 |                           1.719796 |             18.79766 |                  18.79766 |     1003.20234 |   1.193016 | 0.0 | nan | 0.0 |             81.263487 |                      0.0 |                 81.263487 |                  81.263487 |                 34.565783 |                     0.0 |                      0.0 |    46.697704 |             0.0 |              0.0 |         46.697704 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.036929 |    1.822 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |                14.905755 |      261.697164 |             69.0 |       1.6 |              1019.0 | 0.02 | 1.728 |    23.2 |                  69.0 |                  12.7 |                     14.905755 |           261.697164 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                28.481304 |                     28.481304 |                         0.0 |                          0.0 |                      1.719796 |                           1.719796 |              19.6521 |                   19.6521 |       999.3479 |    1.18911 | 0.0 | nan | 0.0 |             78.509149 |                      0.0 |                 78.509149 |                  78.509149 |                  31.34697 |                     0.0 |                      0.0 |    47.162179 |             0.0 |              0.0 |         47.162179 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.808559 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |                14.836934 |      227.962116 |             75.6 |       2.7 |              1015.0 | 0.02 | 1.728 |    21.7 |                  75.6 |                  10.2 |                     14.836934 |           227.962116 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 26.00118 |                      26.00118 |                         0.0 |                          0.0 |                        1.5885 |                             1.5885 |            19.656892 |                 19.656892 |     995.343108 |   1.190432 | 0.0 | nan | 0.0 |             68.388635 |                      0.0 |                 68.388635 |                  68.388635 |                 25.918437 |                     0.0 |                      0.0 |    42.470198 |             0.0 |              0.0 |         42.470198 |             78.890988 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.672736 | 1.672736 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.055264 |  0.00064 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |  0.00064 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |                  14.7677 |      216.356321 |             81.4 |       3.2 |              1016.0 | 0.02 | 1.728 |    20.8 |                  81.4 |                   9.4 |                       14.7677 |           216.356321 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                24.605348 |                     24.605348 |                         0.0 |                          0.0 |                      1.513854 |                           1.513854 |            20.028753 |                 20.028753 |     995.971247 |   1.195095 | 0.0 | nan | 0.0 |             64.906896 |                      0.0 |                 64.906896 |                  64.906896 |                 23.397902 |                     0.0 |                      0.0 |    41.508994 |             0.0 |              0.0 |         41.508994 |             66.564271 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.529542 | 1.529542 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  2.736458 | 0.031672 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.031672 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |                14.698075 |      226.637904 |             79.0 |       1.3 |              1019.0 | 0.02 | 1.728 |    19.5 |                  79.0 |                  10.3 |                     14.698075 |           226.637904 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                22.704676 |                     22.704676 |                         0.0 |                          0.0 |                      1.411269 |                           1.411269 |            17.936694 |                 17.936694 |    1001.063306 |   1.204916 | 0.0 | nan | 0.0 |             67.991371 |                      0.0 |                 67.991371 |                  67.991371 |                 28.794151 |                     0.0 |                      0.0 |    39.197221 |             0.0 |              0.0 |         39.197221 |            163.850513 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.283901 | 1.283901 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.444099 |  0.00514 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |  0.00514 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |                14.628078 |      235.528503 |             75.9 |       1.9 |              1019.0 | 0.02 | 1.728 |    21.0 |                  75.9 |                  11.1 |                     14.628078 |           235.528503 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                24.909749 |                     24.909749 |                         0.0 |                          0.0 |                      1.530181 |                           1.530181 |            18.906499 |                 18.906499 |    1000.093501 |   1.198338 | 0.0 | nan | 0.0 |             70.658551 |                      0.0 |                 70.658551 |                  70.658551 |                 29.177702 |                     0.0 |                      0.0 |    41.480849 |             0.0 |              0.0 |         41.480849 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.510743 | 1.510743 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.217257 | 0.002515 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.002515 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |                14.557728 |      232.849428 |             70.0 |       2.0 |              1018.0 | 0.02 | 1.728 |    21.6 |                  70.0 |                  11.0 |                     14.557728 |           232.849428 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                25.842752 |                     25.842752 |                         0.0 |                          0.0 |                      1.580055 |                           1.580055 |            18.089927 |                 18.089927 |     999.910073 |   1.195081 | 0.0 | nan | 0.0 |             69.854829 |                      0.0 |                 69.854829 |                  69.854829 |                 31.281288 |                     0.0 |                      0.0 |    38.573541 |             0.0 |              0.0 |         38.573541 |            106.502833 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.594572 | 1.594572 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.133428 | 0.001544 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.001544 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |                14.487046 |       199.71349 |             70.3 |       2.6 |              1019.0 | 0.02 | 1.728 |    21.1 |                  70.3 |                   8.5 |                     14.487046 |            199.71349 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                25.063177 |                     25.063177 |                         0.0 |                          0.0 |                        1.5384 |                             1.5384 |            17.619414 |                 17.619414 |    1001.380586 |   1.198506 | 0.0 | nan | 0.0 |             59.914047 |                      0.0 |                 59.914047 |                  59.914047 |                 26.711296 |                     0.0 |                      0.0 |    33.202751 |             0.0 |              0.0 |         33.202751 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.541567 | 1.541567 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.186433 | 0.002158 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.002158 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |                14.416048 |      233.771902 |             69.0 |       3.6 |              1019.0 | 0.02 | 1.728 |    22.4 |                  69.0 |                  11.3 |                     14.416048 |           233.771902 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                27.134082 |                     27.134082 |                         0.0 |                          0.0 |                      1.648679 |                           1.648679 |            18.722517 |                 18.722517 |    1000.277483 |   1.192743 | 0.0 | nan | 0.0 |             70.131571 |                      0.0 |                 70.131571 |                  70.131571 |                 31.116622 |                     0.0 |                      0.0 |    39.014949 |             0.0 |              0.0 |         39.014949 |             59.168241 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.939383 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |                14.344752 |      246.121286 |             62.1 |       2.5 |              1017.0 | 0.02 | 1.728 |    22.5 |                  62.1 |                  12.4 |                     14.344752 |           246.121286 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                27.299387 |                     27.299387 |                         0.0 |                          0.0 |                      1.657431 |                           1.657431 |             16.95292 |                  16.95292 |     1000.04708 |   1.190771 | 0.0 | nan | 0.0 |             73.836386 |                      0.0 |                 73.836386 |                  73.836386 |                 38.297604 |                     0.0 |                      0.0 |    35.538782 |             0.0 |              0.0 |         35.538782 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.826811 |    1.728 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |       0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |      0.0 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |                14.273175 |        90.58551 |             84.5 |       2.1 |              1015.0 | 0.02 | 1.728 |    19.6 |                  84.5 |                   0.1 |                     14.273175 |             90.58551 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                22.846186 |                     22.846186 |                         0.0 |                          0.0 |                      1.418946 |                           1.418946 |            19.305027 |                 19.305027 |     995.694973 |   1.199129 | 0.0 | nan | 0.0 |             27.175653 |                      0.0 |                 27.175653 |                  27.175653 |                   7.02118 |                     0.0 |                      0.0 |    20.154473 |             0.0 |              0.0 |         20.154473 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.806188 | 0.806188 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  2.801812 | 0.032428 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.032428 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |                14.201333 |      171.944993 |             83.6 |       1.5 |              1015.0 | 0.02 | 1.728 |    20.3 |                  83.6 |                   6.7 |                     14.201333 |           171.944993 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                23.858503 |                     23.858503 |                         0.0 |                          0.0 |                      1.473678 |                           1.473678 |            19.945708 |                 19.945708 |     995.054292 |   1.195981 | 0.0 | nan | 0.0 |             51.583498 |                      0.0 |                 51.583498 |                  51.583498 |                 18.994889 |                     0.0 |                      0.0 |    32.588609 |             0.0 |              0.0 |         32.588609 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.08421 |  1.08421 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |   0.92579 | 0.010715 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.010715 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |                14.129243 |      216.549813 |             76.5 |       2.2 |              1011.0 | 0.02 | 1.728 |    24.1 |                  76.5 |                  10.4 |                     14.129243 |           216.549813 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                30.066256 |                     30.066256 |                         0.0 |                          0.0 |                       1.80287 |                            1.80287 |            23.000686 |                 23.000686 |     987.999314 |   1.174651 | 0.0 | nan | 0.0 |             64.964944 |                      0.0 |                 64.964944 |                  64.964944 |                 21.629737 |                     0.0 |                      0.0 |    43.335207 |             0.0 |              0.0 |         43.335207 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.672232 | 1.672232 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.055768 | 0.000645 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.000645 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |                14.056918 |       166.91683 |             76.6 |       1.9 |              1008.0 | 0.02 | 1.728 |    23.7 |                  76.6 |                   6.5 |                     14.056918 |            166.91683 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                29.352596 |                     29.352596 |                         0.0 |                          0.0 |                      1.765541 |                           1.765541 |            22.484088 |                 22.484088 |     985.515912 |   1.172942 | 0.0 | nan | 0.0 |             50.075049 |                      0.0 |                 50.075049 |                  50.075049 |                 16.222254 |                     0.0 |                      0.0 |    33.852795 |             0.0 |              0.0 |         33.852795 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.374102 | 1.374102 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.447898 | 0.005184 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.005184 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |                13.984374 |      145.985611 |             84.2 |       2.3 |              1006.0 | 0.02 | 1.728 |    21.1 |                  84.2 |                   4.9 |                     13.984374 |           145.985611 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                25.063177 |                     25.063177 |                         0.0 |                          0.0 |                        1.5384 |                             1.5384 |            21.103195 |                 21.103195 |     984.896805 |   1.181557 | 0.0 | nan | 0.0 |             43.795683 |                      0.0 |                 43.795683 |                  43.795683 |                 14.602883 |                     0.0 |                      0.0 |      29.1928 |             0.0 |              0.0 |           29.1928 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.077096 | 1.077096 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 10.520904 |  0.12177 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 |  0.12177 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |                13.911624 |       165.54389 |             85.8 |       2.1 |              1005.0 | 0.02 | 1.728 |    20.9 |                  85.8 |                   6.6 |                     13.911624 |            165.54389 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 24.75714 |                      24.75714 |                         0.0 |                          0.0 |                      1.521999 |                           1.521999 |            21.241626 |                 21.241626 |     983.758374 |   1.181114 | 0.0 | nan | 0.0 |             49.663167 |                      0.0 |                 49.663167 |                  49.663167 |                 17.349121 |                     0.0 |                      0.0 |    32.314046 |             0.0 |              0.0 |         32.314046 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.09756 |  1.09756 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |   6.74044 | 0.078014 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.078014 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |                13.838682 |       87.693346 |             86.5 |       2.6 |              1000.0 | 0.02 | 1.728 |    20.2 |                  86.5 |                   0.3 |                     13.838682 |            87.693346 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                23.711531 |                     23.711531 |                         0.0 |                          0.0 |                      1.465752 |                           1.465752 |            20.510475 |                 20.510475 |     979.489525 |   1.178323 | 0.0 | nan | 0.0 |             26.308004 |                      0.0 |                 26.308004 |                  26.308004 |                  6.818024 |                     0.0 |                      0.0 |     19.48998 |             0.0 |              0.0 |          19.48998 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.794859 | 0.794859 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 15.033141 | 0.173995 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.173995 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |                13.765561 |       84.356885 |             89.6 |       2.6 |              1004.0 | 0.02 | 1.728 |    16.1 |                  89.6 |                   0.1 |                     13.765561 |            84.356885 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                18.326791 |                     18.326791 |                         0.0 |                          0.0 |                      1.170311 |                           1.170311 |            16.420805 |                 16.420805 |     987.579195 |   1.201704 | 0.0 | nan | 0.0 |             25.307066 |                      0.0 |                 25.307066 |                  25.307066 |                  8.199212 |                     0.0 |                      0.0 |    17.107854 |             0.0 |              0.0 |         17.107854 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.607776 | 0.607776 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  9.768224 | 0.113058 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.113058 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |                13.692273 |      142.309202 |             78.3 |       2.1 |              1011.0 | 0.02 | 1.728 |    17.0 |                  78.3 |                   5.0 |                     13.692273 |           142.309202 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                19.406929 |                     19.406929 |                         0.0 |                          0.0 |                      1.230421 |                           1.230421 |            15.195625 |                 15.195625 |     995.804375 |   1.206937 | 0.0 | nan | 0.0 |             42.692761 |                      0.0 |                 42.692761 |                  42.692761 |                 21.775852 |                     0.0 |                      0.0 |    20.916908 |             0.0 |              0.0 |         20.916908 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.90065 |  0.90065 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |   0.82735 | 0.009576 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.009576 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |                13.618831 |      126.722078 |             78.5 |       2.1 |              1014.0 | 0.02 | 1.728 |    18.2 |                  78.5 |                   3.8 |                     13.618831 |           126.722078 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                20.933479 |                     20.933479 |                         0.0 |                          0.0 |                      1.314615 |                           1.314615 |            16.432781 |                 16.432781 |     997.567219 |   1.204994 | 0.0 | nan | 0.0 |             38.016624 |                      0.0 |                 38.016624 |                  38.016624 |                 17.491881 |                     0.0 |                      0.0 |    20.524743 |             0.0 |              0.0 |         20.524743 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.912078 | 0.912078 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.815922 | 0.009444 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02 | 0.009444 |

>>> round_(model.check_waterbalance(conditions))
0.0

.. _lland_v4_wasser_negq_daily:

water (negative runoff)
-----------------------

.. integration-test::

    >>> negq(True)
    >>> inlet.sequences.sim.series = 0.0
    >>> test("lland_v4_wasser_negq_daily",
    ...      axis1=(fluxes.nkor, fluxes.evi, fluxes.qah))
    |       date | nied | teml | sunshineduration | possiblesunshineduration | globalradiation | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration | dailypossiblesunshineduration | dailyglobalradiation |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair |   g |  tz |  wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |     evpo |      evi | evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatinz | wlatsnow | wsensinz | wsenssnow | wsurfinz | wsurf | sff | fvg | wada | qdb | qib1 | qib2 | qbb | qkap | qdgz |       qah |        qa | inzp | stinz | sinz | esnowinz | asinz | wats | waes | esnow | taus | ebdn | bowa | qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz | qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet |    outlet |
    --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |                15.690785 |       190.25149 |             86.2 |       2.4 |              1007.0 | 0.0 | 0.0 |    16.1 |                  86.2 |                   6.3 |                     15.690785 |            190.25149 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                18.326791 |                     18.326791 |                         0.0 |                          0.0 |                      1.170311 |                           1.170311 |            15.797694 |                 15.797694 |     991.202306 |   1.205601 | 0.0 | nan | 0.0 |             57.075447 |                      0.0 |                 57.075447 |                  57.075447 |                 21.732733 |                     0.0 |                      0.0 |    35.342714 |             0.0 |              0.0 |         35.342714 |             88.752361 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.077197 | 1.077197 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 11.424803 |  0.132232 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.132232 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |                15.628821 |      127.660782 |             85.8 |       1.6 |              1013.0 | 0.0 | 0.0 |    17.3 |                  85.8 |                   1.7 |                     15.628821 |           127.660782 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                19.779139 |                     19.779139 |                         0.0 |                          0.0 |                      1.251029 |                           1.251029 |            16.970501 |                 16.970501 |     996.029499 |   1.207284 | 0.0 | nan | 0.0 |             38.298235 |                      0.0 |                 38.298235 |                  38.298235 |                 11.219797 |                     0.0 |                      0.0 |    27.078438 |             0.0 |              0.0 |         27.078438 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.86545 |  0.86545 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  -0.77145 | -0.008929 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.008929 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |                15.566134 |      164.401298 |             82.4 |       1.5 |              1016.0 | 0.0 | 0.0 |    18.9 |                  82.4 |                   4.5 |                     15.566134 |           164.401298 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                21.871576 |                     21.871576 |                         0.0 |                          0.0 |                      1.365939 |                           1.365939 |            18.022179 |                 18.022179 |     997.977821 |   1.203775 | 0.0 | nan | 0.0 |             49.320389 |                      0.0 |                 49.320389 |                  49.320389 |                 16.044181 |                     0.0 |                      0.0 |    33.276209 |             0.0 |              0.0 |         33.276209 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.092365 | 1.092365 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.092365 | -0.012643 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.012643 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |                15.502754 |      269.020502 |             77.6 |       2.8 |              1017.0 | 0.0 | 0.0 |    17.8 |                  77.6 |                  12.4 |                     15.502754 |           269.020502 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                20.413369 |                     20.413369 |                         0.0 |                          0.0 |                      1.286025 |                           1.286025 |            15.840774 |                 15.840774 |    1001.159226 |    1.21051 | 0.0 | nan | 0.0 |              80.70615 |                      0.0 |                  80.70615 |                   80.70615 |                 35.980407 |                     0.0 |                      0.0 |    44.725744 |             0.0 |              0.0 |         44.725744 |             76.073452 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.565894 | 1.565894 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.471894 | -0.017036 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.017036 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |                15.438712 |      287.848667 |             74.9 |       3.0 |              1018.0 | 0.0 | 0.0 |    18.5 |                  74.9 |                  13.9 |                     15.438712 |           287.848667 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                21.331125 |                     21.331125 |                         0.0 |                          0.0 |                      1.336407 |                           1.336407 |            15.977013 |                 15.977013 |    1002.022987 |   1.208738 | 0.0 | nan | 0.0 |               86.3546 |                      0.0 |                   86.3546 |                    86.3546 |                 39.523937 |                     0.0 |                      0.0 |    46.830664 |             0.0 |              0.0 |         46.830664 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.729733 | 1.729733 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.729733 |  -0.02002 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  -0.02002 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |                15.374034 |      274.674263 |             77.9 |       3.5 |              1018.0 | 0.0 | 0.0 |    19.9 |                  77.9 |                  13.0 |                     15.374034 |           274.674263 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                23.275348 |                     23.275348 |                         0.0 |                          0.0 |                      1.442189 |                           1.442189 |            18.131496 |                 18.131496 |     999.868504 |   1.201995 | 0.0 | nan | 0.0 |             82.402279 |                      0.0 |                 82.402279 |                  82.402279 |                 32.936813 |                     0.0 |                      0.0 |    49.465466 |             0.0 |              0.0 |         49.465466 |             60.858762 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.826087 | 1.826087 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.826087 | -0.021135 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.021135 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |                 15.30875 |      284.047892 |             73.5 |       3.8 |              1019.0 | 0.0 | 0.0 |    20.8 |                  73.5 |                  13.8 |                      15.30875 |           284.047892 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                24.605348 |                     24.605348 |                         0.0 |                          0.0 |                      1.513854 |                           1.513854 |            18.084931 |                 18.084931 |    1000.915069 |   1.199521 | 0.0 | nan | 0.0 |             85.214368 |                      0.0 |                 85.214368 |                  85.214368 |                  35.31351 |                     0.0 |                      0.0 |    49.900858 |             0.0 |              0.0 |         49.900858 |             56.054123 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.041956 | 2.041956 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -2.041956 | -0.023634 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.023634 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |                15.242885 |      262.971263 |             80.1 |       3.3 |              1019.0 | 0.0 | 0.0 |    20.9 |                  80.1 |                  12.3 |                     15.242885 |           262.971263 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 24.75714 |                      24.75714 |                         0.0 |                          0.0 |                      1.521999 |                           1.521999 |            19.830469 |                 19.830469 |     999.169531 |   1.198331 | 0.0 | nan | 0.0 |             78.891379 |                      0.0 |                 78.891379 |                  78.891379 |                 28.392598 |                     0.0 |                      0.0 |    50.498781 |             0.0 |              0.0 |         50.498781 |             64.547172 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.800053 | 1.800053 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.800053 | -0.020834 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.020834 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |                15.176467 |      272.256414 |             72.9 |       1.8 |              1019.0 | 0.0 | 0.0 |    22.3 |                  72.9 |                  13.1 |                     15.176467 |           272.256414 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 26.96965 |                      26.96965 |                         0.0 |                          0.0 |                      1.639967 |                           1.639967 |            19.660875 |                 19.660875 |     999.339125 |   1.192729 | 0.0 | nan | 0.0 |             81.676924 |                      0.0 |                 81.676924 |                  81.676924 |                 31.114889 |                     0.0 |                      0.0 |    50.562035 |             0.0 |              0.0 |         50.562035 |            118.336481 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.82186 |  1.82186 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  -1.82186 | -0.021086 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.021086 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |                15.109521 |      267.019218 |             68.5 |       1.9 |              1020.0 | 0.0 | 0.0 |    23.0 |                  68.5 |                  12.8 |                     15.109521 |           267.019218 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                28.139149 |                     28.139149 |                         0.0 |                          0.0 |                      1.701779 |                           1.701779 |            19.275317 |                 19.275317 |    1000.724683 |   1.191257 | 0.0 | nan | 0.0 |             80.105765 |                      0.0 |                 80.105765 |                  80.105765 |                 32.010643 |                     0.0 |                      0.0 |    48.095123 |             0.0 |              0.0 |         48.095123 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.891469 | 1.891469 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.891469 | -0.021892 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.021892 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |                15.042071 |      269.626547 |             68.6 |       3.0 |              1021.0 | 0.0 | 0.0 |    23.1 |                  68.6 |                  13.1 |                     15.042071 |           269.626547 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                28.309776 |                     28.309776 |                         0.0 |                          0.0 |                      1.710768 |                           1.710768 |            19.420506 |                 19.420506 |    1001.579494 |   1.191966 | 0.0 | nan | 0.0 |             80.887964 |                      0.0 |                 80.887964 |                  80.887964 |                 32.400014 |                     0.0 |                      0.0 |     48.48795 |             0.0 |              0.0 |          48.48795 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.106575 | 2.106575 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -2.106575 | -0.024382 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.024382 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |                14.974141 |      270.878291 |             66.0 |       2.5 |              1022.0 | 0.0 | 0.0 |    23.2 |                  66.0 |                  13.3 |                     14.974141 |           270.878291 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                28.481304 |                     28.481304 |                         0.0 |                          0.0 |                      1.719796 |                           1.719796 |             18.79766 |                  18.79766 |     1003.20234 |   1.193016 | 0.0 | nan | 0.0 |             81.263487 |                      0.0 |                 81.263487 |                  81.263487 |                 34.565783 |                     0.0 |                      0.0 |    46.697704 |             0.0 |              0.0 |         46.697704 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.036929 | 2.036929 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.942929 | -0.022488 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.022488 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |                14.905755 |      261.697164 |             69.0 |       1.6 |              1019.0 | 0.0 | 0.0 |    23.2 |                  69.0 |                  12.7 |                     14.905755 |           261.697164 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                28.481304 |                     28.481304 |                         0.0 |                          0.0 |                      1.719796 |                           1.719796 |              19.6521 |                   19.6521 |       999.3479 |    1.18911 | 0.0 | nan | 0.0 |             78.509149 |                      0.0 |                 78.509149 |                  78.509149 |                  31.34697 |                     0.0 |                      0.0 |    47.162179 |             0.0 |              0.0 |         47.162179 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.808559 | 1.808559 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.808559 | -0.020932 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.020932 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |                14.836934 |      227.962116 |             75.6 |       2.7 |              1015.0 | 0.0 | 0.0 |    21.7 |                  75.6 |                  10.2 |                     14.836934 |           227.962116 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 26.00118 |                      26.00118 |                         0.0 |                          0.0 |                        1.5885 |                             1.5885 |            19.656892 |                 19.656892 |     995.343108 |   1.190432 | 0.0 | nan | 0.0 |             68.388635 |                      0.0 |                 68.388635 |                  68.388635 |                 25.918437 |                     0.0 |                      0.0 |    42.470198 |             0.0 |              0.0 |         42.470198 |             78.890988 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.672736 | 1.672736 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.672736 |  -0.01936 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  -0.01936 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |                  14.7677 |      216.356321 |             81.4 |       3.2 |              1016.0 | 0.0 | 0.0 |    20.8 |                  81.4 |                   9.4 |                       14.7677 |           216.356321 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                24.605348 |                     24.605348 |                         0.0 |                          0.0 |                      1.513854 |                           1.513854 |            20.028753 |                 20.028753 |     995.971247 |   1.195095 | 0.0 | nan | 0.0 |             64.906896 |                      0.0 |                 64.906896 |                  64.906896 |                 23.397902 |                     0.0 |                      0.0 |    41.508994 |             0.0 |              0.0 |         41.508994 |             66.564271 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.529542 | 1.529542 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  1.008458 |  0.011672 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.011672 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |                14.698075 |      226.637904 |             79.0 |       1.3 |              1019.0 | 0.0 | 0.0 |    19.5 |                  79.0 |                  10.3 |                     14.698075 |           226.637904 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                22.704676 |                     22.704676 |                         0.0 |                          0.0 |                      1.411269 |                           1.411269 |            17.936694 |                 17.936694 |    1001.063306 |   1.204916 | 0.0 | nan | 0.0 |             67.991371 |                      0.0 |                 67.991371 |                  67.991371 |                 28.794151 |                     0.0 |                      0.0 |    39.197221 |             0.0 |              0.0 |         39.197221 |            163.850513 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.283901 | 1.283901 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.283901 |  -0.01486 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  -0.01486 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |                14.628078 |      235.528503 |             75.9 |       1.9 |              1019.0 | 0.0 | 0.0 |    21.0 |                  75.9 |                  11.1 |                     14.628078 |           235.528503 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                24.909749 |                     24.909749 |                         0.0 |                          0.0 |                      1.530181 |                           1.530181 |            18.906499 |                 18.906499 |    1000.093501 |   1.198338 | 0.0 | nan | 0.0 |             70.658551 |                      0.0 |                 70.658551 |                  70.658551 |                 29.177702 |                     0.0 |                      0.0 |    41.480849 |             0.0 |              0.0 |         41.480849 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.510743 | 1.510743 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.510743 | -0.017485 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.017485 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |                14.557728 |      232.849428 |             70.0 |       2.0 |              1018.0 | 0.0 | 0.0 |    21.6 |                  70.0 |                  11.0 |                     14.557728 |           232.849428 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                25.842752 |                     25.842752 |                         0.0 |                          0.0 |                      1.580055 |                           1.580055 |            18.089927 |                 18.089927 |     999.910073 |   1.195081 | 0.0 | nan | 0.0 |             69.854829 |                      0.0 |                 69.854829 |                  69.854829 |                 31.281288 |                     0.0 |                      0.0 |    38.573541 |             0.0 |              0.0 |         38.573541 |            106.502833 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.594572 | 1.594572 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.594572 | -0.018456 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.018456 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |                14.487046 |       199.71349 |             70.3 |       2.6 |              1019.0 | 0.0 | 0.0 |    21.1 |                  70.3 |                   8.5 |                     14.487046 |            199.71349 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                25.063177 |                     25.063177 |                         0.0 |                          0.0 |                        1.5384 |                             1.5384 |            17.619414 |                 17.619414 |    1001.380586 |   1.198506 | 0.0 | nan | 0.0 |             59.914047 |                      0.0 |                 59.914047 |                  59.914047 |                 26.711296 |                     0.0 |                      0.0 |    33.202751 |             0.0 |              0.0 |         33.202751 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.541567 | 1.541567 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.541567 | -0.017842 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.017842 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |                14.416048 |      233.771902 |             69.0 |       3.6 |              1019.0 | 0.0 | 0.0 |    22.4 |                  69.0 |                  11.3 |                     14.416048 |           233.771902 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                27.134082 |                     27.134082 |                         0.0 |                          0.0 |                      1.648679 |                           1.648679 |            18.722517 |                 18.722517 |    1000.277483 |   1.192743 | 0.0 | nan | 0.0 |             70.131571 |                      0.0 |                 70.131571 |                  70.131571 |                 31.116622 |                     0.0 |                      0.0 |    39.014949 |             0.0 |              0.0 |         39.014949 |             59.168241 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.939383 | 1.939383 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.939383 | -0.022447 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.022447 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |                14.344752 |      246.121286 |             62.1 |       2.5 |              1017.0 | 0.0 | 0.0 |    22.5 |                  62.1 |                  12.4 |                     14.344752 |           246.121286 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                27.299387 |                     27.299387 |                         0.0 |                          0.0 |                      1.657431 |                           1.657431 |             16.95292 |                  16.95292 |     1000.04708 |   1.190771 | 0.0 | nan | 0.0 |             73.836386 |                      0.0 |                 73.836386 |                  73.836386 |                 38.297604 |                     0.0 |                      0.0 |    35.538782 |             0.0 |              0.0 |         35.538782 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.826811 | 1.826811 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.826811 | -0.021144 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.021144 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |                14.273175 |        90.58551 |             84.5 |       2.1 |              1015.0 | 0.0 | 0.0 |    19.6 |                  84.5 |                   0.1 |                     14.273175 |             90.58551 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                22.846186 |                     22.846186 |                         0.0 |                          0.0 |                      1.418946 |                           1.418946 |            19.305027 |                 19.305027 |     995.694973 |   1.199129 | 0.0 | nan | 0.0 |             27.175653 |                      0.0 |                 27.175653 |                  27.175653 |                   7.02118 |                     0.0 |                      0.0 |    20.154473 |             0.0 |              0.0 |         20.154473 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.806188 | 0.806188 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  1.073812 |  0.012428 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.012428 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |                14.201333 |      171.944993 |             83.6 |       1.5 |              1015.0 | 0.0 | 0.0 |    20.3 |                  83.6 |                   6.7 |                     14.201333 |           171.944993 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                23.858503 |                     23.858503 |                         0.0 |                          0.0 |                      1.473678 |                           1.473678 |            19.945708 |                 19.945708 |     995.054292 |   1.195981 | 0.0 | nan | 0.0 |             51.583498 |                      0.0 |                 51.583498 |                  51.583498 |                 18.994889 |                     0.0 |                      0.0 |    32.588609 |             0.0 |              0.0 |         32.588609 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.08421 |  1.08421 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  -0.80221 | -0.009285 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.009285 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |                14.129243 |      216.549813 |             76.5 |       2.2 |              1011.0 | 0.0 | 0.0 |    24.1 |                  76.5 |                  10.4 |                     14.129243 |           216.549813 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                30.066256 |                     30.066256 |                         0.0 |                          0.0 |                       1.80287 |                            1.80287 |            23.000686 |                 23.000686 |     987.999314 |   1.174651 | 0.0 | nan | 0.0 |             64.964944 |                      0.0 |                 64.964944 |                  64.964944 |                 21.629737 |                     0.0 |                      0.0 |    43.335207 |             0.0 |              0.0 |         43.335207 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.672232 | 1.672232 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.672232 | -0.019355 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.019355 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |                14.056918 |       166.91683 |             76.6 |       1.9 |              1008.0 | 0.0 | 0.0 |    23.7 |                  76.6 |                   6.5 |                     14.056918 |            166.91683 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                29.352596 |                     29.352596 |                         0.0 |                          0.0 |                      1.765541 |                           1.765541 |            22.484088 |                 22.484088 |     985.515912 |   1.172942 | 0.0 | nan | 0.0 |             50.075049 |                      0.0 |                 50.075049 |                  50.075049 |                 16.222254 |                     0.0 |                      0.0 |    33.852795 |             0.0 |              0.0 |         33.852795 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.374102 | 1.374102 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -1.280102 | -0.014816 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.014816 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |                13.984374 |      145.985611 |             84.2 |       2.3 |              1006.0 | 0.0 | 0.0 |    21.1 |                  84.2 |                   4.9 |                     13.984374 |           145.985611 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                25.063177 |                     25.063177 |                         0.0 |                          0.0 |                        1.5384 |                             1.5384 |            21.103195 |                 21.103195 |     984.896805 |   1.181557 | 0.0 | nan | 0.0 |             43.795683 |                      0.0 |                 43.795683 |                  43.795683 |                 14.602883 |                     0.0 |                      0.0 |      29.1928 |             0.0 |              0.0 |           29.1928 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.077096 | 1.077096 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  8.792904 |   0.10177 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.10177 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |                13.911624 |       165.54389 |             85.8 |       2.1 |              1005.0 | 0.0 | 0.0 |    20.9 |                  85.8 |                   6.6 |                     13.911624 |            165.54389 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 24.75714 |                      24.75714 |                         0.0 |                          0.0 |                      1.521999 |                           1.521999 |            21.241626 |                 21.241626 |     983.758374 |   1.181114 | 0.0 | nan | 0.0 |             49.663167 |                      0.0 |                 49.663167 |                  49.663167 |                 17.349121 |                     0.0 |                      0.0 |    32.314046 |             0.0 |              0.0 |         32.314046 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.09756 |  1.09756 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |   5.01244 |  0.058014 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.058014 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |                13.838682 |       87.693346 |             86.5 |       2.6 |              1000.0 | 0.0 | 0.0 |    20.2 |                  86.5 |                   0.3 |                     13.838682 |            87.693346 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                23.711531 |                     23.711531 |                         0.0 |                          0.0 |                      1.465752 |                           1.465752 |            20.510475 |                 20.510475 |     979.489525 |   1.178323 | 0.0 | nan | 0.0 |             26.308004 |                      0.0 |                 26.308004 |                  26.308004 |                  6.818024 |                     0.0 |                      0.0 |     19.48998 |             0.0 |              0.0 |          19.48998 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.794859 | 0.794859 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 13.305141 |  0.153995 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.153995 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |                13.765561 |       84.356885 |             89.6 |       2.6 |              1004.0 | 0.0 | 0.0 |    16.1 |                  89.6 |                   0.1 |                     13.765561 |            84.356885 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                18.326791 |                     18.326791 |                         0.0 |                          0.0 |                      1.170311 |                           1.170311 |            16.420805 |                 16.420805 |     987.579195 |   1.201704 | 0.0 | nan | 0.0 |             25.307066 |                      0.0 |                 25.307066 |                  25.307066 |                  8.199212 |                     0.0 |                      0.0 |    17.107854 |             0.0 |              0.0 |         17.107854 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.607776 | 0.607776 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  8.040224 |  0.093058 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.093058 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |                13.692273 |      142.309202 |             78.3 |       2.1 |              1011.0 | 0.0 | 0.0 |    17.0 |                  78.3 |                   5.0 |                     13.692273 |           142.309202 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                19.406929 |                     19.406929 |                         0.0 |                          0.0 |                      1.230421 |                           1.230421 |            15.195625 |                 15.195625 |     995.804375 |   1.206937 | 0.0 | nan | 0.0 |             42.692761 |                      0.0 |                 42.692761 |                  42.692761 |                 21.775852 |                     0.0 |                      0.0 |    20.916908 |             0.0 |              0.0 |         20.916908 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.90065 |  0.90065 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  -0.90065 | -0.010424 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.010424 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |                13.618831 |      126.722078 |             78.5 |       2.1 |              1014.0 | 0.0 | 0.0 |    18.2 |                  78.5 |                   3.8 |                     13.618831 |           126.722078 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                20.933479 |                     20.933479 |                         0.0 |                          0.0 |                      1.314615 |                           1.314615 |            16.432781 |                 16.432781 |     997.567219 |   1.204994 | 0.0 | nan | 0.0 |             38.016624 |                      0.0 |                 38.016624 |                  38.016624 |                 17.491881 |                     0.0 |                      0.0 |    20.524743 |             0.0 |              0.0 |         20.524743 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.912078 | 0.912078 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | -0.912078 | -0.010556 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | -0.010556 |

>>> round_(model.check_waterbalance(conditions))
0.0

.. _lland_v4_see_daily:

lakes
-----

.. integration-test::

    >>> lnk(SEE)
    >>> surfaceresistance.see_aug = 0.0
    >>> cropheight.see_aug = 0.05
    >>> albedo.see_aug = 0.7
    >>> negq(False)
    >>> test("lland_v4_see_daily",
    ...      axis1=(fluxes.nkor, fluxes.evi, fluxes.qah))
    |       date | nied | teml | sunshineduration | possiblesunshineduration | globalradiation | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration | dailypossiblesunshineduration | dailyglobalradiation |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair |   g |  tz |  wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |     evpo |      evi | evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatinz | wlatsnow | wsensinz | wsenssnow | wsurfinz | wsurf | sff | fvg | wada | qdb | qib1 | qib2 | qbb | qkap | qdgz |      qah |       qa | inzp | stinz | sinz | esnowinz | asinz | wats | waes | esnow | taus | ebdn | bowa | qdgz1 | qdgz2 | qigz1 | qigz2 |      qbgz | qdga1 | qdga2 | qiga1 | qiga2 |      qbga | inlet |   outlet |
    ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |                15.690785 |       190.25149 |             86.2 |       2.4 |              1007.0 | 0.0 | 0.0 |    16.1 |                  86.2 |                   6.3 |                     15.690785 |            190.25149 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                18.326791 |                     18.326791 |                         0.0 |                          0.0 |                      1.170311 |                           1.170311 |            15.797694 |                 15.797694 |     991.202306 |   1.205601 | 0.0 | nan | 0.0 |             57.075447 |                      0.0 |                 57.075447 |                  57.075447 |                 21.732733 |                     0.0 |                      0.0 |    35.342714 |             0.0 |              0.0 |         35.342714 |             88.752361 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.077197 | 1.077197 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.020478 | 0.000237 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | 11.424803 |   0.0 |   0.0 |   0.0 |   0.0 |  0.020478 |   0.0 | 0.000237 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |                15.628821 |      127.660782 |             85.8 |       1.6 |              1013.0 | 0.0 | 0.0 |    17.3 |                  85.8 |                   1.7 |                     15.628821 |           127.660782 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                19.779139 |                     19.779139 |                         0.0 |                          0.0 |                      1.251029 |                           1.251029 |            16.970501 |                 16.970501 |     996.029499 |   1.207284 | 0.0 | nan | 0.0 |             38.298235 |                      0.0 |                 38.298235 |                  38.298235 |                 11.219797 |                     0.0 |                      0.0 |    27.078438 |             0.0 |              0.0 |         27.078438 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.86545 |  0.86545 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.039475 | 0.000457 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  -0.77145 |   0.0 |   0.0 |   0.0 |   0.0 |  0.039475 |   0.0 | 0.000457 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |                15.566134 |      164.401298 |             82.4 |       1.5 |              1016.0 | 0.0 | 0.0 |    18.9 |                  82.4 |                   4.5 |                     15.566134 |           164.401298 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                21.871576 |                     21.871576 |                         0.0 |                          0.0 |                      1.365939 |                           1.365939 |            18.022179 |                 18.022179 |     997.977821 |   1.203775 | 0.0 | nan | 0.0 |             49.320389 |                      0.0 |                 49.320389 |                  49.320389 |                 16.044181 |                     0.0 |                      0.0 |    33.276209 |             0.0 |              0.0 |         33.276209 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.092365 | 1.092365 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.035994 | 0.000417 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.092365 |   0.0 |   0.0 |   0.0 |   0.0 |  0.035994 |   0.0 | 0.000417 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |                15.502754 |      269.020502 |             77.6 |       2.8 |              1017.0 | 0.0 | 0.0 |    17.8 |                  77.6 |                  12.4 |                     15.502754 |           269.020502 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                20.413369 |                     20.413369 |                         0.0 |                          0.0 |                      1.286025 |                           1.286025 |            15.840774 |                 15.840774 |    1001.159226 |    1.21051 | 0.0 | nan | 0.0 |              80.70615 |                      0.0 |                  80.70615 |                   80.70615 |                 35.980407 |                     0.0 |                      0.0 |    44.725744 |             0.0 |              0.0 |         44.725744 |             76.073452 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.565894 | 1.565894 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.031272 | 0.000362 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.471894 |   0.0 |   0.0 |   0.0 |   0.0 |  0.031272 |   0.0 | 0.000362 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |                15.438712 |      287.848667 |             74.9 |       3.0 |              1018.0 | 0.0 | 0.0 |    18.5 |                  74.9 |                  13.9 |                     15.438712 |           287.848667 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                21.331125 |                     21.331125 |                         0.0 |                          0.0 |                      1.336407 |                           1.336407 |            15.977013 |                 15.977013 |    1002.022987 |   1.208738 | 0.0 | nan | 0.0 |               86.3546 |                      0.0 |                   86.3546 |                    86.3546 |                 39.523937 |                     0.0 |                      0.0 |    46.830664 |             0.0 |              0.0 |         46.830664 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.729733 | 1.729733 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.025424 | 0.000294 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.729733 |   0.0 |   0.0 |   0.0 |   0.0 |  0.025424 |   0.0 | 0.000294 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |                15.374034 |      274.674263 |             77.9 |       3.5 |              1018.0 | 0.0 | 0.0 |    19.9 |                  77.9 |                  13.0 |                     15.374034 |           274.674263 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                23.275348 |                     23.275348 |                         0.0 |                          0.0 |                      1.442189 |                           1.442189 |            18.131496 |                 18.131496 |     999.868504 |   1.201995 | 0.0 | nan | 0.0 |             82.402279 |                      0.0 |                 82.402279 |                  82.402279 |                 32.936813 |                     0.0 |                      0.0 |    49.465466 |             0.0 |              0.0 |         49.465466 |             60.858762 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.826087 | 1.826087 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.018963 | 0.000219 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.826087 |   0.0 |   0.0 |   0.0 |   0.0 |  0.018963 |   0.0 | 0.000219 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |                 15.30875 |      284.047892 |             73.5 |       3.8 |              1019.0 | 0.0 | 0.0 |    20.8 |                  73.5 |                  13.8 |                      15.30875 |           284.047892 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                24.605348 |                     24.605348 |                         0.0 |                          0.0 |                      1.513854 |                           1.513854 |            18.084931 |                 18.084931 |    1000.915069 |   1.199521 | 0.0 | nan | 0.0 |             85.214368 |                      0.0 |                 85.214368 |                  85.214368 |                  35.31351 |                     0.0 |                      0.0 |    49.900858 |             0.0 |              0.0 |         49.900858 |             56.054123 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.041956 | 2.041956 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.011966 | 0.000138 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -2.041956 |   0.0 |   0.0 |   0.0 |   0.0 |  0.011966 |   0.0 | 0.000138 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |                15.242885 |      262.971263 |             80.1 |       3.3 |              1019.0 | 0.0 | 0.0 |    20.9 |                  80.1 |                  12.3 |                     15.242885 |           262.971263 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 24.75714 |                      24.75714 |                         0.0 |                          0.0 |                      1.521999 |                           1.521999 |            19.830469 |                 19.830469 |     999.169531 |   1.198331 | 0.0 | nan | 0.0 |             78.891379 |                      0.0 |                 78.891379 |                  78.891379 |                 28.392598 |                     0.0 |                      0.0 |    50.498781 |             0.0 |              0.0 |         50.498781 |             64.547172 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.800053 | 1.800053 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.005041 | 0.000058 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.800053 |   0.0 |   0.0 |   0.0 |   0.0 |  0.005041 |   0.0 | 0.000058 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |                15.176467 |      272.256414 |             72.9 |       1.8 |              1019.0 | 0.0 | 0.0 |    22.3 |                  72.9 |                  13.1 |                     15.176467 |           272.256414 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 26.96965 |                      26.96965 |                         0.0 |                          0.0 |                      1.639967 |                           1.639967 |            19.660875 |                 19.660875 |     999.339125 |   1.192729 | 0.0 | nan | 0.0 |             81.676924 |                      0.0 |                 81.676924 |                  81.676924 |                 31.114889 |                     0.0 |                      0.0 |    50.562035 |             0.0 |              0.0 |         50.562035 |            118.336481 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.82186 | 1.820396 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  -1.82186 |   0.0 |   0.0 |   0.0 |   0.0 | -0.001465 |   0.0 |      0.0 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |                15.109521 |      267.019218 |             68.5 |       1.9 |              1020.0 | 0.0 | 0.0 |    23.0 |                  68.5 |                  12.8 |                     15.109521 |           267.019218 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                28.139149 |                     28.139149 |                         0.0 |                          0.0 |                      1.701779 |                           1.701779 |            19.275317 |                 19.275317 |    1000.724683 |   1.191257 | 0.0 | nan | 0.0 |             80.105765 |                      0.0 |                 80.105765 |                  80.105765 |                 32.010643 |                     0.0 |                      0.0 |    48.095123 |             0.0 |              0.0 |         48.095123 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.891469 | 1.883357 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.891469 |   0.0 |   0.0 |   0.0 |   0.0 | -0.008111 |   0.0 |      0.0 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |                15.042071 |      269.626547 |             68.6 |       3.0 |              1021.0 | 0.0 | 0.0 |    23.1 |                  68.6 |                  13.1 |                     15.042071 |           269.626547 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                28.309776 |                     28.309776 |                         0.0 |                          0.0 |                      1.710768 |                           1.710768 |            19.420506 |                 19.420506 |    1001.579494 |   1.191966 | 0.0 | nan | 0.0 |             80.887964 |                      0.0 |                 80.887964 |                  80.887964 |                 32.400014 |                     0.0 |                      0.0 |     48.48795 |             0.0 |              0.0 |          48.48795 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.106575 | 2.091331 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -2.106575 |   0.0 |   0.0 |   0.0 |   0.0 | -0.015244 |   0.0 |      0.0 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |                14.974141 |      270.878291 |             66.0 |       2.5 |              1022.0 | 0.0 | 0.0 |    23.2 |                  66.0 |                  13.3 |                     14.974141 |           270.878291 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                28.481304 |                     28.481304 |                         0.0 |                          0.0 |                      1.719796 |                           1.719796 |             18.79766 |                  18.79766 |     1003.20234 |   1.193016 | 0.0 | nan | 0.0 |             81.263487 |                      0.0 |                 81.263487 |                  81.263487 |                 34.565783 |                     0.0 |                      0.0 |    46.697704 |             0.0 |              0.0 |         46.697704 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.036929 | 2.014486 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.942929 |   0.0 |   0.0 |   0.0 |   0.0 | -0.022443 |   0.0 |      0.0 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |                14.905755 |      261.697164 |             69.0 |       1.6 |              1019.0 | 0.0 | 0.0 |    23.2 |                  69.0 |                  12.7 |                     14.905755 |           261.697164 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                28.481304 |                     28.481304 |                         0.0 |                          0.0 |                      1.719796 |                           1.719796 |              19.6521 |                   19.6521 |       999.3479 |    1.18911 | 0.0 | nan | 0.0 |             78.509149 |                      0.0 |                 78.509149 |                  78.509149 |                  31.34697 |                     0.0 |                      0.0 |    47.162179 |             0.0 |              0.0 |         47.162179 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.808559 | 1.779476 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.808559 |   0.0 |   0.0 |   0.0 |   0.0 | -0.029083 |   0.0 |      0.0 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |                14.836934 |      227.962116 |             75.6 |       2.7 |              1015.0 | 0.0 | 0.0 |    21.7 |                  75.6 |                  10.2 |                     14.836934 |           227.962116 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 26.00118 |                      26.00118 |                         0.0 |                          0.0 |                        1.5885 |                             1.5885 |            19.656892 |                 19.656892 |     995.343108 |   1.190432 | 0.0 | nan | 0.0 |             68.388635 |                      0.0 |                 68.388635 |                  68.388635 |                 25.918437 |                     0.0 |                      0.0 |    42.470198 |             0.0 |              0.0 |         42.470198 |             78.890988 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.672736 | 1.637521 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.672736 |   0.0 |   0.0 |   0.0 |   0.0 | -0.035215 |   0.0 |      0.0 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |                  14.7677 |      216.356321 |             81.4 |       3.2 |              1016.0 | 0.0 | 0.0 |    20.8 |                  81.4 |                   9.4 |                       14.7677 |           216.356321 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                24.605348 |                     24.605348 |                         0.0 |                          0.0 |                      1.513854 |                           1.513854 |            20.028753 |                 20.028753 |     995.971247 |   1.195095 | 0.0 | nan | 0.0 |             64.906896 |                      0.0 |                 64.906896 |                  64.906896 |                 23.397902 |                     0.0 |                      0.0 |    41.508994 |             0.0 |              0.0 |         41.508994 |             66.564271 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.529542 | 1.493266 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  1.008458 |   0.0 |   0.0 |   0.0 |   0.0 | -0.036276 |   0.0 |      0.0 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |                14.698075 |      226.637904 |             79.0 |       1.3 |              1019.0 | 0.0 | 0.0 |    19.5 |                  79.0 |                  10.3 |                     14.698075 |           226.637904 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                22.704676 |                     22.704676 |                         0.0 |                          0.0 |                      1.411269 |                           1.411269 |            17.936694 |                 17.936694 |    1001.063306 |   1.204916 | 0.0 | nan | 0.0 |             67.991371 |                      0.0 |                 67.991371 |                  67.991371 |                 28.794151 |                     0.0 |                      0.0 |    39.197221 |             0.0 |              0.0 |         39.197221 |            163.850513 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.283901 | 1.247259 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.283901 |   0.0 |   0.0 |   0.0 |   0.0 | -0.036642 |   0.0 |      0.0 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |                14.628078 |      235.528503 |             75.9 |       1.9 |              1019.0 | 0.0 | 0.0 |    21.0 |                  75.9 |                  11.1 |                     14.628078 |           235.528503 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                24.909749 |                     24.909749 |                         0.0 |                          0.0 |                      1.530181 |                           1.530181 |            18.906499 |                 18.906499 |    1000.093501 |   1.198338 | 0.0 | nan | 0.0 |             70.658551 |                      0.0 |                 70.658551 |                  70.658551 |                 29.177702 |                     0.0 |                      0.0 |    41.480849 |             0.0 |              0.0 |         41.480849 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.510743 | 1.469226 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.510743 |   0.0 |   0.0 |   0.0 |   0.0 | -0.041517 |   0.0 |      0.0 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |                14.557728 |      232.849428 |             70.0 |       2.0 |              1018.0 | 0.0 | 0.0 |    21.6 |                  70.0 |                  11.0 |                     14.557728 |           232.849428 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                25.842752 |                     25.842752 |                         0.0 |                          0.0 |                      1.580055 |                           1.580055 |            18.089927 |                 18.089927 |     999.910073 |   1.195081 | 0.0 | nan | 0.0 |             69.854829 |                      0.0 |                 69.854829 |                  69.854829 |                 31.281288 |                     0.0 |                      0.0 |    38.573541 |             0.0 |              0.0 |         38.573541 |            106.502833 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.594572 | 1.547641 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.594572 |   0.0 |   0.0 |   0.0 |   0.0 | -0.046931 |   0.0 |      0.0 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |                14.487046 |       199.71349 |             70.3 |       2.6 |              1019.0 | 0.0 | 0.0 |    21.1 |                  70.3 |                   8.5 |                     14.487046 |            199.71349 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                25.063177 |                     25.063177 |                         0.0 |                          0.0 |                        1.5384 |                             1.5384 |            17.619414 |                 17.619414 |    1001.380586 |   1.198506 | 0.0 | nan | 0.0 |             59.914047 |                      0.0 |                 59.914047 |                  59.914047 |                 26.711296 |                     0.0 |                      0.0 |    33.202751 |             0.0 |              0.0 |         33.202751 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.541567 | 1.489186 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.541567 |   0.0 |   0.0 |   0.0 |   0.0 |  -0.05238 |   0.0 |      0.0 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |                14.416048 |      233.771902 |             69.0 |       3.6 |              1019.0 | 0.0 | 0.0 |    22.4 |                  69.0 |                  11.3 |                     14.416048 |           233.771902 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                27.134082 |                     27.134082 |                         0.0 |                          0.0 |                      1.648679 |                           1.648679 |            18.722517 |                 18.722517 |    1000.277483 |   1.192743 | 0.0 | nan | 0.0 |             70.131571 |                      0.0 |                 70.131571 |                  70.131571 |                 31.116622 |                     0.0 |                      0.0 |    39.014949 |             0.0 |              0.0 |         39.014949 |             59.168241 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.939383 | 1.880954 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.939383 |   0.0 |   0.0 |   0.0 |   0.0 | -0.058428 |   0.0 |      0.0 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |                14.344752 |      246.121286 |             62.1 |       2.5 |              1017.0 | 0.0 | 0.0 |    22.5 |                  62.1 |                  12.4 |                     14.344752 |           246.121286 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                27.299387 |                     27.299387 |                         0.0 |                          0.0 |                      1.657431 |                           1.657431 |             16.95292 |                  16.95292 |     1000.04708 |   1.190771 | 0.0 | nan | 0.0 |             73.836386 |                      0.0 |                 73.836386 |                  73.836386 |                 38.297604 |                     0.0 |                      0.0 |    35.538782 |             0.0 |              0.0 |         35.538782 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.826811 | 1.761846 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.826811 |   0.0 |   0.0 |   0.0 |   0.0 | -0.064965 |   0.0 |      0.0 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |                14.273175 |        90.58551 |             84.5 |       2.1 |              1015.0 | 0.0 | 0.0 |    19.6 |                  84.5 |                   0.1 |                     14.273175 |             90.58551 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                22.846186 |                     22.846186 |                         0.0 |                          0.0 |                      1.418946 |                           1.418946 |            19.305027 |                 19.305027 |     995.694973 |   1.199129 | 0.0 | nan | 0.0 |             27.175653 |                      0.0 |                 27.175653 |                  27.175653 |                   7.02118 |                     0.0 |                      0.0 |    20.154473 |             0.0 |              0.0 |         20.154473 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.806188 | 0.740109 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  1.073812 |   0.0 |   0.0 |   0.0 |   0.0 | -0.066078 |   0.0 |      0.0 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |                14.201333 |      171.944993 |             83.6 |       1.5 |              1015.0 | 0.0 | 0.0 |    20.3 |                  83.6 |                   6.7 |                     14.201333 |           171.944993 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                23.858503 |                     23.858503 |                         0.0 |                          0.0 |                      1.473678 |                           1.473678 |            19.945708 |                 19.945708 |     995.054292 |   1.195981 | 0.0 | nan | 0.0 |             51.583498 |                      0.0 |                 51.583498 |                  51.583498 |                 18.994889 |                     0.0 |                      0.0 |    32.588609 |             0.0 |              0.0 |         32.588609 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.08421 | 1.018853 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  -0.80221 |   0.0 |   0.0 |   0.0 |   0.0 | -0.065357 |   0.0 |      0.0 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |                14.129243 |      216.549813 |             76.5 |       2.2 |              1011.0 | 0.0 | 0.0 |    24.1 |                  76.5 |                  10.4 |                     14.129243 |           216.549813 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                30.066256 |                     30.066256 |                         0.0 |                          0.0 |                       1.80287 |                            1.80287 |            23.000686 |                 23.000686 |     987.999314 |   1.174651 | 0.0 | nan | 0.0 |             64.964944 |                      0.0 |                 64.964944 |                  64.964944 |                 21.629737 |                     0.0 |                      0.0 |    43.335207 |             0.0 |              0.0 |         43.335207 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.672232 | 1.602675 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.672232 |   0.0 |   0.0 |   0.0 |   0.0 | -0.069557 |   0.0 |      0.0 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |                14.056918 |       166.91683 |             76.6 |       1.9 |              1008.0 | 0.0 | 0.0 |    23.7 |                  76.6 |                   6.5 |                     14.056918 |            166.91683 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                29.352596 |                     29.352596 |                         0.0 |                          0.0 |                      1.765541 |                           1.765541 |            22.484088 |                 22.484088 |     985.515912 |   1.172942 | 0.0 | nan | 0.0 |             50.075049 |                      0.0 |                 50.075049 |                  50.075049 |                 16.222254 |                     0.0 |                      0.0 |    33.852795 |             0.0 |              0.0 |         33.852795 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.374102 | 1.299507 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -1.280102 |   0.0 |   0.0 |   0.0 |   0.0 | -0.074595 |   0.0 |      0.0 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |                13.984374 |      145.985611 |             84.2 |       2.3 |              1006.0 | 0.0 | 0.0 |    21.1 |                  84.2 |                   4.9 |                     13.984374 |           145.985611 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                25.063177 |                     25.063177 |                         0.0 |                          0.0 |                        1.5384 |                             1.5384 |            21.103195 |                 21.103195 |     984.896805 |   1.181557 | 0.0 | nan | 0.0 |             43.795683 |                      0.0 |                 43.795683 |                  43.795683 |                 14.602883 |                     0.0 |                      0.0 |      29.1928 |             0.0 |              0.0 |           29.1928 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.077096 | 1.016237 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  8.792904 |   0.0 |   0.0 |   0.0 |   0.0 |  -0.06086 |   0.0 |      0.0 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |                13.911624 |       165.54389 |             85.8 |       2.1 |              1005.0 | 0.0 | 0.0 |    20.9 |                  85.8 |                   6.6 |                     13.911624 |            165.54389 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 24.75714 |                      24.75714 |                         0.0 |                          0.0 |                      1.521999 |                           1.521999 |            21.241626 |                 21.241626 |     983.758374 |   1.181114 | 0.0 | nan | 0.0 |             49.663167 |                      0.0 |                 49.663167 |                  49.663167 |                 17.349121 |                     0.0 |                      0.0 |    32.314046 |             0.0 |              0.0 |         32.314046 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.09756 | 1.061645 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |   5.01244 |   0.0 |   0.0 |   0.0 |   0.0 | -0.035916 |   0.0 |      0.0 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |                13.838682 |       87.693346 |             86.5 |       2.6 |              1000.0 | 0.0 | 0.0 |    20.2 |                  86.5 |                   0.3 |                     13.838682 |            87.693346 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                23.711531 |                     23.711531 |                         0.0 |                          0.0 |                      1.465752 |                           1.465752 |            20.510475 |                 20.510475 |     979.489525 |   1.178323 | 0.0 | nan | 0.0 |             26.308004 |                      0.0 |                 26.308004 |                  26.308004 |                  6.818024 |                     0.0 |                      0.0 |     19.48998 |             0.0 |              0.0 |          19.48998 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.794859 | 0.791893 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | 13.305141 |   0.0 |   0.0 |   0.0 |   0.0 | -0.002966 |   0.0 |      0.0 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |                13.765561 |       84.356885 |             89.6 |       2.6 |              1004.0 | 0.0 | 0.0 |    16.1 |                  89.6 |                   0.1 |                     13.765561 |            84.356885 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                18.326791 |                     18.326791 |                         0.0 |                          0.0 |                      1.170311 |                           1.170311 |            16.420805 |                 16.420805 |     987.579195 |   1.201704 | 0.0 | nan | 0.0 |             25.307066 |                      0.0 |                 25.307066 |                  25.307066 |                  8.199212 |                     0.0 |                      0.0 |    17.107854 |             0.0 |              0.0 |         17.107854 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.607776 | 0.607776 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.035276 | 0.000408 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  8.040224 |   0.0 |   0.0 |   0.0 |   0.0 |  0.035276 |   0.0 | 0.000408 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |                13.692273 |      142.309202 |             78.3 |       2.1 |              1011.0 | 0.0 | 0.0 |    17.0 |                  78.3 |                   5.0 |                     13.692273 |           142.309202 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                19.406929 |                     19.406929 |                         0.0 |                          0.0 |                      1.230421 |                           1.230421 |            15.195625 |                 15.195625 |     995.804375 |   1.206937 | 0.0 | nan | 0.0 |             42.692761 |                      0.0 |                 42.692761 |                  42.692761 |                 21.775852 |                     0.0 |                      0.0 |    20.916908 |             0.0 |              0.0 |         20.916908 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.90065 |  0.90065 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.047929 | 0.000555 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  -0.90065 |   0.0 |   0.0 |   0.0 |   0.0 |  0.047929 |   0.0 | 0.000555 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |                13.618831 |      126.722078 |             78.5 |       2.1 |              1014.0 | 0.0 | 0.0 |    18.2 |                  78.5 |                   3.8 |                     13.618831 |           126.722078 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                20.933479 |                     20.933479 |                         0.0 |                          0.0 |                      1.314615 |                           1.314615 |            16.432781 |                 16.432781 |     997.567219 |   1.204994 | 0.0 | nan | 0.0 |             38.016624 |                      0.0 |                 38.016624 |                  38.016624 |                 17.491881 |                     0.0 |                      0.0 |    20.524743 |             0.0 |              0.0 |         20.524743 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.912078 | 0.912078 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.04451 | 0.000515 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 | -0.912078 |   0.0 |   0.0 |   0.0 |   0.0 |   0.04451 |   0.0 | 0.000515 |

>>> round_(model.check_waterbalance(conditions))
0.0

.. _lland_v4_fluss_daily:

streams
-------

.. integration-test::

    >>> lnk(FLUSS)
    >>> surfaceresistance.fluss_aug = 0.0
    >>> cropheight.fluss_aug = 0.05
    >>> albedo.fluss_aug = 0.7
    >>> test("lland_v4_fluss_daily",
    ...      axis1=(fluxes.nkor, fluxes.evi, fluxes.qah))
    |       date | nied | teml | sunshineduration | possiblesunshineduration | globalradiation | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration | dailypossiblesunshineduration | dailyglobalradiation |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair |   g |  tz |  wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |     evpo |       evi | evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatinz | wlatsnow | wsensinz | wsenssnow | wsurfinz | wsurf | sff | fvg | wada | qdb | qib1 | qib2 | qbb | qkap |      qdgz |      qah |       qa | inzp | stinz | sinz | esnowinz | asinz | wats | waes | esnow | taus | ebdn | bowa |     qdgz1 |    qdgz2 | qigz1 | qigz2 | qbgz |     qdga1 |    qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    -------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |                15.690785 |       190.25149 |             86.2 |       2.4 |              1007.0 | 0.0 | 0.0 |    16.1 |                  86.2 |                   6.3 |                     15.690785 |            190.25149 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                18.326791 |                     18.326791 |                         0.0 |                          0.0 |                      1.170311 |                           1.170311 |            15.797694 |                 15.797694 |     991.202306 |   1.205601 | 0.0 | nan | 0.0 |             57.075447 |                      0.0 |                 57.075447 |                  57.075447 |                 21.732733 |                     0.0 |                      0.0 |    35.342714 |             0.0 |              0.0 |         35.342714 |             88.752361 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.077197 |  1.077197 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | 11.424803 | 2.941276 | 0.034043 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | 11.424803 |      0.0 |   0.0 |   0.0 |  0.0 |  2.941276 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.034043 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |                15.628821 |      127.660782 |             85.8 |       1.6 |              1013.0 | 0.0 | 0.0 |    17.3 |                  85.8 |                   1.7 |                     15.628821 |           127.660782 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                19.779139 |                     19.779139 |                         0.0 |                          0.0 |                      1.251029 |                           1.251029 |            16.970501 |                 16.970501 |     996.029499 |   1.207284 | 0.0 | nan | 0.0 |             38.298235 |                      0.0 |                 38.298235 |                  38.298235 |                 11.219797 |                     0.0 |                      0.0 |    27.078438 |             0.0 |              0.0 |         27.078438 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.86545 |   0.86545 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  -0.77145 | 3.758016 | 0.043496 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |  -0.77145 |      0.0 |   0.0 |   0.0 |  0.0 |  3.758016 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.043496 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |                15.566134 |      164.401298 |             82.4 |       1.5 |              1016.0 | 0.0 | 0.0 |    18.9 |                  82.4 |                   4.5 |                     15.566134 |           164.401298 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                21.871576 |                     21.871576 |                         0.0 |                          0.0 |                      1.365939 |                           1.365939 |            18.022179 |                 18.022179 |     997.977821 |   1.203775 | 0.0 | nan | 0.0 |             49.320389 |                      0.0 |                 49.320389 |                  49.320389 |                 16.044181 |                     0.0 |                      0.0 |    33.276209 |             0.0 |              0.0 |         33.276209 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.092365 |  1.092365 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.092365 | 1.562905 | 0.018089 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.092365 |      0.0 |   0.0 |   0.0 |  0.0 |  1.562905 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.018089 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |                15.502754 |      269.020502 |             77.6 |       2.8 |              1017.0 | 0.0 | 0.0 |    17.8 |                  77.6 |                  12.4 |                     15.502754 |           269.020502 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                20.413369 |                     20.413369 |                         0.0 |                          0.0 |                      1.286025 |                           1.286025 |            15.840774 |                 15.840774 |    1001.159226 |    1.21051 | 0.0 | nan | 0.0 |              80.70615 |                      0.0 |                  80.70615 |                   80.70615 |                 35.980407 |                     0.0 |                      0.0 |    44.725744 |             0.0 |              0.0 |         44.725744 |             76.073452 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.565894 |  1.565894 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.471894 | 0.226808 | 0.002625 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.471894 |      0.0 |   0.0 |   0.0 |  0.0 |  0.226808 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.002625 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |                15.438712 |      287.848667 |             74.9 |       3.0 |              1018.0 | 0.0 | 0.0 |    18.5 |                  74.9 |                  13.9 |                     15.438712 |           287.848667 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                21.331125 |                     21.331125 |                         0.0 |                          0.0 |                      1.336407 |                           1.336407 |            15.977013 |                 15.977013 |    1002.022987 |   1.208738 | 0.0 | nan | 0.0 |               86.3546 |                      0.0 |                   86.3546 |                    86.3546 |                 39.523937 |                     0.0 |                      0.0 |    46.830664 |             0.0 |              0.0 |         46.830664 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.729733 |  1.097905 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.729733 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.729733 |      0.0 |   0.0 |   0.0 |  0.0 | -0.631827 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |                15.374034 |      274.674263 |             77.9 |       3.5 |              1018.0 | 0.0 | 0.0 |    19.9 |                  77.9 |                  13.0 |                     15.374034 |           274.674263 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                23.275348 |                     23.275348 |                         0.0 |                          0.0 |                      1.442189 |                           1.442189 |            18.131496 |                 18.131496 |     999.868504 |   1.201995 | 0.0 | nan | 0.0 |             82.402279 |                      0.0 |                 82.402279 |                  82.402279 |                 32.936813 |                     0.0 |                      0.0 |    49.465466 |             0.0 |              0.0 |         49.465466 |             60.858762 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.826087 |  0.657403 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.826087 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.826087 |      0.0 |   0.0 |   0.0 |  0.0 | -1.168684 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |                 15.30875 |      284.047892 |             73.5 |       3.8 |              1019.0 | 0.0 | 0.0 |    20.8 |                  73.5 |                  13.8 |                      15.30875 |           284.047892 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                24.605348 |                     24.605348 |                         0.0 |                          0.0 |                      1.513854 |                           1.513854 |            18.084931 |                 18.084931 |    1000.915069 |   1.199521 | 0.0 | nan | 0.0 |             85.214368 |                      0.0 |                 85.214368 |                  85.214368 |                  35.31351 |                     0.0 |                      0.0 |    49.900858 |             0.0 |              0.0 |         49.900858 |             56.054123 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.041956 |  0.511091 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -2.041956 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -2.041956 |      0.0 |   0.0 |   0.0 |  0.0 | -1.530864 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |                15.242885 |      262.971263 |             80.1 |       3.3 |              1019.0 | 0.0 | 0.0 |    20.9 |                  80.1 |                  12.3 |                     15.242885 |           262.971263 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 24.75714 |                      24.75714 |                         0.0 |                          0.0 |                      1.521999 |                           1.521999 |            19.830469 |                 19.830469 |     999.169531 |   1.198331 | 0.0 | nan | 0.0 |             78.891379 |                      0.0 |                 78.891379 |                  78.891379 |                 28.392598 |                     0.0 |                      0.0 |    50.498781 |             0.0 |              0.0 |         50.498781 |             64.547172 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.800053 |  0.093098 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.800053 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.800053 |      0.0 |   0.0 |   0.0 |  0.0 | -1.706955 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |                15.176467 |      272.256414 |             72.9 |       1.8 |              1019.0 | 0.0 | 0.0 |    22.3 |                  72.9 |                  13.1 |                     15.176467 |           272.256414 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 26.96965 |                      26.96965 |                         0.0 |                          0.0 |                      1.639967 |                           1.639967 |            19.660875 |                 19.660875 |     999.339125 |   1.192729 | 0.0 | nan | 0.0 |             81.676924 |                      0.0 |                 81.676924 |                  81.676924 |                 31.114889 |                     0.0 |                      0.0 |    50.562035 |             0.0 |              0.0 |         50.562035 |            118.336481 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.82186 |  0.065871 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  -1.82186 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |  -1.82186 |      0.0 |   0.0 |   0.0 |  0.0 | -1.755989 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |                15.109521 |      267.019218 |             68.5 |       1.9 |              1020.0 | 0.0 | 0.0 |    23.0 |                  68.5 |                  12.8 |                     15.109521 |           267.019218 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                28.139149 |                     28.139149 |                         0.0 |                          0.0 |                      1.701779 |                           1.701779 |            19.275317 |                 19.275317 |    1000.724683 |   1.191257 | 0.0 | nan | 0.0 |             80.105765 |                      0.0 |                 80.105765 |                  80.105765 |                 32.010643 |                     0.0 |                      0.0 |    48.095123 |             0.0 |              0.0 |         48.095123 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.891469 |  0.086838 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.891469 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.891469 |      0.0 |   0.0 |   0.0 |  0.0 | -1.804631 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |                15.042071 |      269.626547 |             68.6 |       3.0 |              1021.0 | 0.0 | 0.0 |    23.1 |                  68.6 |                  13.1 |                     15.042071 |           269.626547 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                28.309776 |                     28.309776 |                         0.0 |                          0.0 |                      1.710768 |                           1.710768 |            19.420506 |                 19.420506 |    1001.579494 |   1.191966 | 0.0 | nan | 0.0 |             80.887964 |                      0.0 |                 80.887964 |                  80.887964 |                 32.400014 |                     0.0 |                      0.0 |     48.48795 |             0.0 |              0.0 |          48.48795 |             71.001889 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.106575 |  0.206065 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -2.106575 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -2.106575 |      0.0 |   0.0 |   0.0 |  0.0 | -1.900509 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |                14.974141 |      270.878291 |             66.0 |       2.5 |              1022.0 | 0.0 | 0.0 |    23.2 |                  66.0 |                  13.3 |                     14.974141 |           270.878291 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                28.481304 |                     28.481304 |                         0.0 |                          0.0 |                      1.719796 |                           1.719796 |             18.79766 |                  18.79766 |     1003.20234 |   1.193016 | 0.0 | nan | 0.0 |             81.263487 |                      0.0 |                 81.263487 |                  81.263487 |                 34.565783 |                     0.0 |                      0.0 |    46.697704 |             0.0 |              0.0 |         46.697704 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 2.036929 |  0.082443 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.942929 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.942929 |      0.0 |   0.0 |   0.0 |  0.0 | -1.954486 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |                14.905755 |      261.697164 |             69.0 |       1.6 |              1019.0 | 0.0 | 0.0 |    23.2 |                  69.0 |                  12.7 |                     14.905755 |           261.697164 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                28.481304 |                     28.481304 |                         0.0 |                          0.0 |                      1.719796 |                           1.719796 |              19.6521 |                   19.6521 |       999.3479 |    1.18911 | 0.0 | nan | 0.0 |             78.509149 |                      0.0 |                 78.509149 |                  78.509149 |                  31.34697 |                     0.0 |                      0.0 |    47.162179 |             0.0 |              0.0 |         47.162179 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.808559 | -0.105944 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.808559 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.808559 |      0.0 |   0.0 |   0.0 |  0.0 | -1.914503 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |                14.836934 |      227.962116 |             75.6 |       2.7 |              1015.0 | 0.0 | 0.0 |    21.7 |                  75.6 |                  10.2 |                     14.836934 |           227.962116 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 26.00118 |                      26.00118 |                         0.0 |                          0.0 |                        1.5885 |                             1.5885 |            19.656892 |                 19.656892 |     995.343108 |   1.190432 | 0.0 | nan | 0.0 |             68.388635 |                      0.0 |                 68.388635 |                  68.388635 |                 25.918437 |                     0.0 |                      0.0 |    42.470198 |             0.0 |              0.0 |         42.470198 |             78.890988 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.672736 | -0.157388 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.672736 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.672736 |      0.0 |   0.0 |   0.0 |  0.0 | -1.830125 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |                  14.7677 |      216.356321 |             81.4 |       3.2 |              1016.0 | 0.0 | 0.0 |    20.8 |                  81.4 |                   9.4 |                       14.7677 |           216.356321 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                24.605348 |                     24.605348 |                         0.0 |                          0.0 |                      1.513854 |                           1.513854 |            20.028753 |                 20.028753 |     995.971247 |   1.195095 | 0.0 | nan | 0.0 |             64.906896 |                      0.0 |                 64.906896 |                  64.906896 |                 23.397902 |                     0.0 |                      0.0 |    41.508994 |             0.0 |              0.0 |         41.508994 |             66.564271 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.529542 |  0.463086 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  1.008458 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |  1.008458 |      0.0 |   0.0 |   0.0 |  0.0 | -1.066456 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |                14.698075 |      226.637904 |             79.0 |       1.3 |              1019.0 | 0.0 | 0.0 |    19.5 |                  79.0 |                  10.3 |                     14.698075 |           226.637904 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                22.704676 |                     22.704676 |                         0.0 |                          0.0 |                      1.411269 |                           1.411269 |            17.936694 |                 17.936694 |    1001.063306 |   1.204916 | 0.0 | nan | 0.0 |             67.991371 |                      0.0 |                 67.991371 |                  67.991371 |                 28.794151 |                     0.0 |                      0.0 |    39.197221 |             0.0 |              0.0 |         39.197221 |            163.850513 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.283901 |  0.595002 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.283901 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.283901 |      0.0 |   0.0 |   0.0 |  0.0 | -0.688899 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |                14.628078 |      235.528503 |             75.9 |       1.9 |              1019.0 | 0.0 | 0.0 |    21.0 |                  75.9 |                  11.1 |                     14.628078 |           235.528503 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                24.909749 |                     24.909749 |                         0.0 |                          0.0 |                      1.530181 |                           1.530181 |            18.906499 |                 18.906499 |    1000.093501 |   1.198338 | 0.0 | nan | 0.0 |             70.658551 |                      0.0 |                 70.658551 |                  70.658551 |                 29.177702 |                     0.0 |                      0.0 |    41.480849 |             0.0 |              0.0 |         41.480849 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.510743 |  0.485942 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.510743 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.510743 |      0.0 |   0.0 |   0.0 |  0.0 | -1.024801 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |                14.557728 |      232.849428 |             70.0 |       2.0 |              1018.0 | 0.0 | 0.0 |    21.6 |                  70.0 |                  11.0 |                     14.557728 |           232.849428 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                25.842752 |                     25.842752 |                         0.0 |                          0.0 |                      1.580055 |                           1.580055 |            18.089927 |                 18.089927 |     999.910073 |   1.195081 | 0.0 | nan | 0.0 |             69.854829 |                      0.0 |                 69.854829 |                  69.854829 |                 31.281288 |                     0.0 |                      0.0 |    38.573541 |             0.0 |              0.0 |         38.573541 |            106.502833 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.594572 |  0.321551 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.594572 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.594572 |      0.0 |   0.0 |   0.0 |  0.0 |  -1.27302 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |                14.487046 |       199.71349 |             70.3 |       2.6 |              1019.0 | 0.0 | 0.0 |    21.1 |                  70.3 |                   8.5 |                     14.487046 |            199.71349 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                25.063177 |                     25.063177 |                         0.0 |                          0.0 |                        1.5384 |                             1.5384 |            17.619414 |                 17.619414 |    1001.380586 |   1.198506 | 0.0 | nan | 0.0 |             59.914047 |                      0.0 |                 59.914047 |                  59.914047 |                 26.711296 |                     0.0 |                      0.0 |    33.202751 |             0.0 |              0.0 |         33.202751 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.541567 |  0.132224 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.541567 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.541567 |      0.0 |   0.0 |   0.0 |  0.0 | -1.409342 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |                14.416048 |      233.771902 |             69.0 |       3.6 |              1019.0 | 0.0 | 0.0 |    22.4 |                  69.0 |                  11.3 |                     14.416048 |           233.771902 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                27.134082 |                     27.134082 |                         0.0 |                          0.0 |                      1.648679 |                           1.648679 |            18.722517 |                 18.722517 |    1000.277483 |   1.192743 | 0.0 | nan | 0.0 |             70.131571 |                      0.0 |                 70.131571 |                  70.131571 |                 31.116622 |                     0.0 |                      0.0 |    39.014949 |             0.0 |              0.0 |         39.014949 |             59.168241 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.939383 |  0.365956 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.939383 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.939383 |      0.0 |   0.0 |   0.0 |  0.0 | -1.573427 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |                14.344752 |      246.121286 |             62.1 |       2.5 |              1017.0 | 0.0 | 0.0 |    22.5 |                  62.1 |                  12.4 |                     14.344752 |           246.121286 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                27.299387 |                     27.299387 |                         0.0 |                          0.0 |                      1.657431 |                           1.657431 |             16.95292 |                  16.95292 |     1000.04708 |   1.190771 | 0.0 | nan | 0.0 |             73.836386 |                      0.0 |                 73.836386 |                  73.836386 |                 38.297604 |                     0.0 |                      0.0 |    35.538782 |             0.0 |              0.0 |         35.538782 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.826811 |  0.111688 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.826811 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.826811 |      0.0 |   0.0 |   0.0 |  0.0 | -1.715123 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |                14.273175 |        90.58551 |             84.5 |       2.1 |              1015.0 | 0.0 | 0.0 |    19.6 |                  84.5 |                   0.1 |                     14.273175 |             90.58551 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                22.846186 |                     22.846186 |                         0.0 |                          0.0 |                      1.418946 |                           1.418946 |            19.305027 |                 19.305027 |     995.694973 |   1.199129 | 0.0 | nan | 0.0 |             27.175653 |                      0.0 |                 27.175653 |                  27.175653 |                   7.02118 |                     0.0 |                      0.0 |    20.154473 |             0.0 |              0.0 |         20.154473 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.806188 |  -0.21427 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  1.073812 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |  1.073812 |      0.0 |   0.0 |   0.0 |  0.0 | -1.020458 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |                14.201333 |      171.944993 |             83.6 |       1.5 |              1015.0 | 0.0 | 0.0 |    20.3 |                  83.6 |                   6.7 |                     14.201333 |           171.944993 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                23.858503 |                     23.858503 |                         0.0 |                          0.0 |                      1.473678 |                           1.473678 |            19.945708 |                 19.945708 |     995.054292 |   1.195981 | 0.0 | nan | 0.0 |             51.583498 |                      0.0 |                 51.583498 |                  51.583498 |                 18.994889 |                     0.0 |                      0.0 |    32.588609 |             0.0 |              0.0 |         32.588609 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.08421 |  0.557521 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  -0.80221 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |  -0.80221 |      0.0 |   0.0 |   0.0 |  0.0 | -0.526689 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |                14.129243 |      216.549813 |             76.5 |       2.2 |              1011.0 | 0.0 | 0.0 |    24.1 |                  76.5 |                  10.4 |                     14.129243 |           216.549813 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                30.066256 |                     30.066256 |                         0.0 |                          0.0 |                       1.80287 |                            1.80287 |            23.000686 |                 23.000686 |     987.999314 |   1.174651 | 0.0 | nan | 0.0 |             64.964944 |                      0.0 |                 64.964944 |                  64.964944 |                 21.629737 |                     0.0 |                      0.0 |    43.335207 |             0.0 |              0.0 |         43.335207 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.672232 |  0.793059 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.672232 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.672232 |      0.0 |   0.0 |   0.0 |  0.0 | -0.879173 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |                14.056918 |       166.91683 |             76.6 |       1.9 |              1008.0 | 0.0 | 0.0 |    23.7 |                  76.6 |                   6.5 |                     14.056918 |            166.91683 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                29.352596 |                     29.352596 |                         0.0 |                          0.0 |                      1.765541 |                           1.765541 |            22.484088 |                 22.484088 |     985.515912 |   1.172942 | 0.0 | nan | 0.0 |             50.075049 |                      0.0 |                 50.075049 |                  50.075049 |                 16.222254 |                     0.0 |                      0.0 |    33.852795 |             0.0 |              0.0 |         33.852795 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.374102 |  0.226008 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -1.280102 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -1.280102 |      0.0 |   0.0 |   0.0 |  0.0 | -1.148095 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |                13.984374 |      145.985611 |             84.2 |       2.3 |              1006.0 | 0.0 | 0.0 |    21.1 |                  84.2 |                   4.9 |                     13.984374 |           145.985611 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                25.063177 |                     25.063177 |                         0.0 |                          0.0 |                        1.5384 |                             1.5384 |            21.103195 |                 21.103195 |     984.896805 |   1.181557 | 0.0 | nan | 0.0 |             43.795683 |                      0.0 |                 43.795683 |                  43.795683 |                 14.602883 |                     0.0 |                      0.0 |      29.1928 |             0.0 |              0.0 |           29.1928 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 1.077096 |  1.077096 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  8.792904 |   1.3836 | 0.016014 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |  8.792904 |      0.0 |   0.0 |   0.0 |  0.0 |    1.3836 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.016014 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |                13.911624 |       165.54389 |             85.8 |       2.1 |              1005.0 | 0.0 | 0.0 |    20.9 |                  85.8 |                   6.6 |                     13.911624 |            165.54389 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 24.75714 |                      24.75714 |                         0.0 |                          0.0 |                      1.521999 |                           1.521999 |            21.241626 |                 21.241626 |     983.758374 |   1.181114 | 0.0 | nan | 0.0 |             49.663167 |                      0.0 |                 49.663167 |                  49.663167 |                 17.349121 |                     0.0 |                      0.0 |    32.314046 |             0.0 |              0.0 |         32.314046 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.09756 |   1.09756 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |   5.01244 |  3.86595 | 0.044745 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   5.01244 |      0.0 |   0.0 |   0.0 |  0.0 |   3.86595 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.044745 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |                13.838682 |       87.693346 |             86.5 |       2.6 |              1000.0 | 0.0 | 0.0 |    20.2 |                  86.5 |                   0.3 |                     13.838682 |            87.693346 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                23.711531 |                     23.711531 |                         0.0 |                          0.0 |                      1.465752 |                           1.465752 |            20.510475 |                 20.510475 |     979.489525 |   1.178323 | 0.0 | nan | 0.0 |             26.308004 |                      0.0 |                 26.308004 |                  26.308004 |                  6.818024 |                     0.0 |                      0.0 |     19.48998 |             0.0 |              0.0 |          19.48998 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.794859 |  0.794859 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | 13.305141 | 7.027354 | 0.081335 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |      12.0 | 1.305141 |   0.0 |   0.0 |  0.0 |  6.199583 |  0.82777 |   0.0 |   0.0 |  0.0 |   0.0 | 0.081335 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |                13.765561 |       84.356885 |             89.6 |       2.6 |              1004.0 | 0.0 | 0.0 |    16.1 |                  89.6 |                   0.1 |                     13.765561 |            84.356885 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                18.326791 |                     18.326791 |                         0.0 |                          0.0 |                      1.170311 |                           1.170311 |            16.420805 |                 16.420805 |     987.579195 |   1.201704 | 0.0 | nan | 0.0 |             25.307066 |                      0.0 |                 25.307066 |                  25.307066 |                  8.199212 |                     0.0 |                      0.0 |    17.107854 |             0.0 |              0.0 |         17.107854 |             81.925256 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.607776 |  0.607776 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  8.040224 |  8.32407 | 0.096343 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |  8.040224 |      0.0 |   0.0 |   0.0 |  0.0 |  7.885403 | 0.438667 |   0.0 |   0.0 |  0.0 |   0.0 | 0.096343 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |                13.692273 |      142.309202 |             78.3 |       2.1 |              1011.0 | 0.0 | 0.0 |    17.0 |                  78.3 |                   5.0 |                     13.692273 |           142.309202 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                19.406929 |                     19.406929 |                         0.0 |                          0.0 |                      1.230421 |                           1.230421 |            15.195625 |                 15.195625 |     995.804375 |   1.206937 | 0.0 | nan | 0.0 |             42.692761 |                      0.0 |                 42.692761 |                  42.692761 |                 21.775852 |                     0.0 |                      0.0 |    20.916908 |             0.0 |              0.0 |         20.916908 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.90065 |   0.90065 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  -0.90065 | 5.691378 | 0.065872 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |  -0.90065 |      0.0 |   0.0 |   0.0 |  0.0 |  5.655812 | 0.035566 |   0.0 |   0.0 |  0.0 |   0.0 | 0.065872 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |                13.618831 |      126.722078 |             78.5 |       2.1 |              1014.0 | 0.0 | 0.0 |    18.2 |                  78.5 |                   3.8 |                     13.618831 |           126.722078 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                20.933479 |                     20.933479 |                         0.0 |                          0.0 |                      1.314615 |                           1.314615 |            16.432781 |                 16.432781 |     997.567219 |   1.204994 | 0.0 | nan | 0.0 |             38.016624 |                      0.0 |                 38.016624 |                  38.016624 |                 17.491881 |                     0.0 |                      0.0 |    20.524743 |             0.0 |              0.0 |         20.524743 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.912078 |  0.912078 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 | -0.912078 | 2.597892 | 0.030068 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 | -0.912078 |      0.0 |   0.0 |   0.0 |  0.0 |  2.595008 | 0.002884 |   0.0 |   0.0 |  0.0 |   0.0 | 0.030068 |

>>> round_(model.check_waterbalance(conditions))
0.0

.. _lland_v4_vers_daily:

sealed surfaces
---------------

.. integration-test::

    >>> lnk(VERS)
    >>> negq(True)
    >>> surfaceresistance.vers_aug = 500
    >>> cropheight.vers_aug = 5.0
    >>> albedo.vers_aug = 0.10
    >>> lai.vers_aug = 10.0
    >>> test("lland_v4_vers_daily",
    ...      axis1=(inputs.nied, fluxes.qah, states.bowa))
    |       date | nied | teml | sunshineduration | possiblesunshineduration | globalradiation | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration | dailypossiblesunshineduration | dailyglobalradiation |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair |    g |        tz |         wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |   nbes | sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |      evpo |   evi | evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatinz | wlatsnow | wsensinz | wsenssnow | wsurfinz | wsurf | sff | fvg |   wada |    qdb | qib1 | qib2 | qbb | qkap |   qdgz |      qah |       qa | inzp | stinz | sinz | esnowinz | asinz | wats | waes | esnow | taus |       ebdn | bowa |  qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 |    qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |                15.690785 |       190.25149 |             86.2 |       2.4 |              1007.0 | 0.0 | 0.0 |    16.1 |                  86.2 |                   6.3 |                     15.690785 |            190.25149 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                18.326791 |                     18.326791 |                         0.0 |                          0.0 |                      1.170311 |                           1.170311 |            15.797694 |                 15.797694 |     991.202306 |   1.205601 | -2.9 | 11.733333 |      -26.2 |            171.226341 |                      0.0 |                171.226341 |                 171.226341 |                 21.732733 |                     0.0 |                      0.0 |   149.493608 |             0.0 |              0.0 |        149.493608 |             16.038785 |                   nan |                    500.0 |                   500.0 | 10.502 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  7.049443 |   2.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 10.502 | 10.502 |  0.0 |  0.0 | 0.0 |  0.0 | 10.502 | 2.703704 | 0.031293 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |      373.3 |  0.0 | 10.502 |   0.0 |   0.0 |   0.0 |  0.0 | 2.703704 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.031293 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |                15.628821 |      127.660782 |             85.8 |       1.6 |              1013.0 | 0.0 | 0.0 |    17.3 |                  85.8 |                   1.7 |                     15.628821 |           127.660782 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                19.779139 |                     19.779139 |                         0.0 |                          0.0 |                      1.251029 |                           1.251029 |            16.970501 |                 16.970501 |     996.029499 |   1.207284 | -2.9 | 18.443733 |     6.8624 |            114.894704 |                      0.0 |                114.894704 |                 114.894704 |                 11.219797 |                     0.0 |                      0.0 |   103.674907 |             0.0 |              0.0 |        103.674907 |             24.058177 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   5.02075 | 0.094 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 3.637039 | 0.042095 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |   363.5376 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 3.637039 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.042095 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |                15.566134 |      164.401298 |             82.4 |       1.5 |              1016.0 | 0.0 | 0.0 |    18.9 |                  82.4 |                   4.5 |                     15.566134 |           164.401298 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                21.871576 |                     21.871576 |                         0.0 |                          0.0 |                      1.365939 |                           1.365939 |            18.022179 |                 18.022179 |     997.977821 |   1.203775 | -2.9 | 15.632162 | -19.607027 |            147.961168 |                      0.0 |                147.961168 |                 147.961168 |                 16.044181 |                     0.0 |                      0.0 |   131.916987 |             0.0 |              0.0 |        131.916987 |             25.662055 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  6.335399 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 1.940764 | 0.022463 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 380.244627 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 1.940764 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.022463 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |                15.502754 |      269.020502 |             77.6 |       2.8 |              1017.0 | 0.0 | 0.0 |    17.8 |                  77.6 |                  12.4 |                     15.502754 |           269.020502 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                20.413369 |                     20.413369 |                         0.0 |                          0.0 |                      1.286025 |                           1.286025 |            15.840774 |                 15.840774 |    1001.159226 |    1.21051 | -2.9 | 20.443786 |  15.862716 |            242.118451 |                      0.0 |                242.118451 |                 242.118451 |                 35.980407 |                     0.0 |                      0.0 |   206.138045 |             0.0 |              0.0 |        206.138045 |              13.74753 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  12.23059 | 0.094 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 1.035613 | 0.011986 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 361.481911 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 1.035613 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.011986 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |                15.438712 |      287.848667 |             74.9 |       3.0 |              1018.0 | 0.0 | 0.0 |    18.5 |                  74.9 |                  13.9 |                     15.438712 |           287.848667 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                21.331125 |                     21.331125 |                         0.0 |                          0.0 |                      1.336407 |                           1.336407 |            15.977013 |                 15.977013 |    1002.022987 |   1.208738 | -2.9 | 15.040124 | -20.759257 |              259.0638 |                      0.0 |                  259.0638 |                   259.0638 |                 39.523937 |                     0.0 |                      0.0 |   219.539864 |             0.0 |              0.0 |        219.539864 |             12.831028 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 14.271812 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.552615 | 0.006396 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 379.341169 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.552615 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.006396 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |                15.374034 |      274.674263 |             77.9 |       3.5 |              1018.0 | 0.0 | 0.0 |    19.9 |                  77.9 |                  13.0 |                     15.374034 |           274.674263 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           2.805569 |          3.5 |                23.275348 |                     23.275348 |                         0.0 |                          0.0 |                      1.442189 |                           1.442189 |            18.131496 |                 18.131496 |     999.868504 |   1.201995 | -2.9 |  20.18359 |   1.701539 |            247.206837 |                      0.0 |                247.206837 |                 247.206837 |                 32.936813 |                     0.0 |                      0.0 |   214.270023 |             0.0 |              0.0 |        214.270023 |             10.998024 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 14.791542 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.294881 | 0.003413 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 374.739629 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.294881 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.003413 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |                 15.30875 |      284.047892 |             73.5 |       3.8 |              1019.0 | 0.0 | 0.0 |    20.8 |                  73.5 |                  13.8 |                      15.30875 |           284.047892 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                24.605348 |                     24.605348 |                         0.0 |                          0.0 |                      1.513854 |                           1.513854 |            18.084931 |                 18.084931 |    1000.915069 |   1.199521 | -2.9 | 18.858347 | -11.649921 |            255.643103 |                      0.0 |                255.643103 |                 255.643103 |                  35.31351 |                     0.0 |                      0.0 |   220.329593 |             0.0 |              0.0 |        220.329593 |             10.129759 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 18.203372 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.157352 | 0.001821 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  383.48955 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.157352 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.001821 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |                15.242885 |      262.971263 |             80.1 |       3.3 |              1019.0 | 0.0 | 0.0 |    20.9 |                  80.1 |                  12.3 |                     15.242885 |           262.971263 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           2.645251 |          3.3 |                 24.75714 |                      24.75714 |                         0.0 |                          0.0 |                      1.521999 |                           1.521999 |            19.830469 |                 19.830469 |     999.169531 |   1.198331 | -2.9 | 21.378324 |   2.869942 |            236.674137 |                      0.0 |                236.674137 |                 236.674137 |                 28.392598 |                     0.0 |                      0.0 |   208.281539 |             0.0 |              0.0 |        208.281539 |             11.664571 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 13.456656 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.083965 | 0.000972 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 377.719608 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.083965 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000972 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |                15.176467 |      272.256414 |             72.9 |       1.8 |              1019.0 | 0.0 | 0.0 |    22.3 |                  72.9 |                  13.1 |                     15.176467 |           272.256414 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 26.96965 |                      26.96965 |                         0.0 |                          0.0 |                      1.639967 |                           1.639967 |            19.660875 |                 19.660875 |     999.339125 |   1.192729 | -2.9 |  19.71658 | -15.500518 |            245.030772 |                      0.0 |                245.030772 |                 245.030772 |                 31.114889 |                     0.0 |                      0.0 |   213.915883 |             0.0 |              0.0 |        213.915883 |             21.385046 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 11.840372 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.044804 | 0.000519 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 390.320126 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.044804 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000519 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |                15.109521 |      267.019218 |             68.5 |       1.9 |              1020.0 | 0.0 | 0.0 |    23.0 |                  68.5 |                  12.8 |                     15.109521 |           267.019218 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                28.139149 |                     28.139149 |                         0.0 |                          0.0 |                      1.701779 |                           1.701779 |            19.275317 |                 19.275317 |    1000.724683 |   1.191257 | -2.9 |  23.34553 |   2.073177 |            240.317296 |                      0.0 |                240.317296 |                 240.317296 |                 32.010643 |                     0.0 |                      0.0 |   208.306653 |             0.0 |              0.0 |        208.306653 |             20.259517 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 13.388635 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.023908 | 0.000277 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 385.346949 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.023908 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000277 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |                15.042071 |      269.626547 |             68.6 |       3.0 |              1021.0 | 0.0 | 0.0 |    23.1 |                  68.6 |                  13.1 |                     15.042071 |           269.626547 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                28.309776 |                     28.309776 |                         0.0 |                          0.0 |                      1.710768 |                           1.710768 |            19.420506 |                 19.420506 |    1001.579494 |   1.191966 | -2.9 | 21.913255 |  -7.120473 |            242.663892 |                      0.0 |                242.663892 |                 242.663892 |                 32.400014 |                     0.0 |                      0.0 |   210.263878 |             0.0 |              0.0 |        210.263878 |             12.831028 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 18.009707 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.012758 | 0.000148 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 389.567421 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.012758 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000148 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |                14.974141 |      270.878291 |             66.0 |       2.5 |              1022.0 | 0.0 | 0.0 |    23.2 |                  66.0 |                  13.3 |                     14.974141 |           270.878291 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                28.481304 |                     28.481304 |                         0.0 |                          0.0 |                      1.719796 |                           1.719796 |             18.79766 |                  18.79766 |     1003.20234 |   1.193016 | -2.9 | 23.128751 |  -0.427496 |            243.790462 |                      0.0 |                243.790462 |                 243.790462 |                 34.565783 |                     0.0 |                      0.0 |   209.224678 |             0.0 |              0.0 |        209.224678 |             15.397233 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 16.851676 | 0.094 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.006808 | 0.000079 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 387.094917 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.006808 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000079 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |                14.905755 |      261.697164 |             69.0 |       1.6 |              1019.0 | 0.0 | 0.0 |    23.2 |                  69.0 |                  12.7 |                     14.905755 |           261.697164 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           1.282546 |          1.6 |                28.481304 |                     28.481304 |                         0.0 |                          0.0 |                      1.719796 |                           1.719796 |              19.6521 |                   19.6521 |       999.3479 |    1.18911 | -2.9 | 22.416669 |  -4.699983 |            235.527447 |                      0.0 |                235.527447 |                 235.527447 |                  31.34697 |                     0.0 |                      0.0 |   204.180477 |             0.0 |              0.0 |        204.180477 |             24.058177 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 11.953889 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.003633 | 0.000042 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |   388.8949 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.003633 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000042 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |                14.836934 |      227.962116 |             75.6 |       2.7 |              1015.0 | 0.0 | 0.0 |    21.7 |                  75.6 |                  10.2 |                     14.836934 |           227.962116 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 26.00118 |                      26.00118 |                         0.0 |                          0.0 |                        1.5885 |                             1.5885 |            19.656892 |                 19.656892 |     995.343108 |   1.190432 | -2.9 | 22.935065 |   7.410388 |            205.165904 |                      0.0 |                205.165904 |                 205.165904 |                 25.918437 |                     0.0 |                      0.0 |   179.247467 |             0.0 |              0.0 |        179.247467 |             14.256697 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 13.005745 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.001938 | 0.000022 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 378.584513 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.001938 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000022 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |                  14.7677 |      216.356321 |             81.4 |       3.2 |              1016.0 | 0.0 | 0.0 |    20.8 |                  81.4 |                   9.4 |                       14.7677 |           216.356321 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           2.565092 |          3.2 |                24.605348 |                     24.605348 |                         0.0 |                          0.0 |                      1.513854 |                           1.513854 |            20.028753 |                 20.028753 |     995.971247 |   1.195095 | -2.9 | 19.965673 |  -5.005962 |            194.720689 |                      0.0 |                194.720689 |                 194.720689 |                 23.397902 |                     0.0 |                      0.0 |   171.322787 |             0.0 |              0.0 |        171.322787 |             12.029088 |                   nan |                    500.0 |                   500.0 |  0.538 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 11.728612 |   2.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.538 |  0.538 |  0.0 |  0.0 | 0.0 |  0.0 |  0.538 | 0.139541 | 0.001615 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 380.690475 |  0.0 |  0.538 |   0.0 |   0.0 |   0.0 |  0.0 | 0.139541 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.001615 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |                14.698075 |      226.637904 |             79.0 |       1.3 |              1019.0 | 0.0 | 0.0 |    19.5 |                  79.0 |                  10.3 |                     14.698075 |           226.637904 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                22.704676 |                     22.704676 |                         0.0 |                          0.0 |                      1.411269 |                           1.411269 |            17.936694 |                 17.936694 |    1001.063306 |   1.204916 | -2.9 |  20.57219 |   6.433141 |            203.974113 |                      0.0 |                203.974113 |                 203.974113 |                 28.794151 |                     0.0 |                      0.0 |   175.179963 |             0.0 |              0.0 |        175.179963 |             29.610064 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  7.560661 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.186871 | 0.002163 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 371.357334 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.186871 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.002163 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |                14.628078 |      235.528503 |             75.9 |       1.9 |              1019.0 | 0.0 | 0.0 |    21.0 |                  75.9 |                  11.1 |                     14.628078 |           235.528503 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                24.909749 |                     24.909749 |                         0.0 |                          0.0 |                      1.530181 |                           1.530181 |            18.906499 |                 18.906499 |    1000.093501 |   1.198338 | -2.9 | 17.884246 | -18.694526 |            211.975652 |                      0.0 |                211.975652 |                 211.975652 |                 29.177702 |                     0.0 |                      0.0 |    182.79795 |             0.0 |              0.0 |         182.79795 |             20.259517 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 10.391615 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.099717 | 0.001154 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 387.151861 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.099717 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.001154 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |                14.557728 |      232.849428 |             70.0 |       2.0 |              1018.0 | 0.0 | 0.0 |    21.6 |                  70.0 |                  11.0 |                     14.557728 |           232.849428 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                25.842752 |                     25.842752 |                         0.0 |                          0.0 |                      1.580055 |                           1.580055 |            18.089927 |                 18.089927 |     999.910073 |   1.195081 | -2.9 | 22.433069 |   4.998415 |            209.564486 |                      0.0 |                209.564486 |                 209.564486 |                 31.281288 |                     0.0 |                      0.0 |   178.283198 |             0.0 |              0.0 |        178.283198 |             19.246541 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 12.298412 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 |  0.05321 | 0.000616 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 379.253445 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.05321 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000616 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |                14.487046 |       199.71349 |             70.3 |       2.6 |              1019.0 | 0.0 | 0.0 |    21.1 |                  70.3 |                   8.5 |                     14.487046 |            199.71349 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                25.063177 |                     25.063177 |                         0.0 |                          0.0 |                        1.5384 |                             1.5384 |            17.619414 |                 17.619414 |    1001.380586 |   1.198506 | -2.9 | 20.158326 |  -5.650046 |            179.742141 |                      0.0 |                179.742141 |                 179.742141 |                 26.711296 |                     0.0 |                      0.0 |   153.030845 |             0.0 |              0.0 |        153.030845 |             14.805032 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 13.758946 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.028393 | 0.000329 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 382.003492 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.028393 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000329 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |                14.416048 |      233.771902 |             69.0 |       3.6 |              1019.0 | 0.0 | 0.0 |    22.4 |                  69.0 |                  11.3 |                     14.416048 |           233.771902 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           2.885728 |          3.6 |                27.134082 |                     27.134082 |                         0.0 |                          0.0 |                      1.648679 |                           1.648679 |            18.722517 |                 18.722517 |    1000.277483 |   1.192743 | -2.9 | 20.950339 |  -8.697966 |            210.394712 |                      0.0 |                210.394712 |                 210.394712 |                 31.116622 |                     0.0 |                      0.0 |    179.27809 |             0.0 |              0.0 |         179.27809 |             10.692523 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 19.233312 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.015151 | 0.000175 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 387.801458 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.015151 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000175 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |                14.344752 |      246.121286 |             62.1 |       2.5 |              1017.0 | 0.0 | 0.0 |    22.5 |                  62.1 |                  12.4 |                     14.344752 |           246.121286 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           2.003978 |          2.5 |                27.299387 |                     27.299387 |                         0.0 |                          0.0 |                      1.657431 |                           1.657431 |             16.95292 |                  16.95292 |     1000.04708 |   1.190771 | -2.9 | 22.620153 |    0.72092 |            221.509157 |                      0.0 |                221.509157 |                 221.509157 |                 38.297604 |                     0.0 |                      0.0 |   183.211553 |             0.0 |              0.0 |        183.211553 |             15.397233 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  17.26053 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.008085 | 0.000094 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 384.180539 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.008085 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000094 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |                14.273175 |        90.58551 |             84.5 |       2.1 |              1015.0 | 0.0 | 0.0 |    19.6 |                  84.5 |                   0.1 |                     14.273175 |             90.58551 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                22.846186 |                     22.846186 |                         0.0 |                          0.0 |                      1.418946 |                           1.418946 |            19.305027 |                 19.305027 |     995.694973 |   1.199129 | -2.9 | 21.577328 |  11.863971 |             81.526959 |                      0.0 |                 81.526959 |                  81.526959 |                   7.02118 |                     0.0 |                      0.0 |    74.505779 |             0.0 |              0.0 |         74.505779 |             18.330039 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  5.830318 |  1.88 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.004314 |  0.00005 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 369.416568 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.004314 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.00005 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |                14.201333 |      171.944993 |             83.6 |       1.5 |              1015.0 | 0.0 | 0.0 |    20.3 |                  83.6 |                   6.7 |                     14.201333 |           171.944993 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                23.858503 |                     23.858503 |                         0.0 |                          0.0 |                      1.473678 |                           1.473678 |            19.945708 |                 19.945708 |     995.054292 |   1.195981 | -2.9 | 17.325305 | -17.848171 |            154.750494 |                      0.0 |                154.750494 |                 154.750494 |                 18.994889 |                     0.0 |                      0.0 |   135.755605 |             0.0 |              0.0 |        135.755605 |             25.662055 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  6.371341 | 0.282 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.002302 | 0.000027 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 384.364739 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.002302 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000027 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |                14.129243 |      216.549813 |             76.5 |       2.2 |              1011.0 | 0.0 | 0.0 |    24.1 |                  76.5 |                  10.4 |                     14.129243 |           216.549813 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                30.066256 |                     30.066256 |                         0.0 |                          0.0 |                       1.80287 |                            1.80287 |            23.000686 |                 23.000686 |     987.999314 |   1.174651 | -2.9 | 21.630378 | -14.817732 |            194.894831 |                      0.0 |                194.894831 |                 194.894831 |                 21.629737 |                     0.0 |                      0.0 |   173.265095 |             0.0 |              0.0 |        173.265095 |             17.496856 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 11.503853 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.001228 | 0.000014 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  396.28247 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.001228 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000014 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |                14.056918 |       166.91683 |             76.6 |       1.9 |              1008.0 | 0.0 | 0.0 |    23.7 |                  76.6 |                   6.5 |                     14.056918 |            166.91683 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                29.352596 |                     29.352596 |                         0.0 |                          0.0 |                      1.765541 |                           1.765541 |            22.484088 |                 22.484088 |     985.515912 |   1.172942 | -2.9 | 25.062685 |   8.176109 |            150.225147 |                      0.0 |                150.225147 |                 150.225147 |                 16.222254 |                     0.0 |                      0.0 |   134.002892 |             0.0 |              0.0 |        134.002892 |             20.259517 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  9.461664 | 0.094 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 0.000655 | 0.000008 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 385.206362 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000655 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000008 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |                13.984374 |      145.985611 |             84.2 |       2.3 |              1006.0 | 0.0 | 0.0 |    21.1 |                  84.2 |                   4.9 |                     13.984374 |           145.985611 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                25.063177 |                     25.063177 |                         0.0 |                          0.0 |                        1.5384 |                             1.5384 |            21.103195 |                 21.103195 |     984.896805 |   1.181557 | -2.9 | 21.872765 |   4.636593 |             131.38705 |                      0.0 |                 131.38705 |                  131.38705 |                 14.602883 |                     0.0 |                      0.0 |   116.784167 |             0.0 |              0.0 |        116.784167 |             16.736123 |                   nan |                    500.0 |                   500.0 |   7.87 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  7.472407 |   2.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |   7.87 |   7.87 |  0.0 |  0.0 | 0.0 |  0.0 |   7.87 | 2.026454 | 0.023454 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 377.669769 |  0.0 |   7.87 |   0.0 |   0.0 |   0.0 |  0.0 | 2.026454 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.023454 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |                13.911624 |       165.54389 |             85.8 |       2.1 |              1005.0 | 0.0 | 0.0 |    20.9 |                  85.8 |                   6.6 |                     13.911624 |            165.54389 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                 24.75714 |                      24.75714 |                         0.0 |                          0.0 |                      1.521999 |                           1.521999 |            21.241626 |                 21.241626 |     983.758374 |   1.181114 | -2.9 | 19.702227 |   -7.18664 |            148.989501 |                      0.0 |                148.989501 |                 148.989501 |                 17.349121 |                     0.0 |                      0.0 |    131.64038 |             0.0 |              0.0 |         131.64038 |             18.330039 |                   nan |                    500.0 |                   500.0 |   4.11 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   6.96061 |   2.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |   4.11 |   4.11 |  0.0 |  0.0 | 0.0 |  0.0 |   4.11 |  3.78382 | 0.043794 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 381.956408 |  0.0 |   4.11 |   0.0 |   0.0 |   0.0 |  0.0 |  3.78382 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.043794 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |                13.838682 |       87.693346 |             86.5 |       2.6 |              1000.0 | 0.0 | 0.0 |    20.2 |                  86.5 |                   0.3 |                     13.838682 |            87.693346 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                23.711531 |                     23.711531 |                         0.0 |                          0.0 |                      1.465752 |                           1.465752 |            20.510475 |                 20.510475 |     979.489525 |   1.178323 | -2.9 | 20.936779 |   4.420674 |             78.924011 |                      0.0 |                 78.924011 |                  78.924011 |                  6.818024 |                     0.0 |                      0.0 |    72.105987 |             0.0 |              0.0 |         72.105987 |             14.805032 |                   nan |                    500.0 |                   500.0 |   12.1 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  6.075004 |   2.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |   12.1 |   12.1 |  0.0 |  0.0 | 0.0 |  0.0 |   12.1 | 6.030624 | 0.069799 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 374.635735 |  0.0 |   12.0 |   0.1 |   0.0 |   0.0 |  0.0 | 5.967201 | 0.063424 |   0.0 |   0.0 |  0.0 |   0.0 | 0.069799 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |                13.765561 |       84.356885 |             89.6 |       2.6 |              1004.0 | 0.0 | 0.0 |    16.1 |                  89.6 |                   0.1 |                     13.765561 |            84.356885 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           2.084137 |          2.6 |                18.326791 |                     18.326791 |                         0.0 |                          0.0 |                      1.170311 |                           1.170311 |            16.420805 |                 16.420805 |     987.579195 |   1.201704 | -2.9 | 18.828425 |   16.37055 |             75.921197 |                      0.0 |                 75.921197 |                  75.921197 |                  8.199212 |                     0.0 |                      0.0 |    67.721985 |             0.0 |              0.0 |         67.721985 |             14.805032 |                   nan |                    500.0 |                   500.0 |  6.648 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  4.534094 |   2.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  6.648 |  6.648 |  0.0 |  0.0 | 0.0 |  0.0 |  6.648 | 7.436588 | 0.086072 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 355.365185 |  0.0 |  6.648 |   0.0 |   0.0 |   0.0 |  0.0 | 7.402977 | 0.033611 |   0.0 |   0.0 |  0.0 |   0.0 | 0.086072 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |                13.692273 |      142.309202 |             78.3 |       2.1 |              1011.0 | 0.0 | 0.0 |    17.0 |                  78.3 |                   5.0 |                     13.692273 |           142.309202 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                19.406929 |                     19.406929 |                         0.0 |                          0.0 |                      1.230421 |                           1.230421 |            15.195625 |                 15.195625 |     995.804375 |   1.206937 | -2.9 | 13.278507 |  -22.32896 |            128.078282 |                      0.0 |                128.078282 |                 128.078282 |                 21.775852 |                     0.0 |                      0.0 |    106.30243 |             0.0 |              0.0 |         106.30243 |             18.330039 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  7.752223 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 5.342084 |  0.06183 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 374.794145 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 5.339359 | 0.002725 |   0.0 |   0.0 |  0.0 |   0.0 |  0.06183 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |                13.618831 |      126.722078 |             78.5 |       2.1 |              1014.0 | 0.0 | 0.0 |    18.2 |                  78.5 |                   3.8 |                     13.618831 |           126.722078 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |           1.683342 |          2.1 |                20.933479 |                     20.933479 |                         0.0 |                          0.0 |                      1.314615 |                           1.314615 |            16.432781 |                 16.432781 |     997.567219 |   1.204994 | -2.9 | 18.874047 |   4.044283 |            114.049871 |                      0.0 |                114.049871 |                 114.049871 |                 17.491881 |                     0.0 |                      0.0 |     96.55799 |             0.0 |              0.0 |          96.55799 |             18.330039 |                   nan |                    500.0 |                   500.0 |    0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  7.710948 |   0.0 | 0.0 | 0.0 |   0.0 |          nan |          0.1 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |    0.0 |  0.0 |  0.0 | 0.0 |  0.0 |    0.0 | 2.849362 | 0.032979 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 367.849862 |  0.0 |    0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 2.849141 | 0.000221 |   0.0 |   0.0 |  0.0 |   0.0 | 0.032979 |

>>> round_(model.check_waterbalance(conditions))
0.0

.. _lland_v4_nadelw_summer_daily:

conifers (summer)
-----------------

.. integration-test::

    >>> lnk(NADELW)
    >>> surfaceresistance.nadelw = 56.0
    >>> cropheight.nadelw = 10.0
    >>> albedo.nadelw = 0.12
    >>> lai.nadelw = 11.0
    >>> conditions_nadelw_summer = test(
    ...     "lland_v4_nadelw_summer_daily",
    ...     axis1=(inputs.nied, fluxes.qah, states.bowa),
    ...     get_conditions="1997-08-03")
    |       date | nied | teml | sunshineduration | possiblesunshineduration | globalradiation | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration | dailypossiblesunshineduration | dailyglobalradiation |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair |    g |        tz |         wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |   nbes | sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |      evpo |   evi |      evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatinz | wlatsnow | wsensinz | wsenssnow | wsurfinz | wsurf | sff | fvg |   wada |      qdb | qib1 | qib2 | qbb |     qkap |     qdgz |       qah |        qa | inzp | stinz | sinz | esnowinz | asinz | wats | waes | esnow | taus |       ebdn |      bowa |    qdgz1 | qdgz2 | qigz1 | qigz2 |      qbgz |    qdga1 | qdga2 | qiga1 | qiga2 |      qbga | inlet |    outlet |
    ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-01-08 | 13.3 | 16.1 |              6.3 |                15.690785 |       190.25149 |             86.2 |       2.4 |              1007.0 | 0.0 | 0.0 |    16.1 |                  86.2 |                   6.3 |                     15.690785 |            190.25149 | 12.502 | 16.1 |    16.1 |    1.923819 |         1.923819 |           0.851977 |          2.4 |                18.326791 |                     18.326791 |                         0.0 |                          0.0 |                      1.170311 |                           1.170311 |            15.797694 |                 15.797694 |     991.202306 |   1.205601 | -2.9 | 11.733333 |      -26.2 |            167.421311 |                      nan |                 31.092529 |                 167.421311 |                 21.732733 |                     0.0 |                      0.0 |   145.688578 |             0.0 |              0.0 |        145.688578 |             39.166667 |                 100.0 |               211.364559 |              122.747678 | 10.302 |  0.0 |       49.0 |        0.42 | 4.32684 |     0.0 |      0.0 |             nan | 4.32684 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   4.68629 |   2.2 | 1.110342 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 10.302 | 0.657815 |  0.0 |  0.0 | 0.0 |      0.0 | 0.657815 |  0.169352 |   0.00196 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |      373.3 | 80.533843 | 0.657815 |   0.0 |   0.0 |   0.0 |       0.0 | 0.169352 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |   0.00196 |
    | 1997-02-08 |  0.1 | 17.3 |              1.7 |                15.628821 |      127.660782 |             85.8 |       1.6 |              1013.0 | 0.0 | 0.0 |    17.3 |                  85.8 |                   1.7 |                     15.628821 |           127.660782 |  0.094 | 17.3 |    17.3 |    1.282546 |         1.282546 |           0.567985 |          1.6 |                19.779139 |                     19.779139 |                         0.0 |                          0.0 |                      1.251029 |                           1.251029 |            16.970501 |                 16.970501 |     996.029499 |   1.207284 | -2.9 | 18.443733 |     6.8624 |            112.341488 |                      nan |                 20.863419 |                 112.341488 |                 11.219797 |                     0.0 |                      0.0 |   101.121692 |             0.0 |              0.0 |        101.121692 |                 58.75 |                 100.0 |               181.457771 |               115.35051 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  3.311865 | 0.094 | 1.809024 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 |  0.227814 |  0.002637 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |   363.5376 | 78.724819 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.227814 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.002637 |
    | 1997-03-08 |  0.0 | 18.9 |              4.5 |                15.566134 |      164.401298 |             82.4 |       1.5 |              1016.0 | 0.0 | 0.0 |    18.9 |                  82.4 |                   4.5 |                     15.566134 |           164.401298 |    0.0 | 18.9 |    18.9 |    1.202387 |         1.202387 |           0.532486 |          1.5 |                21.871576 |                     21.871576 |                         0.0 |                          0.0 |                      1.365939 |                           1.365939 |            18.022179 |                 18.022179 |     997.977821 |   1.203775 | -2.9 | 15.632162 | -19.607027 |            144.673142 |                      nan |                 26.867869 |                 144.673142 |                 16.044181 |                     0.0 |                      0.0 |   128.628961 |             0.0 |              0.0 |        128.628961 |             62.666667 |                 100.0 |               184.734223 |              115.873858 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  4.247909 |   0.0 | 2.494966 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 |  0.121564 |  0.001407 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 380.244627 | 76.229853 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.121564 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.001407 |
    | 1997-04-08 |  0.1 | 17.8 |             12.4 |                15.502754 |      269.020502 |             77.6 |       2.8 |              1017.0 | 0.0 | 0.0 |    17.8 |                  77.6 |                  12.4 |                     15.502754 |           269.020502 |  0.094 | 17.8 |    17.8 |    2.244455 |         2.244455 |           0.993973 |          2.8 |                20.413369 |                     20.413369 |                         0.0 |                          0.0 |                      1.286025 |                           1.286025 |            15.840774 |                 15.840774 |    1001.159226 |    1.21051 | -2.9 | 20.443786 |  15.862716 |            236.738041 |                      nan |                 43.965636 |                 236.738041 |                 35.980407 |                     0.0 |                      0.0 |   200.757635 |             0.0 |              0.0 |        200.757635 |             33.571429 |                 100.0 |               209.622013 |              121.261638 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  7.655384 | 0.094 | 3.243031 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 |  0.064868 |  0.000751 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 361.481911 | 72.986823 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.064868 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.000751 |
    | 1997-05-08 |  0.0 | 18.5 |             13.9 |                15.438712 |      287.848667 |             74.9 |       3.0 |              1018.0 | 0.0 | 0.0 |    18.5 |                  74.9 |                  13.9 |                     15.438712 |           287.848667 |    0.0 | 18.5 |    18.5 |    2.404774 |         2.404774 |           1.064971 |          3.0 |                21.331125 |                     21.331125 |                         0.0 |                          0.0 |                      1.336407 |                           1.336407 |            15.977013 |                 15.977013 |    1002.022987 |   1.208738 | -2.9 | 15.040124 | -20.759257 |            253.306827 |                      nan |                 47.042696 |                 253.306827 |                 39.523937 |                     0.0 |                      0.0 |   213.782891 |             0.0 |              0.0 |        213.782891 |             31.333333 |                 100.0 |               223.523903 |               123.68636 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  8.731356 |   0.0 | 3.618047 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 |  0.034614 |  0.000401 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 379.341169 | 69.368775 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.034614 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.000401 |
    | 1997-06-08 |  0.0 | 19.9 |             13.0 |                15.374034 |      274.674263 |             77.9 |       3.5 |              1018.0 | 0.0 | 0.0 |    19.9 |                  77.9 |                  13.0 |                     15.374034 |           274.674263 |    0.0 | 19.9 |    19.9 |    2.805569 |         2.805569 |           1.242466 |          3.5 |                23.275348 |                     23.275348 |                         0.0 |                          0.0 |                      1.442189 |                           1.442189 |            18.131496 |                 18.131496 |     999.868504 |   1.201995 | -2.9 |  20.18359 |   1.701539 |            241.713352 |                      nan |                 44.889622 |                 241.713352 |                 32.936813 |                     0.0 |                      0.0 |   208.776538 |             0.0 |              0.0 |        208.776538 |             26.857143 |                 100.0 |               217.805293 |              122.172051 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   8.96754 |   0.0 | 3.538533 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 |  0.018471 |  0.000214 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 374.739629 | 65.830242 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.018471 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.000214 |
    | 1997-07-08 |  0.0 | 20.8 |             13.8 |                 15.30875 |      284.047892 |             73.5 |       3.8 |              1019.0 | 0.0 | 0.0 |    20.8 |                  73.5 |                  13.8 |                      15.30875 |           284.047892 |    0.0 | 20.8 |    20.8 |    3.046047 |         3.046047 |           1.348964 |          3.8 |                24.605348 |                     24.605348 |                         0.0 |                          0.0 |                      1.513854 |                           1.513854 |            18.084931 |                 18.084931 |    1000.915069 |   1.199521 | -2.9 | 18.858347 | -11.649921 |            249.962145 |                      nan |                 46.421541 |                 249.962145 |                  35.31351 |                     0.0 |                      0.0 |   214.648635 |             0.0 |              0.0 |        214.648635 |             24.736842 |                 100.0 |               249.895771 |              127.503409 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 10.546373 |   0.0 | 3.943085 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 |  0.009856 |  0.000114 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  383.48955 | 61.887157 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.009856 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.000114 |
    | 1997-08-08 |  0.0 | 20.9 |             12.3 |                15.242885 |      262.971263 |             80.1 |       3.3 |              1019.0 | 0.0 | 0.0 |    20.9 |                  80.1 |                  12.3 |                     15.242885 |           262.971263 |    0.0 | 20.9 |    20.9 |    2.645251 |         2.645251 |           1.171468 |          3.3 |                 24.75714 |                      24.75714 |                         0.0 |                          0.0 |                      1.521999 |                           1.521999 |            19.830469 |                 19.830469 |     999.169531 |   1.198331 | -2.9 | 21.378324 |   2.869942 |            231.414712 |                      nan |                 42.977018 |                 231.414712 |                 28.392598 |                     0.0 |                      0.0 |   203.022114 |             0.0 |              0.0 |        203.022114 |             28.484848 |                 100.0 |               234.558928 |              124.470477 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  8.374876 |   0.0 | 3.447369 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 |  0.005259 |  0.000061 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 377.719608 | 58.439788 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.005259 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.000061 |
    | 1997-09-08 |  0.0 | 22.3 |             13.1 |                15.176467 |      272.256414 |             72.9 |       1.8 |              1019.0 | 0.0 | 0.0 |    22.3 |                  72.9 |                  13.1 |                     15.176467 |           272.256414 |    0.0 | 22.3 |    22.3 |    1.442864 |         1.442864 |           0.638983 |          1.8 |                 26.96965 |                      26.96965 |                         0.0 |                          0.0 |                      1.639967 |                           1.639967 |            19.660875 |                 19.660875 |     999.339125 |   1.192729 | -2.9 |  19.71658 | -15.500518 |            239.585644 |                      nan |                 44.494477 |                 239.585644 |                 31.114889 |                     0.0 |                      0.0 |   208.470755 |             0.0 |              0.0 |        208.470755 |             52.222222 |                 100.0 |               290.491428 |              132.398577 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  7.801219 |   0.0 | 4.238236 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 |  0.002806 |  0.000032 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 390.320126 | 54.201552 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.002806 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.000032 |
    | 1997-10-08 |  0.0 | 23.0 |             12.8 |                15.109521 |      267.019218 |             68.5 |       1.9 |              1020.0 | 0.0 | 0.0 |    23.0 |                  68.5 |                  12.8 |                     15.109521 |           267.019218 |    0.0 | 23.0 |    23.0 |    1.523023 |         1.523023 |           0.674482 |          1.9 |                28.139149 |                     28.139149 |                         0.0 |                          0.0 |                      1.701779 |                           1.701779 |            19.275317 |                 19.275317 |    1000.724683 |   1.191257 | -2.9 |  23.34553 |   2.073177 |            234.976912 |                      nan |                 43.638569 |                 234.976912 |                 32.010643 |                     0.0 |                      0.0 |   202.966269 |             0.0 |              0.0 |        202.966269 |             49.473684 |                 100.0 |               348.464183 |              138.317772 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  8.457253 |   0.0 | 4.453704 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 |  0.001498 |  0.000017 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 385.346949 | 49.747848 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.001498 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.000017 |
    | 1997-11-08 |  0.0 | 23.1 |             13.1 |                15.042071 |      269.626547 |             68.6 |       3.0 |              1021.0 | 0.0 | 0.0 |    23.1 |                  68.6 |                  13.1 |                     15.042071 |           269.626547 |    0.0 | 23.1 |    23.1 |    2.404774 |         2.404774 |           1.064971 |          3.0 |                28.309776 |                     28.309776 |                         0.0 |                          0.0 |                      1.710768 |                           1.710768 |            19.420506 |                 19.420506 |    1001.579494 |   1.191966 | -2.9 | 21.913255 |  -7.120473 |            237.271361 |                      nan |                 44.064681 |                 237.271361 |                 32.400014 |                     0.0 |                      0.0 |   204.871347 |             0.0 |              0.0 |        204.871347 |             31.333333 |                 100.0 |               368.722677 |              139.569846 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 10.478603 |   0.0 | 4.442272 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 |  0.000799 |  0.000009 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 389.567421 | 45.305576 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000799 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.000009 |
    | 1997-12-08 |  0.1 | 23.2 |             13.3 |                14.974141 |      270.878291 |             66.0 |       2.5 |              1022.0 | 0.0 | 0.0 |    23.2 |                  66.0 |                  13.3 |                     14.974141 |           270.878291 |  0.094 | 23.2 |    23.2 |    2.003978 |         2.003978 |           0.887476 |          2.5 |                28.481304 |                     28.481304 |                         0.0 |                          0.0 |                      1.719796 |                           1.719796 |             18.79766 |                  18.79766 |     1003.20234 |   1.193016 | -2.9 | 23.128751 |  -0.427496 |            238.372896 |                      nan |                 44.269252 |                 238.372896 |                 34.565783 |                     0.0 |                      0.0 |   203.807113 |             0.0 |              0.0 |        203.807113 |                  37.6 |                 100.0 |               419.707371 |              142.937668 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  9.985446 | 0.094 |  4.54598 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 |  0.000426 |  0.000005 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 387.094917 | 40.759596 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000426 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.000005 |
    | 1997-13-08 |  0.0 | 23.2 |             12.7 |                14.905755 |      261.697164 |             69.0 |       1.6 |              1019.0 | 0.0 | 0.0 |    23.2 |                  69.0 |                  12.7 |                     14.905755 |           261.697164 |    0.0 | 23.2 |    23.2 |    1.282546 |         1.282546 |           0.567985 |          1.6 |                28.481304 |                     28.481304 |                         0.0 |                          0.0 |                      1.719796 |                           1.719796 |              19.6521 |                   19.6521 |       999.3479 |    1.18911 | -2.9 | 22.416669 |  -4.699983 |            230.293504 |                      nan |                 42.768794 |                 230.293504 |                  31.34697 |                     0.0 |                      0.0 |   198.946534 |             0.0 |              0.0 |        198.946534 |                 58.75 |                 100.0 |               411.248888 |              141.648997 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  7.775446 |   0.0 | 4.347763 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 |  0.000228 |  0.000003 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |   388.8949 | 36.411834 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000228 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.000003 |
    | 1997-14-08 |  0.0 | 21.7 |             10.2 |                14.836934 |      227.962116 |             75.6 |       2.7 |              1015.0 | 0.0 | 0.0 |    21.7 |                  75.6 |                  10.2 |                     14.836934 |           227.962116 |    0.0 | 21.7 |    21.7 |    2.164296 |         2.164296 |           0.958474 |          2.7 |                 26.00118 |                      26.00118 |                         0.0 |                          0.0 |                        1.5885 |                             1.5885 |            19.656892 |                 19.656892 |     995.343108 |   1.190432 | -2.9 | 22.935065 |   7.410388 |            200.606662 |                      nan |                 37.255523 |                 200.606662 |                 25.918437 |                     0.0 |                      0.0 |   174.688225 |             0.0 |              0.0 |        174.688225 |             34.814815 |                 100.0 |               357.544238 |               136.70684 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  7.856911 |   0.0 | 3.459634 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 |  0.000121 |  0.000001 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 378.584513 | 32.952199 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.000121 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.000001 |
    | 1997-15-08 |  2.7 | 20.8 |              9.4 |                  14.7677 |      216.356321 |             81.4 |       3.2 |              1016.0 | 0.0 | 0.0 |    20.8 |                  81.4 |                   9.4 |                       14.7677 |           216.356321 |  2.538 | 20.8 |    20.8 |    2.565092 |         2.565092 |           1.135969 |          3.2 |                24.605348 |                     24.605348 |                         0.0 |                          0.0 |                      1.513854 |                           1.513854 |            20.028753 |                 20.028753 |     995.971247 |   1.195095 | -2.9 | 19.965673 |  -5.005962 |            190.393563 |                      nan |                 35.358805 |                 190.393563 |                 23.397902 |                     0.0 |                      0.0 |   166.995661 |             0.0 |              0.0 |        166.995661 |                29.375 |                 100.0 |               333.903254 |              133.915021 |  0.338 |  0.0 |       49.0 |        0.42 | 0.14196 |     0.0 |      0.0 |             nan | 0.14196 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  7.161899 |   2.2 | 1.984444 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.338 | 0.008728 |  0.0 |  0.0 | 0.0 |      0.0 | 0.008728 |  0.002312 |  0.000027 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 380.690475 | 31.297027 | 0.008728 |   0.0 |   0.0 |   0.0 |       0.0 | 0.002312 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.000027 |
    | 1997-16-08 |  0.0 | 19.5 |             10.3 |                14.698075 |      226.637904 |             79.0 |       1.3 |              1019.0 | 0.0 | 0.0 |    19.5 |                  79.0 |                  10.3 |                     14.698075 |           226.637904 |    0.0 | 19.5 |    19.5 |    1.042069 |         1.042069 |           0.461488 |          1.3 |                22.704676 |                     22.704676 |                         0.0 |                          0.0 |                      1.411269 |                           1.411269 |            17.936694 |                 17.936694 |    1001.063306 |   1.204916 | -2.9 |  20.57219 |   6.433141 |            199.441355 |                      nan |                 37.039109 |                 199.441355 |                 28.794151 |                     0.0 |                      0.0 |   170.647205 |             0.0 |              0.0 |        170.647205 |             72.307692 |                 100.0 |               354.689619 |              135.231167 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  5.274133 |   0.0 |  3.08192 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 |  0.003057 |  0.000035 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 371.357334 | 28.215107 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.003057 |   0.0 |   0.0 |   0.0 |       0.0 |   0.0 |  0.000035 |
    | 1997-17-08 |  0.0 | 21.0 |             11.1 |                14.628078 |      235.528503 |             75.9 |       1.9 |              1019.0 | 0.0 | 0.0 |    21.0 |                  75.9 |                  11.1 |                     14.628078 |           235.528503 |    0.0 | 21.0 |    21.0 |    1.523023 |         1.523023 |           0.674482 |          1.9 |                24.909749 |                     24.909749 |                         0.0 |                          0.0 |                      1.530181 |                           1.530181 |            18.906499 |                 18.906499 |    1000.093501 |   1.198338 | -2.9 | 17.884246 | -18.694526 |            207.265082 |                      nan |                 38.492087 |                 207.265082 |                 29.177702 |                     0.0 |                      0.0 |    178.08738 |             0.0 |              0.0 |         178.08738 |             49.473684 |                 100.0 |               400.417627 |              138.202815 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  6.725408 |   0.0 | 3.429476 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.166828 |      0.0 |  0.001332 |  0.000015 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 387.151861 |  24.95246 |      0.0 |   0.0 |   0.0 |   0.0 | -0.166828 | 0.001631 |   0.0 |   0.0 |   0.0 | -0.000299 |   0.0 |  0.000015 |
    | 1997-18-08 |  0.0 | 21.6 |             11.0 |                14.557728 |      232.849428 |             70.0 |       2.0 |              1018.0 | 0.0 | 0.0 |    21.6 |                  70.0 |                  11.0 |                     14.557728 |           232.849428 |    0.0 | 21.6 |    21.6 |    1.603182 |         1.603182 |           0.709981 |          2.0 |                25.842752 |                     25.842752 |                         0.0 |                          0.0 |                      1.580055 |                           1.580055 |            18.089927 |                 18.089927 |     999.910073 |   1.195081 | -2.9 | 22.433069 |   4.998415 |            204.907497 |                      nan |                 38.054249 |                 204.907497 |                 31.281288 |                     0.0 |                      0.0 |   173.626209 |             0.0 |              0.0 |        173.626209 |                  47.0 |                 100.0 |               491.312196 |              142.949327 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  7.539559 |   0.0 | 3.733791 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.369556 |      0.0 | -0.000389 | -0.000004 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 379.253445 | 21.588224 |      0.0 |   0.0 |   0.0 |   0.0 | -0.369556 |  0.00087 |   0.0 |   0.0 |   0.0 | -0.001259 |   0.0 | -0.000004 |
    | 1997-19-08 |  0.0 | 21.1 |              8.5 |                14.487046 |       199.71349 |             70.3 |       2.6 |              1019.0 | 0.0 | 0.0 |    21.1 |                  70.3 |                   8.5 |                     14.487046 |            199.71349 |    0.0 | 21.1 |    21.1 |    2.084137 |         2.084137 |           0.922975 |          2.6 |                25.063177 |                     25.063177 |                         0.0 |                          0.0 |                        1.5384 |                             1.5384 |            17.619414 |                 17.619414 |    1001.380586 |   1.198506 | -2.9 | 20.158326 |  -5.650046 |            175.747871 |                      nan |                  32.63889 |                 175.747871 |                 26.711296 |                     0.0 |                      0.0 |   149.036575 |             0.0 |              0.0 |        149.036575 |             36.153846 |                 100.0 |               525.371793 |               143.81531 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  7.831862 |   0.0 | 3.377664 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.578596 |      0.0 | -0.002489 | -0.000029 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 382.003492 | 18.789156 |      0.0 |   0.0 |   0.0 |   0.0 | -0.578596 | 0.000465 |   0.0 |   0.0 |   0.0 | -0.002953 |   0.0 | -0.000029 |
    | 1997-20-08 |  0.0 | 22.4 |             11.3 |                14.416048 |      233.771902 |             69.0 |       3.6 |              1019.0 | 0.0 | 0.0 |    22.4 |                  69.0 |                  11.3 |                     14.416048 |           233.771902 |    0.0 | 22.4 |    22.4 |    2.885728 |         2.885728 |           1.277965 |          3.6 |                27.134082 |                     27.134082 |                         0.0 |                          0.0 |                      1.648679 |                           1.648679 |            18.722517 |                 18.722517 |    1000.277483 |   1.192743 | -2.9 | 20.950339 |  -8.697966 |            205.719274 |                      nan |                 38.205008 |                 205.719274 |                 31.116622 |                     0.0 |                      0.0 |   174.602652 |             0.0 |              0.0 |        174.602652 |             26.111111 |                 100.0 |               629.512602 |              146.962511 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 10.542746 |   0.0 | 3.855538 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.752518 |      0.0 | -0.005079 | -0.000059 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 387.801458 | 15.686136 |      0.0 |   0.0 |   0.0 |   0.0 | -0.752518 | 0.000248 |   0.0 |   0.0 |   0.0 | -0.005327 |   0.0 | -0.000059 |
    | 1997-21-08 |  0.0 | 22.5 |             12.4 |                14.344752 |      246.121286 |             62.1 |       2.5 |              1017.0 | 0.0 | 0.0 |    22.5 |                  62.1 |                  12.4 |                     14.344752 |           246.121286 |    0.0 | 22.5 |    22.5 |    2.003978 |         2.003978 |           0.887476 |          2.5 |                27.299387 |                     27.299387 |                         0.0 |                          0.0 |                      1.657431 |                           1.657431 |             16.95292 |                  16.95292 |     1000.04708 |   1.190771 | -2.9 | 22.620153 |    0.72092 |            216.586732 |                      nan |                  40.22325 |                 216.586732 |                 38.297604 |                     0.0 |                      0.0 |   178.289128 |             0.0 |              0.0 |        178.289128 |                  37.6 |                 100.0 |               884.445242 |              152.119959 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  9.796296 |   0.0 | 4.291217 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 0.945327 |      0.0 | -0.008217 | -0.000095 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 384.180539 | 12.340246 |      0.0 |   0.0 |   0.0 |   0.0 | -0.945327 | 0.000132 |   0.0 |   0.0 |   0.0 |  -0.00835 |   0.0 | -0.000095 |
    | 1997-22-08 |  2.0 | 19.6 |              0.1 |                14.273175 |        90.58551 |             84.5 |       2.1 |              1015.0 | 0.0 | 0.0 |    19.6 |                  84.5 |                   0.1 |                     14.273175 |             90.58551 |   1.88 | 19.6 |    19.6 |    1.683342 |         1.683342 |            0.74548 |          2.1 |                22.846186 |                     22.846186 |                         0.0 |                          0.0 |                      1.418946 |                           1.418946 |            19.305027 |                 19.305027 |     995.694973 |   1.199129 | -2.9 | 21.577328 |  11.863971 |             79.715249 |                      nan |                  14.80426 |                  79.715249 |                   7.02118 |                     0.0 |                      0.0 |    72.694069 |             0.0 |              0.0 |         72.694069 |             44.761905 |                 100.0 |               695.922288 |              147.245828 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  3.396916 |  1.88 | 0.698036 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 1.153227 |      0.0 | -0.012009 | -0.000139 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 369.416568 | 12.795437 |      0.0 |   0.0 |   0.0 |   0.0 | -1.153227 | 0.000071 |   0.0 |   0.0 |   0.0 | -0.012079 |   0.0 | -0.000139 |
    | 1997-23-08 |  0.3 | 20.3 |              6.7 |                14.201333 |      171.944993 |             83.6 |       1.5 |              1015.0 | 0.0 | 0.0 |    20.3 |                  83.6 |                   6.7 |                     14.201333 |           171.944993 |  0.282 | 20.3 |    20.3 |    1.202387 |         1.202387 |           0.532486 |          1.5 |                23.858503 |                     23.858503 |                         0.0 |                          0.0 |                      1.473678 |                           1.473678 |            19.945708 |                 19.945708 |     995.054292 |   1.195981 | -2.9 | 17.325305 | -17.848171 |            151.311594 |                      nan |                 28.100725 |                 151.311594 |                 18.994889 |                     0.0 |                      0.0 |   132.316705 |             0.0 |              0.0 |        132.316705 |             62.666667 |                 100.0 |               667.605003 |              145.666152 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  4.351769 | 0.282 | 2.206422 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 1.124944 |      0.0 | -0.016079 | -0.000186 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 384.364739 | 11.713959 |      0.0 |   0.0 |   0.0 |   0.0 | -1.124944 | 0.000038 |   0.0 |   0.0 |   0.0 | -0.016117 |   0.0 | -0.000186 |
    | 1997-24-08 |  0.0 | 24.1 |             10.4 |                14.129243 |      216.549813 |             76.5 |       2.2 |              1011.0 | 0.0 | 0.0 |    24.1 |                  76.5 |                  10.4 |                     14.129243 |           216.549813 |    0.0 | 24.1 |    24.1 |    1.763501 |         1.763501 |           0.780979 |          2.2 |                30.066256 |                     30.066256 |                         0.0 |                          0.0 |                       1.80287 |                            1.80287 |            23.000686 |                 23.000686 |     987.999314 |   1.174651 | -2.9 | 21.630378 | -14.817732 |            190.563835 |                      nan |                 35.390427 |                 190.563835 |                 21.629737 |                     0.0 |                      0.0 |   168.934098 |             0.0 |              0.0 |        168.934098 |             42.727273 |                 100.0 |               938.332509 |              150.288557 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  7.235935 |   0.0 | 3.494124 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 1.192142 |      0.0 |  -0.02019 | -0.000234 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  396.28247 |  9.411977 |      0.0 |   0.0 |   0.0 |   0.0 | -1.192142 |  0.00002 |   0.0 |   0.0 |   0.0 |  -0.02021 |   0.0 | -0.000234 |
    | 1997-25-08 |  0.1 | 23.7 |              6.5 |                14.056918 |       166.91683 |             76.6 |       1.9 |              1008.0 | 0.0 | 0.0 |    23.7 |                  76.6 |                   6.5 |                     14.056918 |            166.91683 |  0.094 | 23.7 |    23.7 |    1.523023 |         1.523023 |           0.674482 |          1.9 |                29.352596 |                     29.352596 |                         0.0 |                          0.0 |                      1.765541 |                           1.765541 |            22.484088 |                 22.484088 |     985.515912 |   1.172942 | -2.9 | 25.062685 |   8.176109 |             146.88681 |                      nan |                 27.278979 |                  146.88681 |                 16.222254 |                     0.0 |                      0.0 |   130.664556 |             0.0 |              0.0 |        130.664556 |             49.473684 |                 100.0 |               1353.41939 |              153.707473 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  5.830169 | 0.094 | 2.901389 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 | 1.335178 |      0.0 | -0.024654 | -0.000285 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 385.206362 |  7.845767 |      0.0 |   0.0 |   0.0 |   0.0 | -1.335178 | 0.000011 |   0.0 |   0.0 |   0.0 | -0.024665 |   0.0 | -0.000285 |
    | 1997-26-08 | 10.5 | 21.1 |              4.9 |                13.984374 |      145.985611 |             84.2 |       2.3 |              1006.0 | 0.0 | 0.0 |    21.1 |                  84.2 |                   4.9 |                     13.984374 |           145.985611 |   9.87 | 21.1 |    21.1 |     1.84366 |          1.84366 |           0.816478 |          2.3 |                25.063177 |                     25.063177 |                         0.0 |                          0.0 |                        1.5384 |                             1.5384 |            21.103195 |                 21.103195 |     984.896805 |   1.181557 | -2.9 | 21.872765 |   4.636593 |            128.467338 |                      nan |                  23.85822 |                 128.467338 |                 14.602883 |                     0.0 |                      0.0 |   113.864455 |             0.0 |              0.0 |        113.864455 |             40.869565 |                 100.0 |              1693.426996 |              154.753913 |   7.67 |  0.0 |       49.0 |        0.42 |  3.2214 |     0.0 |      0.0 |             nan |  3.2214 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  4.656397 |   2.2 | 1.080885 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |   7.67 | 0.067763 |  0.0 |  0.0 | 0.0 | 1.432496 | 0.067763 | -0.012083 |  -0.00014 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 377.669769 | 15.799615 | 0.067763 |   0.0 |   0.0 |   0.0 | -1.432496 | 0.017451 |   0.0 |   0.0 |   0.0 | -0.029534 |   0.0 |  -0.00014 |
    | 1997-27-08 |  6.5 | 20.9 |              6.6 |                13.911624 |       165.54389 |             85.8 |       2.1 |              1005.0 | 0.0 | 0.0 |    20.9 |                  85.8 |                   6.6 |                     13.911624 |            165.54389 |   6.11 | 20.9 |    20.9 |    1.683342 |         1.683342 |            0.74548 |          2.1 |                 24.75714 |                      24.75714 |                         0.0 |                          0.0 |                      1.521999 |                           1.521999 |            21.241626 |                 21.241626 |     983.758374 |   1.181114 | -2.9 | 19.702227 |   -7.18664 |            145.678623 |                      nan |                 27.054602 |                 145.678623 |                 17.349121 |                     0.0 |                      0.0 |   128.329502 |             0.0 |              0.0 |        128.329502 |             44.761905 |                 100.0 |               514.330605 |              137.506981 |   3.91 |  0.0 |       49.0 |        0.42 |  1.6422 |     0.0 |      0.0 |             nan |  1.6422 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  4.603714 |   2.2 | 1.172574 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |   3.91 | 0.052958 |  0.0 |  0.0 | 0.0 | 0.938276 | 0.052958 |   0.00343 |   0.00004 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 381.956408 | 19.422358 | 0.052958 |   0.0 |   0.0 |   0.0 | -0.938276 | 0.037105 |   0.0 |   0.0 |   0.0 | -0.033675 |   0.0 |   0.00004 |
    | 1997-28-08 | 15.0 | 20.2 |              0.3 |                13.838682 |       87.693346 |             86.5 |       2.6 |              1000.0 | 0.0 | 0.0 |    20.2 |                  86.5 |                   0.3 |                     13.838682 |            87.693346 |   14.1 | 20.2 |    20.2 |    2.084137 |         2.084137 |           0.922975 |          2.6 |                23.711531 |                     23.711531 |                         0.0 |                          0.0 |                      1.465752 |                           1.465752 |            20.510475 |                 20.510475 |     979.489525 |   1.178323 | -2.9 | 20.936779 |   4.420674 |             77.170144 |                      nan |                 14.331598 |                  77.170144 |                  6.818024 |                     0.0 |                      0.0 |     70.35212 |             0.0 |              0.0 |          70.35212 |             36.153846 |                 100.0 |               423.374916 |              132.523752 |   11.9 |  0.0 |       49.0 |        0.42 |   4.998 |     0.0 |      0.0 |             nan |   4.998 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  3.480087 |   2.2 | 0.567488 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |   11.9 | 0.232109 |  0.0 |  0.0 | 0.0 | 0.713174 | 0.232109 |  0.054108 |  0.000626 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 374.635735 | 31.235935 | 0.232109 |   0.0 |   0.0 |   0.0 | -0.713174 |  0.09062 |   0.0 |   0.0 |   0.0 | -0.036512 |   0.0 |  0.000626 |
    | 1997-29-08 |  9.2 | 16.1 |              0.1 |                13.765561 |       84.356885 |             89.6 |       2.6 |              1004.0 | 0.0 | 0.0 |    16.1 |                  89.6 |                   0.1 |                     13.765561 |            84.356885 |  8.648 | 16.1 |    16.1 |    2.084137 |         2.084137 |           0.922975 |          2.6 |                18.326791 |                     18.326791 |                         0.0 |                          0.0 |                      1.170311 |                           1.170311 |            16.420805 |                 16.420805 |     987.579195 |   1.201704 | -2.9 | 18.828425 |   16.37055 |             74.234059 |                      nan |                 13.786325 |                  74.234059 |                  8.199212 |                     0.0 |                      0.0 |    66.034847 |             0.0 |              0.0 |         66.034847 |             36.153846 |                 100.0 |               347.066882 |              127.000167 |  6.448 |  0.0 |       49.0 |        0.42 | 2.70816 |     0.0 |      0.0 |             nan | 2.70816 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  2.684886 |   2.2 | 0.203734 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  6.448 | 0.173151 |  0.0 |  0.0 | 0.0 |      0.0 | 0.173151 |  0.103772 |  0.001201 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 355.365185 | 37.307051 | 0.173151 |   0.0 |   0.0 |   0.0 |       0.0 |  0.14143 |   0.0 |   0.0 |   0.0 | -0.037658 |   0.0 |  0.001201 |
    | 1997-30-08 |  0.0 | 17.0 |              5.0 |                13.692273 |      142.309202 |             78.3 |       2.1 |              1011.0 | 0.0 | 0.0 |    17.0 |                  78.3 |                   5.0 |                     13.692273 |           142.309202 |    0.0 | 17.0 |    17.0 |    1.683342 |         1.683342 |            0.74548 |          2.1 |                19.406929 |                     19.406929 |                         0.0 |                          0.0 |                      1.230421 |                           1.230421 |            15.195625 |                 15.195625 |     995.804375 |   1.206937 | -2.9 | 13.278507 |  -22.32896 |            125.232098 |                      nan |                  23.25739 |                 125.232098 |                 21.775852 |                     0.0 |                      0.0 |   103.456246 |             0.0 |              0.0 |        103.456246 |             44.761905 |                 100.0 |               346.870428 |              126.427648 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   4.53294 |   0.0 | 2.160883 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 |  0.074124 |  0.000858 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 374.794145 | 35.146168 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.111647 |   0.0 |   0.0 |   0.0 | -0.037523 |   0.0 |  0.000858 |
    | 1997-31-08 |  0.0 | 18.2 |              3.8 |                13.618831 |      126.722078 |             78.5 |       2.1 |              1014.0 | 0.0 | 0.0 |    18.2 |                  78.5 |                   3.8 |                     13.618831 |           126.722078 |    0.0 | 18.2 |    18.2 |    1.683342 |         1.683342 |            0.74548 |          2.1 |                20.933479 |                     20.933479 |                         0.0 |                          0.0 |                      1.314615 |                           1.314615 |            16.432781 |                 16.432781 |     997.567219 |   1.204994 | -2.9 | 18.874047 |   4.044283 |            111.515429 |                      nan |                 20.710008 |                 111.515429 |                 17.491881 |                     0.0 |                      0.0 |    94.023548 |             0.0 |              0.0 |         94.023548 |             44.761905 |                 100.0 |               345.930896 |              125.804674 |    0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  4.439668 |   0.0 | 2.165731 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |    0.0 |      0.0 |  0.0 |  0.0 | 0.0 |      0.0 |      0.0 |  0.022187 |  0.000257 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 367.849862 | 32.980437 |      0.0 |   0.0 |   0.0 |   0.0 |       0.0 | 0.059576 |   0.0 |   0.0 |   0.0 | -0.037389 |   0.0 |  0.000257 |

>>> round_(model.check_waterbalance(conditions))
0.0

.. _lland_v4_acker_winter_daily:

acre (winter)
-------------

>>> pub.timegrids = "2010-12-01", "2011-01-01", "1d"
>>> lnk(ACKER)
>>> lai.acker_dec = 0.3
>>> cropheight.acker_dec = 0.05
>>> albedo.acker_dec = 0.1
>>> wg2z.dec = 8.583333333333
>>> test = IntegrationTest(land)
>>> test.dateformat = "%Y-%d-%m"

>>> test.inits = (
...     (states.inzp, 0.0),
...     (states.stinz, 0.0),
...     (states.sinz, 0.0),
...     (states.esnowinz, 0.0),
...     (states.asinz, 0.0),
...     (states.wats, 0.0),
...     (states.waes, 0.0),
...     (states.esnow, 0.0),
...     (states.taus, 0.0),
...     (states.ebdn, 0.0),
...     (states.bowa, 72.0),
...     (states.qdgz1, 0.0),
...     (states.qdgz2, 0.0),
...     (states.qigz1, 0.0),
...     (states.qigz2, 0.0),
...     (states.qbgz, 0.0),
...     (states.qdga1, 0.0),
...     (states.qdga2, 0.0),
...     (states.qiga1, 0.0),
...     (states.qiga2, 0.0),
...     (states.qbga, 0.0))

>>> inputs.nied.series = (
...     0.0, 11.1, 0.4, 1.2, 6.4, 0.0, 0.0, 0.0, 0.7, 3.3, 9.4, 0.2, 2.9, 0.0, 0.0,
...     6.7, 0.0, 0.1, 0.0, 0.2, 0.2, 2.4, 7.0, 0.2, 0.0, 1.2, 0.9, 0.0, 0.0, 0.2, 0.9)
>>> inputs.teml.series = (
...     -5.8, -3.0, -5.6, -3.8, 0.2, -1.1, -2.9, -2.9, -2.1, -3.1, 5.1, 0.2, -3.5,
...     -3.9, -6.3, -3.5, -7.3, -5.8, -7.9, -5.7, -5.6, -4.8, -1.1, -2.9, -9.1, -8.6,
...     -5.3, -7.7, -9.2, -7.4, 1.7)
>>> inputs.sunshineduration.series = (
...     3.5, 0.0, 0.1, 3.7, 0.0, 0.1, 0.0, 0.0, 4.4, 0.1, 0.2, 0.7, 0.0, 1.2, 3.0, 0.0,
...     0.0, 0.0, 0.0, 0.2, 0.8, 0.0, 0.0, 0.0, 3.2, 4.3, 0.0, 3.4, 0.0, 0.0, 4.0)
>>> inputs.windspeed.series = (
...     6.9, 5.2, 2.0, 4.3, 4.0, 2.4, 1.9, 1.8, 2.7, 3.0, 7.4, 5.7, 2.2, 3.1, 2.8, 3.7,
...     0.9, 1.3, 1.9, 2.4, 2.3, 4.6, 8.6, 7.6, 3.8, 1.5, 1.3, 3.0, 2.7, 1.4, 4.2)
>>> inputs.relativehumidity.series = (
...     70.3, 85.0, 95.9, 90.6, 96.5, 99.0, 99.2, 99.1, 96.9, 95.0, 90.6, 78.7, 90.5,
...     88.1, 92.8, 98.1, 96.6, 97.7, 95.8, 97.5, 95.0, 93.8, 94.9, 85.4, 87.2, 92.7,
...     97.8, 95.4, 94.9, 96.0, 96.5)
>>> inputs.atmosphericpressure.series = (
...     1015.0, 1007.0, 1005.0, 1008.0, 996.0, 998.0, 1004.0, 1006.0, 1009.0, 1018.0,
...     1004.0, 1012.0, 1020.0, 1025.0, 1023.0, 1000.0, 991.0, 997.0, 997.0, 997.0,
...     1004.0, 1007.0, 1004.0, 1008.0, 1018.0, 1019.0, 1017.0, 1022.0, 1024.0, 1021.0,
...     1016.0)
>>> inlet.sequences.sim.series = 0.0

>>> inputs.possiblesunshineduration.series = (
...     7.64102946, 7.6096444, 7.57982377, 7.55159603, 7.52498861, 7.50002776,
...     7.47673851, 7.45514453, 7.43526809, 7.41712991, 7.40074913, 7.38614323,
...     7.3733279, 7.36231706, 7.35312273, 7.34575502, 7.34022207, 7.33653002,
...     7.33468296, 7.33468296, 7.33653002, 7.34022207, 7.34575502, 7.35312273,
...     7.36231706, 7.3733279, 7.38614323, 7.40074913, 7.41712991, 7.43526809,
...     7.45514453)
>>> inputs.globalradiation.series = (
...     31.9427981, 9.8738984, 16.6834998, 31.8074275, 9.5325105, 16.1414844,
...     9.3414287, 9.2567644, 33.405815, 15.5911584, 15.8915442, 17.8235189, 8.941324,
...     19.6680679, 26.8315815, 8.8377384, 8.817416, 8.8041854, 8.7980439, 15.4647459,
...     17.8795412, 8.8221612, 8.8443961, 8.8737425, 27.7596417, 32.3287553, 9.0045823,
...     28.9823209, 9.1276454, 9.1999895, 32.0622345)

.. integration-test::

    >>> conditions_acker_winter = test(
    ...     "lland_v4_acker_winter_daily",
    ...     axis1=(inputs.nied, fluxes.wada), axis2=(states.waes, states.wats),
    ...     get_conditions="2010-12-10")
    |       date | nied | teml | sunshineduration | possiblesunshineduration | globalradiation | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration | dailypossiblesunshineduration | dailyglobalradiation |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair |        g |         tz |         wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |   nbes |    sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |      evpo | evi |     evb |       evs |     wnied | tempssurface | actualalbedo |   schmpot |      schm |  gefrpot |     gefr | wlatinz |   wlatsnow | wsensinz |  wsenssnow | wsurfinz |       wsurf |      sff |      fvg |      wada |      qdb | qib1 | qib2 | qbb | qkap |     qdgz |      qah |       qa | inzp | stinz | sinz | esnowinz | asinz |      wats |      waes |     esnow |     taus |       ebdn |      bowa |    qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    -----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 2010-01-12 |  0.0 | -5.8 |              3.5 |                 7.641029 |       31.942798 |             70.3 |       6.9 |              1015.0 | 0.0 | 0.0 |    -5.8 |                  70.3 |                   3.5 |                      7.641029 |            31.942798 |    0.0 | -5.8 |    -5.8 |     5.53098 |          5.53098 |            5.53098 |          6.9 |                 3.958114 |                      3.958114 |                         0.0 |                          0.0 |                      0.303557 |                           0.303557 |             2.782554 |                  2.782554 |    1012.217446 |   1.321192 | 8.583333 |        0.0 |       34.8 |             28.748518 |                      0.0 |                 28.748518 |                  28.748518 |                 46.472798 |                     0.0 |                      0.0 |    -17.72428 |             0.0 |              0.0 |         -17.72428 |             30.870386 |                 100.0 |               111.309423 |               99.512695 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.808179 | 0.0 | 0.55374 |       0.0 |       0.0 |          nan |          0.1 |       0.0 |       0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan | -26.216667 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-02-12 | 11.1 | -3.0 |              0.0 |                 7.609644 |        9.873898 |             85.0 |       5.2 |              1007.0 | 0.0 | 0.0 |    -3.0 |                  85.0 |                   0.0 |                      7.609644 |             9.873898 | 10.434 | -3.0 |    -3.0 |    4.168274 |         4.168274 |           4.168274 |          5.2 |                 4.893489 |                      4.893489 |                         0.0 |                     3.818807 |                      0.366257 |                           0.366257 |             4.159465 |                  4.159465 |    1002.840535 |   1.296511 | 8.583333 |    -7.5504 |   -7.70501 |              1.229506 |                      0.0 |                  1.229506 |                   1.229506 |                  14.78976 |                     0.0 |                27.857608 |   -13.560254 |             0.0 |       -26.628102 |        -13.560254 |             40.962628 |                 100.0 |               112.102145 |               99.530051 | 10.374 |  10.374 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.779906 | 0.0 |     0.0 | -0.168357 | -0.752835 |    -6.266232 |     0.875479 |       0.0 |       0.0 | 0.406773 |      0.0 |     0.0 |  -5.197724 |      0.0 | -28.315743 |      0.0 |   -6.885365 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 10.542357 | 10.542357 |  -1.57248 | 0.084415 |  -9.928323 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-03-12 |  0.4 | -5.6 |              0.1 |                 7.579824 |         16.6835 |             95.9 |       2.0 |              1005.0 | 0.0 | 0.0 |    -5.6 |                  95.9 |                   0.1 |                      7.579824 |              16.6835 |  0.376 | -5.6 |    -5.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 4.019239 |                      4.019239 |                         0.0 |                      3.57079 |                      0.307706 |                           0.307706 |              3.85445 |                   3.85445 |     1001.14555 |   1.306656 | 8.583333 |  -2.859357 |  25.653408 |              2.321007 |                      0.0 |                  2.321007 |                   2.321007 |                 15.387201 |                     0.0 |                37.521585 |   -13.066194 |             0.0 |       -35.200578 |        -13.066194 |            106.502833 |                 100.0 |               112.102145 |               99.527226 |  0.376 |   0.376 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.039557 | 0.0 |     0.0 | -0.073821 | -0.050934 |    -7.134925 |      0.86088 |       0.0 |       0.0 | 0.487468 |      0.0 |     0.0 |  -2.279085 |      0.0 |  -7.007074 |      0.0 |    25.91442 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 10.992177 | 10.992177 | -1.884427 |   0.1418 | -26.998397 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-04-12 |  1.2 | -3.8 |              3.7 |                 7.551596 |       31.807427 |             90.6 |       4.3 |              1008.0 | 0.0 | 0.0 |    -3.8 |                  90.6 |                   3.7 |                      7.551596 |            31.807427 |  1.128 | -3.8 |    -3.8 |    3.446842 |         3.446842 |           3.446842 |          4.3 |                 4.608124 |                      4.608124 |                         0.0 |                     3.236114 |                      0.347298 |                           0.347298 |              4.17496 |                   4.17496 |     1003.82504 |   1.301648 | 8.583333 |  -7.775538 |    3.72366 |               4.73374 |                      0.0 |                   4.73374 |                    4.73374 |                  43.14024 |                     0.0 |                55.888201 |     -38.4065 |             0.0 |       -51.154461 |          -38.4065 |             49.536202 |                 100.0 |               112.102145 |               99.524551 |  1.128 |   1.128 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.092108 | 0.0 |     0.0 | -0.402208 | -0.103687 |    -8.396148 |     0.851175 |       0.0 |       0.0 | 0.636779 |      0.0 |     0.0 | -12.417475 |      0.0 | -34.539816 |      0.0 |     4.19717 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 12.522386 | 12.522386 | -2.461624 | 0.183432 | -22.138724 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-05-12 |  6.4 |  0.2 |              0.0 |                 7.524989 |        9.532511 |             96.5 |       4.0 |               996.0 | 0.0 | 0.0 |     0.2 |                  96.5 |                   0.0 |                      7.524989 |             9.532511 |  6.016 |  0.2 |     0.2 |    3.206365 |         3.206365 |           3.206365 |          4.0 |                 6.197477 |                      6.197477 |                         0.0 |                      4.91255 |                      0.451275 |                           0.451275 |             5.980565 |                  5.980565 |     990.019435 |   1.266438 | 8.583333 |  -6.375953 |  -20.56744 |              1.460936 |                      0.0 |                  1.460936 |                   1.460936 |                 13.205056 |                     0.0 |                 15.88459 |   -11.744121 |             0.0 |       -14.423654 |        -11.744121 |             53.251417 |                 100.0 |               112.102145 |               99.522031 |  6.016 |  2.7072 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.134777 | 0.0 |     0.0 | -0.434119 |  0.045165 |    -2.948046 |     0.846742 |       0.0 |       0.0 | 0.403221 | 0.403221 |     0.0 | -13.402657 |      0.0 | -22.446147 |      0.0 |  -21.425149 | 0.977326 |  0.28655 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 15.998885 | 18.972504 |       0.0 | 0.203476 |   7.012049 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-06-12 |  0.0 | -1.1 |              0.1 |                 7.500028 |       16.141484 |             99.0 |       2.4 |               998.0 | 0.0 | 0.0 |    -1.1 |                  99.0 |                   0.1 |                      7.500028 |            16.141484 |    0.0 | -1.1 |    -1.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 5.634754 |                      5.634754 |                         0.0 |                     5.151567 |                       0.41489 |                            0.41489 |             5.578406 |                  5.578406 |     992.421594 |   1.275245 | 8.583333 |        0.0 |  13.867616 |              3.011222 |                      0.0 |                  3.011222 |                   3.011222 |                 14.093918 |                     0.0 |                28.072188 |   -11.082695 |             0.0 |       -25.060966 |        -11.082695 |             88.752361 |                 100.0 |               112.102145 |               99.519666 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.000586 | 0.0 |     0.0 | -0.123565 |       0.0 |    -2.311269 |     0.813448 |       0.0 |       0.0 | 0.317541 | 0.317541 |     0.0 |   -3.81486 |      0.0 |  -6.150959 |      0.0 |   15.095147 | 0.994413 | 0.296657 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 16.422692 | 19.096069 |       0.0 | 0.378868 |   1.727766 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-07-12 |  0.0 | -2.9 |              0.0 |                 7.476739 |        9.341429 |             99.2 |       1.9 |              1004.0 | 0.0 | 0.0 |    -2.9 |                  99.2 |                   0.0 |                      7.476739 |             9.341429 |    0.0 | -2.9 |    -2.9 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 4.930236 |                      4.930236 |                         0.0 |                     4.623648 |                      0.368688 |                           0.368688 |             4.890794 |                  4.890794 |     999.109206 |   1.291808 | 8.583333 |        0.0 |   22.53221 |              1.983477 |                      0.0 |                  1.983477 |                   1.983477 |                 13.815559 |                     0.0 |                32.375282 |   -11.832082 |             0.0 |       -30.391805 |        -11.832082 |            112.108246 |                 100.0 |               112.102145 |                99.51746 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.015911 | 0.0 |     0.0 |  -0.06757 |       0.0 |    -3.755368 |     0.787669 |       0.0 |       0.0 | 0.511768 | 0.511768 |     0.0 |  -2.086102 |      0.0 |   -3.79513 |      0.0 |   24.510573 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 16.994381 | 19.163639 |       0.0 |  0.55426 | -12.221111 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-08-12 |  0.0 | -2.9 |              0.0 |                 7.455145 |        9.256764 |             99.1 |       1.8 |              1006.0 | 0.0 | 0.0 |    -2.9 |                  99.1 |                   0.0 |                      7.455145 |             9.256764 |    0.0 | -2.9 |    -2.9 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 4.930236 |                      4.930236 |                         0.0 |                     4.223622 |                      0.368688 |                           0.368688 |             4.885864 |                  4.885864 |    1001.114136 |   1.294388 | 8.583333 |   -3.51968 |   8.583604 |              2.155738 |                      0.0 |                  2.155738 |                   2.155738 |                 13.821847 |                     0.0 |                27.150377 |   -11.666109 |             0.0 |       -24.994639 |        -11.666109 |            118.336481 |                 100.0 |               112.102145 |               99.515415 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.012641 | 0.0 |     0.0 |  -0.16266 |       0.0 |    -4.950281 |     0.767118 |       0.0 |       0.0 | 0.661033 | 0.661033 |     0.0 |   -5.02185 |      0.0 |  -8.833803 |      0.0 |   11.138986 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 17.805273 | 19.326299 |       0.0 | 0.729652 | -12.221381 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-09-12 |  0.7 | -2.1 |              4.4 |                 7.435268 |       33.405815 |             96.9 |       2.7 |              1009.0 | 0.0 | 0.0 |    -2.1 |                  96.9 |                   4.4 |                      7.435268 |            33.405815 |  0.658 | -2.1 |    -2.1 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 5.233109 |                      5.233109 |                         0.0 |                     3.897182 |                      0.388644 |                           0.388644 |             5.070883 |                  5.070883 |    1003.929117 |   1.294334 | 8.583333 |  -3.519758 |  14.894088 |              8.184668 |                      0.0 |                  8.184668 |                   8.184668 |                 46.364476 |                     0.0 |                58.769843 |   -38.179808 |             0.0 |       -50.585175 |        -38.179808 |             78.890988 |                 100.0 |               112.102145 |               99.513532 |  0.658 |   0.658 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.235007 | 0.0 |     0.0 | -0.365518 | -0.033425 |    -6.002106 |     0.754993 |       0.0 |       0.0 | 0.807885 | 0.807885 |     0.0 | -11.284727 |      0.0 | -21.316713 |      0.0 |   17.983735 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 19.623632 | 20.349817 |       0.0 |  0.85302 | -18.532136 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-10-12 |  3.3 | -3.1 |              0.1 |                  7.41713 |       15.591158 |             95.0 |       3.0 |              1018.0 | 0.0 | 0.0 |    -3.1 |                  95.0 |                   0.1 |                       7.41713 |            15.591158 |  3.102 | -3.1 |    -3.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 4.856984 |                      4.856984 |                         0.0 |                     4.026976 |                       0.36384 |                            0.36384 |             4.614135 |                  4.614135 |    1013.385865 |   1.310959 | 8.583333 |  -5.337255 |   1.425738 |              3.649886 |                      0.0 |                  3.649886 |                   3.649886 |                 14.892721 |                     0.0 |                28.618996 |   -11.242835 |             0.0 |        -24.96911 |        -11.242835 |             71.001889 |                 100.0 |               112.102145 |               99.511814 |  3.102 |   3.102 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.137189 | 0.0 |     0.0 | -0.195734 | -0.232614 |    -5.574878 |       0.7659 |       0.0 |       0.0 | 0.843525 | 0.726185 |     0.0 |  -6.042946 |      0.0 | -14.472191 |      0.0 |    4.453973 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 23.647551 | 23.647551 | -0.453606 | 0.741291 |  -11.37454 |  71.44626 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-11-12 |  9.4 |  5.1 |              0.2 |                 7.400749 |       15.891544 |             90.6 |       7.4 |              1004.0 | 0.0 | 0.0 |     5.1 |                  90.6 |                   0.2 |                      7.400749 |            15.891544 |  8.836 |  5.1 |     5.1 |    5.931775 |         5.931775 |           5.931775 |          7.4 |                 8.790172 |                      8.790172 |                         0.0 |                       6.1078 |                      0.614118 |                           0.614118 |             7.963896 |                  7.963896 |     996.036104 |   1.253213 | 8.583333 |  -3.275868 | -19.655205 |              3.997567 |                      0.0 |                  3.997567 |                   3.997567 |                 13.525343 |                     0.0 |                -0.754591 |    -9.527776 |             0.0 |         4.752158 |         -9.527776 |              28.78455 |                 100.0 |               112.102145 |               99.510262 |  8.836 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.033005 | 0.0 |     0.0 | -1.215856 |  2.183707 |          0.0 |     0.748447 | 21.462327 | 21.462327 |      0.0 |      0.0 |     0.0 | -37.537452 |      0.0 | -58.603285 |      0.0 | -100.892896 |  0.94547 | 0.268174 | 30.457573 | 9.778899 |  0.0 |  0.0 | 0.0 |  0.0 | 9.778899 | 2.517544 | 0.029138 | 0.06 |   0.0 |  0.0 |      0.0 |   nan |  2.267017 |  3.241834 |       0.0 | 0.927231 |  16.863998 | 92.124934 | 9.778899 |   0.0 |   0.0 |   0.0 |  0.0 | 2.517544 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.029138 |
    | 2010-12-12 |  0.2 |  0.2 |              0.7 |                 7.386143 |       17.823519 |             78.7 |       5.7 |              1012.0 | 0.0 | 0.0 |     0.2 |                  78.7 |                   0.7 |                      7.386143 |            17.823519 |  0.188 |  0.2 |     0.2 |     4.56907 |          4.56907 |            4.56907 |          5.7 |                 6.197477 |                      6.197477 |                         0.0 |                     5.176451 |                      0.451275 |                           0.451275 |             4.877414 |                  4.877414 |    1007.122586 |    1.28736 | 8.583333 |        0.0 |  13.478921 |              4.716622 |                      0.0 |                  4.716622 |                   4.716622 |                 20.069431 |                     0.0 |                36.312063 |   -15.352808 |             0.0 |       -31.595441 |        -15.352808 |             37.369415 |                 100.0 |                83.817217 |               98.586605 |  0.188 |  0.0846 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.429349 | 0.0 |     0.0 |  0.158718 |  0.000762 |    -2.246487 |     0.735371 |       0.0 |       0.0 | 0.061533 | 0.061533 |     0.0 |   4.900158 |      0.0 | -22.778045 |      0.0 |   13.717554 |   0.9613 | 0.277229 |  0.067022 | 0.022382 |  0.0 |  0.0 | 0.0 |  0.0 | 0.022382 | 3.392378 | 0.039264 | 0.06 |   0.0 |  0.0 |      0.0 |   nan |  2.299254 |  3.204093 |       0.0 | 1.094821 |  11.968411 | 92.169574 | 0.022382 |   0.0 |   0.0 |   0.0 |  0.0 | 3.392378 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.039264 |
    | 2010-13-12 |  2.9 | -3.5 |              0.0 |                 7.373328 |        8.941324 |             90.5 |       2.2 |              1020.0 | 0.0 | 0.0 |    -3.5 |                  90.5 |                   0.0 |                      7.373328 |             8.941324 |  2.726 | -3.5 |    -3.5 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 4.713362 |                      4.713362 |                         0.0 |                     4.320066 |                      0.354306 |                           0.354306 |             4.265593 |                  4.265593 |    1015.734407 |   1.315658 | 8.583333 |        0.0 |  27.919864 |              2.237089 |                      0.0 |                  2.237089 |                   2.237089 |                 14.516908 |                     0.0 |                35.794391 |   -12.279819 |             0.0 |       -33.557302 |        -12.279819 |             96.820757 |                 100.0 |                83.758368 |               98.584459 |  2.726 |   2.726 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.193889 | 0.0 |     0.0 |  0.014973 | -0.230795 |    -4.653311 |     0.749803 |       0.0 |       0.0 | 0.199105 | 0.199105 |     0.0 |   0.462261 |      0.0 |  -5.560804 |      0.0 |   28.458759 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 1.814887 | 0.021006 | 0.06 |   0.0 |  0.0 |      0.0 |   nan |  5.211168 |   5.91512 |       0.0 | 0.911369 |   -7.36812 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 1.814887 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.021006 |
    | 2010-14-12 |  0.0 | -3.9 |              1.2 |                 7.362317 |       19.668068 |             88.1 |       3.1 |              1025.0 | 0.0 | 0.0 |    -3.9 |                  88.1 |                   1.2 |                      7.362317 |            19.668068 |    0.0 | -3.9 |    -3.9 |    2.484933 |         2.484933 |           2.484933 |          3.1 |                  4.57351 |                       4.57351 |                         0.0 |                     3.838228 |                      0.344989 |                           0.344989 |             4.029262 |                  4.029262 |    1020.970738 |   1.324197 | 8.583333 |  -2.122018 |  24.469911 |              5.193321 |                      0.0 |                  5.193321 |                   5.193321 |                 24.390008 |                     0.0 |                46.412117 |   -19.196687 |             0.0 |       -41.218796 |        -19.196687 |             68.711505 |                 100.0 |                83.758368 |               98.584802 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.281648 | 0.0 |     0.0 | -0.065079 |       0.0 |    -6.200337 |     0.735952 |       0.0 |       0.0 | 0.256899 | 0.256899 |     0.0 |  -2.009214 |      0.0 | -13.746566 |      0.0 |   25.463016 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.968444 | 0.011209 | 0.06 |   0.0 |  0.0 |      0.0 |   nan |  5.528228 |    5.9802 |       0.0 | 1.086761 | -23.254697 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.968444 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.011209 |
    | 2010-15-12 |  0.0 | -6.3 |              3.0 |                 7.353123 |       26.831581 |             92.8 |       2.8 |              1023.0 | 0.0 | 0.0 |    -6.3 |                  92.8 |                   3.0 |                      7.353123 |            26.831581 |    0.0 | -6.3 |    -6.3 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 3.808888 |                      3.808888 |                         0.0 |                     2.826082 |                      0.293396 |                           0.293396 |             3.534648 |                  3.534648 |    1019.465352 |    1.33374 | 8.583333 |  -6.697353 |  20.473897 |              7.398865 |                      0.0 |                  7.398865 |                   7.398865 |                 39.328283 |                     0.0 |                57.719027 |   -31.929417 |             0.0 |       -50.320161 |        -31.929417 |             76.073452 |                 100.0 |                83.758368 |               98.585088 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.028526 | 0.0 |     0.0 | -0.225845 |       0.0 |   -10.109669 |     0.724248 |       0.0 |       0.0 | 0.406999 | 0.406999 |     0.0 |  -6.972561 |      0.0 |  -21.30035 |      0.0 |   22.047251 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.516772 | 0.005981 | 0.06 |   0.0 |  0.0 |      0.0 |   nan |  6.159374 |  6.206044 |       0.0 | 1.262153 | -35.145261 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.516772 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.005981 |
    | 2010-16-12 |  6.7 | -3.5 |              0.0 |                 7.345755 |        8.837738 |             98.1 |       3.7 |              1000.0 | 0.0 | 0.0 |    -3.5 |                  98.1 |                   0.0 |                      7.345755 |             8.837738 |  6.298 | -3.5 |    -3.5 |    2.965888 |         2.965888 |           2.965888 |          3.7 |                 4.713362 |                      4.713362 |                         0.0 |                     3.578581 |                      0.354306 |                           0.354306 |             4.623808 |                  4.623808 |     995.376192 |   1.289645 | 8.583333 | -10.121835 | -18.089989 |              1.897885 |                      0.0 |                  1.897885 |                   1.897885 |                   14.0252 |                     0.0 |                22.170845 |   -12.127315 |             0.0 |        -20.27296 |        -12.127315 |             57.569099 |                 100.0 |                83.758368 |               98.585317 |  6.298 |   6.298 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.035293 | 0.0 |     0.0 |  -0.40193 | -0.533216 |    -7.106837 |     0.785252 |       0.0 |       0.0 | 0.558149 | 0.046671 |     0.0 | -12.408877 |      0.0 |  -24.32963 |      0.0 |  -16.465547 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.275755 | 0.003192 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 12.905974 | 12.905974 |  -1.97724 | 0.573017 |  -8.471938 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.275755 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.003192 |
    | 2010-17-12 |  0.0 | -7.3 |              0.0 |                 7.340222 |        8.817416 |             96.6 |       0.9 |               991.0 | 0.0 | 0.0 |    -7.3 |                  96.6 |                   0.0 |                      7.340222 |             8.817416 |    0.0 | -7.3 |    -7.3 |    0.721432 |         0.721432 |           0.721432 |          0.9 |                  3.52531 |                       3.52531 |                         0.0 |                     3.261149 |                      0.273951 |                           0.273951 |             3.405449 |                  3.405449 |     987.594551 |   1.296889 | 8.583333 |  -2.439918 |  35.147834 |              1.958046 |                      0.0 |                  1.958046 |                   1.958046 |                 14.906398 |                     0.0 |                41.667909 |   -12.948351 |             0.0 |       -39.709863 |        -12.948351 |            236.672963 |                 100.0 |                83.758368 |               98.585489 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.001256 | 0.0 |     0.0 | -0.025948 |       0.0 |    -8.297891 |     0.777934 |       0.0 |       0.0 | 0.670129 |      0.0 |     0.0 |  -0.801088 |      0.0 |  -3.147637 |      0.0 |   35.761138 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.147146 | 0.001703 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 12.931922 | 12.931922 | -2.590544 | 0.632681 | -35.036439 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.147146 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.001703 |
    | 2010-18-12 |  0.1 | -5.8 |              0.0 |                  7.33653 |        8.804185 |             97.7 |       1.3 |               997.0 | 0.0 | 0.0 |    -5.8 |                  97.7 |                   0.0 |                       7.33653 |             8.804185 |  0.094 | -5.8 |    -5.8 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 3.958114 |                      3.958114 |                         0.0 |                     2.837121 |                      0.303557 |                           0.303557 |             3.867077 |                  3.867077 |     993.132923 |   1.297204 | 8.583333 | -10.090494 |  -0.178602 |              2.001401 |                      0.0 |                  2.001401 |                   2.001401 |                 14.549007 |                     0.0 |                24.662372 |   -12.547606 |             0.0 |       -22.660971 |        -12.547606 |            163.850513 |                 100.0 |                83.758368 |               98.585603 |  0.094 |   0.094 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.002818 | 0.0 |     0.0 | -0.215326 | -0.013188 |   -10.060727 |     0.772676 |       0.0 |       0.0 | 0.820045 |      0.0 |     0.0 |  -6.647818 |      0.0 | -15.625407 |      0.0 |    0.387746 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.078519 | 0.000909 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 13.241248 | 13.241248 | -3.170081 | 0.678422 | -26.274503 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.078519 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000909 |
    | 2010-19-12 |  0.0 | -7.9 |              0.0 |                 7.334683 |        8.798044 |             95.8 |       1.9 |               997.0 | 0.0 | 0.0 |    -7.9 |                  95.8 |                   0.0 |                      7.334683 |             8.798044 |    0.0 | -7.9 |    -7.9 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 3.364296 |                      3.364296 |                         0.0 |                     2.722791 |                      0.262827 |                           0.262827 |             3.222996 |                  3.222996 |     993.777004 |   1.307793 | 8.583333 |  -7.567057 |  18.053896 |               2.04414 |                      0.0 |                   2.04414 |                    2.04414 |                 15.062045 |                     0.0 |                 36.09466 |   -13.017906 |             0.0 |        -34.05052 |        -13.017906 |            112.108246 |                 100.0 |                83.758368 |               98.585661 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.032531 | 0.0 |     0.0 | -0.126518 |       0.0 |    -10.57604 |      0.76766 |       0.0 |       0.0 | 0.876298 |      0.0 |     0.0 |  -3.906012 |      0.0 | -11.873153 |      0.0 |   18.271356 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.041898 | 0.000485 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 13.367765 | 13.367765 | -3.387541 | 0.724518 | -35.745066 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.041898 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000485 |
    | 2010-20-12 |  0.2 | -5.7 |              0.2 |                 7.334683 |       15.464746 |             97.5 |       2.4 |               997.0 | 0.0 | 0.0 |    -5.7 |                  97.5 |                   0.2 |                      7.334683 |            15.464746 |  0.188 | -5.7 |    -5.7 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 3.988573 |                      3.988573 |                         0.0 |                     2.930412 |                      0.305625 |                           0.305625 |             3.888858 |                  3.888858 |     993.111142 |   1.296708 | 8.583333 | -10.294579 |  -3.845125 |              3.642756 |                      0.0 |                  3.642756 |                   3.642756 |                  16.12634 |                     0.0 |                28.193275 |   -12.483583 |             0.0 |       -24.550519 |        -12.483583 |             88.752361 |                 100.0 |                83.758368 |               98.585661 |  0.188 |   0.188 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.020944 | 0.0 |     0.0 |  -0.27746 | -0.025922 |    -9.653725 |     0.764448 |       0.0 |       0.0 | 0.818877 |      0.0 |     0.0 |  -8.566089 |      0.0 | -20.077451 |      0.0 |   -4.093021 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.022357 | 0.000259 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 13.833225 | 13.833225 | -3.165567 | 0.755386 | -23.316607 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.022357 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000259 |
    | 2010-21-12 |  0.2 | -5.6 |              0.8 |                  7.33653 |       17.879541 |             95.0 |       2.3 |              1004.0 | 0.0 | 0.0 |    -5.6 |                  95.0 |                   0.8 |                       7.33653 |            17.879541 |  0.188 | -5.6 |    -5.6 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 4.019239 |                      4.019239 |                         0.0 |                      3.12843 |                      0.307706 |                           0.307706 |             3.818277 |                  3.818277 |    1000.181723 |   1.305372 | 8.583333 |  -6.715183 |  12.668619 |              4.271628 |                      0.0 |                  4.271628 |                   4.271628 |                 21.069426 |                     0.0 |                38.724024 |   -16.797798 |             0.0 |       -34.452396 |        -16.797798 |             92.611159 |                 100.0 |                83.758368 |               98.585603 |  0.188 |   0.188 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.034932 | 0.0 |     0.0 |  -0.19466 | -0.025467 |     -8.82662 |     0.761089 |       0.0 |       0.0 | 0.774426 |      0.0 |     0.0 |  -6.009778 |      0.0 | -15.971301 |      0.0 |   12.471317 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |  0.01193 | 0.000138 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 14.215885 | 14.215885 | -2.993732 | 0.788873 | -27.401894 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.01193 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000138 |
    | 2010-22-12 |  2.4 | -4.8 |              0.0 |                 7.340222 |        8.822161 |             93.8 |       4.6 |              1007.0 | 0.0 | 0.0 |    -4.8 |                  93.8 |                   0.0 |                      7.340222 |             8.822161 |  2.256 | -4.8 |    -4.8 |     3.68732 |          3.68732 |            3.68732 |          4.6 |                 4.272186 |                      4.272186 |                         0.0 |                      3.44223 |                      0.324793 |                           0.324793 |             4.007311 |                  4.007311 |    1002.992689 |   1.305282 | 8.583333 |  -7.891745 |  -1.711598 |              1.990066 |                      0.0 |                  1.990066 |                   1.990066 |                 14.580265 |                     0.0 |                30.080567 |   -12.590199 |             0.0 |       -28.090501 |        -12.590199 |              46.30558 |                 100.0 |                83.758368 |               98.585489 |  2.256 |   2.256 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.238205 | 0.0 |     0.0 | -0.254479 | -0.261947 |    -7.606479 |     0.774424 |       0.0 |       0.0 | 0.784019 |      0.0 |     0.0 |  -7.856587 |      0.0 | -22.170375 |      0.0 |   -1.936461 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.006366 | 0.000074 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 16.726363 | 16.726363 | -3.030816 | 0.662934 | -17.106962 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.006366 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000074 |
    | 2010-23-12 |  7.0 | -1.1 |              0.0 |                 7.345755 |        8.844396 |             94.9 |       8.6 |              1004.0 | 0.0 | 0.0 |    -1.1 |                  94.9 |                   0.0 |                      7.345755 |             8.844396 |   6.58 | -1.1 |    -1.1 |    6.893685 |         6.893685 |           6.893685 |          8.6 |                 5.634754 |                      5.634754 |                         0.0 |                     4.876556 |                       0.41489 |                            0.41489 |             5.347381 |                  5.347381 |     998.652619 |    1.28304 | 8.583333 |  -4.926805 | -11.283011 |              1.665079 |                      0.0 |                  1.665079 |                   1.665079 |                 13.655186 |                     0.0 |                25.120919 |   -11.990107 |             0.0 |        -23.45584 |        -11.990107 |             24.768101 |                 100.0 |                83.758368 |               98.585317 |   6.58 |  5.0995 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.461634 | 0.0 |     0.0 | -0.349728 | -0.214608 |    -3.046303 |     0.811736 |       0.0 |       0.0 | 0.472584 | 0.472584 |     0.0 |  -10.79726 |      0.0 | -25.360128 |      0.0 |  -12.701548 | 0.991077 |  0.29467 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.003397 | 0.000039 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 22.633051 | 23.656092 |       0.0 | 0.389279 |   2.759382 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.003397 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000039 |
    | 2010-24-12 |  0.2 | -2.9 |              0.0 |                 7.353123 |        8.873743 |             85.4 |       7.6 |              1008.0 | 0.0 | 0.0 |    -2.9 |                  85.4 |                   0.0 |                      7.353123 |             8.873743 |  0.188 | -2.9 |    -2.9 |    6.092093 |         6.092093 |           6.092093 |          7.6 |                 4.930236 |                      4.930236 |                         0.0 |                     4.510219 |                      0.368688 |                           0.368688 |             4.210422 |                  4.210422 |    1003.789578 |   1.297296 | 8.583333 |        0.0 |  24.507545 |              1.886144 |                      0.0 |                  1.886144 |                   1.886144 |                 14.739672 |                     0.0 |                 36.55523 |   -12.853528 |             0.0 |       -34.669086 |        -12.853528 |             28.027061 |                 100.0 |                83.758368 |               98.585088 |  0.188 |   0.188 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.132438 | 0.0 |     0.0 |   0.20077 | -0.013188 |    -4.084591 |     0.787447 |       0.0 |       0.0 | 0.635671 | 0.635671 |     0.0 |   6.198411 |      0.0 | -13.915802 |      0.0 |   26.951695 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.001813 | 0.000021 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 23.259214 | 23.643322 |       0.0 | 0.555966 | -13.164829 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.001813 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000021 |
    | 2010-25-12 |  0.0 | -9.1 |              3.2 |                 7.362317 |       27.759642 |             87.2 |       3.8 |              1018.0 | 0.0 | 0.0 |    -9.1 |                  87.2 |                   3.2 |                      7.362317 |            27.759642 |    0.0 | -9.1 |    -9.1 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                 3.061705 |                      3.061705 |                         0.0 |                     2.608187 |                      0.241745 |                           0.241745 |             2.669806 |                  2.669806 |    1015.330194 |   1.341718 | 8.583333 |  -3.791471 |  43.924699 |              6.469714 |                      0.0 |                  6.469714 |                   6.469714 |                 43.173862 |                     0.0 |                71.430435 |   -36.704147 |             0.0 |        -64.96072 |        -36.704147 |             56.054123 |                 100.0 |                83.758368 |               98.584802 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.146109 | 0.0 |     0.0 | -0.024146 |       0.0 |   -11.112254 |     0.766938 |       0.0 |       0.0 | 1.670829 | 0.384108 |     0.0 |  -0.745452 |      0.0 | -13.831579 |      0.0 |   50.383689 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.000967 | 0.000011 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 23.667468 | 23.667468 | -4.974127 | 0.731358 | -48.506195 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000967 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000011 |
    | 2010-26-12 |  1.2 | -8.6 |              4.3 |                 7.373328 |       32.328755 |             92.7 |       1.5 |              1019.0 | 0.0 | 0.0 |    -8.6 |                  92.7 |                   4.3 |                      7.373328 |            32.328755 |  1.128 | -8.6 |    -8.6 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 3.184716 |                      3.184716 |                         0.0 |                     1.819064 |                      0.250344 |                           0.250344 |             2.952232 |                  2.952232 |    1016.047768 |   1.340358 | 8.583333 | -13.969784 |   9.219314 |              7.422688 |                      0.0 |                  7.422688 |                   7.422688 |                 51.194924 |                     0.0 |                55.658664 |   -43.772236 |             0.0 |       -48.235975 |        -43.772236 |            142.003778 |                 100.0 |                83.758368 |               98.584459 |  1.128 |   1.128 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  -0.16868 | 0.0 |     0.0 | -0.253474 | -0.234661 |   -15.506336 |       0.7704 |       0.0 |       0.0 | 2.405921 |      0.0 |     0.0 |  -7.825569 |      0.0 | -27.099214 |      0.0 |   13.311192 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.000516 | 0.000006 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 25.048942 | 25.048942 | -9.300666 | 0.699031 | -49.142175 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000516 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000006 |
    | 2010-27-12 |  0.9 | -5.3 |              0.0 |                 7.386143 |        9.004582 |             97.8 |       1.3 |              1017.0 | 0.0 | 0.0 |    -5.3 |                  97.8 |                   0.0 |                      7.386143 |             9.004582 |  0.846 | -5.3 |    -5.3 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 4.112496 |                      4.112496 |                         0.0 |                     2.537355 |                       0.31402 |                            0.31402 |             4.022021 |                  4.022021 |    1012.977979 |   1.320718 | 8.583333 | -14.152947 |  -16.19232 |              2.040733 |                      0.0 |                  2.040733 |                   2.040733 |                 14.439522 |                     0.0 |                15.766318 |   -12.398789 |             0.0 |       -13.725584 |        -12.398789 |            163.850513 |                 100.0 |                83.758368 |               98.584061 |  0.846 |   0.846 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.002438 | 0.0 |     0.0 | -0.310389 | -0.108462 |   -11.454227 |     0.773367 |       0.0 |       0.0 | 1.856011 |      0.0 |     0.0 |   -9.58272 |      0.0 | -22.569455 |      0.0 |  -18.426591 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.000275 | 0.000003 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 26.205331 | 26.205331 | -7.174857 | 0.672265 | -24.366522 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000275 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000003 |
    | 2010-28-12 |  0.0 | -7.7 |              3.4 |                 7.400749 |       28.982321 |             95.4 |       3.0 |              1022.0 | 0.0 | 0.0 |    -7.7 |                  95.4 |                   3.4 |                      7.400749 |            28.982321 |    0.0 | -7.7 |    -7.7 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 3.417227 |                      3.417227 |                         0.0 |                     2.553035 |                      0.266491 |                           0.266491 |             3.260035 |                  3.260035 |    1018.739965 |   1.339598 | 8.583333 |  -7.017558 |  26.161471 |               6.70143 |                      0.0 |                   6.70143 |                    6.70143 |                 42.703706 |                     0.0 |                61.683256 |   -36.002275 |             0.0 |       -54.981825 |        -36.002275 |             71.001889 |                 100.0 |                83.758368 |               98.583607 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.122157 | 0.0 |     0.0 | -0.235684 |       0.0 |   -11.377804 |     0.768775 |       0.0 |       0.0 | 1.865726 |      0.0 |     0.0 |  -7.276335 |      0.0 | -21.506464 |      0.0 |   26.199027 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.000147 | 0.000002 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 26.441015 | 26.441015 | -7.212414 |  0.71405 |  -41.94466 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000147 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000002 |
    | 2010-29-12 |  0.0 | -9.2 |              0.0 |                  7.41713 |        9.127645 |             94.9 |       2.7 |              1024.0 | 0.0 | 0.0 |    -9.2 |                  94.9 |                   0.0 |                       7.41713 |             9.127645 |    0.0 | -9.2 |    -9.2 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 3.037615 |                      3.037615 |                         0.0 |                     2.252914 |                      0.240056 |                           0.240056 |             2.882696 |                  2.882696 |    1021.117304 |   1.350039 | 8.583333 | -12.080062 |   5.033832 |               2.15057 |                      0.0 |                   2.15057 |                    2.15057 |                 15.336383 |                     0.0 |                34.606803 |   -13.185813 |             0.0 |       -32.456233 |        -13.185813 |             78.890988 |                 100.0 |                83.758368 |               98.583098 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.072905 | 0.0 |     0.0 | -0.196129 |       0.0 |   -12.919034 |     0.764389 |       0.0 |       0.0 | 2.137509 |      0.0 |     0.0 |  -6.055144 |      0.0 | -20.316614 |      0.0 |    6.084474 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.000078 | 0.000001 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 26.637144 | 26.637144 | -8.263056 | 0.755957 | -38.395158 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000078 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000001 |
    | 2010-30-12 |  0.2 | -7.4 |              0.0 |                 7.435268 |        9.199989 |             96.0 |       1.4 |              1021.0 | 0.0 | 0.0 |    -7.4 |                  96.0 |                   0.0 |                      7.435268 |             9.199989 |  0.188 | -7.4 |    -7.4 |    1.122228 |         1.122228 |           1.122228 |          1.4 |                 3.498009 |                      3.498009 |                         0.0 |                     2.475478 |                      0.272069 |                           0.272069 |             3.358088 |                  3.358088 |    1017.641912 |   1.336727 | 8.583333 | -11.057806 |   4.213054 |              2.189463 |                      0.0 |                  2.189463 |                   2.189463 |                 14.959543 |                     0.0 |                28.181399 |   -12.770081 |             0.0 |       -25.991936 |        -12.770081 |            152.146905 |                 100.0 |                83.758368 |               98.582534 |  0.188 |   0.188 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.018935 | 0.0 |     0.0 | -0.190974 | -0.033653 |   -11.759981 |     0.762015 |       0.0 |       0.0 | 1.974006 |      0.0 |     0.0 |  -5.896006 |      0.0 | -16.548589 |      0.0 |    3.547341 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.000042 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 27.016118 | 27.016118 | -7.630996 | 0.779513 | -34.024879 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000042 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-31-12 |  0.9 |  1.7 |              4.0 |                 7.455145 |       32.062235 |             96.5 |       4.2 |              1016.0 | 0.0 | 0.0 |     1.7 |                  96.5 |                   4.0 |                      7.455145 |            32.062235 |  0.846 |  1.7 |     1.7 |    3.366683 |         3.366683 |           3.366683 |          4.2 |                 6.907942 |                      6.907942 |                         0.0 |                     4.469848 |                      0.496631 |                           0.496631 |             6.666164 |                  6.666164 |    1009.333836 |   1.284547 | 8.583333 |  -9.799165 | -33.573784 |              7.749055 |                      0.0 |                  7.749055 |                   7.749055 |                 40.225917 |                     0.0 |                41.685254 |   -32.476862 |             0.0 |       -33.936199 |        -32.476862 |             50.715635 |                 100.0 |                83.758368 |               98.581916 |  0.846 | 0.06345 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.128252 | 0.0 |     0.0 | -0.924858 |  0.067075 |    -4.203534 |     0.758312 |       0.0 |       0.0 | 0.753524 | 0.753524 |     0.0 |  -28.55338 |      0.0 | -43.607597 |      0.0 |  -38.224778 | 0.973704 |  0.28443 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.000022 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 28.756986 | 28.786976 |       0.0 | 0.817531 |   8.132239 | 92.169574 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000022 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |

>>> round_(model.check_waterbalance(conditions))
0.0

.. _lland_v4_nadelw_winter_daily:

conifers (winter)
-----------------

.. integration-test::

    >>> lnk(NADELW)
    >>> conditions_nadelw_winter = test(
    ...     "lland_v4_nadelw_winter_daily",
    ...     axis1=(inputs.nied, fluxes.wada), axis2=(states.waes, states.wats),
    ...     get_conditions="2010-12-10")
    |       date | nied | teml | sunshineduration | possiblesunshineduration | globalradiation | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration | dailypossiblesunshineduration | dailyglobalradiation |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair |        g |        tz |         wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |  nbes |    sbes | snowintmax | snowintrate |  nbesinz |  sbesinz |  wniedinz | actualalbedoinz |  wadainz | schmpotinz |  schminz | gefrpotinz |  gefrinz |   evsinz |      evpo | evi |       evb |       evs |     wnied | tempssurface | actualalbedo |   schmpot |      schm |  gefrpot |     gefr |   wlatinz |   wlatsnow |   wsensinz |  wsenssnow |   wsurfinz |      wsurf |      sff |      fvg |      wada |      qdb | qib1 | qib2 | qbb | qkap |     qdgz |      qah |       qa | inzp |    stinz |     sinz |  esnowinz |    asinz |      wats |      waes |     esnow |     taus |       ebdn |      bowa |    qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    -----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 2010-01-12 |  0.0 | -5.8 |              3.5 |                 7.641029 |       31.942798 |             70.3 |       6.9 |              1015.0 | 0.0 | 0.0 |    -5.8 |                  70.3 |                   3.5 |                      7.641029 |            31.942798 |    0.0 | -5.8 |    -5.8 |     5.53098 |          5.53098 |           2.449434 |          6.9 |                 3.958114 |                      3.958114 |                         0.0 |                          0.0 |                      0.303557 |                           0.303557 |             2.782554 |                  2.782554 |    1012.217446 |   1.321192 | 8.583333 |       0.0 |       34.8 |             28.109662 |                      nan |                  5.220366 |                  28.109662 |                 46.472798 |                     0.0 |                      0.0 |   -18.363136 |             0.0 |              0.0 |        -18.363136 |             13.623188 |                 100.0 |             27827.355778 |              101.116829 |   0.0 |     0.0 |       24.5 |        0.42 |      0.0 |      0.0 |       0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  4.142976 | 0.0 |  0.674992 |       0.0 |       0.0 |          nan |         0.12 |       0.0 |       0.0 |      0.0 |      0.0 |       0.0 |        0.0 |        0.0 |        0.0 |        0.0 |        0.0 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |  0.0 |      0.0 |      0.0 |       0.0 |      nan |       0.0 |       0.0 |       0.0 |      nan | -26.216667 | 71.325008 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-02-12 | 11.1 | -3.0 |              0.0 |                 7.609644 |        9.873898 |             85.0 |       5.2 |              1007.0 | 0.0 | 0.0 |    -3.0 |                  85.0 |                   0.0 |                      7.609644 |             9.873898 | 10.434 | -3.0 |    -3.0 |    4.168274 |         4.168274 |            1.84595 |          5.2 |                 4.893489 |                      4.893489 |                    4.893489 |                     4.000777 |                      0.366257 |                           0.366257 |             4.159465 |                  4.159465 |    1002.840535 |   1.296511 | 8.583333 |   -7.5504 | -11.341661 |               1.17941 |                 1.001169 |                  0.219033 |                    1.17941 |                  14.78976 |               17.929265 |                 3.494169 |    -13.61035 |      -16.928095 |        -3.275135 |         -13.61035 |             18.076923 |                 100.0 |              2039.690308 |               99.509579 | 8.234 |   8.234 |       24.5 |        0.42 |  3.45828 |  3.45828 | -0.250965 |        0.875479 |      0.0 |        0.0 |      0.0 |   0.165137 |      0.0 | 0.207278 |  1.793874 | 0.0 |  0.389762 | -0.044812 | -0.597537 |    -5.660123 |     0.880553 |       0.0 |       0.0 | 0.169147 |      0.0 |  6.399359 |   -1.38348 | -22.940042 | -13.176974 |   0.387412 | -11.285319 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |  2.2 | 3.251002 | 3.251002 | -0.638377 | 0.084415 |  4.820532 |  4.820532 | -0.653879 | 0.065799 |  -6.291673 | 70.935246 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-03-12 |  0.4 | -5.6 |              0.1 |                 7.579824 |         16.6835 |             95.9 |       2.0 |              1005.0 | 0.0 | 0.0 |    -5.6 |                  95.9 |                   0.1 |                      7.579824 |              16.6835 |  0.376 | -5.6 |    -5.6 |    1.603182 |         1.603182 |           0.709981 |          2.0 |                 4.019239 |                      4.019239 |                    4.019239 |                     4.160243 |                      0.307706 |                           0.307706 |              3.85445 |                   3.85445 |     1001.14555 |   1.306656 | 8.583333 | -1.812002 |  20.020333 |              2.255166 |                 1.869927 |                  0.418817 |                   2.255166 |                 15.387201 |                17.78068 |                17.258471 |   -13.132035 |      -15.910754 |       -16.839654 |        -13.132035 |                  47.0 |                 100.0 |             28209.005837 |              100.750293 | 0.376 |   0.376 |       24.5 |    0.429753 | 0.161587 | 0.161587 | -0.021889 |        0.862355 |      0.0 |        0.0 |      0.0 |   0.232496 |      0.0 |  0.02946 |  0.128653 | 0.0 |  0.051032 |  0.054667 | -0.050934 |    -5.148724 |     0.864827 |       0.0 |       0.0 | 0.162216 |      0.0 |   0.90952 |   1.687763 | -16.581769 |   1.415188 |   0.238504 |  19.942605 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |  2.2 | 3.383129 | 3.383129 | -0.898771 | 0.135729 |  4.980277 |  4.980277 | -0.627086 | 0.125697 | -17.728672 | 70.884214 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-04-12 |  1.2 | -3.8 |              3.7 |                 7.551596 |       31.807427 |             90.6 |       4.3 |              1008.0 | 0.0 | 0.0 |    -3.8 |                  90.6 |                   3.7 |                      7.551596 |            31.807427 |  1.128 | -3.8 |    -3.8 |    3.446842 |         3.446842 |           1.526459 |          4.3 |                 4.608124 |                      4.608124 |                    4.608124 |                     4.012494 |                      0.347298 |                           0.347298 |              4.17496 |                   4.17496 |     1003.82504 |   1.301648 | 8.583333 | -5.105858 |   3.096474 |              4.681319 |                 3.803097 |                  0.869388 |                   4.681319 |                  43.14024 |               34.929625 |                13.363685 |   -38.458921 |      -31.126528 |       -12.494297 |        -38.458921 |             21.860465 |                 100.0 |              3366.095402 |               99.771654 | 1.128 |   1.128 |       24.5 |    0.430149 | 0.485209 | 0.485209 | -0.044601 |        0.853164 |      0.0 |        0.0 |      0.0 |   0.279301 |      0.0 | 0.109697 |  0.581314 | 0.0 |   0.14346 | -0.041144 | -0.103687 |    -5.621937 |     0.852823 |       0.0 |       0.0 | 0.197815 |      0.0 |  3.386694 |  -1.270243 | -34.376888 |  -8.093651 |   0.136334 |   3.130403 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |  2.2 | 3.758641 | 3.758641 | -1.079705 | 0.174653 |  5.664212 |  5.664212 | -0.764701 | 0.176149 | -12.241813 | 70.740754 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-05-12 |  6.4 |  0.2 |              0.0 |                 7.524989 |        9.532511 |             96.5 |       4.0 |               996.0 | 0.0 | 0.0 |     0.2 |                  96.5 |                   0.0 |                      7.524989 |             9.532511 |  6.016 |  0.2 |     0.2 |    3.206365 |         3.206365 |           1.419962 |          4.0 |                 6.197477 |                      6.197477 |                    6.197477 |                     5.324062 |                      0.451275 |                           0.451275 |             5.980565 |                  5.980565 |     990.019435 |   1.266438 | 8.583333 | -3.525642 |  -9.947345 |              1.488187 |                 1.212963 |                  0.276378 |                   1.488187 |                 13.205056 |                 12.7506 |                 3.874587 |   -11.716869 |      -11.537636 |         -3.59821 |        -11.716869 |                  23.5 |                 100.0 |               769.705161 |               96.652893 | 6.016 |  2.7072 |       49.0 |    0.431276 | 2.594556 |  1.16755 |  0.019479 |        0.843734 |      0.0 |        0.0 |      0.0 |   0.146071 | 0.146071 | 0.052825 |  0.341603 | 0.0 |  0.097263 | -0.159879 |  0.045165 |    -1.867751 |     0.843883 |       0.0 |       0.0 | 0.128253 | 0.128253 |  1.630875 |  -4.935991 | -13.664067 |  -8.833307 |  -0.495555 | -10.171089 | 0.979665 | 0.287923 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |  2.2 | 5.030087 | 6.300372 |       0.0 | 0.217465 |  7.461137 |  9.245535 |       0.0 | 0.216766 |   6.288865 | 70.643491 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-06-12 |  0.0 | -1.1 |              0.1 |                 7.500028 |       16.141484 |             99.0 |       2.4 |               998.0 | 0.0 | 0.0 |    -1.1 |                  99.0 |                   0.1 |                      7.500028 |            16.141484 |    0.0 | -1.1 |    -1.1 |    1.923819 |         1.923819 |           0.851977 |          2.4 |                 5.634754 |                      5.634754 |                    5.634754 |                     5.469703 |                       0.41489 |                            0.41489 |             5.578406 |                  5.578406 |     992.421594 |   1.275245 | 8.583333 |       0.0 |   9.018379 |              3.046424 |                 2.482152 |                  0.565764 |                   3.046424 |                 14.093918 |               14.665301 |                11.984392 |   -11.047494 |      -12.183149 |       -11.418628 |        -11.047494 |             39.166667 |                 100.0 |              1019.272046 |               97.443993 |   0.0 |     0.0 |     47.775 |    0.438901 |      0.0 |      0.0 |       0.0 |        0.811154 |      0.0 |        0.0 |      0.0 |   0.216736 | 0.216736 | 0.010803 |  0.032333 | 0.0 |  0.012457 | -0.020841 |       0.0 |    -1.503063 |     0.811267 |       0.0 |       0.0 | 0.103795 | 0.103795 |   0.33353 |  -0.643435 | -11.678834 |  -1.355567 |   0.837845 |   9.419626 | 0.981071 |  0.28875 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |  2.2 | 5.237827 | 6.289569 |       0.0 | 0.392857 |  7.581986 |  9.266377 |       0.0 | 0.392158 |   5.853819 | 70.631034 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-07-12 |  0.0 | -2.9 |              0.0 |                 7.476739 |        9.341429 |             99.2 |       1.9 |              1004.0 | 0.0 | 0.0 |    -2.9 |                  99.2 |                   0.0 |                      7.476739 |             9.341429 |    0.0 | -2.9 |    -2.9 |    1.523023 |         1.523023 |           0.674482 |          1.9 |                 4.930236 |                      4.930236 |                    4.930236 |                     5.029253 |                      0.368688 |                           0.368688 |             4.890794 |                  4.890794 |     999.109206 |   1.291808 | 8.583333 |       0.0 |  15.802737 |              1.999528 |                 1.628868 |                  0.371341 |                   1.999528 |                 13.815559 |               15.077555 |                15.302045 |   -11.816031 |      -13.448687 |       -14.930704 |        -11.816031 |             49.473684 |                 100.0 |              1891.630333 |               98.648619 |   0.0 |     0.0 |     25.725 |    0.438869 |      0.0 |      0.0 |       0.0 |        0.785861 |      0.0 |        0.0 |      0.0 |   0.318812 | 0.318812 | 0.006923 |  0.000338 | 0.0 |  0.000144 |  0.024304 |       0.0 |    -2.633789 |     0.785951 |       0.0 |       0.0 | 0.180569 | 0.180569 |  0.213749 |   0.750357 | -12.429991 |   0.819708 |   1.232445 |  16.500769 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |  2.2 | 5.550522 | 6.282645 |       0.0 | 0.568249 |  7.742194 |  9.242072 |       0.0 |  0.56755 |  -1.365584 | 70.630889 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-08-12 |  0.0 | -2.9 |              0.0 |                 7.455145 |        9.256764 |             99.1 |       1.8 |              1006.0 | 0.0 | 0.0 |    -2.9 |                  99.1 |                   0.0 |                      7.455145 |             9.256764 |    0.0 | -2.9 |    -2.9 |    1.442864 |         1.442864 |           0.638983 |          1.8 |                 4.930236 |                      4.930236 |                    4.930236 |                      4.97365 |                      0.368688 |                           0.368688 |             4.885864 |                  4.885864 |    1001.114136 |   1.294388 | 8.583333 | -0.393288 |  14.336484 |              2.168592 |                   1.7664 |                  0.402739 |                   2.168592 |                 13.821847 |               15.004299 |                14.641138 |   -11.653255 |      -13.237899 |       -14.238399 |        -11.653255 |             52.222222 |                 100.0 |              1892.101154 |               98.529337 |   0.0 |     0.0 |     25.725 |    0.438848 |      0.0 |      0.0 |       0.0 |        0.765656 |      0.0 |        0.0 |      0.0 |    0.30591 |  0.30591 | 0.007645 |  0.005011 | 0.0 |  0.002204 |  0.015125 |       0.0 |    -2.782702 |     0.765729 |       0.0 |       0.0 | 0.187132 | 0.187132 |  0.236032 |    0.46697 | -12.291363 |   0.354517 |   1.182568 |  15.059886 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |  2.2 | 5.849306 |    6.275 |       0.0 | 0.743641 |  7.916349 |  9.226947 |       0.0 | 0.742942 |  -7.118735 | 70.628686 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-09-12 |  0.7 | -2.1 |              4.4 |                 7.435268 |       33.405815 |             96.9 |       2.7 |              1009.0 | 0.0 | 0.0 |    -2.1 |                  96.9 |                   4.4 |                      7.435268 |            33.405815 |  0.658 | -2.1 |    -2.1 |    2.164296 |         2.164296 |           0.958474 |          2.7 |                 5.233109 |                      5.233109 |                    5.233109 |                      4.77755 |                      0.388644 |                           0.388644 |             5.070883 |                  5.070883 |    1003.929117 |   1.294334 | 8.583333 | -2.050196 |    7.61832 |              8.211096 |                 6.760354 |                  1.524918 |                   8.211096 |                 46.364476 |               34.687643 |                16.120955 |    -38.15338 |      -27.927288 |       -14.596037 |         -38.15338 |             34.814815 |                 100.0 |              1378.320731 |               97.843326 | 0.658 |   0.658 |     35.525 |    0.438825 | 0.288747 | 0.288747 | -0.014668 |        0.751475 |      0.0 |        0.0 |      0.0 |   0.429708 | 0.425694 | 0.032679 | -0.148485 | 0.0 | -0.052391 | -0.059088 | -0.033425 |    -3.319916 |     0.754202 |       0.0 |       0.0 |  0.22667 |  0.22667 |  1.008896 |  -1.824252 | -27.289714 |  -4.310643 |    1.64647 |   8.461141 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |  2.2 | 6.531068 | 6.531068 | -0.015514 | 0.892179 |  8.564686 |  9.655288 |       0.0 | 0.861685 |  -6.153721 | 70.681077 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-10-12 |  3.3 | -3.1 |              0.1 |                  7.41713 |       15.591158 |             95.0 |       3.0 |              1018.0 | 0.0 | 0.0 |    -3.1 |                  95.0 |                   0.1 |                       7.41713 |            15.591158 |  3.102 | -3.1 |    -3.1 |    2.404774 |         2.404774 |           1.064971 |          3.0 |                 4.856984 |                      4.856984 |                    4.856984 |                     4.708644 |                       0.36384 |                            0.36384 |             4.614135 |                  4.614135 |    1013.385865 |   1.310959 | 8.583333 | -1.772272 |  10.446299 |              3.632331 |                  3.17831 |                  0.674576 |                   3.632331 |                 14.892721 |               17.810271 |                12.864381 |    -11.26039 |      -14.631961 |       -12.189805 |         -11.26039 |             31.333333 |                 100.0 |              2110.929844 |               98.480067 | 3.102 |   3.102 |       24.5 |    0.439593 | 1.363618 | 1.363618 | -0.102256 |        0.749654 |      0.0 |        0.0 |      0.0 |    0.35279 |      0.0 | 0.051278 |   0.33207 | 0.0 |  0.107502 |  0.019956 | -0.232614 |    -3.513322 |     0.767026 |       0.0 |       0.0 | 0.274539 | 0.274539 |  1.583124 |   0.616102 | -14.969059 |  -1.530924 |   1.246026 |  11.274983 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |  2.2 | 7.843408 | 7.843408 | -1.363795 | 0.913105 | 10.559081 | 11.373714 |       0.0 |  0.73052 |  -8.016687 | 70.573575 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-11-12 |  9.4 |  5.1 |              0.2 |                 7.400749 |       15.891544 |             90.6 |       7.4 |              1004.0 | 0.0 | 0.0 |     5.1 |                  90.6 |                   0.2 |                      7.400749 |            15.891544 |  8.836 |  5.1 |     5.1 |    5.931775 |         5.931775 |           2.626929 |          7.4 |                 8.790172 |                      8.790172 |                    8.790172 |                       6.1078 |                      0.614118 |                           0.614118 |             7.963896 |                  7.963896 |     996.036104 |   1.253213 | 8.583333 | -2.308806 | -13.852836 |              3.891541 |                 3.357472 |                  0.722715 |                   3.891541 |                 13.525343 |               -0.614453 |               -11.576952 |    -9.633802 |        3.971925 |        12.299667 |         -9.633802 |             12.702703 |                 100.0 |               409.748902 |                92.95224 | 8.836 |     0.0 |       49.0 |     0.44353 | 3.919033 |      0.0 |  0.833512 |        0.740541 | 9.934016 |   6.740382 | 6.740382 |        0.0 |      0.0 | 0.292189 |  2.297709 | 0.0 |  0.472765 | -0.656354 |  -0.56575 |          0.0 |     0.755119 | 12.877382 | 10.559081 |      0.0 |      0.0 |  9.020814 | -20.263798 | -31.635741 | -31.635741 | -26.586852 | -64.199205 | 0.953374 | 0.272677 | 26.881051 | 8.695463 |  0.0 |  0.0 | 0.0 |  0.0 | 8.695463 | 2.238617 |  0.02591 |  2.2 | 1.074292 | 1.536237 |       0.0 | 1.025195 |       0.0 |       0.0 |       0.0 | 0.851648 |  14.419482 | 88.286398 | 8.695463 |   0.0 |   0.0 |   0.0 |  0.0 | 2.238617 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.02591 |
    | 2010-12-12 |  0.2 |  0.2 |              0.7 |                 7.386143 |       17.823519 |             78.7 |       5.7 |              1012.0 | 0.0 | 0.0 |     0.2 |                  78.7 |                   0.7 |                      7.386143 |            17.823519 |  0.188 |  0.2 |     0.2 |     4.56907 |          4.56907 |           2.023445 |          5.7 |                 6.197477 |                      6.197477 |                    6.197477 |                     5.565605 |                      0.451275 |                           0.451275 |             4.877414 |                  4.877414 |    1007.122586 |    1.28736 | 8.583333 |       0.0 |   7.605462 |              4.619919 |                 3.942123 |                  0.857985 |                   4.619919 |                 20.069431 |               19.469698 |                 9.809681 |   -15.449512 |      -15.527574 |        -8.951696 |        -15.449512 |             16.491228 |                 100.0 |               640.690101 |               95.237913 | 0.188 |  0.0846 |       49.0 |    0.424609 | 0.079826 | 0.035922 |  0.000323 |        0.728381 | 0.028458 |        0.0 |      0.0 |   0.065172 | 0.065172 | 0.394139 |  3.262355 | 0.0 |  0.725773 |   0.06961 |  0.000762 |    -1.267577 |     0.740796 |       0.0 |       0.0 | 0.000719 | 0.000719 | 12.168366 |   6.343759 | -27.443679 |  -7.686453 |   0.252262 |   7.609002 | 0.950212 | 0.270871 |  0.067022 | 0.021807 |  0.0 |  0.0 | 0.0 |  0.0 | 0.021807 | 3.017016 | 0.034919 |  2.2 | 0.883584 | 1.193466 |       0.0 | 1.196942 |       0.0 |       0.0 |       0.0 | 1.021868 |  15.397353 |  87.60584 | 0.021807 |   0.0 |   0.0 |   0.0 |  0.0 | 3.017016 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.034919 |
    | 2010-13-12 |  2.9 | -3.5 |              0.0 |                 7.373328 |        8.941324 |             90.5 |       2.2 |              1020.0 | 0.0 | 0.0 |    -3.5 |                  90.5 |                   0.0 |                      7.373328 |             8.941324 |  2.726 | -3.5 |    -3.5 |    1.763501 |         1.763501 |           0.780979 |          2.2 |                 4.713362 |                      4.713362 |                    4.713362 |                     4.799895 |                      0.354306 |                           0.354306 |             4.265593 |                  4.265593 |    1015.734407 |   1.315658 | 8.583333 |       0.0 |  19.546376 |              2.149589 |                 1.964804 |                  0.399209 |                   2.149589 |                 14.516908 |               15.898175 |                15.995542 |    -12.36732 |      -13.933371 |       -15.596333 |         -12.36732 |             42.727273 |                 100.0 |              2143.647518 |               98.258519 | 2.726 |   2.726 |       24.5 |     0.42358 |  1.15468 |  1.15468 |  -0.09776 |        0.730139 |      0.0 |        0.0 |      0.0 |   0.140414 | 0.140414 | 0.082949 |  0.460106 | 0.0 |  0.179863 |  0.098979 | -0.230795 |    -3.257729 |     0.759589 |       0.0 |       0.0 | 0.032032 |      0.0 |  2.560899 |   3.055801 | -16.049228 |   0.787274 |   0.445042 |  19.439408 | 0.985662 | 0.291459 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 1.614469 | 0.018686 |  2.2 | 2.101716 | 2.265198 |       0.0 | 1.170338 |  1.472341 |  1.472341 | -0.123826 | 0.804233 |    4.43431 | 87.425977 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 1.614469 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.018686 |
    | 2010-14-12 |  0.0 | -3.9 |              1.2 |                 7.362317 |       19.668068 |             88.1 |       3.1 |              1025.0 | 0.0 | 0.0 |    -3.9 |                  88.1 |                   1.2 |                      7.362317 |            19.668068 |    0.0 | -3.9 |    -3.9 |    2.484933 |         2.484933 |            1.10047 |          3.1 |                  4.57351 |                       4.57351 |                     4.57351 |                     4.635498 |                      0.344989 |                           0.344989 |             4.029262 |                  4.029262 |    1020.970738 |   1.324197 | 8.583333 |       0.0 |  22.328328 |              4.872436 |                 4.495746 |                  0.904881 |                   4.872436 |                 24.390008 |                 24.5834 |                 18.55758 |   -19.517572 |      -20.087654 |       -17.652699 |        -19.517572 |             30.322581 |                 100.0 |              2945.513093 |                98.56729 |   0.0 |     0.0 |       24.5 |    0.426796 |      0.0 |      0.0 |       0.0 |        0.719287 |      0.0 |        0.0 |      0.0 |   0.152651 | 0.152651 | 0.116681 |  0.739655 | 0.0 |  0.231474 |  0.129971 |       0.0 |    -3.721388 |     0.752267 |       0.0 |       0.0 | 0.034286 |      0.0 |  3.602334 |   4.012627 | -23.099877 |   0.671716 |   0.590111 |  22.337041 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.861499 | 0.009971 |  2.2 | 2.138244 | 2.148516 |       0.0 |  1.34573 |   1.34237 |   1.34237 |  -0.13254 |  0.88322 |  -9.310684 | 87.194503 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.861499 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.009971 |
    | 2010-15-12 |  0.0 | -6.3 |              3.0 |                 7.353123 |       26.831581 |             92.8 |       2.8 |              1023.0 | 0.0 | 0.0 |    -6.3 |                  92.8 |                   3.0 |                      7.353123 |            26.831581 |    0.0 | -6.3 |    -6.3 |    2.244455 |         2.244455 |           0.993973 |          2.8 |                 3.808888 |                      3.808888 |                    3.808888 |                     3.819843 |                      0.293396 |                           0.293396 |             3.534648 |                  3.534648 |    1019.465352 |    1.33374 | 8.583333 | -2.681477 |  21.487389 |              6.815485 |                 6.337302 |                  1.265733 |                   6.815485 |                 39.328283 |               33.634876 |                20.887899 |   -32.512798 |      -27.297574 |       -19.622166 |        -32.512798 |             33.571429 |                 100.0 |             22590.916958 |               99.383804 |   0.0 |     0.0 |       24.5 |    0.426446 |      0.0 |      0.0 |       0.0 |        0.709944 |      0.0 |        0.0 |      0.0 |   0.191547 | 0.010273 |  0.05613 |  0.174533 | 0.0 |  0.056208 |  0.058373 |       0.0 |    -6.262709 |      0.74599 |       0.0 |       0.0 | 0.052606 |      0.0 |   1.73293 |   1.802155 | -28.290032 |   0.133889 |   0.740472 |   21.55821 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.459705 | 0.005321 |  2.2 | 2.092386 | 2.092386 | -0.700761 | 1.521122 |  1.283997 |  1.283997 | -0.203361 | 0.956641 |  -22.21474 | 87.138295 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.459705 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.005321 |
    | 2010-16-12 |  6.7 | -3.5 |              0.0 |                 7.345755 |        8.837738 |             98.1 |       3.7 |              1000.0 | 0.0 | 0.0 |    -3.5 |                  98.1 |                   0.0 |                      7.345755 |             8.837738 |  6.298 | -3.5 |    -3.5 |    2.965888 |         2.965888 |           1.313465 |          3.7 |                 4.713362 |                      4.713362 |                    4.713362 |                      4.08012 |                      0.354306 |                           0.354306 |             4.623808 |                  4.623808 |     995.376192 |   1.289645 | 8.583333 | -6.397845 |  -5.966306 |              1.699365 |                 1.939612 |                  0.315596 |                   1.699365 |                   14.0252 |               17.117087 |                  5.98469 |   -12.325835 |      -15.177474 |        -5.669094 |        -12.325835 |             25.405405 |                 100.0 |              2119.731735 |               98.091099 | 6.298 |   6.298 |       24.5 |    0.426277 | 2.684694 | 2.684694 | -0.227298 |        0.730476 |      0.0 |        0.0 |      0.0 |   0.220502 |      0.0 | 0.020939 |  0.119463 | 0.0 |  0.033418 | -0.127124 | -0.533216 |    -5.403461 |     0.807715 |       0.0 |       0.0 | 0.165588 |      0.0 |  0.646464 |  -3.924729 | -15.899594 |  -7.807127 |  -0.075656 |  -6.062762 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.245304 | 0.002839 |  2.2 |  4.75614 |  4.75614 | -0.852403 | 1.165301 |  5.024428 |  5.024428 |  -0.64012 | 0.414364 |  -7.665101 | 87.104877 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.245304 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.002839 |
    | 2010-17-12 |  0.0 | -7.3 |              0.0 |                 7.340222 |        8.817416 |             96.6 |       0.9 |               991.0 | 0.0 | 0.0 |    -7.3 |                  96.6 |                   0.0 |                      7.340222 |             8.817416 |    0.0 | -7.3 |    -7.3 |    0.721432 |         0.721432 |           0.319491 |          0.9 |                  3.52531 |                       3.52531 |                     3.52531 |                     3.829309 |                      0.273951 |                           0.273951 |             3.405449 |                  3.405449 |     987.594551 |   1.296889 | 8.583333 | -2.207549 |  24.138067 |              1.782047 |                 1.961018 |                  0.330952 |                   1.782047 |                 14.906398 |                17.20791 |                20.027269 |    -13.12435 |      -15.246892 |       -19.696317 |         -13.12435 |            104.444444 |                 100.0 |             22620.895638 |               99.308862 |   0.0 |     0.0 |       24.5 |    0.434268 |      0.0 |      0.0 |       0.0 |        0.726874 |      0.0 |        0.0 |      0.0 |   0.396187 |      0.0 | 0.017159 |  0.022654 | 0.0 |  0.013151 |  0.060678 |       0.0 |     -6.23056 |     0.797895 |       0.0 |       0.0 | 0.195891 |      0.0 |  0.529746 |    1.87333 | -15.097483 |   2.685562 |   0.679155 |  24.255209 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.130897 | 0.001515 |  2.2 | 4.738981 | 4.738981 | -1.531558 | 1.220283 |   4.96375 |   4.96375 | -0.757262 | 0.479605 | -23.219835 | 87.091725 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.130897 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.001515 |
    | 2010-18-12 |  0.1 | -5.8 |              0.0 |                  7.33653 |        8.804185 |             97.7 |       1.3 |               997.0 | 0.0 | 0.0 |    -5.8 |                  97.7 |                   0.0 |                       7.33653 |             8.804185 |  0.094 | -5.8 |    -5.8 |    1.042069 |         1.042069 |           0.461488 |          1.3 |                 3.958114 |                      3.958114 |                    3.958114 |                     3.550629 |                      0.303557 |                           0.303557 |             3.867077 |                  3.867077 |     993.132923 |   1.297204 | 8.583333 | -6.687312 |   3.123354 |               1.84665 |                 1.972296 |                  0.342949 |                    1.84665 |                 14.549007 |               16.169634 |                 8.951016 |   -12.702357 |      -14.197337 |        -8.608066 |        -12.702357 |             72.307692 |                 100.0 |             22625.295215 |                99.28741 | 0.094 |   0.094 |       24.5 |    0.434217 | 0.040816 | 0.040816 | -0.005727 |         0.72489 |      0.0 |        0.0 |      0.0 |   0.335942 |      0.0 | 0.014212 |  0.024249 | 0.0 |  0.012143 |   -0.0494 | -0.013188 |    -7.207871 |     0.790253 |       0.0 |       0.0 | 0.226279 |      0.0 |  0.438755 |  -1.525138 | -14.874714 |  -3.855287 |  -0.238621 |   3.227641 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.069848 | 0.000808 |  2.2 | 4.765586 | 4.765586 | -1.298664 | 1.251767 |  5.066333 |  5.066333 | -0.874738 |  0.53469 | -17.759856 | 87.079583 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.069848 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000808 |
    | 2010-19-12 |  0.0 | -7.9 |              0.0 |                 7.334683 |        8.798044 |             95.8 |       1.9 |               997.0 | 0.0 | 0.0 |    -7.9 |                  95.8 |                   0.0 |                      7.334683 |             8.798044 |    0.0 | -7.9 |    -7.9 |    1.523023 |         1.523023 |           0.674482 |          1.9 |                 3.364296 |                      3.364296 |                    3.364296 |                     3.377786 |                      0.262827 |                           0.262827 |             3.222996 |                  3.222996 |     993.777004 |   1.307793 | 8.583333 | -5.114838 |  16.403556 |              1.909043 |                 1.989245 |                  0.354537 |                   1.909043 |                 15.062045 |               19.248216 |                15.848624 |   -13.153003 |      -17.258971 |       -15.494087 |        -13.153003 |             49.473684 |                 100.0 |             22629.357677 |               99.276693 |   0.0 |     0.0 |       24.5 |    0.434297 |      0.0 |      0.0 |       0.0 |        0.722332 |      0.0 |        0.0 |      0.0 |   0.407467 |      0.0 | 0.024803 |  0.107681 | 0.0 |  0.043279 |  0.027171 |       0.0 |    -7.848764 |     0.783015 |       0.0 |       0.0 | 0.248825 |      0.0 |  0.765755 |   0.838862 | -17.748227 |   0.157763 |   0.276499 |  16.490713 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.037272 | 0.000431 |  2.2 | 4.740783 | 4.740783 | -1.575163 | 1.293705 |  5.039162 |  5.039162 | -0.961895 | 0.590788 | -25.580078 | 87.036303 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.037272 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000431 |
    | 2010-20-12 |  0.2 | -5.7 |              0.2 |                 7.334683 |       15.464746 |             97.5 |       2.4 |               997.0 | 0.0 | 0.0 |    -5.7 |                  97.5 |                   0.2 |                      7.334683 |            15.464746 |  0.188 | -5.7 |    -5.7 |    1.923819 |         1.923819 |           0.851977 |          2.4 |                 3.988573 |                      3.988573 |                    3.988573 |                      3.50984 |                      0.305625 |                           0.305625 |             3.888858 |                  3.888858 |     993.111142 |   1.296708 | 8.583333 | -7.367063 |  -0.062903 |              3.434033 |                 3.515347 |                  0.637749 |                   3.434033 |                  16.12634 |               19.423467 |                 8.317516 |   -12.692307 |       -15.90812 |        -7.679767 |        -12.692307 |             39.166667 |                 100.0 |             22643.838175 |               99.276776 | 0.188 |   0.188 |       24.5 |    0.434222 | 0.081634 | 0.081634 | -0.011256 |        0.720843 |      0.0 |        0.0 |      0.0 |   0.323065 |      0.0 | 0.019118 |  0.083293 | 0.0 |  0.029842 | -0.072667 | -0.025922 |    -7.356579 |     0.777944 |       0.0 |       0.0 | 0.236868 |      0.0 |  0.590226 |  -2.243473 | -16.835878 |  -5.571345 |  -0.337532 |  -0.135051 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.019889 |  0.00023 |  2.2 | 4.803299 | 4.803299 | -1.248887 | 1.318854 |  5.218195 |  5.218195 | -0.915669 | 0.632595 | -16.933842 | 87.006461 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.019889 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.00023 |
    | 2010-21-12 |  0.2 | -5.6 |              0.8 |                  7.33653 |       17.879541 |             95.0 |       2.3 |              1004.0 | 0.0 | 0.0 |    -5.6 |                  95.0 |                   0.8 |                       7.33653 |            17.879541 |  0.188 | -5.6 |    -5.6 |     1.84366 |          1.84366 |           0.816478 |          2.3 |                 4.019239 |                      4.019239 |                    4.019239 |                     3.765689 |                      0.307706 |                           0.307706 |             3.818277 |                  3.818277 |    1000.181723 |   1.305372 | 8.583333 | -4.876947 |   9.426214 |              4.059792 |                  4.09197 |                  0.753961 |                   4.059792 |                 21.069426 |               22.540759 |                13.179295 |   -17.009634 |      -18.448789 |       -12.425334 |        -17.009634 |             40.869565 |                 100.0 |             22653.824092 |               99.287575 | 0.188 |   0.188 |       24.5 |     0.43441 | 0.081669 | 0.081669 | -0.011063 |         0.71894 |      0.0 |        0.0 |      0.0 |   0.351632 |      0.0 | 0.037879 |  0.157096 | 0.0 |  0.057855 | -0.009912 | -0.025467 |    -6.447982 |     0.772936 |       0.0 |       0.0 | 0.214835 |      0.0 |  1.169438 |  -0.306022 | -19.518859 |  -2.803738 |   0.099369 |   9.315574 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.010613 | 0.000123 |  2.2 | 4.847089 | 4.847089 | -1.359319 | 1.351802 |  5.334439 |  5.334439 | -0.830496 | 0.676098 | -17.776723 | 86.948606 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.010613 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000123 |
    | 2010-22-12 |  2.4 | -4.8 |              0.0 |                 7.340222 |        8.822161 |             93.8 |       4.6 |              1007.0 | 0.0 | 0.0 |    -4.8 |                  93.8 |                   0.0 |                      7.340222 |             8.822161 |  2.256 | -4.8 |    -4.8 |     3.68732 |          3.68732 |           1.632956 |          4.6 |                 4.272186 |                      4.272186 |                    4.272186 |                     3.934375 |                      0.324793 |                           0.324793 |             4.007311 |                  4.007311 |    1002.992689 |   1.305282 | 8.583333 | -5.119696 |   4.552284 |              1.903431 |                 1.981596 |                  0.353494 |                   1.903431 |                 14.580265 |               19.415762 |                10.319812 |   -12.676833 |      -17.434166 |        -9.966318 |        -12.676833 |             20.434783 |                 100.0 |             16084.246759 |               99.255771 | 2.256 |   2.256 |       24.5 |    0.434541 | 0.980325 | 0.980325 | -0.113827 |        0.724156 |      0.0 |        0.0 |      0.0 |   0.331372 |      0.0 | 0.069651 |  0.575838 | 0.0 |  0.133482 | -0.019179 | -0.261947 |     -5.87841 |     0.784244 |       0.0 |       0.0 | 0.243147 |      0.0 |  2.150367 |   -0.59212 | -19.776679 |  -4.974414 |  -0.192146 |   4.399784 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.005663 | 0.000066 |  2.2 | 5.757763 | 5.757763 | -1.280999 | 1.263645 |  6.629293 |  6.629293 | -0.939943 | 0.580974 | -13.745673 | 86.815125 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.005663 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000066 |
    | 2010-23-12 |  7.0 | -1.1 |              0.0 |                 7.345755 |        8.844396 |             94.9 |       8.6 |              1004.0 | 0.0 | 0.0 |    -1.1 |                  94.9 |                   0.0 |                      7.345755 |             8.844396 |   6.58 | -1.1 |    -1.1 |    6.893685 |         6.893685 |           3.052917 |          8.6 |                 5.634754 |                      5.634754 |                    5.634754 |                     5.018803 |                       0.41489 |                            0.41489 |             5.347381 |                  5.347381 |     998.652619 |    1.28304 | 8.583333 | -3.958754 |  -7.782537 |              1.616765 |                 1.878109 |                  0.300256 |                   1.616765 |                 13.655186 |               17.299788 |                 6.815137 |   -12.038422 |      -15.421679 |        -6.514881 |        -12.038422 |             10.930233 |                 100.0 |               827.402013 |               96.009151 |  6.58 |  5.0995 |     47.775 |    0.437273 | 2.877258 | 2.229875 | -0.093843 |        0.739219 |      0.0 |        0.0 |      0.0 |   0.227608 | 0.227608 | 0.112787 |  1.081358 | 0.0 |  0.167586 | -0.128959 | -0.214608 |    -2.661664 |     0.817199 |       0.0 |       0.0 | 0.186004 | 0.186004 |  3.482096 |  -3.981384 | -19.398744 |  -10.75154 |  -0.494969 |  -8.218043 | 0.991528 | 0.294938 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.003022 | 0.000035 |  2.2 | 8.107942 | 8.522235 |       0.0 |  1.04255 |  9.805804 | 10.460993 |       0.0 | 0.356595 |   2.620197 | 86.647539 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.003022 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000035 |
    | 2010-24-12 |  0.2 | -2.9 |              0.0 |                 7.353123 |        8.873743 |             85.4 |       7.6 |              1008.0 | 0.0 | 0.0 |    -2.9 |                  85.4 |                   0.0 |                      7.353123 |             8.873743 |  0.188 | -2.9 |    -2.9 |    6.092093 |         6.092093 |           2.697927 |          7.6 |                 4.930236 |                      4.930236 |                    4.930236 |                     4.849838 |                      0.368688 |                           0.368688 |             4.210422 |                  4.210422 |    1003.789578 |   1.297296 | 8.583333 |       0.0 |  18.717925 |              1.846336 |                 1.967612 |                  0.342891 |                   1.846336 |                 14.739672 |               20.121642 |                14.173388 |   -12.893336 |       -18.15403 |       -13.830497 |        -12.893336 |             12.368421 |                 100.0 |              1574.950553 |               97.662026 | 0.188 |   0.188 |     25.725 |    0.445567 | 0.083767 | 0.083767 | -0.005876 |        0.727695 |      0.0 |        0.0 |      0.0 |   0.384253 | 0.384253 | 0.259203 |  2.584493 | 0.0 |   0.42204 |  0.230252 | -0.013188 |    -3.119654 |     0.791933 |       0.0 |       0.0 | 0.219077 | 0.219077 |  8.002438 |   7.108622 | -24.676922 |  -1.387486 |   1.479546 |  19.551633 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.001613 | 0.000019 |  2.2 | 8.317664 | 8.346798 |       0.0 | 1.207507 |  9.908367 | 10.334975 |       0.0 | 0.522234 |  -7.514394 | 86.225499 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.001613 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000019 |
    | 2010-25-12 |  0.0 | -9.1 |              3.2 |                 7.362317 |       27.759642 |             87.2 |       3.8 |              1018.0 | 0.0 | 0.0 |    -9.1 |                  87.2 |                   3.2 |                      7.362317 |            27.759642 |    0.0 | -9.1 |    -9.1 |    3.046047 |         3.046047 |           1.348964 |          3.8 |                 3.061705 |                      3.061705 |                    3.061705 |                     3.380113 |                      0.241745 |                           0.241745 |             2.669806 |                  2.669806 |    1015.330194 |   1.341718 | 8.583333 | -2.164145 |  34.054801 |               6.36936 |                 6.392673 |                  1.182881 |                    6.36936 |                 43.173862 |               39.487127 |                26.840603 |   -36.804501 |      -33.094453 |       -25.657722 |        -36.804501 |             24.736842 |                 100.0 |             22915.499647 |               99.439273 |   0.0 |     0.0 |       24.5 |     0.44504 |      0.0 |      0.0 |       0.0 |        0.717192 |      0.0 |        0.0 |      0.0 |    0.88702 | 0.029135 | 0.092902 |  0.597976 | 0.0 |  0.149835 |  0.168382 |       0.0 |    -7.839946 |     0.770553 |       0.0 |       0.0 | 0.528014 | 0.426608 |  2.868178 |   5.198506 | -32.533642 |   5.239736 |   3.428989 |  36.095964 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |  0.00086 |  0.00001 |  2.2 | 8.253897 | 8.253897 | -3.316362 | 1.382899 | 10.166593 | 10.166593 | -0.392009 | 0.697626 | -32.985862 | 86.075664 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.00086 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.00001 |
    | 2010-26-12 |  1.2 | -8.6 |              4.3 |                 7.373328 |       32.328755 |             92.7 |       1.5 |              1019.0 | 0.0 | 0.0 |    -8.6 |                  92.7 |                   4.3 |                      7.373328 |            32.328755 |  1.128 | -8.6 |    -8.6 |    1.202387 |         1.202387 |           0.532486 |          1.5 |                 3.184716 |                      3.184716 |                    3.184716 |                     2.763681 |                      0.250344 |                           0.250344 |             2.952232 |                  2.952232 |    1016.047768 |   1.340358 | 8.583333 | -9.499928 |   5.337832 |              7.468211 |                 7.386314 |                  1.386953 |                   7.468211 |                 51.194924 |               35.492318 |                14.860836 |   -43.726713 |      -28.106004 |       -13.473883 |        -43.726713 |             62.666667 |                 100.0 |             22965.781967 |               99.503835 | 1.128 |   1.128 |       24.5 |    0.444762 | 0.501691 | 0.501691 | -0.104368 |        0.719417 |      0.0 |        0.0 |      0.0 |    1.02325 |      0.0 | 0.037798 |  -0.09817 | 0.0 | -0.044608 | -0.030655 | -0.234661 |   -10.389567 |     0.768992 |       0.0 |       0.0 | 0.701674 |      0.0 |   1.16695 |  -0.946426 | -28.738067 |  -5.103804 |   0.534888 |   7.423653 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.000459 | 0.000005 |  2.2 |  8.71779 |  8.71779 | -3.955618 | 1.343461 | 10.823557 | 10.823557 |  -2.71249 | 0.712034 | -29.740361 | 86.120272 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000459 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000005 |
    | 2010-27-12 |  0.9 | -5.3 |              0.0 |                 7.386143 |        9.004582 |             97.8 |       1.3 |              1017.0 | 0.0 | 0.0 |    -5.3 |                  97.8 |                   0.0 |                      7.386143 |             9.004582 |  0.846 | -5.3 |    -5.3 |    1.042069 |         1.042069 |           0.461488 |          1.3 |                 4.112496 |                      4.112496 |                    4.112496 |                     3.406464 |                       0.31402 |                            0.31402 |             4.022021 |                  4.022021 |    1012.977979 |   1.320718 | 8.583333 | -8.565224 |  -4.948671 |              2.065001 |                 2.046813 |                    0.3835 |                   2.065001 |                 14.439522 |                15.02886 |                 4.377783 |   -12.374521 |      -12.982048 |        -3.994283 |        -12.374521 |             72.307692 |                 100.0 |             22950.809628 |               99.578666 | 0.846 |   0.846 |       24.5 |    0.446153 | 0.377446 | 0.377446 | -0.048391 |         0.72085 |      0.0 |        0.0 |      0.0 |   0.614373 |      0.0 | 0.014124 |  0.026546 | 0.0 |   0.01334 | -0.096093 | -0.108462 |    -7.740445 |     0.770672 |       0.0 |       0.0 | 0.546942 |      0.0 |  0.436048 |  -2.966706 | -15.047098 |  -6.682867 |  -1.629003 |   -5.65529 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.000245 | 0.000003 |  2.2 | 9.081112 | 9.081112 | -2.375006 | 1.318735 | 11.388204 | 11.388204 | -2.114334 | 0.696539 | -16.208356 | 86.106931 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000245 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000003 |
    | 2010-28-12 |  0.0 | -7.7 |              3.4 |                 7.400749 |       28.982321 |             95.4 |       3.0 |              1022.0 | 0.0 | 0.0 |    -7.7 |                  95.4 |                   3.4 |                      7.400749 |            28.982321 |    0.0 | -7.7 |    -7.7 |    2.404774 |         2.404774 |           1.064971 |          3.0 |                 3.417227 |                      3.417227 |                    3.417227 |                     3.371846 |                      0.266491 |                           0.266491 |             3.260035 |                  3.260035 |    1018.739965 |   1.339598 | 8.583333 | -4.668007 |  19.219765 |              6.812219 |                 6.646645 |                  1.265126 |                   6.812219 |                 42.703706 |               36.290071 |                20.444526 |   -35.891487 |      -29.643426 |         -19.1794 |        -35.891487 |             31.333333 |                 100.0 |             22955.287001 |               99.664212 |   0.0 |     0.0 |       24.5 |    0.447243 |      0.0 |      0.0 |       0.0 |        0.718361 |      0.0 |        0.0 |      0.0 |   0.891611 |      0.0 | 0.033192 |  0.003934 | 0.0 |  0.001185 |  0.023609 |       0.0 |    -7.871301 |     0.764953 |       0.0 |       0.0 | 0.560921 |      0.0 |  1.024731 |   0.728896 | -29.596426 |   -0.63449 |   1.071731 |  19.273806 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.000131 | 0.000002 |  2.2 |  9.04792 |  9.04792 | -3.446737 | 1.362006 | 11.364595 | 11.364595 | -2.168375 | 0.750461 | -26.844788 | 86.105746 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000131 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000002 |
    | 2010-29-12 |  0.0 | -9.2 |              0.0 |                  7.41713 |        9.127645 |             94.9 |       2.7 |              1024.0 | 0.0 | 0.0 |    -9.2 |                  94.9 |                   0.0 |                       7.41713 |             9.127645 |    0.0 | -9.2 |    -9.2 |    2.164296 |         2.164296 |           0.958474 |          2.7 |                 3.037615 |                      3.037615 |                    3.037615 |                     2.925149 |                      0.240056 |                           0.240056 |             2.882696 |                  2.882696 |    1021.117304 |   1.350039 | 8.583333 | -7.731299 |  11.670462 |              2.193788 |                 2.106037 |                  0.407418 |                   2.193788 |                 15.336383 |               21.481317 |                13.988897 |   -13.142596 |      -19.375281 |       -13.581479 |        -13.142596 |             34.814815 |                 100.0 |             22955.684868 |               99.760303 |   0.0 |     0.0 |       24.5 |    0.447144 |      0.0 |      0.0 |       0.0 |        0.716645 |      0.0 |        0.0 |      0.0 |   0.849376 |      0.0 | 0.031207 |  0.202749 | 0.0 |  0.064362 |  0.008552 |       0.0 |    -9.676376 |     0.759655 |       0.0 |       0.0 | 0.688124 |      0.0 |  0.963447 |   0.264019 | -20.501997 |  -1.683303 |  -0.163269 |  12.162196 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |  0.00007 | 0.000001 |  2.2 | 9.016714 | 9.016714 | -3.283468 | 1.392793 | 11.356043 | 11.356043 | -2.660108 |  0.80356 | -29.931917 | 86.041384 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.00007 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000001 |
    | 2010-30-12 |  0.2 | -7.4 |              0.0 |                 7.435268 |        9.199989 |             96.0 |       1.4 |              1021.0 | 0.0 | 0.0 |    -7.4 |                  96.0 |                   0.0 |                      7.435268 |             9.199989 |  0.188 | -7.4 |    -7.4 |    1.122228 |         1.122228 |           0.496987 |          1.4 |                 3.498009 |                      3.498009 |                    3.498009 |                     3.078393 |                      0.272069 |                           0.272069 |             3.358088 |                  3.358088 |    1017.641912 |   1.336727 | 8.583333 | -8.620392 |   2.464458 |              2.238816 |                 2.131208 |                   0.41578 |                   2.238816 |                 14.959543 |               17.732181 |                 8.625328 |   -12.720728 |      -15.600973 |        -8.209548 |        -12.720728 |             67.142857 |                 100.0 |             22977.289238 |                99.86704 | 0.188 |   0.188 |       24.5 |     0.44705 | 0.084045 | 0.084045 | -0.015045 |        0.715513 |      0.0 |        0.0 |      0.0 |   0.760586 |      0.0 | 0.022296 |  0.074177 | 0.0 |  0.035231 | -0.044568 | -0.033653 |    -9.031135 |      0.75665 |       0.0 |       0.0 |  0.64763 |      0.0 |   0.68834 |  -1.375968 | -16.647598 |  -4.559314 |  -0.358285 |   2.274266 |      1.0 |      0.3 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.000037 |      0.0 |  2.2 | 9.078463 | 9.078463 | -2.940227 | 1.413544 | 11.504566 | 11.504566 | -2.503568 | 0.835128 | -23.813042 | 86.006153 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000037 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-31-12 |  0.9 |  1.7 |              4.0 |                 7.455145 |       32.062235 |             96.5 |       4.2 |              1016.0 | 0.0 | 0.0 |     1.7 |                  96.5 |                   4.0 |                      7.455145 |            32.062235 |  0.846 |  1.7 |     1.7 |    3.366683 |         3.366683 |            1.49096 |          4.2 |                 6.907942 |                      6.907942 |                    6.907942 |                     5.142434 |                      0.496631 |                           0.496631 |             6.666164 |                  6.666164 |    1009.333836 |   1.284547 | 8.583333 | -6.858156 |  -27.13824 |              7.939559 |                 7.476212 |                   1.47449 |                   7.939559 |                 40.225917 |               32.041784 |                  2.05765 |   -32.286358 |      -24.565572 |        -0.583161 |        -32.286358 |             22.380952 |                 100.0 |               486.247816 |               94.210429 | 0.846 | 0.06345 |       49.0 |    0.447235 | 0.378361 | 0.028377 |  0.029998 |        0.713641 |      0.0 |        0.0 |      0.0 |   0.290571 | 0.290571 | 0.060446 |  0.095015 | 0.0 |  0.027375 | -0.380944 |  0.067075 |    -2.335116 |      0.75237 |       0.0 |       0.0 | 0.181279 | 0.181279 |  1.866174 | -11.760972 | -28.218705 | -17.696145 |  -1.786959 | -28.873957 | 0.961493 | 0.277341 |       0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |  0.00002 |      0.0 |  2.2 | 9.337344 | 9.396378 |       0.0 | 1.448677 | 12.093865 | 12.353148 |       0.0 | 0.882053 |  11.908531 | 85.978778 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.00002 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |

>>> round_(model.check_waterbalance(conditions))
0.0

.. _lland_v4_acker_winter_ktschnee_daily:

acre (snow surface temperature)
-------------------------------

.. integration-test::

    >>> lnk(ACKER)
    >>> ktschnee(5.0)
    >>> conditions_acker_winter_ktschnee = test(
    ...     "lland_v4_acker_winter_ktschnee_daily",
    ...     axis1=(inputs.nied, fluxes.wada), axis2=(states.waes, states.wats),
    ...     get_conditions="2010-12-10")
    |       date | nied | teml | sunshineduration | possiblesunshineduration | globalradiation | relativehumidity | windspeed | atmosphericpressure |  qz | qzh | temltag | dailyrelativehumidity | dailysunshineduration | dailypossiblesunshineduration | dailyglobalradiation |   nkor | tkor | tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair |        g |        tz |         wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |   nbes |    sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |      evpo |  evi |      evb |       evs |     wnied | tempssurface | actualalbedo |   schmpot |      schm |  gefrpot |     gefr | wlatinz |   wlatsnow | wsensinz |  wsenssnow | wsurfinz |       wsurf |      sff |      fvg |      wada |       qdb | qib1 | qib2 | qbb | qkap |      qdgz |      qah |       qa | inzp | stinz | sinz | esnowinz | asinz |      wats |      waes |     esnow |     taus |       ebdn |      bowa |     qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 2010-01-12 |  0.0 | -5.8 |              3.5 |                 7.641029 |       31.942798 |             70.3 |       6.9 |              1015.0 | 0.0 | 0.0 |    -5.8 |                  70.3 |                   3.5 |                      7.641029 |            31.942798 |    0.0 | -5.8 |    -5.8 |     5.53098 |          5.53098 |            5.53098 |          6.9 |                 3.958114 |                      3.958114 |                         0.0 |                          0.0 |                      0.303557 |                           0.303557 |             2.782554 |                  2.782554 |    1012.217446 |   1.321192 | 8.583333 |       0.0 |       34.8 |             28.748518 |                      0.0 |                 28.748518 |                  28.748518 |                 46.472798 |                     0.0 |                      0.0 |    -17.72428 |             0.0 |              0.0 |         -17.72428 |             30.870386 |                 100.0 |               111.309423 |               99.512695 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.808179 |  0.0 |  0.55374 |       0.0 |       0.0 |          nan |          0.1 |       0.0 |       0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |      nan | -26.216667 |  71.44626 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-02-12 | 11.1 | -3.0 |              0.0 |                 7.609644 |        9.873898 |             85.0 |       5.2 |              1007.0 | 0.0 | 0.0 |    -3.0 |                  85.0 |                   0.0 |                      7.609644 |             9.873898 | 10.434 | -3.0 |    -3.0 |    4.168274 |         4.168274 |           4.168274 |          5.2 |                 4.893489 |                      4.893489 |                         0.0 |                     3.874639 |                      0.366257 |                           0.366257 |             4.159465 |                  4.159465 |    1002.840535 |   1.296511 | 8.583333 |   -7.5504 |  -4.536126 |              1.229506 |                      0.0 |                  1.229506 |                   1.229506 |                  14.78976 |                     0.0 |                28.671763 |   -13.560254 |             0.0 |       -27.442257 |        -13.560254 |             40.962628 |                 100.0 |               112.102145 |               99.530051 | 10.374 |  10.374 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.779906 |  0.0 |      0.0 | -0.140764 | -0.752835 |     -6.07759 |     0.875479 |       0.0 |       0.0 | 0.441058 |      0.0 |     0.0 |  -4.345837 |      0.0 | -26.680365 |      0.0 |   -3.583945 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 10.514764 | 10.514764 | -1.705016 | 0.084415 | -13.097208 |  71.44626 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-03-12 |  0.4 | -5.6 |              0.1 |                 7.579824 |         16.6835 |             95.9 |       2.0 |              1005.0 | 0.0 | 0.0 |    -5.6 |                  95.9 |                   0.1 |                      7.579824 |              16.6835 |  0.376 | -5.6 |    -5.6 |    1.603182 |         1.603182 |           1.603182 |          2.0 |                 4.019239 |                      4.019239 |                         0.0 |                     3.243359 |                      0.307706 |                           0.307706 |              3.85445 |                   3.85445 |     1001.14555 |   1.306656 | 8.583333 | -3.771996 |  12.646472 |              2.311234 |                      0.0 |                  2.311234 |                   2.311234 |                 15.387201 |                     0.0 |                32.295147 |   -13.075967 |             0.0 |       -29.983913 |        -13.075967 |            106.502833 |                 100.0 |               112.102145 |               99.527226 |  0.376 |   0.376 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.039468 |  0.0 |      0.0 | -0.159032 | -0.050934 |    -8.367644 |     0.861466 |       0.0 |       0.0 | 0.400697 |      0.0 |     0.0 |  -4.909851 |      0.0 | -12.634548 |      0.0 |   12.439513 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 11.049796 | 11.049796 | -1.548992 | 0.139381 | -17.160346 |  71.44626 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-04-12 |  1.2 | -3.8 |              3.7 |                 7.551596 |       31.807427 |             90.6 |       4.3 |              1008.0 | 0.0 | 0.0 |    -3.8 |                  90.6 |                   3.7 |                      7.551596 |            31.807427 |  1.128 | -3.8 |    -3.8 |    3.446842 |         3.446842 |           3.446842 |          4.3 |                 4.608124 |                      4.608124 |                         0.0 |                     3.312979 |                      0.347298 |                           0.347298 |              4.17496 |                   4.17496 |     1003.82504 |   1.301648 | 8.583333 |  -4.94218 |     8.4831 |              4.758969 |                      0.0 |                  4.758969 |                   4.758969 |                  43.14024 |                     0.0 |                57.151182 |   -38.381271 |             0.0 |       -52.392213 |        -38.381271 |             49.536202 |                 100.0 |               112.102145 |               99.524551 |  1.128 |   1.128 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.092386 |  0.0 |      0.0 | -0.369279 | -0.103687 |    -8.096584 |     0.850382 |       0.0 |       0.0 | 0.484345 |      0.0 |     0.0 | -11.400835 |      0.0 | -32.288606 |      0.0 |    8.702772 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 12.547075 | 12.547075 | -1.872351 | 0.186969 | -17.060113 |  71.44626 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-05-12 |  6.4 |  0.2 |              0.0 |                 7.524989 |        9.532511 |             96.5 |       4.0 |               996.0 | 0.0 | 0.0 |     0.2 |                  96.5 |                   0.0 |                      7.524989 |             9.532511 |  6.016 |  0.2 |     0.2 |    3.206365 |         3.206365 |           3.206365 |          4.0 |                 6.197477 |                      6.197477 |                         0.0 |                     5.231858 |                      0.451275 |                           0.451275 |             5.980565 |                  5.980565 |     990.019435 |   1.266438 | 8.583333 | -4.913313 |   -7.71462 |              1.485816 |                      0.0 |                  1.485816 |                   1.485816 |                 13.205056 |                     0.0 |                19.682213 |    -11.71924 |             0.0 |       -18.196397 |         -11.71924 |             53.251417 |                 100.0 |               112.102145 |               99.522031 |  6.016 |  2.7072 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.135099 |  0.0 |      0.0 | -0.304329 |  0.045165 |     -2.10322 |     0.844132 |       0.0 |       0.0 |  0.49672 |  0.49672 |     0.0 |  -9.395626 |      0.0 | -16.422384 |      0.0 |   -7.621612 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 16.009222 | 18.867404 |       0.0 | 0.215598 |   -0.76216 |  71.44626 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-06-12 |  0.0 | -1.1 |              0.1 |                 7.500028 |       16.141484 |             99.0 |       2.4 |               998.0 | 0.0 | 0.0 |    -1.1 |                  99.0 |                   0.1 |                      7.500028 |            16.141484 |    0.0 | -1.1 |    -1.1 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 5.634754 |                      5.634754 |                         0.0 |                      4.93464 |                       0.41489 |                            0.41489 |             5.578406 |                  5.578406 |     992.421594 |   1.275245 | 8.583333 | -0.219502 |   6.885814 |              3.043358 |                      0.0 |                  3.043358 |                   3.043358 |                 14.093918 |                     0.0 |                25.481561 |    -11.05056 |             0.0 |       -22.438203 |         -11.05056 |             88.752361 |                 100.0 |               112.102145 |               99.519666 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.000956 |  0.0 |      0.0 | -0.186363 |       0.0 |    -2.888059 |     0.811457 |       0.0 |       0.0 | 0.185939 | 0.185939 |     0.0 |   -5.75364 |      0.0 |  -9.079959 |      0.0 |    7.604604 | 0.996975 | 0.298188 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 16.355129 | 19.053767 |       0.0 |  0.39099 |   0.935359 |  71.44626 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-07-12 |  0.0 | -2.9 |              0.0 |                 7.476739 |        9.341429 |             99.2 |       1.9 |              1004.0 | 0.0 | 0.0 |    -2.9 |                  99.2 |                   0.0 |                      7.476739 |             9.341429 |    0.0 | -2.9 |    -2.9 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 4.930236 |                      4.930236 |                         0.0 |                     4.296908 |                      0.368688 |                           0.368688 |             4.890794 |                  4.890794 |     999.109206 |   1.291808 | 8.583333 |       0.0 |  12.295456 |              1.998128 |                      0.0 |                  1.998128 |                   1.998128 |                 13.815559 |                     0.0 |                28.102997 |    -11.81743 |             0.0 |       -26.104869 |         -11.81743 |            112.108246 |                 100.0 |               112.102145 |                99.51746 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.015761 |  0.0 |      0.0 | -0.150213 |       0.0 |    -4.724078 |       0.7861 |       0.0 |       0.0 | 0.279046 | 0.279046 |     0.0 |  -4.637556 |      0.0 |  -8.093137 |      0.0 |   13.374176 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 16.765312 |  19.20398 |       0.0 | 0.566382 |  -2.776764 |  71.44626 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-08-12 |  0.0 | -2.9 |              0.0 |                 7.455145 |        9.256764 |             99.1 |       1.8 |              1006.0 | 0.0 | 0.0 |    -2.9 |                  99.1 |                   0.0 |                      7.455145 |             9.256764 |    0.0 | -2.9 |    -2.9 |    1.442864 |         1.442864 |           1.442864 |          1.8 |                 4.930236 |                      4.930236 |                         0.0 |                     4.242152 |                      0.368688 |                           0.368688 |             4.885864 |                  4.885864 |    1001.114136 |   1.294388 | 8.583333 | -0.799708 |  10.437477 |              2.167471 |                      0.0 |                  2.167471 |                   2.167471 |                 13.821847 |                     0.0 |                27.402135 |   -11.654376 |             0.0 |       -25.234664 |        -11.654376 |            118.336481 |                 100.0 |               112.102145 |               99.515415 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.012523 |  0.0 |      0.0 | -0.158109 |       0.0 |     -4.89276 |      0.76585 |       0.0 |       0.0 | 0.344018 | 0.344018 |     0.0 |  -4.881333 |      0.0 |   -8.58597 |      0.0 |   11.767361 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 17.250194 | 19.362088 |       0.0 | 0.741774 |  -4.630907 |  71.44626 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-09-12 |  0.7 | -2.1 |              4.4 |                 7.435268 |       33.405815 |             96.9 |       2.7 |              1009.0 | 0.0 | 0.0 |    -2.1 |                  96.9 |                   4.4 |                      7.435268 |            33.405815 |  0.658 | -2.1 |    -2.1 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 5.233109 |                      5.233109 |                         0.0 |                     3.812253 |                      0.388644 |                           0.388644 |             5.070883 |                  5.070883 |    1003.929117 |   1.294334 | 8.583333 | -1.333701 |  12.531717 |                8.2195 |                      0.0 |                    8.2195 |                     8.2195 |                 46.364476 |                     0.0 |                57.533285 |   -38.144976 |             0.0 |       -49.313785 |        -38.144976 |             78.890988 |                 100.0 |               112.102145 |               99.513532 |  0.658 |   0.658 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.234616 |  0.0 |      0.0 | -0.391967 | -0.033425 |    -6.288533 |      0.75395 |       0.0 |       0.0 | 0.474106 | 0.474106 |     0.0 | -12.101292 |      0.0 |  -22.88143 |      0.0 |   14.331063 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 18.742201 | 20.412055 |       0.0 | 0.864457 |   -8.57929 |  71.44626 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-10-12 |  3.3 | -3.1 |              0.1 |                  7.41713 |       15.591158 |             95.0 |       3.0 |              1018.0 | 0.0 | 0.0 |    -3.1 |                  95.0 |                   0.1 |                       7.41713 |            15.591158 |  3.102 | -3.1 |    -3.1 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 4.856984 |                      4.856984 |                         0.0 |                     4.121719 |                       0.36384 |                            0.36384 |             4.614135 |                  4.614135 |    1013.385865 |   1.310959 | 8.583333 | -2.470836 |   6.577217 |              3.663893 |                      0.0 |                  3.663893 |                   3.663893 |                 14.892721 |                     0.0 |                29.943065 |   -11.228828 |             0.0 |       -26.279171 |        -11.228828 |             71.001889 |                 100.0 |               112.102145 |               99.511814 |  3.102 |   3.102 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.137342 |  0.0 |      0.0 | -0.164151 | -0.232614 |    -5.270657 |     0.765002 |       0.0 |       0.0 | 0.562244 | 0.562244 |     0.0 |  -5.067866 |      0.0 | -12.693215 |      0.0 |    8.518091 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 |      0.0 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 22.562863 | 23.678206 |       0.0 | 0.749982 |  -6.573175 |  71.44626 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 |      0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-11-12 |  9.4 |  5.1 |              0.2 |                 7.400749 |       15.891544 |             90.6 |       7.4 |              1004.0 | 0.0 | 0.0 |     5.1 |                  90.6 |                   0.2 |                      7.400749 |            15.891544 |  8.836 |  5.1 |     5.1 |    5.931775 |         5.931775 |           5.931775 |          7.4 |                 8.790172 |                      8.790172 |                         0.0 |                       6.1078 |                      0.614118 |                           0.614118 |             7.963896 |                  7.963896 |     996.036104 |   1.253213 | 8.583333 | -1.893074 | -11.358446 |              4.012333 |                      0.0 |                  4.012333 |                   4.012333 |                 13.525343 |                     0.0 |                -0.754591 |     -9.51301 |             0.0 |         4.766924 |          -9.51301 |              28.78455 |                 100.0 |               112.102145 |               99.510262 |  8.836 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.033243 | 0.06 | 0.338985 | -1.215856 |  2.122093 |          0.0 |     0.747518 | 23.713776 | 22.562863 |      0.0 |      0.0 |     0.0 | -37.537452 |      0.0 | -58.603285 |      0.0 | -100.907662 | 0.956772 | 0.274624 | 33.730062 | 11.067544 |  0.0 |  0.0 | 0.0 |  0.0 | 11.067544 | 2.849301 | 0.032978 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 |  0.93825 |  13.368604 | 93.769793 | 11.067544 |   0.0 |   0.0 |   0.0 |  0.0 | 2.849301 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.032978 |
    | 2010-12-12 |  0.2 |  0.2 |              0.7 |                 7.386143 |       17.823519 |             78.7 |       5.7 |              1012.0 | 0.0 | 0.0 |     0.2 |                  78.7 |                   0.7 |                      7.386143 |            17.823519 |  0.188 |  0.2 |     0.2 |     4.56907 |          4.56907 |            4.56907 |          5.7 |                 6.197477 |                      6.197477 |                         0.0 |                     5.049902 |                      0.451275 |                           0.451275 |             4.877414 |                  4.877414 |    1007.122586 |    1.28736 | 8.583333 |       0.0 |   7.031308 |               4.73515 |                      0.0 |                   4.73515 |                    4.73515 |                 20.069431 |                     0.0 |                34.816134 |   -15.334281 |             0.0 |       -30.080984 |        -15.334281 |             37.369415 |                 100.0 |                81.653408 |               98.490732 |  0.128 |  0.0576 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.429595 | 0.06 | 0.519628 |  0.082368 |  0.000519 |    -2.578858 |     0.734331 |       0.0 |       0.0 | 0.000786 | 0.000786 |     0.0 |   2.826474 |      0.0 | -25.872591 |      0.0 |    7.034866 | 0.951754 | 0.271751 |  0.045632 |  0.015062 |  0.0 |  0.0 | 0.0 |  0.0 |  0.015062 | 3.836775 | 0.044407 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |       0.0 | 1.109403 |  14.920629 | 93.280735 |  0.015062 |   0.0 |   0.0 |   0.0 |  0.0 | 3.836775 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.044407 |
    | 2010-13-12 |  2.9 | -3.5 |              0.0 |                 7.373328 |        8.941324 |             90.5 |       2.2 |              1020.0 | 0.0 | 0.0 |    -3.5 |                  90.5 |                   0.0 |                      7.373328 |             8.941324 |  2.726 | -3.5 |    -3.5 |    1.763501 |         1.763501 |           1.763501 |          2.2 |                 4.713362 |                      4.713362 |                         0.0 |                     3.977478 |                      0.354306 |                           0.354306 |             4.265593 |                  4.265593 |    1015.734407 |   1.315658 | 8.583333 |       0.0 |  15.675798 |              2.222307 |                      0.0 |                  2.222307 |                   2.222307 |                 14.516908 |                     0.0 |                31.068601 |   -12.294601 |             0.0 |       -28.846294 |        -12.294601 |             96.820757 |                 100.0 |                82.295785 |                98.52021 |  2.666 |   2.666 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.193737 |  0.0 |      0.0 | -0.079193 | -0.225715 |    -5.736347 |     0.751457 |       0.0 |       0.0 | 0.043585 |      0.0 |     0.0 |  -2.444947 |      0.0 | -10.782775 |      0.0 |   15.618572 | 0.974687 | 0.285005 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 2.050493 | 0.023733 | 0.06 |   0.0 |  0.0 |      0.0 |   nan |  2.745193 |  2.745193 | -0.168489 | 0.892385 |   7.828164 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 2.050493 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.023733 |
    | 2010-14-12 |  0.0 | -3.9 |              1.2 |                 7.362317 |       19.668068 |             88.1 |       3.1 |              1025.0 | 0.0 | 0.0 |    -3.9 |                  88.1 |                   1.2 |                      7.362317 |            19.668068 |    0.0 | -3.9 |    -3.9 |    2.484933 |         2.484933 |           2.484933 |          3.1 |                  4.57351 |                       4.57351 |                         0.0 |                     3.681219 |                      0.344989 |                           0.344989 |             4.029262 |                  4.029262 |    1020.970738 |   1.324197 | 8.583333 |       0.0 |  18.367591 |              5.038413 |                      0.0 |                  5.038413 |                   5.038413 |                 24.390008 |                     0.0 |                44.083316 |   -19.351595 |             0.0 |       -39.044903 |        -19.351595 |             68.711505 |                 100.0 |                82.295785 |               98.520648 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.280006 |  0.0 |      0.0 | -0.118568 |       0.0 |    -6.741743 |     0.743828 |       0.0 |       0.0 | 0.052586 |      0.0 |     0.0 |  -3.660568 |      0.0 | -16.981948 |      0.0 |   18.402387 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 1.094166 | 0.012664 | 0.06 |   0.0 |  0.0 |      0.0 |   nan |  2.863761 |  2.863761 | -0.203285 |  0.98328 |  -1.956093 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 1.094166 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.012664 |
    | 2010-15-12 |  0.0 | -6.3 |              3.0 |                 7.353123 |       26.831581 |             92.8 |       2.8 |              1023.0 | 0.0 | 0.0 |    -6.3 |                  92.8 |                   3.0 |                      7.353123 |            26.831581 |    0.0 | -6.3 |    -6.3 |    2.244455 |         2.244455 |           2.244455 |          2.8 |                 3.808888 |                      3.808888 |                         0.0 |                     2.883255 |                      0.293396 |                           0.293396 |             3.534648 |                  3.534648 |    1019.465352 |    1.33374 | 8.583333 | -0.563355 |  25.279421 |              7.049141 |                      0.0 |                  7.049141 |                   7.049141 |                 39.328283 |                     0.0 |                58.759349 |   -32.279141 |             0.0 |       -51.710208 |        -32.279141 |             76.073452 |                 100.0 |                82.295785 |               98.521014 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.031797 |  0.0 |      0.0 | -0.207622 |       0.0 |    -9.857997 |     0.737282 |       0.0 |       0.0 | 0.085596 |      0.0 |     0.0 |   -6.40996 |      0.0 |  -19.89322 |      0.0 |   25.407028 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 0.583859 | 0.006758 | 0.06 |   0.0 |  0.0 |      0.0 |   nan |  3.071382 |  3.071382 | -0.330893 | 1.068532 |  -18.65218 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.583859 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.006758 |
    | 2010-16-12 |  6.7 | -3.5 |              0.0 |                 7.345755 |        8.837738 |             98.1 |       3.7 |              1000.0 | 0.0 | 0.0 |    -3.5 |                  98.1 |                   0.0 |                      7.345755 |             8.837738 |  6.298 | -3.5 |    -3.5 |    2.965888 |         2.965888 |           2.965888 |          3.7 |                 4.713362 |                      4.713362 |                         0.0 |                     3.936324 |                      0.354306 |                           0.354306 |             4.623808 |                  4.623808 |     995.376192 |   1.289645 | 8.583333 | -5.371828 |   1.147589 |              1.775552 |                      0.0 |                  1.775552 |                   1.775552 |                   14.0252 |                     0.0 |                27.481485 |   -12.249648 |             0.0 |       -25.705933 |        -12.249648 |             57.569099 |                 100.0 |                82.295785 |               98.521307 |  6.298 |   6.298 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.03395 |  0.0 |      0.0 | -0.264364 | -0.533216 |    -5.871957 |     0.799094 |       0.0 |       0.0 | 0.326157 |      0.0 |     0.0 |  -8.161769 |      0.0 | -15.999844 |      0.0 |     1.54432 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 0.311554 | 0.003606 | 0.06 |   0.0 |  0.0 |      0.0 |   nan |  9.633746 |  9.633746 | -1.260839 | 0.471314 | -11.216436 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.311554 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.003606 |
    | 2010-17-12 |  0.0 | -7.3 |              0.0 |                 7.340222 |        8.817416 |             96.6 |       0.9 |               991.0 | 0.0 | 0.0 |    -7.3 |                  96.6 |                   0.0 |                      7.340222 |             8.817416 |    0.0 | -7.3 |    -7.3 |    0.721432 |         0.721432 |           0.721432 |          0.9 |                  3.52531 |                       3.52531 |                         0.0 |                     2.799842 |                      0.273951 |                           0.273951 |             3.405449 |                  3.405449 |     987.594551 |   1.296889 | 8.583333 | -3.230334 |  18.956495 |               1.85065 |                      0.0 |                   1.85065 |                    1.85065 |                 14.906398 |                     0.0 |                33.629061 |   -13.055748 |             0.0 |       -31.778411 |        -13.055748 |            236.672963 |                 100.0 |                82.295785 |               98.521527 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.001988 |  0.0 |      0.0 | -0.108899 |       0.0 |   -10.226692 |     0.790114 |       0.0 |       0.0 | 0.385193 |      0.0 |     0.0 |  -3.362059 |      0.0 |  -9.231639 |      0.0 |   19.184713 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 0.166248 | 0.001924 | 0.06 |   0.0 |  0.0 |      0.0 |   nan |  9.742645 |  9.742645 | -1.489058 |  0.53573 | -21.589597 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.166248 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.001924 |
    | 2010-18-12 |  0.1 | -5.8 |              0.0 |                  7.33653 |        8.804185 |             97.7 |       1.3 |               997.0 | 0.0 | 0.0 |    -5.8 |                  97.7 |                   0.0 |                       7.33653 |             8.804185 |  0.094 | -5.8 |    -5.8 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 3.958114 |                      3.958114 |                         0.0 |                     3.032971 |                      0.303557 |                           0.303557 |             3.867077 |                  3.867077 |     993.132923 |   1.297204 | 8.583333 | -6.217804 |   8.024872 |              1.910008 |                      0.0 |                  1.910008 |                   1.910008 |                 14.549007 |                     0.0 |                28.153948 |   -12.638999 |             0.0 |        -26.24394 |        -12.638999 |            163.850513 |                 100.0 |                82.295785 |               98.521673 |  0.094 |   0.094 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.003569 |  0.0 |      0.0 | -0.174381 | -0.013188 |    -9.219359 |     0.783057 |       0.0 |       0.0 | 0.465048 |      0.0 |     0.0 |  -5.383712 |      0.0 | -12.539848 |      0.0 |     8.32038 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 0.088712 | 0.001027 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 10.011026 | 10.011026 | -1.797754 | 0.590452 | -21.031136 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.088712 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.001027 |
    | 2010-19-12 |  0.0 | -7.9 |              0.0 |                 7.334683 |        8.798044 |             95.8 |       1.9 |               997.0 | 0.0 | 0.0 |    -7.9 |                  95.8 |                   0.0 |                      7.334683 |             8.798044 |    0.0 | -7.9 |    -7.9 |    1.523023 |         1.523023 |           1.523023 |          1.9 |                 3.364296 |                      3.364296 |                         0.0 |                     2.624563 |                      0.262827 |                           0.262827 |             3.222996 |                  3.222996 |     993.777004 |   1.307793 | 8.583333 | -6.056967 |  13.458957 |               1.96684 |                      0.0 |                   1.96684 |                    1.96684 |                 15.062045 |                     0.0 |                 34.21778 |   -13.095205 |             0.0 |        -32.25094 |        -13.095205 |            112.108246 |                 100.0 |                82.295785 |               98.521747 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.031907 |  0.0 |      0.0 | -0.151363 |       0.0 |   -11.034367 |     0.776446 |       0.0 |       0.0 | 0.519952 |      0.0 |     0.0 |  -4.673061 |      0.0 | -13.906677 |      0.0 |   13.671202 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 0.047338 | 0.000548 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 10.162388 | 10.162388 | -2.009999 | 0.645375 |  -25.90676 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.047338 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000548 |
    | 2010-20-12 |  0.2 | -5.7 |              0.2 |                 7.334683 |       15.464746 |             97.5 |       2.4 |               997.0 | 0.0 | 0.0 |    -5.7 |                  97.5 |                   0.2 |                      7.334683 |            15.464746 |  0.188 | -5.7 |    -5.7 |    1.923819 |         1.923819 |           1.923819 |          2.4 |                 3.988573 |                      3.988573 |                         0.0 |                     3.098752 |                      0.305625 |                           0.305625 |             3.888858 |                  3.888858 |     993.111142 |   1.296708 | 8.583333 | -7.461147 |    4.05317 |              3.527553 |                      0.0 |                  3.527553 |                   3.527553 |                  16.12634 |                     0.0 |                31.135077 |   -12.598787 |             0.0 |       -27.607524 |        -12.598787 |             88.752361 |                 100.0 |                82.295785 |               98.521747 |  0.188 |   0.188 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.019864 |  0.0 |      0.0 | -0.228727 | -0.025922 |    -8.947567 |     0.771897 |       0.0 |       0.0 | 0.526991 |      0.0 |     0.0 |  -7.061557 |      0.0 | -16.491506 |      0.0 |    4.054461 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 |  0.02526 | 0.000292 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 10.579116 | 10.579116 | -2.037212 | 0.685419 | -21.376597 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.02526 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000292 |
    | 2010-21-12 |  0.2 | -5.6 |              0.8 |                  7.33653 |       17.879541 |             95.0 |       2.3 |              1004.0 | 0.0 | 0.0 |    -5.6 |                  95.0 |                   0.8 |                       7.33653 |            17.879541 |  0.188 | -5.6 |    -5.6 |     1.84366 |          1.84366 |            1.84366 |          2.3 |                 4.019239 |                      4.019239 |                         0.0 |                     3.035378 |                      0.307706 |                           0.307706 |             3.818277 |                  3.818277 |    1000.181723 |   1.305372 | 8.583333 |  -6.15646 |   8.378078 |              4.156016 |                      0.0 |                  4.156016 |                   4.156016 |                 21.069426 |                     0.0 |                  37.1246 |    -16.91341 |             0.0 |       -32.968584 |         -16.91341 |             92.611159 |                 100.0 |                82.295785 |               98.521673 |  0.188 |   0.188 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.03385 |  0.0 |      0.0 | -0.220917 | -0.025467 |    -9.209318 |     0.767555 |       0.0 |       0.0 | 0.508871 |      0.0 |     0.0 |  -6.820417 |      0.0 | -17.865606 |      0.0 |    8.282561 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 0.013479 | 0.000156 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 10.988032 | 10.988032 | -1.967163 |  0.72551 | -21.171341 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.013479 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000156 |
    | 2010-22-12 |  2.4 | -4.8 |              0.0 |                 7.340222 |        8.822161 |             93.8 |       4.6 |              1007.0 | 0.0 | 0.0 |    -4.8 |                  93.8 |                   0.0 |                      7.340222 |             8.822161 |  2.256 | -4.8 |    -4.8 |     3.68732 |          3.68732 |            3.68732 |          4.6 |                 4.272186 |                      4.272186 |                         0.0 |                     3.528908 |                      0.324793 |                           0.324793 |             4.007311 |                  4.007311 |    1002.992689 |   1.305282 | 8.583333 | -6.097346 |   3.298469 |              1.944393 |                      0.0 |                  1.944393 |                   1.944393 |                 14.580265 |                     0.0 |                  31.4403 |   -12.635872 |             0.0 |       -29.495907 |        -12.635872 |              46.30558 |                 100.0 |                82.295785 |               98.521527 |  2.256 |   2.256 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.237721 |  0.0 |      0.0 | -0.215444 | -0.261947 |     -7.28687 |     0.779601 |       0.0 |       0.0 | 0.550873 |      0.0 |     0.0 |   -6.65146 |      0.0 | -19.645554 |      0.0 |    3.198893 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 0.007193 | 0.000083 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 13.459476 | 13.459476 | -2.129533 | 0.618694 | -15.886477 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.007193 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000083 |
    | 2010-23-12 |  7.0 | -1.1 |              0.0 |                 7.345755 |        8.844396 |             94.9 |       8.6 |              1004.0 | 0.0 | 0.0 |    -1.1 |                  94.9 |                   0.0 |                      7.345755 |             8.844396 |   6.58 | -1.1 |    -1.1 |    6.893685 |         6.893685 |           6.893685 |          8.6 |                 5.634754 |                      5.634754 |                         0.0 |                     4.983706 |                       0.41489 |                            0.41489 |             5.347381 |                  5.347381 |     998.652619 |    1.28304 | 8.583333 | -4.575305 |  -4.758233 |              1.641862 |                      0.0 |                  1.641862 |                   1.641862 |                 13.655186 |                     0.0 |                26.421965 |   -12.013324 |             0.0 |       -24.780102 |        -12.013324 |             24.768101 |                 100.0 |                82.295785 |               98.521307 |   6.58 |  5.0995 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.461334 |  0.0 |      0.0 | -0.270137 | -0.214608 |    -2.755662 |     0.814361 |       0.0 |       0.0 | 0.509438 | 0.509438 |     0.0 |  -8.340017 |      0.0 | -21.573106 |      0.0 |   -5.133021 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 0.003838 | 0.000044 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 19.325461 | 20.309613 |       0.0 | 0.373379 |  -2.544911 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.003838 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000044 |
    | 2010-24-12 |  0.2 | -2.9 |              0.0 |                 7.353123 |        8.873743 |             85.4 |       7.6 |              1008.0 | 0.0 | 0.0 |    -2.9 |                  85.4 |                   0.0 |                      7.353123 |             8.873743 |  0.188 | -2.9 |    -2.9 |    6.092093 |         6.092093 |           6.092093 |          7.6 |                 4.930236 |                      4.930236 |                         0.0 |                     4.287029 |                      0.368688 |                           0.368688 |             4.210422 |                  4.210422 |    1003.789578 |   1.297296 | 8.583333 | -0.732934 |  10.281215 |              1.867694 |                      0.0 |                  1.867694 |                   1.867694 |                 14.739672 |                     0.0 |                33.607237 |   -12.871979 |             0.0 |       -31.739544 |        -12.871979 |             28.027061 |                 100.0 |                82.295785 |               98.521014 |  0.188 |   0.188 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.132218 |  0.0 |      0.0 |  0.051303 | -0.013188 |     -4.75437 |     0.789526 |       0.0 |       0.0 | 0.328908 | 0.328908 |     0.0 |   1.583889 |      0.0 | -21.783933 |      0.0 |     11.5395 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 0.002048 | 0.000024 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 19.792706 | 20.446311 |       0.0 | 0.540148 |  -4.242792 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.002048 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000024 |
    | 2010-25-12 |  0.0 | -9.1 |              3.2 |                 7.362317 |       27.759642 |             87.2 |       3.8 |              1018.0 | 0.0 | 0.0 |    -9.1 |                  87.2 |                   3.2 |                      7.362317 |            27.759642 |    0.0 | -9.1 |    -9.1 |    3.046047 |         3.046047 |           3.046047 |          3.8 |                 3.061705 |                      3.061705 |                         0.0 |                     2.334119 |                      0.241745 |                           0.241745 |             2.669806 |                  2.669806 |    1015.330194 |   1.341718 | 8.583333 | -1.221924 |  29.029614 |              6.423146 |                      0.0 |                  6.423146 |                   6.423146 |                 43.173862 |                     0.0 |                65.873313 |   -36.750716 |             0.0 |       -59.450167 |        -36.750716 |             56.054123 |                 100.0 |                82.295785 |               98.520648 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.145715 |  0.0 |      0.0 | -0.131539 |       0.0 |   -12.484807 |     0.768616 |       0.0 |       0.0 | 0.800223 | 0.653605 |     0.0 |  -4.061041 |      0.0 |  -23.26606 |      0.0 |   32.123067 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 0.001093 | 0.000013 | 0.06 |   0.0 |  0.0 |      0.0 |   nan |  20.57785 |  20.57785 | -0.566787 |  0.71554 | -24.689073 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.001093 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000013 |
    | 2010-26-12 |  1.2 | -8.6 |              4.3 |                 7.373328 |       32.328755 |             92.7 |       1.5 |              1019.0 | 0.0 | 0.0 |    -8.6 |                  92.7 |                   4.3 |                      7.373328 |            32.328755 |  1.128 | -8.6 |    -8.6 |    1.202387 |         1.202387 |           1.202387 |          1.5 |                 3.184716 |                      3.184716 |                         0.0 |                     1.979012 |                      0.250344 |                           0.250344 |             2.952232 |                  2.952232 |    1016.047768 |   1.340358 | 8.583333 | -7.110453 |  17.693718 |              7.584795 |                      0.0 |                  7.584795 |                   7.584795 |                 51.194924 |                     0.0 |                59.607694 |   -43.610129 |             0.0 |       -52.022899 |        -43.610129 |            142.003778 |                 100.0 |                82.295785 |                98.52021 |  1.128 |   1.128 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.167474 |  0.0 |      0.0 | -0.217696 | -0.234661 |   -14.494219 |     0.765385 |       0.0 |       0.0 | 1.366309 |      0.0 |     0.0 |  -6.720982 |      0.0 |  -23.12785 |      0.0 |   22.174067 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 0.000583 | 0.000007 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 21.923545 | 21.923545 | -5.281797 |  0.74626 | -33.799457 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000583 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000007 |
    | 2010-27-12 |  0.9 | -5.3 |              0.0 |                 7.386143 |        9.004582 |             97.8 |       1.3 |              1017.0 | 0.0 | 0.0 |    -5.3 |                  97.8 |                   0.0 |                      7.386143 |             9.004582 |  0.846 | -5.3 |    -5.3 |    1.042069 |         1.042069 |           1.042069 |          1.3 |                 4.112496 |                      4.112496 |                         0.0 |                     2.963217 |                       0.31402 |                            0.31402 |             4.022021 |                  4.022021 |    1012.977979 |   1.320718 | 8.583333 | -9.734244 |  -0.558819 |              2.096939 |                      0.0 |                  2.096939 |                   2.096939 |                 14.439522 |                     0.0 |                23.743729 |   -12.342583 |             0.0 |       -21.646791 |        -12.342583 |            163.850513 |                 100.0 |                82.295785 |               98.519701 |  0.846 |   0.846 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.001963 |  0.0 |      0.0 | -0.221357 | -0.108462 |    -9.513339 |     0.767125 |       0.0 |       0.0 | 1.373666 |      0.0 |     0.0 |  -6.834013 |      0.0 | -15.451619 |      0.0 |   -0.638841 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 0.000311 | 0.000004 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 22.990902 | 22.990902 | -5.310236 | 0.729578 | -24.657306 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000311 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000004 |
    | 2010-28-12 |  0.0 | -7.7 |              3.4 |                 7.400749 |       28.982321 |             95.4 |       3.0 |              1022.0 | 0.0 | 0.0 |    -7.7 |                  95.4 |                   3.4 |                      7.400749 |            28.982321 |    0.0 | -7.7 |    -7.7 |    2.404774 |         2.404774 |           2.404774 |          3.0 |                 3.417227 |                      3.417227 |                         0.0 |                     2.358487 |                      0.266491 |                           0.266491 |             3.260035 |                  3.260035 |    1018.739965 |   1.339598 | 8.583333 | -7.101304 |  14.350787 |              6.889543 |                      0.0 |                  6.889543 |                   6.889543 |                 42.703706 |                     0.0 |                57.721392 |   -35.814163 |             0.0 |       -50.831849 |        -35.814163 |             71.001889 |                 100.0 |                82.295785 |                98.51912 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.120498 |  0.0 |      0.0 | -0.300538 |       0.0 |   -12.357122 |     0.762285 |       0.0 |       0.0 | 1.365727 |      0.0 |     0.0 |  -9.278588 |      0.0 | -27.233162 |      0.0 |   14.320099 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 0.000166 | 0.000002 | 0.06 |   0.0 |  0.0 |      0.0 |   nan |  23.29144 |  23.29144 | -5.279548 | 0.776804 |  -30.42476 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000166 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000002 |
    | 2010-29-12 |  0.0 | -9.2 |              0.0 |                  7.41713 |        9.127645 |             94.9 |       2.7 |              1024.0 | 0.0 | 0.0 |    -9.2 |                  94.9 |                   0.0 |                       7.41713 |             9.127645 |    0.0 | -9.2 |    -9.2 |    2.164296 |         2.164296 |           2.164296 |          2.7 |                 3.037615 |                      3.037615 |                         0.0 |                     2.327418 |                      0.240056 |                           0.240056 |             2.882696 |                  2.882696 |    1021.117304 |   1.350039 | 8.583333 | -8.762331 |   9.927002 |              2.212186 |                      0.0 |                  2.212186 |                   2.212186 |                 15.336383 |                     0.0 |                36.204868 |   -13.124197 |             0.0 |       -33.992682 |        -13.124197 |             78.890988 |                 100.0 |                82.295785 |               98.518469 |    0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.073401 |  0.0 |      0.0 | -0.172927 |       0.0 |   -12.520123 |     0.757639 |       0.0 |       0.0 |  1.51821 |      0.0 |     0.0 |  -5.338809 |      0.0 | -18.137414 |      0.0 |   10.516459 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 0.000089 | 0.000001 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 23.464367 | 23.464367 | -5.869005 | 0.824618 | -31.768428 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000089 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000001 |
    | 2010-30-12 |  0.2 | -7.4 |              0.0 |                 7.435268 |        9.199989 |             96.0 |       1.4 |              1021.0 | 0.0 | 0.0 |    -7.4 |                  96.0 |                   0.0 |                      7.435268 |             9.199989 |  0.188 | -7.4 |    -7.4 |    1.122228 |         1.122228 |           1.122228 |          1.4 |                 3.498009 |                      3.498009 |                         0.0 |                     2.536678 |                      0.272069 |                           0.272069 |             3.358088 |                  3.358088 |    1017.641912 |   1.336727 | 8.583333 | -9.149307 |   6.330234 |              2.254799 |                      0.0 |                  2.254799 |                   2.254799 |                 14.959543 |                     0.0 |                29.408579 |   -12.704745 |             0.0 |       -27.153781 |        -12.704745 |            152.146905 |                 100.0 |                82.295785 |               98.517748 |  0.188 |   0.188 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.019443 |  0.0 |      0.0 | -0.177732 | -0.033653 |   -11.457537 |     0.754913 |       0.0 |       0.0 | 1.510287 |      0.0 |     0.0 |   -5.48718 |      0.0 | -15.400644 |      0.0 |    6.265956 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 0.000047 | 0.000001 | 0.06 |   0.0 |  0.0 |      0.0 |   nan |   23.8301 |   23.8301 |  -5.83838 | 0.853891 | -29.515329 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000047 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000001 |
    | 2010-31-12 |  0.9 |  1.7 |              4.0 |                 7.455145 |       32.062235 |             96.5 |       4.2 |              1016.0 | 0.0 | 0.0 |     1.7 |                  96.5 |                   4.0 |                      7.455145 |            32.062235 |  0.846 |  1.7 |     1.7 |    3.366683 |         3.366683 |           3.366683 |          4.2 |                 6.907942 |                      6.907942 |                         0.0 |                     4.946677 |                      0.496631 |                           0.496631 |             6.666164 |                  6.666164 |    1009.333836 |   1.284547 | 8.583333 | -8.500415 |   -14.3022 |              7.979839 |                      0.0 |                  7.979839 |                   7.979839 |                 40.225917 |                     0.0 |                 47.67784 |   -32.246079 |             0.0 |       -39.698001 |        -32.246079 |             50.715635 |                 100.0 |                82.295785 |               98.516958 |  0.846 | 0.06345 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.125078 |  0.0 |      0.0 | -0.724068 |  0.067075 |    -2.855473 |     0.751114 |       0.0 |       0.0 | 0.974534 |  0.78255 |     0.0 | -22.354332 |      0.0 | -33.649879 |      0.0 |   -16.30621 |      1.0 |      0.3 |       0.0 |       0.0 |  0.0 |  0.0 | 0.0 |  0.0 |       0.0 | 0.000025 |      0.0 | 0.06 |   0.0 |  0.0 |      0.0 |   nan | 25.400167 | 25.400167 | -0.742159 | 0.896287 |  -6.629796 | 93.280735 |       0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.000025 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |

>>> round_(model.check_waterbalance(conditions))
0.0

>>> ktschnee(inf)

hourly simulation
_________________

>>> pub.timegrids = "1997-08-03", "1997-08-06", "1h"
>>> for parameter in control:
...     if parameter.TIME == True:
...         parameter.value = parameter.value/24
...     if parameter.TIME == False:
...         parameter.value = parameter.value*24
>>> for parameter in fixed:
...     parameter.restore()
>>> test = IntegrationTest(land)

.. _lland_v4_acker_summer_hourly:

acre (summer)
-------------

>>> inputs.nied.series = 0.0
>>> inputs.teml.series = (
...     16.9, 16.6, 16.4, 16.3, 16.0, 15.9, 16.0, 16.6, 17.4, 19.0, 20.3, 21.4, 21.3,
...     21.8, 22.9, 22.7, 22.5, 21.9, 21.4, 20.7, 19.4, 17.8, 17.0, 16.4, 15.9, 15.4,
...     15.2, 14.9, 14.3, 14.1, 14.3, 15.4, 16.8, 18.2, 19.5, 20.3, 21.0, 21.8, 21.9,
...     21.8, 22.2, 21.8, 20.9, 19.7, 17.9, 15.8, 14.8, 14.1, 13.4, 12.8, 12.3, 12.0,
...     11.7, 11.6, 12.6, 14.8, 17.1, 19.4, 20.6, 21.4, 22.5, 23.1, 23.4, 23.8, 24.0,
...     24.0, 23.7, 22.8, 21.3, 19.3, 18.3, 17.4)
>>> inputs.sunshineduration.series = (
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.2, 0.5, 0.7, 0.8, 0.5, 0.4, 0.5,
...     0.5, 0.3, 0.1, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.1, 0.9,
...     1.0, 1.0, 0.9, 0.8, 0.9, 0.8, 0.9, 0.9, 0.9, 1.0, 1.0, 1.0, 0.3, 0.0, 0.0, 0.0,
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.6, 0.9, 1.0, 0.9, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
...     1.0, 1.0, 1.0, 1.0, 0.5, 0.0, 0.0, 0.0)
>>> inputs.windspeed.series = (
...     0.8, 0.8, 0.8, 0.8, 0.8, 0.6, 0.9, 0.9, 0.9, 1.3, 1.5, 1.2, 1.3, 1.5, 1.9, 1.9,
...     2.3, 2.4, 2.5, 2.5, 2.2, 1.7, 1.7, 2.3, 2.3, 2.2, 2.3, 2.0, 2.3, 2.5, 2.4, 2.5,
...     2.5, 2.9, 3.1, 3.3, 3.3, 3.2, 2.5, 2.9, 3.6, 4.2, 4.2, 3.6, 3.2, 2.2, 2.2, 2.1,
...     1.8, 1.8, 1.6, 1.2, 1.9, 1.3, 1.6, 1.9, 3.2, 4.0, 4.6, 4.1, 4.1, 4.4, 4.6, 4.5,
...     4.7, 4.5, 4.1, 3.8, 2.5, 1.9, 2.2, 2.7)
>>> inputs.relativehumidity.series = (
...     95.1, 94.9, 95.9, 96.7, 97.2, 97.5, 97.7, 97.4, 96.8, 86.1, 76.8, 71.8, 67.5,
...     66.1, 63.4, 62.4, 61.1, 62.1, 67.0, 74.5, 81.2, 86.9, 90.1, 90.9, 88.0, 89.2,
...     88.8, 89.7, 93.0, 93.6, 94.6, 93.3, 86.9, 78.9, 75.8, 73.4, 68.3, 61.3, 60.6,
...     58.8, 57.2, 57.2, 59.9, 66.1, 71.7, 80.6, 85.9, 89.3, 91.4, 93.1, 94.6, 95.6,
...     96.2, 95.7, 95.1, 96.3, 89.9, 79.2, 73.4, 67.3, 62.2, 59.6, 55.4, 52.1, 49.5,
...     48.1, 49.4, 57.0, 65.9, 73.0, 76.7, 81.8)
>>> inputs.atmosphericpressure.series = (
...     1015.0, 1015.0, 1015.0, 1015.0, 1015.0, 1015.0, 1015.0, 1015.0, 1016.0, 1016.0,
...     1016.0, 1016.0, 1016.0, 1016.0, 1016.0, 1016.0, 1016.0, 1016.0, 1016.0, 1016.0,
...     1016.0, 1016.0, 1017.0, 1017.0, 1017.0, 1016.0, 1016.0, 1016.0, 1016.0, 1017.0,
...     1017.0, 1017.0, 1017.0, 1017.0, 1017.0, 1018.0, 1018.0, 1018.0, 1018.0, 1017.0,
...     1017.0, 1017.0, 1017.0, 1017.0, 1017.0, 1018.0, 1018.0, 1018.0, 1018.0, 1018.0,
...     1018.0, 1018.0, 1018.0, 1018.0, 1018.0, 1019.0, 1019.0, 1019.0, 1019.0, 1019.0,
...     1019.0, 1019.0, 1019.0, 1019.0, 1018.0, 1018.0, 1018.0, 1018.0, 1018.0, 1018.0,
...     1018.0, 1018.0)
>>> inlet.sequences.sim.series = 0.0

>>> inputs.possiblesunshineduration.series = (
...     0.0, 0.0, 0.0, 0.0, 0.42973358, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
...     1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.13640025, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...     0.39804385, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
...     1.0, 1.0, 0.10471052, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.36602248, 1.0, 1.0,
...     1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.07268915,
...     0.0, 0.0, 0.0)
>>> inputs.globalradiation.series = (
...     0.0, 0.0, 0.0, 0.0, 1.9436861, 21.9324417, 57.2561861, 109.3328444,
...     170.9491528, 311.762625, 501.5833, 615.0187278, 626.544325, 496.1334167,
...     419.5209944, 387.8873528, 278.4968722, 137.1386083, 51.0807139, 13.6328167,
...     0.1859444, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.5905167, 19.9509028,
...     62.985775, 285.7763889, 475.2252472, 614.0067194, 643.8065083, 600.2629806,
...     620.5141389, 599.4413861, 613.5512222, 517.1070472, 377.8366167, 252.2046389,
...     128.5767139, 42.0725861, 0.8128972, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...     1.4823472, 21.1355667, 125.88345, 305.9549444, 511.104575, 629.5192389,
...     765.8165889, 767.1458417, 736.2638528, 768.0061833, 735.9293694, 620.3749528,
...     450.0952639, 273.6141417, 133.3681056, 40.3265028, 0.8671222, 0.0, 0.0, 0.0)

>>> test.inits = (
...     (logs.loggedsunshineduration,
...      [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.2, 0.1, 0.2, 0.1, 0.2, 0.2, 0.3, 0.0,
...       0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
...     (logs.loggedpossiblesunshineduration,
...      [0.0, 0.0, 0.0, 0.0, 0.5, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
...       1.0, 1.0, 1.0, 1.0, 1.0, 0.2, 0.0, 0.0, 0.0]),
...     (logs.loggedglobalradiation,
...      [0.0, 0.0, 0.0, 0.0, 0.0, 27.8, 55.6, 138.9, 222.2, 305.6, 333.3, 388.9,
...       527.8, 444.4, 250.0, 222.2, 166.7, 111.1, 55.6, 27.8, 0.0, 0.0, 0.0, 0.0]),
...     (logs.loggedteml,
...      [13.2, 13.2, 13.1, 12.6, 12.7, 13.0, 13.5, 14.8, 16.2, 17.7, 18.8, 19.4, 20.4,
...       21.0, 21.5, 21.2, 20.4, 20.7, 20.2, 19.7, 19.0, 18.0, 17.5, 17.1]),
...     (logs.loggedrelativehumidity,
...      [95.1, 94.5, 94.8, 96.4, 96.6, 97.1, 97.1, 96.7, 92.2, 88.5, 81.1, 76.5, 75.1,
...       70.8, 68.9, 69.2, 75.0, 74.0, 77.4, 81.4, 85.3, 90.1, 92.3, 93.8]),
...     (logs.loggedwindspeed2m,
...      [0.8, 1.0, 1.2, 1.3, 0.9, 1.1, 1.3, 1.3, 1.9, 2.2, 1.8, 2.3, 2.4, 2.5, 2.4,
...       2.5, 2.1, 2.2, 1.7, 1.7, 1.3, 1.3, 0.7, 0.8]))

>>> for key, value in locals().copy().items():
...     if key.startswith("conditions_") and "states" in value:
...         for name in ("esnowinz", "esnow", "ebdn"):
...             value["states"][name] *= 24

.. integration-test::

    >>> test("lland_v4_acker_summer_hourly",
    ...      axis1=(fluxes.evb, fluxes.qah), axis2=states.bowa,
    ...      use_conditions=conditions_acker_summer)
    |                date | nied | teml | sunshineduration | possiblesunshineduration | globalradiation | relativehumidity | windspeed | atmosphericpressure |  qz | qzh |   temltag | dailyrelativehumidity | dailysunshineduration | dailypossiblesunshineduration | dailyglobalradiation | nkor | tkor |   tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair |    g |        tz |         wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |      evpo |       evi |      evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatinz | wlatsnow | wsensinz | wsenssnow | wsurfinz | wsurf | sff | fvg | wada | qdb | qib1 | qib2 | qbb | qkap | qdgz |      qah |       qa |     inzp | stinz | sinz | esnowinz | asinz | wats | waes | esnow | taus |        ebdn |      bowa | qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-08-03 00:00:00 |  0.0 | 16.9 |              0.0 |                      0.0 |             0.0 |             95.1 |       0.8 |              1015.0 | 0.0 | 0.0 | 17.279167 |             85.883333 |                   1.3 |                          15.7 |           136.579167 |  0.0 | 16.9 | 17.279167 |    0.641273 |         1.605886 |           0.641273 |          0.8 |                19.284227 |                     19.753091 |                         0.0 |                          0.0 |                      1.223615 |                           1.249589 |              18.3393 |                 16.964613 |       996.6607 |    1.21073 | -2.9 | 15.632162 |  -7.607027 |                   0.0 |                      0.0 |                       0.0 |                 103.800167 |                 10.408237 |                     0.0 |                      0.0 |   -10.408237 |             0.0 |              0.0 |          93.39193 |            176.542034 |                 100.0 |               100.628883 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.002953 | -0.002953 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.256857 | 0.071349 | 0.002953 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8729.609427 | 79.623355 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.256857 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.071349 |
    | 1997-08-03 01:00:00 |  0.0 | 16.6 |              0.0 |                      0.0 |             0.0 |             94.9 |       0.8 |              1015.0 | 0.0 | 0.0 | 17.241667 |             85.991667 |                   1.3 |                          15.7 |           136.579167 |  0.0 | 16.6 | 17.241667 |    0.641273 |         1.603439 |           0.641273 |          0.8 |                18.920184 |                      19.70628 |                         0.0 |                          0.0 |                       1.20339 |                           1.246999 |            17.955254 |                 16.945759 |     997.044746 |   1.212158 | -2.9 | 15.688646 |  -5.468121 |                   0.0 |                      0.0 |                       0.0 |                 103.800167 |                  10.41572 |                     0.0 |                      0.0 |    -10.41572 |             0.0 |              0.0 |         93.384446 |            176.542034 |                 100.0 |               100.628883 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.002673 | -0.002673 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.250222 | 0.069506 | 0.005626 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8732.177548 | 79.623355 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.250222 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.069506 |
    | 1997-08-03 02:00:00 |  0.0 | 16.4 |              0.0 |                      0.0 |             0.0 |             95.9 |       0.8 |              1015.0 | 0.0 | 0.0 |    17.175 |             86.233333 |                   1.3 |                          15.7 |           136.579167 |  0.0 | 16.4 |    17.175 |    0.641273 |         1.575992 |           0.641273 |          0.8 |                 18.68084 |                       19.6233 |                         0.0 |                          0.0 |                      1.190065 |                           1.242407 |            17.914926 |                 16.921826 |     997.085074 |   1.213014 | -2.9 | 15.719464 |  -4.083217 |                   0.0 |                      0.0 |                       0.0 |                 103.800167 |                 10.421627 |                     0.0 |                      0.0 |   -10.421627 |             0.0 |              0.0 |         93.378539 |            176.542034 |                 100.0 |               100.628883 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.004096 | -0.004096 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.243758 | 0.067711 | 0.009722 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8733.360765 | 79.623355 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.243758 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.067711 |
    | 1997-08-03 03:00:00 |  0.0 | 16.3 |              0.0 |                      0.0 |             0.0 |             96.7 |       0.8 |              1015.0 | 0.0 | 0.0 |   17.0625 |             86.708333 |                   1.3 |                          15.5 |           136.579167 |  0.0 | 16.3 |   17.0625 |    0.641273 |         1.548545 |           0.641273 |          0.8 |                18.562165 |                     19.483964 |                         0.0 |                          0.0 |                      1.183449 |                            1.23469 |            17.949613 |                  16.89422 |     997.050387 |   1.213417 | -2.9 | 15.733663 |  -3.398025 |                   0.0 |                      0.0 |                       0.0 |                 103.800167 |                 10.455184 |                     0.0 |                      0.0 |   -10.455184 |             0.0 |              0.0 |         93.344983 |            176.542034 |                 100.0 |               100.628883 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.005245 | -0.005245 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.237462 | 0.065962 | 0.014967 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  8733.85879 | 79.623355 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.237462 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.065962 |
    | 1997-08-03 04:00:00 |  0.0 | 16.0 |              0.0 |                 0.429734 |        1.943686 |             97.2 |       0.8 |              1015.0 | 0.0 | 0.0 | 16.908333 |             87.366667 |                   1.3 |                     14.929734 |            135.50182 |  0.0 | 16.0 | 16.908333 |    0.641273 |         1.504432 |           0.641273 |          0.8 |                18.210086 |                     19.294427 |                         0.0 |                          0.0 |                      1.163788 |                           1.224181 |            17.700204 |                 16.856897 |     997.299796 |    1.21479 | -2.9 | 15.739639 |  -1.562167 |              1.477201 |                      0.0 |                  1.477201 |                 102.981383 |                 10.555242 |                     0.0 |                      0.0 |    -9.078041 |             0.0 |              0.0 |         92.426141 |            176.542034 |                 100.0 |               100.628883 |               92.771059 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.004905 | -0.004905 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.231328 | 0.064258 | 0.019872 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8732.520957 | 79.623355 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.231328 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.064258 |
    | 1997-08-03 05:00:00 |  0.0 | 15.9 |              0.0 |                      1.0 |       21.932442 |             97.5 |       0.6 |              1015.0 | 0.0 | 0.0 | 16.729167 |             88.204167 |                   1.3 |                     14.929734 |           134.099005 |  0.0 | 15.9 | 16.729167 |    0.480955 |         1.453638 |           0.480955 |          0.6 |                18.094032 |                     19.076181 |                         0.0 |                          0.0 |                      1.157296 |                           1.212063 |            17.641681 |                 16.825987 |     997.358319 |   1.215237 | -2.9 | 15.723585 |  -1.058491 |             16.668656 |                      0.0 |                 16.668656 |                 101.915244 |                 10.545193 |                     0.0 |                      0.0 |     6.123463 |             0.0 |              0.0 |         91.370051 |            235.389378 |                 100.0 |               100.628883 |              100.447602 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.004992 |  0.004992 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.225353 | 0.062598 |  0.01488 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8730.679448 | 79.623355 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.225353 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.062598 |
    | 1997-08-03 06:00:00 |  0.0 | 16.0 |              0.0 |                      1.0 |       57.256186 |             97.7 |       0.9 |              1015.0 | 0.0 | 0.0 | 16.533333 |             89.191667 |                   1.3 |                     14.929734 |           131.855513 |  0.0 | 16.0 | 16.533333 |    0.721432 |         1.392031 |           0.721432 |          0.9 |                18.210086 |                     18.840106 |                         0.0 |                          0.0 |                      1.163788 |                           1.198935 |            17.791254 |                 16.803805 |     997.208746 |   1.214748 | -2.9 | 15.701487 |   -1.79108 |             43.514701 |                      0.0 |                 43.514701 |                  100.21019 |                  10.52516 |                     0.0 |                      0.0 |    32.989541 |             0.0 |              0.0 |          89.68503 |            156.926252 |                 100.0 |               100.628883 |              100.447602 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.026203 |   0.01488 | 0.008643 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.219531 | 0.060981 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8729.570528 | 79.614713 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.219531 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.060981 |
    | 1997-08-03 07:00:00 |  0.0 | 16.6 |              0.0 |                      1.0 |      109.332844 |             97.4 |       0.9 |              1015.0 | 0.0 | 0.0 |    16.375 |                90.125 |                   1.3 |                     14.929734 |           129.465215 |  0.0 | 16.6 |    16.375 |    0.721432 |         1.334591 |           0.721432 |          0.9 |                18.920184 |                     18.651109 |                         0.0 |                          0.0 |                       1.20339 |                           1.188408 |            18.428259 |                 16.809312 |     996.571741 |   1.211943 | -2.9 |  15.68818 |  -5.470922 |             83.092962 |                      0.0 |                 83.092962 |                  98.393563 |                  10.49073 |                     0.0 |                      0.0 |    72.602232 |             0.0 |              0.0 |         87.902834 |            156.926252 |                 100.0 |               100.640774 |               100.45605 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.057366 |       0.0 | 0.043948 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.213861 | 0.059406 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  8732.14145 | 79.570765 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.213861 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.059406 |
    | 1997-08-03 08:00:00 |  0.0 | 17.4 |              0.0 |                      1.0 |      170.949153 |             96.8 |       0.9 |              1016.0 | 0.0 | 0.0 | 16.216667 |                91.275 |                   1.3 |                     14.929734 |           127.329763 |  0.0 | 17.4 | 16.216667 |    0.721432 |         1.260484 |           0.721432 |          0.9 |                19.904589 |                     18.463773 |                         0.0 |                          0.0 |                      1.257963 |                           1.177959 |            19.267642 |                 16.852809 |     996.732358 |   1.209425 | -2.9 | 15.719031 | -10.085816 |            129.921356 |                      0.0 |                129.921356 |                   96.77062 |                 10.426918 |                     0.0 |                      0.0 |   119.494439 |             0.0 |              0.0 |         86.343702 |            156.926252 |                 100.0 |                100.70125 |              100.499005 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.096008 |       0.0 | 0.073907 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.208337 | 0.057871 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8739.327265 | 79.496858 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.208337 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.057871 |
    | 1997-08-03 09:00:00 |  0.0 | 19.0 |              0.2 |                      1.0 |      311.762625 |             86.1 |       1.3 |              1016.0 | 0.0 | 0.0 |   16.1125 |             91.991667 |                   1.5 |                     14.929734 |           129.903206 |  0.0 | 19.0 |   16.1125 |    1.042069 |         1.203904 |           1.042069 |          1.3 |                22.008543 |                     18.341425 |                         0.0 |                          0.0 |                      1.373407 |                           1.171128 |            18.949356 |                 16.872582 |     997.050644 |   1.202945 | -2.9 | 15.805261 | -19.168437 |            236.939595 |                      0.0 |                236.939595 |                  98.726436 |                 10.804921 |                     0.0 |                      0.0 |   226.134674 |             0.0 |              0.0 |         87.921515 |            108.641252 |                 100.0 |               100.802976 |              100.571226 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.222649 |       0.0 | 0.160936 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.202955 | 0.056376 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8755.595702 | 79.335922 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.202955 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.056376 |
    | 1997-08-03 10:00:00 |  0.0 | 20.3 |              0.5 |                      1.0 |        501.5833 |             76.8 |       1.5 |              1016.0 | 0.0 | 0.0 | 16.083333 |             92.241667 |                   1.7 |                     14.929734 |           132.285843 |  0.0 | 20.3 | 16.083333 |    1.202387 |         1.149836 |           1.202387 |          1.5 |                23.858503 |                     18.307295 |                         0.0 |                          0.0 |                      1.473678 |                           1.169221 |             18.32333 |                 16.886954 |      997.67667 |   1.197896 | -2.9 | 16.000482 | -25.797109 |            381.203308 |                      0.0 |                381.203308 |                 100.537241 |                 11.199993 |                     0.0 |                      0.0 |   370.003315 |             0.0 |              0.0 |         89.337248 |             94.155751 |                 100.0 |               101.024603 |              100.728426 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.389789 |       0.0 | 0.275473 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.197713 |  0.05492 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8778.492812 | 79.060449 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.197713 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.05492 |
    | 1997-08-03 11:00:00 |  0.0 | 21.4 |              0.7 |                      1.0 |      615.018728 |             71.8 |       1.2 |              1016.0 | 0.0 | 0.0 |    16.125 |             92.104167 |                   2.2 |                     14.929734 |           135.919957 |  0.0 | 21.4 |    16.125 |    0.961909 |         1.089916 |           0.961909 |          1.2 |                25.528421 |                     18.356069 |                         0.0 |                          0.0 |                      1.563281 |                           1.171946 |            18.329406 |                 16.906704 |     997.670594 |    1.19342 | -2.9 | 16.275247 | -30.748518 |            467.414233 |                      0.0 |                467.414233 |                 103.299167 |                 12.222123 |                     0.0 |                      0.0 |    455.19211 |             0.0 |              0.0 |         91.077044 |            117.694689 |                 100.0 |               101.404317 |              100.997295 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.46611 |       0.0 | 0.348731 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.192605 | 0.053502 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8806.341329 | 78.711718 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.192605 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.053502 |
    | 1997-08-03 12:00:00 |  0.0 | 21.3 |              0.8 |                      1.0 |      626.544325 |             67.5 |       1.3 |              1016.0 | 0.0 | 0.0 | 16.204167 |             91.729167 |                   2.8 |                     14.929734 |           145.821804 |  0.0 | 21.3 | 16.204167 |    1.042069 |         1.037502 |           1.042069 |          1.3 |                 25.37251 |                     18.449053 |                         0.0 |                          0.0 |                       1.55495 |                           1.177138 |            17.126444 |                 16.923163 |     998.873556 |   1.194363 | -2.9 | 16.609429 | -28.143424 |            476.173687 |                      0.0 |                476.173687 |                 110.824571 |                 13.461134 |                     0.0 |                      0.0 |   462.712553 |             0.0 |              0.0 |         97.363437 |            108.641252 |                 100.0 |               101.885667 |              101.337292 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.490849 |       0.0 | 0.360693 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.18763 |  0.05212 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8831.584754 | 78.351026 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.18763 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.05212 |
    | 1997-08-03 13:00:00 |  0.0 | 21.8 |              0.5 |                      1.0 |      496.133417 |             66.1 |       1.5 |              1016.0 | 0.0 | 0.0 | 16.329167 |             91.104167 |                   3.2 |                     14.929734 |           152.606529 |  0.0 | 21.8 | 16.329167 |    1.202387 |         1.012602 |           1.202387 |          1.5 |                26.160453 |                      18.59671 |                         0.0 |                          0.0 |                      1.596982 |                           1.185375 |             17.29206 |                 16.942378 |      998.70794 |   1.192265 | -2.9 |  16.91235 | -29.325898 |            377.061397 |                      0.0 |                377.061397 |                 115.980962 |                 14.297428 |                     0.0 |                      0.0 |   362.763969 |             0.0 |              0.0 |        101.683534 |             94.155751 |                 100.0 |               102.384309 |               101.68852 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.424547 |       0.0 | 0.303239 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.182784 | 0.050773 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8858.010651 | 78.047787 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.182784 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.050773 |
    | 1997-08-03 14:00:00 |  0.0 | 22.9 |              0.4 |                      1.0 |      419.520994 |             63.4 |       1.9 |              1016.0 | 0.0 | 0.0 | 16.545833 |             90.058333 |                   3.4 |                     14.929734 |           157.353237 |  0.0 | 22.9 | 16.545833 |    1.523023 |         0.984394 |           1.523023 |          1.9 |                27.969419 |                     18.855098 |                         0.0 |                          0.0 |                      1.692831 |                           1.199769 |            17.732611 |                 16.980587 |     998.267389 |   1.187639 | -2.9 | 17.229461 | -34.023233 |            318.835956 |                      0.0 |                318.835956 |                  119.58846 |                 14.726001 |                     0.0 |                      0.0 |   304.109955 |             0.0 |              0.0 |         104.86246 |             74.333488 |                 100.0 |               102.804149 |              101.983469 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.415176 |       0.0 | 0.282369 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.178062 | 0.049462 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8889.133884 | 77.765418 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.178062 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.049462 |
    | 1997-08-03 15:00:00 |  0.0 | 22.7 |              0.5 |                      1.0 |      387.887353 |             62.4 |       1.9 |              1016.0 | 0.0 | 0.0 | 16.816667 |             88.816667 |                   3.8 |                     14.929734 |           164.256877 |  0.0 | 22.7 | 16.816667 |    1.523023 |         0.968687 |           1.523023 |          1.9 |                27.632633 |                     19.182495 |                         0.0 |                          0.0 |                      1.675052 |                           1.217969 |            17.242763 |                 17.037252 |     998.757237 |    1.18866 | -2.9 |  17.60294 |  -30.58236 |            294.794388 |                      0.0 |                294.794388 |                 124.835227 |                 15.562805 |                     0.0 |                      0.0 |   279.231583 |             0.0 |              0.0 |        109.272421 |             74.333488 |                 100.0 |               103.195613 |              102.257848 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.393194 |       0.0 | 0.266636 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.173463 | 0.048184 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8916.816245 | 77.498782 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.173463 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.048184 |
    | 1997-08-03 16:00:00 |  0.0 | 22.5 |              0.5 |                      1.0 |      278.496872 |             61.1 |       2.3 |              1016.0 | 0.0 | 0.0 |   17.1375 |             87.333333 |                   4.1 |                     14.929734 |           170.073414 |  0.0 | 22.5 |   17.1375 |     1.84366 |         0.991339 |            1.84366 |          2.3 |                27.299387 |                     19.576758 |                         0.0 |                          0.0 |                      1.657431 |                            1.23983 |            16.679926 |                 17.097035 |     999.320074 |   1.189715 | -2.9 | 17.935128 |  -27.38923 |            211.657623 |                      0.0 |                211.657623 |                 129.255794 |                 16.204965 |                     0.0 |                      0.0 |   195.452657 |             0.0 |              0.0 |        113.050829 |             61.405925 |                 100.0 |               103.565733 |              102.516706 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.34217 |       0.0 | 0.218982 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.168982 | 0.046939 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8941.305475 | 77.279799 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.168982 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.046939 |
    | 1997-08-03 17:00:00 |  0.0 | 21.9 |              0.3 |                      1.0 |      137.138608 |             62.1 |       2.4 |              1016.0 | 0.0 | 0.0 |   17.4875 |                85.875 |                   4.4 |                     14.929734 |           173.470856 |  0.0 | 21.9 |   17.4875 |    1.923819 |         1.017332 |           1.923819 |          2.4 |                26.320577 |                     20.014927 |                         0.0 |                          0.0 |                      1.605502 |                           1.264057 |            16.345078 |                 17.187818 |     999.654922 |   1.192283 | -2.9 | 18.228999 | -22.026006 |            104.225342 |                      0.0 |                104.225342 |                  131.83785 |                 16.818923 |                     0.0 |                      0.0 |     87.40642 |             0.0 |              0.0 |        115.018928 |             58.847345 |                 100.0 |               103.870047 |              102.729133 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.236219 |       0.0 | 0.147975 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.164617 | 0.045727 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8960.431481 | 77.131824 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.164617 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.045727 |
    | 1997-08-03 18:00:00 |  0.0 | 21.4 |              0.1 |                      1.0 |       51.080714 |             67.0 |       2.5 |              1016.0 | 0.0 | 0.0 |   17.8375 |             84.620833 |                   4.5 |                     14.929734 |           174.440885 |  0.0 | 21.4 |   17.8375 |    2.003978 |         1.054998 |           2.003978 |          2.5 |                25.528421 |                     20.461645 |                         0.0 |                          0.0 |                      1.563281 |                           1.288683 |            17.104042 |                 17.314814 |     998.895958 |   1.193968 | -2.9 | 18.458511 | -17.648933 |             38.821343 |                      0.0 |                 38.821343 |                 132.575073 |                 16.973018 |                     0.0 |                      0.0 |    21.848325 |             0.0 |              0.0 |        115.602055 |             56.493451 |                 100.0 |                104.07586 |              102.872593 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.156077 |       0.0 | 0.095816 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.160365 | 0.044546 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8975.180414 | 77.036008 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.160365 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.044546 |
    | 1997-08-03 19:00:00 |  0.0 | 20.7 |              0.0 |                      1.0 |       13.632817 |             74.5 |       2.5 |              1016.0 | 0.0 | 0.0 | 18.170833 |                  83.7 |                   4.5 |                     15.429734 |           175.008919 |  0.0 | 20.7 | 18.170833 |    2.003978 |         1.100997 |           2.003978 |          2.5 |                24.454368 |                     20.895167 |                         0.0 |                          0.0 |                      1.505746 |                           1.312512 |            18.218504 |                 17.489255 |     997.781496 |   1.196313 | -2.9 | 18.635498 |  -12.38701 |             10.360941 |                      0.0 |                 10.360941 |                 133.006779 |                 16.556776 |                     0.0 |                      0.0 |    -6.195835 |             0.0 |              0.0 |        116.450003 |             56.493451 |                 100.0 |               104.209204 |              102.965449 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.09611 |       0.0 | 0.058452 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.156222 | 0.043395 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8984.667424 | 76.977556 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.156222 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.043395 |
    | 1997-08-03 20:00:00 |  0.0 | 19.4 |              0.0 |                   0.1364 |        0.185944 |             81.2 |       2.2 |              1016.0 | 0.0 | 0.0 | 18.454167 |             83.066667 |                   4.5 |                     15.566134 |           175.016667 |  0.0 | 19.4 | 18.454167 |    1.763501 |         1.120309 |           1.763501 |          2.2 |                22.563931 |                      21.26995 |                         0.0 |                          0.0 |                      1.403627 |                           1.333058 |            18.321912 |                 17.668238 |     997.678088 |   1.201582 | -2.9 | 18.749342 |  -3.903945 |              0.141318 |                      0.0 |                  0.141318 |                 133.012667 |                 16.343497 |                     0.0 |                      0.0 |    -16.20218 |             0.0 |              0.0 |        116.669169 |             64.197103 |                 100.0 |               104.290579 |               89.533309 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.049374 |       0.0 | 0.032413 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.152187 | 0.042274 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  8985.67137 | 76.945143 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.152187 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.042274 |
    | 1997-08-03 21:00:00 |  0.0 | 17.8 |              0.0 |                      0.0 |             0.0 |             86.9 |       1.7 |              1016.0 | 0.0 | 0.0 |     18.65 |               82.7375 |                   4.5 |                     15.566134 |           175.016667 |  0.0 | 17.8 |     18.65 |    1.362705 |         1.127089 |           1.362705 |          1.7 |                20.413369 |                     21.532411 |                         0.0 |                          0.0 |                      1.286025 |                           1.347418 |            17.739217 |                 17.815378 |     998.260783 |   1.208454 | -2.9 |  18.76139 |   5.768339 |                   0.0 |                      0.0 |                       0.0 |                 133.012667 |                 16.221809 |                     0.0 |                      0.0 |   -16.221809 |             0.0 |              0.0 |        116.790858 |             83.078604 |                 100.0 |               104.335713 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.019253 |       0.0 | 0.013439 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.148256 | 0.041182 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8977.003031 | 76.931705 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.148256 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.041182 |
    | 1997-08-03 22:00:00 |  0.0 | 17.0 |              0.0 |                      0.0 |             0.0 |             90.1 |       1.7 |              1017.0 | 0.0 | 0.0 | 18.808333 |             82.554167 |                   4.5 |                     15.566134 |           175.016667 |  0.0 | 17.0 | 18.808333 |    1.362705 |         1.142201 |           1.362705 |          1.7 |                19.406929 |                     21.746678 |                         0.0 |                          0.0 |                      1.230421 |                           1.359123 |            17.485643 |                 17.952788 |     999.514357 |   1.213101 | -2.9 |  18.65737 |   9.944218 |                   0.0 |                      0.0 |                       0.0 |                 133.012667 |                 16.101364 |                     0.0 |                      0.0 |   -16.101364 |             0.0 |              0.0 |        116.911302 |             83.078604 |                 100.0 |               104.354428 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.010292 |       0.0 | 0.007132 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.144426 | 0.040118 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8964.158813 | 76.924573 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.144426 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.040118 |
    | 1997-08-03 23:00:00 |  0.0 | 16.4 |              0.0 |                      0.0 |             0.0 |             90.9 |       2.3 |              1017.0 | 0.0 | 0.0 | 18.941667 |             82.379167 |                   4.5 |                     15.566134 |           175.016667 |  0.0 | 16.4 | 18.941667 |     1.84366 |         1.185687 |            1.84366 |          2.3 |                 18.68084 |                     21.928555 |                         0.0 |                          0.0 |                      1.190065 |                           1.369047 |            16.980884 |                 18.064561 |    1000.019116 |   1.215845 | -2.9 | 18.503239 |  12.619435 |                   0.0 |                      0.0 |                       0.0 |                 133.012667 |                 16.005095 |                     0.0 |                      0.0 |   -16.005095 |             0.0 |              0.0 |        117.007571 |             61.405925 |                 100.0 |                104.36436 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.014655 |       0.0 | 0.009242 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.140696 | 0.039082 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8948.639378 | 76.915331 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.140696 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.039082 |
    | 1997-08-04 00:00:00 |  0.0 | 15.9 |              0.0 |                      0.0 |             0.0 |             88.0 |       2.3 |              1017.0 | 0.0 | 0.0 |      18.9 |             82.083333 |                   4.5 |                     15.566134 |           175.016667 |  0.0 | 15.9 |      18.9 |     1.84366 |         1.235786 |            1.84366 |          2.3 |                18.094032 |                     21.871576 |                         0.0 |                          0.0 |                      1.157296 |                           1.365939 |            15.922748 |                 17.952919 |    1001.077252 |    1.21843 | -2.9 | 18.317006 |  14.502035 |                   0.0 |                      0.0 |                       0.0 |                 133.012667 |                 16.128293 |                     0.0 |                      0.0 |   -16.128293 |             0.0 |              0.0 |        116.884374 |             61.405925 |                 100.0 |               104.377231 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.024014 |       0.0 | 0.015057 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.137061 | 0.038073 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8931.237343 | 76.900275 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.137061 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.038073 |
    | 1997-08-04 01:00:00 |  0.0 | 15.4 |              0.0 |                      0.0 |             0.0 |             89.2 |       2.2 |              1016.0 | 0.0 | 0.0 |     18.85 |             81.845833 |                   4.5 |                     15.566134 |           175.016667 |  0.0 | 15.4 |     18.85 |    1.763501 |         1.282546 |           1.763501 |          2.2 |                17.523416 |                     21.803372 |                         0.0 |                          0.0 |                      1.125295 |                           1.362218 |            15.630887 |                 17.845152 |    1000.369113 |   1.219467 | -2.9 | 18.108181 |  16.249089 |                   0.0 |                      0.0 |                       0.0 |                 133.012667 |                 16.244859 |                     0.0 |                      0.0 |   -16.244859 |             0.0 |              0.0 |        116.767808 |             64.197103 |                 100.0 |               104.398203 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.018207 |       0.0 | 0.011515 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.133521 | 0.037089 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8912.088255 | 76.888759 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.133521 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.037089 |
    | 1997-08-04 02:00:00 |  0.0 | 15.2 |              0.0 |                      0.0 |             0.0 |             88.8 |       2.3 |              1016.0 | 0.0 | 0.0 |      18.8 |                 81.55 |                   4.5 |                     15.566134 |           175.016667 |  0.0 | 15.2 |      18.8 |     1.84366 |         1.332645 |            1.84366 |          2.3 |                17.299618 |                     21.735354 |                         0.0 |                          0.0 |                      1.112706 |                           1.358505 |            15.362061 |                 17.725181 |    1000.637939 |   1.220436 | -2.9 | 17.878392 |  16.070354 |                   0.0 |                      0.0 |                       0.0 |                 133.012667 |                 16.376926 |                     0.0 |                      0.0 |   -16.376926 |             0.0 |              0.0 |        116.635741 |             61.405925 |                 100.0 |               104.414243 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.020574 |       0.0 | 0.012795 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.130072 | 0.036131 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |   8893.1179 | 76.875964 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.130072 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.036131 |
    | 1997-08-04 03:00:00 |  0.0 | 14.9 |              0.0 |                      0.0 |             0.0 |             89.7 |       2.0 |              1016.0 | 0.0 | 0.0 | 18.741667 |             81.258333 |                   4.5 |                     15.566134 |           175.016667 |  0.0 | 14.9 | 18.741667 |    1.603182 |         1.372725 |           1.603182 |          2.0 |                16.968612 |                     21.656234 |                         0.0 |                          0.0 |                      1.094046 |                           1.354184 |            15.220845 |                 17.597495 |    1000.779155 |   1.221771 | -2.9 | 17.650748 |  16.504489 |                   0.0 |                      0.0 |                       0.0 |                 133.012667 |                 16.516685 |                     0.0 |                      0.0 |   -16.516685 |             0.0 |              0.0 |        116.495982 |             70.616813 |                 100.0 |               104.432066 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.013655 |       0.0 | 0.008851 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.126712 | 0.035198 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8873.713411 | 76.867113 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.126712 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.035198 |
    | 1997-08-04 04:00:00 |  0.0 | 14.3 |              0.0 |                 0.398044 |        1.590517 |             93.0 |       2.3 |              1016.0 | 0.0 | 0.0 | 18.670833 |             81.083333 |                   4.5 |                     15.534444 |           175.001952 |  0.0 | 14.3 | 18.670833 |     1.84366 |         1.422824 |            1.84366 |          2.3 |                16.323194 |                     21.560498 |                         0.0 |                          0.0 |                      1.057522 |                           1.348953 |             15.18057 |                  17.48197 |     1000.81943 |    1.22434 | -2.9 | 17.417894 |  18.707366 |              1.208793 |                      0.0 |                  1.208793 |                 133.001483 |                  16.65652 |                     0.0 |                      0.0 |   -15.447727 |             0.0 |              0.0 |        116.344963 |             61.405925 |                 100.0 |               104.444396 |               93.266466 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.007264 |       0.0 | 0.004365 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.123439 | 0.034289 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8852.106046 | 76.862748 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.123439 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.034289 |
    | 1997-08-04 05:00:00 |  0.0 | 14.1 |              0.0 |                      1.0 |       19.950903 |             93.6 |       2.5 |              1017.0 | 0.0 | 0.0 | 18.595833 |             80.920833 |                   4.5 |                     15.534444 |           174.919387 |  0.0 | 14.1 | 18.595833 |    2.003978 |         1.486284 |           2.003978 |          2.5 |                16.112886 |                     21.459534 |                         0.0 |                          0.0 |                      1.045579 |                           1.343433 |            15.081661 |                 17.365233 |    1001.918339 |   1.226451 | -2.9 | 17.158606 |  18.351635 |             15.162686 |                      0.0 |                 15.162686 |                 132.938735 |                 16.779096 |                     0.0 |                      0.0 |     -1.61641 |             0.0 |              0.0 |        116.159638 |             56.493451 |                 100.0 |               104.450477 |              103.133286 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.018172 |       0.0 | 0.010118 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.120251 | 0.033403 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  8830.85441 |  76.85263 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.120251 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.033403 |
    | 1997-08-04 06:00:00 |  0.0 | 14.3 |              0.1 |                      1.0 |       62.985775 |             94.6 |       2.4 |              1017.0 | 0.0 | 0.0 |    18.525 |             80.791667 |                   4.6 |                     15.534444 |            175.15812 |  0.0 | 14.3 |    18.525 |    1.923819 |         1.536383 |           1.923819 |          2.4 |                16.323194 |                     21.364558 |                         0.0 |                          0.0 |                      1.057522 |                           1.338237 |            15.441741 |                 17.260782 |    1001.558259 |   1.225432 | -2.9 | 16.903586 |  15.621518 |             47.869189 |                      0.0 |                 47.869189 |                 133.120171 |                 17.089561 |                     0.0 |                      0.0 |    30.779628 |             0.0 |              0.0 |         116.03061 |             58.847345 |                 100.0 |               104.464573 |              103.143085 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.04094 |       0.0 | 0.023223 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.117144 |  0.03254 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8812.332893 | 76.829407 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.117144 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.03254 |
    | 1997-08-04 07:00:00 |  0.0 | 15.4 |              0.9 |                      1.0 |      285.776389 |             93.3 |       2.5 |              1017.0 | 0.0 | 0.0 |    18.475 |             80.620833 |                   5.5 |                     15.534444 |           182.509935 |  0.0 | 15.4 |    18.475 |    2.003978 |         1.589823 |           2.003978 |          2.5 |                17.523416 |                     21.297737 |                         0.0 |                          0.0 |                      1.125295 |                            1.33458 |            16.349347 |                 17.170413 |    1000.650653 |   1.220347 | -2.9 | 16.681328 |   7.687968 |            217.190056 |                      0.0 |                217.190056 |                  138.70755 |                 19.012403 |                     0.0 |                      0.0 |   198.177653 |             0.0 |              0.0 |        119.695148 |             56.493451 |                 100.0 |               104.496928 |              103.165573 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.189733 |       0.0 |   0.1075 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.114118 |   0.0317 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8801.744925 | 76.721908 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.114118 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0317 |
    | 1997-08-04 08:00:00 |  0.0 | 16.8 |              1.0 |                      1.0 |      475.225247 |             86.9 |       2.5 |              1017.0 | 0.0 | 0.0 |     18.45 |             80.208333 |                   6.5 |                     15.534444 |           195.188105 |  0.0 | 16.8 |     18.45 |    2.003978 |         1.643262 |           2.003978 |          2.5 |                19.162205 |                     21.264396 |                         0.0 |                          0.0 |                      1.216842 |                           1.332754 |            16.651956 |                 17.055817 |    1000.348044 |   1.214317 | -2.9 | 16.554272 |  -1.474365 |            361.171188 |                      0.0 |                361.171188 |                  148.34296 |                 21.208047 |                     0.0 |                      0.0 |   339.963141 |             0.0 |              0.0 |        127.134913 |             56.493451 |                 100.0 |               104.646751 |              103.269652 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.343753 |       0.0 | 0.198377 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.111171 | 0.030881 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  8800.31929 | 76.523531 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.111171 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.030881 |
    | 1997-08-04 09:00:00 |  0.0 | 18.2 |              1.0 |                      1.0 |      614.006719 |             78.9 |       2.9 |              1017.0 | 0.0 | 0.0 | 18.416667 |             79.908333 |                   7.3 |                     15.534444 |           207.781609 |  0.0 | 18.2 | 18.416667 |    2.324615 |         1.696701 |           2.324615 |          2.9 |                20.933479 |                     21.220011 |                         0.0 |                          0.0 |                      1.314615 |                           1.330323 |            16.516515 |                 16.956557 |    1000.483485 |   1.208543 | -2.9 | 16.537165 |  -9.977011 |            466.645107 |                      0.0 |                466.645107 |                 157.914023 |                 22.996116 |                     0.0 |                      0.0 |   443.648991 |             0.0 |              0.0 |        134.917907 |             48.701251 |                 100.0 |               104.923431 |              103.461623 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.494205 |       0.0 | 0.274475 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.108299 | 0.030083 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8807.396301 | 76.249056 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.108299 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.030083 |
    | 1997-08-04 10:00:00 |  0.0 | 19.5 |              0.9 |                      1.0 |      643.806508 |             75.8 |       3.1 |              1017.0 | 0.0 | 0.0 | 18.383333 |             79.866667 |                   7.7 |                     15.534444 |           213.707576 |  0.0 | 19.5 | 18.383333 |    2.484933 |         1.750141 |           2.484933 |          3.1 |                22.704676 |                     21.175708 |                         0.0 |                          0.0 |                      1.411269 |                           1.327896 |            17.210144 |                 16.912332 |     999.789856 |   1.202862 | -2.9 | 16.622089 | -17.267466 |            489.292946 |                      0.0 |                489.292946 |                 162.417758 |                 23.882018 |                     0.0 |                      0.0 |   465.410928 |             0.0 |              0.0 |         138.53574 |             45.559235 |                 100.0 |               105.306685 |              103.727044 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.550473 |       0.0 |  0.30313 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.105502 | 0.029306 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8821.763767 | 75.945926 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.105502 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.029306 |
    | 1997-08-04 11:00:00 |  0.0 | 20.3 |              0.8 |                      1.0 |      600.262981 |             73.4 |       3.3 |              1018.0 | 0.0 | 0.0 |   18.3375 |             79.933333 |                   7.8 |                     15.534444 |           213.092753 |  0.0 | 20.3 |   18.3375 |    2.645251 |          1.82028 |           2.645251 |          3.3 |                23.858503 |                     21.114922 |                         0.0 |                          0.0 |                      1.473678 |                           1.324565 |            17.512141 |                 16.877861 |    1000.487859 |   1.200635 | -2.9 | 16.794499 | -21.033009 |            456.199865 |                      0.0 |                456.199865 |                 161.950493 |                   24.1295 |                     0.0 |                      0.0 |   432.070365 |             0.0 |              0.0 |        137.820992 |             42.798069 |                 100.0 |               105.730548 |              104.019919 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.548409 |       0.0 | 0.297686 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.102776 | 0.028549 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8839.896776 |  75.64824 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.102776 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.028549 |
    | 1997-08-04 12:00:00 |  0.0 | 21.0 |              0.9 |                      1.0 |      620.514139 |             68.3 |       3.3 |              1018.0 | 0.0 | 0.0 |    18.325 |             79.966667 |                   7.9 |                     15.534444 |           212.841496 |  0.0 | 21.0 |    18.325 |    2.645251 |         1.887079 |           2.645251 |          3.3 |                24.909749 |                     21.098371 |                         0.0 |                          0.0 |                      1.530181 |                           1.323658 |            17.013358 |                 16.871664 |    1000.986642 |   1.198001 | -2.9 | 17.012095 | -23.927432 |            471.590746 |                      0.0 |                471.590746 |                 161.759537 |                 24.341655 |                     0.0 |                      0.0 |    447.24909 |             0.0 |              0.0 |        137.417882 |             42.798069 |                 100.0 |               106.147416 |              104.307278 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.597585 |       0.0 | 0.327488 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.100122 | 0.027812 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8860.924208 | 75.320752 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.100122 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.027812 |
    | 1997-08-04 13:00:00 |  0.0 | 21.8 |              0.8 |                      1.0 |      599.441386 |             61.3 |       3.2 |              1018.0 | 0.0 | 0.0 |    18.325 |             79.766667 |                   8.2 |                     15.534444 |           217.145994 |  0.0 | 21.8 |    18.325 |    2.565092 |         1.943859 |           2.565092 |          3.2 |                26.160453 |                     21.098371 |                         0.0 |                          0.0 |                      1.596982 |                           1.323658 |            16.036358 |                 16.829467 |    1001.963642 |   1.195188 | -2.9 | 17.264424 | -27.213457 |            455.575453 |                      0.0 |                455.575453 |                 165.030956 |                 25.038979 |                     0.0 |                      0.0 |   430.536475 |             0.0 |              0.0 |        139.991977 |             44.135508 |                 100.0 |               106.606735 |              104.623118 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.621828 |       0.0 | 0.348705 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.097535 | 0.027093 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8885.237665 | 74.972048 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.097535 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.027093 |
    | 1997-08-04 14:00:00 |  0.0 | 21.9 |              0.9 |                      1.0 |      613.551222 |             60.6 |       2.5 |              1018.0 | 0.0 | 0.0 | 18.283333 |                 79.65 |                   8.7 |                     15.534444 |           225.230587 |  0.0 | 21.9 | 18.283333 |    2.003978 |         1.963899 |           2.003978 |          2.5 |                26.320577 |                     21.043281 |                         0.0 |                          0.0 |                      1.605502 |                           1.320638 |             15.95027 |                 16.760973 |     1002.04973 |   1.194821 | -2.9 | 17.556185 | -26.062888 |            466.298929 |                      0.0 |                466.298929 |                 171.175246 |                 26.187015 |                     0.0 |                      0.0 |   440.111914 |             0.0 |              0.0 |        144.988231 |             56.493451 |                 100.0 |                107.09665 |              104.959099 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.589213 |       0.0 | 0.361278 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.095016 | 0.026393 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8908.400553 | 74.610769 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.095016 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.026393 |
    | 1997-08-04 15:00:00 |  0.0 | 21.8 |              0.9 |                      1.0 |      517.107047 |             58.8 |       2.9 |              1017.0 | 0.0 | 0.0 | 18.245833 |                  79.5 |                   9.1 |                     15.534444 |           230.614741 |  0.0 | 21.8 | 18.245833 |    2.324615 |         1.997298 |           2.324615 |          2.9 |                26.160453 |                     20.993808 |                         0.0 |                          0.0 |                      1.596982 |                           1.317924 |            15.382347 |                 16.690077 |    1001.617653 |   1.194298 | -2.9 |  17.83414 |  -23.79516 |            393.001356 |                      0.0 |                393.001356 |                 175.267203 |                 27.142101 |                     0.0 |                      0.0 |   365.859255 |             0.0 |              0.0 |        148.125102 |             48.701251 |                 100.0 |               107.605159 |              105.306852 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.551455 |       0.0 | 0.320086 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.092562 | 0.025712 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8929.295714 | 74.290683 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.092562 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.025712 |
    | 1997-08-04 16:00:00 |  0.0 | 22.2 |              0.9 |                      1.0 |      377.836617 |             57.2 |       3.6 |              1017.0 | 0.0 | 0.0 | 18.233333 |               79.3375 |                   9.5 |                     15.534444 |           234.753897 |  0.0 | 22.2 | 18.233333 |    2.885728 |         2.040718 |           2.885728 |          3.6 |                26.806088 |                      20.97734 |                         0.0 |                          0.0 |                      1.631293 |                           1.317021 |            15.333082 |                 16.642897 |    1001.666918 |   1.192703 | -2.9 | 18.084882 | -24.690709 |            287.155829 |                      0.0 |                287.155829 |                 178.412962 |                 28.068928 |                     0.0 |                      0.0 |   259.086901 |             0.0 |              0.0 |        150.344034 |             39.231563 |                 100.0 |               108.056488 |              105.614671 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.505795 |       0.0 | 0.271017 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.090171 | 0.025047 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8951.086422 | 74.019667 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.090171 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.025047 |
    | 1997-08-04 17:00:00 |  0.0 | 21.8 |              1.0 |                      1.0 |      252.204639 |             57.2 |       4.2 |              1017.0 | 0.0 | 0.0 | 18.229167 |             79.133333 |                  10.2 |                     15.534444 |           239.548315 |  0.0 | 21.8 | 18.229167 |    3.366683 |         2.100837 |           3.366683 |          4.2 |                26.160453 |                     20.971853 |                         0.0 |                          0.0 |                      1.596982 |                            1.31672 |            14.963779 |                 16.595726 |    1002.036221 |   1.194485 | -2.9 |  18.34637 | -20.721778 |            191.675526 |                      0.0 |                191.675526 |                  182.05672 |                 29.636995 |                     0.0 |                      0.0 |    162.03853 |             0.0 |              0.0 |        152.419724 |             33.627054 |                 100.0 |               108.439226 |              105.875099 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.44464 |       0.0 | 0.220995 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.087841 |   0.0244 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |   8968.9082 | 73.798672 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.087841 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0244 |
    | 1997-08-04 18:00:00 |  0.0 | 20.9 |              1.0 |                      1.0 |      128.576714 |             59.9 |       4.2 |              1017.0 | 0.0 | 0.0 | 18.208333 |               78.8375 |                  11.1 |                     15.534444 |           242.777315 |  0.0 | 20.9 | 18.208333 |    3.366683 |         2.157616 |           3.366683 |          4.2 |                 24.75714 |                     20.944437 |                         0.0 |                          0.0 |                      1.521999 |                           1.315216 |            14.829527 |                  16.51207 |    1002.170473 |   1.198201 | -2.9 | 18.560232 |  -14.03861 |             97.718303 |                      0.0 |                 97.718303 |                  184.51076 |                 31.712576 |                     0.0 |                      0.0 |    66.005727 |             0.0 |              0.0 |        152.798184 |             33.627054 |                 100.0 |               108.751732 |              106.087324 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.32482 |       0.0 | 0.158739 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.085572 |  0.02377 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8980.046809 | 73.639933 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.085572 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.02377 |
    | 1997-08-04 19:00:00 |  0.0 | 19.7 |              1.0 |                      1.0 |       42.072586 |             66.1 |       3.6 |              1017.0 | 0.0 | 0.0 | 18.166667 |               78.4875 |                  12.1 |                     15.534444 |           243.962306 |  0.0 | 19.7 | 18.166667 |    2.885728 |         2.194356 |           2.885728 |          3.6 |                22.988466 |                     20.889699 |                         0.0 |                          0.0 |                      1.426658 |                           1.312212 |            15.195376 |                 16.395802 |    1001.804624 |   1.202947 | -2.9 | 18.693895 |   -6.03663 |             31.975165 |                      0.0 |                 31.975165 |                 185.411352 |                 34.091963 |                     0.0 |                      0.0 |    -2.116797 |             0.0 |              0.0 |        151.319389 |             39.231563 |                 100.0 |               108.976433 |              106.239691 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.183135 |       0.0 | 0.093951 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.083362 | 0.023156 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8983.183439 | 73.545981 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.083362 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.023156 |
    | 1997-08-04 20:00:00 |  0.0 | 17.9 |              0.3 |                 0.104711 |        0.812897 |             71.7 |       3.2 |              1017.0 | 0.0 | 0.0 | 18.104167 |             78.091667 |                  12.4 |                     15.502754 |           243.988429 |  0.0 | 17.9 | 18.104167 |    2.565092 |         2.227756 |           2.565092 |          3.2 |                20.542326 |                     20.807826 |                         0.0 |                          0.0 |                      1.293123 |                           1.307717 |            14.728848 |                 16.249178 |    1002.271152 |   1.210597 | -2.9 | 18.731535 |   4.989208 |              0.617802 |                      0.0 |                  0.617802 |                 185.431206 |                 35.120093 |                     0.0 |                      0.0 |   -34.502291 |             0.0 |              0.0 |        150.311113 |             44.135508 |                 100.0 |               109.109515 |               89.356954 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.100625 |       0.0 | 0.057093 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.081209 | 0.022558 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8975.294232 | 73.488889 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.081209 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.022558 |
    | 1997-08-04 21:00:00 |  0.0 | 15.8 |              0.0 |                      0.0 |             0.0 |             80.6 |       2.2 |              1018.0 | 0.0 | 0.0 | 18.020833 |             77.829167 |                  12.4 |                     15.502754 |           243.988429 |  0.0 | 15.8 | 18.020833 |    1.763501 |         2.244455 |           1.763501 |          2.2 |                17.978626 |                     20.699099 |                         0.0 |                          0.0 |                      1.150835 |                           1.301743 |            14.490772 |                 16.109936 |    1003.509228 |    1.22071 | -2.9 | 18.636864 |  17.021185 |                   0.0 |                      0.0 |                       0.0 |                 185.431206 |                 35.424509 |                     0.0 |                      0.0 |   -35.424509 |             0.0 |              0.0 |        150.006697 |             64.197103 |                 100.0 |                109.19042 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.029975 |       0.0 | 0.019046 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.079111 | 0.021975 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8955.373047 | 73.469843 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.079111 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.021975 |
    | 1997-08-04 22:00:00 |  0.0 | 14.8 |              0.0 |                      0.0 |             0.0 |             85.9 |       2.2 |              1018.0 | 0.0 | 0.0 | 17.929167 |             77.654167 |                  12.4 |                     15.502754 |           243.988429 |  0.0 | 14.8 | 17.929167 |    1.763501 |         2.261155 |           1.763501 |          2.2 |                16.859515 |                     20.580072 |                         0.0 |                          0.0 |                      1.087886 |                           1.295199 |            14.482324 |                 15.981284 |    1003.517676 |   1.224953 | -2.9 |  18.39781 |  21.586859 |                   0.0 |                      0.0 |                       0.0 |                 185.431206 |                 35.698198 |                     0.0 |                      0.0 |   -35.698198 |             0.0 |              0.0 |        149.733007 |             64.197103 |                 100.0 |               109.217416 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.012059 |       0.0 | 0.007576 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.077068 | 0.021408 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8930.886187 | 73.462266 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.077068 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.021408 |
    | 1997-08-04 23:00:00 |  0.0 | 14.1 |              0.0 |                      0.0 |             0.0 |             89.3 |       2.1 |              1018.0 | 0.0 | 0.0 | 17.833333 |               77.5875 |                  12.4 |                     15.502754 |           243.988429 |  0.0 | 14.1 | 17.833333 |    1.683342 |         2.254475 |           1.683342 |          2.1 |                16.112886 |                     20.456276 |                         0.0 |                          0.0 |                      1.045579 |                           1.288387 |            14.388807 |                 15.871513 |    1003.611193 |   1.227981 | -2.9 | 18.103968 |  24.023805 |                   0.0 |                      0.0 |                       0.0 |                 185.431206 |                 35.921753 |                     0.0 |                      0.0 |   -35.921753 |             0.0 |              0.0 |        149.509453 |             67.254108 |                 100.0 |               109.228155 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.000116 | -0.000116 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.075077 | 0.020855 | 0.000116 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8903.962382 | 73.462266 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.075077 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.020855 |
    | 1997-08-05 00:00:00 |  0.0 | 13.4 |              0.0 |                      0.0 |             0.0 |             91.4 |       1.8 |              1018.0 | 0.0 | 0.0 | 17.729167 |             77.729167 |                  12.4 |                     15.502754 |           243.988429 |  0.0 | 13.4 | 17.729167 |    1.442864 |         2.237776 |           1.442864 |          1.8 |                15.395377 |                     20.322453 |                         0.0 |                          0.0 |                      1.004677 |                           1.281018 |            14.071374 |                 15.796473 |    1003.928626 |   1.231127 | -2.9 | 17.780882 |  26.285292 |                   0.0 |                      0.0 |                       0.0 |                 185.431206 |                 36.050541 |                     0.0 |                      0.0 |   -36.050541 |             0.0 |              0.0 |        149.380664 |             78.463126 |                 100.0 |               109.228155 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.009023 | -0.009023 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.073138 | 0.020316 | 0.009139 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  8874.77709 | 73.462266 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.073138 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.020316 |
    | 1997-08-05 01:00:00 |  0.0 | 12.8 |              0.0 |                      0.0 |             0.0 |             93.1 |       1.8 |              1018.0 | 0.0 | 0.0 | 17.620833 |             77.891667 |                  12.4 |                     15.502754 |           243.988429 |  0.0 | 12.8 | 17.620833 |    1.442864 |         2.224416 |           1.442864 |          1.8 |                14.802811 |                     20.184089 |                         0.0 |                          0.0 |                      0.970706 |                           1.273391 |            13.781417 |                 15.721724 |    1004.218583 |   1.233843 | -2.9 | 17.430658 |  27.783951 |                   0.0 |                      0.0 |                       0.0 |                 185.431206 |                 36.176259 |                     0.0 |                      0.0 |   -36.176259 |             0.0 |              0.0 |        149.254947 |             78.463126 |                 100.0 |               109.228155 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.013264 | -0.013264 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.071248 | 0.019791 | 0.022403 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  8844.09314 | 73.462266 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.071248 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.019791 |
    | 1997-08-05 02:00:00 |  0.0 | 12.3 |              0.0 |                      0.0 |             0.0 |             94.6 |       1.6 |              1018.0 | 0.0 | 0.0 |      17.5 |             78.133333 |                  12.4 |                     15.502754 |           243.988429 |  0.0 | 12.3 |      17.5 |    1.282546 |         2.201036 |           1.282546 |          1.6 |                14.324376 |                     20.030733 |                         0.0 |                          0.0 |                      0.943147 |                            1.26493 |            13.550859 |                 15.650679 |    1004.449141 |   1.236111 | -2.9 | 17.062451 |  28.574706 |                   0.0 |                      0.0 |                       0.0 |                 185.431206 |                 36.284611 |                     0.0 |                      0.0 |   -36.284611 |             0.0 |              0.0 |        149.146595 |             88.271017 |                 100.0 |               109.228155 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.017718 | -0.017718 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.069408 |  0.01928 | 0.040121 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8812.618434 | 73.462266 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.069408 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.01928 |
    | 1997-08-05 03:00:00 |  0.0 | 12.0 |              0.0 |                      0.0 |             0.0 |             95.6 |       1.2 |              1018.0 | 0.0 | 0.0 | 17.379167 |             78.379167 |                  12.4 |                     15.502754 |           243.988429 |  0.0 | 12.0 | 17.379167 |    0.961909 |         2.174316 |           0.961909 |          1.2 |                 14.04387 |                     19.878396 |                         0.0 |                          0.0 |                      0.926933 |                           1.256516 |            13.425939 |                 15.580521 |    1004.574061 |   1.237469 | -2.9 | 16.684755 |  28.108527 |                   0.0 |                      0.0 |                       0.0 |                 185.431206 |                 36.390891 |                     0.0 |                      0.0 |   -36.390891 |             0.0 |              0.0 |        149.040314 |            117.694689 |                 100.0 |               109.228155 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.020716 | -0.020716 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.067615 | 0.018782 | 0.060837 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8781.609907 | 73.462266 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.067615 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.018782 |
    | 1997-08-05 04:00:00 |  0.0 | 11.7 |              0.0 |                 0.366022 |        1.482347 |             96.2 |       1.9 |              1018.0 | 0.0 | 0.0 | 17.270833 |               78.5125 |                  12.4 |                     15.470733 |           243.983922 |  0.0 | 11.7 | 17.270833 |    1.523023 |         2.160956 |           1.523023 |          1.9 |                13.768192 |                      19.74268 |                         0.0 |                          0.0 |                      0.910957 |                           1.249013 |            13.245001 |                 15.500472 |    1004.754999 |   1.238856 | -2.9 | 16.312652 |  27.675913 |              1.126584 |                      0.0 |                  1.126584 |                  185.42778 |                 36.589063 |                     0.0 |                      0.0 |   -35.462479 |             0.0 |              0.0 |        148.838717 |             74.333488 |                 100.0 |               109.228155 |               93.746399 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.019578 | -0.019578 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.065868 | 0.018297 | 0.080415 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8751.033993 | 73.462266 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.065868 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.018297 |
    | 1997-08-05 05:00:00 |  0.0 | 11.6 |              0.0 |                      1.0 |       21.135567 |             95.7 |       1.3 |              1018.0 | 0.0 | 0.0 | 17.166667 |                  78.6 |                  12.4 |                     15.470733 |           244.033283 |  0.0 | 11.6 | 17.166667 |    1.042069 |         2.120877 |           1.042069 |          1.3 |                 13.67736 |                     19.612949 |                         0.0 |                          0.0 |                      0.905683 |                           1.241834 |            13.089234 |                 15.415778 |    1004.910766 |   1.239363 | -2.9 | 15.945741 |  26.074448 |             16.063031 |                      0.0 |                 16.063031 |                 185.465295 |                  36.74524 |                     0.0 |                      0.0 |   -20.682209 |             0.0 |              0.0 |        148.720055 |            108.641252 |                 100.0 |               109.228155 |              106.410154 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.009263 | -0.009263 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.064167 | 0.017824 | 0.089679 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8722.059546 | 73.462266 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.064167 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.017824 |
    | 1997-08-05 06:00:00 |  0.0 | 12.6 |              0.6 |                      1.0 |       125.88345 |             95.1 |       1.6 |              1018.0 | 0.0 | 0.0 | 17.095833 |             78.620833 |                  12.9 |                     15.470733 |           246.654019 |  0.0 | 12.6 | 17.095833 |    1.282546 |         2.094157 |           1.282546 |          1.6 |                14.609782 |                     19.525158 |                         0.0 |                          0.0 |                      0.959602 |                           1.236973 |            13.893903 |                 15.350842 |    1004.106097 |   1.234655 | -2.9 | 15.598048 |  17.988287 |             95.671422 |                      0.0 |                 95.671422 |                 187.457054 |                  38.00446 |                     0.0 |                      0.0 |    57.666962 |             0.0 |              0.0 |        149.452594 |             88.271017 |                 100.0 |               109.228155 |              106.410154 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.052361 |  0.052361 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.06251 | 0.017364 | 0.037317 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8701.171259 | 73.462266 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.06251 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.017364 |
    | 1997-08-05 07:00:00 |  0.0 | 14.8 |              0.9 |                      1.0 |      305.954944 |             96.3 |       1.9 |              1019.0 | 0.0 | 0.0 | 17.070833 |             78.745833 |                  12.9 |                     15.470733 |           247.494792 |  0.0 | 14.8 | 17.070833 |    1.523023 |         2.074117 |           1.523023 |          1.9 |                16.859515 |                     19.494255 |                         0.0 |                          0.0 |                      1.087886 |                           1.235261 |            16.235713 |                 15.350914 |    1002.764287 |   1.225361 | -2.9 | 15.347388 |   3.284331 |            232.525758 |                      0.0 |                232.525758 |                 188.096042 |                 37.987562 |                     0.0 |                      0.0 |   194.538195 |             0.0 |              0.0 |         150.10848 |             74.333488 |                 100.0 |               109.228155 |              106.410154 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.166779 |  0.037317 | 0.079358 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.060895 | 0.016915 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8694.986928 | 73.382908 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.060895 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.016915 |
    | 1997-08-05 08:00:00 |  0.0 | 17.1 |              1.0 |                      1.0 |      511.104575 |             89.9 |       3.2 |              1019.0 | 0.0 | 0.0 | 17.083333 |             78.870833 |                  12.9 |                     15.470733 |           248.989764 |  0.0 | 17.1 | 17.083333 |    2.565092 |         2.097497 |           2.565092 |          3.2 |                19.530313 |                     19.509701 |                         0.0 |                          0.0 |                      1.237258 |                           1.236116 |            17.557751 |                 15.387464 |    1001.442249 |   1.215051 | -2.9 | 15.273176 | -10.960941 |            388.439477 |                      0.0 |                388.439477 |                 189.232221 |                 37.896217 |                     0.0 |                      0.0 |    350.54326 |             0.0 |              0.0 |        151.336004 |             44.135508 |                 100.0 |                109.34067 |              106.486271 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.358863 |       0.0 | 0.185679 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.059322 | 0.016478 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8703.047869 | 73.197229 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.059322 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.016478 |
    | 1997-08-05 09:00:00 |  0.0 | 19.4 |              0.9 |                      1.0 |      629.519239 |             79.2 |       4.0 |              1019.0 | 0.0 | 0.0 | 17.133333 |             78.883333 |                  12.8 |                     15.470733 |           249.636119 |  0.0 | 19.4 | 17.133333 |    3.206365 |         2.134237 |           3.206365 |          4.0 |                22.563931 |                     19.571593 |                         0.0 |                          0.0 |                      1.403627 |                           1.239544 |            17.870634 |                 15.438725 |    1001.129366 |   1.205358 | -2.9 | 15.369908 | -24.180553 |            478.434622 |                      0.0 |                478.434622 |                 189.723451 |                  37.56462 |                     0.0 |                      0.0 |   440.870002 |             0.0 |              0.0 |        152.158831 |             35.308407 |                 100.0 |               109.604122 |              106.664308 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.541806 |       0.0 |   0.2631 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.05779 | 0.016053 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8724.328423 |  72.93413 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.05779 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.016053 |
    | 1997-08-05 10:00:00 |  0.0 | 20.6 |              1.0 |                      1.0 |      765.816589 |             73.4 |       4.6 |              1019.0 | 0.0 | 0.0 | 17.179167 |             78.783333 |                  12.9 |                     15.470733 |           254.719872 |  0.0 | 20.6 | 17.179167 |     3.68732 |         2.184336 |            3.68732 |          4.6 |                24.304197 |                     19.628477 |                         0.0 |                          0.0 |                      1.497674 |                           1.242694 |            17.839281 |                 15.463969 |    1001.160719 |   1.200448 | -2.9 | 15.625274 | -29.848354 |            582.020608 |                      0.0 |                582.020608 |                 193.587103 |                 37.751868 |                     0.0 |                      0.0 |    544.26874 |             0.0 |              0.0 |        155.835235 |             30.702962 |                 100.0 |               109.977887 |              106.916443 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.716842 |       0.0 |   0.3322 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.056297 | 0.015638 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8751.276776 |  72.60193 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.056297 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.015638 |
    | 1997-08-05 11:00:00 |  0.0 | 21.4 |              1.0 |                      1.0 |      767.145842 |             67.3 |       4.1 |              1019.0 | 0.0 | 0.0 |    17.225 |             78.529167 |                  13.1 |                     15.470733 |           261.673325 |  0.0 | 21.4 |    17.225 |    3.286524 |         2.211056 |           3.286524 |          4.1 |                25.528421 |                     19.685506 |                         0.0 |                          0.0 |                      1.563281 |                            1.24585 |            17.180627 |                 15.458864 |    1001.819373 |   1.197482 | -2.9 | 15.948655 | -32.708072 |             583.03084 |                      0.0 |                 583.03084 |                 198.871727 |                 38.247049 |                     0.0 |                      0.0 |   544.783791 |             0.0 |              0.0 |        160.624678 |             34.447226 |                 100.0 |               110.450603 |              107.234579 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.746089 |       0.0 | 0.369947 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.054843 | 0.015234 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8781.084848 | 72.231982 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.054843 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.015234 |
    | 1997-08-05 12:00:00 |  0.0 | 22.5 |              1.0 |                      1.0 |      736.263853 |             62.2 |       4.1 |              1019.0 | 0.0 | 0.0 |   17.2875 |                78.275 |                  13.2 |                     15.470733 |            266.49623 |  0.0 | 22.5 |   17.2875 |    3.286524 |         2.237776 |           3.286524 |          4.1 |                27.299387 |                     19.763507 |                         0.0 |                          0.0 |                      1.657431 |                           1.250165 |            16.980219 |                 15.469885 |    1002.019781 |   1.193116 | -2.9 | 16.306352 | -37.161891 |            559.560528 |                      0.0 |                559.560528 |                 202.537135 |                  38.48429 |                     0.0 |                      0.0 |   521.076238 |             0.0 |              0.0 |        164.052844 |             34.447226 |                 100.0 |               110.978084 |              107.588582 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.771457 |       0.0 | 0.389289 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.053426 | 0.014841 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8815.346739 | 71.842694 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.053426 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.014841 |
    | 1997-08-05 13:00:00 |  0.0 | 23.1 |              1.0 |                      1.0 |      768.006183 |             59.6 |       4.4 |              1019.0 | 0.0 | 0.0 | 17.341667 |             78.204167 |                  13.4 |                     15.470733 |           273.519763 |  0.0 | 23.1 | 17.341667 |    3.527001 |         2.277855 |           3.527001 |          4.4 |                28.309776 |                     19.831326 |                         0.0 |                          0.0 |                      1.710768 |                           1.253914 |            16.872626 |                 15.508923 |    1002.127374 |   1.190747 | -2.9 | 16.717494 | -38.295035 |            583.684699 |                      0.0 |                583.684699 |                  207.87502 |                  38.86309 |                     0.0 |                      0.0 |   544.821609 |             0.0 |              0.0 |         169.01193 |             32.098552 |                 100.0 |               111.534357 |              107.960785 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.841038 |       0.0 |  0.41464 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.052046 | 0.014457 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8850.741774 | 71.428054 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.052046 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.014457 |
    | 1997-08-05 14:00:00 |  0.0 | 23.4 |              1.0 |                      1.0 |      735.929369 |             55.4 |       4.6 |              1019.0 | 0.0 | 0.0 | 17.404167 |               77.9875 |                  13.5 |                     15.470733 |           278.618852 |  0.0 | 23.4 | 17.404167 |     3.68732 |         2.347994 |            3.68732 |          4.6 |                28.827078 |                     19.909831 |                         0.0 |                          0.0 |                      1.737973 |                           1.258253 |            15.970201 |                 15.527179 |    1003.029799 |   1.189943 | -2.9 | 17.142235 | -37.546592 |            559.306321 |                      0.0 |                559.306321 |                 211.750328 |                 39.079824 |                     0.0 |                      0.0 |   520.226497 |             0.0 |              0.0 |        172.670504 |             30.702962 |                 100.0 |               112.128253 |              108.356894 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.865229 |       0.0 | 0.419328 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.050702 | 0.014084 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8885.388366 | 71.008726 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.050702 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.014084 |
    | 1997-08-05 15:00:00 |  0.0 | 23.8 |              1.0 |                      1.0 |      620.374953 |             52.1 |       4.5 |              1019.0 | 0.0 | 0.0 |   17.4875 |             77.708333 |                  13.6 |                     15.470733 |           282.921682 |  0.0 | 23.8 |   17.4875 |    3.607161 |         2.401434 |           3.607161 |          4.5 |                29.529613 |                     20.014927 |                         0.0 |                          0.0 |                      1.774812 |                           1.264057 |            15.384928 |                 15.553266 |    1003.615072 |     1.1886 | -2.9 | 17.557994 | -37.452038 |            471.484964 |                      0.0 |                471.484964 |                 215.020478 |                 39.288738 |                     0.0 |                      0.0 |   432.196226 |             0.0 |              0.0 |         175.73174 |              31.38525 |                 100.0 |               112.730358 |              108.757148 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.79847 |       0.0 |  0.39307 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.049392 |  0.01372 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8919.940404 | 70.615656 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.049392 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.01372 |
    | 1997-08-05 16:00:00 |  0.0 | 24.0 |              1.0 |                      1.0 |      450.095264 |             49.5 |       4.7 |              1018.0 | 0.0 | 0.0 |   17.5625 |               77.3875 |                  13.7 |                     15.470733 |           285.932459 |  0.0 | 24.0 |   17.5625 |    3.767479 |         2.438173 |           3.767479 |          4.7 |                29.886439 |                     20.109928 |                         0.0 |                          0.0 |                      1.793476 |                             1.2693 |            14.793787 |                  15.56257 |    1003.206213 |   1.186889 | -2.9 | 17.972618 | -36.164291 |            342.072401 |                      0.0 |                342.072401 |                 217.308669 |                 39.539407 |                     0.0 |                      0.0 |   302.532993 |             0.0 |              0.0 |        177.769261 |             30.049708 |                 100.0 |               113.296152 |              109.132049 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.70277 |       0.0 |  0.33964 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.048116 | 0.013366 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8953.204695 | 70.276016 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.048116 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.013366 |
    | 1997-08-05 17:00:00 |  0.0 | 24.0 |              1.0 |                      1.0 |      273.614142 |             48.1 |       4.5 |              1018.0 | 0.0 | 0.0 | 17.654167 |             77.008333 |                  13.7 |                     15.470733 |           286.824521 |  0.0 | 24.0 | 17.654167 |    3.607161 |         2.448193 |           3.607161 |          4.5 |                29.886439 |                     20.226575 |                         0.0 |                          0.0 |                      1.793476 |                           1.275734 |            14.375377 |                 15.576148 |    1003.624623 |   1.187075 | -2.9 |  18.37179 | -33.769262 |            207.946748 |                      0.0 |                207.946748 |                 217.986636 |                 39.564612 |                     0.0 |                      0.0 |   168.382136 |             0.0 |              0.0 |        178.422024 |              31.38525 |                 100.0 |               113.786144 |              109.455777 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.559369 |       0.0 | 0.275457 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.046873 |  0.01302 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8984.073957 | 70.000559 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.046873 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.01302 |
    | 1997-08-05 18:00:00 |  0.0 | 23.7 |              1.0 |                      1.0 |      133.368106 |             49.4 |       4.1 |              1018.0 | 0.0 | 0.0 | 17.770833 |             76.570833 |                  13.7 |                     15.470733 |           287.024163 |  0.0 | 23.7 | 17.770833 |    3.286524 |         2.444853 |           3.286524 |          4.1 |                29.352596 |                      20.37589 |                         0.0 |                          0.0 |                      1.765541 |                           1.283961 |            14.500182 |                 15.601989 |    1003.499818 |   1.188219 | -2.9 | 18.742221 | -29.746675 |             101.35976 |                      0.0 |                 101.35976 |                 218.138364 |                 39.572333 |                     0.0 |                      0.0 |    61.787427 |             0.0 |              0.0 |        178.566031 |             34.447226 |                 100.0 |               114.184306 |              109.718191 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.406457 |       0.0 | 0.207398 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.045662 | 0.012684 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 9010.920632 | 69.793161 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.045662 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.012684 |
    | 1997-08-05 19:00:00 |  0.0 | 22.8 |              1.0 |                      1.0 |       40.326503 |             57.0 |       3.8 |              1018.0 | 0.0 | 0.0 |      17.9 |             76.191667 |                  13.7 |                     15.470733 |           286.951409 |  0.0 | 22.8 |      17.9 |    3.046047 |         2.451533 |           3.046047 |          3.8 |                27.800581 |                     20.542326 |                         0.0 |                          0.0 |                      1.683922 |                           1.293123 |            15.846331 |                 15.651541 |    1002.153669 |   1.191234 | -2.9 | 19.064381 | -22.413715 |             30.648142 |                      0.0 |                 30.648142 |                 218.083071 |                 39.521285 |                     0.0 |                      0.0 |    -8.873143 |             0.0 |              0.0 |        178.561786 |             37.166744 |                 100.0 |               114.484549 |               109.91569 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.257526 |       0.0 | 0.133656 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.044483 | 0.012356 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 9030.434346 | 69.659505 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.044483 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.012356 |
    | 1997-08-05 20:00:00 |  0.0 | 21.3 |              0.5 |                 0.072689 |        0.867122 |             65.9 |       2.5 |              1018.0 | 0.0 | 0.0 | 18.041667 |                 75.95 |                  13.9 |                     15.438712 |           286.953669 |  0.0 | 21.3 | 18.041667 |    2.003978 |         2.428153 |           2.003978 |          2.5 |                 25.37251 |                     20.726234 |                         0.0 |                          0.0 |                       1.55495 |                           1.303234 |            16.720484 |                 15.741575 |    1001.279516 |   1.196911 | -2.9 | 19.298545 | -12.008727 |              0.659013 |                      0.0 |                  0.659013 |                 218.084788 |                 39.876809 |                     0.0 |                      0.0 |   -39.217796 |             0.0 |              0.0 |        178.207979 |             56.493451 |                 100.0 |                114.67825 |               89.032161 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.102206 |       0.0 | 0.066112 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.043334 | 0.012037 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 9039.543073 | 69.593394 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.043334 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.012037 |
    | 1997-08-05 21:00:00 |  0.0 | 19.3 |              0.0 |                      0.0 |             0.0 |             73.0 |       1.9 |              1018.0 | 0.0 | 0.0 |   18.1875 |             75.633333 |                  13.9 |                     15.438712 |           286.953669 |  0.0 | 19.3 |   18.1875 |    1.523023 |         2.418134 |           1.523023 |          1.9 |                22.423949 |                     20.917052 |                         0.0 |                          0.0 |                       1.39602 |                           1.313713 |            16.369483 |                 15.820264 |    1001.630517 |   1.205255 | -2.9 |  19.40785 |   0.647101 |                   0.0 |                      0.0 |                       0.0 |                 218.084788 |                 39.753451 |                     0.0 |                      0.0 |   -39.753451 |             0.0 |              0.0 |        178.331337 |             74.333488 |                 100.0 |               114.774124 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.046254 |       0.0 |  0.03182 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.042215 | 0.011726 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 9035.995972 | 69.561574 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.042215 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.011726 |
    | 1997-08-05 22:00:00 |  0.0 | 18.3 |              0.0 |                      0.0 |             0.0 |             76.7 |       2.2 |              1018.0 | 0.0 | 0.0 | 18.333333 |                 75.25 |                  13.9 |                     15.438712 |           286.953669 |  0.0 | 18.3 | 18.333333 |    1.763501 |         2.418134 |           1.763501 |          2.2 |                21.065302 |                     21.109404 |                         0.0 |                          0.0 |                      1.321845 |                           1.324263 |            16.157086 |                 15.884826 |    1001.842914 |   1.209486 | -2.9 | 19.365285 |    6.39171 |                   0.0 |                      0.0 |                       0.0 |                 218.084788 |                 39.670785 |                     0.0 |                      0.0 |   -39.670785 |             0.0 |              0.0 |        178.414003 |             64.197103 |                 100.0 |               114.820283 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.042508 |       0.0 |  0.02777 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.041124 | 0.011423 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 9026.704262 | 69.533804 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.041124 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.011423 |
    | 1997-08-05 23:00:00 |  0.0 | 17.4 |              0.0 |                      0.0 |             0.0 |             81.8 |       2.7 |              1018.0 | 0.0 | 0.0 | 18.470833 |               74.9375 |                  13.9 |                     15.438712 |           286.953669 |  0.0 | 17.4 | 18.470833 |    2.164296 |         2.438173 |           2.164296 |          2.7 |                19.904589 |                     21.292177 |                         0.0 |                          0.0 |                      1.257963 |                           1.334275 |            16.281953 |                 15.955825 |    1001.718047 |   1.213176 | -2.9 | 19.253784 |  11.122707 |                   0.0 |                      0.0 |                       0.0 |                 218.084788 |                 39.564038 |                     0.0 |                      0.0 |   -39.564038 |             0.0 |              0.0 |        178.520751 |             52.308751 |                 100.0 |               114.860575 |               87.719298 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.035577 |       0.0 | 0.021514 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.040062 | 0.011128 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 9012.681555 |  69.51229 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.040062 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.011128 |

>>> round_(model.check_waterbalance(conditions_acker_summer))
0.0

.. _lland_v4_acker_heavy_rain_hourly:

acre (heavy rain)
-----------------

.. integration-test::

    >>> inputs.nied.series = 20.0 / 24.0
    >>> control.negq(False)
    >>> test("lland_v4_acker_heavy_rain_hourly",
    ...      axis1=(inputs.nied, fluxes.qah), axis2=states.bowa,
    ...      use_conditions=conditions_acker_heavy_rain)
    |                date |     nied | teml | sunshineduration | possiblesunshineduration | globalradiation | relativehumidity | windspeed | atmosphericpressure |  qz | qzh |   temltag | dailyrelativehumidity | dailysunshineduration | dailypossiblesunshineduration | dailyglobalradiation |     nkor | tkor |   tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair |    g |        tz |         wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |     nbes | sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |      evpo |       evi |      evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatinz | wlatsnow | wsensinz | wsenssnow | wsurfinz | wsurf | sff | fvg |     wada |      qdb |     qib1 | qib2 |      qbb | qkap |     qdgz |      qah |       qa |     inzp | stinz | sinz | esnowinz | asinz | wats | waes | esnow | taus |        ebdn |       bowa |    qdgz1 | qdgz2 |    qigz1 | qigz2 |     qbgz |    qdga1 | qdga2 |    qiga1 | qiga2 |     qbga | inlet |   outlet |
    ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-08-03 00:00:00 | 0.833333 | 16.9 |              0.0 |                      0.0 |             0.0 |             95.1 |       0.8 |              1015.0 | 0.0 | 0.0 | 17.279167 |             85.883333 |                   1.3 |                          15.7 |           136.579167 | 0.783333 | 16.9 | 17.279167 |    0.641273 |         1.605886 |           0.641273 |          0.8 |                19.284227 |                     19.753091 |                         0.0 |                          0.0 |                      1.223615 |                           1.249589 |              18.3393 |                 16.964613 |       996.6607 |    1.21073 | -2.9 | 15.632162 |  -7.607027 |                   0.0 |                      0.0 |                       0.0 |                 103.800167 |                 10.408237 |                     0.0 |                      0.0 |   -10.408237 |             0.0 |              0.0 |          93.39193 |            176.542034 |                 100.0 |                70.326844 |               87.719298 | 0.083333 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.002953 | -0.002953 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.083333 | 0.007403 |      0.0 |  0.0 |      0.0 |  0.0 | 0.007403 |  0.80139 | 0.222608 | 0.702953 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8729.609427 | 102.562669 | 0.007403 |   0.0 |      0.0 |   0.0 |      0.0 |  0.80139 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.222608 |
    | 1997-08-03 01:00:00 | 0.833333 | 16.6 |              0.0 |                      0.0 |             0.0 |             94.9 |       0.8 |              1015.0 | 0.0 | 0.0 | 17.241667 |             85.991667 |                   1.3 |                          15.7 |           136.579167 | 0.783333 | 16.6 | 17.241667 |    0.641273 |         1.603439 |           0.641273 |          0.8 |                18.920184 |                      19.70628 |                         0.0 |                          0.0 |                       1.20339 |                           1.246999 |            17.955254 |                 16.945759 |     997.044746 |   1.212158 | -2.9 | 15.688646 |  -5.468121 |                   0.0 |                      0.0 |                       0.0 |                 103.800167 |                  10.41572 |                     0.0 |                      0.0 |    -10.41572 |             0.0 |              0.0 |         93.384446 |            176.542034 |                 100.0 |                70.229104 |               87.719298 | 0.786286 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.002673 | -0.002673 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.786286 | 0.070171 |      0.0 |  0.0 |      0.0 |  0.0 | 0.070171 | 0.781695 | 0.217138 | 0.702673 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8732.177548 | 103.278784 | 0.070171 |   0.0 |      0.0 |   0.0 |      0.0 | 0.781695 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.217138 |
    | 1997-08-03 02:00:00 | 0.833333 | 16.4 |              0.0 |                      0.0 |             0.0 |             95.9 |       0.8 |              1015.0 | 0.0 | 0.0 |    17.175 |             86.233333 |                   1.3 |                          15.7 |           136.579167 | 0.783333 | 16.4 |    17.175 |    0.641273 |         1.575992 |           0.641273 |          0.8 |                 18.68084 |                       19.6233 |                         0.0 |                          0.0 |                      1.190065 |                           1.242407 |            17.914926 |                 16.921826 |     997.085074 |   1.213014 | -2.9 | 15.719464 |  -4.083217 |                   0.0 |                      0.0 |                       0.0 |                 103.800167 |                 10.421627 |                     0.0 |                      0.0 |   -10.421627 |             0.0 |              0.0 |         93.378539 |            176.542034 |                 100.0 |                69.307997 |               87.719298 | 0.786007 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.004096 | -0.004096 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.786007 | 0.070721 |      0.0 |  0.0 |      0.0 |  0.0 | 0.070721 | 0.763323 | 0.212034 | 0.704096 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8733.360765 |  103.99407 | 0.070721 |   0.0 |      0.0 |   0.0 |      0.0 | 0.763323 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.212034 |
    | 1997-08-03 03:00:00 | 0.833333 | 16.3 |              0.0 |                      0.0 |             0.0 |             96.7 |       0.8 |              1015.0 | 0.0 | 0.0 |   17.0625 |             86.708333 |                   1.3 |                          15.5 |           136.579167 | 0.783333 | 16.3 |   17.0625 |    0.641273 |         1.548545 |           0.641273 |          0.8 |                18.562165 |                     19.483964 |                         0.0 |                          0.0 |                      1.183449 |                            1.23469 |            17.949613 |                  16.89422 |     997.050387 |   1.213417 | -2.9 | 15.733663 |  -3.398025 |                   0.0 |                      0.0 |                       0.0 |                 103.800167 |                 10.455184 |                     0.0 |                      0.0 |   -10.455184 |             0.0 |              0.0 |         93.344983 |            176.542034 |                 100.0 |                68.389204 |               87.719298 | 0.787429 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.005245 | -0.005245 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.787429 | 0.071426 |      0.0 |  0.0 |      0.0 |  0.0 | 0.071426 | 0.745442 | 0.207067 | 0.705245 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  8733.85879 | 104.710073 | 0.071426 |   0.0 |      0.0 |   0.0 |      0.0 | 0.745442 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.207067 |
    | 1997-08-03 04:00:00 | 0.833333 | 16.0 |              0.0 |                 0.429734 |        1.943686 |             97.2 |       0.8 |              1015.0 | 0.0 | 0.0 | 16.908333 |             87.366667 |                   1.3 |                     14.929734 |            135.50182 | 0.783333 | 16.0 | 16.908333 |    0.641273 |         1.504432 |           0.641273 |          0.8 |                18.210086 |                     19.294427 |                         0.0 |                          0.0 |                      1.163788 |                           1.224181 |            17.700204 |                 16.856897 |     997.299796 |    1.21479 | -2.9 | 15.739639 |  -1.562167 |              1.477201 |                      0.0 |                  1.477201 |                 102.981383 |                 10.555242 |                     0.0 |                      0.0 |    -9.078041 |             0.0 |              0.0 |         92.426141 |            176.542034 |                 100.0 |                67.470711 |               81.462042 | 0.788579 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.004905 | -0.004905 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.788579 |  0.07211 |      0.0 |  0.0 |      0.0 |  0.0 |  0.07211 |  0.72804 | 0.202233 | 0.704905 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8732.520957 | 105.426542 |  0.07211 |   0.0 |      0.0 |   0.0 |      0.0 |  0.72804 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.202233 |
    | 1997-08-03 05:00:00 | 0.833333 | 15.9 |              0.0 |                      1.0 |       21.932442 |             97.5 |       0.6 |              1015.0 | 0.0 | 0.0 | 16.729167 |             88.204167 |                   1.3 |                     14.929734 |           134.099005 | 0.783333 | 15.9 | 16.729167 |    0.480955 |         1.453638 |           0.480955 |          0.6 |                18.094032 |                     19.076181 |                         0.0 |                          0.0 |                      1.157296 |                           1.212063 |            17.641681 |                 16.825987 |     997.358319 |   1.215237 | -2.9 | 15.723585 |  -1.058491 |             16.668656 |                      0.0 |                 16.668656 |                 101.915244 |                 10.545193 |                     0.0 |                      0.0 |     6.123463 |             0.0 |              0.0 |         91.370051 |            235.389378 |                 100.0 |                66.552817 |               73.619139 | 0.788239 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.004992 |  0.004992 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.788239 |  0.07266 |      0.0 |  0.0 |      0.0 |  0.0 |  0.07266 | 0.711104 | 0.197529 | 0.695008 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8730.679448 |  106.14212 |  0.07266 |   0.0 |      0.0 |   0.0 |      0.0 | 0.711104 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.197529 |
    | 1997-08-03 06:00:00 | 0.833333 | 16.0 |              0.0 |                      1.0 |       57.256186 |             97.7 |       0.9 |              1015.0 | 0.0 | 0.0 | 16.533333 |             89.191667 |                   1.3 |                     14.929734 |           131.855513 | 0.783333 | 16.0 | 16.533333 |    0.721432 |         1.392031 |           0.721432 |          0.9 |                18.210086 |                     18.840106 |                         0.0 |                          0.0 |                      1.163788 |                           1.198935 |            17.791254 |                 16.803805 |     997.208746 |   1.214748 | -2.9 | 15.701487 |   -1.79108 |             43.514701 |                      0.0 |                 43.514701 |                  100.21019 |                  10.52516 |                     0.0 |                      0.0 |    32.989541 |             0.0 |              0.0 |          89.68503 |            156.926252 |                 100.0 |                65.637228 |               72.817981 | 0.778341 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.026203 |  0.026203 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.778341 | 0.072319 |      0.0 |  0.0 |      0.0 |  0.0 | 0.072319 | 0.694608 | 0.192947 | 0.673797 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8729.570528 | 106.848143 | 0.072319 |   0.0 |      0.0 |   0.0 |      0.0 | 0.694608 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.192947 |
    | 1997-08-03 07:00:00 | 0.833333 | 16.6 |              0.0 |                      1.0 |      109.332844 |             97.4 |       0.9 |              1015.0 | 0.0 | 0.0 |    16.375 |                90.125 |                   1.3 |                     14.929734 |           129.465215 | 0.783333 | 16.6 |    16.375 |    0.721432 |         1.334591 |           0.721432 |          0.9 |                18.920184 |                     18.651109 |                         0.0 |                          0.0 |                       1.20339 |                           1.188408 |            18.428259 |                 16.809312 |     996.571741 |   1.211943 | -2.9 |  15.68818 |  -5.470922 |             83.092962 |                      0.0 |                 83.092962 |                  98.393563 |                  10.49073 |                     0.0 |                      0.0 |    72.602232 |             0.0 |              0.0 |         87.902834 |            156.926252 |                 100.0 |                64.734984 |               72.023917 |  0.75713 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.057366 |  0.057366 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.75713 | 0.070894 |      0.0 |  0.0 |      0.0 |  0.0 | 0.070894 | 0.678515 | 0.188476 | 0.642634 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  8732.14145 |  107.53438 | 0.070894 |   0.0 |      0.0 |   0.0 |      0.0 | 0.678515 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.188476 |
    | 1997-08-03 08:00:00 | 0.833333 | 17.4 |              0.0 |                      1.0 |      170.949153 |             96.8 |       0.9 |              1016.0 | 0.0 | 0.0 | 16.216667 |                91.275 |                   1.3 |                     14.929734 |           127.329763 | 0.783333 | 17.4 | 16.216667 |    0.721432 |         1.260484 |           0.721432 |          0.9 |                19.904589 |                     18.463773 |                         0.0 |                          0.0 |                      1.257963 |                           1.177959 |            19.267642 |                 16.852809 |     996.732358 |   1.209425 | -2.9 | 15.719031 | -10.085816 |            129.921356 |                      0.0 |                129.921356 |                   96.77062 |                 10.426918 |                     0.0 |                      0.0 |   119.494439 |             0.0 |              0.0 |         86.343702 |            156.926252 |                 100.0 |                63.859065 |               71.248633 | 0.725967 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.096008 |  0.096008 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.725967 | 0.068482 |      0.0 |  0.0 |      0.0 |  0.0 | 0.068482 | 0.662789 | 0.184108 | 0.603992 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8739.327265 | 108.191865 | 0.068482 |   0.0 |      0.0 |   0.0 |      0.0 | 0.662789 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.184108 |
    | 1997-08-03 09:00:00 | 0.833333 | 19.0 |              0.2 |                      1.0 |      311.762625 |             86.1 |       1.3 |              1016.0 | 0.0 | 0.0 |   16.1125 |             91.991667 |                   1.5 |                     14.929734 |           129.903206 | 0.783333 | 19.0 |   16.1125 |    1.042069 |         1.203904 |           1.042069 |          1.3 |                22.008543 |                     18.341425 |                         0.0 |                          0.0 |                      1.373407 |                           1.171128 |            18.949356 |                 16.872582 |     997.050644 |   1.202945 | -2.9 | 15.805261 | -19.168437 |            236.939595 |                      0.0 |                236.939595 |                  98.726436 |                 10.804921 |                     0.0 |                      0.0 |   226.134674 |             0.0 |              0.0 |         87.921515 |            108.641252 |                 100.0 |                63.020785 |               70.502581 | 0.687326 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.222649 |  0.222649 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.687326 | 0.065294 |      0.0 |  0.0 |      0.0 |  0.0 | 0.065294 | 0.647396 | 0.179832 | 0.477351 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8755.595702 | 108.813897 | 0.065294 |   0.0 |      0.0 |   0.0 |      0.0 | 0.647396 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.179832 |
    | 1997-08-03 10:00:00 | 0.833333 | 20.3 |              0.5 |                      1.0 |        501.5833 |             76.8 |       1.5 |              1016.0 | 0.0 | 0.0 | 16.083333 |             92.241667 |                   1.7 |                     14.929734 |           132.285843 | 0.783333 | 20.3 | 16.083333 |    1.202387 |         1.149836 |           1.202387 |          1.5 |                23.858503 |                     18.307295 |                         0.0 |                          0.0 |                      1.473678 |                           1.169221 |             18.32333 |                 16.886954 |      997.67667 |   1.197896 | -2.9 | 16.000482 | -25.797109 |            381.203308 |                      0.0 |                381.203308 |                 100.537241 |                 11.199993 |                     0.0 |                      0.0 |   370.003315 |             0.0 |              0.0 |         89.337248 |             94.155751 |                 100.0 |                62.228541 |               69.793799 | 0.560684 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.389789 |  0.389789 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.560684 | 0.053594 |      0.0 |  0.0 |      0.0 |  0.0 | 0.053594 | 0.632208 | 0.175613 | 0.310211 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8778.492812 | 109.320987 | 0.053594 |   0.0 |      0.0 |   0.0 |      0.0 | 0.632208 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.175613 |
    | 1997-08-03 11:00:00 | 0.833333 | 21.4 |              0.7 |                      1.0 |      615.018728 |             71.8 |       1.2 |              1016.0 | 0.0 | 0.0 |    16.125 |             92.104167 |                   2.2 |                     14.929734 |           135.919957 | 0.783333 | 21.4 |    16.125 |    0.961909 |         1.089916 |           0.961909 |          1.2 |                25.528421 |                     18.356069 |                         0.0 |                          0.0 |                      1.563281 |                           1.171946 |            18.329406 |                 16.906704 |     997.670594 |    1.19342 | -2.9 | 16.275247 | -30.748518 |            467.414233 |                      0.0 |                467.414233 |                 103.299167 |                 12.222123 |                     0.0 |                      0.0 |    455.19211 |             0.0 |              0.0 |         91.077044 |            117.694689 |                 100.0 |                61.583277 |               69.213837 | 0.393544 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.46611 |   0.46611 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.393544 | 0.037795 |      0.0 |  0.0 |      0.0 |  0.0 | 0.037795 | 0.617057 | 0.171405 |  0.23389 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8806.341329 | 109.676737 | 0.037795 |   0.0 |      0.0 |   0.0 |      0.0 | 0.617057 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.171405 |
    | 1997-08-03 12:00:00 | 0.833333 | 21.3 |              0.8 |                      1.0 |      626.544325 |             67.5 |       1.3 |              1016.0 | 0.0 | 0.0 | 16.204167 |             91.729167 |                   2.8 |                     14.929734 |           145.821804 | 0.783333 | 21.3 | 16.204167 |    1.042069 |         1.037502 |           1.042069 |          1.3 |                 25.37251 |                     18.449053 |                         0.0 |                          0.0 |                       1.55495 |                           1.177138 |            17.126444 |                 16.923163 |     998.873556 |   1.194363 | -2.9 | 16.609429 | -28.143424 |            476.173687 |                      0.0 |                476.173687 |                 110.824571 |                 13.461134 |                     0.0 |                      0.0 |   462.712553 |             0.0 |              0.0 |         97.363437 |            108.641252 |                 100.0 |                61.130903 |               68.805802 | 0.317224 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.490849 |  0.490849 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.317224 | 0.030572 |      0.0 |  0.0 |      0.0 |  0.0 | 0.030572 |    0.602 | 0.167222 | 0.209151 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8831.584754 | 109.963389 | 0.030572 |   0.0 |      0.0 |   0.0 |      0.0 |    0.602 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.167222 |
    | 1997-08-03 13:00:00 | 0.833333 | 21.8 |              0.5 |                      1.0 |      496.133417 |             66.1 |       1.5 |              1016.0 | 0.0 | 0.0 | 16.329167 |             91.104167 |                   3.2 |                     14.929734 |           152.606529 | 0.783333 | 21.8 | 16.329167 |    1.202387 |         1.012602 |           1.202387 |          1.5 |                26.160453 |                      18.59671 |                         0.0 |                          0.0 |                      1.596982 |                           1.185375 |             17.29206 |                 16.942378 |      998.70794 |   1.192265 | -2.9 |  16.91235 | -29.325898 |            377.061397 |                      0.0 |                377.061397 |                 115.980962 |                 14.297428 |                     0.0 |                      0.0 |   362.763969 |             0.0 |              0.0 |        101.683534 |             94.155751 |                 100.0 |                60.766577 |               68.476317 | 0.292485 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.424547 |  0.424547 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.292485 | 0.028272 |      0.0 |  0.0 |      0.0 |  0.0 | 0.028272 |  0.58721 | 0.163114 | 0.275453 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8858.010651 | 110.227602 | 0.028272 |   0.0 |      0.0 |   0.0 |      0.0 |  0.58721 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.163114 |
    | 1997-08-03 14:00:00 | 0.833333 | 22.9 |              0.4 |                      1.0 |      419.520994 |             63.4 |       1.9 |              1016.0 | 0.0 | 0.0 | 16.545833 |             90.058333 |                   3.4 |                     14.929734 |           157.353237 | 0.783333 | 22.9 | 16.545833 |    1.523023 |         0.984394 |           1.523023 |          1.9 |                27.969419 |                     18.855098 |                         0.0 |                          0.0 |                      1.692831 |                           1.199769 |            17.732611 |                 16.980587 |     998.267389 |   1.187639 | -2.9 | 17.229461 | -34.023233 |            318.835956 |                      0.0 |                318.835956 |                  119.58846 |                 14.726001 |                     0.0 |                      0.0 |   304.109955 |             0.0 |              0.0 |         104.86246 |             74.333488 |                 100.0 |                60.430915 |               68.172068 | 0.358786 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.415176 |  0.415176 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.358786 | 0.034791 |      0.0 |  0.0 |      0.0 |  0.0 | 0.034791 | 0.572857 | 0.159127 | 0.284824 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8889.133884 | 110.551596 | 0.034791 |   0.0 |      0.0 |   0.0 |      0.0 | 0.572857 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.159127 |
    | 1997-08-03 15:00:00 | 0.833333 | 22.7 |              0.5 |                      1.0 |      387.887353 |             62.4 |       1.9 |              1016.0 | 0.0 | 0.0 | 16.816667 |             88.816667 |                   3.8 |                     14.929734 |           164.256877 | 0.783333 | 22.7 | 16.816667 |    1.523023 |         0.968687 |           1.523023 |          1.9 |                27.632633 |                     19.182495 |                         0.0 |                          0.0 |                      1.675052 |                           1.217969 |            17.242763 |                 17.037252 |     998.757237 |    1.18866 | -2.9 |  17.60294 |  -30.58236 |            294.794388 |                      0.0 |                294.794388 |                 124.835227 |                 15.562805 |                     0.0 |                      0.0 |   279.231583 |             0.0 |              0.0 |        109.272421 |             74.333488 |                 100.0 |                60.019492 |               67.798243 | 0.368157 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.393194 |  0.393194 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.368157 | 0.035827 |      0.0 |  0.0 |      0.0 |  0.0 | 0.035827 | 0.558971 |  0.15527 | 0.306806 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8916.816245 | 110.883927 | 0.035827 |   0.0 |      0.0 |   0.0 |      0.0 | 0.558971 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 |  0.15527 |
    | 1997-08-03 16:00:00 | 0.833333 | 22.5 |              0.5 |                      1.0 |      278.496872 |             61.1 |       2.3 |              1016.0 | 0.0 | 0.0 |   17.1375 |             87.333333 |                   4.1 |                     14.929734 |           170.073414 | 0.783333 | 22.5 |   17.1375 |     1.84366 |         0.991339 |            1.84366 |          2.3 |                27.299387 |                     19.576758 |                         0.0 |                          0.0 |                      1.657431 |                            1.23983 |            16.679926 |                 17.097035 |     999.320074 |   1.189715 | -2.9 | 17.935128 |  -27.38923 |            211.657623 |                      0.0 |                211.657623 |                 129.255794 |                 16.204965 |                     0.0 |                      0.0 |   195.452657 |             0.0 |              0.0 |        113.050829 |             61.405925 |                 100.0 |                59.597696 |               67.413957 | 0.390139 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.34217 |   0.34217 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.390139 | 0.038106 |      0.0 |  0.0 |      0.0 |  0.0 | 0.038106 | 0.545488 | 0.151524 |  0.35783 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8941.305475 | 111.235959 | 0.038106 |   0.0 |      0.0 |   0.0 |      0.0 | 0.545488 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.151524 |
    | 1997-08-03 17:00:00 | 0.833333 | 21.9 |              0.3 |                      1.0 |      137.138608 |             62.1 |       2.4 |              1016.0 | 0.0 | 0.0 |   17.4875 |                85.875 |                   4.4 |                     14.929734 |           173.470856 | 0.783333 | 21.9 |   17.4875 |    1.923819 |         1.017332 |           1.923819 |          2.4 |                26.320577 |                     20.014927 |                         0.0 |                          0.0 |                      1.605502 |                           1.264057 |            16.345078 |                 17.187818 |     999.654922 |   1.192283 | -2.9 | 18.228999 | -22.026006 |            104.225342 |                      0.0 |                104.225342 |                  131.83785 |                 16.818923 |                     0.0 |                      0.0 |     87.40642 |             0.0 |              0.0 |        115.018928 |             58.847345 |                 100.0 |                59.151125 |               67.005952 | 0.441163 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.236219 |  0.236219 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.441163 | 0.043264 |      0.0 |  0.0 |      0.0 |  0.0 | 0.043264 | 0.532448 | 0.147902 | 0.463781 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8960.431481 | 111.633858 | 0.043264 |   0.0 |      0.0 |   0.0 |      0.0 | 0.532448 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.147902 |
    | 1997-08-03 18:00:00 | 0.833333 | 21.4 |              0.1 |                      1.0 |       51.080714 |             67.0 |       2.5 |              1016.0 | 0.0 | 0.0 |   17.8375 |             84.620833 |                   4.5 |                     14.929734 |           174.440885 | 0.783333 | 21.4 |   17.8375 |    2.003978 |         1.054998 |           2.003978 |          2.5 |                25.528421 |                     20.461645 |                         0.0 |                          0.0 |                      1.563281 |                           1.288683 |            17.104042 |                 17.314814 |     998.895958 |   1.193968 | -2.9 | 18.458511 | -17.648933 |             38.821343 |                      0.0 |                 38.821343 |                 132.575073 |                 16.973018 |                     0.0 |                      0.0 |    21.848325 |             0.0 |              0.0 |        115.602055 |             56.493451 |                 100.0 |                58.646654 |               66.543619 | 0.547114 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.156077 |  0.156077 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.547114 | 0.053912 |      0.0 |  0.0 |      0.0 |  0.0 | 0.053912 | 0.519951 | 0.144431 | 0.543923 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8975.180414 | 112.127061 | 0.053912 |   0.0 |      0.0 |   0.0 |      0.0 | 0.519951 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.144431 |
    | 1997-08-03 19:00:00 | 0.833333 | 20.7 |              0.0 |                      1.0 |       13.632817 |             74.5 |       2.5 |              1016.0 | 0.0 | 0.0 | 18.170833 |                  83.7 |                   4.5 |                     15.429734 |           175.008919 | 0.783333 | 20.7 | 18.170833 |    2.003978 |         1.100997 |           2.003978 |          2.5 |                24.454368 |                     20.895167 |                         0.0 |                          0.0 |                      1.505746 |                           1.312512 |            18.218504 |                 17.489255 |     997.781496 |   1.196313 | -2.9 | 18.635498 |  -12.38701 |             10.360941 |                      0.0 |                 10.360941 |                 133.006779 |                 16.556776 |                     0.0 |                      0.0 |    -6.195835 |             0.0 |              0.0 |        116.450003 |             56.493451 |                 100.0 |                58.021768 |               65.968819 | 0.627256 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.09611 |   0.09611 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.627256 |  0.06216 |      0.0 |  0.0 |      0.0 |  0.0 |  0.06216 | 0.508019 | 0.141116 |  0.60389 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8984.667424 | 112.692157 |  0.06216 |   0.0 |      0.0 |   0.0 |      0.0 | 0.508019 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.141116 |
    | 1997-08-03 20:00:00 | 0.833333 | 19.4 |              0.0 |                   0.1364 |        0.185944 |             81.2 |       2.2 |              1016.0 | 0.0 | 0.0 | 18.454167 |             83.066667 |                   4.5 |                     15.566134 |           175.016667 | 0.783333 | 19.4 | 18.454167 |    1.763501 |         1.120309 |           1.763501 |          2.2 |                22.563931 |                      21.26995 |                         0.0 |                          0.0 |                      1.403627 |                           1.333058 |            18.321912 |                 17.668238 |     997.678088 |   1.201582 | -2.9 | 18.749342 |  -3.903945 |              0.141318 |                      0.0 |                  0.141318 |                 133.012667 |                 16.343497 |                     0.0 |                      0.0 |    -16.20218 |             0.0 |              0.0 |        116.669169 |             64.197103 |                 100.0 |                57.306347 |               83.796936 | 0.687223 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.049374 |  0.049374 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.687223 | 0.068533 |      0.0 |  0.0 |      0.0 |  0.0 | 0.068533 | 0.496585 |  0.13794 | 0.650626 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  8985.67137 | 113.310848 | 0.068533 |   0.0 |      0.0 |   0.0 |      0.0 | 0.496585 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 |  0.13794 |
    | 1997-08-03 21:00:00 | 0.833333 | 17.8 |              0.0 |                      0.0 |             0.0 |             86.9 |       1.7 |              1016.0 | 0.0 | 0.0 |     18.65 |               82.7375 |                   4.5 |                     15.566134 |           175.016667 | 0.783333 | 17.8 |     18.65 |    1.362705 |         1.127089 |           1.362705 |          1.7 |                20.413369 |                     21.532411 |                         0.0 |                          0.0 |                      1.286025 |                           1.347418 |            17.739217 |                 17.815378 |     998.260783 |   1.208454 | -2.9 |  18.76139 |   5.768339 |                   0.0 |                      0.0 |                       0.0 |                 133.012667 |                 16.221809 |                     0.0 |                      0.0 |   -16.221809 |             0.0 |              0.0 |        116.790858 |             83.078604 |                 100.0 |                56.523742 |               87.719298 | 0.733959 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.019253 |  0.019253 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.733959 | 0.073692 |      0.0 |  0.0 |      0.0 |  0.0 | 0.073692 | 0.485595 | 0.134888 | 0.680747 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8977.003031 | 113.971116 | 0.073692 |   0.0 |      0.0 |   0.0 |      0.0 | 0.485595 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.134888 |
    | 1997-08-03 22:00:00 | 0.833333 | 17.0 |              0.0 |                      0.0 |             0.0 |             90.1 |       1.7 |              1017.0 | 0.0 | 0.0 | 18.808333 |             82.554167 |                   4.5 |                     15.566134 |           175.016667 | 0.783333 | 17.0 | 18.808333 |    1.362705 |         1.142201 |           1.362705 |          1.7 |                19.406929 |                     21.746678 |                         0.0 |                          0.0 |                      1.230421 |                           1.359123 |            17.485643 |                 17.952788 |     999.514357 |   1.213101 | -2.9 |  18.65737 |   9.944218 |                   0.0 |                      0.0 |                       0.0 |                 133.012667 |                 16.101364 |                     0.0 |                      0.0 |   -16.101364 |             0.0 |              0.0 |        116.911302 |             83.078604 |                 100.0 |                55.689302 |               87.719298 | 0.764081 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.010292 |  0.010292 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.764081 | 0.077264 |      0.0 |  0.0 |      0.0 |  0.0 | 0.077264 | 0.475002 | 0.131945 | 0.689708 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8964.158813 | 114.657933 | 0.077264 |   0.0 |      0.0 |   0.0 |      0.0 | 0.475002 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.131945 |
    | 1997-08-03 23:00:00 | 0.833333 | 16.4 |              0.0 |                      0.0 |             0.0 |             90.9 |       2.3 |              1017.0 | 0.0 | 0.0 | 18.941667 |             82.379167 |                   4.5 |                     15.566134 |           175.016667 | 0.783333 | 16.4 | 18.941667 |     1.84366 |         1.185687 |            1.84366 |          2.3 |                 18.68084 |                     21.928555 |                         0.0 |                          0.0 |                      1.190065 |                           1.369047 |            16.980884 |                 18.064561 |    1000.019116 |   1.215845 | -2.9 | 18.503239 |  12.619435 |                   0.0 |                      0.0 |                       0.0 |                 133.012667 |                 16.005095 |                     0.0 |                      0.0 |   -16.005095 |             0.0 |              0.0 |        117.007571 |             61.405925 |                 100.0 |                54.822121 |               87.719298 | 0.773041 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.014655 |  0.014655 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.773041 | 0.078739 |      0.0 |  0.0 |      0.0 |  0.0 | 0.078739 | 0.464747 | 0.129096 | 0.685345 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8948.639378 | 115.352235 | 0.078739 |   0.0 |      0.0 |   0.0 |      0.0 | 0.464747 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.129096 |
    | 1997-08-04 00:00:00 | 0.833333 | 15.9 |              0.0 |                      0.0 |             0.0 |             88.0 |       2.3 |              1017.0 | 0.0 | 0.0 |      18.9 |             82.083333 |                   4.5 |                     15.566134 |           175.016667 | 0.783333 | 15.9 |      18.9 |     1.84366 |         1.235786 |            1.84366 |          2.3 |                18.094032 |                     21.871576 |                         0.0 |                          0.0 |                      1.157296 |                           1.365939 |            15.922748 |                 17.952919 |    1001.077252 |    1.21843 | -2.9 | 18.317006 |  14.502035 |                   0.0 |                      0.0 |                       0.0 |                 133.012667 |                 16.128293 |                     0.0 |                      0.0 |   -16.128293 |             0.0 |              0.0 |        116.884374 |             61.405925 |                 100.0 |                53.946317 |               87.719298 | 0.768678 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.024014 |  0.024014 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.768678 | 0.078864 |      0.0 |  0.0 |      0.0 |  0.0 | 0.078864 | 0.454777 | 0.126327 | 0.675986 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8931.237343 | 116.042049 | 0.078864 |   0.0 |      0.0 |   0.0 |      0.0 | 0.454777 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.126327 |
    | 1997-08-04 01:00:00 | 0.833333 | 15.4 |              0.0 |                      0.0 |             0.0 |             89.2 |       2.2 |              1016.0 | 0.0 | 0.0 |     18.85 |             81.845833 |                   4.5 |                     15.566134 |           175.016667 | 0.783333 | 15.4 |     18.85 |    1.763501 |         1.282546 |           1.763501 |          2.2 |                17.523416 |                     21.803372 |                         0.0 |                          0.0 |                      1.125295 |                           1.362218 |            15.630887 |                 17.845152 |    1000.369113 |   1.219467 | -2.9 | 18.108181 |  16.249089 |                   0.0 |                      0.0 |                       0.0 |                 133.012667 |                 16.244859 |                     0.0 |                      0.0 |   -16.244859 |             0.0 |              0.0 |        116.767808 |             64.197103 |                 100.0 |                 53.07698 |               87.719298 | 0.759319 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.018207 |  0.018207 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.759319 | 0.078462 |      0.0 |  0.0 |      0.0 |  0.0 | 0.078462 | 0.445062 | 0.123628 | 0.681793 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8912.088255 | 116.722906 | 0.078462 |   0.0 |      0.0 |   0.0 |      0.0 | 0.445062 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.123628 |
    | 1997-08-04 02:00:00 | 0.833333 | 15.2 |              0.0 |                      0.0 |             0.0 |             88.8 |       2.3 |              1016.0 | 0.0 | 0.0 |      18.8 |                 81.55 |                   4.5 |                     15.566134 |           175.016667 | 0.783333 | 15.2 |      18.8 |     1.84366 |         1.332645 |            1.84366 |          2.3 |                17.299618 |                     21.735354 |                         0.0 |                          0.0 |                      1.112706 |                           1.358505 |            15.362061 |                 17.725181 |    1000.637939 |   1.220436 | -2.9 | 17.878392 |  16.070354 |                   0.0 |                      0.0 |                       0.0 |                 133.012667 |                 16.376926 |                     0.0 |                      0.0 |   -16.376926 |             0.0 |              0.0 |        116.635741 |             61.405925 |                 100.0 |                52.219703 |               87.719298 | 0.765127 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.020574 |  0.020574 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.765127 | 0.079624 |      0.0 |  0.0 |      0.0 |  0.0 | 0.079624 | 0.435608 | 0.121002 | 0.679426 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |   8893.1179 | 117.408409 | 0.079624 |   0.0 |      0.0 |   0.0 |      0.0 | 0.435608 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.121002 |
    | 1997-08-04 03:00:00 | 0.833333 | 14.9 |              0.0 |                      0.0 |             0.0 |             89.7 |       2.0 |              1016.0 | 0.0 | 0.0 | 18.741667 |             81.258333 |                   4.5 |                     15.566134 |           175.016667 | 0.783333 | 14.9 | 18.741667 |    1.603182 |         1.372725 |           1.603182 |          2.0 |                16.968612 |                     21.656234 |                         0.0 |                          0.0 |                      1.094046 |                           1.354184 |            15.220845 |                 17.597495 |    1000.779155 |   1.221771 | -2.9 | 17.650748 |  16.504489 |                   0.0 |                      0.0 |                       0.0 |                 133.012667 |                 16.516685 |                     0.0 |                      0.0 |   -16.516685 |             0.0 |              0.0 |        116.495982 |             70.616813 |                 100.0 |                 51.35734 |               87.719298 | 0.762759 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.013655 |  0.013655 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.762759 | 0.079941 |      0.0 |  0.0 |      0.0 |  0.0 | 0.079941 | 0.426416 | 0.118449 | 0.686345 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8873.713411 | 118.091227 | 0.079941 |   0.0 |      0.0 |   0.0 |      0.0 | 0.426416 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.118449 |
    | 1997-08-04 04:00:00 | 0.833333 | 14.3 |              0.0 |                 0.398044 |        1.590517 |             93.0 |       2.3 |              1016.0 | 0.0 | 0.0 | 18.670833 |             81.083333 |                   4.5 |                     15.534444 |           175.001952 | 0.783333 | 14.3 | 18.670833 |     1.84366 |         1.422824 |            1.84366 |          2.3 |                16.323194 |                     21.560498 |                         0.0 |                          0.0 |                      1.057522 |                           1.348953 |             15.18057 |                  17.48197 |     1000.81943 |    1.22434 | -2.9 | 17.417894 |  18.707366 |              1.208793 |                      0.0 |                  1.208793 |                 133.001483 |                  16.65652 |                     0.0 |                      0.0 |   -15.447727 |             0.0 |              0.0 |        116.344963 |             61.405925 |                 100.0 |                50.499098 |               73.395613 | 0.769679 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.007264 |  0.007264 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.769679 | 0.081237 |      0.0 |  0.0 |      0.0 |  0.0 | 0.081237 | 0.417483 | 0.115968 | 0.692736 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8852.106046 | 118.779668 | 0.081237 |   0.0 |      0.0 |   0.0 |      0.0 | 0.417483 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.115968 |
    | 1997-08-04 05:00:00 | 0.833333 | 14.1 |              0.0 |                      1.0 |       19.950903 |             93.6 |       2.5 |              1017.0 | 0.0 | 0.0 | 18.595833 |             80.920833 |                   4.5 |                     15.534444 |           174.919387 | 0.783333 | 14.1 | 18.595833 |    2.003978 |         1.486284 |           2.003978 |          2.5 |                16.112886 |                     21.459534 |                         0.0 |                          0.0 |                      1.045579 |                           1.343433 |            15.081661 |                 17.365233 |    1001.918339 |   1.226451 | -2.9 | 17.158606 |  18.351635 |             15.162686 |                      0.0 |                 15.162686 |                 132.938735 |                 16.779096 |                     0.0 |                      0.0 |     -1.61641 |             0.0 |              0.0 |        116.159638 |             56.493451 |                 100.0 |                49.634525 |                58.02059 | 0.776069 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.018172 |  0.018172 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.776069 | 0.082493 |      0.0 |  0.0 |      0.0 |  0.0 | 0.082493 | 0.408814 | 0.113559 | 0.681828 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  8830.85441 | 119.473244 | 0.082493 |   0.0 |      0.0 |   0.0 |      0.0 | 0.408814 |   0.0 |      0.0 |   0.0 |      0.0 |   0.0 | 0.113559 |
    | 1997-08-04 06:00:00 | 0.833333 | 14.3 |              0.1 |                      1.0 |       62.985775 |             94.6 |       2.4 |              1017.0 | 0.0 | 0.0 |    18.525 |             80.791667 |                   4.6 |                     15.534444 |            175.15812 | 0.783333 | 14.3 |    18.525 |    1.923819 |         1.536383 |           1.923819 |          2.4 |                16.323194 |                     21.364558 |                         0.0 |                          0.0 |                      1.057522 |                           1.338237 |            15.441741 |                 17.260782 |    1001.558259 |   1.225432 | -2.9 | 16.903586 |  15.621518 |             47.869189 |                      0.0 |                 47.869189 |                 133.120171 |                 17.089561 |                     0.0 |                      0.0 |    30.779628 |             0.0 |              0.0 |         116.03061 |             58.847345 |                 100.0 |                 48.85611 |               57.260179 | 0.765161 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.04094 |   0.04094 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.765161 | 0.081907 | 0.003897 |  0.0 | 0.000003 |  0.0 | 0.081907 | 0.400403 | 0.111223 |  0.65906 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8812.332893 | 120.152598 | 0.081907 |   0.0 | 0.003897 |   0.0 | 0.000003 | 0.400377 |   0.0 | 0.000025 |   0.0 |      0.0 |   0.0 | 0.111223 |
    | 1997-08-04 07:00:00 | 0.833333 | 15.4 |              0.9 |                      1.0 |      285.776389 |             93.3 |       2.5 |              1017.0 | 0.0 | 0.0 |    18.475 |             80.620833 |                   5.5 |                     15.534444 |           182.509935 | 0.783333 | 15.4 |    18.475 |    2.003978 |         1.589823 |           2.003978 |          2.5 |                17.523416 |                     21.297737 |                         0.0 |                          0.0 |                      1.125295 |                            1.33458 |            16.349347 |                 17.170413 |    1000.650653 |   1.220347 | -2.9 | 16.681328 |   7.687968 |            217.190056 |                      0.0 |                217.190056 |                  138.70755 |                 19.012403 |                     0.0 |                      0.0 |   198.177653 |             0.0 |              0.0 |        119.695148 |             56.493451 |                 100.0 |                 48.85611 |               57.260179 | 0.742393 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.189733 |  0.189733 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.742393 | 0.080011 |  0.00392 |  0.0 | 0.000031 |  0.0 | 0.080011 | 0.392202 | 0.108945 | 0.510267 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8801.744925 | 120.811029 | 0.080011 |   0.0 |  0.00392 |   0.0 | 0.000031 | 0.392126 |   0.0 | 0.000076 |   0.0 |      0.0 |   0.0 | 0.108945 |
    | 1997-08-04 08:00:00 | 0.833333 | 16.8 |              1.0 |                      1.0 |      475.225247 |             86.9 |       2.5 |              1017.0 | 0.0 | 0.0 |     18.45 |             80.208333 |                   6.5 |                     15.534444 |           195.188105 | 0.783333 | 16.8 |     18.45 |    2.003978 |         1.643262 |           2.003978 |          2.5 |                19.162205 |                     21.264396 |                         0.0 |                          0.0 |                      1.216842 |                           1.332754 |            16.651956 |                 17.055817 |    1000.348044 |   1.214317 | -2.9 | 16.554272 |  -1.474365 |            361.171188 |                      0.0 |                361.171188 |                  148.34296 |                 21.208047 |                     0.0 |                      0.0 |   339.963141 |             0.0 |              0.0 |        127.134913 |             56.493451 |                 100.0 |                 48.85611 |               57.260179 |   0.5936 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.343753 |  0.343753 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |   0.5936 | 0.064359 | 0.003941 |  0.0 | 0.000059 |  0.0 | 0.064359 | 0.383987 | 0.106663 | 0.356247 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  8800.31929 |  121.33627 | 0.064359 |   0.0 | 0.003941 |   0.0 | 0.000059 | 0.383861 |   0.0 | 0.000126 |   0.0 |      0.0 |   0.0 | 0.106663 |
    | 1997-08-04 09:00:00 | 0.833333 | 18.2 |              1.0 |                      1.0 |      614.006719 |             78.9 |       2.9 |              1017.0 | 0.0 | 0.0 | 18.416667 |             79.908333 |                   7.3 |                     15.534444 |           207.781609 | 0.783333 | 18.2 | 18.416667 |    2.324615 |         1.696701 |           2.324615 |          2.9 |                20.933479 |                     21.220011 |                         0.0 |                          0.0 |                      1.314615 |                           1.330323 |            16.516515 |                 16.956557 |    1000.483485 |   1.208543 | -2.9 | 16.537165 |  -9.977011 |            466.645107 |                      0.0 |                466.645107 |                 157.914023 |                 22.996116 |                     0.0 |                      0.0 |   443.648991 |             0.0 |              0.0 |        134.917907 |             48.701251 |                 100.0 |                 48.85611 |               57.260179 | 0.439581 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.494205 |  0.494205 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.439581 |  0.04788 | 0.003958 |  0.0 | 0.000081 |  0.0 |  0.04788 |  0.37557 | 0.104325 | 0.205795 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8807.396301 | 121.723932 |  0.04788 |   0.0 | 0.003958 |   0.0 | 0.000081 | 0.375394 |   0.0 | 0.000176 |   0.0 |      0.0 |   0.0 | 0.104325 |
    | 1997-08-04 10:00:00 | 0.833333 | 19.5 |              0.9 |                      1.0 |      643.806508 |             75.8 |       3.1 |              1017.0 | 0.0 | 0.0 | 18.383333 |             79.866667 |                   7.7 |                     15.534444 |           213.707576 | 0.783333 | 19.5 | 18.383333 |    2.484933 |         1.750141 |           2.484933 |          3.1 |                22.704676 |                     21.175708 |                         0.0 |                          0.0 |                      1.411269 |                           1.327896 |            17.210144 |                 16.912332 |     999.789856 |   1.202862 | -2.9 | 16.622089 | -17.267466 |            489.292946 |                      0.0 |                489.292946 |                 162.417758 |                 23.882018 |                     0.0 |                      0.0 |   465.410928 |             0.0 |              0.0 |         138.53574 |             45.559235 |                 100.0 |                 48.85611 |               57.260179 | 0.289128 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.550473 |  0.550473 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.289128 | 0.031594 | 0.003971 |  0.0 | 0.000097 |  0.0 | 0.031594 | 0.366948 |  0.10193 | 0.149527 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8821.763767 | 121.977399 | 0.031594 |   0.0 | 0.003971 |   0.0 | 0.000097 | 0.366723 |   0.0 | 0.000225 |   0.0 |      0.0 |   0.0 |  0.10193 |
    | 1997-08-04 11:00:00 | 0.833333 | 20.3 |              0.8 |                      1.0 |      600.262981 |             73.4 |       3.3 |              1018.0 | 0.0 | 0.0 |   18.3375 |             79.933333 |                   7.8 |                     15.534444 |           213.092753 | 0.783333 | 20.3 |   18.3375 |    2.645251 |          1.82028 |           2.645251 |          3.3 |                23.858503 |                     21.114922 |                         0.0 |                          0.0 |                      1.473678 |                           1.324565 |            17.512141 |                 16.877861 |    1000.487859 |   1.200635 | -2.9 | 16.794499 | -21.033009 |            456.199865 |                      0.0 |                456.199865 |                 161.950493 |                   24.1295 |                     0.0 |                      0.0 |   432.070365 |             0.0 |              0.0 |        137.820992 |             42.798069 |                 100.0 |                 48.85611 |               57.260179 |  0.23286 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.548409 |  0.548409 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.23286 | 0.025504 | 0.003979 |  0.0 | 0.000107 |  0.0 | 0.025504 | 0.358261 | 0.099517 | 0.151591 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8839.896776 | 122.180668 | 0.025504 |   0.0 | 0.003979 |   0.0 | 0.000107 | 0.357987 |   0.0 | 0.000274 |   0.0 |      0.0 |   0.0 | 0.099517 |
    | 1997-08-04 12:00:00 | 0.833333 | 21.0 |              0.9 |                      1.0 |      620.514139 |             68.3 |       3.3 |              1018.0 | 0.0 | 0.0 |    18.325 |             79.966667 |                   7.9 |                     15.534444 |           212.841496 | 0.783333 | 21.0 |    18.325 |    2.645251 |         1.887079 |           2.645251 |          3.3 |                24.909749 |                     21.098371 |                         0.0 |                          0.0 |                      1.530181 |                           1.323658 |            17.013358 |                 16.871664 |    1000.986642 |   1.198001 | -2.9 | 17.012095 | -23.927432 |            471.590746 |                      0.0 |                471.590746 |                 161.759537 |                 24.341655 |                     0.0 |                      0.0 |    447.24909 |             0.0 |              0.0 |        137.417882 |             42.798069 |                 100.0 |                 48.85611 |               57.260179 | 0.234924 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.597585 |  0.597585 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.234924 | 0.025783 | 0.003986 |  0.0 | 0.000116 |  0.0 | 0.025783 | 0.349725 | 0.097146 | 0.102415 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8860.924208 | 122.385708 | 0.025783 |   0.0 | 0.003986 |   0.0 | 0.000116 | 0.349403 |   0.0 | 0.000322 |   0.0 |      0.0 |   0.0 | 0.097146 |
    | 1997-08-04 13:00:00 | 0.833333 | 21.8 |              0.8 |                      1.0 |      599.441386 |             61.3 |       3.2 |              1018.0 | 0.0 | 0.0 |    18.325 |             79.766667 |                   8.2 |                     15.534444 |           217.145994 | 0.783333 | 21.8 |    18.325 |    2.565092 |         1.943859 |           2.565092 |          3.2 |                26.160453 |                     21.098371 |                         0.0 |                          0.0 |                      1.596982 |                           1.323658 |            16.036358 |                 16.829467 |    1001.963642 |   1.195188 | -2.9 | 17.264424 | -27.213457 |            455.575453 |                      0.0 |                455.575453 |                 165.030956 |                 25.038979 |                     0.0 |                      0.0 |   430.536475 |             0.0 |              0.0 |        139.991977 |             44.135508 |                 100.0 |                 48.85611 |               57.260179 | 0.185749 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.621828 |  0.621828 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.185749 | 0.020423 | 0.003992 |  0.0 | 0.000124 |  0.0 | 0.020423 | 0.341344 | 0.094818 | 0.078172 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8885.237665 | 122.546917 | 0.020423 |   0.0 | 0.003992 |   0.0 | 0.000124 | 0.340974 |   0.0 |  0.00037 |   0.0 |      0.0 |   0.0 | 0.094818 |
    | 1997-08-04 14:00:00 | 0.833333 | 21.9 |              0.9 |                      1.0 |      613.551222 |             60.6 |       2.5 |              1018.0 | 0.0 | 0.0 | 18.283333 |                 79.65 |                   8.7 |                     15.534444 |           225.230587 | 0.783333 | 21.9 | 18.283333 |    2.003978 |         1.963899 |           2.003978 |          2.5 |                26.320577 |                     21.043281 |                         0.0 |                          0.0 |                      1.605502 |                           1.320638 |             15.95027 |                 16.760973 |     1002.04973 |   1.194821 | -2.9 | 17.556185 | -26.062888 |            466.298929 |                      0.0 |                466.298929 |                 171.175246 |                 26.187015 |                     0.0 |                      0.0 |   440.111914 |             0.0 |              0.0 |        144.988231 |             56.493451 |                 100.0 |                 48.85611 |               57.260179 | 0.161505 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.589213 |  0.589213 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.161505 | 0.017784 | 0.003998 |  0.0 | 0.000131 |  0.0 | 0.017784 | 0.333076 | 0.092521 | 0.110787 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8908.400553 | 122.686509 | 0.017784 |   0.0 | 0.003998 |   0.0 | 0.000131 |  0.33266 |   0.0 | 0.000417 |   0.0 |      0.0 |   0.0 | 0.092521 |
    | 1997-08-04 15:00:00 | 0.833333 | 21.8 |              0.9 |                      1.0 |      517.107047 |             58.8 |       2.9 |              1017.0 | 0.0 | 0.0 | 18.245833 |                  79.5 |                   9.1 |                     15.534444 |           230.614741 | 0.783333 | 21.8 | 18.245833 |    2.324615 |         1.997298 |           2.324615 |          2.9 |                26.160453 |                     20.993808 |                         0.0 |                          0.0 |                      1.596982 |                           1.317924 |            15.382347 |                 16.690077 |    1001.617653 |   1.194298 | -2.9 |  17.83414 |  -23.79516 |            393.001356 |                      0.0 |                393.001356 |                 175.267203 |                 27.142101 |                     0.0 |                      0.0 |   365.859255 |             0.0 |              0.0 |        148.125102 |             48.701251 |                 100.0 |                 48.85611 |               57.260179 | 0.194121 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.551455 |  0.551455 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.194121 | 0.021409 | 0.004002 |  0.0 | 0.000137 |  0.0 | 0.021409 | 0.325036 | 0.090288 | 0.148545 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8929.295714 | 122.855082 | 0.021409 |   0.0 | 0.004002 |   0.0 | 0.000137 | 0.324573 |   0.0 | 0.000463 |   0.0 |      0.0 |   0.0 | 0.090288 |
    | 1997-08-04 16:00:00 | 0.833333 | 22.2 |              0.9 |                      1.0 |      377.836617 |             57.2 |       3.6 |              1017.0 | 0.0 | 0.0 | 18.233333 |               79.3375 |                   9.5 |                     15.534444 |           234.753897 | 0.783333 | 22.2 | 18.233333 |    2.885728 |         2.040718 |           2.885728 |          3.6 |                26.806088 |                      20.97734 |                         0.0 |                          0.0 |                      1.631293 |                           1.317021 |            15.333082 |                 16.642897 |    1001.666918 |   1.192703 | -2.9 | 18.084882 | -24.690709 |            287.155829 |                      0.0 |                287.155829 |                 178.412962 |                 28.068928 |                     0.0 |                      0.0 |   259.086901 |             0.0 |              0.0 |        150.344034 |             39.231563 |                 100.0 |                 48.85611 |               57.260179 | 0.231878 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.505795 |  0.505795 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.231878 |  0.02562 | 0.004008 |  0.0 | 0.000144 |  0.0 |  0.02562 | 0.317306 | 0.088141 | 0.194205 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8951.086422 | 123.057187 |  0.02562 |   0.0 | 0.004008 |   0.0 | 0.000144 | 0.316797 |   0.0 | 0.000509 |   0.0 |      0.0 |   0.0 | 0.088141 |
    | 1997-08-04 17:00:00 | 0.833333 | 21.8 |              1.0 |                      1.0 |      252.204639 |             57.2 |       4.2 |              1017.0 | 0.0 | 0.0 | 18.229167 |             79.133333 |                  10.2 |                     15.534444 |           239.548315 | 0.783333 | 21.8 | 18.229167 |    3.366683 |         2.100837 |           3.366683 |          4.2 |                26.160453 |                     20.971853 |                         0.0 |                          0.0 |                      1.596982 |                            1.31672 |            14.963779 |                 16.595726 |    1002.036221 |   1.194485 | -2.9 |  18.34637 | -20.721778 |            191.675526 |                      0.0 |                191.675526 |                  182.05672 |                 29.636995 |                     0.0 |                      0.0 |    162.03853 |             0.0 |              0.0 |        152.419724 |             33.627054 |                 100.0 |                 48.85611 |               57.260179 | 0.277538 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.44464 |   0.44464 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.277538 | 0.030734 | 0.004014 |  0.0 | 0.000152 |  0.0 | 0.030734 | 0.309897 | 0.086082 |  0.25536 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |   8968.9082 | 123.299825 | 0.030734 |   0.0 | 0.004014 |   0.0 | 0.000152 | 0.309342 |   0.0 | 0.000555 |   0.0 |      0.0 |   0.0 | 0.086082 |
    | 1997-08-04 18:00:00 | 0.833333 | 20.9 |              1.0 |                      1.0 |      128.576714 |             59.9 |       4.2 |              1017.0 | 0.0 | 0.0 | 18.208333 |               78.8375 |                  11.1 |                     15.534444 |           242.777315 | 0.783333 | 20.9 | 18.208333 |    3.366683 |         2.157616 |           3.366683 |          4.2 |                 24.75714 |                     20.944437 |                         0.0 |                          0.0 |                      1.521999 |                           1.315216 |            14.829527 |                  16.51207 |    1002.170473 |   1.198201 | -2.9 | 18.560232 |  -14.03861 |             97.718303 |                      0.0 |                 97.718303 |                  184.51076 |                 31.712576 |                     0.0 |                      0.0 |    66.005727 |             0.0 |              0.0 |        152.798184 |             33.627054 |                 100.0 |                 48.85611 |               57.260179 | 0.338693 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.32482 |   0.32482 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.338693 | 0.037607 | 0.004022 |  0.0 | 0.000162 |  0.0 | 0.037607 | 0.302834 | 0.084121 |  0.37518 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8980.046809 | 123.596727 | 0.037607 |   0.0 | 0.004022 |   0.0 | 0.000162 | 0.302234 |   0.0 |   0.0006 |   0.0 |      0.0 |   0.0 | 0.084121 |
    | 1997-08-04 19:00:00 | 0.833333 | 19.7 |              1.0 |                      1.0 |       42.072586 |             66.1 |       3.6 |              1017.0 | 0.0 | 0.0 | 18.166667 |               78.4875 |                  12.1 |                     15.534444 |           243.962306 | 0.783333 | 19.7 | 18.166667 |    2.885728 |         2.194356 |           2.885728 |          3.6 |                22.988466 |                     20.889699 |                         0.0 |                          0.0 |                      1.426658 |                           1.312212 |            15.195376 |                 16.395802 |    1001.804624 |   1.202947 | -2.9 | 18.693895 |   -6.03663 |             31.975165 |                      0.0 |                 31.975165 |                 185.411352 |                 34.091963 |                     0.0 |                      0.0 |    -2.116797 |             0.0 |              0.0 |        151.319389 |             39.231563 |                 100.0 |                 48.85611 |               57.260179 | 0.458513 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.183135 |  0.183135 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.458513 | 0.051088 | 0.004032 |  0.0 | 0.000175 |  0.0 | 0.051088 | 0.296218 | 0.082283 | 0.516865 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8983.183439 | 123.999945 | 0.051088 |   0.0 | 0.004032 |   0.0 | 0.000175 | 0.295573 |   0.0 | 0.000644 |   0.0 |      0.0 |   0.0 | 0.082283 |
    | 1997-08-04 20:00:00 | 0.833333 | 17.9 |              0.3 |                 0.104711 |        0.812897 |             71.7 |       3.2 |              1017.0 | 0.0 | 0.0 | 18.104167 |             78.091667 |                  12.4 |                     15.502754 |           243.988429 | 0.783333 | 17.9 | 18.104167 |    2.565092 |         2.227756 |           2.565092 |          3.2 |                20.542326 |                     20.807826 |                         0.0 |                          0.0 |                      1.293123 |                           1.307717 |            14.728848 |                 16.249178 |    1002.271152 |   1.210597 | -2.9 | 18.731535 |   4.989208 |              0.617802 |                      0.0 |                  0.617802 |                 185.431206 |                 35.120093 |                     0.0 |                      0.0 |   -34.502291 |             0.0 |              0.0 |        150.311113 |             44.135508 |                 100.0 |                 48.85611 |               83.091125 | 0.600198 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.100625 |  0.100625 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.600198 | 0.067185 | 0.004045 |  0.0 | 0.000192 |  0.0 | 0.067185 | 0.290156 | 0.080599 | 0.599375 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8975.294232 | 124.528721 | 0.067185 |   0.0 | 0.004045 |   0.0 | 0.000192 | 0.289467 |   0.0 | 0.000689 |   0.0 |      0.0 |   0.0 | 0.080599 |
    | 1997-08-04 21:00:00 | 0.833333 | 15.8 |              0.0 |                      0.0 |             0.0 |             80.6 |       2.2 |              1018.0 | 0.0 | 0.0 | 18.020833 |             77.829167 |                  12.4 |                     15.502754 |           243.988429 | 0.783333 | 15.8 | 18.020833 |    1.763501 |         2.244455 |           1.763501 |          2.2 |                17.978626 |                     20.699099 |                         0.0 |                          0.0 |                      1.150835 |                           1.301743 |            14.490772 |                 16.109936 |    1003.509228 |    1.22071 | -2.9 | 18.636864 |  17.021185 |                   0.0 |                      0.0 |                       0.0 |                 185.431206 |                 35.424509 |                     0.0 |                      0.0 |   -35.424509 |             0.0 |              0.0 |        150.006697 |             64.197103 |                 100.0 |                 48.85611 |               87.719298 | 0.682708 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.029975 |  0.029975 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.682708 |  0.07685 | 0.004062 |  0.0 | 0.000214 |  0.0 |  0.07685 | 0.284583 | 0.079051 | 0.670025 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8955.373047 | 125.130304 |  0.07685 |   0.0 | 0.004062 |   0.0 | 0.000214 | 0.283851 |   0.0 | 0.000732 |   0.0 |      0.0 |   0.0 | 0.079051 |
    | 1997-08-04 22:00:00 | 0.833333 | 14.8 |              0.0 |                      0.0 |             0.0 |             85.9 |       2.2 |              1018.0 | 0.0 | 0.0 | 17.929167 |             77.654167 |                  12.4 |                     15.502754 |           243.988429 | 0.783333 | 14.8 | 17.929167 |    1.763501 |         2.261155 |           1.763501 |          2.2 |                16.859515 |                     20.580072 |                         0.0 |                          0.0 |                      1.087886 |                           1.295199 |            14.482324 |                 15.981284 |    1003.517676 |   1.224953 | -2.9 |  18.39781 |  21.586859 |                   0.0 |                      0.0 |                       0.0 |                 185.431206 |                 35.698198 |                     0.0 |                      0.0 |   -35.698198 |             0.0 |              0.0 |        149.733007 |             64.197103 |                 100.0 |                 48.85611 |               87.719298 | 0.753358 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.012059 |  0.012059 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.753358 | 0.085333 | 0.004082 |  0.0 | 0.000239 |  0.0 | 0.085333 |  0.27939 | 0.077608 | 0.687941 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8930.886187 | 125.794008 | 0.085333 |   0.0 | 0.004082 |   0.0 | 0.000239 | 0.278614 |   0.0 | 0.000776 |   0.0 |      0.0 |   0.0 | 0.077608 |
    | 1997-08-04 23:00:00 | 0.833333 | 14.1 |              0.0 |                      0.0 |             0.0 |             89.3 |       2.1 |              1018.0 | 0.0 | 0.0 | 17.833333 |               77.5875 |                  12.4 |                     15.502754 |           243.988429 | 0.783333 | 14.1 | 17.833333 |    1.683342 |         2.254475 |           1.683342 |          2.1 |                16.112886 |                     20.456276 |                         0.0 |                          0.0 |                      1.045579 |                           1.288387 |            14.388807 |                 15.871513 |    1003.611193 |   1.227981 | -2.9 | 18.103968 |  24.023805 |                   0.0 |                      0.0 |                       0.0 |                 185.431206 |                 35.921753 |                     0.0 |                      0.0 |   -35.921753 |             0.0 |              0.0 |        149.509453 |             67.254108 |                 100.0 |                 48.85611 |               87.719298 | 0.771275 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.000116 | -0.000116 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.771275 | 0.087941 | 0.004104 |  0.0 | 0.000266 |  0.0 | 0.087941 | 0.274474 | 0.076243 | 0.700116 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8903.962382 | 126.472972 | 0.087941 |   0.0 | 0.004104 |   0.0 | 0.000266 | 0.273655 |   0.0 | 0.000819 |   0.0 |      0.0 |   0.0 | 0.076243 |
    | 1997-08-05 00:00:00 | 0.833333 | 13.4 |              0.0 |                      0.0 |             0.0 |             91.4 |       1.8 |              1018.0 | 0.0 | 0.0 | 17.729167 |             77.729167 |                  12.4 |                     15.502754 |           243.988429 | 0.783333 | 13.4 | 17.729167 |    1.442864 |         2.237776 |           1.442864 |          1.8 |                15.395377 |                     20.322453 |                         0.0 |                          0.0 |                      1.004677 |                           1.281018 |            14.071374 |                 15.796473 |    1003.928626 |   1.231127 | -2.9 | 17.780882 |  26.285292 |                   0.0 |                      0.0 |                       0.0 |                 185.431206 |                 36.050541 |                     0.0 |                      0.0 |   -36.050541 |             0.0 |              0.0 |        149.380664 |             78.463126 |                 100.0 |                 48.85611 |               87.719298 |  0.78345 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.009023 | -0.009023 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.78345 | 0.089929 | 0.004126 |  0.0 | 0.000295 |  0.0 | 0.089929 | 0.269746 | 0.074929 | 0.709023 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  8874.77709 | 127.162072 | 0.089929 |   0.0 | 0.004126 |   0.0 | 0.000295 | 0.268884 |   0.0 | 0.000862 |   0.0 |      0.0 |   0.0 | 0.074929 |
    | 1997-08-05 01:00:00 | 0.833333 | 12.8 |              0.0 |                      0.0 |             0.0 |             93.1 |       1.8 |              1018.0 | 0.0 | 0.0 | 17.620833 |             77.891667 |                  12.4 |                     15.502754 |           243.988429 | 0.783333 | 12.8 | 17.620833 |    1.442864 |         2.224416 |           1.442864 |          1.8 |                14.802811 |                     20.184089 |                         0.0 |                          0.0 |                      0.970706 |                           1.273391 |            13.781417 |                 15.721724 |    1004.218583 |   1.233843 | -2.9 | 17.430658 |  27.783951 |                   0.0 |                      0.0 |                       0.0 |                 185.431206 |                 36.176259 |                     0.0 |                      0.0 |   -36.176259 |             0.0 |              0.0 |        149.254947 |             78.463126 |                 100.0 |                 48.85611 |               87.719298 | 0.792356 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.013264 | -0.013264 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.792356 | 0.091568 | 0.004148 |  0.0 | 0.000323 |  0.0 | 0.091568 | 0.265187 | 0.073663 | 0.713264 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  8844.09314 | 127.858389 | 0.091568 |   0.0 | 0.004148 |   0.0 | 0.000323 | 0.264282 |   0.0 | 0.000904 |   0.0 |      0.0 |   0.0 | 0.073663 |
    | 1997-08-05 02:00:00 | 0.833333 | 12.3 |              0.0 |                      0.0 |             0.0 |             94.6 |       1.6 |              1018.0 | 0.0 | 0.0 |      17.5 |             78.133333 |                  12.4 |                     15.502754 |           243.988429 | 0.783333 | 12.3 |      17.5 |    1.282546 |         2.201036 |           1.282546 |          1.6 |                14.324376 |                     20.030733 |                         0.0 |                          0.0 |                      0.943147 |                            1.26493 |            13.550859 |                 15.650679 |    1004.449141 |   1.236111 | -2.9 | 17.062451 |  28.574706 |                   0.0 |                      0.0 |                       0.0 |                 185.431206 |                 36.284611 |                     0.0 |                      0.0 |   -36.284611 |             0.0 |              0.0 |        149.146595 |             88.271017 |                 100.0 |                 48.85611 |               87.719298 | 0.796598 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.017718 | -0.017718 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.796598 | 0.092684 | 0.004171 |  0.0 | 0.000352 |  0.0 | 0.092684 | 0.260782 |  0.07244 | 0.717718 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8812.618434 | 128.557779 | 0.092684 |   0.0 | 0.004171 |   0.0 | 0.000352 | 0.259835 |   0.0 | 0.000947 |   0.0 |      0.0 |   0.0 |  0.07244 |
    | 1997-08-05 03:00:00 | 0.833333 | 12.0 |              0.0 |                      0.0 |             0.0 |             95.6 |       1.2 |              1018.0 | 0.0 | 0.0 | 17.379167 |             78.379167 |                  12.4 |                     15.502754 |           243.988429 | 0.783333 | 12.0 | 17.379167 |    0.961909 |         2.174316 |           0.961909 |          1.2 |                 14.04387 |                     19.878396 |                         0.0 |                          0.0 |                      0.926933 |                           1.256516 |            13.425939 |                 15.580521 |    1004.574061 |   1.237469 | -2.9 | 16.684755 |  28.108527 |                   0.0 |                      0.0 |                       0.0 |                 185.431206 |                 36.390891 |                     0.0 |                      0.0 |   -36.390891 |             0.0 |              0.0 |        149.040314 |            117.694689 |                 100.0 |                 48.85611 |               87.719298 | 0.801051 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.020716 | -0.020716 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.801051 | 0.093837 | 0.004194 |  0.0 | 0.000382 |  0.0 | 0.093837 | 0.256522 | 0.071256 | 0.720716 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8781.609907 | 129.260418 | 0.093837 |   0.0 | 0.004194 |   0.0 | 0.000382 | 0.255533 |   0.0 | 0.000989 |   0.0 | 0.000001 |   0.0 | 0.071256 |
    | 1997-08-05 04:00:00 | 0.833333 | 11.7 |              0.0 |                 0.366022 |        1.482347 |             96.2 |       1.9 |              1018.0 | 0.0 | 0.0 | 17.270833 |               78.5125 |                  12.4 |                     15.470733 |           243.983922 | 0.783333 | 11.7 | 17.270833 |    1.523023 |         2.160956 |           1.523023 |          1.9 |                13.768192 |                      19.74268 |                         0.0 |                          0.0 |                      0.910957 |                           1.249013 |            13.245001 |                 15.500472 |    1004.754999 |   1.238856 | -2.9 | 16.312652 |  27.675913 |              1.126584 |                      0.0 |                  1.126584 |                  185.42778 |                 36.589063 |                     0.0 |                      0.0 |   -35.462479 |             0.0 |              0.0 |        148.838717 |             74.333488 |                 100.0 |                 48.85611 |               73.423526 | 0.804049 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.019578 | -0.019578 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.804049 | 0.094829 | 0.004217 |  0.0 | 0.000411 |  0.0 | 0.094829 |   0.2524 | 0.070111 | 0.719578 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8751.033993 | 129.965011 | 0.094829 |   0.0 | 0.004217 |   0.0 | 0.000411 | 0.251369 |   0.0 |  0.00103 |   0.0 | 0.000001 |   0.0 | 0.070111 |
    | 1997-08-05 05:00:00 | 0.833333 | 11.6 |              0.0 |                      1.0 |       21.135567 |             95.7 |       1.3 |              1018.0 | 0.0 | 0.0 | 17.166667 |                  78.6 |                  12.4 |                     15.470733 |           244.033283 | 0.783333 | 11.6 | 17.166667 |    1.042069 |         2.120877 |           1.042069 |          1.3 |                 13.67736 |                     19.612949 |                         0.0 |                          0.0 |                      0.905683 |                           1.241834 |            13.089234 |                 15.415778 |    1004.910766 |   1.239363 | -2.9 | 15.945741 |  26.074448 |             16.063031 |                      0.0 |                 16.063031 |                 185.465295 |                  36.74524 |                     0.0 |                      0.0 |   -20.682209 |             0.0 |              0.0 |        148.720055 |            108.641252 |                 100.0 |                 48.85611 |               57.260179 | 0.802911 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.009263 | -0.009263 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.802911 | 0.095337 |  0.00424 |  0.0 |  0.00044 |  0.0 | 0.095337 | 0.248404 | 0.069001 | 0.709263 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8722.059546 | 130.667906 | 0.095337 |   0.0 |  0.00424 |   0.0 |  0.00044 | 0.247332 |   0.0 | 0.001072 |   0.0 | 0.000001 |   0.0 | 0.069001 |
    | 1997-08-05 06:00:00 | 0.833333 | 12.6 |              0.6 |                      1.0 |       125.88345 |             95.1 |       1.6 |              1018.0 | 0.0 | 0.0 | 17.095833 |             78.620833 |                  12.9 |                     15.470733 |           246.654019 | 0.783333 | 12.6 | 17.095833 |    1.282546 |         2.094157 |           1.282546 |          1.6 |                14.609782 |                     19.525158 |                         0.0 |                          0.0 |                      0.959602 |                           1.236973 |            13.893903 |                 15.350842 |    1004.106097 |   1.234655 | -2.9 | 15.598048 |  17.988287 |             95.671422 |                      0.0 |                 95.671422 |                 187.457054 |                  38.00446 |                     0.0 |                      0.0 |    57.666962 |             0.0 |              0.0 |        149.452594 |             88.271017 |                 100.0 |                 48.85611 |               57.260179 | 0.792597 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.052361 |  0.052361 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.792597 | 0.094743 | 0.004263 |  0.0 | 0.000469 |  0.0 | 0.094743 | 0.244512 |  0.06792 | 0.647639 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8701.171259 | 131.361028 | 0.094743 |   0.0 | 0.004263 |   0.0 | 0.000469 | 0.243398 |   0.0 | 0.001113 |   0.0 | 0.000001 |   0.0 |  0.06792 |
    | 1997-08-05 07:00:00 | 0.833333 | 14.8 |              0.9 |                      1.0 |      305.954944 |             96.3 |       1.9 |              1019.0 | 0.0 | 0.0 | 17.070833 |             78.745833 |                  12.9 |                     15.470733 |           247.494792 | 0.783333 | 14.8 | 17.070833 |    1.523023 |         2.074117 |           1.523023 |          1.9 |                16.859515 |                     19.494255 |                         0.0 |                          0.0 |                      1.087886 |                           1.235261 |            16.235713 |                 15.350914 |    1002.764287 |   1.225361 | -2.9 | 15.347388 |   3.284331 |            232.525758 |                      0.0 |                232.525758 |                 188.096042 |                 37.987562 |                     0.0 |                      0.0 |   194.538195 |             0.0 |              0.0 |         150.10848 |             74.333488 |                 100.0 |                 48.85611 |               57.260179 | 0.730972 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.166779 |  0.166779 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.730972 | 0.087933 | 0.004285 |  0.0 | 0.000498 |  0.0 | 0.087933 | 0.240625 |  0.06684 | 0.533221 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8694.986928 | 131.999284 | 0.087933 |   0.0 | 0.004285 |   0.0 | 0.000498 |  0.23947 |   0.0 | 0.001154 |   0.0 | 0.000001 |   0.0 |  0.06684 |
    | 1997-08-05 08:00:00 | 0.833333 | 17.1 |              1.0 |                      1.0 |      511.104575 |             89.9 |       3.2 |              1019.0 | 0.0 | 0.0 | 17.083333 |             78.870833 |                  12.9 |                     15.470733 |           248.989764 | 0.783333 | 17.1 | 17.083333 |    2.565092 |         2.097497 |           2.565092 |          3.2 |                19.530313 |                     19.509701 |                         0.0 |                          0.0 |                      1.237258 |                           1.236116 |            17.557751 |                 15.387464 |    1001.442249 |   1.215051 | -2.9 | 15.273176 | -10.960941 |            388.439477 |                      0.0 |                388.439477 |                 189.232221 |                 37.896217 |                     0.0 |                      0.0 |    350.54326 |             0.0 |              0.0 |        151.336004 |             44.135508 |                 100.0 |                 48.85611 |               57.260179 | 0.616554 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.358863 |  0.358863 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.616554 | 0.074584 | 0.004306 |  0.0 | 0.000525 |  0.0 | 0.074584 | 0.236578 | 0.065716 | 0.341137 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8703.047869 | 132.536423 | 0.074584 |   0.0 | 0.004306 |   0.0 | 0.000525 | 0.235382 |   0.0 | 0.001195 |   0.0 | 0.000001 |   0.0 | 0.065716 |
    | 1997-08-05 09:00:00 | 0.833333 | 19.4 |              0.9 |                      1.0 |      629.519239 |             79.2 |       4.0 |              1019.0 | 0.0 | 0.0 | 17.133333 |             78.883333 |                  12.8 |                     15.470733 |           249.636119 | 0.783333 | 19.4 | 17.133333 |    3.206365 |         2.134237 |           3.206365 |          4.0 |                22.563931 |                     19.571593 |                         0.0 |                          0.0 |                      1.403627 |                           1.239544 |            17.870634 |                 15.438725 |    1001.129366 |   1.205358 | -2.9 | 15.369908 | -24.180553 |            478.434622 |                      0.0 |                478.434622 |                 189.723451 |                  37.56462 |                     0.0 |                      0.0 |   440.870002 |             0.0 |              0.0 |        152.158831 |             35.308407 |                 100.0 |                 48.85611 |               57.260179 | 0.424471 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.541806 |  0.541806 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.424471 | 0.051569 | 0.004324 |  0.0 | 0.000547 |  0.0 | 0.051569 | 0.232167 | 0.064491 | 0.158194 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8724.328423 | 132.904454 | 0.051569 |   0.0 | 0.004324 |   0.0 | 0.000547 |  0.23093 |   0.0 | 0.001236 |   0.0 | 0.000001 |   0.0 | 0.064491 |
    | 1997-08-05 10:00:00 | 0.833333 | 20.6 |              1.0 |                      1.0 |      765.816589 |             73.4 |       4.6 |              1019.0 | 0.0 | 0.0 | 17.179167 |             78.783333 |                  12.9 |                     15.470733 |           254.719872 | 0.783333 | 20.6 | 17.179167 |     3.68732 |         2.184336 |            3.68732 |          4.6 |                24.304197 |                     19.628477 |                         0.0 |                          0.0 |                      1.497674 |                           1.242694 |            17.839281 |                 15.463969 |    1001.160719 |   1.200448 | -2.9 | 15.625274 | -29.848354 |            582.020608 |                      0.0 |                582.020608 |                 193.587103 |                 37.751868 |                     0.0 |                      0.0 |    544.26874 |             0.0 |              0.0 |        155.835235 |             30.702962 |                 100.0 |                 48.85611 |               57.260179 | 0.241527 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.716842 |       0.7 | 0.010395 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.241527 | 0.029423 | 0.004336 |  0.0 | 0.000563 |  0.0 | 0.029423 | 0.227287 | 0.063135 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8751.276776 | 133.101265 | 0.029423 |   0.0 | 0.004336 |   0.0 | 0.000563 |  0.22601 |   0.0 | 0.001276 |   0.0 | 0.000001 |   0.0 | 0.063135 |
    | 1997-08-05 11:00:00 | 0.833333 | 21.4 |              1.0 |                      1.0 |      767.145842 |             67.3 |       4.1 |              1019.0 | 0.0 | 0.0 |    17.225 |             78.529167 |                  13.1 |                     15.470733 |           261.673325 | 0.783333 | 21.4 |    17.225 |    3.286524 |         2.211056 |           3.286524 |          4.1 |                25.528421 |                     19.685506 |                         0.0 |                          0.0 |                      1.563281 |                            1.24585 |            17.180627 |                 15.458864 |    1001.819373 |   1.197482 | -2.9 | 15.948655 | -32.708072 |             583.03084 |                      0.0 |                 583.03084 |                 198.871727 |                 38.247049 |                     0.0 |                      0.0 |   544.783791 |             0.0 |              0.0 |        160.624678 |             34.447226 |                 100.0 |                 48.85611 |               57.260179 | 0.083333 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.746089 |       0.7 | 0.029871 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.083333 | 0.010164 | 0.004342 |  0.0 | 0.000571 |  0.0 | 0.010164 | 0.221999 | 0.061666 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8781.084848 |  133.13965 | 0.010164 |   0.0 | 0.004342 |   0.0 | 0.000571 | 0.220682 |   0.0 | 0.001316 |   0.0 | 0.000001 |   0.0 | 0.061666 |
    | 1997-08-05 12:00:00 | 0.833333 | 22.5 |              1.0 |                      1.0 |      736.263853 |             62.2 |       4.1 |              1019.0 | 0.0 | 0.0 |   17.2875 |                78.275 |                  13.2 |                     15.470733 |            266.49623 | 0.783333 | 22.5 |   17.2875 |    3.286524 |         2.237776 |           3.286524 |          4.1 |                27.299387 |                     19.763507 |                         0.0 |                          0.0 |                      1.657431 |                           1.250165 |            16.980219 |                 15.469885 |    1002.019781 |   1.193116 | -2.9 | 16.306352 | -37.161891 |            559.560528 |                      0.0 |                559.560528 |                 202.537135 |                  38.48429 |                     0.0 |                      0.0 |   521.076238 |             0.0 |              0.0 |        164.052844 |             34.447226 |                 100.0 |                 48.85611 |               57.260179 | 0.083333 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.771457 |       0.7 | 0.046934 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.083333 | 0.010168 | 0.004343 |  0.0 | 0.000572 |  0.0 | 0.010168 | 0.216601 | 0.060167 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8815.346739 | 133.160966 | 0.010168 |   0.0 | 0.004343 |   0.0 | 0.000572 | 0.215244 |   0.0 | 0.001355 |   0.0 | 0.000001 |   0.0 | 0.060167 |
    | 1997-08-05 13:00:00 | 0.833333 | 23.1 |              1.0 |                      1.0 |      768.006183 |             59.6 |       4.4 |              1019.0 | 0.0 | 0.0 | 17.341667 |             78.204167 |                  13.4 |                     15.470733 |           273.519763 | 0.783333 | 23.1 | 17.341667 |    3.527001 |         2.277855 |           3.527001 |          4.4 |                28.309776 |                     19.831326 |                         0.0 |                          0.0 |                      1.710768 |                           1.253914 |            16.872626 |                 15.508923 |    1002.127374 |   1.190747 | -2.9 | 16.717494 | -38.295035 |            583.684699 |                      0.0 |                583.684699 |                  207.87502 |                  38.86309 |                     0.0 |                      0.0 |   544.821609 |             0.0 |              0.0 |         169.01193 |             32.098552 |                 100.0 |                 48.85611 |               57.260179 | 0.083333 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.841038 |       0.7 | 0.091262 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.083333 |  0.01017 | 0.004344 |  0.0 | 0.000573 |  0.0 |  0.01017 | 0.211342 | 0.058706 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8850.741774 |  133.13795 |  0.01017 |   0.0 | 0.004344 |   0.0 | 0.000573 | 0.209947 |   0.0 | 0.001394 |   0.0 | 0.000001 |   0.0 | 0.058706 |
    | 1997-08-05 14:00:00 | 0.833333 | 23.4 |              1.0 |                      1.0 |      735.929369 |             55.4 |       4.6 |              1019.0 | 0.0 | 0.0 | 17.404167 |               77.9875 |                  13.5 |                     15.470733 |           278.618852 | 0.783333 | 23.4 | 17.404167 |     3.68732 |         2.347994 |            3.68732 |          4.6 |                28.827078 |                     19.909831 |                         0.0 |                          0.0 |                      1.737973 |                           1.258253 |            15.970201 |                 15.527179 |    1003.029799 |   1.189943 | -2.9 | 17.142235 | -37.546592 |            559.306321 |                      0.0 |                559.306321 |                 211.750328 |                 39.079824 |                     0.0 |                      0.0 |   520.226497 |             0.0 |              0.0 |        172.670504 |             30.702962 |                 100.0 |                 48.85611 |               57.260179 | 0.083333 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.865229 |       0.7 | 0.105785 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.083333 | 0.010167 | 0.004343 |  0.0 | 0.000572 |  0.0 | 0.010167 |  0.20622 | 0.057283 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8885.388366 | 133.100415 | 0.010167 |   0.0 | 0.004343 |   0.0 | 0.000572 | 0.204786 |   0.0 | 0.001432 |   0.0 | 0.000001 |   0.0 | 0.057283 |
    | 1997-08-05 15:00:00 | 0.833333 | 23.8 |              1.0 |                      1.0 |      620.374953 |             52.1 |       4.5 |              1019.0 | 0.0 | 0.0 |   17.4875 |             77.708333 |                  13.6 |                     15.470733 |           282.921682 | 0.783333 | 23.8 |   17.4875 |    3.607161 |         2.401434 |           3.607161 |          4.5 |                29.529613 |                     20.014927 |                         0.0 |                          0.0 |                      1.774812 |                           1.264057 |            15.384928 |                 15.553266 |    1003.615072 |     1.1886 | -2.9 | 17.557994 | -37.452038 |            471.484964 |                      0.0 |                471.484964 |                 215.020478 |                 39.288738 |                     0.0 |                      0.0 |   432.196226 |             0.0 |              0.0 |         175.73174 |              31.38525 |                 100.0 |                 48.85611 |               57.260179 | 0.083333 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.79847 |       0.7 | 0.063817 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.083333 | 0.010164 | 0.004342 |  0.0 | 0.000571 |  0.0 | 0.010164 | 0.201231 | 0.055897 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8919.940404 | 133.104855 | 0.010164 |   0.0 | 0.004342 |   0.0 | 0.000571 | 0.199759 |   0.0 |  0.00147 |   0.0 | 0.000001 |   0.0 | 0.055897 |
    | 1997-08-05 16:00:00 | 0.833333 | 24.0 |              1.0 |                      1.0 |      450.095264 |             49.5 |       4.7 |              1018.0 | 0.0 | 0.0 |   17.5625 |               77.3875 |                  13.7 |                     15.470733 |           285.932459 | 0.783333 | 24.0 |   17.5625 |    3.767479 |         2.438173 |           3.767479 |          4.7 |                29.886439 |                     20.109928 |                         0.0 |                          0.0 |                      1.793476 |                             1.2693 |            14.793787 |                  15.56257 |    1003.206213 |   1.186889 | -2.9 | 17.972618 | -36.164291 |            342.072401 |                      0.0 |                342.072401 |                 217.308669 |                 39.539407 |                     0.0 |                      0.0 |   302.532993 |             0.0 |              0.0 |        177.769261 |             30.049708 |                 100.0 |                 48.85611 |               57.260179 | 0.083333 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.70277 |       0.7 | 0.001774 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.083333 | 0.010164 | 0.004342 |  0.0 | 0.000571 |  0.0 | 0.010164 | 0.196371 | 0.054547 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8953.204695 | 133.171337 | 0.010164 |   0.0 | 0.004342 |   0.0 | 0.000571 | 0.194862 |   0.0 | 0.001508 |   0.0 | 0.000002 |   0.0 | 0.054547 |
    | 1997-08-05 17:00:00 | 0.833333 | 24.0 |              1.0 |                      1.0 |      273.614142 |             48.1 |       4.5 |              1018.0 | 0.0 | 0.0 | 17.654167 |             77.008333 |                  13.7 |                     15.470733 |           286.824521 | 0.783333 | 24.0 | 17.654167 |    3.607161 |         2.448193 |           3.607161 |          4.5 |                29.886439 |                     20.226575 |                         0.0 |                          0.0 |                      1.793476 |                           1.275734 |            14.375377 |                 15.576148 |    1003.624623 |   1.187075 | -2.9 |  18.37179 | -33.769262 |            207.946748 |                      0.0 |                207.946748 |                 217.986636 |                 39.564612 |                     0.0 |                      0.0 |   168.382136 |             0.0 |              0.0 |        178.422024 |              31.38525 |                 100.0 |                 48.85611 |               57.260179 | 0.083333 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.559369 |  0.559369 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.083333 | 0.010171 | 0.004344 |  0.0 | 0.000574 |  0.0 | 0.010171 | 0.191637 | 0.053233 | 0.140631 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8984.073957 | 133.239582 | 0.010171 |   0.0 | 0.004344 |   0.0 | 0.000574 | 0.190091 |   0.0 | 0.001544 |   0.0 | 0.000002 |   0.0 | 0.053233 |
    | 1997-08-05 18:00:00 | 0.833333 | 23.7 |              1.0 |                      1.0 |      133.368106 |             49.4 |       4.1 |              1018.0 | 0.0 | 0.0 | 17.770833 |             76.570833 |                  13.7 |                     15.470733 |           287.024163 | 0.783333 | 23.7 | 17.770833 |    3.286524 |         2.444853 |           3.286524 |          4.1 |                29.352596 |                      20.37589 |                         0.0 |                          0.0 |                      1.765541 |                           1.283961 |            14.500182 |                 15.601989 |    1003.499818 |   1.188219 | -2.9 | 18.742221 | -29.746675 |             101.35976 |                      0.0 |                 101.35976 |                 218.138364 |                 39.572333 |                     0.0 |                      0.0 |    61.787427 |             0.0 |              0.0 |        178.566031 |             34.447226 |                 100.0 |                 48.85611 |               57.260179 | 0.223964 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.406457 |  0.406457 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.223964 | 0.027368 | 0.004346 |  0.0 | 0.000577 |  0.0 | 0.027368 | 0.187249 | 0.052014 | 0.293543 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 9010.920632 | 133.431255 | 0.027368 |   0.0 | 0.004346 |   0.0 | 0.000577 | 0.185667 |   0.0 | 0.001581 |   0.0 | 0.000002 |   0.0 | 0.052014 |
    | 1997-08-05 19:00:00 | 0.833333 | 22.8 |              1.0 |                      1.0 |       40.326503 |             57.0 |       3.8 |              1018.0 | 0.0 | 0.0 |      17.9 |             76.191667 |                  13.7 |                     15.470733 |           286.951409 | 0.783333 | 22.8 |      17.9 |    3.046047 |         2.451533 |           3.046047 |          3.8 |                27.800581 |                     20.542326 |                         0.0 |                          0.0 |                      1.683922 |                           1.293123 |            15.846331 |                 15.651541 |    1002.153669 |   1.191234 | -2.9 | 19.064381 | -22.413715 |             30.648142 |                      0.0 |                 30.648142 |                 218.083071 |                 39.521285 |                     0.0 |                      0.0 |    -8.873143 |             0.0 |              0.0 |        178.561786 |             37.166744 |                 100.0 |                 48.85611 |               57.260179 | 0.376876 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.257526 |  0.257526 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.376876 | 0.046166 | 0.004353 |  0.0 | 0.000585 |  0.0 | 0.046166 |  0.18344 | 0.050956 | 0.442474 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 9030.434346 | 133.757028 | 0.046166 |   0.0 | 0.004353 |   0.0 | 0.000585 | 0.181821 |   0.0 | 0.001617 |   0.0 | 0.000002 |   0.0 | 0.050956 |
    | 1997-08-05 20:00:00 | 0.833333 | 21.3 |              0.5 |                 0.072689 |        0.867122 |             65.9 |       2.5 |              1018.0 | 0.0 | 0.0 | 18.041667 |                 75.95 |                  13.9 |                     15.438712 |           286.953669 | 0.783333 | 21.3 | 18.041667 |    2.003978 |         2.428153 |           2.003978 |          2.5 |                 25.37251 |                     20.726234 |                         0.0 |                          0.0 |                       1.55495 |                           1.303234 |            16.720484 |                 15.741575 |    1001.279516 |   1.196911 | -2.9 | 19.298545 | -12.008727 |              0.659013 |                      0.0 |                  0.659013 |                 218.084788 |                 39.876809 |                     0.0 |                      0.0 |   -39.217796 |             0.0 |              0.0 |        178.207979 |             56.493451 |                 100.0 |                 48.85611 |               84.453771 | 0.525807 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.102206 |  0.102206 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.525807 | 0.064647 | 0.004363 |  0.0 | 0.000598 |  0.0 | 0.064647 | 0.180211 | 0.050059 | 0.597794 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 9039.543073 | 134.213226 | 0.064647 |   0.0 | 0.004363 |   0.0 | 0.000598 | 0.178557 |   0.0 | 0.001652 |   0.0 | 0.000002 |   0.0 | 0.050059 |
    | 1997-08-05 21:00:00 | 0.833333 | 19.3 |              0.0 |                      0.0 |             0.0 |             73.0 |       1.9 |              1018.0 | 0.0 | 0.0 |   18.1875 |             75.633333 |                  13.9 |                     15.438712 |           286.953669 | 0.783333 | 19.3 |   18.1875 |    1.523023 |         2.418134 |           1.523023 |          1.9 |                22.423949 |                     20.917052 |                         0.0 |                          0.0 |                       1.39602 |                           1.313713 |            16.369483 |                 15.820264 |    1001.630517 |   1.205255 | -2.9 |  19.40785 |   0.647101 |                   0.0 |                      0.0 |                       0.0 |                 218.084788 |                 39.753451 |                     0.0 |                      0.0 |   -39.753451 |             0.0 |              0.0 |        178.331337 |             74.333488 |                 100.0 |                 48.85611 |               87.719298 | 0.681127 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.046254 |  0.046254 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.681127 | 0.084156 | 0.004378 |  0.0 | 0.000617 |  0.0 | 0.084156 | 0.177557 | 0.049322 | 0.653746 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 9035.995972 | 134.805202 | 0.084156 |   0.0 | 0.004378 |   0.0 | 0.000617 | 0.175868 |   0.0 | 0.001688 |   0.0 | 0.000002 |   0.0 | 0.049322 |
    | 1997-08-05 22:00:00 | 0.833333 | 18.3 |              0.0 |                      0.0 |             0.0 |             76.7 |       2.2 |              1018.0 | 0.0 | 0.0 | 18.333333 |                 75.25 |                  13.9 |                     15.438712 |           286.953669 | 0.783333 | 18.3 | 18.333333 |    1.763501 |         2.418134 |           1.763501 |          2.2 |                21.065302 |                     21.109404 |                         0.0 |                          0.0 |                      1.321845 |                           1.324263 |            16.157086 |                 15.884826 |    1001.842914 |   1.209486 | -2.9 | 19.365285 |    6.39171 |                   0.0 |                      0.0 |                       0.0 |                 218.084788 |                 39.670785 |                     0.0 |                      0.0 |   -39.670785 |             0.0 |              0.0 |        178.414003 |             64.197103 |                 100.0 |                 48.85611 |               87.719298 |  0.73708 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.042508 |  0.042508 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.73708 | 0.091597 | 0.004398 |  0.0 | 0.000642 |  0.0 | 0.091597 |  0.17532 |   0.0487 | 0.657492 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 9026.704262 | 135.445645 | 0.091597 |   0.0 | 0.004398 |   0.0 | 0.000642 | 0.173595 |   0.0 | 0.001723 |   0.0 | 0.000002 |   0.0 |   0.0487 |
    | 1997-08-05 23:00:00 | 0.833333 | 17.4 |              0.0 |                      0.0 |             0.0 |             81.8 |       2.7 |              1018.0 | 0.0 | 0.0 | 18.470833 |               74.9375 |                  13.9 |                     15.438712 |           286.953669 | 0.783333 | 17.4 | 18.470833 |    2.164296 |         2.438173 |           2.164296 |          2.7 |                19.904589 |                     21.292177 |                         0.0 |                          0.0 |                      1.257963 |                           1.334275 |            16.281953 |                 15.955825 |    1001.718047 |   1.213176 | -2.9 | 19.253784 |  11.122707 |                   0.0 |                      0.0 |                       0.0 |                 218.084788 |                 39.564038 |                     0.0 |                      0.0 |   -39.564038 |             0.0 |              0.0 |        178.520751 |             52.308751 |                 100.0 |                 48.85611 |               87.719298 | 0.740825 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.035577 |  0.035577 |      0.0 | 0.0 |   0.0 |          nan |         0.24 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 | 0.740825 | 0.092616 | 0.004418 |  0.0 | 0.000669 |  0.0 | 0.092616 |  0.17325 | 0.048125 | 0.664423 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 9012.681555 | 136.088768 | 0.092616 |   0.0 | 0.004418 |   0.0 | 0.000669 |  0.17149 |   0.0 | 0.001758 |   0.0 | 0.000002 |   0.0 | 0.048125 |

>>> round_(model.check_waterbalance(conditions_acker_heavy_rain))
0.0

.. _lland_v4_water_hourly:

water
-----

.. integration-test::

    >>> lnk(WASSER)
    >>> inputs.nied.series = 0.05
    >>> test("lland_v4_water_hourly",
    ...      axis1=(fluxes.nkor, fluxes.evpo, fluxes.evi, fluxes.qah),
    ...      use_conditions=conditions_wasser)
    |                date | nied | teml | sunshineduration | possiblesunshineduration | globalradiation | relativehumidity | windspeed | atmosphericpressure |  qz | qzh |   temltag | dailyrelativehumidity | dailysunshineduration | dailypossiblesunshineduration | dailyglobalradiation |  nkor | tkor |   tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair |   g |  tz |  wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |     evpo |      evi | evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatinz | wlatsnow | wsensinz | wsenssnow | wsurfinz | wsurf | sff | fvg | wada | qdb | qib1 | qib2 | qbb | qkap | qdgz |      qah |       qa | inzp | stinz | sinz | esnowinz | asinz | wats | waes | esnow | taus | ebdn | bowa | qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz | qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    -----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-08-03 00:00:00 | 0.05 | 16.9 |              0.0 |                      0.0 |             0.0 |             95.1 |       0.8 |              1015.0 | 0.0 | 0.0 | 17.279167 |             85.883333 |                   1.3 |                          15.7 |           136.579167 | 0.047 | 16.9 | 17.279167 |    0.641273 |         1.605886 |           0.641273 |          0.8 |                19.284227 |                     19.753091 |                         0.0 |                          0.0 |                      1.223615 |                           1.249589 |              18.3393 |                 16.964613 |       996.6607 |    1.21073 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                   40.97375 |                 10.408237 |                     0.0 |                      0.0 |   -10.408237 |             0.0 |              0.0 |         30.565513 |            266.257083 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.040545 | 0.040545 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.006455 | 0.001793 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.001793 |
    | 1997-08-03 01:00:00 | 0.05 | 16.6 |              0.0 |                      0.0 |             0.0 |             94.9 |       0.8 |              1015.0 | 0.0 | 0.0 | 17.241667 |             85.991667 |                   1.3 |                          15.7 |           136.579167 | 0.047 | 16.6 | 17.241667 |    0.641273 |         1.603439 |           0.641273 |          0.8 |                18.920184 |                      19.70628 |                         0.0 |                          0.0 |                       1.20339 |                           1.246999 |            17.955254 |                 16.945759 |     997.044746 |   1.212158 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                   40.97375 |                  10.41572 |                     0.0 |                      0.0 |    -10.41572 |             0.0 |              0.0 |          30.55803 |            266.257083 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.04041 |  0.04041 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.00659 | 0.001831 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.001831 |
    | 1997-08-03 02:00:00 | 0.05 | 16.4 |              0.0 |                      0.0 |             0.0 |             95.9 |       0.8 |              1015.0 | 0.0 | 0.0 |    17.175 |             86.233333 |                   1.3 |                          15.7 |           136.579167 | 0.047 | 16.4 |    17.175 |    0.641273 |         1.575992 |           0.641273 |          0.8 |                 18.68084 |                       19.6233 |                         0.0 |                          0.0 |                      1.190065 |                           1.242407 |            17.914926 |                 16.921826 |     997.085074 |   1.213014 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                   40.97375 |                 10.421627 |                     0.0 |                      0.0 |   -10.421627 |             0.0 |              0.0 |         30.552123 |            266.257083 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.040055 | 0.040055 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.006945 | 0.001929 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.001929 |
    | 1997-08-03 03:00:00 | 0.05 | 16.3 |              0.0 |                      0.0 |             0.0 |             96.7 |       0.8 |              1015.0 | 0.0 | 0.0 |   17.0625 |             86.708333 |                   1.3 |                          15.5 |           136.579167 | 0.047 | 16.3 |   17.0625 |    0.641273 |         1.548545 |           0.641273 |          0.8 |                18.562165 |                     19.483964 |                         0.0 |                          0.0 |                      1.183449 |                            1.23469 |            17.949613 |                  16.89422 |     997.050387 |   1.213417 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                   40.97375 |                 10.455184 |                     0.0 |                      0.0 |   -10.455184 |             0.0 |              0.0 |         30.518566 |            266.257083 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.039459 | 0.039459 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.007541 | 0.002095 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.002095 |
    | 1997-08-03 04:00:00 | 0.05 | 16.0 |              0.0 |                 0.429734 |        1.943686 |             97.2 |       0.8 |              1015.0 | 0.0 | 0.0 | 16.908333 |             87.366667 |                   1.3 |                     14.929734 |            135.50182 | 0.047 | 16.0 | 16.908333 |    0.641273 |         1.504432 |           0.641273 |          0.8 |                18.210086 |                     19.294427 |                         0.0 |                          0.0 |                      1.163788 |                           1.224181 |            17.700204 |                 16.856897 |     997.299796 |    1.21479 | 0.0 | nan | 0.0 |              0.583106 |                      0.0 |                  0.583106 |                  40.650546 |                 10.555242 |                     0.0 |                      0.0 |    -9.972137 |             0.0 |              0.0 |         30.095304 |            266.257083 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.038271 | 0.038271 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.008729 | 0.002425 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.002425 |
    | 1997-08-03 05:00:00 | 0.05 | 15.9 |              0.0 |                      1.0 |       21.932442 |             97.5 |       0.6 |              1015.0 | 0.0 | 0.0 | 16.729167 |             88.204167 |                   1.3 |                     14.929734 |           134.099005 | 0.047 | 15.9 | 16.729167 |    0.480955 |         1.453638 |           0.480955 |          0.6 |                18.094032 |                     19.076181 |                         0.0 |                          0.0 |                      1.157296 |                           1.212063 |            17.641681 |                 16.825987 |     997.358319 |   1.215237 | 0.0 | nan | 0.0 |              6.579733 |                      0.0 |                  6.579733 |                  40.229702 |                 10.545193 |                     0.0 |                      0.0 |    -3.965461 |             0.0 |              0.0 |         29.684509 |            355.009444 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.036944 | 0.036944 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.010056 | 0.002793 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.002793 |
    | 1997-08-03 06:00:00 | 0.05 | 16.0 |              0.0 |                      1.0 |       57.256186 |             97.7 |       0.9 |              1015.0 | 0.0 | 0.0 | 16.533333 |             89.191667 |                   1.3 |                     14.929734 |           131.855513 | 0.047 | 16.0 | 16.533333 |    0.721432 |         1.392031 |           0.721432 |          0.9 |                18.210086 |                     18.840106 |                         0.0 |                          0.0 |                      1.163788 |                           1.198935 |            17.791254 |                 16.803805 |     997.208746 |   1.214748 | 0.0 | nan | 0.0 |             17.176856 |                      0.0 |                 17.176856 |                  39.556654 |                  10.52516 |                     0.0 |                      0.0 |     6.651696 |             0.0 |              0.0 |         29.031494 |            236.672963 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.035268 | 0.035268 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.011732 | 0.003259 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.003259 |
    | 1997-08-03 07:00:00 | 0.05 | 16.6 |              0.0 |                      1.0 |      109.332844 |             97.4 |       0.9 |              1015.0 | 0.0 | 0.0 |    16.375 |                90.125 |                   1.3 |                     14.929734 |           129.465215 | 0.047 | 16.6 |    16.375 |    0.721432 |         1.334591 |           0.721432 |          0.9 |                18.920184 |                     18.651109 |                         0.0 |                          0.0 |                       1.20339 |                           1.188408 |            18.428259 |                 16.809312 |     996.571741 |   1.211943 | 0.0 | nan | 0.0 |             32.799853 |                      0.0 |                 32.799853 |                  38.839564 |                  10.49073 |                     0.0 |                      0.0 |    22.309124 |             0.0 |              0.0 |         28.348835 |            236.672963 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.033685 | 0.033685 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.013315 | 0.003699 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.003699 |
    | 1997-08-03 08:00:00 | 0.05 | 17.4 |              0.0 |                      1.0 |      170.949153 |             96.8 |       0.9 |              1016.0 | 0.0 | 0.0 | 16.216667 |                91.275 |                   1.3 |                     14.929734 |           127.329763 | 0.047 | 17.4 | 16.216667 |    0.721432 |         1.260484 |           0.721432 |          0.9 |                19.904589 |                     18.463773 |                         0.0 |                          0.0 |                      1.257963 |                           1.177959 |            19.267642 |                 16.852809 |     996.732358 |   1.209425 | 0.0 | nan | 0.0 |             51.284746 |                      0.0 |                 51.284746 |                  38.198929 |                 10.426918 |                     0.0 |                      0.0 |    40.857828 |             0.0 |              0.0 |         27.772011 |            236.672963 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.032054 | 0.032054 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.014946 | 0.004152 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.004152 |
    | 1997-08-03 09:00:00 | 0.05 | 19.0 |              0.2 |                      1.0 |      311.762625 |             86.1 |       1.3 |              1016.0 | 0.0 | 0.0 |   16.1125 |             91.991667 |                   1.5 |                     14.929734 |           129.903206 | 0.047 | 19.0 |   16.1125 |    1.042069 |         1.203904 |           1.042069 |          1.3 |                22.008543 |                     18.341425 |                         0.0 |                          0.0 |                      1.373407 |                           1.171128 |            18.949356 |                 16.872582 |     997.050644 |   1.202945 | 0.0 | nan | 0.0 |             93.528788 |                      0.0 |                 93.528788 |                  38.970962 |                 10.804921 |                     0.0 |                      0.0 |    82.723866 |             0.0 |              0.0 |         28.166041 |            163.850513 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.031746 | 0.031746 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.015254 | 0.004237 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.004237 |
    | 1997-08-03 10:00:00 | 0.05 | 20.3 |              0.5 |                      1.0 |        501.5833 |             76.8 |       1.5 |              1016.0 | 0.0 | 0.0 | 16.083333 |             92.241667 |                   1.7 |                     14.929734 |           132.285843 | 0.047 | 20.3 | 16.083333 |    1.202387 |         1.149836 |           1.202387 |          1.5 |                23.858503 |                     18.307295 |                         0.0 |                          0.0 |                      1.473678 |                           1.169221 |             18.32333 |                 16.886954 |      997.67667 |   1.197896 | 0.0 | nan | 0.0 |             150.47499 |                      0.0 |                 150.47499 |                  39.685753 |                 11.199993 |                     0.0 |                      0.0 |   139.274997 |             0.0 |              0.0 |          28.48576 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.031752 | 0.031752 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.015248 | 0.004236 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.004236 |
    | 1997-08-03 11:00:00 | 0.05 | 21.4 |              0.7 |                      1.0 |      615.018728 |             71.8 |       1.2 |              1016.0 | 0.0 | 0.0 |    16.125 |             92.104167 |                   2.2 |                     14.929734 |           135.919957 | 0.047 | 21.4 |    16.125 |    0.961909 |         1.089916 |           0.961909 |          1.2 |                25.528421 |                     18.356069 |                         0.0 |                          0.0 |                      1.563281 |                           1.171946 |            18.329406 |                 16.906704 |     997.670594 |    1.19342 | 0.0 | nan | 0.0 |            184.505618 |                      0.0 |                184.505618 |                  40.775987 |                 12.222123 |                     0.0 |                      0.0 |   172.283495 |             0.0 |              0.0 |         28.553864 |            177.504722 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.031812 | 0.031812 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.015188 | 0.004219 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.004219 |
    | 1997-08-03 12:00:00 | 0.05 | 21.3 |              0.8 |                      1.0 |      626.544325 |             67.5 |       1.3 |              1016.0 | 0.0 | 0.0 | 16.204167 |             91.729167 |                   2.8 |                     14.929734 |           145.821804 | 0.047 | 21.3 | 16.204167 |    1.042069 |         1.037502 |           1.042069 |          1.3 |                 25.37251 |                     18.449053 |                         0.0 |                          0.0 |                       1.55495 |                           1.177138 |            17.126444 |                 16.923163 |     998.873556 |   1.194363 | 0.0 | nan | 0.0 |            187.963298 |                      0.0 |                187.963298 |                  43.746541 |                 13.461134 |                     0.0 |                      0.0 |   174.502163 |             0.0 |              0.0 |         30.285407 |            163.850513 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.033619 | 0.033619 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.013381 | 0.003717 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.003717 |
    | 1997-08-03 13:00:00 | 0.05 | 21.8 |              0.5 |                      1.0 |      496.133417 |             66.1 |       1.5 |              1016.0 | 0.0 | 0.0 | 16.329167 |             91.104167 |                   3.2 |                     14.929734 |           152.606529 | 0.047 | 21.8 | 16.329167 |    1.202387 |         1.012602 |           1.202387 |          1.5 |                26.160453 |                      18.59671 |                         0.0 |                          0.0 |                      1.596982 |                           1.185375 |             17.29206 |                 16.942378 |      998.70794 |   1.192265 | 0.0 | nan | 0.0 |            148.840025 |                      0.0 |                148.840025 |                  45.781959 |                 14.297428 |                     0.0 |                      0.0 |   134.542597 |             0.0 |              0.0 |         31.484531 |            142.003778 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.035172 | 0.035172 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.011828 | 0.003286 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.003286 |
    | 1997-08-03 14:00:00 | 0.05 | 22.9 |              0.4 |                      1.0 |      419.520994 |             63.4 |       1.9 |              1016.0 | 0.0 | 0.0 | 16.545833 |             90.058333 |                   3.4 |                     14.929734 |           157.353237 | 0.047 | 22.9 | 16.545833 |    1.523023 |         0.984394 |           1.523023 |          1.9 |                27.969419 |                     18.855098 |                         0.0 |                          0.0 |                      1.692831 |                           1.199769 |            17.732611 |                 16.980587 |     998.267389 |   1.187639 | 0.0 | nan | 0.0 |            125.856298 |                      0.0 |                125.856298 |                  47.205971 |                 14.726001 |                     0.0 |                      0.0 |   111.130298 |             0.0 |              0.0 |         32.479971 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.036854 | 0.036854 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.010146 | 0.002818 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.002818 |
    | 1997-08-03 15:00:00 | 0.05 | 22.7 |              0.5 |                      1.0 |      387.887353 |             62.4 |       1.9 |              1016.0 | 0.0 | 0.0 | 16.816667 |             88.816667 |                   3.8 |                     14.929734 |           164.256877 | 0.047 | 22.7 | 16.816667 |    1.523023 |         0.968687 |           1.523023 |          1.9 |                27.632633 |                     19.182495 |                         0.0 |                          0.0 |                      1.675052 |                           1.217969 |            17.242763 |                 17.037252 |     998.757237 |    1.18866 | 0.0 | nan | 0.0 |            116.366206 |                      0.0 |                116.366206 |                  49.277063 |                 15.562805 |                     0.0 |                      0.0 |     100.8034 |             0.0 |              0.0 |         33.714258 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.038963 | 0.038963 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.008037 | 0.002233 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.002233 |
    | 1997-08-03 16:00:00 | 0.05 | 22.5 |              0.5 |                      1.0 |      278.496872 |             61.1 |       2.3 |              1016.0 | 0.0 | 0.0 |   17.1375 |             87.333333 |                   4.1 |                     14.929734 |           170.073414 | 0.047 | 22.5 |   17.1375 |     1.84366 |         0.991339 |            1.84366 |          2.3 |                27.299387 |                     19.576758 |                         0.0 |                          0.0 |                      1.657431 |                            1.23983 |            16.679926 |                 17.097035 |     999.320074 |   1.189715 | 0.0 | nan | 0.0 |             83.549062 |                      0.0 |                 83.549062 |                  51.022024 |                 16.204965 |                     0.0 |                      0.0 |    67.344096 |             0.0 |              0.0 |         34.817059 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.041278 | 0.041278 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.005722 |  0.00159 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.00159 |
    | 1997-08-03 17:00:00 | 0.05 | 21.9 |              0.3 |                      1.0 |      137.138608 |             62.1 |       2.4 |              1016.0 | 0.0 | 0.0 |   17.4875 |                85.875 |                   4.4 |                     14.929734 |           173.470856 | 0.047 | 21.9 |   17.4875 |    1.923819 |         1.017332 |           1.923819 |          2.4 |                26.320577 |                     20.014927 |                         0.0 |                          0.0 |                      1.605502 |                           1.264057 |            16.345078 |                 17.187818 |     999.654922 |   1.192283 | 0.0 | nan | 0.0 |             41.141582 |                      0.0 |                 41.141582 |                  52.041257 |                 16.818923 |                     0.0 |                      0.0 |     24.32266 |             0.0 |              0.0 |         35.222334 |             88.752361 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.04299 |  0.04299 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.00401 | 0.001114 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.001114 |
    | 1997-08-03 18:00:00 | 0.05 | 21.4 |              0.1 |                      1.0 |       51.080714 |             67.0 |       2.5 |              1016.0 | 0.0 | 0.0 |   17.8375 |             84.620833 |                   4.5 |                     14.929734 |           174.440885 | 0.047 | 21.4 |   17.8375 |    2.003978 |         1.054998 |           2.003978 |          2.5 |                25.528421 |                     20.461645 |                         0.0 |                          0.0 |                      1.563281 |                           1.288683 |            17.104042 |                 17.314814 |     998.895958 |   1.193968 | 0.0 | nan | 0.0 |             15.324214 |                      0.0 |                 15.324214 |                  52.332266 |                 16.973018 |                     0.0 |                      0.0 |    -1.648804 |             0.0 |              0.0 |         35.359248 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.0444 |   0.0444 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |   0.0026 | 0.000722 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000722 |
    | 1997-08-03 19:00:00 | 0.05 | 20.7 |              0.0 |                      1.0 |       13.632817 |             74.5 |       2.5 |              1016.0 | 0.0 | 0.0 | 18.170833 |                  83.7 |                   4.5 |                     15.429734 |           175.008919 | 0.047 | 20.7 | 18.170833 |    2.003978 |         1.100997 |           2.003978 |          2.5 |                24.454368 |                     20.895167 |                         0.0 |                          0.0 |                      1.505746 |                           1.312512 |            18.218504 |                 17.489255 |     997.781496 |   1.196313 | 0.0 | nan | 0.0 |              4.089845 |                      0.0 |                  4.089845 |                  52.502676 |                 16.556776 |                     0.0 |                      0.0 |   -12.466931 |             0.0 |              0.0 |           35.9459 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.046089 | 0.046089 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.000911 | 0.000253 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000253 |
    | 1997-08-03 20:00:00 | 0.05 | 19.4 |              0.0 |                   0.1364 |        0.185944 |             81.2 |       2.2 |              1016.0 | 0.0 | 0.0 | 18.454167 |             83.066667 |                   4.5 |                     15.566134 |           175.016667 | 0.047 | 19.4 | 18.454167 |    1.763501 |         1.120309 |           1.763501 |          2.2 |                22.563931 |                      21.26995 |                         0.0 |                          0.0 |                      1.403627 |                           1.333058 |            18.321912 |                 17.668238 |     997.678088 |   1.201582 | 0.0 | nan | 0.0 |              0.055783 |                      0.0 |                  0.055783 |                     52.505 |                 16.343497 |                     0.0 |                      0.0 |   -16.287714 |             0.0 |              0.0 |         36.161503 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.047084 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-03 21:00:00 | 0.05 | 17.8 |              0.0 |                      0.0 |             0.0 |             86.9 |       1.7 |              1016.0 | 0.0 | 0.0 |     18.65 |               82.7375 |                   4.5 |                     15.566134 |           175.016667 | 0.047 | 17.8 |     18.65 |    1.362705 |         1.127089 |           1.362705 |          1.7 |                20.413369 |                     21.532411 |                         0.0 |                          0.0 |                      1.286025 |                           1.347418 |            17.739217 |                 17.815378 |     998.260783 |   1.208454 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                     52.505 |                 16.221809 |                     0.0 |                      0.0 |   -16.221809 |             0.0 |              0.0 |         36.283192 |            125.297451 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.047647 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-03 22:00:00 | 0.05 | 17.0 |              0.0 |                      0.0 |             0.0 |             90.1 |       1.7 |              1017.0 | 0.0 | 0.0 | 18.808333 |             82.554167 |                   4.5 |                     15.566134 |           175.016667 | 0.047 | 17.0 | 18.808333 |    1.362705 |         1.142201 |           1.362705 |          1.7 |                19.406929 |                     21.746678 |                         0.0 |                          0.0 |                      1.230421 |                           1.359123 |            17.485643 |                 17.952788 |     999.514357 |   1.213101 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                     52.505 |                 16.101364 |                     0.0 |                      0.0 |   -16.101364 |             0.0 |              0.0 |         36.403636 |            125.297451 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.048116 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-03 23:00:00 | 0.05 | 16.4 |              0.0 |                      0.0 |             0.0 |             90.9 |       2.3 |              1017.0 | 0.0 | 0.0 | 18.941667 |             82.379167 |                   4.5 |                     15.566134 |           175.016667 | 0.047 | 16.4 | 18.941667 |     1.84366 |         1.185687 |            1.84366 |          2.3 |                 18.68084 |                     21.928555 |                         0.0 |                          0.0 |                      1.190065 |                           1.369047 |            16.980884 |                 18.064561 |    1000.019116 |   1.215845 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                     52.505 |                 16.005095 |                     0.0 |                      0.0 |   -16.005095 |             0.0 |              0.0 |         36.499905 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.048674 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 00:00:00 | 0.05 | 15.9 |              0.0 |                      0.0 |             0.0 |             88.0 |       2.3 |              1017.0 | 0.0 | 0.0 |      18.9 |             82.083333 |                   4.5 |                     15.566134 |           175.016667 | 0.047 | 15.9 |      18.9 |     1.84366 |         1.235786 |            1.84366 |          2.3 |                18.094032 |                     21.871576 |                         0.0 |                          0.0 |                      1.157296 |                           1.365939 |            15.922748 |                 17.952919 |    1001.077252 |    1.21843 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                     52.505 |                 16.128293 |                     0.0 |                      0.0 |   -16.128293 |             0.0 |              0.0 |         36.376707 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.048972 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 01:00:00 | 0.05 | 15.4 |              0.0 |                      0.0 |             0.0 |             89.2 |       2.2 |              1016.0 | 0.0 | 0.0 |     18.85 |             81.845833 |                   4.5 |                     15.566134 |           175.016667 | 0.047 | 15.4 |     18.85 |    1.763501 |         1.282546 |           1.763501 |          2.2 |                17.523416 |                     21.803372 |                         0.0 |                          0.0 |                      1.125295 |                           1.362218 |            15.630887 |                 17.845152 |    1000.369113 |   1.219467 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                     52.505 |                 16.244859 |                     0.0 |                      0.0 |   -16.244859 |             0.0 |              0.0 |         36.260141 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.049217 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 02:00:00 | 0.05 | 15.2 |              0.0 |                      0.0 |             0.0 |             88.8 |       2.3 |              1016.0 | 0.0 | 0.0 |      18.8 |                 81.55 |                   4.5 |                     15.566134 |           175.016667 | 0.047 | 15.2 |      18.8 |     1.84366 |         1.332645 |            1.84366 |          2.3 |                17.299618 |                     21.735354 |                         0.0 |                          0.0 |                      1.112706 |                           1.358505 |            15.362061 |                 17.725181 |    1000.637939 |   1.220436 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                     52.505 |                 16.376926 |                     0.0 |                      0.0 |   -16.376926 |             0.0 |              0.0 |         36.128074 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.049512 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 03:00:00 | 0.05 | 14.9 |              0.0 |                      0.0 |             0.0 |             89.7 |       2.0 |              1016.0 | 0.0 | 0.0 | 18.741667 |             81.258333 |                   4.5 |                     15.566134 |           175.016667 | 0.047 | 14.9 | 18.741667 |    1.603182 |         1.372725 |           1.603182 |          2.0 |                16.968612 |                     21.656234 |                         0.0 |                          0.0 |                      1.094046 |                           1.354184 |            15.220845 |                 17.597495 |    1000.779155 |   1.221771 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                     52.505 |                 16.516685 |                     0.0 |                      0.0 |   -16.516685 |             0.0 |              0.0 |         35.988315 |            106.502833 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.049743 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 04:00:00 | 0.05 | 14.3 |              0.0 |                 0.398044 |        1.590517 |             93.0 |       2.3 |              1016.0 | 0.0 | 0.0 | 18.670833 |             81.083333 |                   4.5 |                     15.534444 |           175.001952 | 0.047 | 14.3 | 18.670833 |     1.84366 |         1.422824 |            1.84366 |          2.3 |                16.323194 |                     21.560498 |                         0.0 |                          0.0 |                      1.057522 |                           1.348953 |             15.18057 |                  17.48197 |     1000.81943 |    1.22434 | 0.0 | nan | 0.0 |              0.477155 |                      0.0 |                  0.477155 |                  52.500585 |                  16.65652 |                     0.0 |                      0.0 |   -16.179365 |             0.0 |              0.0 |         35.844065 |             92.611159 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.049925 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 05:00:00 | 0.05 | 14.1 |              0.0 |                      1.0 |       19.950903 |             93.6 |       2.5 |              1017.0 | 0.0 | 0.0 | 18.595833 |             80.920833 |                   4.5 |                     15.534444 |           174.919387 | 0.047 | 14.1 | 18.595833 |    2.003978 |         1.486284 |           2.003978 |          2.5 |                16.112886 |                     21.459534 |                         0.0 |                          0.0 |                      1.045579 |                           1.343433 |            15.081661 |                 17.365233 |    1001.918339 |   1.226451 | 0.0 | nan | 0.0 |              5.985271 |                      0.0 |                  5.985271 |                  52.475816 |                 16.779096 |                     0.0 |                      0.0 |   -10.793826 |             0.0 |              0.0 |          35.69672 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.050164 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 06:00:00 | 0.05 | 14.3 |              0.1 |                      1.0 |       62.985775 |             94.6 |       2.4 |              1017.0 | 0.0 | 0.0 |    18.525 |             80.791667 |                   4.6 |                     15.534444 |            175.15812 | 0.047 | 14.3 |    18.525 |    1.923819 |         1.536383 |           1.923819 |          2.4 |                16.323194 |                     21.364558 |                         0.0 |                          0.0 |                      1.057522 |                           1.338237 |            15.441741 |                 17.260782 |    1001.558259 |   1.225432 | 0.0 | nan | 0.0 |             18.895733 |                      0.0 |                 18.895733 |                  52.547436 |                 17.089561 |                     0.0 |                      0.0 |     1.806172 |             0.0 |              0.0 |         35.457875 |             88.752361 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.050224 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 07:00:00 | 0.05 | 15.4 |              0.9 |                      1.0 |      285.776389 |             93.3 |       2.5 |              1017.0 | 0.0 | 0.0 |    18.475 |             80.620833 |                   5.5 |                     15.534444 |           182.509935 | 0.047 | 15.4 |    18.475 |    2.003978 |         1.589823 |           2.003978 |          2.5 |                17.523416 |                     21.297737 |                         0.0 |                          0.0 |                      1.125295 |                            1.33458 |            16.349347 |                 17.170413 |    1000.650653 |   1.220347 | 0.0 | nan | 0.0 |             85.732917 |                      0.0 |                 85.732917 |                   54.75298 |                 19.012403 |                     0.0 |                      0.0 |    66.720514 |             0.0 |              0.0 |         35.740578 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.050871 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 08:00:00 | 0.05 | 16.8 |              1.0 |                      1.0 |      475.225247 |             86.9 |       2.5 |              1017.0 | 0.0 | 0.0 |     18.45 |             80.208333 |                   6.5 |                     15.534444 |           195.188105 | 0.047 | 16.8 |     18.45 |    2.003978 |         1.643262 |           2.003978 |          2.5 |                19.162205 |                     21.264396 |                         0.0 |                          0.0 |                      1.216842 |                           1.332754 |            16.651956 |                 17.055817 |    1000.348044 |   1.214317 | 0.0 | nan | 0.0 |            142.567574 |                      0.0 |                142.567574 |                  58.556432 |                 21.208047 |                     0.0 |                      0.0 |   121.359527 |             0.0 |              0.0 |         37.348385 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.053048 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 09:00:00 | 0.05 | 18.2 |              1.0 |                      1.0 |      614.006719 |             78.9 |       2.9 |              1017.0 | 0.0 | 0.0 | 18.416667 |             79.908333 |                   7.3 |                     15.534444 |           207.781609 | 0.047 | 18.2 | 18.416667 |    2.324615 |         1.696701 |           2.324615 |          2.9 |                20.933479 |                     21.220011 |                         0.0 |                          0.0 |                      1.314615 |                           1.330323 |            16.516515 |                 16.956557 |    1000.483485 |   1.208543 | 0.0 | nan | 0.0 |            184.202016 |                      0.0 |                184.202016 |                  62.334483 |                 22.996116 |                     0.0 |                      0.0 |     161.2059 |             0.0 |              0.0 |         39.338367 |              73.45023 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.055504 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 10:00:00 | 0.05 | 19.5 |              0.9 |                      1.0 |      643.806508 |             75.8 |       3.1 |              1017.0 | 0.0 | 0.0 | 18.383333 |             79.866667 |                   7.7 |                     15.534444 |           213.707576 | 0.047 | 19.5 | 18.383333 |    2.484933 |         1.750141 |           2.484933 |          3.1 |                22.704676 |                     21.175708 |                         0.0 |                          0.0 |                      1.411269 |                           1.327896 |            17.210144 |                 16.912332 |     999.789856 |   1.202862 | 0.0 | nan | 0.0 |            193.141952 |                      0.0 |                193.141952 |                  64.112273 |                 23.882018 |                     0.0 |                      0.0 |   169.259935 |             0.0 |              0.0 |         40.230255 |             68.711505 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.05667 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 11:00:00 | 0.05 | 20.3 |              0.8 |                      1.0 |      600.262981 |             73.4 |       3.3 |              1018.0 | 0.0 | 0.0 |   18.3375 |             79.933333 |                   7.8 |                     15.534444 |           213.092753 | 0.047 | 20.3 |   18.3375 |    2.645251 |          1.82028 |           2.645251 |          3.3 |                23.858503 |                     21.114922 |                         0.0 |                          0.0 |                      1.473678 |                           1.324565 |            17.512141 |                 16.877861 |    1000.487859 |   1.200635 | 0.0 | nan | 0.0 |            180.078894 |                      0.0 |                180.078894 |                  63.927826 |                   24.1295 |                     0.0 |                      0.0 |   155.949394 |             0.0 |              0.0 |         39.798326 |             64.547172 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.056522 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 12:00:00 | 0.05 | 21.0 |              0.9 |                      1.0 |      620.514139 |             68.3 |       3.3 |              1018.0 | 0.0 | 0.0 |    18.325 |             79.966667 |                   7.9 |                     15.534444 |           212.841496 | 0.047 | 21.0 |    18.325 |    2.645251 |         1.887079 |           2.645251 |          3.3 |                24.909749 |                     21.098371 |                         0.0 |                          0.0 |                      1.530181 |                           1.323658 |            17.013358 |                 16.871664 |    1000.986642 |   1.198001 | 0.0 | nan | 0.0 |            186.154242 |                      0.0 |                186.154242 |                  63.852449 |                 24.341655 |                     0.0 |                      0.0 |   161.812586 |             0.0 |              0.0 |         39.510794 |             64.547172 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.056563 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 13:00:00 | 0.05 | 21.8 |              0.8 |                      1.0 |      599.441386 |             61.3 |       3.2 |              1018.0 | 0.0 | 0.0 |    18.325 |             79.766667 |                   8.2 |                     15.534444 |           217.145994 | 0.047 | 21.8 |    18.325 |    2.565092 |         1.943859 |           2.565092 |          3.2 |                26.160453 |                     21.098371 |                         0.0 |                          0.0 |                      1.596982 |                           1.323658 |            16.036358 |                 16.829467 |    1001.963642 |   1.195188 | 0.0 | nan | 0.0 |            179.832416 |                      0.0 |                179.832416 |                  65.143798 |                 25.038979 |                     0.0 |                      0.0 |   154.793437 |             0.0 |              0.0 |          40.10482 |             66.564271 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.057637 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 14:00:00 | 0.05 | 21.9 |              0.9 |                      1.0 |      613.551222 |             60.6 |       2.5 |              1018.0 | 0.0 | 0.0 | 18.283333 |                 79.65 |                   8.7 |                     15.534444 |           225.230587 | 0.047 | 21.9 | 18.283333 |    2.003978 |         1.963899 |           2.003978 |          2.5 |                26.320577 |                     21.043281 |                         0.0 |                          0.0 |                      1.605502 |                           1.320638 |             15.95027 |                 16.760973 |     1002.04973 |   1.194821 | 0.0 | nan | 0.0 |            184.065367 |                      0.0 |                184.065367 |                  67.569176 |                 26.187015 |                     0.0 |                      0.0 |   157.878352 |             0.0 |              0.0 |         41.382161 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.059053 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 15:00:00 | 0.05 | 21.8 |              0.9 |                      1.0 |      517.107047 |             58.8 |       2.9 |              1017.0 | 0.0 | 0.0 | 18.245833 |                  79.5 |                   9.1 |                     15.534444 |           230.614741 | 0.047 | 21.8 | 18.245833 |    2.324615 |         1.997298 |           2.324615 |          2.9 |                26.160453 |                     20.993808 |                         0.0 |                          0.0 |                      1.596982 |                           1.317924 |            15.382347 |                 16.690077 |    1001.617653 |   1.194298 | 0.0 | nan | 0.0 |            155.132114 |                      0.0 |                155.132114 |                  69.184422 |                 27.142101 |                     0.0 |                      0.0 |   127.990013 |             0.0 |              0.0 |         42.042321 |              73.45023 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.059976 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 16:00:00 | 0.05 | 22.2 |              0.9 |                      1.0 |      377.836617 |             57.2 |       3.6 |              1017.0 | 0.0 | 0.0 | 18.233333 |               79.3375 |                   9.5 |                     15.534444 |           234.753897 | 0.047 | 22.2 | 18.233333 |    2.885728 |         2.040718 |           2.885728 |          3.6 |                26.806088 |                      20.97734 |                         0.0 |                          0.0 |                      1.631293 |                           1.317021 |            15.333082 |                 16.642897 |    1001.666918 |   1.192703 | 0.0 | nan | 0.0 |            113.350985 |                      0.0 |                113.350985 |                  70.426169 |                 28.068928 |                     0.0 |                      0.0 |    85.282057 |             0.0 |              0.0 |         42.357241 |             59.168241 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.060663 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 17:00:00 | 0.05 | 21.8 |              1.0 |                      1.0 |      252.204639 |             57.2 |       4.2 |              1017.0 | 0.0 | 0.0 | 18.229167 |             79.133333 |                  10.2 |                     15.534444 |           239.548315 | 0.047 | 21.8 | 18.229167 |    3.366683 |         2.100837 |           3.366683 |          4.2 |                26.160453 |                     20.971853 |                         0.0 |                          0.0 |                      1.596982 |                            1.31672 |            14.963779 |                 16.595726 |    1002.036221 |   1.194485 | 0.0 | nan | 0.0 |             75.661392 |                      0.0 |                 75.661392 |                  71.864495 |                 29.636995 |                     0.0 |                      0.0 |    46.024396 |             0.0 |              0.0 |         42.227499 |             50.715635 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.061064 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 18:00:00 | 0.05 | 20.9 |              1.0 |                      1.0 |      128.576714 |             59.9 |       4.2 |              1017.0 | 0.0 | 0.0 | 18.208333 |               78.8375 |                  11.1 |                     15.534444 |           242.777315 | 0.047 | 20.9 | 18.208333 |    3.366683 |         2.157616 |           3.366683 |          4.2 |                 24.75714 |                     20.944437 |                         0.0 |                          0.0 |                      1.521999 |                           1.315216 |            14.829527 |                  16.51207 |    1002.170473 |   1.198201 | 0.0 | nan | 0.0 |             38.573014 |                      0.0 |                 38.573014 |                  72.833195 |                 31.712576 |                     0.0 |                      0.0 |     6.860438 |             0.0 |              0.0 |         41.120619 |             50.715635 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.060566 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 19:00:00 | 0.05 | 19.7 |              1.0 |                      1.0 |       42.072586 |             66.1 |       3.6 |              1017.0 | 0.0 | 0.0 | 18.166667 |               78.4875 |                  12.1 |                     15.534444 |           243.962306 | 0.047 | 19.7 | 18.166667 |    2.885728 |         2.194356 |           2.885728 |          3.6 |                22.988466 |                     20.889699 |                         0.0 |                          0.0 |                      1.426658 |                           1.312212 |            15.195376 |                 16.395802 |    1001.804624 |   1.202947 | 0.0 | nan | 0.0 |             12.621776 |                      0.0 |                 12.621776 |                  73.188692 |                 34.091963 |                     0.0 |                      0.0 |   -21.470187 |             0.0 |              0.0 |         39.096729 |             59.168241 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.059092 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 20:00:00 | 0.05 | 17.9 |              0.3 |                 0.104711 |        0.812897 |             71.7 |       3.2 |              1017.0 | 0.0 | 0.0 | 18.104167 |             78.091667 |                  12.4 |                     15.502754 |           243.988429 | 0.047 | 17.9 | 18.104167 |    2.565092 |         2.227756 |           2.565092 |          3.2 |                20.542326 |                     20.807826 |                         0.0 |                          0.0 |                      1.293123 |                           1.307717 |            14.728848 |                 16.249178 |    1002.271152 |   1.210597 | 0.0 | nan | 0.0 |              0.243869 |                      0.0 |                  0.243869 |                  73.196529 |                 35.120093 |                     0.0 |                      0.0 |   -34.876224 |             0.0 |              0.0 |         38.076436 |             66.564271 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.058604 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 21:00:00 | 0.05 | 15.8 |              0.0 |                      0.0 |             0.0 |             80.6 |       2.2 |              1018.0 | 0.0 | 0.0 | 18.020833 |             77.829167 |                  12.4 |                     15.502754 |           243.988429 | 0.047 | 15.8 | 18.020833 |    1.763501 |         2.244455 |           1.763501 |          2.2 |                17.978626 |                     20.699099 |                         0.0 |                          0.0 |                      1.150835 |                           1.301743 |            14.490772 |                 16.109936 |    1003.509228 |    1.22071 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                  73.196529 |                 35.424509 |                     0.0 |                      0.0 |   -35.424509 |             0.0 |              0.0 |          37.77202 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.058562 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 22:00:00 | 0.05 | 14.8 |              0.0 |                      0.0 |             0.0 |             85.9 |       2.2 |              1018.0 | 0.0 | 0.0 | 17.929167 |             77.654167 |                  12.4 |                     15.502754 |           243.988429 | 0.047 | 14.8 | 17.929167 |    1.763501 |         2.261155 |           1.763501 |          2.2 |                16.859515 |                     20.580072 |                         0.0 |                          0.0 |                      1.087886 |                           1.295199 |            14.482324 |                 15.981284 |    1003.517676 |   1.224953 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                  73.196529 |                 35.698198 |                     0.0 |                      0.0 |   -35.698198 |             0.0 |              0.0 |          37.49833 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.058455 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-04 23:00:00 | 0.05 | 14.1 |              0.0 |                      0.0 |             0.0 |             89.3 |       2.1 |              1018.0 | 0.0 | 0.0 | 17.833333 |               77.5875 |                  12.4 |                     15.502754 |           243.988429 | 0.047 | 14.1 | 17.833333 |    1.683342 |         2.254475 |           1.683342 |          2.1 |                16.112886 |                     20.456276 |                         0.0 |                          0.0 |                      1.045579 |                           1.288387 |            14.388807 |                 15.871513 |    1003.611193 |   1.227981 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                  73.196529 |                 35.921753 |                     0.0 |                      0.0 |   -35.921753 |             0.0 |              0.0 |         37.274776 |             101.43127 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.058143 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 00:00:00 | 0.05 | 13.4 |              0.0 |                      0.0 |             0.0 |             91.4 |       1.8 |              1018.0 | 0.0 | 0.0 | 17.729167 |             77.729167 |                  12.4 |                     15.502754 |           243.988429 | 0.047 | 13.4 | 17.729167 |    1.442864 |         2.237776 |           1.442864 |          1.8 |                15.395377 |                     20.322453 |                         0.0 |                          0.0 |                      1.004677 |                           1.281018 |            14.071374 |                 15.796473 |    1003.928626 |   1.231127 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                  73.196529 |                 36.050541 |                     0.0 |                      0.0 |   -36.050541 |             0.0 |              0.0 |         37.145987 |            118.336481 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.057649 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 01:00:00 | 0.05 | 12.8 |              0.0 |                      0.0 |             0.0 |             93.1 |       1.8 |              1018.0 | 0.0 | 0.0 | 17.620833 |             77.891667 |                  12.4 |                     15.502754 |           243.988429 | 0.047 | 12.8 | 17.620833 |    1.442864 |         2.224416 |           1.442864 |          1.8 |                14.802811 |                     20.184089 |                         0.0 |                          0.0 |                      0.970706 |                           1.273391 |            13.781417 |                 15.721724 |    1004.218583 |   1.233843 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                  73.196529 |                 36.176259 |                     0.0 |                      0.0 |   -36.176259 |             0.0 |              0.0 |          37.02027 |            118.336481 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.057155 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 02:00:00 | 0.05 | 12.3 |              0.0 |                      0.0 |             0.0 |             94.6 |       1.6 |              1018.0 | 0.0 | 0.0 |      17.5 |             78.133333 |                  12.4 |                     15.502754 |           243.988429 | 0.047 | 12.3 |      17.5 |    1.282546 |         2.201036 |           1.282546 |          1.6 |                14.324376 |                     20.030733 |                         0.0 |                          0.0 |                      0.943147 |                            1.26493 |            13.550859 |                 15.650679 |    1004.449141 |   1.236111 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                  73.196529 |                 36.284611 |                     0.0 |                      0.0 |   -36.284611 |             0.0 |              0.0 |         36.911917 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.05653 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 03:00:00 | 0.05 | 12.0 |              0.0 |                      0.0 |             0.0 |             95.6 |       1.2 |              1018.0 | 0.0 | 0.0 | 17.379167 |             78.379167 |                  12.4 |                     15.502754 |           243.988429 | 0.047 | 12.0 | 17.379167 |    0.961909 |         2.174316 |           0.961909 |          1.2 |                 14.04387 |                     19.878396 |                         0.0 |                          0.0 |                      0.926933 |                           1.256516 |            13.425939 |                 15.580521 |    1004.574061 |   1.237469 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                  73.196529 |                 36.390891 |                     0.0 |                      0.0 |   -36.390891 |             0.0 |              0.0 |         36.805637 |            177.504722 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.05589 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 04:00:00 | 0.05 | 11.7 |              0.0 |                 0.366022 |        1.482347 |             96.2 |       1.9 |              1018.0 | 0.0 | 0.0 | 17.270833 |               78.5125 |                  12.4 |                     15.470733 |           243.983922 | 0.047 | 11.7 | 17.270833 |    1.523023 |         2.160956 |           1.523023 |          1.9 |                13.768192 |                      19.74268 |                         0.0 |                          0.0 |                      0.910957 |                           1.249013 |            13.245001 |                 15.500472 |    1004.754999 |   1.238856 | 0.0 | nan | 0.0 |              0.444704 |                      0.0 |                  0.444704 |                  73.195176 |                 36.589063 |                     0.0 |                      0.0 |   -36.144359 |             0.0 |              0.0 |         36.606113 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.055364 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 05:00:00 | 0.05 | 11.6 |              0.0 |                      1.0 |       21.135567 |             95.7 |       1.3 |              1018.0 | 0.0 | 0.0 | 17.166667 |                  78.6 |                  12.4 |                     15.470733 |           244.033283 | 0.047 | 11.6 | 17.166667 |    1.042069 |         2.120877 |           1.042069 |          1.3 |                 13.67736 |                     19.612949 |                         0.0 |                          0.0 |                      0.905683 |                           1.241834 |            13.089234 |                 15.415778 |    1004.910766 |   1.239363 | 0.0 | nan | 0.0 |               6.34067 |                      0.0 |                   6.34067 |                  73.209985 |                  36.74524 |                     0.0 |                      0.0 |    -30.40457 |             0.0 |              0.0 |         36.464745 |            163.850513 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.054792 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 06:00:00 | 0.05 | 12.6 |              0.6 |                      1.0 |       125.88345 |             95.1 |       1.6 |              1018.0 | 0.0 | 0.0 | 17.095833 |             78.620833 |                  12.9 |                     15.470733 |           246.654019 | 0.047 | 12.6 | 17.095833 |    1.282546 |         2.094157 |           1.282546 |          1.6 |                14.609782 |                     19.525158 |                         0.0 |                          0.0 |                      0.959602 |                           1.236973 |            13.893903 |                 15.350842 |    1004.106097 |   1.234655 | 0.0 | nan | 0.0 |             37.765035 |                      0.0 |                 37.765035 |                  73.996206 |                  38.00446 |                     0.0 |                      0.0 |    -0.239425 |             0.0 |              0.0 |         35.991746 |            133.128542 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.054084 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 07:00:00 | 0.05 | 14.8 |              0.9 |                      1.0 |      305.954944 |             96.3 |       1.9 |              1019.0 | 0.0 | 0.0 | 17.070833 |             78.745833 |                  12.9 |                     15.470733 |           247.494792 | 0.047 | 14.8 | 17.070833 |    1.523023 |         2.074117 |           1.523023 |          1.9 |                16.859515 |                     19.494255 |                         0.0 |                          0.0 |                      1.087886 |                           1.235261 |            16.235713 |                 15.350914 |    1002.764287 |   1.225361 | 0.0 | nan | 0.0 |             91.786483 |                      0.0 |                 91.786483 |                  74.248438 |                 37.987562 |                     0.0 |                      0.0 |    53.798921 |             0.0 |              0.0 |         36.260875 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.054083 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 08:00:00 | 0.05 | 17.1 |              1.0 |                      1.0 |      511.104575 |             89.9 |       3.2 |              1019.0 | 0.0 | 0.0 | 17.083333 |             78.870833 |                  12.9 |                     15.470733 |           248.989764 | 0.047 | 17.1 | 17.083333 |    2.565092 |         2.097497 |           2.565092 |          3.2 |                19.530313 |                     19.509701 |                         0.0 |                          0.0 |                      1.237258 |                           1.236116 |            17.557751 |                 15.387464 |    1001.442249 |   1.215051 | 0.0 | nan | 0.0 |            153.331373 |                      0.0 |                153.331373 |                  74.696929 |                 37.896217 |                     0.0 |                      0.0 |   115.435156 |             0.0 |              0.0 |         36.800712 |             66.564271 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.054631 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 09:00:00 | 0.05 | 19.4 |              0.9 |                      1.0 |      629.519239 |             79.2 |       4.0 |              1019.0 | 0.0 | 0.0 | 17.133333 |             78.883333 |                  12.8 |                     15.470733 |           249.636119 | 0.047 | 19.4 | 17.133333 |    3.206365 |         2.134237 |           3.206365 |          4.0 |                22.563931 |                     19.571593 |                         0.0 |                          0.0 |                      1.403627 |                           1.239544 |            17.870634 |                 15.438725 |    1001.129366 |   1.205358 | 0.0 | nan | 0.0 |            188.855772 |                      0.0 |                188.855772 |                  74.890836 |                  37.56462 |                     0.0 |                      0.0 |   151.291152 |             0.0 |              0.0 |         37.326216 |             53.251417 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.055387 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 10:00:00 | 0.05 | 20.6 |              1.0 |                      1.0 |      765.816589 |             73.4 |       4.6 |              1019.0 | 0.0 | 0.0 | 17.179167 |             78.783333 |                  12.9 |                     15.470733 |           254.719872 | 0.047 | 20.6 | 17.179167 |     3.68732 |         2.184336 |            3.68732 |          4.6 |                24.304197 |                     19.628477 |                         0.0 |                          0.0 |                      1.497674 |                           1.242694 |            17.839281 |                 15.463969 |    1001.160719 |   1.200448 | 0.0 | nan | 0.0 |            229.744977 |                      0.0 |                229.744977 |                  76.415962 |                 37.751868 |                     0.0 |                      0.0 |   191.993109 |             0.0 |              0.0 |         38.664094 |              46.30558 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.057099 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 11:00:00 | 0.05 | 21.4 |              1.0 |                      1.0 |      767.145842 |             67.3 |       4.1 |              1019.0 | 0.0 | 0.0 |    17.225 |             78.529167 |                  13.1 |                     15.470733 |           261.673325 | 0.047 | 21.4 |    17.225 |    3.286524 |         2.211056 |           3.286524 |          4.1 |                25.528421 |                     19.685506 |                         0.0 |                          0.0 |                      1.563281 |                            1.24585 |            17.180627 |                 15.458864 |    1001.819373 |   1.197482 | 0.0 | nan | 0.0 |            230.143753 |                      0.0 |                230.143753 |                  78.501998 |                 38.247049 |                     0.0 |                      0.0 |   191.896704 |             0.0 |              0.0 |         40.254949 |             51.952602 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.059074 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 12:00:00 | 0.05 | 22.5 |              1.0 |                      1.0 |      736.263853 |             62.2 |       4.1 |              1019.0 | 0.0 | 0.0 |   17.2875 |                78.275 |                  13.2 |                     15.470733 |            266.49623 | 0.047 | 22.5 |   17.2875 |    3.286524 |         2.237776 |           3.286524 |          4.1 |                27.299387 |                     19.763507 |                         0.0 |                          0.0 |                      1.657431 |                           1.250165 |            16.980219 |                 15.469885 |    1002.019781 |   1.193116 | 0.0 | nan | 0.0 |            220.879156 |                      0.0 |                220.879156 |                  79.948869 |                  38.48429 |                     0.0 |                      0.0 |   182.394866 |             0.0 |              0.0 |         41.464579 |             51.952602 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.060713 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 13:00:00 | 0.05 | 23.1 |              1.0 |                      1.0 |      768.006183 |             59.6 |       4.4 |              1019.0 | 0.0 | 0.0 | 17.341667 |             78.204167 |                  13.4 |                     15.470733 |           273.519763 | 0.047 | 23.1 | 17.341667 |    3.527001 |         2.277855 |           3.527001 |          4.4 |                28.309776 |                     19.831326 |                         0.0 |                          0.0 |                      1.710768 |                           1.253914 |            16.872626 |                 15.508923 |    1002.127374 |   1.190747 | 0.0 | nan | 0.0 |            230.401855 |                      0.0 |                230.401855 |                  82.055929 |                  38.86309 |                     0.0 |                      0.0 |   191.538765 |             0.0 |              0.0 |         43.192839 |             48.410379 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.062745 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 14:00:00 | 0.05 | 23.4 |              1.0 |                      1.0 |      735.929369 |             55.4 |       4.6 |              1019.0 | 0.0 | 0.0 | 17.404167 |               77.9875 |                  13.5 |                     15.470733 |           278.618852 | 0.047 | 23.4 | 17.404167 |     3.68732 |         2.347994 |            3.68732 |          4.6 |                28.827078 |                     19.909831 |                         0.0 |                          0.0 |                      1.737973 |                           1.258253 |            15.970201 |                 15.527179 |    1003.029799 |   1.189943 | 0.0 | nan | 0.0 |            220.778811 |                      0.0 |                220.778811 |                  83.585656 |                 39.079824 |                     0.0 |                      0.0 |   181.698987 |             0.0 |              0.0 |         44.505832 |              46.30558 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.064717 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 15:00:00 | 0.05 | 23.8 |              1.0 |                      1.0 |      620.374953 |             52.1 |       4.5 |              1019.0 | 0.0 | 0.0 |   17.4875 |             77.708333 |                  13.6 |                     15.470733 |           282.921682 | 0.047 | 23.8 |   17.4875 |    3.607161 |         2.401434 |           3.607161 |          4.5 |                29.529613 |                     20.014927 |                         0.0 |                          0.0 |                      1.774812 |                           1.264057 |            15.384928 |                 15.553266 |    1003.615072 |     1.1886 | 0.0 | nan | 0.0 |            186.112486 |                      0.0 |                186.112486 |                  84.876505 |                 39.288738 |                     0.0 |                      0.0 |   146.823747 |             0.0 |              0.0 |         45.587766 |             47.334593 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.066472 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 16:00:00 | 0.05 | 24.0 |              1.0 |                      1.0 |      450.095264 |             49.5 |       4.7 |              1018.0 | 0.0 | 0.0 |   17.5625 |               77.3875 |                  13.7 |                     15.470733 |           285.932459 | 0.047 | 24.0 |   17.5625 |    3.767479 |         2.438173 |           3.767479 |          4.7 |                29.886439 |                     20.109928 |                         0.0 |                          0.0 |                      1.793476 |                             1.2693 |            14.793787 |                  15.56257 |    1003.206213 |   1.186889 | 0.0 | nan | 0.0 |            135.028579 |                      0.0 |                135.028579 |                  85.779738 |                 39.539407 |                     0.0 |                      0.0 |    95.489172 |             0.0 |              0.0 |          46.24033 |             45.320355 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.067757 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 17:00:00 | 0.05 | 24.0 |              1.0 |                      1.0 |      273.614142 |             48.1 |       4.5 |              1018.0 | 0.0 | 0.0 | 17.654167 |             77.008333 |                  13.7 |                     15.470733 |           286.824521 | 0.047 | 24.0 | 17.654167 |    3.607161 |         2.448193 |           3.607161 |          4.5 |                29.886439 |                     20.226575 |                         0.0 |                          0.0 |                      1.793476 |                           1.275734 |            14.375377 |                 15.576148 |    1003.624623 |   1.187075 | 0.0 | nan | 0.0 |             82.084243 |                      0.0 |                 82.084243 |                  86.047356 |                 39.564612 |                     0.0 |                      0.0 |    42.519631 |             0.0 |              0.0 |         46.482744 |             47.334593 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.068576 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 18:00:00 | 0.05 | 23.7 |              1.0 |                      1.0 |      133.368106 |             49.4 |       4.1 |              1018.0 | 0.0 | 0.0 | 17.770833 |             76.570833 |                  13.7 |                     15.470733 |           287.024163 | 0.047 | 23.7 | 17.770833 |    3.286524 |         2.444853 |           3.286524 |          4.1 |                29.352596 |                      20.37589 |                         0.0 |                          0.0 |                      1.765541 |                           1.283961 |            14.500182 |                 15.601989 |    1003.499818 |   1.188219 | 0.0 | nan | 0.0 |             40.010432 |                      0.0 |                 40.010432 |                  86.107249 |                 39.572333 |                     0.0 |                      0.0 |     0.438099 |             0.0 |              0.0 |         46.534916 |             51.952602 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.069229 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 19:00:00 | 0.05 | 22.8 |              1.0 |                      1.0 |       40.326503 |             57.0 |       3.8 |              1018.0 | 0.0 | 0.0 |      17.9 |             76.191667 |                  13.7 |                     15.470733 |           286.951409 | 0.047 | 22.8 |      17.9 |    3.046047 |         2.451533 |           3.046047 |          3.8 |                27.800581 |                     20.542326 |                         0.0 |                          0.0 |                      1.683922 |                           1.293123 |            15.846331 |                 15.651541 |    1002.153669 |   1.191234 | 0.0 | nan | 0.0 |             12.097951 |                      0.0 |                 12.097951 |                  86.085423 |                 39.521285 |                     0.0 |                      0.0 |   -27.423334 |             0.0 |              0.0 |         46.564138 |             56.054123 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.069884 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 20:00:00 | 0.05 | 21.3 |              0.5 |                 0.072689 |        0.867122 |             65.9 |       2.5 |              1018.0 | 0.0 | 0.0 | 18.041667 |                 75.95 |                  13.9 |                     15.438712 |           286.953669 | 0.047 | 21.3 | 18.041667 |    2.003978 |         2.428153 |           2.003978 |          2.5 |                 25.37251 |                     20.726234 |                         0.0 |                          0.0 |                       1.55495 |                           1.303234 |            16.720484 |                 15.741575 |    1001.279516 |   1.196911 | 0.0 | nan | 0.0 |              0.260137 |                      0.0 |                  0.260137 |                  86.086101 |                 39.876809 |                     0.0 |                      0.0 |   -39.616673 |             0.0 |              0.0 |         46.209291 |             85.202267 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.069848 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 21:00:00 | 0.05 | 19.3 |              0.0 |                      0.0 |             0.0 |             73.0 |       1.9 |              1018.0 | 0.0 | 0.0 |   18.1875 |             75.633333 |                  13.9 |                     15.438712 |           286.953669 | 0.047 | 19.3 |   18.1875 |    1.523023 |         2.418134 |           1.523023 |          1.9 |                22.423949 |                     20.917052 |                         0.0 |                          0.0 |                       1.39602 |                           1.313713 |            16.369483 |                 15.820264 |    1001.630517 |   1.205255 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                  86.086101 |                 39.753451 |                     0.0 |                      0.0 |   -39.753451 |             0.0 |              0.0 |          46.33265 |            112.108246 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.070447 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 22:00:00 | 0.05 | 18.3 |              0.0 |                      0.0 |             0.0 |             76.7 |       2.2 |              1018.0 | 0.0 | 0.0 | 18.333333 |                 75.25 |                  13.9 |                     15.438712 |           286.953669 | 0.047 | 18.3 | 18.333333 |    1.763501 |         2.418134 |           1.763501 |          2.2 |                21.065302 |                     21.109404 |                         0.0 |                          0.0 |                      1.321845 |                           1.324263 |            16.157086 |                 15.884826 |    1001.842914 |   1.209486 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                  86.086101 |                 39.670785 |                     0.0 |                      0.0 |   -39.670785 |             0.0 |              0.0 |         46.415315 |             96.820757 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.071143 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 1997-08-05 23:00:00 | 0.05 | 17.4 |              0.0 |                      0.0 |             0.0 |             81.8 |       2.7 |              1018.0 | 0.0 | 0.0 | 18.470833 |               74.9375 |                  13.9 |                     15.438712 |           286.953669 | 0.047 | 17.4 | 18.470833 |    2.164296 |         2.438173 |           2.164296 |          2.7 |                19.904589 |                     21.292177 |                         0.0 |                          0.0 |                      1.257963 |                           1.334275 |            16.281953 |                 15.955825 |    1001.718047 |   1.213176 | 0.0 | nan | 0.0 |                   0.0 |                      0.0 |                       0.0 |                  86.086101 |                 39.564038 |                     0.0 |                      0.0 |   -39.564038 |             0.0 |              0.0 |         46.522063 |             78.890988 |                   nan |                      0.0 |                     0.0 |  0.0 |  0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | 0.071917 |    0.047 | 0.0 | 0.0 |   0.0 |          nan |          0.7 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |      0.0 |      0.0 |  0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |

>>> inputs.nied.series = 0.0
>>> round_(model.check_waterbalance(conditions_wasser))
0.0


.. _lland_v4_nadelw_summer_hourly:

conifers (summer)
-----------------

.. integration-test::

    >>> lnk(NADELW)
    >>> test("lland_v4_nadelw_summer_hourly",
    ...      axis1=(fluxes.evi, fluxes.evb), axis2=states.inzp,
    ...      use_conditions=conditions_nadelw_summer)
    |                date | nied | teml | sunshineduration | possiblesunshineduration | globalradiation | relativehumidity | windspeed | atmosphericpressure |  qz | qzh |   temltag | dailyrelativehumidity | dailysunshineduration | dailypossiblesunshineduration | dailyglobalradiation | nkor | tkor |   tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair |    g |        tz |         wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance | nbes | sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |      evpo |       evi |      evb | evs | wnied | tempssurface | actualalbedo | schmpot | schm | gefrpot | gefr | wlatinz | wlatsnow | wsensinz | wsenssnow | wsurfinz | wsurf | sff | fvg | wada | qdb | qib1 | qib2 | qbb | qkap | qdgz |      qah |       qa |     inzp | stinz | sinz | esnowinz | asinz | wats | waes | esnow | taus |        ebdn |      bowa | qdgz1 | qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 | qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 1997-08-03 00:00:00 |  0.0 | 16.9 |              0.0 |                      0.0 |             0.0 |             95.1 |       0.8 |              1015.0 | 0.0 | 0.0 | 17.279167 |             85.883333 |                   1.3 |                          15.7 |           136.579167 |  0.0 | 16.9 | 17.279167 |    0.641273 |         1.605886 |           0.283992 |          0.8 |                19.284227 |                     19.753091 |                         0.0 |                          0.0 |                      1.223615 |                           1.249589 |              18.3393 |                 16.964613 |       996.6607 |    1.21073 | -2.9 | 15.632162 |  -7.607027 |                   0.0 |                      nan |                       0.0 |                 120.189667 |                 10.408237 |                     0.0 |                      0.0 |   -10.408237 |             0.0 |              0.0 |         109.78143 |                 117.5 |                 100.0 |               170.875289 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.001012 | -0.001012 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.221929 | 0.061647 | 0.001012 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8729.609427 | 78.724819 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.221929 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.061647 |
    | 1997-08-03 01:00:00 |  0.0 | 16.6 |              0.0 |                      0.0 |             0.0 |             94.9 |       0.8 |              1015.0 | 0.0 | 0.0 | 17.241667 |             85.991667 |                   1.3 |                          15.7 |           136.579167 |  0.0 | 16.6 | 17.241667 |    0.641273 |         1.603439 |           0.283992 |          0.8 |                18.920184 |                      19.70628 |                         0.0 |                          0.0 |                       1.20339 |                           1.246999 |            17.955254 |                 16.945759 |     997.044746 |   1.212158 | -2.9 | 15.688646 |  -5.468121 |                   0.0 |                      nan |                       0.0 |                 120.189667 |                  10.41572 |                     0.0 |                      0.0 |    -10.41572 |             0.0 |              0.0 |        109.773946 |                 117.5 |                 100.0 |               173.430608 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.000643 | -0.000643 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.216196 | 0.060055 | 0.001655 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8732.177548 | 78.724819 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.216196 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.060055 |
    | 1997-08-03 02:00:00 |  0.0 | 16.4 |              0.0 |                      0.0 |             0.0 |             95.9 |       0.8 |              1015.0 | 0.0 | 0.0 |    17.175 |             86.233333 |                   1.3 |                          15.7 |           136.579167 |  0.0 | 16.4 |    17.175 |    0.641273 |         1.575992 |           0.283992 |          0.8 |                 18.68084 |                       19.6233 |                         0.0 |                          0.0 |                      1.190065 |                           1.242407 |            17.914926 |                 16.921826 |     997.085074 |   1.213014 | -2.9 | 15.719464 |  -4.083217 |                   0.0 |                      nan |                       0.0 |                 120.189667 |                 10.421627 |                     0.0 |                      0.0 |   -10.421627 |             0.0 |              0.0 |        109.768039 |                 117.5 |                 100.0 |               173.240199 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.002622 | -0.002622 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.210612 | 0.058503 | 0.004277 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8733.360765 | 78.724819 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.210612 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.058503 |
    | 1997-08-03 03:00:00 |  0.0 | 16.3 |              0.0 |                      0.0 |             0.0 |             96.7 |       0.8 |              1015.0 | 0.0 | 0.0 |   17.0625 |             86.708333 |                   1.3 |                          15.5 |           136.579167 |  0.0 | 16.3 |   17.0625 |    0.641273 |         1.548545 |           0.283992 |          0.8 |                18.562165 |                     19.483964 |                         0.0 |                          0.0 |                      1.183449 |                            1.23469 |            17.949613 |                  16.89422 |     997.050387 |   1.213417 | -2.9 | 15.733663 |  -3.398025 |                   0.0 |                      nan |                       0.0 |                 120.189667 |                 10.455184 |                     0.0 |                      0.0 |   -10.455184 |             0.0 |              0.0 |        109.734483 |                 117.5 |                 100.0 |               172.676695 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.004212 | -0.004212 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.205172 | 0.056992 | 0.008489 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  8733.85879 | 78.724819 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.205172 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.056992 |
    | 1997-08-03 04:00:00 |  0.0 | 16.0 |              0.0 |                 0.429734 |        1.943686 |             97.2 |       0.8 |              1015.0 | 0.0 | 0.0 | 16.908333 |             87.366667 |                   1.3 |                     14.929734 |            135.50182 |  0.0 | 16.0 | 16.908333 |    0.641273 |         1.504432 |           0.283992 |          0.8 |                18.210086 |                     19.294427 |                         0.0 |                          0.0 |                      1.163788 |                           1.224181 |            17.700204 |                 16.856897 |     997.299796 |    1.21479 | -2.9 | 15.739639 |  -1.562167 |              1.710444 |                      nan |                  0.317654 |                 119.241602 |                 10.555242 |                     0.0 |                      0.0 |    -8.844799 |             0.0 |              0.0 |        108.686359 |                 117.5 |                 100.0 |               174.220893 |               93.330265 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.003891 | -0.003891 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.199872 |  0.05552 |  0.01238 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8732.520957 | 78.724819 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.199872 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.05552 |
    | 1997-08-03 05:00:00 |  0.0 | 15.9 |              0.0 |                      1.0 |       21.932442 |             97.5 |       0.6 |              1015.0 | 0.0 | 0.0 | 16.729167 |             88.204167 |                   1.3 |                     14.929734 |           134.099005 |  0.0 | 15.9 | 16.729167 |    0.480955 |         1.453638 |           0.212994 |          0.6 |                18.094032 |                     19.076181 |                         0.0 |                          0.0 |                      1.157296 |                           1.212063 |            17.641681 |                 16.825987 |     997.358319 |   1.215237 | -2.9 | 15.723585 |  -1.058491 |             19.300549 |                      nan |                  3.584388 |                 118.007125 |                 10.545193 |                     0.0 |                      0.0 |     8.755356 |             0.0 |              0.0 |        107.461932 |            156.666667 |                 100.0 |               174.539275 |              172.004132 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.008226 |  0.008226 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.194709 | 0.054086 | 0.004154 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8730.679448 | 78.724819 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.194709 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.054086 |
    | 1997-08-03 06:00:00 |  0.0 | 16.0 |              0.0 |                      1.0 |       57.256186 |             97.7 |       0.9 |              1015.0 | 0.0 | 0.0 | 16.533333 |             89.191667 |                   1.3 |                     14.929734 |           131.855513 |  0.0 | 16.0 | 16.533333 |    0.721432 |         1.392031 |           0.319491 |          0.9 |                18.210086 |                     18.840106 |                         0.0 |                          0.0 |                      1.163788 |                           1.198935 |            17.791254 |                 16.803805 |     997.208746 |   1.214748 | -2.9 | 15.701487 |   -1.79108 |             50.385444 |                      nan |                  9.357297 |                 116.032852 |                  10.52516 |                     0.0 |                      0.0 |    39.860284 |             0.0 |              0.0 |        105.507691 |            104.444444 |                 100.0 |               173.410784 |              170.929619 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.034696 |  0.004154 | 0.017612 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.189679 | 0.052689 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8729.570528 | 78.707207 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.189679 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.052689 |
    | 1997-08-03 07:00:00 |  0.0 | 16.6 |              0.0 |                      1.0 |      109.332844 |             97.4 |       0.9 |              1015.0 | 0.0 | 0.0 |    16.375 |                90.125 |                   1.3 |                     14.929734 |           129.465215 |  0.0 | 16.6 |    16.375 |    0.721432 |         1.334591 |           0.319491 |          0.9 |                18.920184 |                     18.651109 |                         0.0 |                          0.0 |                       1.20339 |                           1.188408 |            18.428259 |                 16.809312 |     996.571741 |   1.211943 | -2.9 |  15.68818 |  -5.470922 |             96.212903 |                      nan |                 17.868111 |                 113.929389 |                  10.49073 |                     0.0 |                      0.0 |    85.722173 |             0.0 |              0.0 |        103.438659 |            104.444444 |                 100.0 |               169.265918 |              166.978951 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.073333 |       0.0 | 0.043009 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.18478 | 0.051328 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  8732.14145 | 78.664198 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.18478 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.051328 |
    | 1997-08-03 08:00:00 |  0.0 | 17.4 |              0.0 |                      1.0 |      170.949153 |             96.8 |       0.9 |              1016.0 | 0.0 | 0.0 | 16.216667 |                91.275 |                   1.3 |                     14.929734 |           127.329763 |  0.0 | 17.4 | 16.216667 |    0.721432 |         1.260484 |           0.319491 |          0.9 |                19.904589 |                     18.463773 |                         0.0 |                          0.0 |                      1.257963 |                           1.177959 |            19.267642 |                 16.852809 |     996.732358 |   1.209425 | -2.9 | 15.719031 | -10.085816 |            150.435254 |                      nan |                 27.937976 |                 112.050191 |                 10.426918 |                     0.0 |                      0.0 |   140.008337 |             0.0 |              0.0 |        101.623274 |            104.444444 |                 100.0 |               164.539063 |              162.465759 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.121207 |       0.0 | 0.072568 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.180007 | 0.050002 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8739.327265 |  78.59163 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.180007 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.050002 |
    | 1997-08-03 09:00:00 |  0.0 | 19.0 |              0.2 |                      1.0 |      311.762625 |             86.1 |       1.3 |              1016.0 | 0.0 | 0.0 |   16.1125 |             91.991667 |                   1.5 |                     14.929734 |           129.903206 |  0.0 | 19.0 |   16.1125 |    1.042069 |         1.203904 |           0.461488 |          1.3 |                22.008543 |                     18.341425 |                         0.0 |                          0.0 |                      1.373407 |                           1.171128 |            18.949356 |                 16.872582 |     997.050644 |   1.202945 | -2.9 | 15.805261 | -19.168437 |             274.35111 |                      nan |                  50.95092 |                 114.314821 |                 10.804921 |                     0.0 |                      0.0 |   263.546189 |             0.0 |              0.0 |          103.5099 |             72.307692 |                 100.0 |               175.700241 |              173.109073 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.277889 |       0.0 | 0.144323 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.175357 |  0.04871 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8755.595702 | 78.447307 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.175357 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.04871 |
    | 1997-08-03 10:00:00 |  0.0 | 20.3 |              0.5 |                      1.0 |        501.5833 |             76.8 |       1.5 |              1016.0 | 0.0 | 0.0 | 16.083333 |             92.241667 |                   1.7 |                     14.929734 |           132.285843 |  0.0 | 20.3 | 16.083333 |    1.202387 |         1.149836 |           0.532486 |          1.5 |                23.858503 |                     18.307295 |                         0.0 |                          0.0 |                      1.473678 |                           1.169221 |             18.32333 |                 16.886954 |      997.67667 |   1.197896 | -2.9 | 16.000482 | -25.797109 |            441.393304 |                      nan |                 81.973042 |                 116.411542 |                 11.199993 |                     0.0 |                      0.0 |   430.193311 |             0.0 |              0.0 |        105.211549 |             62.666667 |                 100.0 |               197.930569 |              194.170635 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.483999 |       0.0 | 0.227714 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.170827 | 0.047452 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8778.492812 | 78.219593 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.170827 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.047452 |
    | 1997-08-03 11:00:00 |  0.0 | 21.4 |              0.7 |                      1.0 |      615.018728 |             71.8 |       1.2 |              1016.0 | 0.0 | 0.0 |    16.125 |             92.104167 |                   2.2 |                     14.929734 |           135.919957 |  0.0 | 21.4 |    16.125 |    0.961909 |         1.089916 |           0.425988 |          1.2 |                25.528421 |                     18.356069 |                         0.0 |                          0.0 |                      1.563281 |                           1.171946 |            18.329406 |                 16.906704 |     997.670594 |    1.19342 | -2.9 | 16.275247 | -30.748518 |             541.21648 |                      nan |                100.511632 |                 119.609562 |                 12.222123 |                     0.0 |                      0.0 |   528.994357 |             0.0 |              0.0 |        107.387439 |             78.333333 |                 100.0 |               224.346316 |               218.96261 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.579295 |       0.0 | 0.287568 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.166415 | 0.046226 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8806.341329 | 77.932025 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.166415 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.046226 |
    | 1997-08-03 12:00:00 |  0.0 | 21.3 |              0.8 |                      1.0 |      626.544325 |             67.5 |       1.3 |              1016.0 | 0.0 | 0.0 | 16.204167 |             91.729167 |                   2.8 |                     14.929734 |           145.821804 |  0.0 | 21.3 | 16.204167 |    1.042069 |         1.037502 |           0.461488 |          1.3 |                 25.37251 |                     18.449053 |                         0.0 |                          0.0 |                       1.55495 |                           1.177138 |            17.126444 |                 16.923163 |     998.873556 |   1.194363 | -2.9 | 16.609429 | -28.143424 |            551.359006 |                      nan |                102.395244 |                 128.323187 |                 13.461134 |                     0.0 |                      0.0 |   537.897872 |             0.0 |              0.0 |        114.862053 |             72.307692 |                 100.0 |               245.280266 |              238.430942 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.610907 |       0.0 | 0.279617 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.162116 | 0.045032 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8831.584754 | 77.652407 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.162116 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.045032 |
    | 1997-08-03 13:00:00 |  0.0 | 21.8 |              0.5 |                      1.0 |      496.133417 |             66.1 |       1.5 |              1016.0 | 0.0 | 0.0 | 16.329167 |             91.104167 |                   3.2 |                     14.929734 |           152.606529 |  0.0 | 21.8 | 16.329167 |    1.202387 |         1.012602 |           0.532486 |          1.5 |                26.160453 |                      18.59671 |                         0.0 |                          0.0 |                      1.596982 |                           1.185375 |             17.29206 |                 16.942378 |      998.70794 |   1.192265 | -2.9 |  16.91235 | -29.325898 |            436.597407 |                      nan |                 81.082376 |                 134.293746 |                 14.297428 |                     0.0 |                      0.0 |   422.299979 |             0.0 |              0.0 |        119.996318 |             62.666667 |                 100.0 |               259.968591 |              251.997649 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.531993 |       0.0 | 0.222516 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.157929 | 0.043869 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8858.010651 | 77.429892 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.157929 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.043869 |
    | 1997-08-03 14:00:00 |  0.0 | 22.9 |              0.4 |                      1.0 |      419.520994 |             63.4 |       1.9 |              1016.0 | 0.0 | 0.0 | 16.545833 |             90.058333 |                   3.4 |                     14.929734 |           157.353237 |  0.0 | 22.9 | 16.545833 |    1.523023 |         0.984394 |           0.674482 |          1.9 |                27.969419 |                     18.855098 |                         0.0 |                          0.0 |                      1.692831 |                           1.199769 |            17.732611 |                 16.980587 |     998.267389 |   1.187639 | -2.9 | 17.229461 | -34.023233 |            369.178475 |                      nan |                 68.561717 |                 138.470849 |                 14.726001 |                     0.0 |                      0.0 |   354.452474 |             0.0 |              0.0 |        123.744848 |             49.473684 |                 100.0 |               297.292109 |              286.129869 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.525997 |       0.0 | 0.183486 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.153849 | 0.042736 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8889.133884 | 77.246406 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.153849 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.042736 |
    | 1997-08-03 15:00:00 |  0.0 | 22.7 |              0.5 |                      1.0 |      387.887353 |             62.4 |       1.9 |              1016.0 | 0.0 | 0.0 | 16.816667 |             88.816667 |                   3.8 |                     14.929734 |           164.256877 |  0.0 | 22.7 | 16.816667 |    1.523023 |         0.968687 |           0.674482 |          1.9 |                27.632633 |                     19.182495 |                         0.0 |                          0.0 |                      1.675052 |                           1.217969 |            17.242763 |                 17.037252 |     998.757237 |    1.18866 | -2.9 |  17.60294 |  -30.58236 |             341.34087 |                      nan |                 63.391876 |                 144.546052 |                 15.562805 |                     0.0 |                      0.0 |   325.778065 |             0.0 |              0.0 |        128.983247 |             49.473684 |                 100.0 |               302.770249 |              291.098843 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.500615 |       0.0 | 0.171914 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.149875 | 0.041632 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8916.816245 | 77.074493 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.149875 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.041632 |
    | 1997-08-03 16:00:00 |  0.0 | 22.5 |              0.5 |                      1.0 |      278.496872 |             61.1 |       2.3 |              1016.0 | 0.0 | 0.0 |   17.1375 |             87.333333 |                   4.1 |                     14.929734 |           170.073414 |  0.0 | 22.5 |   17.1375 |     1.84366 |         0.991339 |           0.816478 |          2.3 |                27.299387 |                     19.576758 |                         0.0 |                          0.0 |                      1.657431 |                            1.23983 |            16.679926 |                 17.097035 |     999.320074 |   1.189715 | -2.9 | 17.935128 |  -27.38923 |            245.077248 |                      nan |                 45.514346 |                 149.664604 |                 16.204965 |                     0.0 |                      0.0 |   228.872282 |             0.0 |              0.0 |        133.459638 |             40.869565 |                 100.0 |               310.894515 |              298.448973 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.446523 |       0.0 | 0.133843 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.146004 | 0.040557 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8941.305475 | 76.940649 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.146004 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.040557 |
    | 1997-08-03 17:00:00 |  0.0 | 21.9 |              0.3 |                      1.0 |      137.138608 |             62.1 |       2.4 |              1016.0 | 0.0 | 0.0 |   17.4875 |                85.875 |                   4.4 |                     14.929734 |           173.470856 |  0.0 | 21.9 |   17.4875 |    1.923819 |         1.017332 |           0.851977 |          2.4 |                26.320577 |                     20.014927 |                         0.0 |                          0.0 |                      1.605502 |                           1.264057 |            16.345078 |                 17.187818 |     999.654922 |   1.192283 | -2.9 | 18.228999 | -22.026006 |            120.681975 |                      nan |                 22.412367 |                 152.654353 |                 16.818923 |                     0.0 |                      0.0 |   103.863053 |             0.0 |              0.0 |         135.83543 |             39.166667 |                 100.0 |               291.443446 |              280.813362 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.321091 |       0.0 | 0.096392 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.142232 | 0.039509 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8960.431481 | 76.844258 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.142232 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.039509 |
    | 1997-08-03 18:00:00 |  0.0 | 21.4 |              0.1 |                      1.0 |       51.080714 |             67.0 |       2.5 |              1016.0 | 0.0 | 0.0 |   17.8375 |             84.620833 |                   4.5 |                     14.929734 |           174.440885 |  0.0 | 21.4 |   17.8375 |    2.003978 |         1.054998 |           0.887476 |          2.5 |                25.528421 |                     20.461645 |                         0.0 |                          0.0 |                      1.563281 |                           1.288683 |            17.104042 |                 17.314814 |     998.895958 |   1.193968 | -2.9 | 18.458511 | -17.648933 |             44.951028 |                      nan |                  8.348048 |                 153.507979 |                 16.973018 |                     0.0 |                      0.0 |     27.97801 |             0.0 |              0.0 |        136.534961 |                  37.6 |                 100.0 |               252.715137 |              245.307638 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.222274 |       0.0 | 0.070579 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.138558 | 0.038488 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8975.180414 | 76.773678 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.138558 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.038488 |
    | 1997-08-03 19:00:00 |  0.0 | 20.7 |              0.0 |                      1.0 |       13.632817 |             74.5 |       2.5 |              1016.0 | 0.0 | 0.0 | 18.170833 |                  83.7 |                   4.5 |                     15.429734 |           175.008919 |  0.0 | 20.7 | 18.170833 |    2.003978 |         1.100997 |           0.887476 |          2.5 |                24.454368 |                     20.895167 |                         0.0 |                          0.0 |                      1.505746 |                           1.312512 |            18.218504 |                 17.489255 |     997.781496 |   1.196313 | -2.9 | 18.635498 |  -12.38701 |             11.996879 |                      nan |                  2.227992 |                 154.007849 |                 16.556776 |                     0.0 |                      0.0 |    -4.559897 |             0.0 |              0.0 |        137.451073 |                  37.6 |                 100.0 |               212.733181 |              208.094539 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.142971 |       0.0 | 0.049855 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.134979 | 0.037494 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8984.667424 | 76.723823 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.134979 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.037494 |
    | 1997-08-03 20:00:00 |  0.0 | 19.4 |              0.0 |                   0.1364 |        0.185944 |             81.2 |       2.2 |              1016.0 | 0.0 | 0.0 | 18.454167 |             83.066667 |                   4.5 |                     15.566134 |           175.016667 |  0.0 | 19.4 | 18.454167 |    1.763501 |         1.120309 |           0.780979 |          2.2 |                22.563931 |                      21.26995 |                         0.0 |                          0.0 |                      1.403627 |                           1.333058 |            18.321912 |                 17.668238 |     997.678088 |   1.201582 | -2.9 | 18.749342 |  -3.903945 |              0.163631 |                      nan |                  0.030389 |                 154.014667 |                 16.343497 |                     0.0 |                      0.0 |   -16.179866 |             0.0 |              0.0 |         137.67117 |             42.727273 |                 100.0 |               190.511989 |               75.961192 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.07766 |       0.0 | 0.047497 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.131493 | 0.036526 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  8985.67137 | 76.676326 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.131493 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.036526 |
    | 1997-08-03 21:00:00 |  0.0 | 17.8 |              0.0 |                      0.0 |             0.0 |             86.9 |       1.7 |              1016.0 | 0.0 | 0.0 |     18.65 |               82.7375 |                   4.5 |                     15.566134 |           175.016667 |  0.0 | 17.8 |     18.65 |    1.362705 |         1.127089 |           0.603484 |          1.7 |                20.413369 |                     21.532411 |                         0.0 |                          0.0 |                      1.286025 |                           1.347418 |            17.739217 |                 17.815378 |     998.260783 |   1.208454 | -2.9 |  18.76139 |   5.768339 |                   0.0 |                      nan |                       0.0 |                 154.014667 |                 16.221809 |                     0.0 |                      0.0 |   -16.221809 |             0.0 |              0.0 |        137.792858 |             55.294118 |                 100.0 |               185.548772 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.033662 |       0.0 | 0.022645 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.128096 | 0.035582 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8977.003031 | 76.653681 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.128096 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.035582 |
    | 1997-08-03 22:00:00 |  0.0 | 17.0 |              0.0 |                      0.0 |             0.0 |             90.1 |       1.7 |              1017.0 | 0.0 | 0.0 | 18.808333 |             82.554167 |                   4.5 |                     15.566134 |           175.016667 |  0.0 | 17.0 | 18.808333 |    1.362705 |         1.142201 |           0.603484 |          1.7 |                19.406929 |                     21.746678 |                         0.0 |                          0.0 |                      1.230421 |                           1.359123 |            17.485643 |                 17.952788 |     999.514357 |   1.213101 | -2.9 |  18.65737 |   9.944218 |                   0.0 |                      nan |                       0.0 |                 154.014667 |                 16.101364 |                     0.0 |                      0.0 |   -16.101364 |             0.0 |              0.0 |        137.913303 |             55.294118 |                 100.0 |               184.343649 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.020833 |       0.0 | 0.013897 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.124787 | 0.034663 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8964.158813 | 76.639784 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.124787 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.034663 |
    | 1997-08-03 23:00:00 |  0.0 | 16.4 |              0.0 |                      0.0 |             0.0 |             90.9 |       2.3 |              1017.0 | 0.0 | 0.0 | 18.941667 |             82.379167 |                   4.5 |                     15.566134 |           175.016667 |  0.0 | 16.4 | 18.941667 |     1.84366 |         1.185687 |           0.816478 |          2.3 |                 18.68084 |                     21.928555 |                         0.0 |                          0.0 |                      1.190065 |                           1.369047 |            16.980884 |                 18.064561 |    1000.019116 |   1.215845 | -2.9 | 18.503239 |  12.619435 |                   0.0 |                      nan |                       0.0 |                 154.014667 |                 16.005095 |                     0.0 |                      0.0 |   -16.005095 |             0.0 |              0.0 |        138.009572 |             40.869565 |                 100.0 |               187.254752 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.027767 |       0.0 | 0.016654 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.121564 | 0.033768 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8948.639378 | 76.623131 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.121564 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.033768 |
    | 1997-08-04 00:00:00 |  0.0 | 15.9 |              0.0 |                      0.0 |             0.0 |             88.0 |       2.3 |              1017.0 | 0.0 | 0.0 |      18.9 |             82.083333 |                   4.5 |                     15.566134 |           175.016667 |  0.0 | 15.9 |      18.9 |     1.84366 |         1.235786 |           0.816478 |          2.3 |                18.094032 |                     21.871576 |                         0.0 |                          0.0 |                      1.157296 |                           1.365939 |            15.922748 |                 17.952919 |    1001.077252 |    1.21843 | -2.9 | 18.317006 |  14.502035 |                   0.0 |                      nan |                       0.0 |                 154.014667 |                 16.128293 |                     0.0 |                      0.0 |   -16.128293 |             0.0 |              0.0 |        137.886374 |             40.869565 |                 100.0 |               196.846938 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.041241 |       0.0 | 0.024575 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.118424 | 0.032895 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8931.237343 | 76.598556 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.118424 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.032895 |
    | 1997-08-04 01:00:00 |  0.0 | 15.4 |              0.0 |                      0.0 |             0.0 |             89.2 |       2.2 |              1016.0 | 0.0 | 0.0 |     18.85 |             81.845833 |                   4.5 |                     15.566134 |           175.016667 |  0.0 | 15.4 |     18.85 |    1.763501 |         1.282546 |           0.780979 |          2.2 |                17.523416 |                     21.803372 |                         0.0 |                          0.0 |                      1.125295 |                           1.362218 |            15.630887 |                 17.845152 |    1000.369113 |   1.219467 | -2.9 | 18.108181 |  16.249089 |                   0.0 |                      nan |                       0.0 |                 154.014667 |                 16.244859 |                     0.0 |                      0.0 |   -16.244859 |             0.0 |              0.0 |        137.769808 |             42.727273 |                 100.0 |               198.631945 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.032757 |       0.0 |  0.01971 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.115365 | 0.032046 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8912.088255 | 76.578846 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.115365 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.032046 |
    | 1997-08-04 02:00:00 |  0.0 | 15.2 |              0.0 |                      0.0 |             0.0 |             88.8 |       2.3 |              1016.0 | 0.0 | 0.0 |      18.8 |                 81.55 |                   4.5 |                     15.566134 |           175.016667 |  0.0 | 15.2 |      18.8 |     1.84366 |         1.332645 |           0.816478 |          2.3 |                17.299618 |                     21.735354 |                         0.0 |                          0.0 |                      1.112706 |                           1.358505 |            15.362061 |                 17.725181 |    1000.637939 |   1.220436 | -2.9 | 17.878392 |  16.070354 |                   0.0 |                      nan |                       0.0 |                 154.014667 |                 16.376926 |                     0.0 |                      0.0 |   -16.376926 |             0.0 |              0.0 |        137.637741 |             40.869565 |                 100.0 |               201.151424 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.036313 |       0.0 |  0.02144 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.112385 | 0.031218 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |   8893.1179 | 76.557406 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.112385 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.031218 |
    | 1997-08-04 03:00:00 |  0.0 | 14.9 |              0.0 |                      0.0 |             0.0 |             89.7 |       2.0 |              1016.0 | 0.0 | 0.0 | 18.741667 |             81.258333 |                   4.5 |                     15.566134 |           175.016667 |  0.0 | 14.9 | 18.741667 |    1.603182 |         1.372725 |           0.709981 |          2.0 |                16.968612 |                     21.656234 |                         0.0 |                          0.0 |                      1.094046 |                           1.354184 |            15.220845 |                 17.597495 |    1000.779155 |   1.221771 | -2.9 | 17.650748 |  16.504489 |                   0.0 |                      nan |                       0.0 |                 154.014667 |                 16.516685 |                     0.0 |                      0.0 |   -16.516685 |             0.0 |              0.0 |        137.497982 |                  47.0 |                 100.0 |               202.118359 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.026013 |       0.0 | 0.016089 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.109482 | 0.030412 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8873.713411 | 76.541317 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.109482 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.030412 |
    | 1997-08-04 04:00:00 |  0.0 | 14.3 |              0.0 |                 0.398044 |        1.590517 |             93.0 |       2.3 |              1016.0 | 0.0 | 0.0 | 18.670833 |             81.083333 |                   4.5 |                     15.534444 |           175.001952 |  0.0 | 14.3 | 18.670833 |     1.84366 |         1.422824 |           0.816478 |          2.3 |                16.323194 |                     21.560498 |                         0.0 |                          0.0 |                      1.057522 |                           1.348953 |             15.18057 |                  17.48197 |     1000.81943 |    1.22434 | -2.9 | 17.417894 |  18.707366 |              1.399655 |                      nan |                  0.259936 |                 154.001717 |                  16.65652 |                     0.0 |                      0.0 |   -15.256865 |             0.0 |              0.0 |        137.345197 |             40.869565 |                 100.0 |               201.757278 |               93.626391 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.016871 |       0.0 | 0.008597 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.106654 | 0.029626 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8852.106046 | 76.532721 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.106654 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.029626 |
    | 1997-08-04 05:00:00 |  0.0 | 14.1 |              0.0 |                      1.0 |       19.950903 |             93.6 |       2.5 |              1017.0 | 0.0 | 0.0 | 18.595833 |             80.920833 |                   4.5 |                     15.534444 |           174.919387 |  0.0 | 14.1 | 18.595833 |    2.003978 |         1.486284 |           0.887476 |          2.5 |                16.112886 |                     21.459534 |                         0.0 |                          0.0 |                      1.045579 |                           1.343433 |            15.081661 |                 17.365233 |    1001.918339 |   1.226451 | -2.9 | 17.158606 |  18.351635 |             17.556794 |                      nan |                  3.260548 |                 153.929061 |                 16.779096 |                     0.0 |                      0.0 |     0.777698 |             0.0 |              0.0 |        137.149965 |                  37.6 |                 100.0 |               202.695817 |              198.661729 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.029889 |       0.0 | 0.009285 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.103899 | 0.028861 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  8830.85441 | 76.523435 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.103899 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.028861 |
    | 1997-08-04 06:00:00 |  0.0 | 14.3 |              0.1 |                      1.0 |       62.985775 |             94.6 |       2.4 |              1017.0 | 0.0 | 0.0 |    18.525 |             80.791667 |                   4.6 |                     15.534444 |            175.15812 |  0.0 | 14.3 |    18.525 |    1.923819 |         1.536383 |           0.851977 |          2.4 |                16.323194 |                     21.364558 |                         0.0 |                          0.0 |                      1.057522 |                           1.338237 |            15.441741 |                 17.260782 |    1001.558259 |   1.225432 | -2.9 | 16.903586 |  15.621518 |             55.427482 |                      nan |                 10.293675 |                 154.139146 |                 17.089561 |                     0.0 |                      0.0 |    38.337921 |             0.0 |              0.0 |        137.049585 |             39.166667 |                 100.0 |               199.048483 |              195.224978 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.056019 |       0.0 | 0.018148 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.101215 | 0.028115 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8812.332893 | 76.505287 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.101215 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.028115 |
    | 1997-08-04 07:00:00 |  0.0 | 15.4 |              0.9 |                      1.0 |      285.776389 |             93.3 |       2.5 |              1017.0 | 0.0 | 0.0 |    18.475 |             80.620833 |                   5.5 |                     15.534444 |           182.509935 |  0.0 | 15.4 |    18.475 |    2.003978 |         1.589823 |           0.887476 |          2.5 |                17.523416 |                     21.297737 |                         0.0 |                          0.0 |                      1.125295 |                            1.33458 |            16.349347 |                 17.170413 |    1000.650653 |   1.220347 | -2.9 | 16.681328 |   7.687968 |            251.483222 |                      nan |                 46.704027 |                 160.608743 |                 19.012403 |                     0.0 |                      0.0 |    232.47082 |             0.0 |              0.0 |         141.59634 |                  37.6 |                 100.0 |               191.288673 |               187.89699 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.234872 |       0.0 | 0.077957 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.098601 | 0.027389 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8801.744925 |  76.42733 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.098601 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.027389 |
    | 1997-08-04 08:00:00 |  0.0 | 16.8 |              1.0 |                      1.0 |      475.225247 |             86.9 |       2.5 |              1017.0 | 0.0 | 0.0 |     18.45 |             80.208333 |                   6.5 |                     15.534444 |           195.188105 |  0.0 | 16.8 |     18.45 |    2.003978 |         1.643262 |           0.887476 |          2.5 |                19.162205 |                     21.264396 |                         0.0 |                          0.0 |                      1.216842 |                           1.332754 |            16.651956 |                 17.055817 |    1000.348044 |   1.214317 | -2.9 | 16.554272 |  -1.474365 |            418.198218 |                      nan |                 77.665383 |                 171.765533 |                 21.208047 |                     0.0 |                      0.0 |   396.990171 |             0.0 |              0.0 |        150.557486 |                  37.6 |                 100.0 |               192.879382 |              189.400978 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.424445 |       0.0 | 0.144461 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.096054 | 0.026682 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  8800.31929 | 76.282869 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.096054 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.026682 |
    | 1997-08-04 09:00:00 |  0.0 | 18.2 |              1.0 |                      1.0 |      614.006719 |             78.9 |       2.9 |              1017.0 | 0.0 | 0.0 | 18.416667 |             79.908333 |                   7.3 |                     15.534444 |           207.781609 |  0.0 | 18.2 | 18.416667 |    2.324615 |         1.696701 |           1.029472 |          2.9 |                20.933479 |                     21.220011 |                         0.0 |                          0.0 |                      1.314615 |                           1.330323 |            16.516515 |                 16.956557 |    1000.483485 |   1.208543 | -2.9 | 16.537165 |  -9.977011 |            540.325913 |                      nan |                100.346241 |                 182.847816 |                 22.996116 |                     0.0 |                      0.0 |   517.329797 |             0.0 |              0.0 |          159.8517 |             32.413793 |                 100.0 |                203.80698 |              199.707772 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.613301 |       0.0 | 0.189606 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.093572 | 0.025992 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8807.396301 | 76.093264 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.093572 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.025992 |
    | 1997-08-04 10:00:00 |  0.0 | 19.5 |              0.9 |                      1.0 |      643.806508 |             75.8 |       3.1 |              1017.0 | 0.0 | 0.0 | 18.383333 |             79.866667 |                   7.7 |                     15.534444 |           213.707576 |  0.0 | 19.5 | 18.383333 |    2.484933 |         1.750141 |            1.10047 |          3.1 |                22.704676 |                     21.175708 |                         0.0 |                          0.0 |                      1.411269 |                           1.327896 |            17.210144 |                 16.912332 |     999.789856 |   1.202862 | -2.9 | 16.622089 | -17.267466 |            566.549727 |                      nan |                105.216378 |                 188.062667 |                 23.882018 |                     0.0 |                      0.0 |   542.667709 |             0.0 |              0.0 |        164.180649 |             30.322581 |                 100.0 |                207.85156 |              203.511519 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.68553 |       0.0 | 0.206717 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.091155 | 0.025321 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8821.763767 | 75.886546 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.091155 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.025321 |
    | 1997-08-04 11:00:00 |  0.0 | 20.3 |              0.8 |                      1.0 |      600.262981 |             73.4 |       3.3 |              1018.0 | 0.0 | 0.0 |   18.3375 |             79.933333 |                   7.8 |                     15.534444 |           213.092753 |  0.0 | 20.3 |   18.3375 |    2.645251 |          1.82028 |           1.171468 |          3.3 |                23.858503 |                     21.114922 |                         0.0 |                          0.0 |                      1.473678 |                           1.324565 |            17.512141 |                 16.877861 |    1000.487859 |   1.200635 | -2.9 | 16.794499 | -21.033009 |            528.231423 |                      nan |                 98.100121 |                 187.521623 |                   24.1295 |                     0.0 |                      0.0 |   504.101923 |             0.0 |              0.0 |        163.392123 |             28.484848 |                 100.0 |               216.995805 |              212.089353 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.688425 |       0.0 | 0.197417 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.088801 | 0.024667 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8839.896776 | 75.689129 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.088801 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.024667 |
    | 1997-08-04 12:00:00 |  0.0 | 21.0 |              0.9 |                      1.0 |      620.514139 |             68.3 |       3.3 |              1018.0 | 0.0 | 0.0 |    18.325 |             79.966667 |                   7.9 |                     15.534444 |           212.841496 |  0.0 | 21.0 |    18.325 |    2.645251 |         1.887079 |           1.171468 |          3.3 |                24.909749 |                     21.098371 |                         0.0 |                          0.0 |                      1.530181 |                           1.323658 |            17.013358 |                 16.871664 |    1000.986642 |   1.198001 | -2.9 | 17.012095 | -23.927432 |            546.052442 |                      nan |                101.409739 |                 187.300516 |                 24.341655 |                     0.0 |                      0.0 |   521.710787 |             0.0 |              0.0 |        162.958861 |             28.484848 |                 100.0 |               245.424582 |              238.564611 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.754088 |       0.0 | 0.202171 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.086507 |  0.02403 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8860.924208 | 75.486958 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.086507 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.02403 |
    | 1997-08-04 13:00:00 |  0.0 | 21.8 |              0.8 |                      1.0 |      599.441386 |             61.3 |       3.2 |              1018.0 | 0.0 | 0.0 |    18.325 |             79.766667 |                   8.2 |                     15.534444 |           217.145994 |  0.0 | 21.8 |    18.325 |    2.565092 |         1.943859 |           1.135969 |          3.2 |                26.160453 |                     21.098371 |                         0.0 |                          0.0 |                      1.596982 |                           1.323658 |            16.036358 |                 16.829467 |    1001.963642 |   1.195188 | -2.9 | 17.264424 | -27.213457 |             527.50842 |                      nan |                 97.965849 |                 191.088475 |                 25.038979 |                     0.0 |                      0.0 |   502.469441 |             0.0 |              0.0 |        166.049497 |                29.375 |                 100.0 |               301.588452 |              290.027764 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.791888 |       0.0 | 0.191434 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.084272 | 0.023409 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8885.237665 | 75.295524 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.084272 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.023409 |
    | 1997-08-04 14:00:00 |  0.0 | 21.9 |              0.9 |                      1.0 |      613.551222 |             60.6 |       2.5 |              1018.0 | 0.0 | 0.0 | 18.283333 |                 79.65 |                   8.7 |                     15.534444 |           225.230587 |  0.0 | 21.9 | 18.283333 |    2.003978 |         1.963899 |           0.887476 |          2.5 |                26.320577 |                     21.043281 |                         0.0 |                          0.0 |                      1.605502 |                           1.320638 |             15.95027 |                 16.760973 |     1002.04973 |   1.194821 | -2.9 | 17.556185 | -26.062888 |            539.925076 |                      nan |                  100.2718 |                 198.202917 |                 26.187015 |                     0.0 |                      0.0 |   513.738061 |             0.0 |              0.0 |        172.015902 |                  37.6 |                 100.0 |               310.080531 |              297.713573 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.744374 |       0.0 | 0.208908 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.082096 | 0.022804 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8908.400553 | 75.086616 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.082096 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.022804 |
    | 1997-08-04 15:00:00 |  0.0 | 21.8 |              0.9 |                      1.0 |      517.107047 |             58.8 |       2.9 |              1017.0 | 0.0 | 0.0 | 18.245833 |                  79.5 |                   9.1 |                     15.534444 |           230.614741 |  0.0 | 21.8 | 18.245833 |    2.324615 |         1.997298 |           1.029472 |          2.9 |                26.160453 |                     20.993808 |                         0.0 |                          0.0 |                      1.596982 |                           1.317924 |            15.382347 |                 16.690077 |    1001.617653 |   1.194298 | -2.9 |  17.83414 |  -23.79516 |            455.054202 |                      nan |                 84.510066 |                 202.940972 |                 27.142101 |                     0.0 |                      0.0 |     427.9121 |             0.0 |              0.0 |        175.798871 |             32.413793 |                 100.0 |               324.683496 |              310.872289 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.707691 |       0.0 |  0.17377 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.079975 | 0.022215 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8929.295714 | 74.912846 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.079975 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.022215 |
    | 1997-08-04 16:00:00 |  0.0 | 22.2 |              0.9 |                      1.0 |      377.836617 |             57.2 |       3.6 |              1017.0 | 0.0 | 0.0 | 18.233333 |               79.3375 |                   9.5 |                     15.534444 |           234.753897 |  0.0 | 22.2 | 18.233333 |    2.885728 |         2.040718 |           1.277965 |          3.6 |                26.806088 |                      20.97734 |                         0.0 |                          0.0 |                      1.631293 |                           1.317021 |            15.333082 |                 16.642897 |    1001.666918 |   1.192703 | -2.9 | 18.084882 | -24.690709 |            332.496223 |                      nan |                 61.749299 |                  206.58343 |                 28.068928 |                     0.0 |                      0.0 |   304.427295 |             0.0 |              0.0 |        178.514502 |             26.111111 |                 100.0 |               351.945528 |               335.24437 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.668882 |       0.0 | 0.134178 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.077909 | 0.021641 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8951.086422 | 74.778667 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.077909 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.021641 |
    | 1997-08-04 17:00:00 |  0.0 | 21.8 |              1.0 |                      1.0 |      252.204639 |             57.2 |       4.2 |              1017.0 | 0.0 | 0.0 | 18.229167 |             79.133333 |                  10.2 |                     15.534444 |           239.548315 |  0.0 | 21.8 | 18.229167 |    3.366683 |         2.100837 |            1.49096 |          4.2 |                26.160453 |                     20.971853 |                         0.0 |                          0.0 |                      1.596982 |                            1.31672 |            14.963779 |                 16.595726 |    1002.036221 |   1.194485 | -2.9 |  18.34637 | -20.721778 |            221.940082 |                      nan |                 41.217444 |                 210.802517 |                 29.636995 |                     0.0 |                      0.0 |   192.303087 |             0.0 |              0.0 |        181.165522 |             22.380952 |                 100.0 |               341.498675 |              325.934588 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.608204 |       0.0 | 0.109857 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.075897 | 0.021082 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |   8968.9082 |  74.66881 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.075897 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.021082 |
    | 1997-08-04 18:00:00 |  0.0 | 20.9 |              1.0 |                      1.0 |      128.576714 |             59.9 |       4.2 |              1017.0 | 0.0 | 0.0 | 18.208333 |               78.8375 |                  11.1 |                     15.534444 |           242.777315 |  0.0 | 20.9 | 18.208333 |    3.366683 |         2.157616 |            1.49096 |          4.2 |                 24.75714 |                     20.944437 |                         0.0 |                          0.0 |                      1.521999 |                           1.315216 |            14.829527 |                  16.51207 |    1002.170473 |   1.198201 | -2.9 | 18.560232 |  -14.03861 |            113.147508 |                      nan |                 21.013109 |                 213.644037 |                 31.712576 |                     0.0 |                      0.0 |    81.434932 |             0.0 |              0.0 |        181.931461 |             22.380952 |                 100.0 |               298.901864 |               287.59108 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.461374 |       0.0 | 0.089881 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.073936 | 0.020538 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8980.046809 |  74.57893 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.073936 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.020538 |
    | 1997-08-04 19:00:00 |  0.0 | 19.7 |              1.0 |                      1.0 |       42.072586 |             66.1 |       3.6 |              1017.0 | 0.0 | 0.0 | 18.166667 |               78.4875 |                  12.1 |                     15.534444 |           243.962306 |  0.0 | 19.7 | 18.166667 |    2.885728 |         2.194356 |           1.277965 |          3.6 |                22.988466 |                     20.889699 |                         0.0 |                          0.0 |                      1.426658 |                           1.312212 |            15.195376 |                 16.395802 |    1001.804624 |   1.202947 | -2.9 | 18.693895 |   -6.03663 |             37.023876 |                      nan |                  6.875863 |                 214.686829 |                 34.091963 |                     0.0 |                      0.0 |     2.931913 |             0.0 |              0.0 |        180.594866 |             26.111111 |                 100.0 |                249.92486 |              242.729144 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.273023 |       0.0 | 0.065773 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.072026 | 0.020007 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8983.183439 | 74.513157 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.072026 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.020007 |
    | 1997-08-04 20:00:00 |  0.0 | 17.9 |              0.3 |                 0.104711 |        0.812897 |             71.7 |       3.2 |              1017.0 | 0.0 | 0.0 | 18.104167 |             78.091667 |                  12.4 |                     15.502754 |           243.988429 |  0.0 | 17.9 | 18.104167 |    2.565092 |         2.227756 |           1.135969 |          3.2 |                20.542326 |                     20.807826 |                         0.0 |                          0.0 |                      1.293123 |                           1.307717 |            14.728848 |                 16.249178 |    1002.271152 |   1.210597 | -2.9 | 18.731535 |   4.989208 |               0.71535 |                      nan |                  0.132851 |                 214.709817 |                 35.120093 |                     0.0 |                      0.0 |   -34.404743 |             0.0 |              0.0 |        179.589724 |                29.375 |                 100.0 |                232.15313 |               74.878349 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.16074 |       0.0 | 0.083415 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.070166 | 0.019491 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8975.294232 | 74.429742 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.070166 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.019491 |
    | 1997-08-04 21:00:00 |  0.0 | 15.8 |              0.0 |                      0.0 |             0.0 |             80.6 |       2.2 |              1018.0 | 0.0 | 0.0 | 18.020833 |             77.829167 |                  12.4 |                     15.502754 |           243.988429 |  0.0 | 15.8 | 18.020833 |    1.763501 |         2.244455 |           0.780979 |          2.2 |                17.978626 |                     20.699099 |                         0.0 |                          0.0 |                      1.150835 |                           1.301743 |            14.490772 |                 16.109936 |    1003.509228 |    1.22071 | -2.9 | 18.636864 |  17.021185 |                   0.0 |                      nan |                       0.0 |                 214.709817 |                 35.424509 |                     0.0 |                      0.0 |   -35.424509 |             0.0 |              0.0 |        179.285308 |             42.727273 |                 100.0 |               219.832992 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.056348 |       0.0 | 0.034081 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.068353 | 0.018987 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8955.373047 | 74.395661 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.068353 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.018987 |
    | 1997-08-04 22:00:00 |  0.0 | 14.8 |              0.0 |                      0.0 |             0.0 |             85.9 |       2.2 |              1018.0 | 0.0 | 0.0 | 17.929167 |             77.654167 |                  12.4 |                     15.502754 |           243.988429 |  0.0 | 14.8 | 17.929167 |    1.763501 |         2.261155 |           0.780979 |          2.2 |                16.859515 |                     20.580072 |                         0.0 |                          0.0 |                      1.087886 |                           1.295199 |            14.482324 |                 15.981284 |    1003.517676 |   1.224953 | -2.9 |  18.39781 |  21.586859 |                   0.0 |                      nan |                       0.0 |                 214.709817 |                 35.698198 |                     0.0 |                      0.0 |   -35.698198 |             0.0 |              0.0 |        179.011619 |             42.727273 |                 100.0 |               216.477578 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.030446 |       0.0 | 0.018181 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.066588 | 0.018497 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8930.886187 |  74.37748 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.066588 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.018497 |
    | 1997-08-04 23:00:00 |  0.0 | 14.1 |              0.0 |                      0.0 |             0.0 |             89.3 |       2.1 |              1018.0 | 0.0 | 0.0 | 17.833333 |               77.5875 |                  12.4 |                     15.502754 |           243.988429 |  0.0 | 14.1 | 17.833333 |    1.683342 |         2.254475 |            0.74548 |          2.1 |                16.112886 |                     20.456276 |                         0.0 |                          0.0 |                      1.045579 |                           1.288387 |            14.388807 |                 15.871513 |    1003.611193 |   1.227981 | -2.9 | 18.103968 |  24.023805 |                   0.0 |                      nan |                       0.0 |                 214.709817 |                 35.921753 |                     0.0 |                      0.0 |   -35.921753 |             0.0 |              0.0 |        178.788064 |             44.761905 |                 100.0 |               216.443117 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.012637 |       0.0 | 0.007608 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.064868 | 0.018019 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8903.962382 | 74.369872 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.064868 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.018019 |
    | 1997-08-05 00:00:00 |  0.0 | 13.4 |              0.0 |                      0.0 |             0.0 |             91.4 |       1.8 |              1018.0 | 0.0 | 0.0 | 17.729167 |             77.729167 |                  12.4 |                     15.502754 |           243.988429 |  0.0 | 13.4 | 17.729167 |    1.442864 |         2.237776 |           0.638983 |          1.8 |                15.395377 |                     20.322453 |                         0.0 |                          0.0 |                      1.004677 |                           1.281018 |            14.071374 |                 15.796473 |    1003.928626 |   1.231127 | -2.9 | 17.780882 |  26.285292 |                   0.0 |                      nan |                       0.0 |                 214.709817 |                 36.050541 |                     0.0 |                      0.0 |   -36.050541 |             0.0 |              0.0 |        178.659276 |             52.222222 |                 100.0 |               219.886198 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.000854 | -0.000854 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.063192 | 0.017553 | 0.000854 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  8874.77709 | 74.369872 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.063192 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.017553 |
    | 1997-08-05 01:00:00 |  0.0 | 12.8 |              0.0 |                      0.0 |             0.0 |             93.1 |       1.8 |              1018.0 | 0.0 | 0.0 | 17.620833 |             77.891667 |                  12.4 |                     15.502754 |           243.988429 |  0.0 | 12.8 | 17.620833 |    1.442864 |         2.224416 |           0.638983 |          1.8 |                14.802811 |                     20.184089 |                         0.0 |                          0.0 |                      0.970706 |                           1.273391 |            13.781417 |                 15.721724 |    1004.218583 |   1.233843 | -2.9 | 17.430658 |  27.783951 |                   0.0 |                      nan |                       0.0 |                 214.709817 |                 36.176259 |                     0.0 |                      0.0 |   -36.176259 |             0.0 |              0.0 |        178.533559 |             52.222222 |                 100.0 |               223.673889 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.007069 | -0.007069 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.06156 |   0.0171 | 0.007924 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan |  8844.09314 | 74.369872 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.06156 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0171 |
    | 1997-08-05 02:00:00 |  0.0 | 12.3 |              0.0 |                      0.0 |             0.0 |             94.6 |       1.6 |              1018.0 | 0.0 | 0.0 |      17.5 |             78.133333 |                  12.4 |                     15.502754 |           243.988429 |  0.0 | 12.3 |      17.5 |    1.282546 |         2.201036 |           0.567985 |          1.6 |                14.324376 |                     20.030733 |                         0.0 |                          0.0 |                      0.943147 |                            1.26493 |            13.550859 |                 15.650679 |    1004.449141 |   1.236111 | -2.9 | 17.062451 |  28.574706 |                   0.0 |                      nan |                       0.0 |                 214.709817 |                 36.284611 |                     0.0 |                      0.0 |   -36.284611 |             0.0 |              0.0 |        178.425206 |                 58.75 |                 100.0 |                227.17139 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.013945 | -0.013945 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 |  0.05997 | 0.016658 | 0.021868 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8812.618434 | 74.369872 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 |  0.05997 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.016658 |
    | 1997-08-05 03:00:00 |  0.0 | 12.0 |              0.0 |                      0.0 |             0.0 |             95.6 |       1.2 |              1018.0 | 0.0 | 0.0 | 17.379167 |             78.379167 |                  12.4 |                     15.502754 |           243.988429 |  0.0 | 12.0 | 17.379167 |    0.961909 |         2.174316 |           0.425988 |          1.2 |                 14.04387 |                     19.878396 |                         0.0 |                          0.0 |                      0.926933 |                           1.256516 |            13.425939 |                 15.580521 |    1004.574061 |   1.237469 | -2.9 | 16.684755 |  28.108527 |                   0.0 |                      nan |                       0.0 |                 214.709817 |                 36.390891 |                     0.0 |                      0.0 |   -36.390891 |             0.0 |              0.0 |        178.318926 |             78.333333 |                 100.0 |               229.324539 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.019233 | -0.019233 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.058421 | 0.016228 | 0.041102 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8781.609907 | 74.369872 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.058421 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.016228 |
    | 1997-08-05 04:00:00 |  0.0 | 11.7 |              0.0 |                 0.366022 |        1.482347 |             96.2 |       1.9 |              1018.0 | 0.0 | 0.0 | 17.270833 |               78.5125 |                  12.4 |                     15.470733 |           243.983922 |  0.0 | 11.7 | 17.270833 |    1.523023 |         2.160956 |           0.674482 |          1.9 |                13.768192 |                      19.74268 |                         0.0 |                          0.0 |                      0.910957 |                           1.249013 |            13.245001 |                 15.500472 |    1004.754999 |   1.238856 | -2.9 | 16.312652 |  27.675913 |              1.304466 |                      nan |                  0.242258 |                 214.705851 |                 36.589063 |                     0.0 |                      0.0 |   -35.284597 |             0.0 |              0.0 |        178.116788 |             49.473684 |                 100.0 |               232.308624 |               93.057159 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.016393 | -0.016393 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.056912 | 0.015809 | 0.057495 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8751.033993 | 74.369872 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.056912 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.015809 |
    | 1997-08-05 05:00:00 |  0.0 | 11.6 |              0.0 |                      1.0 |       21.135567 |             95.7 |       1.3 |              1018.0 | 0.0 | 0.0 | 17.166667 |                  78.6 |                  12.4 |                     15.470733 |           244.033283 |  0.0 | 11.6 | 17.166667 |    1.042069 |         2.120877 |           0.461488 |          1.3 |                 13.67736 |                     19.612949 |                         0.0 |                          0.0 |                      0.905683 |                           1.241834 |            13.089234 |                 15.415778 |    1004.910766 |   1.239363 | -2.9 | 15.945741 |  26.074448 |             18.599299 |                      nan |                  3.454155 |                 214.749289 |                  36.74524 |                     0.0 |                      0.0 |   -18.145941 |             0.0 |              0.0 |        178.004048 |             72.307692 |                 100.0 |               234.489857 |              228.415592 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.005017 | -0.005017 |      0.0 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.055442 |   0.0154 | 0.062512 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8722.059546 | 74.369872 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.055442 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0154 |
    | 1997-08-05 06:00:00 |  0.0 | 12.6 |              0.6 |                      1.0 |       125.88345 |             95.1 |       1.6 |              1018.0 | 0.0 | 0.0 | 17.095833 |             78.620833 |                  12.9 |                     15.470733 |           246.654019 |  0.0 | 12.6 | 17.095833 |    1.282546 |         2.094157 |           0.567985 |          1.6 |                14.609782 |                     19.525158 |                         0.0 |                          0.0 |                      0.959602 |                           1.236973 |            13.893903 |                 15.350842 |    1004.106097 |   1.234655 | -2.9 | 15.598048 |  17.988287 |            110.777436 |                      nan |                 20.572952 |                 217.055537 |                  38.00446 |                     0.0 |                      0.0 |    72.772976 |             0.0 |              0.0 |        179.051077 |                 58.75 |                 100.0 |               222.631747 |              217.361109 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.070823 |  0.062512 | 0.003085 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.054009 | 0.015003 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8701.171259 | 74.366788 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.054009 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.015003 |
    | 1997-08-05 07:00:00 |  0.0 | 14.8 |              0.9 |                      1.0 |      305.954944 |             96.3 |       1.9 |              1019.0 | 0.0 | 0.0 | 17.070833 |             78.745833 |                  12.9 |                     15.470733 |           247.494792 |  0.0 | 14.8 | 17.070833 |    1.523023 |         2.074117 |           0.674482 |          1.9 |                16.859515 |                     19.494255 |                         0.0 |                          0.0 |                      1.087886 |                           1.235261 |            16.235713 |                 15.350914 |    1002.764287 |   1.225361 | -2.9 | 15.347388 |   3.284331 |            269.240351 |                      nan |                 50.001779 |                 217.795417 |                 37.987562 |                     0.0 |                      0.0 |   231.252789 |             0.0 |              0.0 |        179.807855 |             49.473684 |                 100.0 |               196.962413 |              193.257163 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.20795 |       0.0 | 0.078657 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.052614 | 0.014615 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8694.986928 |  74.28813 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.052614 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.014615 |
    | 1997-08-05 08:00:00 |  0.0 | 17.1 |              1.0 |                      1.0 |      511.104575 |             89.9 |       3.2 |              1019.0 | 0.0 | 0.0 | 17.083333 |             78.870833 |                  12.9 |                     15.470733 |           248.989764 |  0.0 | 17.1 | 17.083333 |    2.565092 |         2.097497 |           1.135969 |          3.2 |                19.530313 |                     19.509701 |                         0.0 |                          0.0 |                      1.237258 |                           1.236116 |            17.557751 |                 15.387464 |    1001.442249 |   1.215051 | -2.9 | 15.273176 | -10.960941 |            449.772026 |                      nan |                 83.529091 |                 219.110992 |                 37.896217 |                     0.0 |                      0.0 |   411.875809 |             0.0 |              0.0 |        181.214776 |                29.375 |                 100.0 |               189.861573 |              186.546902 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.442904 |       0.0 | 0.131319 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.051255 | 0.014238 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8703.047869 | 74.156811 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.051255 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.014238 |
    | 1997-08-05 09:00:00 |  0.0 | 19.4 |              0.9 |                      1.0 |      629.519239 |             79.2 |       4.0 |              1019.0 | 0.0 | 0.0 | 17.133333 |             78.883333 |                  12.8 |                     15.470733 |           249.636119 |  0.0 | 19.4 | 17.133333 |    3.206365 |         2.134237 |           1.419962 |          4.0 |                22.563931 |                     19.571593 |                         0.0 |                          0.0 |                      1.403627 |                           1.239544 |            17.870634 |                 15.438725 |    1001.129366 |   1.205358 | -2.9 | 15.369908 | -24.180553 |             553.97693 |                      nan |                 102.88143 |                 219.679785 |                  37.56462 |                     0.0 |                      0.0 |   516.412311 |             0.0 |              0.0 |        182.115165 |                  23.5 |                 100.0 |               202.878648 |              198.833875 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.679064 |       0.0 | 0.175108 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.049931 |  0.01387 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8724.328423 | 73.981703 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.049931 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.01387 |
    | 1997-08-05 10:00:00 |  0.0 | 20.6 |              1.0 |                      1.0 |      765.816589 |             73.4 |       4.6 |              1019.0 | 0.0 | 0.0 | 17.179167 |             78.783333 |                  12.9 |                     15.470733 |           254.719872 |  0.0 | 20.6 | 17.179167 |     3.68732 |         2.184336 |           1.632956 |          4.6 |                24.304197 |                     19.628477 |                         0.0 |                          0.0 |                      1.497674 |                           1.242694 |            17.839281 |                 15.463969 |    1001.160719 |   1.200448 | -2.9 | 15.625274 | -29.848354 |            673.918598 |                      nan |                125.156311 |                 224.153488 |                 37.751868 |                     0.0 |                      0.0 |   636.166731 |             0.0 |              0.0 |         186.40162 |             20.434783 |                 100.0 |                224.43901 |              219.049161 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.901936 |       0.0 | 0.201966 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.048642 | 0.013512 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8751.276776 | 73.779737 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.048642 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.013512 |
    | 1997-08-05 11:00:00 |  0.0 | 21.4 |              1.0 |                      1.0 |      767.145842 |             67.3 |       4.1 |              1019.0 | 0.0 | 0.0 |    17.225 |             78.529167 |                  13.1 |                     15.470733 |           261.673325 |  0.0 | 21.4 |    17.225 |    3.286524 |         2.211056 |           1.455461 |          4.1 |                25.528421 |                     19.685506 |                         0.0 |                          0.0 |                      1.563281 |                            1.24585 |            17.180627 |                 15.458864 |    1001.819373 |   1.197482 | -2.9 | 15.948655 | -32.708072 |            675.088341 |                      nan |                125.373549 |                 230.272526 |                 38.247049 |                     0.0 |                      0.0 |   636.841292 |             0.0 |              0.0 |        192.025477 |             22.926829 |                 100.0 |               261.392444 |              253.308714 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.943031 |       0.0 | 0.209736 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.047385 | 0.013163 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8781.084848 | 73.570001 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.047385 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.013163 |
    | 1997-08-05 12:00:00 |  0.0 | 22.5 |              1.0 |                      1.0 |      736.263853 |             62.2 |       4.1 |              1019.0 | 0.0 | 0.0 |   17.2875 |                78.275 |                  13.2 |                     15.470733 |            266.49623 |  0.0 | 22.5 |   17.2875 |    3.286524 |         2.237776 |           1.455461 |          4.1 |                27.299387 |                     19.763507 |                         0.0 |                          0.0 |                      1.657431 |                           1.250165 |            16.980219 |                 15.469885 |    1002.019781 |   1.193116 | -2.9 | 16.306352 | -37.161891 |             647.91219 |                      nan |                 120.32655 |                 234.516682 |                  38.48429 |                     0.0 |                      0.0 |     609.4279 |             0.0 |              0.0 |        196.032392 |             22.926829 |                 100.0 |               315.480499 |              302.587971 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.983449 |       0.0 | 0.196091 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.046161 | 0.012823 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8815.346739 |  73.37391 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.046161 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.012823 |
    | 1997-08-05 13:00:00 |  0.0 | 23.1 |              1.0 |                      1.0 |      768.006183 |             59.6 |       4.4 |              1019.0 | 0.0 | 0.0 | 17.341667 |             78.204167 |                  13.4 |                     15.470733 |           273.519763 |  0.0 | 23.1 | 17.341667 |    3.527001 |         2.277855 |           1.561958 |          4.4 |                28.309776 |                     19.831326 |                         0.0 |                          0.0 |                      1.710768 |                           1.253914 |            16.872626 |                 15.508923 |    1002.127374 |   1.190747 | -2.9 | 16.717494 | -38.295035 |            675.845441 |                      nan |                125.514153 |                 240.697391 |                  38.86309 |                     0.0 |                      0.0 |   636.982351 |             0.0 |              0.0 |        201.834301 |             21.363636 |                 100.0 |               357.579065 |              340.249545 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.076073 |       0.0 | 0.188131 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.044969 | 0.012491 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8850.741774 | 73.185779 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.044969 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.012491 |
    | 1997-08-05 14:00:00 |  0.0 | 23.4 |              1.0 |                      1.0 |      735.929369 |             55.4 |       4.6 |              1019.0 | 0.0 | 0.0 | 17.404167 |               77.9875 |                  13.5 |                     15.470733 |           278.618852 |  0.0 | 23.4 | 17.404167 |     3.68732 |         2.347994 |           1.632956 |          4.6 |                28.827078 |                     19.909831 |                         0.0 |                          0.0 |                      1.737973 |                           1.258253 |            15.970201 |                 15.527179 |    1003.029799 |   1.189943 | -2.9 | 17.142235 | -37.546592 |            647.617845 |                      nan |                120.271886 |                  245.18459 |                 39.079824 |                     0.0 |                      0.0 |   608.538021 |             0.0 |              0.0 |        206.104766 |             20.434783 |                 100.0 |               429.695868 |              403.397707 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.117532 |       0.0 | 0.165058 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.043807 | 0.012169 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8885.388366 | 73.020721 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.043807 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.012169 |
    | 1997-08-05 15:00:00 |  0.0 | 23.8 |              1.0 |                      1.0 |      620.374953 |             52.1 |       4.5 |              1019.0 | 0.0 | 0.0 |   17.4875 |             77.708333 |                  13.6 |                     15.470733 |           282.921682 |  0.0 | 23.8 |   17.4875 |    3.607161 |         2.401434 |           1.597457 |          4.5 |                29.529613 |                     20.014927 |                         0.0 |                          0.0 |                      1.774812 |                           1.264057 |            15.384928 |                 15.553266 |    1003.615072 |     1.1886 | -2.9 | 17.557994 | -37.452038 |            545.929958 |                      nan |                101.386992 |                  248.97108 |                 39.288738 |                     0.0 |                      0.0 |    506.64122 |             0.0 |              0.0 |        209.682342 |             20.888889 |                 100.0 |               525.323496 |              484.570912 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  1.044853 |       0.0 | 0.135831 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.042676 | 0.011854 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8919.940404 |  72.88489 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.042676 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.011854 |
    | 1997-08-05 16:00:00 |  0.0 | 24.0 |              1.0 |                      1.0 |      450.095264 |             49.5 |       4.7 |              1018.0 | 0.0 | 0.0 |   17.5625 |               77.3875 |                  13.7 |                     15.470733 |           285.932459 |  0.0 | 24.0 |   17.5625 |    3.767479 |         2.438173 |           1.668455 |          4.7 |                29.886439 |                     20.109928 |                         0.0 |                          0.0 |                      1.793476 |                             1.2693 |            14.793787 |                  15.56257 |    1003.206213 |   1.186889 | -2.9 | 17.972618 | -36.164291 |            396.083832 |                      nan |                 73.558426 |                 251.620564 |                 39.539407 |                     0.0 |                      0.0 |   356.544425 |             0.0 |              0.0 |        212.081156 |                  20.0 |                 100.0 |               627.903494 |              568.555525 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.943219 |       0.0 | 0.103426 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.041573 | 0.011548 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8953.204695 | 72.781464 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.041573 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.011548 |
    | 1997-08-05 17:00:00 |  0.0 | 24.0 |              1.0 |                      1.0 |      273.614142 |             48.1 |       4.5 |              1018.0 | 0.0 | 0.0 | 17.654167 |             77.008333 |                  13.7 |                     15.470733 |           286.824521 |  0.0 | 24.0 | 17.654167 |    3.607161 |         2.448193 |           1.597457 |          4.5 |                29.886439 |                     20.226575 |                         0.0 |                          0.0 |                      1.793476 |                           1.275734 |            14.375377 |                 15.576148 |    1003.624623 |   1.187075 | -2.9 |  18.37179 | -33.769262 |            240.780445 |                      nan |                 44.716368 |                 252.405579 |                 39.564612 |                     0.0 |                      0.0 |   201.215833 |             0.0 |              0.0 |        212.840967 |             20.888889 |                 100.0 |               687.347517 |              615.826756 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.773212 |       0.0 | 0.081876 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.040499 |  0.01125 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 8984.073957 | 72.699588 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.040499 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.01125 |
    | 1997-08-05 18:00:00 |  0.0 | 23.7 |              1.0 |                      1.0 |      133.368106 |             49.4 |       4.1 |              1018.0 | 0.0 | 0.0 | 17.770833 |             76.570833 |                  13.7 |                     15.470733 |           287.024163 |  0.0 | 23.7 | 17.770833 |    3.286524 |         2.444853 |           1.455461 |          4.1 |                29.352596 |                      20.37589 |                         0.0 |                          0.0 |                      1.765541 |                           1.283961 |            14.500182 |                 15.601989 |    1003.499818 |   1.188219 | -2.9 | 18.742221 | -29.746675 |            117.363933 |                      nan |                 21.796159 |                 252.581263 |                 39.572333 |                     0.0 |                      0.0 |      77.7916 |             0.0 |              0.0 |         213.00893 |             22.926829 |                 100.0 |               600.033284 |              546.044291 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.580667 |       0.0 | 0.073141 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.039453 | 0.010959 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 9010.920632 | 72.626447 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.039453 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.010959 |
    | 1997-08-05 19:00:00 |  0.0 | 22.8 |              1.0 |                      1.0 |       40.326503 |             57.0 |       3.8 |              1018.0 | 0.0 | 0.0 |      17.9 |             76.191667 |                  13.7 |                     15.470733 |           286.951409 |  0.0 | 22.8 |      17.9 |    3.046047 |         2.451533 |           1.348964 |          3.8 |                27.800581 |                     20.542326 |                         0.0 |                          0.0 |                      1.683922 |                           1.293123 |            15.846331 |                 15.651541 |    1002.153669 |   1.191234 | -2.9 | 19.064381 | -22.413715 |             35.487322 |                      nan |                  6.590503 |                  252.51724 |                 39.521285 |                     0.0 |                      0.0 |    -4.033963 |             0.0 |              0.0 |        212.995955 |             24.736842 |                 100.0 |               384.257428 |              363.808832 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.383669 |       0.0 | 0.070396 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.038434 | 0.010676 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 9030.434346 | 72.556051 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.038434 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.010676 |
    | 1997-08-05 20:00:00 |  0.0 | 21.3 |              0.5 |                 0.072689 |        0.867122 |             65.9 |       2.5 |              1018.0 | 0.0 | 0.0 | 18.041667 |                 75.95 |                  13.9 |                     15.438712 |           286.953669 |  0.0 | 21.3 | 18.041667 |    2.003978 |         2.428153 |           0.887476 |          2.5 |                 25.37251 |                     20.726234 |                         0.0 |                          0.0 |                       1.55495 |                           1.303234 |            16.720484 |                 15.741575 |    1001.279516 |   1.196911 | -2.9 | 19.298545 | -12.008727 |              0.763068 |                      nan |                  0.141713 |                 252.519228 |                 39.876809 |                     0.0 |                      0.0 |   -39.113742 |             0.0 |              0.0 |        212.642419 |                  37.6 |                 100.0 |               272.687161 |               73.373254 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.163068 |       0.0 | 0.099133 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.037441 |   0.0104 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 9039.543073 | 72.456918 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.037441 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |   0.0104 |
    | 1997-08-05 21:00:00 |  0.0 | 19.3 |              0.0 |                      0.0 |             0.0 |             73.0 |       1.9 |              1018.0 | 0.0 | 0.0 |   18.1875 |             75.633333 |                  13.9 |                     15.438712 |           286.953669 |  0.0 | 19.3 |   18.1875 |    1.523023 |         2.418134 |           0.674482 |          1.9 |                22.423949 |                     20.917052 |                         0.0 |                          0.0 |                       1.39602 |                           1.313713 |            16.369483 |                 15.820264 |    1001.630517 |   1.205255 | -2.9 |  19.40785 |   0.647101 |                   0.0 |                      nan |                       0.0 |                 252.519228 |                 39.753451 |                     0.0 |                      0.0 |   -39.753451 |             0.0 |              0.0 |        212.765778 |             49.473684 |                 100.0 |               228.579054 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.080796 |       0.0 | 0.053472 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.036474 | 0.010132 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 9035.995972 | 72.403446 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.036474 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.010132 |
    | 1997-08-05 22:00:00 |  0.0 | 18.3 |              0.0 |                      0.0 |             0.0 |             76.7 |       2.2 |              1018.0 | 0.0 | 0.0 | 18.333333 |                 75.25 |                  13.9 |                     15.438712 |           286.953669 |  0.0 | 18.3 | 18.333333 |    1.763501 |         2.418134 |           0.780979 |          2.2 |                21.065302 |                     21.109404 |                         0.0 |                          0.0 |                      1.321845 |                           1.324263 |            16.157086 |                 15.884826 |    1001.842914 |   1.209486 | -2.9 | 19.365285 |    6.39171 |                   0.0 |                      nan |                       0.0 |                 252.519228 |                 39.670785 |                     0.0 |                      0.0 |   -39.670785 |             0.0 |              0.0 |        212.848443 |             42.727273 |                 100.0 |               220.434933 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.076273 |       0.0 | 0.047609 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.035532 |  0.00987 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 9026.704262 | 72.355838 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.035532 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 |  0.00987 |
    | 1997-08-05 23:00:00 |  0.0 | 17.4 |              0.0 |                      0.0 |             0.0 |             81.8 |       2.7 |              1018.0 | 0.0 | 0.0 | 18.470833 |               74.9375 |                  13.9 |                     15.438712 |           286.953669 |  0.0 | 17.4 | 18.470833 |    2.164296 |         2.438173 |           0.958474 |          2.7 |                19.904589 |                     21.292177 |                         0.0 |                          0.0 |                      1.257963 |                           1.334275 |            16.281953 |                 15.955825 |    1001.718047 |   1.213176 | -2.9 | 19.253784 |  11.122707 |                   0.0 |                      nan |                       0.0 |                 252.519228 |                 39.564038 |                     0.0 |                      0.0 |   -39.564038 |             0.0 |              0.0 |        212.955191 |             34.814815 |                 100.0 |               211.422451 |               69.444444 |  0.0 |  0.0 |       49.0 |        0.42 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.067289 |       0.0 | 0.038507 | 0.0 |   0.0 |          nan |         0.12 |     0.0 |  0.0 |     0.0 |  0.0 |     0.0 |      0.0 |      0.0 |       0.0 |      0.0 |   0.0 | 0.0 | 0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.0 |  0.0 |  0.0 | 0.034614 | 0.009615 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |  0.0 |  0.0 |   0.0 |  nan | 9012.681555 |  72.31733 |   0.0 |   0.0 |   0.0 |   0.0 |  0.0 | 0.034614 |   0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.009615 |

>>> round_(model.check_waterbalance(conditions_nadelw_summer))
0.0

.. _lland_v4_acker_winter_hourly:

acre (winter)
-------------

>>> pub.timegrids = "2010-12-10", "2010-12-13", "1h"

>>> inputs.nied.series = (
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...     0.1, 0.0, 0.3, 0.2, 0.4, 0.7, 0.5, 1.0, 2.0, 2.1, 1.8, 1.3, 0.5, 0.2, 0.5, 0.2,
...     0.1, 0.3, 0.1, 0.1, 0.0, 0.1, 0.0, 0.1, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.1, 0.1,
...     0.0, 0.0, 0.1, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.1, 0.0, 0.0, 0.0, 0.0, 0.0,
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0)
>>> inputs.teml.series = (
...     -3.7, -4.8, -5.6, -6.2, -7.3, -8.3, -8.6, -8.5, -9.5, -9.2, -6.9, -4.0, -2.8,
...     -1.4, -0.4, 0.5, 1.0, 1.1, 1.0, 1.4, 1.6, 2.0, 2.4, 2.5, 2.7, 3.2, 3.9, 4.4,
...     4.8, 4.8, 5.1, 5.1, 5.4, 5.3, 5.5, 6.1, 6.4, 6.4, 6.5, 6.3, 5.9, 5.5, 5.1, 4.8,
...     4.9, 4.9, 4.7, 3.8, 3.9, 3.7, 3.5, 3.3, 3.3, 3.0, 1.9, 1.2, 0.8, 0.8, 0.5, 0.3,
...     -0.1, -0.4, -0.6, -0.8, -1.3, -1.6, -2.0, -2.1, -2.2, -2.6, -3.1, -3.9)
>>> inputs.sunshineduration.series = (
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.1, 0.0, 0.0, 0.0, 0.0,
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.1, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.2, 0.2,
...     0.2, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0)
>>> inputs.windspeed.series = (
...     3.4, 2.9, 2.5, 3.0, 2.5, 2.3, 2.1, 1.4, 1.1, 1.0, 1.4, 1.3, 1.3, 1.4, 2.3, 2.8,
...     3.1, 3.5, 4.4, 3.8, 5.7, 5.6, 5.8, 6.2, 5.6, 5.2, 5.7, 6.6, 7.0, 7.3, 7.2, 7.6,
...     7.3, 7.1, 7.1, 8.9, 9.5, 7.9, 9.6, 9.1, 8.7, 8.9, 7.9, 7.1, 6.4, 7.2, 7.0, 6.8,
...     6.4, 6.3, 5.7, 5.2, 4.6, 4.9, 6.3, 5.8, 6.1, 5.9, 6.4, 6.8, 7.0, 6.3, 7.3, 7.5,
...     7.0, 6.0, 4.8, 4.2, 4.9, 4.0, 3.4, 3.4)
>>> inputs.relativehumidity.series = (
...     87.5, 90.7, 89.7, 91.3, 94.2, 93.2, 96.5, 95.9, 96.8, 94.3, 96.7, 97.4, 93.9,
...     93.4, 93.3, 94.1, 94.6, 94.4, 96.9, 98.3, 98.7, 99.6, 99.4, 99.2, 99.2, 99.6,
...     99.8, 99.8, 99.0, 97.6, 97.4, 96.3, 95.2, 96.3, 95.5, 91.6, 90.0, 85.7, 82.3,
...     82.5, 82.2, 76.5, 81.7, 83.4, 82.8, 85.8, 87.6, 85.6, 85.3, 86.2, 89.3, 91.4,
...     90.9, 89.2, 85.6, 86.3, 91.2, 87.5, 84.5, 76.7, 74.8, 69.7, 66.9, 64.4, 65.4,
...     67.6, 70.2, 70.4, 69.5, 71.6, 74.5, 80.2)
>>> inputs.atmosphericpressure.series = (
...     1017.0, 1018.0, 1018.0, 1019.0, 1020.0, 1020.0, 1021.0, 1021.0, 1022.0, 1022.0,
...     1022.0, 1022.0, 1022.0, 1021.0, 1020.0, 1020.0, 1019.0, 1018.0, 1016.0, 1015.0,
...     1014.0, 1013.0, 1011.0, 1010.0, 1008.0, 1007.0, 1006.0, 1005.0, 1005.0, 1004.0,
...     1004.0, 1004.0, 1003.0, 1003.0, 1003.0, 1002.0, 1002.0, 1002.0, 1002.0, 1003.0,
...     1003.0, 1003.0, 1003.0, 1003.0, 1003.0, 1003.0, 1003.0, 1002.0, 1002.0, 1002.0,
...     1002.0, 1003.0, 1004.0, 1005.0, 1006.0, 1008.0, 1009.0, 1010.0, 1011.0, 1012.0,
...     1013.0, 1014.0, 1015.0, 1015.0, 1017.0, 1017.0, 1018.0, 1019.0, 1019.0, 1020.0,
...     1020.0, 1020.0)

>>> inputs.possiblesunshineduration.series = (
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.35523162, 1.0, 1.0, 1.0, 1.0, 1.0,
...     1.0, 1.0, 0.06189829, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...     0.0, 0.0, 0.0, 0.0, 0.34704123, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0537079,
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...     0.33973828, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.04640495, 0.0, 0.0, 0.0, 0.0,
...     0.0, 0.0, 0.0)
>>> inputs.globalradiation.series = (
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.1076111, 18.6632028, 56.2679083,
...     107.1682361, 91.4733333, 80.4755361, 43.1205917, 10.3200917, 0.0281028, 0.0,
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...     0.9007333, 17.2675111, 52.6176111, 80.1356333, 80.2761167, 75.8354028,
...     42.2279111, 12.6139194, 0.0219111, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...     0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.8941444, 17.7125389, 54.3025861,
...     81.5041472, 81.6652944, 77.0789583, 60.2540417, 14.8373, 0.1688556, 0.0, 0.0,
...     0.0, 0.0, 0.0, 0.0, 0.0)

>>> test.inits = (
...     (logs.loggedsunshineduration,
...      [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.04, 0.25, 0.59, 0.91,
...       0.97, 1.0, 0.65, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
...     (logs.loggedpossiblesunshineduration,
...      [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.4, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
...       1.0, 0.1, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
...     (logs.loggedglobalradiation,
...      [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 27.8, 55.6, 111.1, 166.7, 138.9,
...       55.6, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]),
...     (logs.loggedteml,
...      [-2.8, -2.6, -2.2, -1.8, -1.8, -2.4, -1.5, -1.4, -0.8, -0.8, -0.9, 0.9, -0.8,
...       -1.0, -1.2, -1.8, -2.8, -3.8, -4.3, -4, -3.9, -1.6, -1.6, 2.6]),
...     (logs.loggedrelativehumidity,
...      [99.4, 99.0, 99.6, 99.6, 99.3, 99.3, 99.8, 99.7, 100.0, 99.6, 99.3, 98.9,
...       96.6, 93.0, 91.4, 92.0, 94.5, 96.5, 98.2, 96.9, 98.0, 95.0, 91.8, 88.5]),
...     (logs.loggedwindspeed2m,
...      [2.2, 2.0, 2.5, 2.6, 2.3, 1.7, 2.8, 1.9, 2.6, 2.9, 3.5, 3.4, 3.5, 3.0, 3.9,
...       3.6, 2.2, 2.4, 2.5, 1.5, 2.5, 3.0, 3.2, 3.1]))

.. integration-test::

    >>> lnk(ACKER)
    >>> test("lland_v4_acker_winter_hourly",
    ...      axis1=(inputs.nied, fluxes.wada), axis2=(states.waes, states.wats),
    ...      use_conditions=conditions_acker_winter)
    |                date | nied | teml | sunshineduration | possiblesunshineduration | globalradiation | relativehumidity | windspeed | atmosphericpressure |  qz | qzh |   temltag | dailyrelativehumidity | dailysunshineduration | dailypossiblesunshineduration | dailyglobalradiation |  nkor | tkor |   tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair |        g |        tz |         wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |     nbes |    sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |      evpo |      evi |      evb |       evs |    wnied | tempssurface | actualalbedo |  schmpot |     schm |  gefrpot |     gefr | wlatinz |   wlatsnow | wsensinz |  wsenssnow | wsurfinz |       wsurf |      sff |      fvg |     wada |      qdb | qib1 | qib2 | qbb | qkap |     qdgz |      qah |       qa |     inzp | stinz | sinz | esnowinz | asinz |      wats |      waes |       esnow |     taus |        ebdn |      bowa |    qdgz1 |    qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 |    qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    ---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 2010-12-10 00:00:00 |  0.0 | -3.7 |              0.0 |                      0.0 |             0.0 |             87.5 |       3.4 |              1017.0 | 0.0 | 0.0 |    -2.025 |             96.870833 |                  4.41 |                           7.5 |            23.154167 |   0.0 | -3.7 |    -2.025 |     2.72541 |         2.684392 |            2.72541 |          3.4 |                  4.64297 |                      5.262329 |                         0.0 |                     4.116122 |                      0.349621 |                           0.390562 |             4.062599 |                  5.097662 |    1012.937401 |   1.312855 | 8.583333 | -5.337255 |  -0.292799 |                   0.0 |                      0.0 |                       0.0 |                   5.688401 |                 46.101494 |                     0.0 |                73.927318 |   -46.101494 |             0.0 |       -73.927318 |        -40.413093 |             62.648725 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.001998 |      0.0 |      0.0 |  0.000809 |      0.0 |    -5.288455 |     0.754325 |      0.0 |      0.0 | 0.697534 | 0.697534 |     0.0 |   0.599182 |      0.0 | -10.103618 |      0.0 |   64.422883 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.320358 | 20.349008 |         0.0 | 0.860328 | -435.895126 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 01:00:00 |  0.0 | -4.8 |              0.0 |                      0.0 |             0.0 |             90.7 |       2.9 |              1018.0 | 0.0 | 0.0 | -2.158333 |                96.825 |                  4.41 |                           7.5 |            23.154167 |   0.0 | -4.8 | -2.158333 |    2.324615 |         2.647918 |           2.324615 |          2.9 |                 4.272186 |                      5.210481 |                         0.0 |                     3.961569 |                      0.324793 |                           0.387158 |             3.874873 |                  5.045049 |    1014.125127 |   1.319627 | 8.583333 | -5.230742 |   3.347272 |                   0.0 |                      0.0 |                       0.0 |                   5.703742 |                 46.220713 |                     0.0 |                76.610661 |   -46.220713 |             0.0 |       -76.610661 |         -40.51697 |              73.45023 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.004391 |      0.0 |      0.0 |  0.001178 |      0.0 |     -5.78862 |     0.753662 |      0.0 |      0.0 | 0.738127 |  0.02865 |     0.0 |   0.872698 |      0.0 |  -5.654298 |      0.0 |   71.829062 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.347831 | 20.347831 |  -65.823721 | 0.867636 | -430.659065 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 02:00:00 |  0.0 | -5.6 |              0.0 |                      0.0 |             0.0 |             89.7 |       2.5 |              1018.0 | 0.0 | 0.0 |    -2.325 |             96.604167 |                  4.41 |                           7.5 |            23.154167 |   0.0 | -5.6 |    -2.325 |    2.003978 |         2.606417 |           2.003978 |          2.5 |                 4.019239 |                      5.146307 |                         0.0 |                     3.241256 |                      0.307706 |                            0.38294 |             3.605258 |                  4.971547 |    1014.394742 |   1.323706 | 8.583333 | -5.167909 |  19.248028 |                   0.0 |                      0.0 |                       0.0 |                   5.709269 |                  46.40442 |                     0.0 |                70.157201 |    -46.40442 |             0.0 |       -70.157201 |        -40.695151 |             85.202267 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.004461 |      0.0 |      0.0 | -0.004502 |      0.0 |    -8.375913 |     0.753424 |      0.0 |      0.0 | 1.066474 |      0.0 |     0.0 |  -3.335424 |      0.0 | -14.452419 |      0.0 |   52.369359 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.352332 | 20.352332 |  -98.945052 | 0.870283 |  -441.32376 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 03:00:00 |  0.0 | -6.2 |              0.0 |                      0.0 |             0.0 |             91.3 |       3.0 |              1019.0 | 0.0 | 0.0 | -2.420833 |                96.325 |                  4.41 |                           7.5 |            23.154167 |   0.0 | -6.2 | -2.420833 |    2.404774 |         2.602449 |           2.404774 |          3.0 |                 3.838328 |                      5.109724 |                         0.0 |                     2.929796 |                      0.295404 |                           0.380532 |             3.504393 |                  4.921942 |    1015.495607 |   1.328036 | 8.583333 | -5.295885 |  26.162948 |                   0.0 |                      0.0 |                       0.0 |                   5.713719 |                 46.542687 |                     0.0 |                67.439656 |   -46.542687 |             0.0 |       -67.439656 |        -40.828968 |             71.001889 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.004771 |      0.0 |      0.0 | -0.007981 |      0.0 |    -9.656376 |     0.753232 |      0.0 |      0.0 | 1.229782 |      0.0 |     0.0 |  -5.913667 |      0.0 | -20.211638 |      0.0 |   41.314351 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.360313 | 20.360313 | -114.096455 |  0.87242 | -458.903374 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 04:00:00 |  0.0 | -7.3 |              0.0 |                      0.0 |             0.0 |             94.2 |       2.5 |              1020.0 | 0.0 | 0.0 | -2.558333 |               96.2125 |                  4.41 |                           7.5 |            23.154167 |   0.0 | -7.3 | -2.558333 |    2.003978 |         2.623448 |           2.003978 |          2.5 |                  3.52531 |                      5.057637 |                         0.0 |                     2.688706 |                      0.273951 |                           0.377101 |             3.320842 |                  4.866079 |    1016.679158 |   1.334932 | 8.583333 |  -5.50684 |  31.359309 |                   0.0 |                      0.0 |                       0.0 |                   5.717779 |                 46.677297 |                     0.0 |                67.802576 |   -46.677297 |             0.0 |       -67.802576 |        -40.959519 |             85.202267 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.008511 |      0.0 |      0.0 | -0.007817 |      0.0 |   -10.733392 |     0.753056 |      0.0 |      0.0 | 1.367481 |      0.0 |     0.0 |  -5.792391 |      0.0 | -17.875492 |      0.0 |   44.134693 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.368131 | 20.368131 | -126.871838 | 0.874373 | -481.679351 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 05:00:00 |  0.0 | -8.3 |              0.0 |                      0.0 |             0.0 |             93.2 |       2.3 |              1020.0 | 0.0 | 0.0 |    -2.725 |             96.004167 |                  4.41 |                           7.5 |            23.154167 |   0.0 | -8.3 |    -2.725 |     1.84366 |         2.596101 |            1.84366 |          2.3 |                  3.26061 |                      4.995131 |                         0.0 |                     2.483084 |                       0.25563 |                           0.372977 |             3.038889 |                  4.795534 |    1016.961111 |   1.340112 | 8.583333 | -5.780152 |    35.6513 |                   0.0 |                      0.0 |                       0.0 |                   5.721539 |                 46.854961 |                     0.0 |                69.171166 |   -46.854961 |             0.0 |       -69.171166 |        -41.133421 |             92.611159 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.007601 |      0.0 |      0.0 | -0.006535 |      0.0 |   -11.722035 |     0.752894 |      0.0 |      0.0 | 1.494012 |      0.0 |     0.0 |   -4.84203 |      0.0 | -16.938582 |      0.0 |   47.390554 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.374665 | 20.374665 | -138.611093 | 0.876187 | -508.747317 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 06:00:00 |  0.0 | -8.6 |              0.0 |                      0.0 |             0.0 |             96.5 |       2.1 |              1021.0 | 0.0 | 0.0 |    -2.925 |             96.004167 |                  4.41 |                           7.5 |            23.154167 |   0.0 | -8.6 |    -2.925 |    1.683342 |          2.56624 |           1.683342 |          2.1 |                 3.184716 |                      4.921026 |                         0.0 |                     2.368455 |                      0.250344 |                           0.368079 |             3.073251 |                   4.72439 |    1017.926749 |   1.342932 | 8.583333 | -6.104968 |  37.201561 |                   0.0 |                      0.0 |                       0.0 |                   5.725046 |                 47.012175 |                     0.0 |                67.356137 |   -47.012175 |             0.0 |       -67.356137 |         -41.28713 |             101.43127 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.010208 |      0.0 |      0.0 | -0.007857 |      0.0 |   -12.305228 |     0.752742 |      0.0 |      0.0 | 1.568845 |      0.0 |     0.0 |   -5.82182 |      0.0 | -17.389919 |      0.0 |   44.144398 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.382523 | 20.382523 | -145.553929 |  0.87788 | -537.365545 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 07:00:00 |  0.0 | -8.5 |              0.0 |                      0.0 |             0.0 |             95.9 |       1.4 |              1021.0 | 0.0 | 0.0 |   -3.1625 |               96.0625 |                  3.76 |                           7.4 |            23.154167 |   0.0 | -8.5 |   -3.1625 |    1.122228 |         2.521333 |           1.122228 |          1.4 |                 3.209838 |                      4.834291 |                         0.0 |                     2.300334 |                      0.252096 |                           0.362336 |             3.078235 |                  4.643941 |    1017.921765 |   1.342422 | 8.583333 | -6.448387 |  37.292525 |                   0.0 |                      0.0 |                       0.0 |                   5.728408 |                  42.68596 |                     0.0 |                62.592321 |    -42.68596 |             0.0 |       -62.592321 |        -36.957552 |            152.146905 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.008397 |      0.0 |      0.0 | -0.007013 |      0.0 |   -12.663807 |     0.752597 |      0.0 |      0.0 | 1.615184 |      0.0 |     0.0 |  -5.196531 |      0.0 | -15.803999 |      0.0 |   41.591791 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.389536 | 20.389536 | -149.853196 | 0.879508 | -566.074736 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 08:00:00 |  0.0 | -9.5 |              0.0 |                 0.355232 |        1.107611 |             96.8 |       1.1 |              1022.0 | 0.0 | 0.0 | -3.483333 |               96.2625 |                  2.76 |                      6.755232 |            23.200317 |   0.0 | -9.5 | -3.483333 |     0.88175 |         2.408072 |            0.88175 |          1.1 |                 2.966351 |                       4.71927 |                         0.0 |                     2.207601 |                      0.235051 |                           0.354699 |             2.871427 |                  4.542888 |    1019.128573 |   1.348938 | 8.583333 | -6.792897 |  38.247153 |              0.274183 |                      0.0 |                  0.274183 |                   5.743107 |                 37.241518 |                     0.0 |                61.028121 |   -36.967335 |             0.0 |       -60.753938 |         -31.49841 |            193.641515 |                 100.0 |               112.102145 |               99.616909 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.006538 |      0.0 |      0.0 | -0.005378 |      0.0 |   -13.167422 |     0.752456 |      0.0 |      0.0 | 1.679995 |      0.0 |     0.0 |  -3.984955 |      0.0 | -12.508846 |      0.0 |   44.260137 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.394914 | 20.394914 | -155.866179 | 0.881095 | -595.738556 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 09:00:00 |  0.0 | -9.2 |              0.0 |                      1.0 |       18.663203 |             94.3 |       1.0 |              1022.0 | 0.0 | 0.0 | -3.816667 |             96.383333 |                  1.79 |                      6.755232 |            21.661284 |   0.0 | -9.2 | -3.816667 |    0.801591 |         2.278972 |           0.801591 |          1.0 |                 3.037615 |                      4.602339 |                         0.0 |                     2.243104 |                      0.240056 |                           0.346913 |             2.864471 |                  4.435888 |    1019.135529 |   1.347408 | 8.583333 | -7.148863 |  34.941383 |              4.622516 |                      0.0 |                  4.622516 |                   5.365082 |                 29.260785 |                     0.0 |                53.269158 |    -24.63827 |             0.0 |       -48.646643 |        -23.895703 |            213.005666 |                 100.0 |               112.102145 |              101.107639 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.002319 |      0.0 |      0.0 | -0.004845 |      0.0 |   -12.972427 |     0.752319 |      0.0 |      0.0 | 1.655552 |      0.0 |     0.0 |  -3.589807 |      0.0 | -12.383164 |      0.0 |   32.673671 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.399759 | 20.399759 | -153.598467 | 0.882628 | -622.096606 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 10:00:00 |  0.0 | -6.9 |              0.0 |                      1.0 |       56.267908 |             96.7 |       1.4 |              1022.0 | 0.0 | 0.0 |   -4.0625 |               96.5375 |                  0.88 |                      6.755232 |             18.21828 |   0.0 | -6.9 |   -4.0625 |    1.122228 |         2.200731 |           1.122228 |          1.4 |                 3.636412 |                      4.517752 |                         0.0 |                     2.591489 |                      0.281591 |                           0.341265 |             3.516411 |                  4.361325 |    1018.483589 |   1.335446 | 8.583333 | -7.465159 |   22.36178 |             13.944241 |                      0.0 |                 13.944241 |                   4.514831 |                 21.675213 |                     0.0 |                 37.72771 |    -7.730972 |             0.0 |       -23.783469 |        -17.160382 |            152.146905 |                 100.0 |               112.102145 |              101.107639 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.002109 |      0.0 |      0.0 | -0.008339 |      0.0 |   -11.192123 |     0.752181 |      0.0 |      0.0 | 1.428688 |      0.0 |     0.0 |  -6.178658 |      0.0 | -16.291028 |      0.0 |    1.313783 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.408097 | 20.408097 | -132.550471 | 0.884182 | -635.875052 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 11:00:00 |  0.0 | -4.0 |              0.1 |                      1.0 |      107.168236 |             97.4 |       1.3 |              1022.0 | 0.0 | 0.0 | -4.195833 |             96.570833 |                  0.39 |                      6.755232 |             15.73779 |   0.0 | -4.0 | -4.195833 |    1.042069 |         2.098318 |           1.042069 |          1.3 |                 4.539126 |                      4.472452 |                         0.0 |                     3.202943 |                      0.342693 |                           0.338235 |             4.421109 |                  4.319085 |    1017.578891 |   1.320615 | 8.583333 | -7.630501 |   5.381266 |             26.574984 |                      0.0 |                 26.574984 |                    3.90257 |                 17.574343 |                     0.0 |                24.904358 |     9.000642 |             0.0 |         1.670627 |        -13.671773 |            163.850513 |                 100.0 |               112.102145 |              101.107639 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.008174 |      0.0 |      0.0 | -0.010611 |      0.0 |    -8.527378 |     0.752026 |      0.0 |      0.0 | 1.088975 |      0.0 |     0.0 |  -7.862607 |      0.0 | -16.603299 |      0.0 |  -26.136533 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.418709 | 20.418709 | -101.032672 | 0.885939 | -632.672985 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 12:00:00 |  0.0 | -2.8 |              0.0 |                      1.0 |       91.473333 |             93.9 |       1.3 |              1022.0 | 0.0 | 0.0 |     -4.35 |               96.3625 |                  0.14 |                      6.755232 |            14.920012 |   0.0 | -2.8 |     -4.35 |    1.042069 |          2.00007 |           1.042069 |          1.3 |                 4.967227 |                      4.420576 |                         0.0 |                     3.625473 |                      0.371134 |                            0.33476 |             4.664226 |                  4.259778 |    1017.335774 |   1.314635 | 8.583333 | -7.592076 |  -3.919048 |             22.700159 |                      0.0 |                 22.700159 |                   3.702573 |                 15.511434 |                     0.0 |                21.886755 |     7.188725 |             0.0 |         0.813404 |        -11.808862 |            163.850513 |                 100.0 |               112.102145 |              101.107639 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.010216 |      0.0 |      0.0 | -0.009049 |      0.0 |    -6.938901 |     0.751839 |      0.0 |      0.0 | 0.886582 |      0.0 |     0.0 |  -6.704598 |      0.0 | -15.178632 |      0.0 |  -22.696634 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.427757 | 20.427757 |  -82.255086 | 0.888053 | -620.170603 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 13:00:00 |  0.0 | -1.4 |              0.0 |                      1.0 |       80.475536 |             93.4 |       1.4 |              1021.0 | 0.0 | 0.0 | -4.370833 |             96.116667 |                   0.1 |                      6.755232 |            15.956493 |   0.0 | -1.4 | -4.370833 |    1.122228 |         1.900997 |           1.122228 |          1.4 |                 5.511494 |                      4.413607 |                         0.0 |                     4.003734 |                      0.406861 |                           0.334293 |             5.147736 |                  4.242212 |    1015.852264 |   1.306346 | 8.583333 | -7.442047 | -10.749432 |             19.987784 |                      0.0 |                 19.987784 |                   3.963129 |                 15.193165 |                     0.0 |                17.769887 |     4.794619 |             0.0 |         2.217896 |        -11.230036 |            152.146905 |                 100.0 |               112.102145 |              101.107639 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.01068 |      0.0 |      0.0 | -0.010314 |      0.0 |    -5.650475 |     0.751629 |      0.0 |      0.0 |  0.72228 |      0.0 |     0.0 |  -7.642151 |      0.0 | -16.132952 |      0.0 |     -25.993 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.438071 | 20.438071 |  -67.011518 | 0.890426 | -600.837838 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 14:00:00 |  0.0 | -0.4 |              0.0 |                      1.0 |       43.120592 |             93.3 |       2.3 |              1020.0 | 0.0 | 0.0 | -4.354167 |             95.854167 |                   0.1 |                      6.755232 |            16.594851 |   0.0 | -0.4 | -4.354167 |     1.84366 |         1.856982 |            1.84366 |          2.3 |                 5.931876 |                      4.419181 |                         0.0 |                     4.318486 |                      0.434155 |                           0.334667 |              5.53444 |                   4.23597 |     1014.46556 |   1.300092 | 8.583333 | -7.210054 | -15.311558 |             10.719882 |                      0.0 |                 10.719882 |                   4.125519 |                 15.206659 |                     0.0 |                15.337726 |    -4.486777 |             0.0 |        -4.617844 |         -11.08114 |             92.611159 |                 100.0 |               112.102145 |              101.107639 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.010323 |      0.0 |      0.0 | -0.014296 |      0.0 |    -4.658128 |     0.751398 |      0.0 |      0.0 | 0.595732 |      0.0 |     0.0 | -10.593086 |      0.0 | -21.077118 |      0.0 |   -27.05236 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.452368 | 20.452368 |  -55.270716 | 0.893056 | -576.942947 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 15:00:00 |  0.0 |  0.5 |              0.0 |                      1.0 |       10.320092 |             94.1 |       2.8 |              1020.0 | 0.0 | 0.0 |      -4.3 |             95.608333 |                   0.1 |                      7.355232 |            17.024855 |   0.0 |  0.5 |      -4.3 |    2.244455 |         1.842168 |           2.244455 |          2.8 |                 6.334173 |                      4.437342 |                         0.0 |                     4.592753 |                      0.460051 |                           0.335883 |             5.960456 |                  4.242469 |    1014.039544 |   1.295611 | 8.583333 | -6.923315 | -18.473943 |              2.568208 |                      0.0 |                  2.568208 |                   4.236722 |                  15.14111 |                     0.0 |                12.218458 |   -12.572902 |             0.0 |         -9.65025 |        -10.904388 |             76.073452 |                 100.0 |               112.102145 |              101.107639 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.007072 |      0.0 |      0.0 | -0.018164 |      0.0 |    -3.844325 |     0.751145 |      0.0 |      0.0 | 0.491998 |      0.0 |     0.0 | -13.458731 |      0.0 |  -24.28968 |      0.0 |  -28.098161 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.470532 | 20.470532 |  -45.646498 | 0.895936 | -549.885671 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 16:00:00 |  0.1 |  1.0 |              0.0 |                 0.061898 |        0.028103 |             94.6 |       3.1 |              1019.0 | 0.0 | 0.0 |      -4.2 |             95.395833 |                   0.1 |                       7.41713 |            17.026026 | 0.094 |  1.0 |      -4.2 |    2.484933 |          1.86654 |           2.484933 |          3.1 |                 6.567919 |                      4.471043 |                         0.0 |                     4.829803 |                      0.475001 |                            0.33814 |             6.213251 |                  4.265189 |    1012.786749 |   1.291856 | 8.583333 | -6.598628 | -20.542414 |              0.006996 |                      0.0 |                  0.006996 |                   4.238586 |                 15.125699 |                     0.0 |                11.447885 |   -15.118703 |             0.0 |       -11.440889 |        -10.887113 |             68.711505 |                 100.0 |               112.102145 |               98.953162 |    0.094 |  0.0235 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.005889 |      0.0 |      0.0 | -0.019637 | 0.095635 |    -3.174892 |     0.751053 |      0.0 |      0.0 | 0.409957 |   0.0705 |     0.0 | -14.550516 |      0.0 | -24.948708 |      0.0 |  -28.058335 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.584169 | 20.584169 |  -31.494109 |  0.89699 | -520.759923 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 17:00:00 |  0.0 |  1.1 |              0.0 |                      0.0 |             0.0 |             94.4 |       3.5 |              1018.0 | 0.0 | 0.0 | -4.091667 |             95.170833 |                   0.1 |                       7.41713 |            17.026026 |   0.0 |  1.1 | -4.091667 |    2.805569 |         1.866772 |           2.805569 |          3.5 |                 6.615571 |                      4.507808 |                         0.0 |                     5.064957 |                      0.478041 |                             0.3406 |             6.245099 |                  4.290119 |    1011.754901 |     1.2901 | 8.583333 | -6.249119 | -22.261106 |                   0.0 |                      0.0 |                       0.0 |                   4.244127 |                 15.115697 |                     0.0 |                13.676209 |   -15.115697 |             0.0 |       -13.676209 |         -10.87157 |             60.858762 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.007388 |      0.0 |      0.0 |  -0.01819 |      0.0 |    -2.538935 |     0.750727 |      0.0 |      0.0 | 0.327028 |      0.0 |     0.0 | -13.477788 |      0.0 | -23.612723 |      0.0 |  -23.414302 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.602359 | 20.602359 |  -30.340914 | 0.900715 | -489.915484 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 18:00:00 |  0.3 |  1.0 |              0.0 |                      0.0 |             0.0 |             96.9 |       4.4 |              1016.0 | 0.0 | 0.0 |     -3.95 |             95.070833 |                   0.1 |                       7.41713 |            17.026026 | 0.282 |  1.0 |     -3.95 |    3.527001 |         1.942897 |           3.527001 |          4.4 |                 6.567919 |                      4.556289 |                         0.0 |                     5.196402 |                      0.475001 |                           0.343839 |             6.364313 |                  4.331702 |    1009.635687 |   1.287972 | 8.583333 | -5.878986 | -22.105464 |                   0.0 |                      0.0 |                       0.0 |                     4.2404 |                  15.08957 |                     0.0 |                14.849938 |    -15.08957 |             0.0 |       -14.849938 |        -10.849171 |             48.410379 |                 100.0 |               112.102145 |               98.814229 |    0.282 |  0.0705 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.003845 |      0.0 |      0.0 | -0.021203 | 0.286904 |    -2.194742 |     0.750946 |      0.0 |      0.0 | 0.289731 |   0.2115 |     0.0 | -15.710776 |      0.0 | -24.418058 |      0.0 |  -25.278896 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.905562 | 20.905562 |   -7.258078 | 0.898208 | -459.226686 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 19:00:00 |  0.2 |  1.4 |              0.0 |                      0.0 |             0.0 |             98.3 |       3.8 |              1015.0 | 0.0 | 0.0 | -3.816667 |             95.029167 |                   0.1 |                       7.41713 |            17.026026 | 0.188 |  1.4 | -3.816667 |    3.046047 |         1.973983 |           3.046047 |          3.8 |                 6.760362 |                      4.602339 |                         0.0 |                     5.553833 |                       0.48726 |                           0.346913 |             6.645436 |                  4.373564 |    1008.354564 |   1.284691 | 8.583333 |  -5.51072 | -25.286578 |                   0.0 |                      0.0 |                       0.0 |                   4.245605 |                 15.061291 |                     0.0 |                15.472735 |   -15.061291 |             0.0 |       -15.472735 |        -10.815686 |             56.054123 |                 100.0 |               112.102145 |               98.814229 |    0.188 |  0.0282 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.000004 |      0.0 |      0.0 | -0.017823 | 0.283107 |    -1.296291 |      0.75064 |      0.0 |      0.0 | 0.172401 |   0.1598 |     0.0 | -13.205853 |      0.0 | -18.533424 |      0.0 |  -16.266543 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 21.111385 | 21.111385 |   -1.169117 | 0.901712 | -425.356775 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 20:00:00 |  0.4 |  1.6 |              0.0 |                      0.0 |             0.0 |             98.7 |       5.7 |              1014.0 | 0.0 | 0.0 |    -3.675 |             94.991667 |                   0.1 |                       7.41713 |            17.026026 | 0.376 |  1.6 |    -3.675 |     4.56907 |         2.056027 |            4.56907 |          5.7 |                 6.858436 |                      4.651718 |                         0.0 |                     5.830906 |                      0.493491 |                           0.350204 |             6.769276 |                  4.418744 |    1007.230724 |   1.282429 | 8.583333 | -5.104281 |  -26.81969 |                   0.0 |                      0.0 |                       0.0 |                   4.251249 |                 15.030582 |                     0.0 |                16.893295 |   -15.030582 |             0.0 |       -16.893295 |        -10.779333 |             37.369415 |                 100.0 |               112.102145 |               98.814229 |    0.376 |  0.0376 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.000346 |      0.0 |      0.0 | -0.020752 | 0.664621 |    -0.634333 |     0.750309 |      0.0 |      0.0 | 0.086638 | 0.086638 |     0.0 | -15.376607 |      0.0 | -20.802784 |      0.0 |  -19.286096 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 21.256132 | 21.508137 |         0.0 | 0.905525 | -389.953752 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 21:00:00 |  0.7 |  2.0 |              0.0 |                      0.0 |             0.0 |             99.6 |       5.6 |              1013.0 | 0.0 | 0.0 |      -3.5 |             94.991667 |                   0.1 |                       7.41713 |            17.026026 | 0.658 |  2.0 |      -3.5 |    4.488911 |         2.138899 |           4.488911 |          5.6 |                 7.058357 |                      4.713362 |                         0.0 |                     6.000247 |                      0.506157 |                           0.354306 |             7.030123 |                  4.477301 |    1005.969877 |   1.279174 | 8.583333 | -4.679445 | -26.616782 |                   0.0 |                      0.0 |                       0.0 |                   4.262042 |                  14.98921 |                     0.0 |                15.416111 |    -14.98921 |             0.0 |       -15.416111 |        -10.727167 |             38.036726 |                 100.0 |               112.102145 |               98.814229 |    0.658 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.002532 |      0.0 |      0.0 | -0.022462 | 1.530508 |    -0.243315 |     0.749675 |      0.0 |      0.0 | 0.035139 | 0.035139 |     0.0 |  -16.64359 |      0.0 | -20.598693 |      0.0 |  -21.826173 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 21.312847 | 22.188599 |         0.0 |  0.91286 | -354.753637 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 22:00:00 |  0.5 |  2.4 |              0.0 |                      0.0 |             0.0 |             99.4 |       5.8 |              1011.0 | 0.0 | 0.0 | -3.291667 |             95.008333 |                   0.1 |                       7.41713 |            17.026026 |  0.47 |  2.4 | -3.291667 |    4.649229 |         2.249283 |           4.649229 |          5.8 |                 7.263399 |                      4.787689 |                         0.0 |                       6.1078 |                      0.519101 |                           0.359244 |             7.219819 |                  4.548704 |    1003.780181 |   1.274698 | 8.583333 | -4.257044 | -25.542262 |                   0.0 |                      0.0 |                       0.0 |                   4.272746 |                 14.938353 |                     0.0 |                 13.71239 |   -14.938353 |             0.0 |        -13.71239 |        -10.665607 |             36.725115 |                 100.0 |               112.102145 |               98.814229 |     0.47 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.001838 |      0.0 |      0.0 | -0.024931 | 1.311864 |          0.0 |     0.749046 | 0.034312 | 0.034312 |      0.0 |      0.0 |     0.0 | -18.473108 |      0.0 |  -22.65304 |      0.0 |  -27.413757 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 21.301948 | 22.683531 |         0.0 | 0.920191 | -320.628041 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 23:00:00 |  1.0 |  2.5 |              0.0 |                      0.0 |             0.0 |             99.2 |       6.2 |              1010.0 | 0.0 | 0.0 | -3.070833 |                  95.0 |                   0.1 |                       7.41713 |            17.026026 |  0.94 |  2.5 | -3.070833 |    4.969866 |         2.364694 |           4.969866 |          6.2 |                 7.315473 |                      4.867606 |                         0.0 |                       6.1078 |                      0.522381 |                           0.364543 |             7.256949 |                  4.624226 |    1002.743051 |   1.272954 | 8.583333 | -3.847536 | -23.085219 |                   0.0 |                      0.0 |                       0.0 |                     4.2834 |                 14.885653 |                     0.0 |                13.067893 |   -14.885653 |             0.0 |       -13.067893 |        -10.602253 |             34.355753 |                 100.0 |               112.102145 |               98.814229 |     0.94 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.000987 |      0.0 |      0.0 | -0.027164 |  2.73305 |          0.0 |      0.74842 | 0.124889 | 0.124889 |      0.0 |      0.0 |     0.0 |  -20.12751 |      0.0 | -24.879463 |      0.0 |   -31.93908 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  21.20141 | 23.650695 |         0.0 | 0.927543 | -288.959489 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-11 00:00:00 |  2.0 |  2.7 |              0.0 |                      0.0 |             0.0 |             99.2 |       5.6 |              1008.0 | 0.0 | 0.0 | -2.804167 |               95.4875 |                   0.1 |                       7.41713 |            17.026026 |  1.88 |  2.7 | -2.804167 |    4.488911 |         2.438173 |           4.488911 |          5.6 |                 7.420609 |                      4.965681 |                         0.0 |                       6.1078 |                      0.528994 |                           0.371032 |             7.361244 |                  4.741604 |    1000.638756 |   1.269455 | 8.583333 | -3.467514 | -20.805083 |                   0.0 |                      0.0 |                       0.0 |                   4.294028 |                 14.787036 |                     0.0 |                11.601106 |   -14.787036 |             0.0 |       -11.601106 |        -10.493009 |             38.036726 |                 100.0 |               112.102145 |               98.814229 |     1.88 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.001174 |      0.0 |      0.0 | -0.027338 | 5.903388 |          0.0 |     0.747796 | 0.199896 | 0.199896 |      0.0 |      0.0 |     0.0 |  -20.25663 |      0.0 | -24.792095 |      0.0 |  -33.447619 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 21.024002 | 25.558033 |         0.0 | 0.934934 | -259.571073 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-11 01:00:00 |  2.1 |  3.2 |              0.0 |                      0.0 |             0.0 |             99.6 |       5.2 |              1007.0 | 0.0 | 0.0 | -2.470833 |             95.858333 |                   0.1 |                       7.41713 |            17.026026 | 1.974 |  3.2 | -2.470833 |    4.168274 |         2.514993 |           4.168274 |          5.2 |                 7.689299 |                      5.090729 |                         0.0 |                       6.1078 |                      0.545841 |                           0.379281 |             7.658542 |                  4.879888 |     999.341458 |   1.265756 | 8.583333 | -3.114853 | -18.689117 |                   0.0 |                      0.0 |                       0.0 |                   4.304586 |                 14.678585 |                     0.0 |                 7.733351 |   -14.678585 |             0.0 |        -7.733351 |        -10.373999 |             40.962628 |                 100.0 |               112.102145 |               98.814229 |    1.974 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.002516 |      0.0 |      0.0 | -0.031933 | 7.346438 |          0.0 |     0.747176 |  0.34843 |  0.34843 |      0.0 |      0.0 |     0.0 | -23.661007 |      0.0 | -27.741565 |      0.0 |  -43.669221 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.699553 | 27.563966 |         0.0 | 0.942333 | -232.298622 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-11 02:00:00 |  1.8 |  3.9 |              0.0 |                      0.0 |             0.0 |             99.8 |       5.7 |              1006.0 | 0.0 | 0.0 |    -2.075 |             96.279167 |                   0.1 |                       7.41713 |            17.026026 | 1.692 |  3.9 |    -2.075 |     4.56907 |         2.621871 |            4.56907 |          5.7 |                  8.07986 |                      5.242833 |                         0.0 |                       6.1078 |                      0.570198 |                           0.389283 |               8.0637 |                  5.047756 |       997.9363 |   1.261108 | 8.583333 | -2.787583 | -16.725501 |                   0.0 |                      0.0 |                       0.0 |                   4.315056 |                 14.548979 |                     0.0 |                 2.415944 |   -14.548979 |             0.0 |        -2.415944 |        -10.233923 |             37.369415 |                 100.0 |               112.102145 |               98.814229 |    1.692 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.003093 |      0.0 |      0.0 | -0.043255 | 7.674404 |          0.0 |     0.746561 | 0.613232 | 0.613232 |      0.0 |      0.0 |     0.0 | -32.050365 |      0.0 | -36.310997 |      0.0 |  -65.945417 |      1.0 |      0.3 | 0.533315 |  0.18206 |  0.0 |  0.0 | 0.0 |  0.0 |  0.18206 | 0.002362 | 0.000656 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.116019 | 28.765907 |         0.0 | 0.949726 | -206.989788 | 71.797515 |  0.18206 |      0.0 |   0.0 |   0.0 |  0.0 | 0.002362 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.000656 |
    | 2010-12-11 03:00:00 |  1.3 |  4.4 |              0.0 |                      0.0 |             0.0 |             99.8 |       6.6 |              1005.0 | 0.0 | 0.0 | -1.633333 |             96.633333 |                   0.1 |                       7.41713 |            17.026026 | 1.222 |  4.4 | -1.633333 |    5.290502 |          2.74211 |           5.290502 |          6.6 |                  8.36943 |                      5.417279 |                         0.0 |                       6.1078 |                       0.58816 |                           0.400709 |             8.352691 |                  5.234898 |     996.647309 |   1.257444 | 8.583333 | -2.483877 | -14.903265 |                   0.0 |                      0.0 |                       0.0 |                   4.325328 |                 14.409644 |                     0.0 |                -1.360594 |   -14.409644 |             0.0 |         1.360594 |        -10.084316 |             32.273586 |                 100.0 |               111.598996 |               98.814229 |    1.222 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.002984 |      0.0 |      0.0 | -0.055801 |      0.0 |          0.0 |     0.745958 | 0.795977 | 0.795977 |      0.0 |      0.0 |     0.0 | -41.346526 |      0.0 | -46.045135 |      0.0 |  -88.752255 |      1.0 |      0.3 | 2.362455 | 0.808317 |  0.0 |  0.0 | 0.0 |  0.0 | 0.808317 | 0.026716 | 0.007421 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 19.357519 | 27.681253 |         0.0 | 0.957034 |  -183.50319 | 73.351654 |      0.5 | 0.308317 |   0.0 |   0.0 |  0.0 | 0.011128 | 0.015589 |   0.0 |   0.0 |  0.0 |   0.0 | 0.007421 |
    | 2010-12-11 04:00:00 |  0.5 |  4.8 |              0.0 |                      0.0 |             0.0 |             99.0 |       7.0 |              1005.0 | 0.0 | 0.0 | -1.129167 |             96.833333 |                   0.1 |                       7.41713 |            17.026026 |  0.47 |  4.8 | -1.129167 |    5.611139 |         2.892408 |           5.611139 |          7.0 |                 8.607627 |                      5.622664 |                         0.0 |                       6.1078 |                      0.602876 |                           0.414103 |             8.521551 |                  5.444613 |     996.478449 |   1.255555 | 8.583333 | -2.202038 |  -13.21223 |                   0.0 |                      0.0 |                       0.0 |                   4.335523 |                 14.261536 |                     0.0 |                -4.039084 |   -14.261536 |             0.0 |         4.039084 |         -9.926013 |             30.429381 |                 100.0 |               109.384997 |               98.814229 |     0.47 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.000598 |      0.0 |      0.0 |  -0.06294 |      0.0 |          0.0 |     0.745359 | 0.971746 | 0.971746 |      0.0 |      0.0 |     0.0 | -46.635995 |      0.0 | -52.693545 |      0.0 | -103.368624 |      1.0 |      0.3 | 1.862756 | 0.638926 |  0.0 |  0.0 | 0.0 |  0.0 | 0.638926 | 0.059874 | 0.016632 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 18.427578 | 26.351437 |         0.0 | 0.964342 | -161.707627 | 74.575484 |      0.5 | 0.138926 |   0.0 |   0.0 |  0.0 | 0.023756 | 0.036118 |   0.0 |   0.0 |  0.0 |   0.0 | 0.016632 |
    | 2010-12-11 05:00:00 |  0.2 |  4.8 |              0.0 |                      0.0 |             0.0 |             97.6 |       7.3 |              1004.0 | 0.0 | 0.0 | -0.583333 |             97.016667 |                   0.1 |                       7.41713 |            17.026026 | 0.188 |  4.8 | -0.583333 |    5.851616 |         3.059407 |           5.851616 |          7.3 |                 8.607627 |                      5.852751 |                         0.0 |                       6.1078 |                      0.602876 |                           0.429037 |             8.401044 |                  5.678144 |     995.598956 |   1.254358 | 8.583333 | -1.940492 | -11.642949 |                   0.0 |                      0.0 |                       0.0 |                   4.345643 |                 14.099661 |                     0.0 |                -3.389325 |   -14.099661 |             0.0 |         3.389325 |         -9.754018 |             29.178858 |                 100.0 |               107.654875 |               98.814229 |    0.188 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.007217 |      0.0 |      0.0 | -0.061894 |      0.0 |          0.0 |     0.744765 | 0.993206 | 0.993206 |      0.0 |      0.0 |     0.0 | -45.860636 |      0.0 | -54.540411 |      0.0 | -103.790373 |      1.0 |      0.3 |  1.61162 | 0.553938 |  0.0 |  0.0 | 0.0 |  0.0 | 0.553938 | 0.078096 | 0.021693 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 17.475322 | 24.989711 |         0.0 |  0.97165 | -141.481344 | 75.633165 |      0.5 | 0.053938 |   0.0 |   0.0 |  0.0 | 0.036057 | 0.042039 |   0.0 |   0.0 |  0.0 |   0.0 | 0.021693 |
    | 2010-12-11 06:00:00 |  0.5 |  5.1 |              0.0 |                      0.0 |             0.0 |             97.4 |       7.2 |              1004.0 | 0.0 | 0.0 |   -0.0125 |             97.054167 |                   0.1 |                       7.41713 |            17.026026 |  0.47 |  5.1 |   -0.0125 |    5.771457 |         3.229745 |           5.771457 |          7.2 |                 8.790172 |                      6.102233 |                         0.0 |                       6.1078 |                      0.614118 |                           0.445147 |             8.561627 |                  5.922472 |     995.438373 |    1.25293 | 8.583333 | -1.697776 | -10.186657 |                   0.0 |                      0.0 |                       0.0 |                   4.355688 |                 13.937223 |                     0.0 |                -5.587229 |   -13.937223 |             0.0 |         5.587229 |         -9.581535 |              29.58412 |                 100.0 |               106.168542 |               98.814229 |     0.47 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.00818 |      0.0 |      0.0 |  -0.06548 |      0.0 |          0.0 |     0.744175 | 1.090927 | 1.090927 |      0.0 |      0.0 |     0.0 |  -48.51811 |      0.0 | -57.295089 |      0.0 | -111.400428 |      1.0 |      0.3 | 2.034113 | 0.700814 |  0.0 |  0.0 | 0.0 |  0.0 | 0.700814 | 0.098689 | 0.027414 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 16.427327 | 23.491078 |         0.0 | 0.978958 | -122.711354 | 76.966465 |      0.5 | 0.200814 |   0.0 |   0.0 |  0.0 | 0.048041 | 0.050648 |   0.0 |   0.0 |  0.0 |   0.0 | 0.027414 |
    | 2010-12-11 07:00:00 |  0.2 |  5.1 |              0.0 |                      0.0 |             0.0 |             96.3 |       7.6 |              1004.0 | 0.0 | 0.0 |  0.554167 |             97.070833 |                   0.1 |                       7.41713 |            17.026026 | 0.188 |  5.1 |  0.554167 |    6.092093 |         3.436822 |           6.092093 |          7.6 |                 8.790172 |                      6.359135 |                         0.0 |                       6.1078 |                      0.614118 |                            0.46165 |             8.464935 |                  6.172866 |     995.535065 |   1.252976 | 8.583333 | -1.472536 |  -8.835218 |                   0.0 |                      0.0 |                       0.0 |                    4.36566 |                 13.773355 |                     0.0 |                -5.066448 |   -13.773355 |             0.0 |         5.066448 |         -9.407695 |             28.027061 |                 100.0 |               104.306022 |               98.814229 |    0.188 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.014147 |      0.0 |      0.0 | -0.065772 |      0.0 |          0.0 |     0.743589 | 1.130414 | 1.130414 |      0.0 |      0.0 |     0.0 | -48.734564 |      0.0 | -59.911482 |      0.0 | -113.712494 |      1.0 |      0.3 | 1.809018 | 0.624709 |  0.0 |  0.0 | 0.0 |  0.0 | 0.624709 | 0.121441 | 0.033733 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 15.339743 | 21.935832 |         0.0 | 0.986266 | -105.292803 | 78.150774 |      0.5 | 0.124709 |   0.0 |   0.0 |  0.0 | 0.059716 | 0.061725 |   0.0 |   0.0 |  0.0 |   0.0 | 0.033733 |
    | 2010-12-11 08:00:00 |  0.1 |  5.4 |              0.0 |                 0.347041 |        0.900733 |             95.2 |       7.3 |              1003.0 | 0.0 | 0.0 |     1.175 |             97.004167 |                   0.1 |                       7.40894 |            17.017406 | 0.094 |  5.4 |     1.175 |    5.851616 |           3.6439 |           5.851616 |          7.3 |                 8.976116 |                       6.65151 |                         0.0 |                       6.1078 |                       0.62554 |                           0.480332 |             8.545263 |                  6.452241 |     994.454737 |   1.250338 | 8.583333 | -1.263514 |  -7.581082 |              0.231482 |                      0.0 |                  0.231482 |                   4.373343 |                  13.59621 |                     0.0 |                -6.835277 |   -13.364729 |             0.0 |         7.066758 |         -9.222868 |             29.178858 |                 100.0 |               102.661497 |               99.312507 |    0.094 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.019008 |      0.0 |      0.0 | -0.065786 |      0.0 |          0.0 |     0.743008 | 1.181192 | 1.181192 |      0.0 |      0.0 |     0.0 | -48.744743 |      0.0 | -61.357963 |      0.0 | -117.169464 |      1.0 |      0.3 |  1.78817 | 0.618883 |  0.0 |  0.0 | 0.0 |  0.0 | 0.618883 | 0.138779 |  0.03855 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 14.201013 | 20.307448 |         0.0 | 0.993574 |  -89.128388 | 79.320061 |      0.5 | 0.118883 |   0.0 |   0.0 |  0.0 | 0.071089 |  0.06769 |   0.0 |   0.0 |  0.0 |   0.0 |  0.03855 |
    | 2010-12-11 09:00:00 |  0.3 |  5.3 |              0.0 |                      1.0 |       17.267511 |             96.3 |       7.1 |              1003.0 | 0.0 | 0.0 |  1.779167 |               97.0875 |                   0.1 |                       7.40894 |            16.959252 | 0.282 |  5.3 |  1.779167 |    5.691298 |         3.847638 |           5.691298 |          7.1 |                 8.913754 |                      6.947358 |                         0.0 |                       6.1078 |                      0.621713 |                            0.49913 |             8.583945 |                  6.745016 |     994.416055 |   1.250768 | 8.583333 | -1.069541 |  -6.417244 |              4.447583 |                      0.0 |                  4.447583 |                   4.368185 |                 13.407231 |                     0.0 |                -6.596815 |    -8.959648 |             0.0 |        11.044398 |         -9.039046 |             30.000798 |                 100.0 |               101.046454 |              100.105201 |    0.282 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.01638 |      0.0 |      0.0 | -0.065321 |      0.0 |          0.0 |     0.742431 | 1.205998 | 1.205998 |      0.0 |      0.0 |     0.0 | -48.400444 |      0.0 | -58.862205 |      0.0 | -118.307048 |      1.0 |      0.3 | 2.012124 | 0.698077 |  0.0 |  0.0 | 0.0 |  0.0 | 0.698077 |  0.15895 | 0.044153 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 13.036815 | 18.642645 |         0.0 | 1.000882 |  -74.127811 | 80.634108 |      0.5 | 0.198077 |   0.0 |   0.0 |  0.0 | 0.082168 | 0.076782 |   0.0 |   0.0 |  0.0 |   0.0 | 0.044153 |
    | 2010-12-11 10:00:00 |  0.1 |  5.5 |              0.0 |                      1.0 |       52.617611 |             95.5 |       7.1 |              1003.0 | 0.0 | 0.0 |  2.295833 |               97.0375 |                   0.1 |                       7.40894 |            16.807156 | 0.094 |  5.5 |  2.295833 |    5.691298 |         4.038016 |           5.691298 |          7.1 |                 9.038862 |                      7.209503 |                         0.0 |                       6.1078 |                      0.629387 |                           0.515703 |             8.632113 |                  6.995922 |     994.367887 |   1.249848 | 8.583333 | -0.889534 |  -5.337202 |             13.582833 |                      0.0 |                 13.582833 |                   4.338638 |                 13.251022 |                     0.0 |                -7.748872 |      0.33181 |             0.0 |        21.331705 |         -8.912384 |             30.000798 |                 100.0 |                99.241158 |               99.922467 |    0.094 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.02643 |      0.0 |      0.0 | -0.066592 |      0.0 |          0.0 |     0.741858 |  1.36261 |  1.36261 |      0.0 |      0.0 |     0.0 | -49.341985 |      0.0 | -61.083421 |      0.0 |  -131.75711 |      1.0 |      0.3 | 2.049492 | 0.712842 |  0.0 |  0.0 | 0.0 |  0.0 | 0.712842 | 0.182545 | 0.050707 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 11.715906 | 16.753745 |         0.0 |  1.00819 |  -60.207275 | 81.970758 |      0.5 | 0.212842 |   0.0 |   0.0 |  0.0 | 0.092961 | 0.089584 |   0.0 |   0.0 |  0.0 |   0.0 | 0.050707 |
    | 2010-12-11 11:00:00 |  0.1 |  6.1 |              0.0 |                      1.0 |       80.135633 |             91.6 |       8.9 |              1002.0 | 0.0 | 0.0 |  2.716667 |             96.795833 |                   0.0 |                       7.40894 |            15.680798 | 0.094 |  6.1 |  2.716667 |    7.134162 |         4.291853 |           7.134162 |          8.9 |                 9.423511 |                       7.42943 |                         0.0 |                       6.1078 |                      0.652898 |                           0.529548 |             8.631936 |                  7.191379 |     993.368064 |   1.245915 | 8.583333 | -0.722487 |  -4.334924 |             20.731976 |                      0.0 |                 20.731976 |                   4.056796 |                 12.463959 |                     0.0 |               -12.026271 |     8.268018 |             0.0 |        32.758247 |         -8.407162 |             23.933221 |                 100.0 |                97.414775 |               99.731421 |    0.094 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.060046 |      0.0 |      0.0 | -0.080429 |      0.0 |          0.0 |     0.741289 | 1.830687 | 1.830687 |      0.0 |      0.0 |     0.0 |  -59.59435 |      0.0 | -81.829421 |      0.0 | -174.182018 |      1.0 |      0.3 |  2.72445 | 0.950584 |  0.0 |  0.0 | 0.0 |  0.0 | 0.950584 |  0.21733 | 0.060369 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  9.932674 | 14.203724 |         0.0 | 1.015498 |  -47.289018 | 83.744624 |      0.5 | 0.450584 |   0.0 |   0.0 |  0.0 | 0.103475 | 0.113855 |   0.0 |   0.0 |  0.0 |   0.0 | 0.060369 |
    | 2010-12-11 12:00:00 |  0.0 |  6.4 |              0.0 |                      1.0 |       80.276117 |             90.0 |       9.5 |              1002.0 | 0.0 | 0.0 |       3.1 |             96.633333 |                   0.0 |                       7.40894 |            15.214247 |   0.0 |  6.4 |       3.1 |    7.615117 |          4.56573 |           7.615117 |          9.5 |                 9.621181 |                      7.634886 |                         0.0 |                       6.1078 |                      0.664931 |                           0.542435 |             8.659063 |                  7.377844 |     993.340937 |   1.244565 | 8.583333 | -0.567468 |  -3.404809 |             20.813648 |                      0.0 |                 20.813648 |                   3.944685 |                 12.359437 |                     0.0 |               -13.533625 |     8.454211 |             0.0 |        34.347272 |         -8.414752 |             22.421649 |                 100.0 |                95.005574 |                99.46937 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.073917 |      0.0 |      0.0 | -0.085956 |      0.0 |          0.0 |     0.740724 | 1.998446 | 1.998446 |      0.0 |      0.0 |     0.0 | -63.690154 |      0.0 | -90.778796 |      0.0 | -188.816222 |      1.0 |      0.3 | 2.875073 | 1.006673 |  0.0 |  0.0 | 0.0 |  0.0 | 1.006673 | 0.263876 | 0.073299 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  7.982243 | 11.414607 |         0.0 | 1.022806 |  -35.300875 | 85.613024 |      0.5 | 0.506673 |   0.0 |   0.0 |  0.0 | 0.113718 | 0.150158 |   0.0 |   0.0 |  0.0 |   0.0 | 0.073299 |
    | 2010-12-11 13:00:00 |  0.1 |  6.4 |              0.0 |                      1.0 |       75.835403 |             85.7 |       7.9 |              1002.0 | 0.0 | 0.0 |     3.425 |               96.3125 |                   0.0 |                       7.40894 |            15.020908 | 0.094 |  6.4 |     3.425 |    6.332571 |         4.782828 |           6.332571 |          7.9 |                 9.621181 |                      7.812981 |                         0.0 |                       6.1078 |                      0.664931 |                           0.553571 |             8.245352 |                  7.524878 |     993.754648 |    1.24476 | 8.583333 | -0.423611 |  -2.541663 |              19.70479 |                      0.0 |                  19.70479 |                   3.902977 |                 12.283332 |                     0.0 |               -11.239491 |     7.421458 |             0.0 |        30.944282 |         -8.380355 |             26.962743 |                 100.0 |                92.484788 |               99.182157 |    0.094 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.085367 |      0.0 |      0.0 | -0.061599 |      0.0 |          0.0 |     0.740164 | 1.634985 | 1.634985 |      0.0 |      0.0 |     0.0 | -45.642119 |      0.0 | -77.645525 |      0.0 | -154.231926 |      1.0 |      0.3 | 2.444646 |   0.8587 |  0.0 |  0.0 | 0.0 |  0.0 |   0.8587 | 0.301805 | 0.083835 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  6.381511 |  9.125561 |         0.0 | 1.030114 |  -24.175879 | 87.198969 |      0.5 |   0.3587 |   0.0 |   0.0 |  0.0 | 0.123696 |  0.17811 |   0.0 |   0.0 |  0.0 |   0.0 | 0.083835 |
    | 2010-12-11 14:00:00 |  0.0 |  6.5 |              0.0 |                      1.0 |       42.227911 |             82.3 |       9.6 |              1002.0 | 0.0 | 0.0 |    3.7125 |             95.854167 |                   0.0 |                       7.40894 |            14.983713 |   0.0 |  6.5 |    3.7125 |    7.695276 |         5.026645 |           7.695276 |          9.6 |                 9.687876 |                      7.973569 |                         0.0 |                       6.1078 |                      0.668985 |                           0.563584 |             7.973122 |                  7.642998 |     994.026878 |   1.244443 | 8.583333 | -0.290111 |  -1.740663 |             10.995846 |                      0.0 |                 10.995846 |                   3.901652 |                 12.228543 |                     0.0 |               -10.124562 |    -1.232697 |             0.0 |        21.120409 |         -8.326891 |              22.18809 |                 100.0 |                90.357693 |               98.928718 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.116273 |      0.0 |      0.0 | -0.063414 |      0.0 |          0.0 |     0.739607 | 1.718061 | 1.718061 |      0.0 |      0.0 |     0.0 | -46.987277 |      0.0 | -93.030869 |      0.0 | -161.138555 |      1.0 |      0.3 |   2.4739 | 0.871645 |  0.0 |  0.0 | 0.0 |  0.0 | 0.871645 | 0.330129 | 0.091702 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  4.695856 |  6.715075 |         0.0 | 1.037422 |  -13.851882 | 88.801224 |      0.5 | 0.371645 |   0.0 |   0.0 |  0.0 | 0.133416 | 0.196713 |   0.0 |   0.0 |  0.0 |   0.0 | 0.091702 |
    | 2010-12-11 15:00:00 |  0.1 |  6.3 |              0.1 |                      1.0 |       12.613919 |             82.5 |       9.1 |              1003.0 | 0.0 | 0.0 |  3.954167 |             95.370833 |                   0.1 |                       7.40894 |            15.079289 | 0.094 |  6.3 |  3.954167 |     7.29448 |         5.237063 |            7.29448 |          9.1 |                 9.554889 |                      8.110798 |                         0.0 |                       6.1078 |                      0.660899 |                           0.572121 |             7.882784 |                  7.735336 |     995.117216 |   1.246623 | 8.583333 | -0.166223 |  -0.997336 |              3.291545 |                      0.0 |                  3.291545 |                   3.934872 |                 12.847805 |                     0.0 |                -7.130286 |    -9.556261 |             0.0 |         10.42183 |         -8.912934 |             23.407216 |                 100.0 |                88.219729 |               98.663066 |    0.094 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.102953 |      0.0 |      0.0 | -0.057639 |      0.0 |          0.0 |     0.739055 |  1.49024 |  1.49024 |      0.0 |      0.0 |     0.0 |  -42.70834 |      0.0 | -86.128361 |      0.0 | -139.258531 |      1.0 |      0.3 | 2.243336 | 0.792685 |  0.0 |  0.0 | 0.0 |  0.0 | 0.792685 | 0.352992 | 0.098053 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  3.233132 |  4.623378 |         0.0 |  1.04473 |   -4.271214 | 90.251875 |      0.5 | 0.292685 |   0.0 |   0.0 |  0.0 | 0.142885 | 0.210107 |   0.0 |   0.0 |  0.0 |   0.0 | 0.098053 |
    | 2010-12-11 16:00:00 |  0.0 |  5.9 |              0.0 |                 0.053708 |        0.021911 |             82.2 |       8.7 |              1003.0 | 0.0 | 0.0 |  4.158333 |             94.854167 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  5.9 |  4.158333 |    6.973844 |         5.424101 |           6.973844 |          8.7 |                 9.293725 |                       8.22835 |                         0.0 |                       6.1078 |                      0.644979 |                           0.579419 |             7.639442 |                  7.804933 |     995.360558 |   1.248525 | 8.583333 | -0.051255 |  -0.307527 |               0.00573 |                      0.0 |                   0.00573 |                   3.943077 |                 12.822679 |                     0.0 |                -3.915548 |   -12.816949 |             0.0 |         3.921278 |         -8.879602 |              24.48341 |                 100.0 |                86.293046 |               98.792635 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.096543 |      0.0 |      0.0 | -0.047871 |      0.0 |          0.0 |     0.738506 | 1.258029 | 1.258029 |      0.0 |      0.0 |     0.0 | -35.470286 |      0.0 | -77.633085 |      0.0 | -117.024649 | 0.999378 | 0.299627 | 1.817608 | 0.643181 |  0.0 |  0.0 | 0.0 |  0.0 | 0.643181 | 0.362865 | 0.100796 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  1.995553 |  2.853641 |         0.0 | 1.052038 |    4.619647 | 91.426302 |      0.5 | 0.143181 |   0.0 |   0.0 |  0.0 |  0.15211 | 0.210755 |   0.0 |   0.0 |  0.0 |   0.0 | 0.100796 |
    | 2010-12-11 17:00:00 |  0.0 |  5.5 |              0.0 |                      0.0 |             0.0 |             76.5 |       8.9 |              1003.0 | 0.0 | 0.0 |  4.341667 |             94.108333 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  5.5 |  4.341667 |    7.134162 |         5.604459 |           7.134162 |          8.9 |                 9.038862 |                      8.335182 |                         0.0 |                       6.1078 |                      0.629387 |                            0.58604 |              6.91473 |                  7.844101 |      996.08527 |   1.250659 | 8.583333 |       0.0 |        0.0 |                   0.0 |                      0.0 |                       0.0 |                   3.951291 |                   12.8228 |                     0.0 |                 2.340465 |     -12.8228 |             0.0 |        -2.340465 |         -8.871509 |             23.933221 |                 100.0 |                84.739162 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.129358 |      0.0 |      0.0 | -0.025712 |      0.0 |          0.0 |     0.737961 | 0.975359 | 0.975359 |      0.0 |      0.0 |     0.0 | -19.051448 |      0.0 | -73.780626 |      0.0 |  -90.491608 | 0.998221 | 0.298934 |  1.40733 |  0.49805 |  0.0 |  0.0 | 0.0 |  0.0 |  0.49805 |  0.35787 | 0.099408 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  1.029387 |  1.472023 |         0.0 | 1.059346 |    13.20298 | 92.335582 |  0.49805 |      0.0 |   0.0 |   0.0 |  0.0 | 0.161071 |   0.1968 |   0.0 |   0.0 |  0.0 |   0.0 | 0.099408 |
    | 2010-12-11 18:00:00 |  0.0 |  5.1 |              0.0 |                      0.0 |             0.0 |             81.7 |       7.9 |              1003.0 | 0.0 | 0.0 |    4.5125 |                93.475 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  5.1 |    4.5125 |    6.332571 |         5.721357 |           6.332571 |          7.9 |                 8.790172 |                      8.435829 |                         0.0 |                       6.1078 |                      0.614118 |                           0.592267 |              7.18157 |                  7.885391 |      995.81843 |   1.252331 | 8.583333 |       0.0 |        0.0 |                   0.0 |                      0.0 |                       0.0 |                   3.959448 |                 12.818218 |                     0.0 |                 2.381726 |   -12.818218 |             0.0 |        -2.381726 |         -8.858771 |             26.962743 |                 100.0 |                83.539582 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.087608 |      0.0 |      0.0 | -0.030943 |      0.0 |          0.0 |      0.73742 | 0.888357 | 0.888357 |      0.0 |      0.0 |     0.0 | -22.927701 |      0.0 | -61.873778 |      0.0 |  -82.419753 | 0.997065 | 0.298241 | 1.297054 | 0.458963 |  0.0 |  0.0 | 0.0 |  0.0 | 0.458963 | 0.346508 | 0.096252 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  0.143995 |  0.205912 |         0.0 | 1.066654 |   21.786314 | 93.173673 | 0.458963 |      0.0 |   0.0 |   0.0 |  0.0 | 0.169268 |  0.17724 |   0.0 |   0.0 |  0.0 |   0.0 | 0.096252 |
    | 2010-12-11 19:00:00 |  0.0 |  4.8 |              0.0 |                      0.0 |             0.0 |             83.4 |       7.1 |              1003.0 | 0.0 | 0.0 |  4.654167 |             92.854167 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  4.8 |  4.654167 |    5.691298 |         5.831576 |           5.691298 |          7.1 |                 8.607627 |                      8.520102 |                         0.0 |                       6.1078 |                      0.602876 |                           0.597475 |             7.178761 |                   7.91127 |     995.821239 |   1.253684 | 8.583333 |       0.0 |        0.0 |                   0.0 |                      0.0 |                       0.0 |                   3.967546 |                 12.822742 |                     0.0 |                 3.699875 |   -12.822742 |             0.0 |        -3.699875 |         -8.855196 |             30.000798 |                 100.0 |                82.436521 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.070419 |     0.06 |  0.00372 | -0.028252 |      0.0 |          0.0 |     0.736883 | 0.760344 | 0.143995 |      0.0 |      0.0 |     0.0 | -20.933744 |      0.0 | -53.309167 |      0.0 |  -70.543036 | 0.995908 |  0.29755 | 0.234165 | 0.082767 |  0.0 |  0.0 | 0.0 |  0.0 | 0.082767 | 0.331496 | 0.092082 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |         0.0 | 1.073962 |   30.369647 | 93.321351 | 0.082767 |      0.0 |   0.0 |   0.0 |  0.0 | 0.171871 | 0.159624 |   0.0 |   0.0 |  0.0 |   0.0 | 0.092082 |
    | 2010-12-11 20:00:00 |  0.0 |  4.9 |              0.0 |                      0.0 |             0.0 |             82.8 |       6.4 |              1003.0 | 0.0 | 0.0 |  4.791667 |             92.191667 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  4.9 |  4.791667 |    5.130184 |         5.854956 |           5.130184 |          6.4 |                 8.668101 |                      8.602604 |                         0.0 |                          0.0 |                      0.606604 |                           0.602566 |             7.177187 |                  7.930884 |     995.822813 |   1.253234 | 8.583333 |       0.0 |      -29.4 |                   0.0 |                      0.0 |                       0.0 |                  13.571128 |                   12.8326 |                     0.0 |                      0.0 |     -12.8326 |             0.0 |              0.0 |          0.738528 |             33.282135 |                 100.0 |                82.242406 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.066191 |      0.0 | 0.025199 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.990791 |   0.2945 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.312256 | 0.086738 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |         0.0 |      nan |    68.35298 | 93.296151 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.168496 | 0.143759 |   0.0 |   0.0 |  0.0 |   0.0 | 0.086738 |
    | 2010-12-11 21:00:00 |  0.0 |  4.9 |              0.0 |                      0.0 |             0.0 |             85.8 |       7.2 |              1003.0 | 0.0 | 0.0 |    4.9125 |             91.616667 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  4.9 |    4.9125 |    5.771457 |         5.908395 |           5.771457 |          7.2 |                 8.668101 |                      8.675686 |                         0.0 |                          0.0 |                      0.606604 |                           0.607071 |              7.43723 |                  7.948375 |      995.56277 |   1.253111 | 8.583333 |       0.0 |      -29.4 |                   0.0 |                      0.0 |                       0.0 |                  13.571128 |                 12.841028 |                     0.0 |                      0.0 |   -12.841028 |             0.0 |              0.0 |          0.730101 |              29.58412 |                 100.0 |                82.275524 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.060894 |      0.0 | 0.021594 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.985673 | 0.291466 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.293615 |  0.08156 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |         0.0 |      nan |  106.336314 | 93.274558 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.164144 | 0.129471 |   0.0 |   0.0 |  0.0 |   0.0 |  0.08156 |
    | 2010-12-11 22:00:00 |  0.1 |  4.7 |              0.0 |                      0.0 |             0.0 |             87.6 |       7.0 |              1003.0 | 0.0 | 0.0 |  5.008333 |                91.125 |                   0.1 |                      7.400749 |            15.079031 | 0.094 |  4.7 |  5.008333 |    5.611139 |         5.948475 |           5.611139 |          7.0 |                 8.547525 |                      8.734036 |                         0.0 |                          0.0 |                      0.599168 |                           0.610664 |             7.487632 |                   7.95889 |     995.512368 |   1.253989 | 8.583333 |       0.0 |      -28.2 |                   0.0 |                      0.0 |                       0.0 |                  13.571128 |                 12.851041 |                     0.0 |                      0.0 |   -12.851041 |             0.0 |              0.0 |          0.720087 |             30.429381 |                 100.0 |                82.303905 |               98.814229 |    0.034 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.050931 | 0.050931 |      0.0 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.980717 | 0.288542 |    0.034 | 0.011736 |  0.0 |  0.0 | 0.0 |  0.0 | 0.011736 | 0.276659 |  0.07685 | 0.009069 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |         0.0 |      nan |  143.119647 | 93.296822 | 0.011736 |      0.0 |   0.0 |   0.0 |  0.0 | 0.160056 | 0.116603 |   0.0 |   0.0 |  0.0 |   0.0 |  0.07685 |
    | 2010-12-11 23:00:00 |  0.1 |  3.8 |              0.0 |                      0.0 |             0.0 |             85.6 |       6.8 |              1002.0 | 0.0 | 0.0 |    5.0625 |             90.558333 |                   0.1 |                      7.400749 |            15.079031 | 0.094 |  3.8 |    5.0625 |     5.45082 |         5.968515 |            5.45082 |          6.8 |                 8.023017 |                      8.767169 |                         0.0 |                          0.0 |                      0.566662 |                           0.612703 |             6.867703 |                  7.939402 |     995.132297 |   1.257101 | 8.583333 |       0.0 |      -22.8 |                   0.0 |                      0.0 |                       0.0 |                  13.571128 |                 12.881907 |                     0.0 |                      0.0 |   -12.881907 |             0.0 |              0.0 |          0.689221 |             31.324363 |                 100.0 |                82.274643 |               98.814229 | 0.043069 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.055782 | 0.055782 |      0.0 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.976489 | 0.286059 | 0.043069 | 0.014768 |  0.0 |  0.0 | 0.0 |  0.0 | 0.014768 | 0.261278 | 0.072577 | 0.004218 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |         0.0 |      nan |   174.50298 | 93.325122 | 0.014768 |      0.0 |   0.0 |   0.0 |  0.0 | 0.156264 | 0.105014 |   0.0 |   0.0 |  0.0 |   0.0 | 0.072577 |
    | 2010-12-12 00:00:00 |  0.0 |  3.9 |              0.0 |                      0.0 |             0.0 |             85.3 |       6.4 |              1002.0 | 0.0 | 0.0 |    5.1125 |             89.979167 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  3.9 |    5.1125 |    5.130184 |         5.995234 |           5.130184 |          6.4 |                  8.07986 |                      8.797851 |                         0.0 |                          0.0 |                      0.570198 |                           0.614591 |             6.892121 |                  7.916233 |     995.107879 |   1.256635 | 8.583333 |       0.0 |      -23.4 |                   0.0 |                      0.0 |                       0.0 |                  13.571128 |                 12.915622 |                     0.0 |                      0.0 |   -12.915622 |             0.0 |              0.0 |          0.655506 |             33.282135 |                 100.0 |                 82.23745 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.05385 | 0.004218 | 0.018582 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 |  0.97218 |  0.28354 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.246995 |  0.06861 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |         0.0 |      nan |  206.486314 | 93.306541 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.152417 | 0.094577 |   0.0 |   0.0 |  0.0 |   0.0 |  0.06861 |
    | 2010-12-12 01:00:00 |  0.0 |  3.7 |              0.0 |                      0.0 |             0.0 |             86.2 |       6.3 |              1002.0 | 0.0 | 0.0 |  5.133333 |             89.420833 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  3.7 |  5.133333 |    5.050025 |         6.031974 |           5.050025 |          6.3 |                 7.966527 |                      8.810663 |                         0.0 |                          0.0 |                      0.563146 |                           0.615379 |             6.867146 |                  7.878569 |     995.132854 |   1.257555 | 8.583333 |       0.0 |      -22.2 |                   0.0 |                      0.0 |                       0.0 |                  13.571128 |                 12.957663 |                     0.0 |                      0.0 |   -12.957663 |             0.0 |              0.0 |          0.613465 |             33.810423 |                 100.0 |                 82.26187 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.049164 |      0.0 | 0.018518 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.968033 | 0.281126 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.233658 | 0.064905 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |         0.0 |      nan |  237.269647 | 93.288022 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |  0.14848 | 0.085177 |   0.0 |   0.0 |  0.0 |   0.0 | 0.064905 |
    | 2010-12-12 02:00:00 |  0.1 |  3.5 |              0.0 |                      0.0 |             0.0 |             89.3 |       5.7 |              1002.0 | 0.0 | 0.0 |  5.116667 |             88.983333 |                   0.1 |                      7.400749 |            15.079031 | 0.094 |  3.5 |  5.116667 |     4.56907 |         6.031974 |            4.56907 |          5.7 |                 7.854597 |                      8.800412 |                         0.0 |                          0.0 |                      0.556168 |                           0.614748 |             7.014155 |                    7.8309 |     994.985845 |   1.258394 | 8.583333 |       0.0 |      -21.0 |                   0.0 |                      0.0 |                       0.0 |                  13.571128 |                 13.001862 |                     0.0 |                      0.0 |   -13.001862 |             0.0 |              0.0 |          0.569266 |             37.369415 |                 100.0 |                82.286208 |               98.814229 |    0.034 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.033549 | 0.033549 |      0.0 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.964047 | 0.278816 |    0.034 | 0.011432 |  0.0 |  0.0 | 0.0 |  0.0 | 0.011432 | 0.221505 | 0.061529 | 0.026451 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |         0.0 |      nan |   266.85298 |  93.31059 | 0.011432 |      0.0 |   0.0 |   0.0 |  0.0 | 0.144793 | 0.076712 |   0.0 |   0.0 |  0.0 |   0.0 | 0.061529 |
    | 2010-12-12 03:00:00 |  0.0 |  3.3 |              0.0 |                      0.0 |             0.0 |             91.4 |       5.2 |              1003.0 | 0.0 | 0.0 |  5.070833 |             88.633333 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  3.3 |  5.070833 |    4.168274 |         5.985215 |           4.168274 |          5.2 |                 7.744054 |                      8.772276 |                         0.0 |                          0.0 |                      0.549265 |                           0.613018 |             7.078066 |                  7.775161 |     995.921934 |   1.260534 | 8.583333 |       0.0 |      -19.8 |                   0.0 |                      0.0 |                       0.0 |                  13.571128 |                 13.048148 |                     0.0 |                      0.0 |   -13.048148 |             0.0 |              0.0 |           0.52298 |             40.962628 |                 100.0 |                82.256548 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.023802 | 0.023802 |      0.0 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.960223 | 0.276608 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.210288 | 0.058413 | 0.002648 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |         0.0 |      nan |  295.236314 |  93.31059 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.1412 | 0.069087 |   0.0 |   0.0 |  0.0 |   0.0 | 0.058413 |
    | 2010-12-12 04:00:00 |  0.0 |  3.3 |              0.0 |                      0.0 |             0.0 |             90.9 |       4.6 |              1004.0 | 0.0 | 0.0 |  5.008333 |             88.295833 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  3.3 |  5.008333 |     3.68732 |         5.905055 |            3.68732 |          4.6 |                 7.744054 |                      8.734036 |                         0.0 |                          0.0 |                      0.549265 |                           0.610664 |             7.039346 |                   7.71179 |     996.960654 |   1.261813 | 8.583333 |       0.0 |      -19.8 |                   0.0 |                      0.0 |                       0.0 |                  13.571128 |                 13.098861 |                     0.0 |                      0.0 |   -13.098861 |             0.0 |              0.0 |          0.472267 |              46.30558 |                 100.0 |                82.256548 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.022297 | 0.002648 | 0.008766 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.956399 | 0.274409 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.199774 | 0.055493 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |         0.0 |      nan |  323.619647 | 93.301824 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.137553 | 0.062221 |   0.0 |   0.0 |  0.0 |   0.0 | 0.055493 |
    | 2010-12-12 05:00:00 |  0.0 |  3.0 |              0.0 |                      0.0 |             0.0 |             89.2 |       4.9 |              1005.0 | 0.0 | 0.0 |  4.933333 |             87.945833 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  3.0 |  4.933333 |    3.927797 |         5.824896 |           3.927797 |          4.9 |                 7.580812 |                      8.688342 |                         0.0 |                          0.0 |                      0.539048 |                           0.607851 |             6.762084 |                  7.641034 |     998.237916 |   1.264577 | 8.583333 |       0.0 |      -18.0 |                   0.0 |                      0.0 |                       0.0 |                  13.571128 |                 13.154767 |                     0.0 |                      0.0 |   -13.154767 |             0.0 |              0.0 |          0.416361 |             43.470544 |                 100.0 |                82.268069 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.028419 |      0.0 |  0.01221 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.952817 | 0.272358 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.190037 | 0.052788 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |         0.0 |      nan |   350.20298 | 93.289614 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |    0.134 | 0.056037 |   0.0 |   0.0 |  0.0 |   0.0 | 0.052788 |
    | 2010-12-12 06:00:00 |  0.0 |  1.9 |              0.0 |                      0.0 |             0.0 |             85.6 |       6.3 |              1006.0 | 0.0 | 0.0 |       4.8 |             87.454167 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  1.9 |       4.8 |    5.050025 |         5.794837 |           5.050025 |          6.3 |                 7.007901 |                      8.607627 |                         0.0 |                          0.0 |                      0.502964 |                           0.602876 |             5.998763 |                  7.527728 |    1000.001237 |   1.271267 | 8.583333 |       0.0 |      -11.4 |                   0.0 |                      0.0 |                       0.0 |                  13.571128 |                  13.24232 |                     0.0 |                      0.0 |    -13.24232 |             0.0 |              0.0 |          0.328808 |             33.810423 |                 100.0 |                82.284117 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.047698 |      0.0 | 0.017452 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.950125 | 0.270821 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.181006 | 0.050279 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |         0.0 |      nan |  370.186314 | 93.272161 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.130538 | 0.050467 |   0.0 |   0.0 |  0.0 |   0.0 | 0.050279 |
    | 2010-12-12 07:00:00 |  0.0 |  1.2 |              0.0 |                      0.0 |             0.0 |             86.3 |       5.8 |              1008.0 | 0.0 | 0.0 |    4.6375 |               87.0375 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  1.2 |    4.6375 |    4.649229 |         5.734717 |           4.649229 |          5.8 |                 6.663527 |                      8.510149 |                         0.0 |                          0.0 |                      0.481097 |                            0.59686 |             5.750624 |                  7.407021 |    1002.249376 |   1.277169 | 8.583333 |       0.0 |       -7.2 |                   0.0 |                      0.0 |                       0.0 |                  13.571128 |                 13.332335 |                     0.0 |                      0.0 |   -13.332335 |             0.0 |              0.0 |          0.238793 |             36.725115 |                 100.0 |                82.307055 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.040382 |      0.0 | 0.015359 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.947998 |  0.26961 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.172618 | 0.047949 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |         0.0 |      nan |  385.969647 | 93.256802 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.127167 | 0.045452 |   0.0 |   0.0 |  0.0 |   0.0 | 0.047949 |
    | 2010-12-12 08:00:00 |  0.0 |  0.8 |              0.0 |                 0.339738 |        0.894144 |             91.2 |       6.1 |              1009.0 | 0.0 | 0.0 |  4.445833 |             86.870833 |                   0.1 |                      7.393446 |            15.078757 |   0.0 |  0.8 |  4.445833 |    4.889707 |         5.694638 |           4.889707 |          6.1 |                 6.473522 |                      8.396426 |                         0.0 |                          0.0 |                      0.468972 |                            0.58983 |             5.903852 |                  7.294045 |    1003.096148 |   1.280232 | 8.583333 |       0.0 |       -4.8 |               0.80473 |                      0.0 |                   0.80473 |                  13.570881 |                 13.409662 |                     0.0 |                      0.0 |   -12.604932 |             0.0 |              0.0 |          0.161219 |             34.918962 |                 100.0 |                82.327243 |               98.490747 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.02629 |      0.0 | 0.009663 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.946195 | 0.268585 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.164816 | 0.045782 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |         0.0 |      nan |   399.35298 | 93.247139 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.123882 | 0.040934 |   0.0 |   0.0 |  0.0 |   0.0 | 0.045782 |
    | 2010-12-12 09:00:00 |  0.0 |  0.8 |              0.0 |                      1.0 |       17.712539 |             87.5 |       5.9 |              1010.0 | 0.0 | 0.0 |  4.258333 |             86.504167 |                   0.1 |                      7.393446 |              15.0973 |   0.0 |  0.8 |  4.258333 |    4.729388 |         5.654558 |           4.729388 |          5.9 |                 6.473522 |                      8.286472 |                         0.0 |                          0.0 |                      0.468972 |                           0.583022 |             5.664332 |                  7.168143 |    1004.335668 |   1.281619 | 8.583333 |       0.0 |       -4.8 |             15.941285 |                      0.0 |                 15.941285 |                   13.58757 |                 13.502128 |                     0.0 |                      0.0 |     2.439157 |             0.0 |              0.0 |          0.085442 |             36.102655 |                 100.0 |                82.339944 |               97.869917 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.045636 |      0.0 | 0.017181 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.944392 | 0.267563 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.157548 | 0.043763 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |         0.0 |      nan |  412.736314 | 93.229958 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.120682 | 0.036866 |   0.0 |   0.0 |  0.0 |   0.0 | 0.043763 |
    | 2010-12-12 10:00:00 |  0.1 |  0.5 |              0.0 |                      1.0 |       54.302586 |             84.5 |       6.4 |              1011.0 | 0.0 | 0.0 |      4.05 |             86.045833 |                   0.1 |                      7.393446 |            15.167507 | 0.094 |  0.5 |      4.05 |    5.130184 |         5.631178 |           5.130184 |          6.4 |                 6.334173 |                       8.16579 |                         0.0 |                     5.554612 |                      0.460051 |                           0.575537 |             5.352376 |                  7.026322 |    1005.647624 |   1.284447 | 8.583333 |       0.0 |   7.766328 |              5.556421 |                      0.0 |                  5.556421 |                    1.55199 |                 13.608639 |                     0.0 |                28.031613 |    -8.052217 |             0.0 |       -22.475192 |        -12.056649 |             33.282135 |                 100.0 |                82.362529 |               97.873154 |    0.034 | 0.01275 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.052426 |      0.0 |      0.0 |  0.004904 | 0.006889 |    -1.294388 |     0.897677 |      0.0 |      0.0 | 0.000192 | 0.000192 |     0.0 |   3.633499 |      0.0 |  -18.31764 |      0.0 |    7.791051 | 0.944282 |   0.2675 | 0.015768 | 0.005136 |  0.0 |  0.0 | 0.0 |  0.0 | 0.005136 | 0.150833 | 0.041898 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  0.009461 |  0.013329 |         0.0 |  0.00743 |   413.55332 | 93.240589 | 0.005136 |      0.0 |   0.0 |   0.0 |  0.0 | 0.117631 | 0.033202 |   0.0 |   0.0 |  0.0 |   0.0 | 0.041898 |
    | 2010-12-12 11:00:00 |  0.0 |  0.3 |              0.0 |                      1.0 |       81.504147 |             76.7 |       6.8 |              1012.0 | 0.0 | 0.0 |  3.808333 |                85.425 |                   0.1 |                      7.393446 |            15.224528 |   0.0 |  0.3 |  3.808333 |     5.45082 |         5.561039 |            5.45082 |          6.8 |                  6.24275 |                      8.027741 |                         0.0 |                     5.354685 |                      0.454184 |                           0.566956 |             4.788189 |                  6.857697 |    1007.211811 |   1.286932 | 8.583333 |       0.0 |  10.742078 |              8.523311 |                      0.0 |                  8.523311 |                   1.592108 |                 13.739115 |                     0.0 |                31.005039 |    -5.215805 |             0.0 |       -22.481728 |        -12.147007 |             31.324363 |                 100.0 |                82.348554 |               97.871152 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.084311 |     0.06 | 0.008323 |  0.013329 |      0.0 |    -1.790346 |     0.895425 |      0.0 |      0.0 | 0.000193 | 0.000193 |     0.0 |  10.689492 |      0.0 | -22.411256 |      0.0 |   10.759964 | 0.944573 | 0.267665 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |  0.14456 | 0.040156 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |         0.0 | 0.014738 |  411.394575 | 93.232266 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.114659 | 0.029902 |   0.0 |   0.0 |  0.0 |   0.0 | 0.040156 |
    | 2010-12-12 12:00:00 |  0.0 | -0.1 |              0.0 |                      1.0 |       81.665294 |             74.8 |       7.0 |              1013.0 | 0.0 | 0.0 |    3.5375 |             84.791667 |                   0.1 |                      7.393446 |            15.282411 |   0.0 | -0.1 |    3.5375 |    5.611139 |          5.47754 |           5.611139 |          7.0 |                 6.063393 |                      7.875477 |                         0.0 |                          0.0 |                      0.442644 |                           0.557471 |             4.535418 |                  6.677749 |    1008.464582 |   1.290215 | 8.583333 |       0.0 |        0.6 |             73.498765 |                      0.0 |                 73.498765 |                   13.75417 |                 13.878543 |                     0.0 |                      0.0 |    59.620222 |             0.0 |              0.0 |         -0.124374 |             30.429381 |                 100.0 |                82.359495 |                97.87272 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.127998 |      0.0 | 0.042746 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.943497 | 0.267056 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.138627 | 0.038507 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |         0.0 |      nan |  419.377908 | 93.189519 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.111697 |  0.02693 |   0.0 |   0.0 |  0.0 |   0.0 | 0.038507 |
    | 2010-12-12 13:00:00 |  0.0 | -0.4 |              0.0 |                      1.0 |       77.078958 |             69.7 |       6.3 |              1014.0 | 0.0 | 0.0 |  3.254167 |                84.125 |                   0.1 |                      7.393446 |            15.334225 |   0.0 | -0.4 |  3.254167 |    5.050025 |         5.424101 |           5.050025 |          6.3 |                 5.931876 |                      7.718916 |                         0.0 |                          0.0 |                      0.434155 |                           0.547693 |             4.134518 |                  6.493538 |    1009.865482 |   1.293105 | 8.583333 |       0.0 |        2.4 |             69.371062 |                      0.0 |                 69.371062 |                  13.800803 |                 14.023359 |                     0.0 |                      0.0 |    55.347703 |             0.0 |              0.0 |         -0.222556 |             33.810423 |                 100.0 |                82.415688 |               97.880768 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.131541 |      0.0 | 0.046731 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.942664 | 0.266585 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.133065 | 0.036963 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |         0.0 |      nan |  425.561242 | 93.142789 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.108812 | 0.024253 |   0.0 |   0.0 |  0.0 |   0.0 | 0.036963 |
    | 2010-12-12 14:00:00 |  0.0 | -0.6 |              0.2 |                      1.0 |       60.254042 |             66.9 |       7.3 |              1015.0 | 0.0 | 0.0 |  2.958333 |             83.483333 |                   0.3 |                      7.393446 |            16.085314 |   0.0 | -0.6 |  2.958333 |    5.851616 |         5.347282 |           5.851616 |          7.3 |                 5.845604 |                       7.55838 |                         0.0 |                          0.0 |                      0.428574 |                           0.537642 |             3.910709 |                  6.309988 |    1011.089291 |    1.29544 | 8.583333 |       0.0 |        3.6 |             54.228638 |                      0.0 |                 54.228638 |                  14.476783 |                 15.622576 |                     0.0 |                      0.0 |    38.606061 |             0.0 |              0.0 |         -1.145794 |             29.178858 |                 100.0 |                82.477126 |               97.889556 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.146499 |      0.0 | 0.047228 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.941992 | 0.266205 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.127844 | 0.035512 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |         0.0 |      nan |  430.544575 | 93.095561 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.106001 | 0.021843 |   0.0 |   0.0 |  0.0 |   0.0 | 0.035512 |
    | 2010-12-12 15:00:00 |  0.0 | -0.8 |              0.2 |                      1.0 |         14.8373 |             64.4 |       7.5 |              1015.0 | 0.0 | 0.0 |    2.6625 |             82.729167 |                   0.4 |                      7.393446 |            16.177955 |   0.0 | -0.8 |    2.6625 |    6.011934 |         5.293842 |           6.011934 |          7.5 |                 5.760442 |                      7.400795 |                         0.0 |                          0.0 |                      0.423055 |                           0.527748 |             3.709725 |                  6.122616 |    1011.290275 |   1.296488 | 8.583333 |       0.0 |        4.8 |              13.35357 |                      0.0 |                  13.35357 |                   14.56016 |                 16.526635 |                     0.0 |                      0.0 |    -3.173065 |             0.0 |              0.0 |         -1.966476 |             28.400756 |                 100.0 |                82.539225 |               97.898427 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.135015 |      0.0 | 0.042618 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.941483 | 0.265917 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.122935 | 0.034149 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |         0.0 |      nan |  434.327908 | 93.052943 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.103263 | 0.019672 |   0.0 |   0.0 |  0.0 |   0.0 | 0.034149 |
    | 2010-12-12 16:00:00 |  0.0 | -1.3 |              0.2 |                 0.046405 |        0.168856 |             65.4 |       7.0 |              1017.0 | 0.0 | 0.0 |    2.3625 |             82.029167 |                   0.6 |                      7.386143 |            16.184078 |   0.0 | -1.3 |    2.3625 |    5.611139 |         5.237063 |           5.611139 |          7.0 |                 5.552313 |                      7.243956 |                         0.0 |                          0.0 |                      0.409522 |                           0.517875 |             3.631213 |                  5.942156 |    1013.368787 |   1.301474 | 8.583333 |       0.0 |        7.8 |               0.15197 |                      0.0 |                   0.15197 |                   14.56567 |                 18.187261 |                     0.0 |                      0.0 |   -18.035291 |             0.0 |              0.0 |         -3.621591 |             30.429381 |                 100.0 |                 82.59527 |                98.77173 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.11254 |      0.0 | 0.036657 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.941377 | 0.265857 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.118312 | 0.032865 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |         0.0 |      nan |  435.111242 | 93.016286 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.100595 | 0.017717 |   0.0 |   0.0 |  0.0 |   0.0 | 0.032865 |
    | 2010-12-12 17:00:00 |  0.0 | -1.6 |              0.0 |                      0.0 |             0.0 |             67.6 |       6.0 |              1017.0 | 0.0 | 0.0 |  2.066667 |             81.658333 |                   0.6 |                      7.386143 |            16.184078 |   0.0 | -1.6 |  2.066667 |    4.809547 |         5.140204 |           4.809547 |          6.0 |                 5.430651 |                      7.092172 |                         0.0 |                          0.0 |                      0.401583 |                           0.508295 |              3.67112 |                  5.791349 |     1013.32888 |   1.302892 | 8.583333 |       0.0 |        9.6 |                   0.0 |                      0.0 |                       0.0 |                   14.56567 |                 18.333316 |                     0.0 |                      0.0 |   -18.333316 |             0.0 |              0.0 |         -3.767646 |             35.500944 |                 100.0 |                 82.64348 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.08847 |      0.0 |  0.03163 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.941514 | 0.265935 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.113953 | 0.031654 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |         0.0 |      nan |  434.094575 | 92.984656 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.097997 | 0.015956 |   0.0 |   0.0 |  0.0 |   0.0 | 0.031654 |
    | 2010-12-12 18:00:00 |  0.0 | -2.0 |              0.0 |                      0.0 |             0.0 |             70.2 |       4.8 |              1018.0 | 0.0 | 0.0 |  1.770833 |             81.179167 |                   0.6 |                      7.386143 |            16.184078 |   0.0 | -2.0 |  1.770833 |    3.847638 |         5.036665 |           3.847638 |          4.8 |                 5.272101 |                      6.943199 |                         0.0 |                          0.0 |                      0.391203 |                           0.498867 |             3.701015 |                  5.636431 |    1014.298985 |   1.306085 | 8.583333 |       0.0 |       12.0 |                   0.0 |                      0.0 |                       0.0 |                   14.56567 |                 18.489582 |                     0.0 |                      0.0 |   -18.489582 |             0.0 |              0.0 |         -3.923912 |             44.376181 |                 100.0 |                82.685083 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.063209 |      0.0 | 0.025658 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.941975 | 0.266195 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.109836 |  0.03051 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |         0.0 |      nan |  430.677908 | 92.958998 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.095466 |  0.01437 |   0.0 |   0.0 |  0.0 |   0.0 |  0.03051 |
    | 2010-12-12 19:00:00 |  0.0 | -2.1 |              0.0 |                      0.0 |             0.0 |             70.4 |       4.2 |              1019.0 | 0.0 | 0.0 |  1.483333 |               80.6375 |                   0.6 |                      7.386143 |            16.184078 |   0.0 | -2.1 |  1.483333 |    3.366683 |         4.939806 |           3.366683 |          4.2 |                 5.233109 |                      6.801075 |                         0.0 |                          0.0 |                      0.388644 |                           0.489848 |             3.684109 |                  5.484217 |    1015.315891 |    1.30786 | 8.583333 |       0.0 |       12.6 |                   0.0 |                      0.0 |                       0.0 |                   14.56567 |                 18.648085 |                     0.0 |                      0.0 |   -18.648085 |             0.0 |              0.0 |         -4.082415 |             50.715635 |                 100.0 |                82.718834 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.054415 |      0.0 |  0.02375 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.942516 | 0.266501 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.105942 | 0.029428 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |         0.0 |      nan |  426.661242 | 92.935248 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |    0.093 | 0.012942 |   0.0 |   0.0 |  0.0 |   0.0 | 0.029428 |
    | 2010-12-12 20:00:00 |  0.0 | -2.2 |              0.0 |                      0.0 |             0.0 |             69.5 |       4.9 |              1019.0 | 0.0 | 0.0 |    1.1875 |             80.083333 |                   0.6 |                      7.386143 |            16.184078 |   0.0 | -2.2 |    1.1875 |    3.927797 |         4.889707 |           3.927797 |          4.9 |                 5.194372 |                      6.657516 |                         0.0 |                          0.0 |                        0.3861 |                           0.480714 |             3.610088 |                  5.331561 |    1015.389912 |   1.308379 | 8.583333 |       0.0 |       13.2 |                   0.0 |                      0.0 |                       0.0 |                   14.56567 |                 18.808978 |                     0.0 |                      0.0 |   -18.808978 |             0.0 |              0.0 |         -4.243308 |             43.470544 |                 100.0 |                82.750077 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.065452 |      0.0 |  0.02619 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.943138 | 0.266853 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.102253 | 0.028404 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |         0.0 |      nan |  422.044575 | 92.909058 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.090597 | 0.011656 |   0.0 |   0.0 |  0.0 |   0.0 | 0.028404 |
    | 2010-12-12 21:00:00 |  0.0 | -2.6 |              0.0 |                      0.0 |             0.0 |             71.6 |       4.0 |              1020.0 | 0.0 | 0.0 |     0.875 |             79.491667 |                   0.6 |                      7.386143 |            16.184078 |   0.0 | -2.6 |     0.875 |    3.206365 |         4.782828 |           3.206365 |          4.0 |                 5.041946 |                       6.50878 |                         0.0 |                          0.0 |                      0.376066 |                           0.471225 |             3.610033 |                  5.173937 |    1016.389967 |   1.311601 | 8.583333 |       0.0 |       15.6 |                   0.0 |                      0.0 |                       0.0 |                   14.56567 |                 18.977604 |                     0.0 |                      0.0 |   -18.977604 |             0.0 |              0.0 |         -4.411934 |             53.251417 |                 100.0 |                82.784532 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.048111 |      0.0 | 0.021424 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.944083 | 0.267388 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.098754 | 0.027432 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |         0.0 |      nan |  415.027908 | 92.887633 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.088257 | 0.010497 |   0.0 |   0.0 |  0.0 |   0.0 | 0.027432 |
    | 2010-12-12 22:00:00 |  0.0 | -3.1 |              0.0 |                      0.0 |             0.0 |             74.5 |       3.4 |              1020.0 | 0.0 | 0.0 |      0.55 |             78.945833 |                   0.6 |                      7.386143 |            16.184078 |   0.0 | -3.1 |      0.55 |     2.72541 |         4.662589 |            2.72541 |          3.4 |                 4.856984 |                      6.357212 |                         0.0 |                          0.0 |                       0.36384 |                           0.461527 |             3.618453 |                  5.018754 |    1016.381547 |   1.314025 | 8.583333 |       0.0 |       18.6 |                   0.0 |                      0.0 |                       0.0 |                   14.56567 |                 19.143113 |                     0.0 |                      0.0 |   -19.143113 |             0.0 |              0.0 |         -4.577443 |             62.648725 |                 100.0 |                82.812719 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.034966 |      0.0 | 0.016822 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.945433 | 0.268153 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.095431 | 0.026509 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |         0.0 |      nan |  405.011242 | 92.870811 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.085977 | 0.009454 |   0.0 |   0.0 |  0.0 |   0.0 | 0.026509 |
    | 2010-12-12 23:00:00 |  0.0 | -3.9 |              0.0 |                      0.0 |             0.0 |             80.2 |       3.4 |              1020.0 | 0.0 | 0.0 |  0.229167 |             78.720833 |                   0.6 |                      7.386143 |            16.184078 |   0.0 | -3.9 |  0.229167 |     2.72541 |          4.54903 |            2.72541 |          3.4 |                  4.57351 |                      6.210652 |                         0.0 |                          0.0 |                      0.344989 |                           0.452122 |             3.667955 |                  4.889077 |    1016.332045 |   1.317905 | 8.583333 |       0.0 |       23.4 |                   0.0 |                      0.0 |                       0.0 |                   14.56567 |                 19.269853 |                     0.0 |                      0.0 |   -19.269853 |             0.0 |              0.0 |         -4.704183 |             62.648725 |                 100.0 |                82.834852 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.024921 |      0.0 |  0.01189 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.947429 | 0.269286 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.092271 | 0.025631 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |         0.0 |      nan |  390.194575 | 92.858921 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.083757 | 0.008514 |   0.0 |   0.0 |  0.0 |   0.0 | 0.025631 |

>>> round_(model.check_waterbalance(conditions_acker_winter))
0.0

.. _lland_v4_nadelw_winter_hourly:

conifers (winter)
-----------------

.. integration-test::

    >>> lnk(NADELW)
    >>> test("lland_v4_nadelw_winter_hourly",
    ...      axis1=(inputs.nied, fluxes.wada), axis2=(states.waes, states.wats),
    ...      use_conditions=conditions_nadelw_winter)
    |                date | nied | teml | sunshineduration | possiblesunshineduration | globalradiation | relativehumidity | windspeed | atmosphericpressure |  qz | qzh |   temltag | dailyrelativehumidity | dailysunshineduration | dailypossiblesunshineduration | dailyglobalradiation |  nkor | tkor |   tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair |        g |        tz |         wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |  nbes |   sbes | snowintmax | snowintrate |  nbesinz |  sbesinz | wniedinz | actualalbedoinz |  wadainz | schmpotinz |  schminz | gefrpotinz |  gefrinz |   evsinz |      evpo |      evi |       evb |       evs |     wnied | tempssurface | actualalbedo |  schmpot |     schm |  gefrpot |     gefr |   wlatinz |   wlatsnow |   wsensinz |  wsenssnow |   wsurfinz |       wsurf |      sff |      fvg |     wada |      qdb | qib1 | qib2 | qbb | qkap |     qdgz |      qah |       qa |     inzp |    stinz |     sinz |   esnowinz |    asinz |     wats |      waes |      esnow |     taus |        ebdn |      bowa |    qdgz1 |    qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 |    qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 2010-12-10 00:00:00 |  0.0 | -3.7 |              0.0 |                      0.0 |             0.0 |             87.5 |       3.4 |              1017.0 | 0.0 | 0.0 |    -2.025 |             96.870833 |                  4.41 |                           7.5 |            23.154167 |   0.0 | -3.7 |    -2.025 |     2.72541 |         2.684392 |           1.206967 |          3.4 |                  4.64297 |                      5.262329 |                     4.64297 |                     4.770504 |                      0.349621 |                           0.390562 |             4.062599 |                  5.097662 |    1012.937401 |   1.312855 | 8.583333 | -1.772272 |   9.403848 |                   0.0 |                      0.0 |                       0.0 |                   5.706576 |                 46.101494 |               48.299114 |                23.930251 |   -46.101494 |      -48.299114 |       -23.930251 |        -40.394918 |             27.647059 |                 100.0 |              3138.827437 |               69.444444 |   0.0 |    0.0 |       24.5 |    0.439593 |      0.0 |      0.0 |      0.0 |        0.750856 |      0.0 |        0.0 |      0.0 |   0.355784 |      0.0 | 0.005419 |  0.023429 |      0.0 |  0.008725 |   0.00661 |       0.0 |     -3.33958 |      0.75354 |      0.0 |      0.0 | 0.224635 | 0.224635 |  4.015484 |   4.897869 | -19.678059 |   1.416865 |   32.63654 |   30.244985 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |      2.2 | 6.525649 | 6.525649 | -33.008866 | 0.899234 | 8.783304 |  9.648678 |        0.0 | 0.868993 |  -148.50983 | 70.672352 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 01:00:00 |  0.0 | -4.8 |              0.0 |                      0.0 |             0.0 |             90.7 |       2.9 |              1018.0 | 0.0 | 0.0 | -2.158333 |                96.825 |                  4.41 |                           7.5 |            23.154167 |   0.0 | -4.8 | -2.158333 |    2.324615 |         2.647918 |           1.029472 |          2.9 |                 4.272186 |                      5.210481 |                    4.272186 |                     4.613345 |                      0.324793 |                           0.387158 |             3.874873 |                  5.045049 |    1014.125127 |   1.319627 | 8.583333 | -1.782118 |  12.017133 |                   0.0 |                      0.0 |                       0.0 |                   5.721775 |                 46.220713 |               39.666214 |                26.684557 |   -46.220713 |      -39.666214 |       -26.684557 |        -40.498937 |             32.413793 |                 100.0 |             20214.096781 |               69.444444 |   0.0 |    0.0 |       24.5 |    0.439577 |      0.0 |      0.0 |      0.0 |        0.750674 |      0.0 |        0.0 |      0.0 |   0.509252 |      0.0 | 0.003442 |  0.007944 |      0.0 |  0.003207 |  0.006397 |       0.0 |    -3.784974 |     0.752884 |      0.0 |      0.0 | 0.249086 | 0.249086 |  2.550352 |   4.740251 | -27.978162 |   3.701959 |  14.238404 |   35.126767 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |      2.2 | 6.522207 | 6.522207 |  -47.24727 |  0.90132 | 9.026401 |  9.642281 |        0.0 | 0.876301 |  -151.94363 | 70.669145 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 02:00:00 |  0.0 | -5.6 |              0.0 |                      0.0 |             0.0 |             89.7 |       2.5 |              1018.0 | 0.0 | 0.0 |    -2.325 |             96.604167 |                  4.41 |                           7.5 |            23.154167 |   0.0 | -5.6 |    -2.325 |    2.003978 |         2.606417 |           0.887476 |          2.5 |                 4.019239 |                      5.146307 |                    4.019239 |                     4.483734 |                      0.307706 |                            0.38294 |             3.605258 |                  4.971547 |    1014.394742 |   1.323706 | 8.583333 | -1.823324 |  14.035154 |                   0.0 |                      0.0 |                       0.0 |                   5.736857 |                  46.40442 |                36.46814 |                 28.62577 |    -46.40442 |       -36.46814 |        -28.62577 |        -40.667563 |                  37.6 |                 100.0 |             28304.769603 |               69.444444 |   0.0 |    0.0 |       24.5 |    0.439567 |      0.0 |      0.0 |      0.0 |        0.750534 |      0.0 |        0.0 |      0.0 |   0.596683 |      0.0 | 0.003363 |  0.006638 |      0.0 |  0.002886 |  0.007136 |       0.0 |    -4.162516 |     0.752232 |      0.0 |      0.0 | 0.267245 | 0.267245 |   2.49181 |   5.287666 | -30.848284 |   4.916141 |   8.111667 |   38.829577 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |      2.2 | 6.518844 | 6.518844 | -55.358937 | 0.902928 | 9.286768 |  9.635145 |        0.0 | 0.883609 |  -157.39545 | 70.666259 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 03:00:00 |  0.0 | -6.2 |              0.0 |                      0.0 |             0.0 |             91.3 |       3.0 |              1019.0 | 0.0 | 0.0 | -2.420833 |                96.325 |                  4.41 |                           7.5 |            23.154167 |   0.0 | -6.2 | -2.420833 |    2.404774 |         2.602449 |           1.064971 |          3.0 |                 3.838328 |                      5.109724 |                    3.838328 |                     4.372555 |                      0.295404 |                           0.380532 |             3.504393 |                  4.921942 |    1015.495607 |   1.328036 | 8.583333 | -1.888745 |  15.632395 |                   0.0 |                      0.0 |                       0.0 |                   5.751822 |                 46.542687 |               35.745783 |                29.705445 |   -46.542687 |      -35.745783 |       -29.705445 |        -40.790865 |             31.333333 |                 100.0 |             28305.809028 |               69.444444 |   0.0 |    0.0 |       24.5 |    0.439557 |      0.0 |      0.0 |      0.0 |        0.750414 |      0.0 |        0.0 |      0.0 |   0.633219 |      0.0 | 0.002938 |  0.006318 |      0.0 |  0.002458 |  0.007638 |       0.0 |    -4.494145 |     0.751586 |      0.0 |      0.0 | 0.280789 | 0.280789 |  2.176905 |   5.659509 | -34.532989 |    6.31841 |   3.389699 |   41.683364 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |      2.2 | 6.515906 | 6.515906 | -58.748636 | 0.904314 | 9.559972 |  9.627506 |        0.0 | 0.890917 | -164.444512 | 70.663801 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 04:00:00 |  0.0 | -7.3 |              0.0 |                      0.0 |             0.0 |             94.2 |       2.5 |              1020.0 | 0.0 | 0.0 | -2.558333 |               96.2125 |                  4.41 |                           7.5 |            23.154167 |   0.0 | -7.3 | -2.558333 |    2.003978 |         2.623448 |           0.887476 |          2.5 |                  3.52531 |                      5.057637 |                     3.52531 |                      4.22261 |                      0.273951 |                           0.377101 |             3.320842 |                  4.866079 |    1016.679158 |   1.334932 | 8.583333 | -1.973334 |  17.880559 |                   0.0 |                      0.0 |                       0.0 |                   5.766672 |                 46.677297 |               35.778488 |                32.308553 |   -46.677297 |      -35.778488 |       -32.308553 |        -40.910626 |                  37.6 |                 100.0 |             28306.694408 |               69.444444 |   0.0 |    0.0 |       24.5 |    0.439548 |      0.0 |      0.0 |      0.0 |        0.750301 |      0.0 |        0.0 |      0.0 |   0.683384 |      0.0 | 0.001661 |  -0.00337 |      0.0 | -0.001439 |  0.007325 |       0.0 |    -4.953427 |     0.750945 |      0.0 |      0.0 | 0.300514 | 0.067534 |   1.23072 |   5.427864 | -32.354981 |   8.025188 |   4.654227 |   45.761605 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |      2.2 | 6.514245 | 6.514245 | -63.402863 | 0.905616 | 9.620181 |  9.620181 | -21.615366 | 0.898225 | -173.741738 |  70.66524 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 05:00:00 |  0.0 | -8.3 |              0.0 |                      0.0 |             0.0 |             93.2 |       2.3 |              1020.0 | 0.0 | 0.0 |    -2.725 |             96.004167 |                  4.41 |                           7.5 |            23.154167 |   0.0 | -8.3 |    -2.725 |     1.84366 |         2.596101 |           0.816478 |          2.3 |                  3.26061 |                      4.995131 |                     3.26061 |                     3.766374 |                       0.25563 |                           0.372977 |             3.038889 |                  4.795534 |    1016.961111 |   1.340112 | 8.583333 | -2.084901 |  26.164341 |                   0.0 |                      0.0 |                       0.0 |                   5.773006 |                 46.854961 |               36.074771 |                30.183482 |   -46.854961 |      -36.074771 |       -30.183482 |        -41.081955 |             40.869565 |                 100.0 |             28306.176285 |               69.444444 |   0.0 |    0.0 |       24.5 |    0.439543 |      0.0 |      0.0 |      0.0 |        0.750197 |      0.0 |        0.0 |      0.0 |   0.737309 |      0.0 | 0.001741 | -0.002347 |      0.0 | -0.001038 |  0.005713 |       0.0 |    -6.445624 |     0.750671 |      0.0 |      0.0 | 0.388015 |      0.0 |  1.290242 |   4.233383 | -32.362003 |   6.131242 |    5.00301 |   40.548106 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |      2.2 | 6.512504 | 6.512504 | -68.405873 | 0.906812 | 9.614468 |  9.614468 | -35.999131 | 0.901359 | -191.322746 | 70.666278 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 06:00:00 |  0.0 | -8.6 |              0.0 |                      0.0 |             0.0 |             96.5 |       2.1 |              1021.0 | 0.0 | 0.0 |    -2.925 |             96.004167 |                  4.41 |                           7.5 |            23.154167 |   0.0 | -8.6 |    -2.925 |    1.683342 |          2.56624 |            0.74548 |          2.1 |                 3.184716 |                      4.921026 |                    3.184716 |                      3.53143 |                      0.250344 |                           0.368079 |             3.073251 |                   4.72439 |    1017.926749 |   1.342932 | 8.583333 | -2.295873 |  29.890831 |                   0.0 |                      0.0 |                       0.0 |                   5.777969 |                 47.012175 |               34.424201 |                27.717256 |   -47.012175 |      -34.424201 |       -27.717256 |        -41.234206 |             44.761905 |                 100.0 |             28305.802404 |               69.444444 |   0.0 |    0.0 |       24.5 |    0.439538 |      0.0 |      0.0 |      0.0 |        0.750103 |      0.0 |        0.0 |      0.0 |   0.765024 |      0.0 | 0.000845 |  -0.00857 |      0.0 | -0.003968 |  0.003475 |       0.0 |    -7.277678 |     0.750457 |      0.0 |      0.0 | 0.437842 |      0.0 |  0.626352 |   2.574628 | -32.479167 |   4.221867 |   2.571387 |   34.513752 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |      2.2 | 6.511659 | 6.511659 |  -70.97726 | 0.907902 | 9.610993 |  9.610993 | -40.622052 | 0.903822 | -212.630243 | 70.670245 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 07:00:00 |  0.0 | -8.5 |              0.0 |                      0.0 |             0.0 |             95.9 |       1.4 |              1021.0 | 0.0 | 0.0 |   -3.1625 |               96.0625 |                  3.76 |                           7.4 |            23.154167 |   0.0 | -8.5 |   -3.1625 |    1.122228 |         2.521333 |           0.496987 |          1.4 |                 3.209838 |                      4.834291 |                    3.209838 |                     3.475879 |                      0.252096 |                           0.362336 |             3.078235 |                  4.643941 |    1017.921765 |   1.342422 | 8.583333 | -2.551563 |  29.580037 |                   0.0 |                      0.0 |                       0.0 |                   5.782618 |                  42.68596 |               30.562509 |                25.900101 |    -42.68596 |      -30.562509 |       -25.900101 |        -36.903342 |             67.142857 |                 100.0 |             28304.373337 |               69.444444 |   0.0 |    0.0 |       24.5 |    0.439535 |      0.0 |      0.0 |      0.0 |        0.750013 |      0.0 |        0.0 |      0.0 |   0.780448 |      0.0 | 0.000874 | -0.007649 |      0.0 |  -0.00429 |   0.00264 |       0.0 |    -7.481569 |     0.750256 |      0.0 |      0.0 | 0.449946 |      0.0 |  0.647425 |   1.956214 |  -29.77897 |   2.846696 |   1.430964 |   30.703011 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |      2.2 | 6.510785 | 6.510785 | -72.408224 | 0.908943 | 9.608353 |  9.608353 | -41.745026 | 0.906135 | -233.626947 | 70.674536 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 08:00:00 |  0.0 | -9.5 |              0.0 |                 0.355232 |        1.107611 |             96.8 |       1.1 |              1022.0 | 0.0 | 0.0 | -3.483333 |               96.2625 |                  2.76 |                      6.755232 |            23.200317 |   0.0 | -9.5 | -3.483333 |     0.88175 |         2.408072 |           0.390489 |          1.1 |                 2.966351 |                       4.71927 |                    2.966351 |                     3.350494 |                      0.235051 |                           0.354699 |             2.871427 |                  4.542888 |    1019.128573 |   1.348938 | 8.583333 | -2.803523 |  30.894515 |              0.276838 |                 0.225545 |                  0.051413 |                   5.798721 |                 37.241518 |               29.067635 |                27.287337 |    -36.96468 |       -28.84209 |       -27.235924 |        -31.442797 |             85.454545 |                 100.0 |             28302.828119 |               106.75539 |   0.0 |    0.0 |       24.5 |    0.439532 |      0.0 |      0.0 |      0.0 |        0.749925 |      0.0 |        0.0 |      0.0 |   0.805479 |      0.0 | 0.000592 |  -0.00665 |      0.0 | -0.003375 |  0.002987 |       0.0 |    -7.952609 |     0.750059 |      0.0 |      0.0 | 0.478144 |      0.0 |  0.438509 |   2.213105 | -26.958254 |   4.061565 |   2.322345 |   33.510595 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |      2.2 | 6.510193 | 6.510193 | -74.730569 | 0.909956 | 9.605366 |  9.605366 | -44.361106 | 0.908414 | -255.938128 | 70.677911 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 09:00:00 |  0.0 | -9.2 |              0.0 |                      1.0 |       18.663203 |             94.3 |       1.0 |              1022.0 | 0.0 | 0.0 | -3.816667 |             96.383333 |                  1.79 |                      6.755232 |            21.661284 |   0.0 | -9.2 | -3.816667 |    0.801591 |         2.278972 |            0.35499 |          1.0 |                 3.037615 |                      4.602339 |                    3.037615 |                     3.350843 |                      0.240056 |                           0.346913 |             2.864471 |                  4.435888 |    1019.135529 |   1.347408 | 8.583333 | -3.071258 |  29.280122 |              4.668258 |                 3.801704 |                  0.866962 |                   5.418173 |                 29.260785 |               24.708347 |                24.720938 |   -24.592528 |      -20.906644 |       -23.853976 |        -23.842613 |                  94.0 |                 100.0 |             28301.612551 |             4303.524339 |   0.0 |    0.0 |       24.5 |    0.439531 |      0.0 |      0.0 |      0.0 |        0.749841 |      0.0 |        0.0 |      0.0 |   0.770433 |      0.0 | 0.001056 | -0.000901 |      0.0 | -0.000025 |  0.002967 |       0.0 |    -7.951278 |     0.749868 |      0.0 |      0.0 | 0.477915 |      0.0 |  0.782551 |   2.198233 | -24.940692 |   3.206699 |  -3.251497 |   29.258908 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |      2.2 | 6.509137 | 6.509137 | -71.479072 | 0.910927 | 9.602399 |  9.602399 | -44.339892 | 0.910615 | -276.634917 | 70.677936 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 10:00:00 |  0.0 | -6.9 |              0.0 |                      1.0 |       56.267908 |             96.7 |       1.4 |              1022.0 | 0.0 | 0.0 |   -4.0625 |               96.5375 |                  0.88 |                      6.755232 |             18.21828 |   0.0 | -6.9 |   -4.0625 |    1.122228 |         2.200731 |           0.496987 |          1.4 |                 3.636412 |                      4.517752 |                    3.636412 |                     3.641135 |                      0.281591 |                           0.341265 |             3.516411 |                  4.361325 |    1018.483589 |   1.335446 | 8.583333 | -3.319619 |  21.381715 |             14.085064 |                11.465871 |                  2.615798 |                   4.560426 |                 21.675213 |               18.472727 |                17.369353 |    -7.590149 |       -7.006856 |       -14.753556 |        -17.114787 |             67.142857 |                 100.0 |             28301.603673 |             4303.524138 |   0.0 |    0.0 |       24.5 |    0.439527 |      0.0 |      0.0 |      0.0 |        0.749753 |      0.0 |        0.0 |      0.0 |   0.609423 |      0.0 | 0.000797 |  0.004327 |      0.0 |   0.00009 |  0.000828 |       0.0 |    -6.883238 |     0.749679 |      0.0 |      0.0 | 0.413592 |      0.0 |   0.59035 |   0.613583 |  -22.53533 |   0.046852 | -14.938124 |   15.413991 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |      2.2 |  6.50834 |  6.50834 | -56.540947 | 0.911957 | 9.601571 |  9.601571 | -38.372168 | 0.912817 | -289.433298 | 70.677846 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 11:00:00 |  0.0 | -4.0 |              0.1 |                      1.0 |      107.168236 |             97.4 |       1.3 |              1022.0 | 0.0 | 0.0 | -4.195833 |             96.570833 |                  0.39 |                      6.755232 |             15.73779 |   0.0 | -4.0 | -4.195833 |    1.042069 |         2.098318 |           0.461488 |          1.3 |                 4.539126 |                      4.472452 |                    4.539126 |                      4.14924 |                      0.342693 |                           0.338235 |             4.421109 |                  4.319085 |    1017.578891 |   1.320615 | 8.583333 |   -3.4732 |   10.26148 |             26.848461 |                 21.84815 |                  4.986143 |                    3.94273 |                 17.574343 |               16.533721 |                10.320264 |     9.274118 |        5.314429 |        -5.334121 |        -13.631612 |             72.307692 |                 100.0 |              3985.748396 |             2253.949862 |   0.0 |    0.0 |       24.5 |    0.439525 |      0.0 |      0.0 |      0.0 |        0.749636 |      0.0 |        0.0 |      0.0 |   0.392138 |      0.0 | 0.000768 |  0.011187 |      0.0 |  0.000489 | -0.001768 |       0.0 |    -5.183446 |     0.749474 |      0.0 |      0.0 |  0.31143 |      0.0 |   0.56879 |  -1.310282 | -15.413603 |  -3.240726 | -20.159242 |    0.783114 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |      2.2 | 6.507572 | 6.507572 | -36.381705 | 0.913311 | 9.603339 |  9.603339 | -28.893802 |   0.9152 | -291.111445 | 70.677357 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 12:00:00 |  0.0 | -2.8 |              0.0 |                      1.0 |       91.473333 |             93.9 |       1.3 |              1022.0 | 0.0 | 0.0 |     -4.35 |               96.3625 |                  0.14 |                      6.755232 |            14.920012 |   0.0 | -2.8 |     -4.35 |    1.042069 |          2.00007 |           0.461488 |          1.3 |                 4.967227 |                      4.420576 |                    4.967227 |                     4.457474 |                      0.371134 |                            0.33476 |             4.664226 |                  4.259778 |    1017.335774 |   1.314635 | 8.583333 | -3.493337 |   4.481067 |             22.937962 |                18.660993 |                  4.259907 |                    3.74136 |                 15.511434 |               18.012332 |                 8.427556 |     7.426528 |        0.648661 |        -4.167649 |        -11.770074 |             72.307692 |                 100.0 |              1828.729747 |              1362.86617 |   0.0 |    0.0 |      26.95 |    0.439523 |      0.0 |      0.0 |      0.0 |        0.749468 |      0.0 |        0.0 |      0.0 |   0.280336 |      0.0 | 0.001971 |  0.015836 |      0.0 |  0.001138 | -0.001345 |       0.0 |    -4.240182 |     0.749239 |      0.0 |      0.0 | 0.254804 |      0.0 |  1.460327 |   -0.99645 | -11.184395 |  -3.943765 | -10.372728 |   -0.772567 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |      2.2 | 6.505602 | 6.505602 | -26.008977 | 0.915268 | 9.604684 |  9.604684 | -23.640168 | 0.917939 | -287.009179 | 70.676218 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 13:00:00 |  0.0 | -1.4 |              0.0 |                      1.0 |       80.475536 |             93.4 |       1.4 |              1021.0 | 0.0 | 0.0 | -4.370833 |             96.116667 |                   0.1 |                      6.755232 |            15.956493 |   0.0 | -1.4 | -4.370833 |    1.122228 |         1.900997 |           0.496987 |          1.4 |                 5.511494 |                      4.413607 |                    5.511494 |                     4.766224 |                      0.406861 |                           0.334293 |             5.147736 |                  4.242212 |    1015.852264 |   1.306346 | 8.583333 |  -3.44411 |  -0.555439 |             20.200827 |                16.430777 |                  3.751582 |                   4.005371 |                 15.193165 |               17.059386 |                 5.576071 |     5.007662 |       -0.628609 |        -1.824489 |        -11.187794 |             67.142857 |                 100.0 |              1121.031211 |               932.72268 |   0.0 |    0.0 |       44.1 |    0.439517 |      0.0 |      0.0 |      0.0 |        0.749264 |      0.0 |        0.0 |      0.0 |   0.200323 |      0.0 | 0.002415 |  0.017223 |      0.0 |  0.001687 | -0.002533 |       0.0 |    -3.351537 |     0.748982 |      0.0 |      0.0 | 0.201431 |      0.0 |  1.789516 |  -1.876854 |   -9.84154 |  -5.454894 |  -7.423416 |   -5.507259 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |      2.2 | 6.503186 | 6.503186 | -18.585561 |  0.91765 | 9.607217 |  9.607217 | -18.688347 | 0.920945 | -277.870407 | 70.674531 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 14:00:00 |  0.0 | -0.4 |              0.0 |                      1.0 |       43.120592 |             93.3 |       2.3 |              1020.0 | 0.0 | 0.0 | -4.354167 |             95.854167 |                   0.1 |                      6.755232 |            16.594851 |   0.0 | -0.4 | -4.354167 |     1.84366 |         1.856982 |           0.816478 |          2.3 |                 5.931876 |                      4.419181 |                    5.931876 |                     5.033102 |                      0.434155 |                           0.334667 |              5.53444 |                   4.23597 |     1014.46556 |   1.300092 | 8.583333 | -3.334445 |  -4.265454 |             10.836352 |                 8.812405 |                  2.012465 |                   4.170343 |                 15.206659 |               14.873485 |                 3.960708 |    -4.370307 |        -6.06108 |        -1.948243 |        -11.036316 |             40.869565 |                 100.0 |               878.854887 |              761.569353 |   0.0 |    0.0 |       49.0 |     0.43951 |      0.0 |      0.0 |      0.0 |        0.749024 |      0.0 |        0.0 |      0.0 |   0.162514 |      0.0 | 0.003121 |  0.020183 |      0.0 |   0.00158 | -0.003937 |       0.0 |    -2.623536 |     0.748697 |      0.0 |      0.0 | 0.157719 |      0.0 |  2.312759 |   -2.91739 | -11.881736 |  -7.351821 |  -3.507898 |   -8.320968 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |      2.2 | 6.500065 | 6.500065 | -15.077663 | 0.920455 | 9.611155 |  9.611155 | -14.632833 | 0.924292 |  -265.02162 | 70.672951 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 15:00:00 |  0.0 |  0.5 |              0.0 |                      1.0 |       10.320092 |             94.1 |       2.8 |              1020.0 | 0.0 | 0.0 |      -4.3 |             95.608333 |                   0.1 |                      7.355232 |            17.024855 |   0.0 |  0.5 |      -4.3 |    2.244455 |         1.842168 |           0.993973 |          2.8 |                 6.334173 |                      4.437342 |                    6.334173 |                     5.290061 |                      0.460051 |                           0.335883 |             5.960456 |                  4.242469 |    1014.039544 |   1.295611 | 8.583333 | -3.180259 |  -7.356597 |              2.596745 |                 2.111295 |                  0.482253 |                     4.2838 |                  15.14111 |               11.539149 |                 2.444218 |   -12.544364 |       -9.427854 |        -1.961966 |        -10.857309 |             33.571429 |                 100.0 |               734.168692 |              652.364835 |   0.0 |    0.0 |       49.0 |      0.4395 |      0.0 |      0.0 |      0.0 |         0.74876 |      0.0 |        0.0 |      0.0 |   0.138481 |      0.0 | 0.003187 |  0.017602 |      0.0 |  0.001363 | -0.005717 |       0.0 |     -1.95416 |      0.74838 |      0.0 |      0.0 | 0.117526 |      0.0 |  2.361523 |  -4.236246 | -14.019087 |   -8.81131 |   -2.22971 |  -11.085591 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |      2.2 | 6.496878 | 6.496878 | -12.847953 | 0.923545 | 9.616872 |  9.616872 | -10.903839 | 0.928025 |  -249.08169 | 70.671588 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 16:00:00 |  0.1 |  1.0 |              0.0 |                 0.061898 |        0.028103 |             94.6 |       3.1 |              1019.0 | 0.0 | 0.0 |      -4.2 |             95.395833 |                   0.1 |                       7.41713 |            17.026026 | 0.094 |  1.0 |      -4.2 |    2.484933 |          1.86654 |            1.10047 |          3.1 |                 6.567919 |                      4.471043 |                    6.567919 |                     5.489944 |                      0.475001 |                            0.33814 |             6.213251 |                  4.265189 |    1012.786749 |   1.291856 | 8.583333 |  -2.98898 |  -9.215531 |              0.007076 |                 0.005754 |                  0.001314 |                   4.287089 |                 15.125699 |               10.405854 |                 2.180939 |   -15.118623 |        -10.4001 |        -2.179625 |        -10.838611 |             30.322581 |                 100.0 |               672.347081 |               73.469196 | 0.094 | 0.0235 |       49.0 |    0.439491 | 0.041312 | 0.010328 |  0.04203 |        0.748557 |      0.0 |        0.0 |      0.0 |   0.118269 | 0.030984 | 0.003168 |  0.016666 |      0.0 |  0.006769 | -0.006461 |  0.095635 |    -1.453058 |     0.748204 |      0.0 |      0.0 | 0.088281 | 0.039516 |  2.347518 |  -4.787513 | -14.580793 |  -9.225346 |  -1.833175 |  -11.833234 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |      2.2 | 6.535022 | 6.535022 |  -8.098112 | 0.925934 | 9.676021 |  9.676021 |  -4.524304 | 0.930102 | -231.282826 | 70.664819 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 17:00:00 |  0.0 |  1.1 |              0.0 |                      0.0 |             0.0 |             94.4 |       3.5 |              1018.0 | 0.0 | 0.0 | -4.091667 |             95.170833 |                   0.1 |                       7.41713 |            17.026026 |   0.0 |  1.1 | -4.091667 |    2.805569 |         1.866772 |           1.242466 |          3.5 |                 6.615571 |                      4.507808 |                    6.615571 |                     5.659133 |                      0.478041 |                             0.3406 |             6.245099 |                  4.290119 |    1011.754901 |     1.2901 | 8.583333 | -2.775394 | -10.404262 |                   0.0 |                      0.0 |                       0.0 |                   4.295147 |                 15.115697 |               11.875706 |                  3.57354 |   -15.115697 |      -11.875706 |         -3.57354 |        -10.820551 |             26.857143 |                 100.0 |               661.914462 |               69.444444 |   0.0 |    0.0 |       49.0 |    0.439605 |      0.0 |      0.0 |      0.0 |        0.748212 |      0.0 |        0.0 |      0.0 |   0.095572 |      0.0 | 0.003509 |  0.020103 |      0.0 |  0.007884 | -0.005551 |       0.0 |     -1.04135 |     0.747731 |      0.0 |      0.0 | 0.063051 |      0.0 |  2.600263 |  -4.112764 | -13.707117 |   -8.53959 |   0.768852 |   -9.078814 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |      2.2 | 6.531513 | 6.531513 |  -8.866965 | 0.930011 | 9.681572 |  9.681572 |  -5.849752 | 0.935716 | -212.295231 | 70.656935 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 18:00:00 |  0.3 |  1.0 |              0.0 |                      0.0 |             0.0 |             96.9 |       4.4 |              1016.0 | 0.0 | 0.0 |     -3.95 |             95.070833 |                   0.1 |                       7.41713 |            17.026026 | 0.282 |  1.0 |     -3.95 |    3.527001 |         1.942897 |           1.561958 |          4.4 |                 6.567919 |                      4.556289 |                    6.567919 |                     5.699177 |                      0.475001 |                           0.343839 |             6.364313 |                  4.331702 |    1009.635687 |   1.287972 | 8.583333 | -2.547543 |  -9.612254 |                   0.0 |                      0.0 |                       0.0 |                   4.293386 |                  15.08957 |               12.252381 |                 4.311207 |    -15.08957 |      -12.252381 |        -4.311207 |        -10.796184 |             21.363636 |                 100.0 |               667.340953 |               69.444444 | 0.282 | 0.0705 |       49.0 |    0.439595 | 0.123966 | 0.030991 | 0.126121 |        0.748117 |      0.0 |        0.0 |      0.0 |   0.090842 | 0.090842 | 0.002176 |  0.012597 |      0.0 |  0.004285 | -0.007108 |  0.286904 |      -0.9455 |     0.747834 |      0.0 |      0.0 | 0.058919 | 0.058919 |  1.612244 |  -5.266861 | -14.177315 |  -8.753064 |  -0.312689 |   -9.708718 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |      2.2 | 6.651171 | 6.653302 |        0.0 |  0.93113 | 9.787064 |  9.846714 |        0.0 | 0.934486 | -194.099643 | 70.652651 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 19:00:00 |  0.2 |  1.4 |              0.0 |                      0.0 |             0.0 |             98.3 |       3.8 |              1015.0 | 0.0 | 0.0 | -3.816667 |             95.029167 |                   0.1 |                       7.41713 |            17.026026 | 0.188 |  1.4 | -3.816667 |    3.046047 |         1.973983 |           1.348964 |          3.8 |                 6.760362 |                      4.602339 |                    6.760362 |                     5.914887 |                       0.48726 |                           0.346913 |             6.645436 |                  4.373564 |    1008.354564 |   1.284691 | 8.583333 | -2.329196 | -11.340094 |                   0.0 |                      0.0 |                       0.0 |                   4.300083 |                 15.061291 |               13.548131 |                 4.513202 |   -15.061291 |      -13.548131 |        -4.513202 |        -10.761208 |             24.736842 |                 100.0 |               622.876036 |               69.444444 | 0.188 | 0.0282 |       49.0 |     0.43996 | 0.082712 | 0.012407 | 0.124556 |        0.747598 |      0.0 |        0.0 |      0.0 |   0.044348 | 0.044348 | 0.001135 |  0.004093 |      0.0 |  0.001535 | -0.007216 |  0.283107 |     -0.43918 |     0.747441 |      0.0 |      0.0 | 0.027761 | 0.027761 |  0.841108 |  -5.346653 | -10.150145 |  -7.647939 |   4.239094 |    -8.48139 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |      2.2 | 6.706796 |  6.73488 |        0.0 |  0.93729 | 9.837746 |  9.959217 |        0.0 | 0.939171 | -174.176216 | 70.651115 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 20:00:00 |  0.4 |  1.6 |              0.0 |                      0.0 |             0.0 |             98.7 |       5.7 |              1014.0 | 0.0 | 0.0 |    -3.675 |             94.991667 |                   0.1 |                       7.41713 |            17.026026 | 0.376 |  1.6 |    -3.675 |     4.56907 |         2.056027 |           2.023445 |          5.7 |                 6.858436 |                      4.651718 |                    6.858436 |                     6.058417 |                      0.493491 |                           0.350204 |             6.769276 |                  4.418744 |    1007.230724 |   1.282429 | 8.583333 | -2.090115 |  -11.87322 |                   0.0 |                      0.0 |                       0.0 |                   4.305497 |                 15.030582 |               13.762728 |                 4.983727 |   -15.030582 |      -13.762728 |        -4.983727 |        -10.725085 |             16.491228 |                 100.0 |               603.231155 |               69.444444 | 0.376 | 0.0376 |       49.0 |    0.440205 | 0.165517 | 0.016552 | 0.292569 |        0.747115 |      0.0 |        0.0 |      0.0 |   0.028015 | 0.028015 | 0.001109 |  0.005159 |      0.0 |  0.001492 | -0.008844 |  0.664621 |    -0.111245 |     0.747123 |      0.0 |      0.0 | 0.007296 | 0.007296 |  0.821876 |  -6.552718 | -11.692875 |  -8.962665 |   2.891729 |  -10.531657 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |      2.2 | 6.750277 | 6.899287 |        0.0 | 0.943064 | 9.874671 | 10.178544 |        0.0 | 0.942974 | -153.719663 | 70.649624 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 21:00:00 |  0.7 |  2.0 |              0.0 |                      0.0 |             0.0 |             99.6 |       5.6 |              1013.0 | 0.0 | 0.0 |      -3.5 |             94.991667 |                   0.1 |                       7.41713 |            17.026026 | 0.658 |  2.0 |      -3.5 |    4.488911 |         2.138899 |           1.987946 |          5.6 |                 7.058357 |                      4.713362 |                    7.058357 |                       6.1078 |                      0.506157 |                           0.354306 |             7.030123 |                  4.477301 |    1005.969877 |   1.279174 | 8.583333 | -1.844636 | -11.067816 |                   0.0 |                      0.0 |                       0.0 |                    4.31592 |                  14.98921 |                12.71147 |                 3.399829 |    -14.98921 |       -12.71147 |        -3.399829 |        -10.673289 |             16.785714 |                 100.0 |               567.036492 |               69.444444 | 0.658 |    0.0 |       49.0 |    0.440698 | 0.289979 |      0.0 | 0.674492 |        0.746505 |      0.0 |        0.0 |      0.0 |   0.009603 | 0.009603 | 0.000347 | -0.001162 |      0.0 | -0.000343 |  -0.01135 |  1.530508 |          0.0 |      0.74651 | 0.062882 | 0.062882 |      0.0 |      0.0 |  0.257434 |  -8.409795 | -11.403474 | -10.361428 |    1.56543 |  -15.371395 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |      2.2 | 6.759554 | 7.188919 |        0.0 | 0.950409 | 9.822347 | 10.557915 |        0.0 | 0.950339 | -134.068514 | 70.649966 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 22:00:00 |  0.5 |  2.4 |              0.0 |                      0.0 |             0.0 |             99.4 |       5.8 |              1011.0 | 0.0 | 0.0 | -3.291667 |             95.008333 |                   0.1 |                       7.41713 |            17.026026 |  0.47 |  2.4 | -3.291667 |    4.649229 |         2.249283 |           2.058944 |          5.8 |                 7.263399 |                      4.787689 |                    7.263399 |                       6.1078 |                      0.519101 |                           0.359244 |             7.219819 |                  4.548704 |    1003.780181 |   1.274698 | 8.583333 | -1.608822 |  -9.652933 |                   0.0 |                      0.0 |                       0.0 |                   4.326251 |                 14.938353 |               11.165804 |                 1.378924 |   -14.938353 |      -11.165804 |        -1.378924 |        -10.612102 |             16.206897 |                 100.0 |               536.796025 |               69.444444 |  0.47 |    0.0 |       49.0 |    0.441567 | 0.207536 |      0.0 | 0.579276 |        0.745899 |      0.0 |   0.018471 | 0.018471 |        0.0 |      0.0 | 0.000548 |  0.000413 |      0.0 |   0.00012 | -0.013984 |  1.311864 |          0.0 |     0.745904 | 0.143872 | 0.143872 |      0.0 |      0.0 |  0.406082 | -10.361775 | -12.706346 | -12.706346 |  -1.134461 |  -21.689197 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |      2.2 | 6.740583 | 7.395907 |        0.0 | 0.957747 | 9.690984 | 10.834363 |        0.0 | 0.957693 | -115.832247 | 70.649846 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 23:00:00 |  1.0 |  2.5 |              0.0 |                      0.0 |             0.0 |             99.2 |       6.2 |              1010.0 | 0.0 | 0.0 | -3.070833 |                  95.0 |                   0.1 |                       7.41713 |            17.026026 |  0.94 |  2.5 | -3.070833 |    4.969866 |         2.364694 |           2.200941 |          6.2 |                 7.315473 |                      4.867606 |                    7.315473 |                       6.1078 |                      0.522381 |                           0.364543 |             7.256949 |                  4.624226 |    1002.743051 |   1.272954 | 8.583333 | -1.389987 |  -8.339922 |                   0.0 |                      0.0 |                       0.0 |                   4.336562 |                 14.885653 |               10.640998 |                 0.884234 |   -14.885653 |      -10.640998 |        -0.884234 |        -10.549091 |              15.16129 |                 100.0 |               530.036444 |               69.444444 |  0.94 |    0.0 |       49.0 |    0.442188 | 0.415656 |      0.0 | 1.208521 |        0.745296 |      0.0 |   0.040986 | 0.040986 |        0.0 |      0.0 | 0.000768 |  0.002303 |      0.0 |  0.000639 | -0.015071 |   2.73305 |          0.0 |     0.745298 | 0.199185 | 0.199185 |      0.0 |      0.0 |  0.568725 | -11.167257 | -13.803762 | -13.803762 |  -2.594038 |  -24.086786 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |      2.2 | 6.698939 | 7.810796 |        0.0 | 0.965112 | 9.504393 | 11.373778 |        0.0 | 0.965089 |  -98.908992 | 70.649208 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-11 00:00:00 |  2.0 |  2.7 |              0.0 |                      0.0 |             0.0 |             99.2 |       5.6 |              1008.0 | 0.0 | 0.0 | -2.804167 |               95.4875 |                   0.1 |                       7.41713 |            17.026026 |  1.88 |  2.7 | -2.804167 |    4.488911 |         2.438173 |           1.987946 |          5.6 |                 7.420609 |                      4.965681 |                    7.420609 |                       6.1078 |                      0.528994 |                           0.371032 |             7.361244 |                  4.741604 |    1000.638756 |   1.269455 | 8.583333 | -1.186908 |  -7.121447 |                   0.0 |                      0.0 |                       0.0 |                   4.346899 |                 14.787036 |                9.446615 |                -0.139391 |   -14.787036 |       -9.446615 |         0.139391 |        -10.440138 |             16.785714 |                 100.0 |               516.295234 |               69.444444 |  1.88 |    0.0 |       49.0 |    0.443432 | 0.833653 |      0.0 | 2.617753 |        0.744694 |      0.0 |   0.071329 | 0.071329 |        0.0 |      0.0 | 0.000731 |  0.001792 |      0.0 |  0.000535 | -0.015425 |  5.903388 |          0.0 |     0.744691 | 0.262328 | 0.262328 |      0.0 |      0.0 |  0.541292 | -11.428977 | -13.987928 | -13.987928 |  -4.000021 |  -25.556296 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |      2.2 | 6.627049 | 8.643719 |        0.0 | 0.972524 | 9.253543 | 12.435549 |        0.0 |  0.97256 |  -83.204212 | 70.648673 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-11 01:00:00 |  2.1 |  3.2 |              0.0 |                      0.0 |             0.0 |             99.6 |       5.2 |              1007.0 | 0.0 | 0.0 | -2.470833 |             95.858333 |                   0.1 |                       7.41713 |            17.026026 | 1.974 |  3.2 | -2.470833 |    4.168274 |         2.514993 |            1.84595 |          5.2 |                 7.689299 |                      5.090729 |                    7.689299 |                       6.1078 |                      0.545841 |                           0.379281 |             7.658542 |                  4.879888 |     999.341458 |   1.265756 | 8.583333 | -0.998451 |  -5.990703 |                   0.0 |                      0.0 |                       0.0 |                   4.357114 |                 14.678585 |                6.297157 |                -2.742905 |   -14.678585 |       -6.297157 |         2.742905 |        -10.321471 |             18.076923 |                 100.0 |               484.122576 |               69.444444 | 1.974 |    0.0 |       49.0 |    0.445931 | 0.880268 |      0.0 |  3.09995 |        0.744096 | 0.238205 |   0.133501 | 0.133501 |        0.0 |      0.0 | 0.000362 | -0.001203 |      0.0 | -0.000381 | -0.018246 |  5.355679 |          0.0 |     0.744091 | 0.339293 | 0.339293 |      0.0 |      0.0 |  0.268147 | -13.519673 | -15.851265 | -15.851265 |   -9.28596 |  -32.113842 |      1.0 |      0.3 | 1.020778 | 0.348096 |  0.0 |  0.0 | 0.0 |  0.0 | 0.348096 | 0.004515 | 0.001254 |      2.2 |   6.4933 |  9.28542 |        0.0 | 0.979938 | 8.926542 | 12.764955 |        0.0 |     0.98 |  -68.630175 | 71.321736 | 0.348096 |      0.0 |   0.0 |   0.0 |  0.0 | 0.004515 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 | 0.001254 |
    | 2010-12-11 02:00:00 |  1.8 |  3.9 |              0.0 |                      0.0 |             0.0 |             99.8 |       5.7 |              1006.0 | 0.0 | 0.0 |    -2.075 |             96.279167 |                   0.1 |                       7.41713 |            17.026026 | 1.692 |  3.9 |    -2.075 |     4.56907 |         2.621871 |           2.023445 |          5.7 |                  8.07986 |                      5.242833 |                     8.07986 |                       6.1078 |                      0.570198 |                           0.389283 |               8.0637 |                  5.047756 |       997.9363 |   1.261108 | 8.583333 | -0.823562 |  -4.941373 |                   0.0 |                      0.0 |                       0.0 |                   4.366615 |                 14.548979 |                1.967269 |                -6.386972 |   -14.548979 |       -1.967269 |         6.386972 |        -10.182364 |             16.491228 |                 100.0 |               441.910645 |               69.444444 | 1.692 |    0.0 |       49.0 |    0.447856 | 0.757773 |      0.0 |      0.0 |        0.743511 | 1.039983 |   0.197354 | 0.197354 |        0.0 |      0.0 | 0.000201 | -0.002412 |      0.0 | -0.000728 | -0.024333 | -1.280021 |          0.0 |     0.743533 | 0.416279 | 0.416279 |      0.0 |      0.0 |  0.148961 | -18.029533 | -20.426299 | -20.426299 | -18.310069 |  -44.842804 |      1.0 |      0.3 | 2.570623 | 0.878919 |  0.0 |  0.0 | 0.0 |  0.0 | 0.878919 | 0.034519 | 0.009589 |      2.2 |  6.29581 | 9.003008 |        0.0 | 0.987246 | 8.526486 | 12.192874 |        0.0 | 0.986969 |  -55.105469 | 73.014169 |      0.5 | 0.378919 |   0.0 |   0.0 |  0.0 | 0.015361 | 0.019159 |   0.0 |   0.0 |  0.0 |   0.0 | 0.009589 |
    | 2010-12-11 03:00:00 |  1.3 |  4.4 |              0.0 |                      0.0 |             0.0 |             99.8 |       6.6 |              1005.0 | 0.0 | 0.0 | -1.633333 |             96.633333 |                   0.1 |                       7.41713 |            17.026026 | 1.222 |  4.4 | -1.633333 |    5.290502 |          2.74211 |           2.342937 |          6.6 |                  8.36943 |                      5.417279 |                     8.36943 |                       6.1078 |                       0.58816 |                           0.400709 |             8.352691 |                  5.234898 |     996.647309 |   1.257444 | 8.583333 | -0.661266 |  -3.967594 |                   0.0 |                      0.0 |                       0.0 |                   4.375755 |                 14.409644 |               -1.107912 |                -8.998234 |   -14.409644 |        1.107912 |         8.998234 |        -10.033889 |             14.242424 |                 100.0 |               409.411198 |               69.444444 | 1.222 |    0.0 |       49.0 |    0.447009 | 0.546245 |      0.0 |      0.0 |         0.74293 | 0.950566 |   0.282749 | 0.282749 |        0.0 |      0.0 | 0.000229 | -0.002134 |      0.0 | -0.000588 | -0.030654 | -2.068991 |          0.0 |     0.742996 | 0.549367 | 0.549367 |      0.0 |      0.0 |  0.169359 | -22.713156 | -25.294274 | -25.294274 | -26.232827 |  -57.005664 |      1.0 |      0.3 | 2.413891 | 0.827844 |  0.0 |  0.0 | 0.0 |  0.0 | 0.827844 | 0.080212 | 0.022281 |      2.2 | 6.012908 | 8.598459 |        0.0 | 0.994554 | 7.997173 | 11.435958 |        0.0 |  0.99372 |  -42.554542 | 74.600804 |      0.5 | 0.327844 |   0.0 |   0.0 |  0.0 |  0.02788 | 0.052332 |   0.0 |   0.0 |  0.0 |   0.0 | 0.022281 |
    | 2010-12-11 04:00:00 |  0.5 |  4.8 |              0.0 |                      0.0 |             0.0 |             99.0 |       7.0 |              1005.0 | 0.0 | 0.0 | -1.129167 |             96.833333 |                   0.1 |                       7.41713 |            17.026026 |  0.47 |  4.8 | -1.129167 |    5.611139 |         2.892408 |           2.484933 |          7.0 |                 8.607627 |                      5.622664 |                    8.607627 |                       6.1078 |                      0.602876 |                           0.414103 |             8.521551 |                  5.444613 |     996.478449 |   1.255555 | 8.583333 | -0.510655 |  -3.063927 |                   0.0 |                      0.0 |                       0.0 |                   4.384576 |                 14.261536 |               -3.288969 |               -11.031041 |   -14.261536 |        3.288969 |        11.031041 |          -9.87696 |             13.428571 |                 100.0 |               386.015617 |               69.444444 |  0.47 |    0.0 |       49.0 |    0.445795 | 0.209524 |      0.0 |      0.0 |        0.742354 | 0.688311 |   0.334864 | 0.334864 |        0.0 |      0.0 | 0.001222 |  0.005738 |      0.0 |  0.001528 | -0.034262 | -2.672783 |          0.0 |     0.742478 | 0.639867 | 0.639867 |      0.0 |      0.0 |  0.905313 | -25.386793 | -28.684284 | -28.684284 |  -31.06794 |  -65.102118 |      1.0 |      0.3 | 1.866538 | 0.641733 |  0.0 |  0.0 | 0.0 |  0.0 | 0.641733 |  0.11038 | 0.030661 |      2.2 | 5.677238 |  8.11845 |        0.0 | 1.001862 | 7.379349 | 10.552469 |        0.0 |  1.00028 |  -30.907282 | 75.824082 |      0.5 | 0.141733 |   0.0 |   0.0 |  0.0 | 0.040075 | 0.070305 |   0.0 |   0.0 |  0.0 |   0.0 | 0.030661 |
    | 2010-12-11 05:00:00 |  0.2 |  4.8 |              0.0 |                      0.0 |             0.0 |             97.6 |       7.3 |              1004.0 | 0.0 | 0.0 | -0.583333 |             97.016667 |                   0.1 |                       7.41713 |            17.026026 | 0.188 |  4.8 | -0.583333 |    5.851616 |         3.059407 |            2.59143 |          7.3 |                 8.607627 |                      5.852751 |                    8.607627 |                       6.1078 |                      0.602876 |                           0.429037 |             8.401044 |                  5.678144 |     995.598956 |   1.254358 | 8.583333 | -0.370887 |  -2.225324 |                   0.0 |                      0.0 |                       0.0 |                   4.393298 |                 14.099661 |               -2.759879 |               -10.910371 |   -14.099661 |        2.759879 |        10.910371 |         -9.706363 |             12.876712 |                 100.0 |                382.16566 |               69.444444 | 0.188 |    0.0 |       49.0 |    0.444355 | 0.083539 |      0.0 |      0.0 |        0.741781 | 0.546184 |   0.323648 | 0.323648 |        0.0 |      0.0 | 0.003016 |  0.020253 |      0.0 |  0.005233 |  -0.03348 | -2.582669 |          0.0 |     0.741966 | 0.651144 | 0.651144 |      0.0 |      0.0 |  2.234705 | -24.807089 | -29.502182 | -29.502182 | -30.027356 |  -65.219643 |      1.0 |      0.3 | 1.584734 | 0.545969 |  0.0 |  0.0 | 0.0 |  0.0 | 0.545969 | 0.124517 | 0.034588 |      2.2 | 5.351601 | 7.652789 |        0.0 |  1.00917 | 6.749552 |  9.651859 |        0.0 | 1.006807 |  -20.098624 | 76.857614 |      0.5 | 0.045969 |   0.0 |   0.0 |  0.0 | 0.051955 | 0.072562 |   0.0 |   0.0 |  0.0 |   0.0 | 0.034588 |
    | 2010-12-11 06:00:00 |  0.5 |  5.1 |              0.0 |                      0.0 |             0.0 |             97.4 |       7.2 |              1004.0 | 0.0 | 0.0 |   -0.0125 |             97.054167 |                   0.1 |                       7.41713 |            17.026026 |  0.47 |  5.1 |   -0.0125 |    5.771457 |         3.229745 |           2.555931 |          7.2 |                 8.790172 |                      6.102233 |                    8.790172 |                       6.1078 |                      0.614118 |                           0.445147 |             8.561627 |                  5.922472 |     995.438373 |    1.25293 | 8.583333 | -0.241183 |  -1.447101 |                   0.0 |                      0.0 |                       0.0 |                   4.401705 |                 13.937223 |               -4.549601 |               -12.474442 |   -13.937223 |        4.549601 |        12.474442 |         -9.535518 |             13.055556 |                 100.0 |               366.166145 |               69.444444 |  0.47 |    0.0 |       49.0 |    0.442958 |  0.20819 |      0.0 |      0.0 |        0.741213 | 0.719018 |   0.357376 | 0.357376 |        0.0 |      0.0 | 0.003306 |  0.022274 |      0.0 |   0.00585 | -0.035493 |  -3.02987 |          0.0 |     0.741472 | 0.704401 | 0.704401 |      0.0 |      0.0 |  2.449426 | -26.298897 | -31.056396 | -31.056396 | -33.156572 |  -69.829736 |      1.0 |      0.3 | 1.991825 | 0.687807 |  0.0 |  0.0 | 0.0 |  0.0 | 0.687807 | 0.140619 | 0.039061 |      2.2 | 4.992067 | 7.138656 |        0.0 | 1.016478 | 6.067381 |  8.676355 |        0.0 |  1.01314 |   -10.06819 | 78.155782 |      0.5 | 0.187807 |   0.0 |   0.0 |  0.0 | 0.063528 | 0.077091 |   0.0 |   0.0 |  0.0 |   0.0 | 0.039061 |
    | 2010-12-11 07:00:00 |  0.2 |  5.1 |              0.0 |                      0.0 |             0.0 |             96.3 |       7.6 |              1004.0 | 0.0 | 0.0 |  0.554167 |             97.070833 |                   0.1 |                       7.41713 |            17.026026 | 0.188 |  5.1 |  0.554167 |    6.092093 |         3.436822 |           2.697927 |          7.6 |                 8.790172 |                      6.359135 |                    8.790172 |                       6.1078 |                      0.614118 |                            0.46165 |             8.464935 |                  6.172866 |     995.535065 |   1.252976 | 8.583333 | -0.120818 |   -0.72491 |                   0.0 |                      0.0 |                       0.0 |                   4.409951 |                 13.773355 |               -4.125536 |               -12.377725 |   -13.773355 |        4.125536 |        12.377725 |         -9.363404 |             12.368421 |                 100.0 |               361.614133 |               69.444444 | 0.188 |    0.0 |       49.0 |    0.441416 | 0.082986 |      0.0 |      0.0 |        0.740649 | 0.587035 |   0.352723 | 0.352723 |        0.0 |      0.0 |  0.00488 |   0.03538 |      0.0 |  0.008934 | -0.035367 | -2.989666 |          0.0 |     0.740988 | 0.723054 | 0.723054 |      0.0 |      0.0 |  3.615771 | -26.205129 | -32.215085 | -32.215085 |  -32.72485 |  -70.797939 |      1.0 |      0.3 | 1.730231 | 0.598806 |  0.0 |  0.0 | 0.0 |  0.0 | 0.598806 | 0.158397 | 0.043999 |      2.2 | 4.636173 | 6.629727 |        0.0 | 1.023786 | 5.366111 |  7.673539 |        0.0 | 1.019391 |   -0.759947 | 79.278274 |      0.5 | 0.098806 |   0.0 |   0.0 |  0.0 | 0.074803 | 0.083594 |   0.0 |   0.0 |  0.0 |   0.0 | 0.043999 |
    | 2010-12-11 08:00:00 |  0.1 |  5.4 |              0.0 |                 0.347041 |        0.900733 |             95.2 |       7.3 |              1003.0 | 0.0 | 0.0 |     1.175 |             97.004167 |                   0.1 |                       7.40894 |            17.017406 | 0.094 |  5.4 |     1.175 |    5.851616 |           3.6439 |            2.59143 |          7.3 |                 8.976116 |                       6.65151 |                    8.976116 |                       6.1078 |                       0.62554 |                           0.480332 |             8.545263 |                  6.452241 |     994.454737 |   1.250338 | 8.583333 | -0.009119 |  -0.054716 |              0.233719 |                 0.190633 |                  0.043405 |                   4.415614 |                  13.59621 |               -5.565868 |               -13.865856 |   -13.362491 |        5.756501 |        13.909261 |         -9.180596 |             12.876712 |                 100.0 |               347.745306 |               95.698636 | 0.094 |    0.0 |       49.0 |    0.439889 |  0.04135 |      0.0 |      0.0 |        0.740089 |   0.5693 |   0.369547 | 0.369547 |        0.0 |      0.0 |  0.00629 |  0.045764 |      0.0 |  0.009363 | -0.035585 | -3.315633 |          0.0 |     0.740524 | 0.755526 | 0.755526 |      0.0 |      0.0 |  4.660744 |  -26.36717 | -33.189955 | -33.189955 | -34.285712 |  -73.466385 | 0.998939 | 0.299363 | 1.707363 | 0.591124 |  0.0 |  0.0 | 0.0 |  0.0 | 0.591124 | 0.170504 | 0.047362 |      2.2 | 4.262578 | 6.095487 |        0.0 | 1.031094 | 4.631966 |  6.623712 |        0.0 | 1.025415 |    7.878103 |  80.38515 |      0.5 | 0.091124 |   0.0 |   0.0 |  0.0 | 0.085786 | 0.084718 |   0.0 |   0.0 |  0.0 |   0.0 | 0.047362 |
    | 2010-12-11 09:00:00 |  0.3 |  5.3 |              0.0 |                      1.0 |       17.267511 |             96.3 |       7.1 |              1003.0 | 0.0 | 0.0 |  1.779167 |               97.0875 |                   0.1 |                       7.40894 |            16.959252 | 0.282 |  5.3 |  1.779167 |    5.691298 |         3.847638 |           2.520432 |          7.1 |                 8.913754 |                      6.947358 |                    8.913754 |                       6.1078 |                      0.621713 |                            0.49913 |             8.583945 |                  6.745016 |     994.416055 |   1.250768 | 8.583333 |       0.0 |        0.0 |              4.488142 |                 3.662347 |                  0.833512 |                    4.40802 |                 13.407231 |               -5.371692 |               -13.434609 |    -8.919089 |         9.03404 |        14.268121 |         -8.999211 |             13.239437 |                 100.0 |                344.06059 |              328.221051 | 0.282 |    0.0 |       49.0 |    0.438286 | 0.123597 |      0.0 |      0.0 |        0.739533 | 0.701228 |   0.404251 | 0.404251 |        0.0 |      0.0 | 0.004726 |  0.036459 |      0.0 |  0.002601 | -0.035482 | -3.560461 |          0.0 |     0.740082 | 0.743405 | 0.743405 |      0.0 |      0.0 |  3.501758 | -26.290556 | -31.973262 | -31.973262 | -37.505545 |  -72.531939 | 0.997782 | 0.298671 | 1.928395 | 0.667945 |  0.0 |  0.0 | 0.0 |  0.0 | 0.667945 | 0.185724 |  0.05159 |      2.2 | 3.855335 | 5.513129 |        0.0 | 1.038402 | 3.909391 |   5.59043 |        0.0 | 1.031187 |   16.461436 |    81.643 |      0.5 | 0.167945 |   0.0 |   0.0 |  0.0 | 0.096486 | 0.089239 |   0.0 |   0.0 |  0.0 |   0.0 |  0.05159 |
    | 2010-12-11 10:00:00 |  0.1 |  5.5 |              0.0 |                      1.0 |       52.617611 |             95.5 |       7.1 |              1003.0 | 0.0 | 0.0 |  2.295833 |               97.0375 |                   0.1 |                       7.40894 |            16.807156 | 0.094 |  5.5 |  2.295833 |    5.691298 |         4.038016 |           2.520432 |          7.1 |                 9.038862 |                      7.209503 |                    9.038862 |                       6.1078 |                      0.629387 |                           0.515703 |             8.632113 |                  6.995922 |     994.367887 |   1.249848 | 8.583333 |       0.0 |        0.0 |             13.697099 |                11.183569 |                  2.543747 |                   4.375137 |                 13.251022 |               -6.309796 |               -14.422901 |     0.446076 |       17.493365 |        16.966648 |         -8.875885 |             13.239437 |                 100.0 |               332.977647 |              318.313792 | 0.094 |    0.0 |       49.0 |    0.436539 | 0.041035 |      0.0 |      0.0 |        0.738981 | 0.754749 |   0.499629 | 0.499629 |        0.0 |      0.0 | 0.005828 |  0.051269 |      0.0 |  0.003784 | -0.036172 | -4.565275 |          0.0 |     0.739686 | 0.780178 | 0.780178 |      0.0 |      0.0 |   4.31867 |  -26.80199 | -33.179801 | -33.179801 | -46.354496 |  -76.948439 | 0.996626 | 0.297979 | 1.930587 | 0.669072 |  0.0 |  0.0 | 0.0 |  0.0 | 0.669072 | 0.204027 | 0.056674 |      2.2 | 3.352158 | 4.793586 |        0.0 |  1.04571 | 3.149461 |  4.503729 |        0.0 | 1.036383 |   25.044769 | 82.900731 |      0.5 | 0.169072 |   0.0 |   0.0 |  0.0 | 0.106909 | 0.097118 |   0.0 |   0.0 |  0.0 |   0.0 | 0.056674 |
    | 2010-12-11 11:00:00 |  0.1 |  6.1 |              0.0 |                      1.0 |       80.135633 |             91.6 |       8.9 |              1002.0 | 0.0 | 0.0 |  2.716667 |             96.795833 |                   0.0 |                       7.40894 |            15.680798 | 0.094 |  6.1 |  2.716667 |    7.134162 |         4.291853 |           3.159415 |          8.9 |                 9.423511 |                       7.42943 |                    9.423511 |                       6.1078 |                      0.652898 |                           0.529548 |             8.631936 |                  7.191379 |     993.368064 |   1.245915 | 8.583333 |       0.0 |        0.0 |             20.886251 |                17.068132 |                  3.878875 |                   4.086984 |                 12.463959 |                -9.79282 |               -17.550317 |     8.422293 |       26.860952 |        21.429192 |         -8.376974 |             10.561798 |                 100.0 |               315.668522 |              302.757514 | 0.094 |    0.0 |       49.0 |    0.434381 | 0.040832 |      0.0 |      0.0 |        0.738433 | 0.964107 |   0.647439 | 0.647439 |        0.0 |      0.0 | 0.013265 |  0.119103 |      0.0 |  0.007618 | -0.042299 | -6.549992 |          0.0 |     0.739364 | 0.962052 | 0.962052 |      0.0 |      0.0 |  9.828916 | -31.341975 | -43.035887 | -43.035887 | -60.067923 |  -95.807053 | 0.995469 | 0.297288 | 2.405931 | 0.834637 |  0.0 |  0.0 | 0.0 |  0.0 | 0.834637 | 0.229703 | 0.063806 |      2.2 | 2.697235 | 3.857046 |        0.0 | 1.053018 | 2.207953 |  3.157372 |        0.0 | 1.040635 |   33.628103 | 84.464406 |      0.5 | 0.334637 |   0.0 |   0.0 |  0.0 | 0.117063 | 0.112641 |   0.0 |   0.0 |  0.0 |   0.0 | 0.063806 |
    | 2010-12-11 12:00:00 |  0.0 |  6.4 |              0.0 |                      1.0 |       80.276117 |             90.0 |       9.5 |              1002.0 | 0.0 | 0.0 |       3.1 |             96.633333 |                   0.0 |                       7.40894 |            15.214247 |   0.0 |  6.4 |       3.1 |    7.615117 |          4.56573 |           3.372409 |          9.5 |                 9.621181 |                      7.634886 |                    9.621181 |                       6.1078 |                      0.664931 |                           0.542435 |             8.659063 |                  7.377844 |     993.340937 |   1.244565 | 8.583333 |       0.0 |        0.0 |             20.944144 |                17.133628 |                  3.889627 |                   3.969417 |                 12.359437 |              -11.020237 |               -19.002428 |     8.584707 |       28.153865 |        22.892055 |          -8.39002 |              9.894737 |                 100.0 |               303.286102 |              291.566218 |   0.0 |    0.0 |       49.0 |    0.431571 |      0.0 |      0.0 |      0.0 |        0.737888 | 0.966218 |   0.678651 | 0.678651 |        0.0 |      0.0 | 0.016902 |  0.149606 |      0.0 |  0.009414 | -0.044819 | -7.191755 |          0.0 |     0.739099 | 1.037348 | 1.037348 |      0.0 |      0.0 | 12.523606 |  -33.20903 | -47.333467 | -47.333467 | -62.963727 | -103.434552 | 0.994313 | 0.296597 | 2.470682 | 0.858159 |  0.0 |  0.0 | 0.0 |  0.0 | 0.858159 | 0.262848 | 0.073013 |      2.2 | 2.009738 | 2.873926 |        0.0 | 1.060326 | 1.187222 |  1.697727 |        0.0 | 1.044145 |   42.211436 | 86.067515 |      0.5 | 0.358159 |   0.0 |   0.0 |  0.0 | 0.126954 | 0.135894 |   0.0 |   0.0 |  0.0 |   0.0 | 0.073013 |
    | 2010-12-11 13:00:00 |  0.1 |  6.4 |              0.0 |                      1.0 |       75.835403 |             85.7 |       7.9 |              1002.0 | 0.0 | 0.0 |     3.425 |               96.3125 |                   0.0 |                       7.40894 |            15.020908 | 0.094 |  6.4 |     3.425 |    6.332571 |         4.782828 |           2.804424 |          7.9 |                 9.621181 |                      7.812981 |                    9.621181 |                       6.1078 |                      0.664931 |                           0.553571 |             8.245352 |                  7.524878 |     993.754648 |    1.24476 | 8.583333 |       0.0 |        0.0 |             19.802182 |                  16.2192 |                  3.677548 |                   3.922268 |                 12.283332 |               -9.152157 |               -18.576375 |     7.518849 |       25.371357 |        22.253923 |         -8.361065 |             11.898734 |                 100.0 |               302.906915 |              291.222675 | 0.094 |    0.0 |       49.0 |    0.428622 |  0.04029 |      0.0 |      0.0 |        0.737348 | 0.823325 |   0.551646 | 0.551646 |        0.0 |      0.0 |   0.0212 |  0.175353 |      0.0 |  0.013073 | -0.032937 | -5.828281 |          0.0 |     0.738879 | 0.887582 | 0.887582 |      0.0 |      0.0 | 15.708194 | -24.404983 | -41.517304 | -41.517304 | -51.180467 |   -88.17621 | 0.993156 | 0.295908 | 2.170901 | 0.754793 |  0.0 |  0.0 | 0.0 |  0.0 | 0.754793 | 0.289349 | 0.080375 |      2.2 | 1.447337 | 2.069692 |        0.0 | 1.067634 | 0.305453 |  0.436797 |        0.0 | 1.047057 |   50.794769 |  87.47055 |      0.5 | 0.254793 |   0.0 |   0.0 |  0.0 |  0.13659 | 0.152758 |   0.0 |   0.0 |  0.0 |   0.0 | 0.080375 |
    | 2010-12-11 14:00:00 |  0.0 |  6.5 |              0.0 |                      1.0 |       42.227911 |             82.3 |       9.6 |              1002.0 | 0.0 | 0.0 |    3.7125 |             95.854167 |                   0.0 |                       7.40894 |            14.983713 |   0.0 |  6.5 |    3.7125 |    7.695276 |         5.026645 |           3.407908 |          9.6 |                 9.687876 |                      7.973569 |                    9.687876 |                       6.1078 |                      0.668985 |                           0.563584 |             7.973122 |                  7.642998 |     994.026878 |   1.244443 | 8.583333 |       0.0 |        0.0 |             11.030848 |                 9.049891 |                  2.048586 |                   3.914072 |                 12.228543 |               -8.244287 |               -18.760881 |    -1.197695 |       17.294177 |        20.809467 |         -8.314471 |              9.791667 |                 100.0 |               299.581924 |              288.208117 |   0.0 |    0.0 |       49.0 |    0.426209 |      0.0 |      0.0 |      0.0 |        0.736811 |  0.65676 |   0.466109 | 0.466109 |        0.0 |      0.0 | 0.030355 |  0.251636 |      0.0 |  0.015898 | -0.033021 | -4.964778 |          0.0 |     0.738778 | 0.956627 | 0.305453 |      0.0 |      0.0 | 22.491862 | -24.466815 | -48.442242 | -48.442242 | -43.244558 |  -93.718524 |    0.992 | 0.295219 | 1.126578 | 0.391681 |  0.0 |  0.0 | 0.0 |  0.0 | 0.391681 | 0.294589 |  0.08183 |      2.2 | 0.966837 | 1.382576 |        0.0 | 1.074942 |      0.0 |       0.0 |        0.0 | 1.048404 |   59.378103 |  88.18955 | 0.391681 |      0.0 |   0.0 |   0.0 |  0.0 | 0.144572 | 0.150017 |   0.0 |   0.0 |  0.0 |   0.0 |  0.08183 |
    | 2010-12-11 15:00:00 |  0.1 |  6.3 |              0.1 |                      1.0 |       12.613919 |             82.5 |       9.1 |              1003.0 | 0.0 | 0.0 |  3.954167 |             95.370833 |                   0.1 |                       7.40894 |            15.079289 | 0.094 |  6.3 |  3.954167 |     7.29448 |         5.237063 |           3.230413 |          9.1 |                 9.554889 |                      8.110798 |                    9.554889 |                          0.0 |                      0.660899 |                           0.572121 |             7.882784 |                  7.735336 |     995.117216 |   1.246623 | 8.583333 |       0.0 |      -37.8 |             11.100249 |                  2.70877 |                  2.061475 |                  13.269775 |                 12.847805 |                -5.80609 |                      0.0 |    -1.747556 |         8.51486 |              0.0 |          0.421969 |              10.32967 |                 100.0 |               300.930422 |              289.431174 | 0.094 |    0.0 |       49.0 |    0.424148 |  0.03987 |      0.0 |      0.0 |        0.736279 | 0.531701 |    0.35117 |  0.35117 |        0.0 |      0.0 | 0.028472 |  0.234548 |      0.0 |  0.015417 |       0.0 | -3.603596 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 | 21.096693 |        0.0 |  -45.16256 |        0.0 | -32.580727 |         0.0 | 0.985751 | 0.291511 | 0.585831 | 0.201855 |  0.0 |  0.0 | 0.0 |  0.0 | 0.201855 |   0.2836 | 0.078778 |      2.2 | 0.602988 | 0.862273 |        0.0 |  1.08225 |      0.0 |       0.0 |        0.0 |      nan |  105.761436 | 88.558108 | 0.201855 |      0.0 |   0.0 |   0.0 |  0.0 | 0.148493 | 0.135107 |   0.0 |   0.0 |  0.0 |   0.0 | 0.078778 |
    | 2010-12-11 16:00:00 |  0.0 |  5.9 |              0.0 |                 0.053708 |        0.021911 |             82.2 |       8.7 |              1003.0 | 0.0 | 0.0 |  4.158333 |             94.854167 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  5.9 |  4.158333 |    6.973844 |         5.424101 |           3.088417 |          8.7 |                 9.293725 |                       8.22835 |                    9.293725 |                          0.0 |                      0.644979 |                           0.579419 |             7.639442 |                  7.804933 |     995.360558 |   1.248525 | 8.583333 |       0.0 |      -35.4 |              0.019282 |                 0.004715 |                  0.003581 |                  13.269547 |                 12.822679 |               -3.188375 |                      0.0 |   -12.803397 |         3.19309 |              0.0 |          0.446868 |             10.804598 |                 100.0 |               309.951162 |               72.426619 |   0.0 |    0.0 |       49.0 |    0.422587 |      0.0 |      0.0 |      0.0 |         0.73575 |  0.35728 |   0.258008 | 0.258008 |        0.0 |      0.0 | 0.027276 |  0.217294 |      0.0 |  0.048899 |       0.0 | -2.451551 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 | 20.210344 |        0.0 | -40.954652 |        0.0 | -23.937398 |         0.0 | 0.979825 | 0.288017 |  0.35728 | 0.122015 |  0.0 |  0.0 | 0.0 |  0.0 | 0.122015 | 0.270515 | 0.075143 |      2.2 | 0.334068 | 0.477717 |        0.0 | 1.089558 |      0.0 |       0.0 |        0.0 |      nan |  149.744769 | 88.744474 | 0.122015 |      0.0 |   0.0 |   0.0 |  0.0 | 0.148836 | 0.121679 |   0.0 |   0.0 |  0.0 |   0.0 | 0.075143 |
    | 2010-12-11 17:00:00 |  0.0 |  5.5 |              0.0 |                      0.0 |             0.0 |             76.5 |       8.9 |              1003.0 | 0.0 | 0.0 |  4.341667 |             94.108333 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  5.5 |  4.341667 |    7.134162 |         5.604459 |           3.159415 |          8.9 |                 9.038862 |                      8.335182 |                    9.038862 |                          0.0 |                      0.629387 |                            0.58604 |              6.91473 |                  7.844101 |      996.08527 |   1.250659 | 8.583333 |       0.0 |      -33.0 |                   0.0 |                      0.0 |                       0.0 |                  13.269547 |                   12.8228 |                1.905808 |                      0.0 |     -12.8228 |       -1.905808 |              0.0 |          0.446747 |             10.561798 |                 100.0 |               329.290397 |               69.444444 |   0.0 |    0.0 |       49.0 |    0.421433 |      0.0 |      0.0 |      0.0 |        0.735224 | 0.150091 |   0.113409 | 0.113409 |        0.0 |      0.0 | 0.035596 |  0.290175 |      0.0 |  0.065707 |       0.0 | -0.960059 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 | 26.375167 |        0.0 | -38.802849 |        0.0 | -10.521874 |         0.0 | 0.974222 | 0.284733 | 0.150091 | 0.050811 |  0.0 |  0.0 | 0.0 |  0.0 | 0.050811 | 0.256805 | 0.071335 |      2.2 | 0.204217 |  0.29203 |        0.0 | 1.096866 |      0.0 |       0.0 |        0.0 |      nan |  191.328103 | 88.778047 | 0.050811 |      0.0 |   0.0 |   0.0 |  0.0 | 0.147219 | 0.109585 |   0.0 |   0.0 |  0.0 |   0.0 | 0.071335 |
    | 2010-12-11 18:00:00 |  0.0 |  5.1 |              0.0 |                      0.0 |             0.0 |             81.7 |       7.9 |              1003.0 | 0.0 | 0.0 |    4.5125 |                93.475 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  5.1 |    4.5125 |    6.332571 |         5.721357 |           2.804424 |          7.9 |                 8.790172 |                      8.435829 |                    8.790172 |                          0.0 |                      0.614118 |                           0.592267 |              7.18157 |                  7.885391 |      995.81843 |   1.252331 | 8.583333 |       0.0 |      -30.6 |                   0.0 |                      0.0 |                       0.0 |                  13.269547 |                 12.818218 |                1.939405 |                      0.0 |   -12.818218 |       -1.939405 |              0.0 |          0.451329 |             11.898734 |                 100.0 |               332.567248 |               69.444444 |   0.0 |    0.0 |       49.0 |    0.420876 |      0.0 |      0.0 |      0.0 |        0.734702 | 0.180245 |   0.137736 | 0.137736 |        0.0 |      0.0 | 0.024787 |  0.197149 |      0.0 |  0.048384 |       0.0 |  -1.06909 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 | 18.365818 |        0.0 | -33.084102 |        0.0 | -12.778878 |         0.0 | 0.968943 | 0.281655 | 0.180245 | 0.060513 |  0.0 |  0.0 | 0.0 |  0.0 | 0.060513 | 0.243549 | 0.067652 |      2.2 | 0.060837 | 0.086998 |        0.0 | 1.104174 |      0.0 |       0.0 |        0.0 |      nan |  230.511436 | 88.849396 | 0.060513 |      0.0 |   0.0 |   0.0 |  0.0 | 0.144855 | 0.098694 |   0.0 |   0.0 |  0.0 |   0.0 | 0.067652 |
    | 2010-12-11 19:00:00 |  0.0 |  4.8 |              0.0 |                      0.0 |             0.0 |             83.4 |       7.1 |              1003.0 | 0.0 | 0.0 |  4.654167 |             92.854167 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  4.8 |  4.654167 |    5.691298 |         5.831576 |           2.520432 |          7.1 |                 8.607627 |                      8.520102 |                    8.607627 |                          0.0 |                      0.602876 |                           0.597475 |             7.178761 |                   7.91127 |     995.821239 |   1.253684 | 8.583333 |       0.0 |      -28.8 |                   0.0 |                      0.0 |                       0.0 |                  13.269547 |                 12.822742 |                3.012755 |                      0.0 |   -12.822742 |       -3.012755 |              0.0 |          0.446806 |             13.239437 |                 100.0 |               339.065437 |               69.444444 |   0.0 |    0.0 |       49.0 |    0.420261 |      0.0 |      0.0 |      0.0 |        0.734184 | 0.066523 |   0.116118 | 0.060837 |        0.0 |      0.0 | 0.020475 |   0.15874 |  0.15874 |       0.0 |       0.0 | -0.371357 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 | 15.171037 |        0.0 | -28.956917 |        0.0 | -10.773124 |         0.0 | 0.963906 | 0.278735 | 0.066523 | 0.022154 |  0.0 |  0.0 | 0.0 |  0.0 | 0.022154 | 0.231063 | 0.064184 |  2.04126 |      0.0 |      0.0 |        0.0 | 1.111482 |      0.0 |       0.0 |        0.0 |      nan |  267.894769 | 88.893765 | 0.022154 |      0.0 |   0.0 |   0.0 |  0.0 | 0.142179 | 0.088885 |   0.0 |   0.0 |  0.0 |   0.0 | 0.064184 |
    | 2010-12-11 20:00:00 |  0.0 |  4.9 |              0.0 |                      0.0 |             0.0 |             82.8 |       6.4 |              1003.0 | 0.0 | 0.0 |  4.791667 |             92.191667 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  4.9 |  4.791667 |    5.130184 |         5.854956 |           2.271939 |          6.4 |                 8.668101 |                      8.602604 |                         0.0 |                          0.0 |                      0.606604 |                           0.602566 |             7.177187 |                  7.930884 |     995.822813 |   1.253234 | 8.583333 |       0.0 |      -29.4 |                   0.0 |                      nan |                       0.0 |                  13.269547 |                   12.8326 |                     0.0 |                      0.0 |     -12.8326 |             0.0 |              0.0 |          0.436947 |               14.6875 |                 100.0 |               336.540136 |               69.444444 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.148995 | 0.148995 |       0.0 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |        0.0 |        0.0 |        0.0 |        0.0 |         0.0 | 0.958789 | 0.275783 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.218842 | 0.060789 | 1.892265 |      0.0 |      0.0 |        0.0 |      nan |      0.0 |       0.0 |        0.0 |      nan |  305.878103 | 88.893765 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.138791 | 0.080051 |   0.0 |   0.0 |  0.0 |   0.0 | 0.060789 |
    | 2010-12-11 21:00:00 |  0.0 |  4.9 |              0.0 |                      0.0 |             0.0 |             85.8 |       7.2 |              1003.0 | 0.0 | 0.0 |    4.9125 |             91.616667 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  4.9 |    4.9125 |    5.771457 |         5.908395 |           2.555931 |          7.2 |                 8.668101 |                      8.675686 |                         0.0 |                          0.0 |                      0.606604 |                           0.607071 |              7.43723 |                  7.948375 |      995.56277 |   1.253111 | 8.583333 |       0.0 |      -29.4 |                   0.0 |                      nan |                       0.0 |                  13.269547 |                 12.841028 |                     0.0 |                      0.0 |   -12.841028 |             0.0 |              0.0 |           0.42852 |             13.055556 |                 100.0 |               331.877431 |               69.444444 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.137771 | 0.137771 |       0.0 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |        0.0 |        0.0 |        0.0 |        0.0 |         0.0 | 0.953671 | 0.272847 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |   0.2073 | 0.057583 | 1.754494 |      0.0 |      0.0 |        0.0 |      nan |      0.0 |       0.0 |        0.0 |      nan |  343.861436 | 88.893765 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.135206 | 0.072094 |   0.0 |   0.0 |  0.0 |   0.0 | 0.057583 |
    | 2010-12-11 22:00:00 |  0.1 |  4.7 |              0.0 |                      0.0 |             0.0 |             87.6 |       7.0 |              1003.0 | 0.0 | 0.0 |  5.008333 |                91.125 |                   0.1 |                      7.400749 |            15.079031 | 0.094 |  4.7 |  5.008333 |    5.611139 |         5.948475 |           2.484933 |          7.0 |                 8.547525 |                      8.734036 |                         0.0 |                          0.0 |                      0.599168 |                           0.610664 |             7.487632 |                   7.95889 |     995.512368 |   1.253989 | 8.583333 |       0.0 |      -28.2 |                   0.0 |                      nan |                       0.0 |                  13.269547 |                 12.851041 |                     0.0 |                      0.0 |   -12.851041 |             0.0 |              0.0 |          0.418506 |             13.428571 |                 100.0 |               335.662549 |               69.444444 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.115706 | 0.115706 |       0.0 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |        0.0 |        0.0 |        0.0 |        0.0 |         0.0 | 0.948716 | 0.270018 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.196642 | 0.054623 | 1.732788 |      0.0 |      0.0 |        0.0 |      nan |      0.0 |       0.0 |        0.0 |      nan |  380.644769 | 88.893765 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.131713 | 0.064929 |   0.0 |   0.0 |  0.0 |   0.0 | 0.054623 |
    | 2010-12-11 23:00:00 |  0.1 |  3.8 |              0.0 |                      0.0 |             0.0 |             85.6 |       6.8 |              1002.0 | 0.0 | 0.0 |    5.0625 |             90.558333 |                   0.1 |                      7.400749 |            15.079031 | 0.094 |  3.8 |    5.0625 |     5.45082 |         5.968515 |           2.413935 |          6.8 |                 8.023017 |                      8.767169 |                         0.0 |                          0.0 |                      0.566662 |                           0.612703 |             6.867703 |                  7.939402 |     995.132297 |   1.257101 | 8.583333 |       0.0 |      -22.8 |                   0.0 |                      nan |                       0.0 |                  13.269547 |                 12.881907 |                     0.0 |                      0.0 |   -12.881907 |             0.0 |              0.0 |           0.38764 |             13.823529 |                 100.0 |               371.865153 |               69.444444 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |   0.12638 |  0.12638 |       0.0 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |        0.0 |        0.0 |        0.0 |        0.0 |         0.0 | 0.944487 | 0.267617 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.186787 | 0.051885 | 1.700408 |      0.0 |      0.0 |        0.0 |      nan |      0.0 |       0.0 |        0.0 |      nan |  412.028103 | 88.893765 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.128311 | 0.058476 |   0.0 |   0.0 |  0.0 |   0.0 | 0.051885 |
    | 2010-12-12 00:00:00 |  0.0 |  3.9 |              0.0 |                      0.0 |             0.0 |             85.3 |       6.4 |              1002.0 | 0.0 | 0.0 |    5.1125 |             89.979167 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  3.9 |    5.1125 |    5.130184 |         5.995234 |           2.271939 |          6.4 |                  8.07986 |                      8.797851 |                         0.0 |                          0.0 |                      0.570198 |                           0.614591 |             6.892121 |                  7.916233 |     995.107879 |   1.256635 | 8.583333 |       0.0 |      -23.4 |                   0.0 |                      nan |                       0.0 |                  13.269547 |                 12.915622 |                     0.0 |                      0.0 |   -12.915622 |             0.0 |              0.0 |          0.353925 |               14.6875 |                 100.0 |               368.320641 |               69.444444 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.121939 | 0.121939 |       0.0 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |        0.0 |        0.0 |        0.0 |        0.0 |         0.0 | 0.940178 |  0.26518 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.177661 |  0.04935 | 1.578469 |      0.0 |      0.0 |        0.0 |      nan |      0.0 |       0.0 |        0.0 |      nan |  444.011436 | 88.893765 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.124997 | 0.052664 |   0.0 |   0.0 |  0.0 |   0.0 |  0.04935 |
    | 2010-12-12 01:00:00 |  0.0 |  3.7 |              0.0 |                      0.0 |             0.0 |             86.2 |       6.3 |              1002.0 | 0.0 | 0.0 |  5.133333 |             89.420833 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  3.7 |  5.133333 |    5.050025 |         6.031974 |            2.23644 |          6.3 |                 7.966527 |                      8.810663 |                         0.0 |                          0.0 |                      0.563146 |                           0.615379 |             6.867146 |                  7.878569 |     995.132854 |   1.257555 | 8.583333 |       0.0 |      -22.2 |                   0.0 |                      nan |                       0.0 |                  13.269547 |                 12.957663 |                     0.0 |                      0.0 |   -12.957663 |             0.0 |              0.0 |          0.311884 |             14.920635 |                 100.0 |               375.026332 |               69.444444 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |    0.1116 |   0.1116 |       0.0 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |        0.0 |        0.0 |        0.0 |        0.0 |         0.0 | 0.936031 | 0.262846 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.169198 | 0.046999 | 1.466869 |      0.0 |      0.0 |        0.0 |      nan |      0.0 |       0.0 |        0.0 |      nan |  474.794769 | 88.893765 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.121768 |  0.04743 |   0.0 |   0.0 |  0.0 |   0.0 | 0.046999 |
    | 2010-12-12 02:00:00 |  0.1 |  3.5 |              0.0 |                      0.0 |             0.0 |             89.3 |       5.7 |              1002.0 | 0.0 | 0.0 |  5.116667 |             88.983333 |                   0.1 |                      7.400749 |            15.079031 | 0.094 |  3.5 |  5.116667 |     4.56907 |         6.031974 |           2.023445 |          5.7 |                 7.854597 |                      8.800412 |                         0.0 |                          0.0 |                      0.556168 |                           0.614748 |             7.014155 |                    7.8309 |     994.985845 |   1.258394 | 8.583333 |       0.0 |      -21.0 |                   0.0 |                      nan |                       0.0 |                  13.269547 |                 13.001862 |                     0.0 |                      0.0 |   -13.001862 |             0.0 |              0.0 |          0.267685 |             16.491228 |                 100.0 |               378.662793 |               69.444444 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.076985 | 0.076985 |       0.0 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |        0.0 |        0.0 |        0.0 |        0.0 |         0.0 | 0.932045 | 0.260612 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.161338 | 0.044816 | 1.483884 |      0.0 |      0.0 |        0.0 |      nan |      0.0 |       0.0 |        0.0 |      nan |  504.378103 | 88.893765 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.118623 | 0.042716 |   0.0 |   0.0 |  0.0 |   0.0 | 0.044816 |
    | 2010-12-12 03:00:00 |  0.0 |  3.3 |              0.0 |                      0.0 |             0.0 |             91.4 |       5.2 |              1003.0 | 0.0 | 0.0 |  5.070833 |             88.633333 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  3.3 |  5.070833 |    4.168274 |         5.985215 |            1.84595 |          5.2 |                 7.744054 |                      8.772276 |                         0.0 |                          0.0 |                      0.549265 |                           0.613018 |             7.078066 |                  7.775161 |     995.921934 |   1.260534 | 8.583333 |       0.0 |      -19.8 |                   0.0 |                      nan |                       0.0 |                  13.269547 |                 13.048148 |                     0.0 |                      0.0 |   -13.048148 |             0.0 |              0.0 |          0.221399 |             18.076923 |                 100.0 |               384.288146 |               69.444444 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.055407 | 0.055407 |       0.0 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |        0.0 |        0.0 |        0.0 |        0.0 |         0.0 | 0.928221 | 0.258478 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.154029 | 0.042786 | 1.428477 |      0.0 |      0.0 |        0.0 |      nan |      0.0 |       0.0 |        0.0 |      nan |  532.761436 | 88.893765 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.115558 |  0.03847 |   0.0 |   0.0 |  0.0 |   0.0 | 0.042786 |
    | 2010-12-12 04:00:00 |  0.0 |  3.3 |              0.0 |                      0.0 |             0.0 |             90.9 |       4.6 |              1004.0 | 0.0 | 0.0 |  5.008333 |             88.295833 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  3.3 |  5.008333 |     3.68732 |         5.905055 |           1.632956 |          4.6 |                 7.744054 |                      8.734036 |                         0.0 |                          0.0 |                      0.549265 |                           0.610664 |             7.039346 |                   7.71179 |     996.960654 |   1.261813 | 8.583333 |       0.0 |      -19.8 |                   0.0 |                      nan |                       0.0 |                  13.269547 |                 13.098861 |                     0.0 |                      0.0 |   -13.098861 |             0.0 |              0.0 |          0.170686 |             20.434783 |                 100.0 |               385.059305 |               69.444444 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.051904 | 0.051904 |       0.0 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |        0.0 |        0.0 |        0.0 |        0.0 |         0.0 | 0.924397 | 0.256353 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |  0.14722 | 0.040895 | 1.376573 |      0.0 |      0.0 |        0.0 |      nan |      0.0 |       0.0 |        0.0 |      nan |  561.144769 | 88.893765 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.112573 | 0.034647 |   0.0 |   0.0 |  0.0 |   0.0 | 0.040895 |
    | 2010-12-12 05:00:00 |  0.0 |  3.0 |              0.0 |                      0.0 |             0.0 |             89.2 |       4.9 |              1005.0 | 0.0 | 0.0 |  4.933333 |             87.945833 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  3.0 |  4.933333 |    3.927797 |         5.824896 |           1.739453 |          4.9 |                 7.580812 |                      8.688342 |                         0.0 |                          0.0 |                      0.539048 |                           0.607851 |             6.762084 |                  7.641034 |     998.237916 |   1.264577 | 8.583333 |       0.0 |      -18.0 |                   0.0 |                      nan |                       0.0 |                  13.269547 |                 13.154767 |                     0.0 |                      0.0 |   -13.154767 |             0.0 |              0.0 |           0.11478 |             19.183673 |                 100.0 |                401.87376 |               69.444444 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.065482 | 0.065482 |       0.0 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |        0.0 |        0.0 |        0.0 |        0.0 |         0.0 | 0.920815 |  0.25437 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.140869 |  0.03913 | 1.311091 |      0.0 |      0.0 |        0.0 |      nan |      0.0 |       0.0 |        0.0 |      nan |  587.728103 | 88.893765 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.109666 | 0.031203 |   0.0 |   0.0 |  0.0 |   0.0 |  0.03913 |
    | 2010-12-12 06:00:00 |  0.0 |  1.9 |              0.0 |                      0.0 |             0.0 |             85.6 |       6.3 |              1006.0 | 0.0 | 0.0 |       4.8 |             87.454167 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  1.9 |       4.8 |    5.050025 |         5.794837 |            2.23644 |          6.3 |                 7.007901 |                      8.607627 |                         0.0 |                          0.0 |                      0.502964 |                           0.602876 |             5.998763 |                  7.527728 |    1000.001237 |   1.271267 | 8.583333 |       0.0 |      -11.4 |                   0.0 |                      nan |                       0.0 |                  13.269547 |                  13.24232 |                     0.0 |                      0.0 |    -13.24232 |             0.0 |              0.0 |          0.027227 |             14.920635 |                 100.0 |                 470.6123 |               69.444444 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.108634 | 0.108634 |       0.0 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |        0.0 |        0.0 |        0.0 |        0.0 |         0.0 | 0.918123 | 0.252885 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.134935 | 0.037482 | 1.202457 |      0.0 |      0.0 |        0.0 |      nan |      0.0 |       0.0 |        0.0 |      nan |  607.711436 | 88.893765 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.106833 | 0.028102 |   0.0 |   0.0 |  0.0 |   0.0 | 0.037482 |
    | 2010-12-12 07:00:00 |  0.0 |  1.2 |              0.0 |                      0.0 |             0.0 |             86.3 |       5.8 |              1008.0 | 0.0 | 0.0 |    4.6375 |               87.0375 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  1.2 |    4.6375 |    4.649229 |         5.734717 |           2.058944 |          5.8 |                 6.663527 |                      8.510149 |                         0.0 |                          0.0 |                      0.481097 |                            0.59686 |             5.750624 |                  7.407021 |    1002.249376 |   1.277169 | 8.583333 |       0.0 |       -7.2 |                   0.0 |                      nan |                       0.0 |                  13.269547 |                 13.332335 |                     0.0 |                      0.0 |   -13.332335 |             0.0 |              0.0 |         -0.062787 |             16.206897 |                 100.0 |               521.105294 |               69.444444 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.092336 | 0.092336 |       0.0 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |        0.0 |        0.0 |        0.0 |        0.0 |         0.0 | 0.915996 | 0.251715 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.129382 |  0.03594 | 1.110122 |      0.0 |      0.0 |        0.0 |      nan |      0.0 |       0.0 |        0.0 |      nan |  623.494769 | 88.893765 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.104073 | 0.025309 |   0.0 |   0.0 |  0.0 |   0.0 |  0.03594 |
    | 2010-12-12 08:00:00 |  0.0 |  0.8 |              0.0 |                 0.339738 |        0.894144 |             91.2 |       6.1 |              1009.0 | 0.0 | 0.0 |  4.445833 |             86.870833 |                   0.1 |                      7.393446 |            15.078757 |   0.0 |  0.8 |  4.445833 |    4.889707 |         5.694638 |           2.165441 |          6.1 |                 6.473522 |                      8.396426 |                         0.0 |                          0.0 |                      0.468972 |                            0.58983 |             5.903852 |                  7.294045 |    1003.096148 |   1.280232 | 8.583333 |       0.0 |       -4.8 |              0.786847 |                      nan |                  0.146129 |                  13.269306 |                 13.409662 |                     0.0 |                      0.0 |   -12.622815 |             0.0 |              0.0 |         -0.140357 |             15.409836 |                 100.0 |               547.203515 |               98.197753 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.060807 | 0.060807 |       0.0 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |        0.0 |        0.0 |        0.0 |        0.0 |         0.0 | 0.914193 | 0.250725 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.124179 | 0.034494 | 1.049315 |      0.0 |      0.0 |        0.0 |      nan |      0.0 |       0.0 |        0.0 |      nan |  636.878103 | 88.893765 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.101385 | 0.022794 |   0.0 |   0.0 |  0.0 |   0.0 | 0.034494 |
    | 2010-12-12 09:00:00 |  0.0 |  0.8 |              0.0 |                      1.0 |       17.712539 |             87.5 |       5.9 |              1010.0 | 0.0 | 0.0 |  4.258333 |             86.504167 |                   0.1 |                      7.393446 |              15.0973 |   0.0 |  0.8 |  4.258333 |    4.729388 |         5.654558 |           2.094443 |          5.9 |                 6.473522 |                      8.286472 |                         0.0 |                          0.0 |                      0.468972 |                           0.583022 |             5.664332 |                  7.168143 |    1004.335668 |   1.281619 | 8.583333 |       0.0 |       -4.8 |             15.587034 |                      nan |                  2.894735 |                  13.285624 |                 13.502128 |                     0.0 |                      0.0 |     2.084906 |             0.0 |              0.0 |         -0.216504 |             15.932203 |                 100.0 |               554.033157 |              508.391221 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.093283 | 0.093283 |       0.0 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |        0.0 |        0.0 |        0.0 |        0.0 |         0.0 |  0.91239 | 0.249737 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.119294 | 0.033137 | 0.956032 |      0.0 |      0.0 |        0.0 |      nan |      0.0 |       0.0 |        0.0 |      nan |  650.261436 | 88.893765 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.098766 | 0.020528 |   0.0 |   0.0 |  0.0 |   0.0 | 0.033137 |
    | 2010-12-12 10:00:00 |  0.1 |  0.5 |              0.0 |                      1.0 |       54.302586 |             84.5 |       6.4 |              1011.0 | 0.0 | 0.0 |      4.05 |             86.045833 |                   0.1 |                      7.393446 |            15.167507 | 0.094 |  0.5 |      4.05 |    5.130184 |         5.631178 |           2.271939 |          6.4 |                 6.334173 |                       8.16579 |                         0.0 |                          0.0 |                      0.460051 |                           0.575537 |             5.352376 |                  7.026322 |    1005.647624 |   1.284447 | 8.583333 |       0.0 |       -3.0 |             47.786276 |                      nan |                  8.874594 |                  13.347406 |                 13.608639 |                     0.0 |                      0.0 |    34.177637 |             0.0 |              0.0 |         -0.261233 |               14.6875 |                 100.0 |               589.555744 |                537.5229 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.140469 | 0.140469 |       0.0 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |        0.0 |        0.0 |        0.0 |        0.0 |         0.0 | 0.910829 | 0.248883 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.114703 | 0.031862 | 0.909563 |      0.0 |      0.0 |        0.0 |      nan |      0.0 |       0.0 |        0.0 |      nan |  661.844769 | 88.893765 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.096215 | 0.018488 |   0.0 |   0.0 |  0.0 |   0.0 | 0.031862 |
    | 2010-12-12 11:00:00 |  0.0 |  0.3 |              0.0 |                      1.0 |       81.504147 |             76.7 |       6.8 |              1012.0 | 0.0 | 0.0 |  3.808333 |                85.425 |                   0.1 |                      7.393446 |            15.224528 |   0.0 |  0.3 |  3.808333 |     5.45082 |         5.561039 |           2.413935 |          6.8 |                  6.24275 |                      8.027741 |                         0.0 |                          0.0 |                      0.454184 |                           0.566956 |             4.788189 |                  6.857697 |    1007.211811 |   1.286932 | 8.583333 |       0.0 |       -1.8 |              71.72365 |                      nan |                 13.320106 |                  13.397585 |                 13.739115 |                     0.0 |                      0.0 |    57.984534 |             0.0 |              0.0 |          -0.34153 |             13.823529 |                 100.0 |               627.399337 |              568.150305 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.221518 | 0.221518 |       0.0 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |        0.0 |        0.0 |        0.0 |        0.0 |         0.0 |  0.90943 | 0.248119 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |  0.11038 | 0.030661 | 0.688045 |      0.0 |      0.0 |        0.0 |      nan |      0.0 |       0.0 |        0.0 |      nan |  672.228103 | 88.893765 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |  0.09373 |  0.01665 |   0.0 |   0.0 |  0.0 |   0.0 | 0.030661 |
    | 2010-12-12 12:00:00 |  0.0 | -0.1 |              0.0 |                      1.0 |       81.665294 |             74.8 |       7.0 |              1013.0 | 0.0 | 0.0 |    3.5375 |             84.791667 |                   0.1 |                      7.393446 |            15.282411 |   0.0 | -0.1 |    3.5375 |    5.611139 |          5.47754 |           2.484933 |          7.0 |                 6.063393 |                      7.875477 |                         0.0 |                          0.0 |                      0.442644 |                           0.557471 |             4.535418 |                  6.677749 |    1008.464582 |   1.290215 | 8.583333 |       0.0 |        0.6 |             71.865459 |                      nan |                 13.346442 |                  13.448521 |                 13.878543 |                     0.0 |                      0.0 |    57.986916 |             0.0 |              0.0 |         -0.430022 |             13.428571 |                 100.0 |               681.312662 |              611.073158 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.238233 | 0.238233 |       0.0 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |        0.0 |        0.0 |        0.0 |        0.0 |         0.0 | 0.908355 | 0.247533 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.106304 | 0.029529 | 0.449813 |      0.0 |      0.0 |        0.0 |      nan |      0.0 |       0.0 |        0.0 |      nan |  680.211436 | 88.893765 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.091308 | 0.014996 |   0.0 |   0.0 |  0.0 |   0.0 | 0.029529 |
    | 2010-12-12 13:00:00 |  0.0 | -0.4 |              0.0 |                      1.0 |       77.078958 |             69.7 |       6.3 |              1014.0 | 0.0 | 0.0 |  3.254167 |                84.125 |                   0.1 |                      7.393446 |            15.334225 |   0.0 | -0.4 |  3.254167 |    5.050025 |         5.424101 |            2.23644 |          6.3 |                 5.931876 |                      7.718916 |                         0.0 |                          0.0 |                      0.434155 |                           0.547693 |             4.134518 |                  6.493538 |    1009.865482 |   1.293105 | 8.583333 |       0.0 |        2.4 |             67.829483 |                      nan |                 12.596904 |                  13.494118 |                 14.023359 |                     0.0 |                      0.0 |    53.806124 |             0.0 |              0.0 |         -0.529241 |             14.920635 |                 100.0 |               736.486517 |              654.157997 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.249511 | 0.249511 |       0.0 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |        0.0 |        0.0 |        0.0 |        0.0 |         0.0 | 0.907522 | 0.247079 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.102455 |  0.02846 | 0.200302 |      0.0 |      0.0 |        0.0 |      nan |      0.0 |       0.0 |        0.0 |      nan |  686.394769 | 88.893765 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |  0.08895 | 0.013505 |   0.0 |   0.0 |  0.0 |   0.0 |  0.02846 |
    | 2010-12-12 14:00:00 |  0.0 | -0.6 |              0.2 |                      1.0 |       60.254042 |             66.9 |       7.3 |              1015.0 | 0.0 | 0.0 |  2.958333 |             83.483333 |                   0.3 |                      7.393446 |            16.085314 |   0.0 | -0.6 |  2.958333 |    5.851616 |         5.347282 |            2.59143 |          7.3 |                 5.845604 |                       7.55838 |                         0.0 |                          0.0 |                      0.428574 |                           0.537642 |             3.910709 |                  6.309988 |    1011.089291 |    1.29544 | 8.583333 |       0.0 |        3.6 |             53.023557 |                      nan |                  9.847232 |                  14.155077 |                 15.622576 |                     0.0 |                      0.0 |     37.40098 |             0.0 |              0.0 |           -1.4675 |             12.876712 |                 100.0 |               775.825198 |              684.370856 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.294254 | 0.200302 |  0.002791 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |        0.0 |        0.0 |        0.0 |        0.0 |         0.0 |  0.90685 | 0.246713 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.098815 | 0.027449 |      0.0 |      0.0 |      0.0 |        0.0 |      nan |      0.0 |       0.0 |        0.0 |      nan |  691.378103 | 88.890973 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.086652 | 0.012163 |   0.0 |   0.0 |  0.0 |   0.0 | 0.027449 |
    | 2010-12-12 15:00:00 |  0.0 | -0.8 |              0.2 |                      1.0 |         14.8373 |             64.4 |       7.5 |              1015.0 | 0.0 | 0.0 |    2.6625 |             82.729167 |                   0.4 |                      7.393446 |            16.177955 |   0.0 | -0.8 |    2.6625 |    6.011934 |         5.293842 |           2.662428 |          7.5 |                 5.760442 |                      7.400795 |                         0.0 |                          0.0 |                      0.423055 |                           0.527748 |             3.709725 |                  6.122616 |    1011.290275 |   1.296488 | 8.583333 |       0.0 |        4.8 |             13.056824 |                      nan |                  2.424839 |                    14.2366 |                 16.526635 |                     0.0 |                      0.0 |    -3.469811 |             0.0 |              0.0 |         -2.290035 |             12.533333 |                 100.0 |               818.048348 |              716.341081 |   0.0 |    0.0 |       49.0 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.296613 |      0.0 |  0.008175 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |        0.0 |        0.0 |        0.0 |        0.0 |         0.0 | 0.906341 | 0.246436 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.095368 | 0.026491 |      0.0 |      0.0 |      0.0 |        0.0 |      nan |      0.0 |       0.0 |        0.0 |      nan |  695.161436 | 88.882798 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.084414 | 0.010954 |   0.0 |   0.0 |  0.0 |   0.0 | 0.026491 |
    | 2010-12-12 16:00:00 |  0.0 | -1.3 |              0.2 |                 0.046405 |        0.168856 |             65.4 |       7.0 |              1017.0 | 0.0 | 0.0 |    2.3625 |             82.029167 |                   0.6 |                      7.386143 |            16.184078 |   0.0 | -1.3 |    2.3625 |    5.611139 |         5.237063 |           2.484933 |          7.0 |                 5.552313 |                      7.243956 |                         0.0 |                          0.0 |                      0.409522 |                           0.517875 |             3.631213 |                  5.942156 |    1013.368787 |   1.301474 | 8.583333 |       0.0 |        7.8 |              0.148593 |                      nan |                  0.027596 |                  14.241988 |                 18.187261 |                     0.0 |                      0.0 |   -18.038668 |             0.0 |              0.0 |         -3.945273 |             13.428571 |                 100.0 |               922.051683 |               72.514858 |   0.0 |    0.0 |     45.325 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.256087 |      0.0 |  0.058495 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |        0.0 |        0.0 |        0.0 |        0.0 |         0.0 | 0.906235 | 0.246379 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.092099 | 0.025583 |      0.0 |      0.0 |      0.0 |        0.0 |      nan |      0.0 |       0.0 |        0.0 |      nan |  695.944769 | 88.824303 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.082233 | 0.009865 |   0.0 |   0.0 |  0.0 |   0.0 | 0.025583 |
    | 2010-12-12 17:00:00 |  0.0 | -1.6 |              0.0 |                      0.0 |             0.0 |             67.6 |       6.0 |              1017.0 | 0.0 | 0.0 |  2.066667 |             81.658333 |                   0.6 |                      7.386143 |            16.184078 |   0.0 | -1.6 |  2.066667 |    4.809547 |         5.140204 |           2.129942 |          6.0 |                 5.430651 |                      7.092172 |                         0.0 |                          0.0 |                      0.401583 |                           0.508295 |              3.67112 |                  5.791349 |     1013.32888 |   1.302892 | 8.583333 |       0.0 |        9.6 |                   0.0 |                      nan |                       0.0 |                  14.241988 |                 18.333316 |                     0.0 |                      0.0 |   -18.333316 |             0.0 |              0.0 |         -4.091328 |             15.666667 |                 100.0 |               995.399897 |               69.444444 |   0.0 |    0.0 |      41.65 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.202105 |      0.0 |  0.053184 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |        0.0 |        0.0 |        0.0 |        0.0 |         0.0 | 0.906372 | 0.246453 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.088994 | 0.024721 |      0.0 |      0.0 |      0.0 |        0.0 |      nan |      0.0 |       0.0 |        0.0 |      nan |  694.928103 |  88.77112 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.080109 | 0.008885 |   0.0 |   0.0 |  0.0 |   0.0 | 0.024721 |
    | 2010-12-12 18:00:00 |  0.0 | -2.0 |              0.0 |                      0.0 |             0.0 |             70.2 |       4.8 |              1018.0 | 0.0 | 0.0 |  1.770833 |             81.179167 |                   0.6 |                      7.386143 |            16.184078 |   0.0 | -2.0 |  1.770833 |    3.847638 |         5.036665 |           1.703954 |          4.8 |                 5.272101 |                      6.943199 |                         0.0 |                          0.0 |                      0.391203 |                           0.498867 |             3.701015 |                  5.636431 |    1014.298985 |   1.306085 | 8.583333 |       0.0 |       12.0 |                   0.0 |                      nan |                       0.0 |                  14.241988 |                 18.489582 |                     0.0 |                      0.0 |   -18.489582 |             0.0 |              0.0 |         -4.247594 |             19.583333 |                 100.0 |              1117.480598 |               69.444444 |   0.0 |    0.0 |      36.75 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |   0.14534 |      0.0 |  0.044429 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |        0.0 |        0.0 |        0.0 |        0.0 |         0.0 | 0.906832 | 0.246703 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.086042 | 0.023901 |      0.0 |      0.0 |      0.0 |        0.0 |      nan |      0.0 |       0.0 |        0.0 |      nan |  691.511436 |  88.72669 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |  0.07804 | 0.008002 |   0.0 |   0.0 |  0.0 |   0.0 | 0.023901 |
    | 2010-12-12 19:00:00 |  0.0 | -2.1 |              0.0 |                      0.0 |             0.0 |             70.4 |       4.2 |              1019.0 | 0.0 | 0.0 |  1.483333 |               80.6375 |                   0.6 |                      7.386143 |            16.184078 |   0.0 | -2.1 |  1.483333 |    3.366683 |         4.939806 |            1.49096 |          4.2 |                 5.233109 |                      6.801075 |                         0.0 |                          0.0 |                      0.388644 |                           0.489848 |             3.684109 |                  5.484217 |    1015.315891 |    1.30786 | 8.583333 |       0.0 |       12.6 |                   0.0 |                      nan |                       0.0 |                  14.241988 |                 18.648085 |                     0.0 |                      0.0 |   -18.648085 |             0.0 |              0.0 |         -4.406096 |             22.380952 |                 100.0 |               1155.40442 |               69.444444 |   0.0 |    0.0 |     35.525 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.125512 |      0.0 |  0.041865 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |        0.0 |        0.0 |        0.0 |        0.0 |         0.0 | 0.907373 | 0.246998 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.083231 |  0.02312 |      0.0 |      0.0 |      0.0 |        0.0 |      nan |      0.0 |       0.0 |        0.0 |      nan |  687.494769 | 88.684826 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.076024 | 0.007207 |   0.0 |   0.0 |  0.0 |   0.0 |  0.02312 |
    | 2010-12-12 20:00:00 |  0.0 | -2.2 |              0.0 |                      0.0 |             0.0 |             69.5 |       4.9 |              1019.0 | 0.0 | 0.0 |    1.1875 |             80.083333 |                   0.6 |                      7.386143 |            16.184078 |   0.0 | -2.2 |    1.1875 |    3.927797 |         4.889707 |           1.739453 |          4.9 |                 5.194372 |                      6.657516 |                         0.0 |                          0.0 |                        0.3861 |                           0.480714 |             3.610088 |                  5.331561 |    1015.389912 |   1.308379 | 8.583333 |       0.0 |       13.2 |                   0.0 |                      nan |                       0.0 |                  14.241988 |                 18.808978 |                     0.0 |                      0.0 |   -18.808978 |             0.0 |              0.0 |         -4.566989 |             19.183673 |                 100.0 |              1199.718279 |               69.444444 |   0.0 |    0.0 |       34.3 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.150562 |      0.0 |  0.045237 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |        0.0 |        0.0 |        0.0 |        0.0 |         0.0 | 0.907995 | 0.247337 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.080551 | 0.022375 |      0.0 |      0.0 |      0.0 |        0.0 |      nan |      0.0 |       0.0 |        0.0 |      nan |  682.878103 | 88.639589 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |  0.07406 |  0.00649 |   0.0 |   0.0 |  0.0 |   0.0 | 0.022375 |
    | 2010-12-12 21:00:00 |  0.0 | -2.6 |              0.0 |                      0.0 |             0.0 |             71.6 |       4.0 |              1020.0 | 0.0 | 0.0 |     0.875 |             79.491667 |                   0.6 |                      7.386143 |            16.184078 |   0.0 | -2.6 |     0.875 |    3.206365 |         4.782828 |           1.419962 |          4.0 |                 5.041946 |                       6.50878 |                         0.0 |                          0.0 |                      0.376066 |                           0.471225 |             3.610033 |                  5.173937 |    1016.389967 |   1.311601 | 8.583333 |       0.0 |       15.6 |                   0.0 |                      nan |                       0.0 |                  14.241988 |                 18.977604 |                     0.0 |                      0.0 |   -18.977604 |             0.0 |              0.0 |         -4.735616 |                  23.5 |                 100.0 |              1389.131752 |               69.444444 |   0.0 |    0.0 |       29.4 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.111614 |      0.0 |  0.038133 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |        0.0 |        0.0 |        0.0 |        0.0 |         0.0 | 0.908941 | 0.247852 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.077993 | 0.021665 |      0.0 |      0.0 |      0.0 |        0.0 |      nan |      0.0 |       0.0 |        0.0 |      nan |  675.861436 | 88.601455 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.072147 | 0.005845 |   0.0 |   0.0 |  0.0 |   0.0 | 0.021665 |
    | 2010-12-12 22:00:00 |  0.0 | -3.1 |              0.0 |                      0.0 |             0.0 |             74.5 |       3.4 |              1020.0 | 0.0 | 0.0 |      0.55 |             78.945833 |                   0.6 |                      7.386143 |            16.184078 |   0.0 | -3.1 |      0.55 |     2.72541 |         4.662589 |           1.206967 |          3.4 |                 4.856984 |                      6.357212 |                         0.0 |                          0.0 |                       0.36384 |                           0.461527 |             3.618453 |                  5.018754 |    1016.381547 |   1.314025 | 8.583333 |       0.0 |       18.6 |                   0.0 |                      nan |                       0.0 |                  14.241988 |                 19.143113 |                     0.0 |                      0.0 |   -19.143113 |             0.0 |              0.0 |         -4.901124 |             27.647059 |                 100.0 |              1737.603702 |               69.444444 |   0.0 |    0.0 |       24.5 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.082224 |      0.0 |  0.030873 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |        0.0 |        0.0 |        0.0 |        0.0 |         0.0 |  0.91029 | 0.248589 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.075548 | 0.020986 |      0.0 |      0.0 |      0.0 |        0.0 |      nan |      0.0 |       0.0 |        0.0 |      nan |  665.844769 | 88.570582 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.070284 | 0.005264 |   0.0 |   0.0 |  0.0 |   0.0 | 0.020986 |
    | 2010-12-12 23:00:00 |  0.0 | -3.9 |              0.0 |                      0.0 |             0.0 |             80.2 |       3.4 |              1020.0 | 0.0 | 0.0 |  0.229167 |             78.720833 |                   0.6 |                      7.386143 |            16.184078 |   0.0 | -3.9 |  0.229167 |     2.72541 |          4.54903 |           1.206967 |          3.4 |                  4.57351 |                      6.210652 |                         0.0 |                          0.0 |                      0.344989 |                           0.452122 |             3.667955 |                  4.889077 |    1016.332045 |   1.317905 | 8.583333 |       0.0 |       23.4 |                   0.0 |                      nan |                       0.0 |                  14.241988 |                 19.269853 |                     0.0 |                      0.0 |   -19.269853 |             0.0 |              0.0 |         -5.027865 |             27.647059 |                 100.0 |              2950.347863 |               69.444444 |   0.0 |    0.0 |       24.5 |        0.42 |      0.0 |      0.0 |      0.0 |             nan |      0.0 |        0.0 |      0.0 |        0.0 |      0.0 |      0.0 |  0.060145 |      0.0 |  0.022339 |       0.0 |       0.0 |          nan |         0.12 |      0.0 |      0.0 |      0.0 |      0.0 |       0.0 |        0.0 |        0.0 |        0.0 |        0.0 |         0.0 | 0.912287 |  0.24968 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.073209 | 0.020336 |      0.0 |      0.0 |      0.0 |        0.0 |      nan |      0.0 |       0.0 |        0.0 |      nan |  651.028103 | 88.548243 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.068468 | 0.004741 |   0.0 |   0.0 |  0.0 |   0.0 | 0.020336 |

>>> round_(model.check_waterbalance(conditions_nadelw_winter))
0.0

.. _lland_v4_acker_winter_ktschnee_hourly:

acre (snow surface temperature)
-------------------------------

.. integration-test::

    >>> lnk(ACKER)
    >>> ktschnee(5.0)
    >>> test("lland_v4_acker_winter_ktschnee_hourly",
    ...      axis1=(inputs.nied, fluxes.wada),
    ...      axis2=(states.waes, states.wats),
    ...      use_conditions=conditions_acker_winter_ktschnee)
    |                date | nied | teml | sunshineduration | possiblesunshineduration | globalradiation | relativehumidity | windspeed | atmosphericpressure |  qz | qzh |   temltag | dailyrelativehumidity | dailysunshineduration | dailypossiblesunshineduration | dailyglobalradiation |  nkor | tkor |   tkortag | windspeed2m | dailywindspeed2m | reducedwindspeed2m | windspeed10m | saturationvapourpressure | dailysaturationvapourpressure | saturationvapourpressureinz | saturationvapourpressuresnow | saturationvapourpressureslope | dailysaturationvapourpressureslope | actualvapourpressure | dailyactualvapourpressure | dryairpressure | densityair |        g |        tz |         wg | netshortwaveradiation | netshortwaveradiationinz | netshortwaveradiationsnow | dailynetshortwaveradiation | dailynetlongwaveradiation | netlongwaveradiationinz | netlongwaveradiationsnow | netradiation | netradiationinz | netradiationsnow | dailynetradiation | aerodynamicresistance | soilsurfaceresistance | landusesurfaceresistance | actualsurfaceresistance |     nbes |    sbes | snowintmax | snowintrate | nbesinz | sbesinz | wniedinz | actualalbedoinz | wadainz | schmpotinz | schminz | gefrpotinz | gefrinz | evsinz |      evpo |      evi |      evb |       evs |    wnied | tempssurface | actualalbedo |  schmpot |     schm |  gefrpot |     gefr | wlatinz |   wlatsnow | wsensinz |  wsenssnow | wsurfinz |       wsurf |      sff |      fvg |     wada |      qdb | qib1 | qib2 | qbb | qkap |     qdgz |      qah |       qa |     inzp | stinz | sinz | esnowinz | asinz |      wats |      waes |      esnow |     taus |        ebdn |      bowa |    qdgz1 |    qdgz2 | qigz1 | qigz2 | qbgz |    qdga1 |    qdga2 | qiga1 | qiga2 | qbga | inlet |   outlet |
    --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 2010-12-10 00:00:00 |  0.0 | -3.7 |              0.0 |                      0.0 |             0.0 |             87.5 |       3.4 |              1017.0 | 0.0 | 0.0 |    -2.025 |             96.870833 |                  4.41 |                           7.5 |            23.154167 |   0.0 | -3.7 |    -2.025 |     2.72541 |         2.684392 |            2.72541 |          3.4 |                  4.64297 |                      5.262329 |                         0.0 |                     3.363108 |                      0.349621 |                           0.390562 |             4.062599 |                  5.097662 |    1012.937401 |   1.312855 | 8.583333 | -2.470836 |  -1.137119 |                   0.0 |                      0.0 |                       0.0 |                   5.712355 |                 46.101494 |                     0.0 |                62.690184 |   -46.101494 |             0.0 |       -62.690184 |        -40.389139 |             62.648725 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.001998 |      0.0 |      0.0 | -0.010568 |      0.0 |    -7.904522 |      0.75329 |      0.0 |      0.0 | 0.315303 | 0.315303 |     0.0 |  -7.830631 |      0.0 |  -26.74352 |      0.0 |   28.116032 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 19.067371 | 20.422623 |        0.0 | 0.871765 | -196.182518 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 01:00:00 |  0.0 | -4.8 |              0.0 |                      0.0 |             0.0 |             90.7 |       2.9 |              1018.0 | 0.0 | 0.0 | -2.158333 |                96.825 |                  4.41 |                           7.5 |            23.154167 |   0.0 | -4.8 | -2.158333 |    2.324615 |         2.647918 |           2.324615 |          2.9 |                 4.272186 |                      5.210481 |                         0.0 |                     3.108746 |                      0.324793 |                           0.387158 |             3.874873 |                  5.045049 |    1014.125127 |   1.319627 | 8.583333 |  -2.35419 |     0.7592 |                   0.0 |                      0.0 |                       0.0 |                    5.72751 |                 46.220713 |                     0.0 |                63.329844 |   -46.220713 |             0.0 |       -63.329844 |        -40.493203 |              73.45023 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.004391 |      0.0 |      0.0 | -0.010408 |      0.0 |    -8.906723 |     0.752636 |      0.0 |      0.0 | 0.338128 | 0.338128 |     0.0 |   -7.71192 |      0.0 | -23.487925 |      0.0 |   32.129999 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 19.415389 | 20.433031 |        0.0 | 0.879073 | -188.358384 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 02:00:00 |  0.0 | -5.6 |              0.0 |                      0.0 |             0.0 |             89.7 |       2.5 |              1018.0 | 0.0 | 0.0 |    -2.325 |             96.604167 |                  4.41 |                           7.5 |            23.154167 |   0.0 | -5.6 |    -2.325 |    2.003978 |         2.606417 |           2.003978 |          2.5 |                 4.019239 |                      5.146307 |                         0.0 |                     2.893794 |                      0.307706 |                            0.38294 |             3.605258 |                  4.971547 |    1014.394742 |   1.323706 | 8.583333 | -2.260301 |   2.457374 |                   0.0 |                      0.0 |                       0.0 |                   5.742547 |                  46.40442 |                     0.0 |                64.160092 |    -46.40442 |             0.0 |       -64.160092 |        -40.661873 |             85.202267 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.004461 |      0.0 |      0.0 | -0.008798 |      0.0 |    -9.812089 |     0.751986 |      0.0 |      0.0 | 0.358424 | 0.358424 |     0.0 |  -6.519287 |      0.0 | -21.929674 |      0.0 |   35.711131 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 19.782327 |  20.44183 |        0.0 | 0.886381 | -182.232425 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 03:00:00 |  0.0 | -6.2 |              0.0 |                      0.0 |             0.0 |             91.3 |       3.0 |              1019.0 | 0.0 | 0.0 | -2.420833 |                96.325 |                  4.41 |                           7.5 |            23.154167 |   0.0 | -6.2 | -2.420833 |    2.404774 |         2.602449 |           2.404774 |          3.0 |                 3.838328 |                      5.109724 |                         0.0 |                     2.841896 |                      0.295404 |                           0.380532 |             3.504393 |                  4.921942 |    1015.495607 |   1.328036 | 8.583333 | -2.186789 |   3.218369 |                   0.0 |                      0.0 |                       0.0 |                   5.757468 |                 46.542687 |                     0.0 |                65.853061 |   -46.542687 |             0.0 |       -65.853061 |        -40.785219 |             71.001889 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.004771 |      0.0 |      0.0 | -0.009202 |      0.0 |   -10.039607 |     0.751342 |      0.0 |      0.0 | 0.359609 | 0.359609 |     0.0 |  -6.818315 |      0.0 | -22.452631 |      0.0 |   36.582115 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.151004 | 20.451032 |        0.0 | 0.893689 | -176.867461 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 04:00:00 |  0.0 | -7.3 |              0.0 |                      0.0 |             0.0 |             94.2 |       2.5 |              1020.0 | 0.0 | 0.0 | -2.558333 |               96.2125 |                  4.41 |                           7.5 |            23.154167 |   0.0 | -7.3 | -2.558333 |    2.003978 |         2.623448 |           2.003978 |          2.5 |                  3.52531 |                      5.057637 |                         0.0 |                      2.62027 |                      0.273951 |                           0.377101 |             3.320842 |                  4.866079 |    1016.679158 |   1.334932 | 8.583333 |  -2.12241 |   4.970078 |                   0.0 |                      0.0 |                       0.0 |                   5.772275 |                 46.677297 |                     0.0 |                66.487965 |   -46.677297 |             0.0 |       -66.487965 |        -40.905023 |             85.202267 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.008511 |      0.0 |      0.0 | -0.008664 |      0.0 |   -11.054742 |     0.750703 |      0.0 |      0.0 | 0.383172 | 0.300028 |     0.0 |  -6.419479 |      0.0 | -19.548556 |      0.0 |    40.51993 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.459696 | 20.459696 |  -7.713912 | 0.900997 | -173.254206 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 05:00:00 |  0.0 | -8.3 |              0.0 |                      0.0 |             0.0 |             93.2 |       2.3 |              1020.0 | 0.0 | 0.0 |    -2.725 |             96.004167 |                  4.41 |                           7.5 |            23.154167 |   0.0 | -8.3 |    -2.725 |     1.84366 |         2.596101 |            1.84366 |          2.3 |                  3.26061 |                      4.995131 |                         0.0 |                     2.401909 |                       0.25563 |                           0.372977 |             3.038889 |                  4.795534 |    1016.961111 |   1.340112 | 8.583333 |  -2.07905 |   8.729642 |                   0.0 |                      0.0 |                       0.0 |                   5.784108 |                 46.854961 |                     0.0 |                67.511852 |   -46.854961 |             0.0 |       -67.511852 |        -41.070852 |             92.611159 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.007601 |      0.0 |      0.0 | -0.007489 |      0.0 |   -12.132478 |     0.750191 |      0.0 |      0.0 | 0.452444 |      0.0 |     0.0 |  -5.549204 |      0.0 | -18.970213 |      0.0 |   42.992436 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.467185 | 20.467185 | -41.976706 | 0.906878 | -173.400514 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 06:00:00 |  0.0 | -8.6 |              0.0 |                      0.0 |             0.0 |             96.5 |       2.1 |              1021.0 | 0.0 | 0.0 |    -2.925 |             96.004167 |                  4.41 |                           7.5 |            23.154167 |   0.0 | -8.6 |    -2.925 |    1.683342 |          2.56624 |           1.683342 |          2.1 |                 3.184716 |                      4.921026 |                         0.0 |                     2.262486 |                      0.250344 |                           0.368079 |             3.073251 |                   4.72439 |    1017.926749 |   1.342932 | 8.583333 | -2.080806 |  18.665091 |                   0.0 |                      0.0 |                       0.0 |                   5.790651 |                 47.012175 |                     0.0 |                65.101655 |   -47.012175 |             0.0 |       -65.101655 |        -41.221524 |             101.43127 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.010208 |      0.0 |      0.0 | -0.009039 |      0.0 |    -12.86713 |     0.749909 |      0.0 |      0.0 | 0.664911 |      0.0 |     0.0 |  -6.697155 |      0.0 | -20.027121 |      0.0 |   38.377378 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.476223 | 20.476223 | -61.688994 | 0.910146 | -183.482272 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 07:00:00 |  0.0 | -8.5 |              0.0 |                      0.0 |             0.0 |             95.9 |       1.4 |              1021.0 | 0.0 | 0.0 |   -3.1625 |               96.0625 |                  3.76 |                           7.4 |            23.154167 |   0.0 | -8.5 |   -3.1625 |    1.122228 |         2.521333 |           1.122228 |          1.4 |                 3.209838 |                      4.834291 |                         0.0 |                     2.180312 |                      0.252096 |                           0.362336 |             3.078235 |                  4.643941 |    1017.921765 |   1.342422 | 8.583333 | -2.201787 |  23.882513 |                   0.0 |                      0.0 |                       0.0 |                   5.796115 |                  42.68596 |                     0.0 |                59.974945 |    -42.68596 |             0.0 |       -59.974945 |        -36.889845 |            152.146905 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.008397 |      0.0 |      0.0 | -0.008095 |      0.0 |    -13.31921 |     0.749673 |      0.0 |      0.0 | 0.792124 |      0.0 |     0.0 |  -5.998301 |      0.0 | -18.291623 |      0.0 |   35.685021 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.484319 | 20.484319 | -73.491502 | 0.912883 | -198.781451 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 08:00:00 |  0.0 | -9.5 |              0.0 |                 0.355232 |        1.107611 |             96.8 |       1.1 |              1022.0 | 0.0 | 0.0 | -3.483333 |               96.2625 |                  2.76 |                      6.755232 |            23.200317 |   0.0 | -9.5 | -3.483333 |     0.88175 |         2.408072 |            0.88175 |          1.1 |                 2.966351 |                       4.71927 |                         0.0 |                     2.038074 |                      0.235051 |                           0.354699 |             2.871427 |                  4.542888 |    1019.128573 |   1.348938 | 8.583333 | -2.385377 |  27.229578 |              0.277505 |                      0.0 |                  0.277505 |                   5.812688 |                 37.241518 |                     0.0 |                57.178258 |   -36.964013 |             0.0 |       -56.900753 |         -31.42883 |            193.641515 |                 100.0 |               112.102145 |               99.616909 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.006537 |      0.0 |      0.0 | -0.006752 |      0.0 |   -14.138836 |     0.749457 |      0.0 |      0.0 | 0.887474 |      0.0 |     0.0 |  -5.002627 |      0.0 | -15.822146 |      0.0 |    36.07598 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  20.49107 |  20.49107 | -82.337904 |   0.9154 | -217.427696 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 09:00:00 |  0.0 | -9.2 |              0.0 |                      1.0 |       18.663203 |             94.3 |       1.0 |              1022.0 | 0.0 | 0.0 | -3.816667 |             96.383333 |                  1.79 |                      6.755232 |            21.661284 |   0.0 | -9.2 | -3.816667 |    0.801591 |         2.278972 |           0.801591 |          1.0 |                 3.037615 |                      4.602339 |                         0.0 |                     2.193836 |                      0.240056 |                           0.346913 |             2.864471 |                  4.435888 |    1019.135529 |   1.347408 | 8.583333 | -2.609132 |  27.377208 |              4.679745 |                      0.0 |                  4.679745 |                   5.431506 |                 29.260785 |                     0.0 |                52.186965 |    -24.58104 |             0.0 |       -47.507219 |        -23.829279 |            213.005666 |                 100.0 |               112.102145 |              101.107639 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.002304 |      0.0 |      0.0 | -0.005229 |      0.0 |   -13.243778 |     0.749253 |      0.0 |      0.0 | 0.919612 |      0.0 |     0.0 |  -3.874446 |      0.0 | -13.273886 |      0.0 |   30.358887 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.496299 | 20.496299 | -85.319584 | 0.917775 |  -236.22157 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 10:00:00 |  0.0 | -6.9 |              0.0 |                      1.0 |       56.267908 |             96.7 |       1.4 |              1022.0 | 0.0 | 0.0 |   -4.0625 |               96.5375 |                  0.88 |                      6.755232 |             18.21828 |   0.0 | -6.9 |   -4.0625 |    1.122228 |         2.200731 |           1.122228 |          1.4 |                 3.636412 |                      4.517752 |                         0.0 |                     2.927254 |                      0.281591 |                           0.341265 |             3.516411 |                  4.361325 |    1018.483589 |   1.335446 | 8.583333 | -2.834659 |  22.468665 |             14.120243 |                      0.0 |                 14.120243 |                   4.571816 |                 21.675213 |                     0.0 |                43.998807 |     -7.55497 |             0.0 |       -29.878564 |        -17.103397 |            152.146905 |                 100.0 |               112.102145 |              101.107639 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.002167 |      0.0 |      0.0 | -0.005312 |      0.0 |    -9.667312 |     0.749053 |      0.0 |      0.0 | 0.843847 |      0.0 |     0.0 |  -3.935678 |      0.0 | -10.503509 |      0.0 |   15.439376 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.501611 | 20.501611 | -78.290295 | 0.920107 | -250.106902 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 11:00:00 |  0.0 | -4.0 |              0.1 |                      1.0 |      107.168236 |             97.4 |       1.3 |              1022.0 | 0.0 | 0.0 | -4.195833 |             96.570833 |                  0.39 |                      6.755232 |             15.73779 |   0.0 | -4.0 | -4.195833 |    1.042069 |         2.098318 |           1.042069 |          1.3 |                 4.539126 |                      4.472452 |                         0.0 |                     3.993365 |                      0.342693 |                           0.338235 |             4.421109 |                  4.319085 |    1017.578891 |   1.320615 | 8.583333 | -3.001283 |   14.66662 |             26.915813 |                      0.0 |                 26.915813 |                   3.952621 |                 17.574343 |                     0.0 |                37.046663 |      9.34147 |             0.0 |        -10.13085 |        -13.621722 |            163.850513 |                 100.0 |               112.102145 |              101.107639 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.008302 |      0.0 |      0.0 | -0.003726 |      0.0 |    -5.684328 |     0.748845 |      0.0 |      0.0 | 0.698623 |      0.0 |     0.0 |  -2.760855 |      0.0 |  -6.176952 |      0.0 |    1.193042 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.505337 | 20.505337 | -64.816717 | 0.922546 | -256.190189 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 12:00:00 |  0.0 | -2.8 |              0.0 |                      1.0 |       91.473333 |             93.9 |       1.3 |              1022.0 | 0.0 | 0.0 |     -4.35 |               96.3625 |                  0.14 |                      6.755232 |            14.920012 |   0.0 | -2.8 |     -4.35 |    1.042069 |          2.00007 |           1.042069 |          1.3 |                 4.967227 |                      4.420576 |                         0.0 |                     4.325457 |                      0.371134 |                            0.33476 |             4.664226 |                  4.259778 |    1017.335774 |   1.314635 | 8.583333 | -3.074282 |   9.438843 |             22.994795 |                      0.0 |                 22.994795 |                    3.75063 |                 15.511434 |                     0.0 |                31.865111 |     7.483361 |             0.0 |        -8.870316 |        -11.760804 |            163.850513 |                 100.0 |               112.102145 |              101.107639 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.010332 |      0.0 |      0.0 | -0.002951 |      0.0 |    -4.636885 |     0.748618 |      0.0 |      0.0 | 0.596319 |      0.0 |     0.0 |  -2.186577 |      0.0 |  -6.736427 |      0.0 |   -0.052688 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.508288 | 20.508288 | -55.325186 | 0.925222 | -257.045699 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 13:00:00 |  0.0 | -1.4 |              0.0 |                      1.0 |       80.475536 |             93.4 |       1.4 |              1021.0 | 0.0 | 0.0 | -4.370833 |             96.116667 |                   0.1 |                      6.755232 |            15.956493 |   0.0 | -1.4 | -4.370833 |    1.122228 |         1.900997 |           1.122228 |          1.4 |                 5.511494 |                      4.413607 |                         0.0 |                     4.749265 |                      0.406861 |                           0.334293 |             5.147736 |                  4.242212 |    1015.852264 |   1.306346 | 8.583333 | -3.084548 |   5.283868 |             20.249823 |                      0.0 |                 20.249823 |                   4.015086 |                 15.193165 |                     0.0 |                27.668109 |     5.056659 |             0.0 |        -7.418286 |        -11.178079 |            152.146905 |                 100.0 |               112.102145 |              101.107639 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.010792 |      0.0 |      0.0 | -0.003592 |      0.0 |    -3.399006 |     0.748373 |      0.0 |      0.0 | 0.508854 |      0.0 |     0.0 |  -2.661861 |      0.0 |  -7.587358 |      0.0 |   -2.830933 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  20.51188 |  20.51188 | -47.210384 | 0.928104 | -253.746234 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 14:00:00 |  0.0 | -0.4 |              0.0 |                      1.0 |       43.120592 |             93.3 |       2.3 |              1020.0 | 0.0 | 0.0 | -4.354167 |             95.854167 |                   0.1 |                      6.755232 |            16.594851 |   0.0 | -0.4 | -4.354167 |     1.84366 |         1.856982 |            1.84366 |          2.3 |                 5.931876 |                      4.419181 |                         0.0 |                     4.984171 |                      0.434155 |                           0.334667 |              5.53444 |                   4.23597 |     1014.46556 |   1.300092 | 8.583333 | -3.044955 |   2.487524 |             10.861617 |                      0.0 |                 10.861617 |                   4.180066 |                 15.206659 |                     0.0 |                23.783791 |    -4.345042 |             0.0 |       -12.922174 |        -11.026593 |             92.611159 |                 100.0 |               112.102145 |              101.107639 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.010392 |      0.0 |      0.0 |  -0.00647 |      0.0 |    -2.754414 |     0.748111 |      0.0 |      0.0 | 0.444042 |      0.0 |     0.0 |  -4.793806 |      0.0 | -11.654009 |      0.0 |   -3.525642 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  20.51835 |  20.51835 | -41.197218 | 0.931204 | -247.650425 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 15:00:00 |  0.0 |  0.5 |              0.0 |                      1.0 |       10.320092 |             94.1 |       2.8 |              1020.0 | 0.0 | 0.0 |      -4.3 |             95.608333 |                   0.1 |                      7.355232 |            17.024855 |   0.0 |  0.5 |      -4.3 |    2.244455 |         1.842168 |           2.244455 |          2.8 |                 6.334173 |                      4.437342 |                         0.0 |                     5.224468 |                      0.460051 |                           0.335883 |             5.960456 |                  4.242469 |    1014.039544 |   1.295611 | 8.583333 | -2.971805 |   0.405685 |              2.602392 |                      0.0 |                  2.602392 |                   4.293115 |                  15.14111 |                     0.0 |                19.920303 |   -12.538718 |             0.0 |       -17.317912 |        -10.847995 |             76.073452 |                 100.0 |               112.102145 |              101.107639 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.00709 |      0.0 |      0.0 | -0.009774 |      0.0 |    -2.122251 |     0.747833 |      0.0 |      0.0 | 0.390241 |      0.0 |     0.0 |  -7.242412 |      0.0 | -14.661342 |      0.0 |   -4.585842 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.528124 | 20.528124 |  -36.20569 | 0.934503 | -239.472776 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 16:00:00 |  0.1 |  1.0 |              0.0 |                 0.061898 |        0.028103 |             94.6 |       3.1 |              1019.0 | 0.0 | 0.0 |      -4.2 |             95.395833 |                   0.1 |                       7.41713 |            17.026026 | 0.094 |  1.0 |      -4.2 |    2.484933 |          1.86654 |           2.484933 |          3.1 |                 6.567919 |                      4.471043 |                         0.0 |                     5.428662 |                      0.475001 |                            0.33814 |             6.213251 |                  4.265189 |    1012.786749 |   1.291856 | 8.583333 | -2.873673 |  -1.239975 |               0.00709 |                      0.0 |                   0.00709 |                   4.295298 |                 15.125699 |                     0.0 |                18.515676 |    -15.11861 |             0.0 |       -18.508587 |        -10.830401 |             68.711505 |                 100.0 |               112.102145 |               98.953162 |    0.094 |  0.0235 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.005889 |      0.0 |      0.0 | -0.011137 | 0.095635 |    -1.604952 |     0.747722 |      0.0 |      0.0 | 0.345338 |   0.0705 |     0.0 |  -8.251965 |      0.0 | -15.566914 |      0.0 |   -5.310293 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.633261 | 20.633261 | -25.498904 | 0.935821 | -229.649468 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 17:00:00 |  0.0 |  1.1 |              0.0 |                      0.0 |             0.0 |             94.4 |       3.5 |              1018.0 | 0.0 | 0.0 | -4.091667 |             95.170833 |                   0.1 |                       7.41713 |            17.026026 |   0.0 |  1.1 | -4.091667 |    2.805569 |         1.866772 |           2.805569 |          3.5 |                 6.615571 |                      4.507808 |                         0.0 |                     5.555272 |                      0.478041 |                             0.3406 |             6.245099 |                  4.290119 |    1011.754901 |     1.2901 | 8.583333 | -2.755794 |   -3.87153 |                   0.0 |                      0.0 |                       0.0 |                   4.301126 |                 15.115697 |                     0.0 |                19.315852 |   -15.115697 |             0.0 |       -19.315852 |        -10.814571 |             60.858762 |                 100.0 |               112.102145 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.007388 |      0.0 |      0.0 | -0.010632 |      0.0 |    -1.292776 |     0.747379 |      0.0 |      0.0 | 0.272497 |      0.0 |     0.0 |   -7.87815 |      0.0 | -15.526513 |      0.0 |   -4.088811 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.643894 | 20.643894 | -25.281624 | 0.939902 | -217.194605 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 18:00:00 |  0.3 |  1.0 |              0.0 |                      0.0 |             0.0 |             96.9 |       4.4 |              1016.0 | 0.0 | 0.0 |     -3.95 |             95.070833 |                   0.1 |                       7.41713 |            17.026026 | 0.282 |  1.0 |     -3.95 |    3.527001 |         1.942897 |           3.527001 |          4.4 |                 6.567919 |                      4.556289 |                         0.0 |                     5.666025 |                      0.475001 |                           0.343839 |             6.364313 |                  4.331702 |    1009.635687 |   1.287972 | 8.583333 | -2.606335 |  -3.840886 |                   0.0 |                      0.0 |                       0.0 |                     4.2976 |                  15.08957 |                     0.0 |                20.162546 |    -15.08957 |             0.0 |       -20.162546 |         -10.79197 |             48.410379 |                 100.0 |               112.102145 |               98.814229 |    0.282 |  0.0705 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.003845 |      0.0 |      0.0 | -0.012677 | 0.286904 |    -1.024811 |     0.747586 |      0.0 |      0.0 |  0.26007 |   0.2115 |     0.0 |  -9.393392 |      0.0 | -15.476039 |      0.0 |   -4.706884 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.938571 | 20.938571 |  -4.506222 | 0.937431 | -204.770386 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 19:00:00 |  0.2 |  1.4 |              0.0 |                      0.0 |             0.0 |             98.3 |       3.8 |              1015.0 | 0.0 | 0.0 | -3.816667 |             95.029167 |                   0.1 |                       7.41713 |            17.026026 | 0.188 |  1.4 | -3.816667 |    3.046047 |         1.973983 |           3.046047 |          3.8 |                 6.760362 |                      4.602339 |                         0.0 |                     5.909622 |                       0.48726 |                           0.346913 |             6.645436 |                  4.373564 |    1008.354564 |   1.284691 | 8.583333 | -2.457245 |  -9.291863 |                   0.0 |                      0.0 |                       0.0 |                   4.303084 |                 15.061291 |                     0.0 |                19.340879 |   -15.061291 |             0.0 |       -19.340879 |        -10.758208 |             56.054123 |                 100.0 |               112.102145 |               98.814229 |    0.188 |  0.0282 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.000004 |      0.0 |      0.0 | -0.012014 | 0.283107 |    -0.451343 |     0.747264 |      0.0 |      0.0 | 0.121028 | 0.121028 |     0.0 |  -8.901633 |      0.0 | -12.725532 |      0.0 |   -2.286287 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  21.09979 | 21.138585 |        0.0 | 0.941277 | -186.895189 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 20:00:00 |  0.4 |  1.6 |              0.0 |                      0.0 |             0.0 |             98.7 |       5.7 |              1014.0 | 0.0 | 0.0 |    -3.675 |             94.991667 |                   0.1 |                       7.41713 |            17.026026 | 0.376 |  1.6 |    -3.675 |     4.56907 |         2.056027 |            4.56907 |          5.7 |                 6.858436 |                      4.651718 |                         0.0 |                       6.1078 |                      0.493491 |                           0.350204 |             6.769276 |                  4.418744 |    1007.230724 |   1.282429 | 8.583333 | -2.242742 | -11.253101 |                   0.0 |                      0.0 |                       0.0 |                   4.308455 |                 15.030582 |                     0.0 |                19.815098 |   -15.030582 |             0.0 |       -19.815098 |        -10.722127 |             37.369415 |                 100.0 |               112.102145 |               98.814229 |    0.376 |  0.0376 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.000346 |      0.0 |      0.0 | -0.014629 | 0.664621 |          0.0 |     0.746949 |      0.0 |      0.0 | 0.050308 | 0.050308 |     0.0 | -10.839285 |      0.0 | -14.896819 |      0.0 |   -5.921007 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 21.202105 | 21.529213 |        0.0 | 0.945058 | -167.058755 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 21:00:00 |  0.7 |  2.0 |              0.0 |                      0.0 |             0.0 |             99.6 |       5.6 |              1013.0 | 0.0 | 0.0 |      -3.5 |             94.991667 |                   0.1 |                       7.41713 |            17.026026 | 0.658 |  2.0 |      -3.5 |    4.488911 |         2.138899 |           4.488911 |          5.6 |                 7.058357 |                      4.713362 |                         0.0 |                       6.1078 |                      0.506157 |                           0.354306 |             7.030123 |                  4.477301 |    1005.969877 |   1.279174 | 8.583333 | -2.004705 |  -12.02823 |                   0.0 |                      0.0 |                       0.0 |                   4.318814 |                  14.98921 |                     0.0 |                16.539253 |    -14.98921 |             0.0 |       -16.539253 |        -10.670395 |             38.036726 |                 100.0 |               112.102145 |               98.814229 |    0.658 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.002532 |      0.0 |      0.0 | -0.020116 | 1.530508 |          0.0 |      0.74634 | 0.067182 | 0.067182 |      0.0 |      0.0 |     0.0 | -14.905456 |      0.0 | -18.364515 |      0.0 |  -16.730719 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 21.154085 |  22.20733 |        0.0 | 0.952394 | -146.447191 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 22:00:00 |  0.5 |  2.4 |              0.0 |                      0.0 |             0.0 |             99.4 |       5.8 |              1011.0 | 0.0 | 0.0 | -3.291667 |             95.008333 |                   0.1 |                       7.41713 |            17.026026 |  0.47 |  2.4 | -3.291667 |    4.649229 |         2.249283 |           4.649229 |          5.8 |                 7.263399 |                      4.787689 |                         0.0 |                       6.1078 |                      0.519101 |                           0.359244 |             7.219819 |                  4.548704 |    1003.780181 |   1.274698 | 8.583333 | -1.757366 | -10.544198 |                   0.0 |                      0.0 |                       0.0 |                    4.32909 |                 14.938353 |                     0.0 |                 13.71239 |   -14.938353 |             0.0 |        -13.71239 |        -10.609263 |             36.725115 |                 100.0 |               112.102145 |               98.814229 |     0.47 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.001838 |      0.0 |      0.0 | -0.024931 | 1.311864 |          0.0 |     0.745737 | 0.195967 | 0.195967 |      0.0 |      0.0 |     0.0 | -18.473108 |      0.0 |  -22.65304 |      0.0 |  -27.413757 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.981159 | 22.702261 |        0.0 | 0.959724 |  -127.31966 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-10 23:00:00 |  1.0 |  2.5 |              0.0 |                      0.0 |             0.0 |             99.2 |       6.2 |              1010.0 | 0.0 | 0.0 | -3.070833 |                  95.0 |                   0.1 |                       7.41713 |            17.026026 |  0.94 |  2.5 | -3.070833 |    4.969866 |         2.364694 |           4.969866 |          6.2 |                 7.315473 |                      4.867606 |                         0.0 |                       6.1078 |                      0.522381 |                           0.364543 |             7.256949 |                  4.624226 |    1002.743051 |   1.272954 | 8.583333 | -1.527836 |  -9.167016 |                   0.0 |                      0.0 |                       0.0 |                   4.339318 |                 14.885653 |                     0.0 |                13.067893 |   -14.885653 |             0.0 |       -13.067893 |        -10.546335 |             34.355753 |                 100.0 |               112.102145 |               98.814229 |     0.94 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.000987 |      0.0 |      0.0 | -0.027164 |  2.73305 |          0.0 |     0.745136 | 0.274905 | 0.274905 |      0.0 |      0.0 |     0.0 |  -20.12751 |      0.0 | -24.879463 |      0.0 |   -31.93908 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.730045 | 23.669425 |        0.0 | 0.967076 | -109.569311 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-11 00:00:00 |  2.0 |  2.7 |              0.0 |                      0.0 |             0.0 |             99.2 |       5.6 |              1008.0 | 0.0 | 0.0 | -2.804167 |               95.4875 |                   0.1 |                       7.41713 |            17.026026 |  1.88 |  2.7 | -2.804167 |    4.488911 |         2.438173 |           4.488911 |          5.6 |                 7.420609 |                      4.965681 |                         0.0 |                       6.1078 |                      0.528994 |                           0.371032 |             7.361244 |                  4.741604 |    1000.638756 |   1.269455 | 8.583333 | -1.314832 |   -7.88899 |                   0.0 |                      0.0 |                       0.0 |                   4.349522 |                 14.787036 |                     0.0 |                11.601106 |   -14.787036 |             0.0 |       -11.601106 |        -10.437515 |             38.036726 |                 100.0 |               112.102145 |               98.814229 |     1.88 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.001174 |      0.0 |      0.0 | -0.027338 | 5.903388 |          0.0 |     0.744537 | 0.339112 | 0.339112 |      0.0 |      0.0 |     0.0 |  -20.25663 |      0.0 | -24.792095 |      0.0 |  -33.447619 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 20.412752 | 25.576764 |        0.0 | 0.974465 |  -93.096987 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-11 01:00:00 |  2.1 |  3.2 |              0.0 |                      0.0 |             0.0 |             99.6 |       5.2 |              1007.0 | 0.0 | 0.0 | -2.470833 |             95.858333 |                   0.1 |                       7.41713 |            17.026026 | 1.974 |  3.2 | -2.470833 |    4.168274 |         2.514993 |           4.168274 |          5.2 |                 7.689299 |                      5.090729 |                         0.0 |                       6.1078 |                      0.545841 |                           0.379281 |             7.658542 |                  4.879888 |     999.341458 |   1.265756 | 8.583333 | -1.117164 |  -6.702983 |                   0.0 |                      0.0 |                       0.0 |                    4.35966 |                 14.678585 |                     0.0 |                 7.733351 |   -14.678585 |             0.0 |        -7.733351 |        -10.318925 |             40.962628 |                 100.0 |               112.102145 |               98.814229 |    1.974 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.002516 |      0.0 |      0.0 | -0.031933 | 7.346438 |          0.0 |     0.743941 | 0.477622 | 0.477622 |      0.0 |      0.0 |     0.0 | -23.661007 |      0.0 | -27.741565 |      0.0 |  -43.669221 |      1.0 |      0.3 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |      0.0 |      0.0 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 19.958236 | 27.582697 |        0.0 | 0.981862 |  -77.810671 |  71.44626 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |   0.0 |      0.0 |
    | 2010-12-11 02:00:00 |  1.8 |  3.9 |              0.0 |                      0.0 |             0.0 |             99.8 |       5.7 |              1006.0 | 0.0 | 0.0 |    -2.075 |             96.279167 |                   0.1 |                       7.41713 |            17.026026 | 1.692 |  3.9 |    -2.075 |     4.56907 |         2.621871 |            4.56907 |          5.7 |                  8.07986 |                      5.242833 |                         0.0 |                       6.1078 |                      0.570198 |                           0.389283 |               8.0637 |                  5.047756 |       997.9363 |   1.261108 | 8.583333 | -0.933728 |  -5.602368 |                   0.0 |                      0.0 |                       0.0 |                   4.369668 |                 14.548979 |                     0.0 |                 2.415944 |   -14.548979 |             0.0 |        -2.415944 |        -10.179312 |             37.369415 |                 100.0 |               112.102145 |               98.814229 |    1.692 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.003093 |      0.0 |      0.0 | -0.043255 | 4.343299 |          0.0 |     0.743354 | 0.697218 | 0.697218 |      0.0 |      0.0 |     0.0 | -32.050365 |      0.0 | -36.310997 |      0.0 |  -65.945417 |      1.0 |      0.3 | 1.732952 | 0.592211 |  0.0 |  0.0 | 0.0 |  0.0 | 0.592211 | 0.011148 | 0.003097 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  19.29021 |    27.585 |        0.0 | 0.989219 |  -63.624969 | 72.587001 |      0.5 | 0.092211 |   0.0 |   0.0 |  0.0 | 0.006486 | 0.004662 |   0.0 |   0.0 |  0.0 |   0.0 | 0.003097 |
    | 2010-12-11 03:00:00 |  1.3 |  4.4 |              0.0 |                      0.0 |             0.0 |             99.8 |       6.6 |              1005.0 | 0.0 | 0.0 | -1.633333 |             96.633333 |                   0.1 |                       7.41713 |            17.026026 | 1.222 |  4.4 | -1.633333 |    5.290502 |          2.74211 |           5.290502 |          6.6 |                  8.36943 |                      5.417279 |                         0.0 |                       6.1078 |                       0.58816 |                           0.400709 |             8.352691 |                  5.234898 |     996.647309 |   1.257444 | 8.583333 |   -0.7635 |  -4.580998 |                   0.0 |                      0.0 |                       0.0 |                   4.379537 |                 14.409644 |                     0.0 |                -1.360594 |   -14.409644 |             0.0 |         1.360594 |        -10.030107 |             32.273586 |                 100.0 |               110.471868 |               98.814229 |    1.222 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 | -0.002984 |      0.0 |      0.0 | -0.055801 |      0.0 |          0.0 |     0.742774 | 0.907235 | 0.907235 |      0.0 |      0.0 |     0.0 | -41.346526 |      0.0 | -46.045135 |      0.0 |  -88.752255 |      1.0 |      0.3 | 2.521971 | 0.864298 |  0.0 |  0.0 | 0.0 |  0.0 | 0.864298 | 0.046354 | 0.012876 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 18.420161 | 26.340831 |        0.0 | 0.996527 |  -50.460638 | 74.244674 |      0.5 | 0.364298 |   0.0 |   0.0 |  0.0 | 0.019234 | 0.027121 |   0.0 |   0.0 |  0.0 |   0.0 | 0.012876 |
    | 2010-12-11 04:00:00 |  0.5 |  4.8 |              0.0 |                      0.0 |             0.0 |             99.0 |       7.0 |              1005.0 | 0.0 | 0.0 | -1.129167 |             96.833333 |                   0.1 |                       7.41713 |            17.026026 |  0.47 |  4.8 | -1.129167 |    5.611139 |         2.892408 |           5.611139 |          7.0 |                 8.607627 |                      5.622664 |                         0.0 |                       6.1078 |                      0.602876 |                           0.414103 |             8.521551 |                  5.444613 |     996.478449 |   1.255555 | 8.583333 | -0.605528 |  -3.633166 |                   0.0 |                      0.0 |                       0.0 |                   4.389334 |                 14.261536 |                     0.0 |                -4.039084 |   -14.261536 |             0.0 |         4.039084 |         -9.872203 |             30.429381 |                 100.0 |               108.121425 |               98.814229 |     0.47 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.000598 |      0.0 |      0.0 |  -0.06294 |      0.0 |          0.0 |     0.742199 | 1.074993 | 1.074993 |      0.0 |      0.0 |     0.0 | -46.635995 |      0.0 | -52.693545 |      0.0 | -103.368624 |      1.0 |      0.3 | 2.010913 | 0.690995 |  0.0 |  0.0 | 0.0 |  0.0 | 0.690995 | 0.083522 | 0.023201 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 17.386614 | 24.862858 |        0.0 | 1.003835 |  -38.244139 | 75.564592 |      0.5 | 0.190995 |   0.0 |   0.0 |  0.0 | 0.031652 |  0.05187 |   0.0 |   0.0 |  0.0 |   0.0 | 0.023201 |
    | 2010-12-11 05:00:00 |  0.2 |  4.8 |              0.0 |                      0.0 |             0.0 |             97.6 |       7.3 |              1004.0 | 0.0 | 0.0 | -0.583333 |             97.016667 |                   0.1 |                       7.41713 |            17.026026 | 0.188 |  4.8 | -0.583333 |    5.851616 |         3.059407 |           5.851616 |          7.3 |                 8.607627 |                      5.852751 |                         0.0 |                       6.1078 |                      0.602876 |                           0.429037 |             8.401044 |                  5.678144 |     995.598956 |   1.254358 | 8.583333 |  -0.45893 |  -2.753578 |                   0.0 |                      0.0 |                       0.0 |                   4.399059 |                 14.099661 |                     0.0 |                -3.389325 |   -14.099661 |             0.0 |         3.389325 |         -9.700602 |             29.178858 |                 100.0 |               106.264665 |               98.814229 |    0.188 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.007217 |      0.0 |      0.0 | -0.061894 |      0.0 |          0.0 |     0.741627 | 1.089019 | 1.089019 |      0.0 |      0.0 |     0.0 | -45.860636 |      0.0 | -54.540411 |      0.0 | -103.790373 |      1.0 |      0.3 | 1.749174 | 0.602414 |  0.0 |  0.0 | 0.0 |  0.0 | 0.602414 | 0.104969 | 0.029158 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 16.338166 | 23.363577 |        0.0 | 1.011143 |  -26.907227 | 76.711353 |      0.5 | 0.102414 |   0.0 |   0.0 |  0.0 |  0.04375 | 0.061219 |   0.0 |   0.0 |  0.0 |   0.0 | 0.029158 |
    | 2010-12-11 06:00:00 |  0.5 |  5.1 |              0.0 |                      0.0 |             0.0 |             97.4 |       7.2 |              1004.0 | 0.0 | 0.0 |   -0.0125 |             97.054167 |                   0.1 |                       7.41713 |            17.026026 |  0.47 |  5.1 |   -0.0125 |    5.771457 |         3.229745 |           5.771457 |          7.2 |                 8.790172 |                      6.102233 |                         0.0 |                       6.1078 |                      0.614118 |                           0.445147 |             8.561627 |                  5.922472 |     995.438373 |    1.25293 | 8.583333 | -0.322887 |   -1.93732 |                   0.0 |                      0.0 |                       0.0 |                   4.408714 |                 13.937223 |                     0.0 |                -5.587229 |   -13.937223 |             0.0 |         5.587229 |         -9.528509 |              29.58412 |                 100.0 |               104.661465 |               98.814229 |     0.47 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.00818 |      0.0 |      0.0 |  -0.06548 |      0.0 |          0.0 |      0.74106 | 1.179842 | 1.179842 |      0.0 |      0.0 |     0.0 |  -48.51811 |      0.0 | -57.295089 |      0.0 | -111.400428 |      1.0 |      0.3 | 2.161902 | 0.746444 |  0.0 |  0.0 | 0.0 |  0.0 | 0.746444 | 0.128131 | 0.035592 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 15.200808 | 21.737155 |        0.0 | 1.018451 |  -16.386574 | 78.126812 |      0.5 | 0.246444 |   0.0 |   0.0 |  0.0 | 0.055536 | 0.072595 |   0.0 |   0.0 |  0.0 |   0.0 | 0.035592 |
    | 2010-12-11 07:00:00 |  0.2 |  5.1 |              0.0 |                      0.0 |             0.0 |             96.3 |       7.6 |              1004.0 | 0.0 | 0.0 |  0.554167 |             97.070833 |                   0.1 |                       7.41713 |            17.026026 | 0.188 |  5.1 |  0.554167 |    6.092093 |         3.436822 |           6.092093 |          7.6 |                 8.790172 |                      6.359135 |                         0.0 |                       6.1078 |                      0.614118 |                            0.46165 |             8.464935 |                  6.172866 |     995.535065 |   1.252976 | 8.583333 | -0.196639 |  -1.179833 |                   0.0 |                      0.0 |                       0.0 |                     4.4183 |                 13.773355 |                     0.0 |                -5.066448 |   -13.773355 |             0.0 |         5.066448 |         -9.355055 |             28.027061 |                 100.0 |               102.694682 |               98.814229 |    0.188 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.014147 |      0.0 |      0.0 | -0.065772 |      0.0 |          0.0 |     0.740497 | 1.212927 | 1.212927 |      0.0 |      0.0 |     0.0 | -48.734564 |      0.0 | -59.911482 |      0.0 | -113.712494 |      1.0 |      0.3 | 1.927734 | 0.667238 |  0.0 |  0.0 | 0.0 |  0.0 | 0.667238 | 0.152885 | 0.042468 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 14.030205 | 20.063193 |        0.0 | 1.025759 |   -6.623407 | 79.387307 |      0.5 | 0.167238 |   0.0 |   0.0 |  0.0 | 0.067016 | 0.085869 |   0.0 |   0.0 |  0.0 |   0.0 | 0.042468 |
    | 2010-12-11 08:00:00 |  0.1 |  5.4 |              0.0 |                 0.347041 |        0.900733 |             95.2 |       7.3 |              1003.0 | 0.0 | 0.0 |     1.175 |             97.004167 |                   0.1 |                       7.40894 |            17.017406 | 0.094 |  5.4 |     1.175 |    5.851616 |           3.6439 |           5.851616 |          7.3 |                 8.976116 |                       6.65151 |                         0.0 |                       6.1078 |                       0.62554 |                           0.480332 |             8.545263 |                  6.452241 |     994.454737 |   1.250338 | 8.583333 | -0.079481 |  -0.476885 |              0.234246 |                      0.0 |                  0.234246 |                   4.425575 |                  13.59621 |                     0.0 |                -6.835277 |   -13.361964 |             0.0 |         7.069523 |         -9.170635 |             29.178858 |                 100.0 |               100.953824 |                99.25531 |    0.094 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.01901 |      0.0 |      0.0 | -0.065786 |      0.0 |          0.0 |     0.739938 | 1.257794 | 1.257794 |      0.0 |      0.0 |     0.0 | -48.744743 |      0.0 | -61.357963 |      0.0 | -117.172228 | 0.999672 | 0.299803 | 1.898543 | 0.658341 |  0.0 |  0.0 | 0.0 |  0.0 | 0.658341 | 0.171707 | 0.047696 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 12.814291 | 18.324436 |        0.0 | 1.033067 |    2.436812 | 80.627509 |      0.5 | 0.158341 |   0.0 |   0.0 |  0.0 | 0.078201 | 0.093506 |   0.0 |   0.0 |  0.0 |   0.0 | 0.047696 |
    | 2010-12-11 09:00:00 |  0.3 |  5.3 |              0.0 |                      1.0 |       17.267511 |             96.3 |       7.1 |              1003.0 | 0.0 | 0.0 |  1.779167 |               97.0875 |                   0.1 |                       7.40894 |            16.959252 | 0.282 |  5.3 |  1.779167 |    5.691298 |         3.847638 |           5.691298 |          7.1 |                 8.913754 |                      6.947358 |                         0.0 |                       6.1078 |                      0.621713 |                            0.49913 |             8.583945 |                  6.745016 |     994.416055 |   1.250768 | 8.583333 |       0.0 |        0.0 |                4.5002 |                      0.0 |                    4.5002 |                   4.419863 |                 13.407231 |                     0.0 |                -6.596815 |    -8.907031 |             0.0 |        11.097015 |         -8.987368 |             30.000798 |                 100.0 |                99.250199 |               99.923397 |    0.282 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.016416 |      0.0 |      0.0 | -0.065321 |      0.0 |          0.0 |     0.739383 | 1.275733 | 1.275733 |      0.0 |      0.0 |     0.0 | -48.400444 |      0.0 | -58.862205 |      0.0 | -118.359665 | 0.998515 |  0.29911 | 2.112801 | 0.733141 |  0.0 |  0.0 | 0.0 |  0.0 | 0.733141 | 0.192828 | 0.053563 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  11.57969 | 16.558957 |        0.0 | 1.040375 |   11.020145 | 82.007169 |      0.5 | 0.233141 |   0.0 |   0.0 |  0.0 | 0.089096 | 0.103732 |   0.0 |   0.0 |  0.0 |   0.0 | 0.053563 |
    | 2010-12-11 10:00:00 |  0.1 |  5.5 |              0.0 |                      1.0 |       52.617611 |             95.5 |       7.1 |              1003.0 | 0.0 | 0.0 |  2.295833 |               97.0375 |                   0.1 |                       7.40894 |            16.807156 | 0.094 |  5.5 |  2.295833 |    5.691298 |         4.038016 |           5.691298 |          7.1 |                 9.038862 |                      7.209503 |                         0.0 |                       6.1078 |                      0.629387 |                           0.515703 |             8.632113 |                  6.995922 |     994.367887 |   1.249848 | 8.583333 |       0.0 |        0.0 |             13.742015 |                      0.0 |                 13.742015 |                   4.389484 |                 13.251022 |                     0.0 |                -7.748872 |     0.490992 |             0.0 |        21.490887 |         -8.861538 |             30.000798 |                 100.0 |                97.365159 |               99.726142 |    0.094 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.026537 |      0.0 |      0.0 | -0.066592 |      0.0 |          0.0 |     0.738832 | 1.421852 | 1.421852 |      0.0 |      0.0 |     0.0 | -49.341985 |      0.0 | -61.083421 |      0.0 | -131.916292 | 0.997359 | 0.298417 | 2.135425 | 0.741587 |  0.0 |  0.0 | 0.0 |  0.0 | 0.741587 | 0.216731 | 0.060203 |     0.06 |   0.0 |  0.0 |      0.0 |   nan | 10.198688 | 14.584123 |        0.0 | 1.047683 |   19.603478 | 83.401008 |      0.5 | 0.241587 |   0.0 |   0.0 |  0.0 |  0.09971 | 0.117021 |   0.0 |   0.0 |  0.0 |   0.0 | 0.060203 |
    | 2010-12-11 11:00:00 |  0.1 |  6.1 |              0.0 |                      1.0 |       80.135633 |             91.6 |       8.9 |              1002.0 | 0.0 | 0.0 |  2.716667 |             96.795833 |                   0.0 |                       7.40894 |            15.680798 | 0.094 |  6.1 |  2.716667 |    7.134162 |         4.291853 |           7.134162 |          8.9 |                 9.423511 |                       7.42943 |                         0.0 |                       6.1078 |                      0.652898 |                           0.529548 |             8.631936 |                  7.191379 |     993.368064 |   1.245915 | 8.583333 |       0.0 |        0.0 |             20.972669 |                      0.0 |                 20.972669 |                   4.103894 |                 12.463959 |                     0.0 |               -12.026271 |     8.508711 |             0.0 |         32.99894 |         -8.360064 |             23.933221 |                 100.0 |                95.471011 |               99.520918 |    0.094 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.060214 |      0.0 |      0.0 | -0.080429 |      0.0 |          0.0 |     0.738285 | 1.880005 | 1.880005 |      0.0 |      0.0 |     0.0 |  -59.59435 |      0.0 | -81.829421 |      0.0 | -174.422711 | 0.996202 | 0.297726 | 2.797234 |   0.9728 |  0.0 |  0.0 | 0.0 |  0.0 |   0.9728 | 0.251142 | 0.069762 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  8.364558 | 11.961318 |        0.0 | 1.054991 |   28.186812 | 85.225441 |      0.5 |   0.4728 |   0.0 |   0.0 |  0.0 |  0.11005 | 0.141092 |   0.0 |   0.0 |  0.0 |   0.0 | 0.069762 |
    | 2010-12-11 12:00:00 |  0.0 |  6.4 |              0.0 |                      1.0 |       80.276117 |             90.0 |       9.5 |              1002.0 | 0.0 | 0.0 |       3.1 |             96.633333 |                   0.0 |                       7.40894 |            15.214247 |   0.0 |  6.4 |       3.1 |    7.615117 |          4.56573 |           7.615117 |          9.5 |                 9.621181 |                      7.634886 |                         0.0 |                       6.1078 |                      0.664931 |                           0.542435 |             8.659063 |                  7.377844 |     993.340937 |   1.244565 | 8.583333 |       0.0 |        0.0 |              21.05304 |                      0.0 |                  21.05304 |                   3.990055 |                 12.359437 |                     0.0 |               -13.533625 |     8.693604 |             0.0 |        34.586665 |         -8.369381 |             22.421649 |                 100.0 |                93.006346 |               99.242721 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.074086 |      0.0 |      0.0 | -0.085956 |      0.0 |          0.0 |     0.737742 | 2.037725 | 2.037725 |      0.0 |      0.0 |     0.0 | -63.690154 |      0.0 | -90.778796 |      0.0 | -189.055614 | 0.995046 | 0.297035 | 2.934887 | 1.022519 |  0.0 |  0.0 | 0.0 |  0.0 | 1.022519 | 0.296696 | 0.082416 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  6.372299 |  9.112388 |        0.0 | 1.062299 |   36.770145 |  87.13781 |      0.5 | 0.522519 |   0.0 |   0.0 |  0.0 | 0.120123 | 0.176574 |   0.0 |   0.0 |  0.0 |   0.0 | 0.082416 |
    | 2010-12-11 13:00:00 |  0.1 |  6.4 |              0.0 |                      1.0 |       75.835403 |             85.7 |       7.9 |              1002.0 | 0.0 | 0.0 |     3.425 |               96.3125 |                   0.0 |                       7.40894 |            15.020908 | 0.094 |  6.4 |     3.425 |    6.332571 |         4.782828 |           6.332571 |          7.9 |                 9.621181 |                      7.812981 |                         0.0 |                       6.1078 |                      0.664931 |                           0.553571 |             8.245352 |                  7.524878 |     993.754648 |    1.24476 | 8.583333 |       0.0 |        0.0 |              19.92933 |                      0.0 |                  19.92933 |                   3.947452 |                 12.283332 |                     0.0 |               -11.239491 |     7.645998 |             0.0 |        31.168822 |          -8.33588 |             26.962743 |                 100.0 |                90.439516 |               98.938663 |    0.094 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.085524 |      0.0 |      0.0 | -0.061599 |      0.0 |          0.0 |     0.737203 |   1.6648 |   1.6648 |      0.0 |      0.0 |     0.0 | -45.642119 |      0.0 | -77.645525 |      0.0 | -154.456466 |  0.99389 | 0.296345 | 2.490757 | 0.869071 |  0.0 |  0.0 | 0.0 |  0.0 | 0.869071 | 0.333133 | 0.092537 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  4.739321 |  6.777229 |        0.0 | 1.069607 |   45.353478 | 88.759496 |      0.5 | 0.369071 |   0.0 |   0.0 |  0.0 | 0.129935 | 0.203198 |   0.0 |   0.0 |  0.0 |   0.0 | 0.092537 |
    | 2010-12-11 14:00:00 |  0.0 |  6.5 |              0.0 |                      1.0 |       42.227911 |             82.3 |       9.6 |              1002.0 | 0.0 | 0.0 |    3.7125 |             95.854167 |                   0.0 |                       7.40894 |            14.983713 |   0.0 |  6.5 |    3.7125 |    7.695276 |         5.026645 |           7.695276 |          9.6 |                 9.687876 |                      7.973569 |                         0.0 |                       6.1078 |                      0.668985 |                           0.563584 |             7.973122 |                  7.642998 |     994.026878 |   1.244443 | 8.583333 |       0.0 |        0.0 |             11.119991 |                      0.0 |                 11.119991 |                   3.945702 |                 12.228543 |                     0.0 |               -10.124562 |    -1.108552 |             0.0 |        21.244554 |         -8.282841 |              22.18809 |                 100.0 |                88.275274 |               98.670112 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.116361 |      0.0 |      0.0 | -0.063414 |      0.0 |          0.0 |     0.736667 | 1.738161 | 1.738161 |      0.0 |      0.0 |     0.0 | -46.987277 |      0.0 | -93.030869 |      0.0 | -161.262701 | 0.992733 | 0.295656 | 2.508827 | 0.876567 |  0.0 |  0.0 | 0.0 |  0.0 | 0.876567 | 0.359557 | 0.099877 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  3.029242 |  4.331816 |        0.0 | 1.076915 |   53.936812 | 90.391756 |      0.5 | 0.376567 |   0.0 |   0.0 |  0.0 | 0.139494 | 0.220063 |   0.0 |   0.0 |  0.0 |   0.0 | 0.099877 |
    | 2010-12-11 15:00:00 |  0.1 |  6.3 |              0.1 |                      1.0 |       12.613919 |             82.5 |       9.1 |              1003.0 | 0.0 | 0.0 |  3.954167 |             95.370833 |                   0.1 |                       7.40894 |            15.079289 | 0.094 |  6.3 |  3.954167 |     7.29448 |         5.237063 |            7.29448 |          9.1 |                 9.554889 |                      8.110798 |                         0.0 |                       6.1078 |                      0.660899 |                           0.572121 |             7.882784 |                  7.735336 |     995.117216 |   1.246623 | 8.583333 |       0.0 |        0.0 |              3.328366 |                      0.0 |                  3.328366 |                    3.97889 |                 12.847805 |                     0.0 |                -7.130286 |    -9.519439 |             0.0 |        10.458652 |         -8.868916 |             23.407216 |                 100.0 |                86.107697 |               98.389186 |    0.094 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.102979 |      0.0 |      0.0 | -0.057639 |      0.0 |          0.0 |     0.736135 | 1.501387 | 1.501387 |      0.0 |      0.0 |     0.0 |  -42.70834 |      0.0 | -86.128361 |      0.0 | -139.295353 | 0.991577 | 0.294967 | 2.269551 | 0.793895 |  0.0 |  0.0 | 0.0 |  0.0 | 0.793895 | 0.380244 | 0.105623 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  1.548184 |  2.213904 |        0.0 | 1.084223 |   62.520145 | 91.867412 |      0.5 | 0.293895 |   0.0 |   0.0 |  0.0 | 0.148807 | 0.231438 |   0.0 |   0.0 |  0.0 |   0.0 | 0.105623 |
    | 2010-12-11 16:00:00 |  0.0 |  5.9 |              0.0 |                 0.053708 |        0.021911 |             82.2 |       8.7 |              1003.0 | 0.0 | 0.0 |  4.158333 |             94.854167 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  5.9 |  4.158333 |    6.973844 |         5.424101 |           6.973844 |          8.7 |                 9.293725 |                       8.22835 |                         0.0 |                       6.1078 |                      0.644979 |                           0.579419 |             7.639442 |                  7.804933 |     995.360558 |   1.248525 | 8.583333 |       0.0 |        0.0 |              0.005793 |                      0.0 |                  0.005793 |                   3.986784 |                 12.822679 |                     0.0 |                -3.915548 |   -12.816886 |             0.0 |         3.921341 |         -8.835895 |              24.48341 |                 100.0 |                84.156849 |                98.77699 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.096543 |      0.0 |      0.0 | -0.047871 |      0.0 |          0.0 |     0.735607 | 1.261344 | 1.261344 |      0.0 |      0.0 |     0.0 | -35.470286 |      0.0 | -77.633085 |      0.0 | -117.024713 |  0.99042 |  0.29428 | 1.842724 |  0.64505 |  0.0 |  0.0 | 0.0 |  0.0 |  0.64505 | 0.387998 | 0.107777 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  0.293043 |  0.419051 |        0.0 | 1.091531 |   71.103478 | 93.065086 |      0.5 |  0.14505 |   0.0 |   0.0 |  0.0 | 0.157878 |  0.23012 |   0.0 |   0.0 |  0.0 |   0.0 | 0.107777 |
    | 2010-12-11 17:00:00 |  0.0 |  5.5 |              0.0 |                      0.0 |             0.0 |             76.5 |       8.9 |              1003.0 | 0.0 | 0.0 |  4.341667 |             94.108333 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  5.5 |  4.341667 |    7.134162 |         5.604459 |           7.134162 |          8.9 |                 9.038862 |                      8.335182 |                         0.0 |                       6.1078 |                      0.629387 |                            0.58604 |              6.91473 |                  7.844101 |      996.08527 |   1.250659 | 8.583333 |       0.0 |        0.0 |                   0.0 |                      0.0 |                       0.0 |                   3.994692 |                   12.8228 |                     0.0 |                 2.340465 |     -12.8228 |             0.0 |        -2.340465 |         -8.828108 |             23.933221 |                 100.0 |                82.579301 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.129358 |     0.06 | 0.021721 | -0.025712 |      0.0 |          0.0 |     0.735083 | 0.975359 | 0.293043 |      0.0 |      0.0 |     0.0 | -19.051448 |      0.0 | -73.780626 |      0.0 |  -90.491608 | 0.989264 | 0.293593 | 0.444763 |  0.15558 |  0.0 |  0.0 | 0.0 |  0.0 |  0.15558 | 0.376578 | 0.104605 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |        0.0 | 1.098839 |   79.686812 | 93.332548 |  0.15558 |      0.0 |   0.0 |   0.0 |  0.0 | 0.162248 | 0.214331 |   0.0 |   0.0 |  0.0 |   0.0 | 0.104605 |
    | 2010-12-11 18:00:00 |  0.0 |  5.1 |              0.0 |                      0.0 |             0.0 |             81.7 |       7.9 |              1003.0 | 0.0 | 0.0 |    4.5125 |                93.475 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  5.1 |    4.5125 |    6.332571 |         5.721357 |           6.332571 |          7.9 |                 8.790172 |                      8.435829 |                         0.0 |                          0.0 |                      0.614118 |                           0.592267 |              7.18157 |                  7.885391 |      995.81843 |   1.252331 | 8.583333 |       0.0 |      -30.6 |                   0.0 |                      0.0 |                       0.0 |                  13.571128 |                 12.818218 |                     0.0 |                      0.0 |   -12.818218 |             0.0 |              0.0 |           0.75291 |             26.962743 |                 100.0 |                82.227691 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.087608 |      0.0 | 0.029418 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.983985 | 0.290468 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.353086 | 0.098079 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |        0.0 |      nan |  118.870145 |  93.30313 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.160057 | 0.193029 |   0.0 |   0.0 |  0.0 |   0.0 | 0.098079 |
    | 2010-12-11 19:00:00 |  0.0 |  4.8 |              0.0 |                      0.0 |             0.0 |             83.4 |       7.1 |              1003.0 | 0.0 | 0.0 |  4.654167 |             92.854167 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  4.8 |  4.654167 |    5.691298 |         5.831576 |           5.691298 |          7.1 |                 8.607627 |                      8.520102 |                         0.0 |                          0.0 |                      0.602876 |                           0.597475 |             7.178761 |                   7.91127 |     995.821239 |   1.253684 | 8.583333 |       0.0 |      -28.8 |                   0.0 |                      0.0 |                       0.0 |                  13.571128 |                 12.822742 |                     0.0 |                      0.0 |   -12.822742 |             0.0 |              0.0 |          0.748386 |             30.000798 |                 100.0 |                82.266353 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.070419 |      0.0 | 0.025143 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.978948 | 0.287502 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.329767 | 0.091602 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |        0.0 |      nan |  156.253478 | 93.277987 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.155923 | 0.173844 |   0.0 |   0.0 |  0.0 |   0.0 | 0.091602 |
    | 2010-12-11 20:00:00 |  0.0 |  4.9 |              0.0 |                      0.0 |             0.0 |             82.8 |       6.4 |              1003.0 | 0.0 | 0.0 |  4.791667 |             92.191667 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  4.9 |  4.791667 |    5.130184 |         5.854956 |           5.130184 |          6.4 |                 8.668101 |                      8.602604 |                         0.0 |                          0.0 |                      0.606604 |                           0.602566 |             7.177187 |                  7.930884 |     995.822813 |   1.253234 | 8.583333 |       0.0 |      -29.4 |                   0.0 |                      0.0 |                       0.0 |                  13.571128 |                   12.8326 |                     0.0 |                      0.0 |     -12.8326 |             0.0 |              0.0 |          0.738528 |             33.282135 |                 100.0 |                82.299398 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.066191 |      0.0 | 0.025199 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 |  0.97383 | 0.284504 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.308461 | 0.085684 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |        0.0 |      nan |  194.236812 | 93.252787 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.151895 | 0.156566 |   0.0 |   0.0 |  0.0 |   0.0 | 0.085684 |
    | 2010-12-11 21:00:00 |  0.0 |  4.9 |              0.0 |                      0.0 |             0.0 |             85.8 |       7.2 |              1003.0 | 0.0 | 0.0 |    4.9125 |             91.616667 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  4.9 |    4.9125 |    5.771457 |         5.908395 |           5.771457 |          7.2 |                 8.668101 |                      8.675686 |                         0.0 |                          0.0 |                      0.606604 |                           0.607071 |              7.43723 |                  7.948375 |      995.56277 |   1.253111 | 8.583333 |       0.0 |      -29.4 |                   0.0 |                      0.0 |                       0.0 |                  13.571128 |                 12.841028 |                     0.0 |                      0.0 |   -12.841028 |             0.0 |              0.0 |          0.730101 |              29.58412 |                 100.0 |                 82.33252 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.060894 |      0.0 | 0.021594 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.968713 | 0.281521 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.288977 | 0.080271 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |        0.0 |      nan |  232.220145 | 93.231194 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.147972 | 0.141005 |   0.0 |   0.0 |  0.0 |   0.0 | 0.080271 |
    | 2010-12-11 22:00:00 |  0.1 |  4.7 |              0.0 |                      0.0 |             0.0 |             87.6 |       7.0 |              1003.0 | 0.0 | 0.0 |  5.008333 |                91.125 |                   0.1 |                      7.400749 |            15.079031 | 0.094 |  4.7 |  5.008333 |    5.611139 |         5.948475 |           5.611139 |          7.0 |                 8.547525 |                      8.734036 |                         0.0 |                          0.0 |                      0.599168 |                           0.610664 |             7.487632 |                   7.95889 |     995.512368 |   1.253989 | 8.583333 |       0.0 |      -28.2 |                   0.0 |                      0.0 |                       0.0 |                  13.571128 |                 12.851041 |                     0.0 |                      0.0 |   -12.851041 |             0.0 |              0.0 |          0.720087 |             30.429381 |                 100.0 |                82.360904 |               98.814229 |    0.034 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.050931 | 0.050931 |      0.0 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.963757 | 0.278648 |    0.034 | 0.011425 |  0.0 |  0.0 | 0.0 |  0.0 | 0.011425 | 0.271288 | 0.075358 | 0.009069 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |        0.0 |      nan |  269.003478 | 93.253769 | 0.011425 |      0.0 |   0.0 |   0.0 |  0.0 | 0.144298 | 0.126991 |   0.0 |   0.0 |  0.0 |   0.0 | 0.075358 |
    | 2010-12-11 23:00:00 |  0.1 |  3.8 |              0.0 |                      0.0 |             0.0 |             85.6 |       6.8 |              1002.0 | 0.0 | 0.0 |    5.0625 |             90.558333 |                   0.1 |                      7.400749 |            15.079031 | 0.094 |  3.8 |    5.0625 |     5.45082 |         5.968515 |            5.45082 |          6.8 |                 8.023017 |                      8.767169 |                         0.0 |                          0.0 |                      0.566662 |                           0.612703 |             6.867703 |                  7.939402 |     995.132297 |   1.257101 | 8.583333 |       0.0 |      -22.8 |                   0.0 |                      0.0 |                       0.0 |                  13.571128 |                 12.881907 |                     0.0 |                      0.0 |   -12.881907 |             0.0 |              0.0 |          0.689221 |             31.324363 |                 100.0 |                 82.33123 |               98.814229 | 0.043069 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.055782 | 0.055782 |      0.0 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.959529 | 0.276209 | 0.043069 | 0.014377 |  0.0 |  0.0 | 0.0 |  0.0 | 0.014377 | 0.255273 | 0.070909 | 0.004218 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |        0.0 |      nan |  300.386812 | 93.282461 | 0.014377 |      0.0 |   0.0 |   0.0 |  0.0 | 0.140904 | 0.114369 |   0.0 |   0.0 |  0.0 |   0.0 | 0.070909 |
    | 2010-12-12 00:00:00 |  0.0 |  3.9 |              0.0 |                      0.0 |             0.0 |             85.3 |       6.4 |              1002.0 | 0.0 | 0.0 |    5.1125 |             89.979167 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  3.9 |    5.1125 |    5.130184 |         5.995234 |           5.130184 |          6.4 |                  8.07986 |                      8.797851 |                         0.0 |                          0.0 |                      0.570198 |                           0.614591 |             6.892121 |                  7.916233 |     995.107879 |   1.256635 | 8.583333 |       0.0 |      -23.4 |                   0.0 |                      0.0 |                       0.0 |                  13.571128 |                 12.915622 |                     0.0 |                      0.0 |   -12.915622 |             0.0 |              0.0 |          0.655506 |             33.282135 |                 100.0 |                82.293518 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.05385 | 0.004218 | 0.018582 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 |  0.95522 | 0.273733 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.240451 | 0.066792 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |        0.0 |      nan |  332.370145 | 93.263879 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.137449 | 0.103002 |   0.0 |   0.0 |  0.0 |   0.0 | 0.066792 |
    | 2010-12-12 01:00:00 |  0.0 |  3.7 |              0.0 |                      0.0 |             0.0 |             86.2 |       6.3 |              1002.0 | 0.0 | 0.0 |  5.133333 |             89.420833 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  3.7 |  5.133333 |    5.050025 |         6.031974 |           5.050025 |          6.3 |                 7.966527 |                      8.810663 |                         0.0 |                          0.0 |                      0.563146 |                           0.615379 |             6.867146 |                  7.878569 |     995.132854 |   1.257555 | 8.583333 |       0.0 |      -22.2 |                   0.0 |                      0.0 |                       0.0 |                  13.571128 |                 12.957663 |                     0.0 |                      0.0 |   -12.957663 |             0.0 |              0.0 |          0.613465 |             33.810423 |                 100.0 |                82.317941 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.049164 |      0.0 | 0.018518 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.951072 | 0.271361 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.226664 | 0.062962 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |        0.0 |      nan |  363.153478 | 93.245361 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.133899 | 0.092765 |   0.0 |   0.0 |  0.0 |   0.0 | 0.062962 |
    | 2010-12-12 02:00:00 |  0.1 |  3.5 |              0.0 |                      0.0 |             0.0 |             89.3 |       5.7 |              1002.0 | 0.0 | 0.0 |  5.116667 |             88.983333 |                   0.1 |                      7.400749 |            15.079031 | 0.094 |  3.5 |  5.116667 |     4.56907 |         6.031974 |            4.56907 |          5.7 |                 7.854597 |                      8.800412 |                         0.0 |                          0.0 |                      0.556168 |                           0.614748 |             7.014155 |                    7.8309 |     994.985845 |   1.258394 | 8.583333 |       0.0 |      -21.0 |                   0.0 |                      0.0 |                       0.0 |                  13.571128 |                 13.001862 |                     0.0 |                      0.0 |   -13.001862 |             0.0 |              0.0 |          0.569266 |             37.369415 |                 100.0 |                82.342282 |               98.814229 |    0.034 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.033549 | 0.033549 |      0.0 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.947086 | 0.269092 |    0.034 | 0.011126 |  0.0 |  0.0 | 0.0 |  0.0 | 0.011126 | 0.214129 |  0.05948 | 0.026451 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |        0.0 |      nan |  392.736812 | 93.268234 | 0.011126 |      0.0 |   0.0 |   0.0 |  0.0 | 0.130584 | 0.083545 |   0.0 |   0.0 |  0.0 |   0.0 |  0.05948 |
    | 2010-12-12 03:00:00 |  0.0 |  3.3 |              0.0 |                      0.0 |             0.0 |             91.4 |       5.2 |              1003.0 | 0.0 | 0.0 |  5.070833 |             88.633333 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  3.3 |  5.070833 |    4.168274 |         5.985215 |           4.168274 |          5.2 |                 7.744054 |                      8.772276 |                         0.0 |                          0.0 |                      0.549265 |                           0.613018 |             7.078066 |                  7.775161 |     995.921934 |   1.260534 | 8.583333 |       0.0 |      -19.8 |                   0.0 |                      0.0 |                       0.0 |                  13.571128 |                 13.048148 |                     0.0 |                      0.0 |   -13.048148 |             0.0 |              0.0 |           0.52298 |             40.962628 |                 100.0 |                82.312216 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.023802 | 0.023802 |      0.0 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.943262 | 0.266923 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.202596 | 0.056277 | 0.002648 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |        0.0 |      nan |  421.120145 | 93.268234 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.127354 | 0.075242 |   0.0 |   0.0 |  0.0 |   0.0 | 0.056277 |
    | 2010-12-12 04:00:00 |  0.0 |  3.3 |              0.0 |                      0.0 |             0.0 |             90.9 |       4.6 |              1004.0 | 0.0 | 0.0 |  5.008333 |             88.295833 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  3.3 |  5.008333 |     3.68732 |         5.905055 |            3.68732 |          4.6 |                 7.744054 |                      8.734036 |                         0.0 |                          0.0 |                      0.549265 |                           0.610664 |             7.039346 |                   7.71179 |     996.960654 |   1.261813 | 8.583333 |       0.0 |      -19.8 |                   0.0 |                      0.0 |                       0.0 |                  13.571128 |                 13.098861 |                     0.0 |                      0.0 |   -13.098861 |             0.0 |              0.0 |          0.472267 |              46.30558 |                 100.0 |                82.312216 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.022297 | 0.002648 | 0.008766 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.939438 | 0.264763 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.191828 | 0.053286 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |        0.0 |      nan |  449.503478 | 93.259468 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.124065 | 0.067764 |   0.0 |   0.0 |  0.0 |   0.0 | 0.053286 |
    | 2010-12-12 05:00:00 |  0.0 |  3.0 |              0.0 |                      0.0 |             0.0 |             89.2 |       4.9 |              1005.0 | 0.0 | 0.0 |  4.933333 |             87.945833 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  3.0 |  4.933333 |    3.927797 |         5.824896 |           3.927797 |          4.9 |                 7.580812 |                      8.688342 |                         0.0 |                          0.0 |                      0.539048 |                           0.607851 |             6.762084 |                  7.641034 |     998.237916 |   1.264577 | 8.583333 |       0.0 |      -18.0 |                   0.0 |                      0.0 |                       0.0 |                  13.571128 |                 13.154767 |                     0.0 |                      0.0 |   -13.154767 |             0.0 |              0.0 |          0.416361 |             43.470544 |                 100.0 |                82.323739 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.028419 |      0.0 |  0.01221 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.935857 | 0.262748 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.181888 | 0.050525 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |        0.0 |      nan |  476.086812 | 93.247257 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |  0.12086 | 0.061029 |   0.0 |   0.0 |  0.0 |   0.0 | 0.050525 |
    | 2010-12-12 06:00:00 |  0.0 |  1.9 |              0.0 |                      0.0 |             0.0 |             85.6 |       6.3 |              1006.0 | 0.0 | 0.0 |       4.8 |             87.454167 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  1.9 |       4.8 |    5.050025 |         5.794837 |           5.050025 |          6.3 |                 7.007901 |                      8.607627 |                         0.0 |                          0.0 |                      0.502964 |                           0.602876 |             5.998763 |                  7.527728 |    1000.001237 |   1.271267 | 8.583333 |       0.0 |      -11.4 |                   0.0 |                      0.0 |                       0.0 |                  13.571128 |                  13.24232 |                     0.0 |                      0.0 |    -13.24232 |             0.0 |              0.0 |          0.328808 |             33.810423 |                 100.0 |                82.339789 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.047698 |      0.0 | 0.017452 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.933164 | 0.261239 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.172701 | 0.047973 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |        0.0 |      nan |  496.070145 | 93.229805 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.117738 | 0.054963 |   0.0 |   0.0 |  0.0 |   0.0 | 0.047973 |
    | 2010-12-12 07:00:00 |  0.0 |  1.2 |              0.0 |                      0.0 |             0.0 |             86.3 |       5.8 |              1008.0 | 0.0 | 0.0 |    4.6375 |               87.0375 |                   0.1 |                      7.400749 |            15.079031 |   0.0 |  1.2 |    4.6375 |    4.649229 |         5.734717 |           4.649229 |          5.8 |                 6.663527 |                      8.510149 |                         0.0 |                          0.0 |                      0.481097 |                            0.59686 |             5.750624 |                  7.407021 |    1002.249376 |   1.277169 | 8.583333 |       0.0 |       -7.2 |                   0.0 |                      0.0 |                       0.0 |                  13.571128 |                 13.332335 |                     0.0 |                      0.0 |   -13.332335 |             0.0 |              0.0 |          0.238793 |             36.725115 |                 100.0 |                 82.36273 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.040382 |      0.0 | 0.015359 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.931038 | 0.260049 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.164197 |  0.04561 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |        0.0 |      nan |  511.853478 | 93.214446 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.114697 |   0.0495 |   0.0 |   0.0 |  0.0 |   0.0 |  0.04561 |
    | 2010-12-12 08:00:00 |  0.0 |  0.8 |              0.0 |                 0.339738 |        0.894144 |             91.2 |       6.1 |              1009.0 | 0.0 | 0.0 |  4.445833 |             86.870833 |                   0.1 |                      7.393446 |            15.078757 |   0.0 |  0.8 |  4.445833 |    4.889707 |         5.694638 |           4.889707 |          6.1 |                 6.473522 |                      8.396426 |                         0.0 |                          0.0 |                      0.468972 |                            0.58983 |             5.903852 |                  7.294045 |    1003.096148 |   1.280232 | 8.583333 |       0.0 |       -4.8 |               0.80473 |                      0.0 |                   0.80473 |                  13.570881 |                 13.409662 |                     0.0 |                      0.0 |   -12.604932 |             0.0 |              0.0 |          0.161219 |             34.918962 |                 100.0 |                 82.38292 |               98.493492 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.02629 |      0.0 | 0.009663 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.929235 | 0.259043 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.156315 | 0.043421 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |        0.0 |      nan |  525.236812 | 93.204783 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.111734 | 0.044581 |   0.0 |   0.0 |  0.0 |   0.0 | 0.043421 |
    | 2010-12-12 09:00:00 |  0.0 |  0.8 |              0.0 |                      1.0 |       17.712539 |             87.5 |       5.9 |              1010.0 | 0.0 | 0.0 |  4.258333 |             86.504167 |                   0.1 |                      7.393446 |              15.0973 |   0.0 |  0.8 |  4.258333 |    4.729388 |         5.654558 |           4.729388 |          5.9 |                 6.473522 |                      8.286472 |                         0.0 |                          0.0 |                      0.468972 |                           0.583022 |             5.664332 |                  7.168143 |    1004.335668 |   1.281619 | 8.583333 |       0.0 |       -4.8 |             15.941285 |                      0.0 |                 15.941285 |                   13.58757 |                 13.502128 |                     0.0 |                      0.0 |     2.439157 |             0.0 |              0.0 |          0.085442 |             36.102655 |                 100.0 |                82.395622 |               97.877895 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.045636 |      0.0 | 0.017181 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.927431 | 0.258039 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.148998 | 0.041388 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |        0.0 |      nan |  538.620145 | 93.187603 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.108848 |  0.04015 |   0.0 |   0.0 |  0.0 |   0.0 | 0.041388 |
    | 2010-12-12 10:00:00 |  0.1 |  0.5 |              0.0 |                      1.0 |       54.302586 |             84.5 |       6.4 |              1011.0 | 0.0 | 0.0 |      4.05 |             86.045833 |                   0.1 |                      7.393446 |            15.167507 | 0.094 |  0.5 |      4.05 |    5.130184 |         5.631178 |           5.130184 |          6.4 |                 6.334173 |                       8.16579 |                         0.0 |                     5.484759 |                      0.460051 |                           0.575537 |             5.352376 |                  7.026322 |    1005.647624 |   1.284447 | 8.583333 |       0.0 |   3.989026 |              5.556421 |                      0.0 |                  5.556421 |                    1.55199 |                 13.608639 |                     0.0 |                27.251025 |    -8.052217 |             0.0 |       -21.694604 |        -12.056649 |             33.282135 |                 100.0 |                82.418208 |               97.881128 |    0.034 | 0.01275 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.052426 |      0.0 |      0.0 |   0.00321 | 0.006889 |    -1.465853 |     0.897677 |      0.0 |      0.0 | 0.000099 | 0.000099 |     0.0 |   2.378474 |      0.0 | -20.068003 |      0.0 |    4.005075 | 0.926812 | 0.257694 | 0.015768 | 0.004994 |  0.0 |  0.0 | 0.0 |  0.0 | 0.004994 |  0.14226 | 0.039517 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  0.010587 |  0.015023 |        0.0 |  0.00743 |  543.214452 | 93.198376 | 0.004994 |      0.0 |   0.0 |   0.0 |  0.0 | 0.106101 | 0.036159 |   0.0 |   0.0 |  0.0 |   0.0 | 0.039517 |
    | 2010-12-12 11:00:00 |  0.0 |  0.3 |              0.0 |                      1.0 |       81.504147 |             76.7 |       6.8 |              1012.0 | 0.0 | 0.0 |  3.808333 |                85.425 |                   0.1 |                      7.393446 |            15.224528 |   0.0 |  0.3 |  3.808333 |     5.45082 |         5.561039 |            5.45082 |          6.8 |                  6.24275 |                      8.027741 |                         0.0 |                     5.263686 |                      0.454184 |                           0.566956 |             4.788189 |                  6.857697 |    1007.211811 |   1.286932 | 8.583333 |       0.0 |   5.507593 |              8.523311 |                      0.0 |                  8.523311 |                   1.592108 |                 13.739115 |                     0.0 |                29.958696 |    -5.215805 |             0.0 |       -21.435386 |        -12.147007 |             31.324363 |                 100.0 |                82.404044 |               97.879101 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.084311 |      0.0 |      0.0 |  0.012109 |      0.0 |    -2.021526 |     0.895425 |      0.0 |      0.0 | 0.000112 | 0.000112 |     0.0 |   8.972394 |      0.0 | -24.889809 |      0.0 |     5.51797 | 0.926398 | 0.257464 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |  0.13599 | 0.037775 |     0.06 |   0.0 |  0.0 |      0.0 |   nan |  0.002075 |  0.002913 |        0.0 | 0.014738 |  546.290193 | 93.198376 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.103424 | 0.032566 |   0.0 |   0.0 |  0.0 |   0.0 | 0.037775 |
    | 2010-12-12 12:00:00 |  0.0 | -0.1 |              0.0 |                      1.0 |       81.665294 |             74.8 |       7.0 |              1013.0 | 0.0 | 0.0 |    3.5375 |             84.791667 |                   0.1 |                      7.393446 |            15.282411 |   0.0 | -0.1 |    3.5375 |    5.611139 |          5.47754 |           5.611139 |          7.0 |                 6.063393 |                      7.875477 |                         0.0 |                      5.07808 |                      0.442644 |                           0.557471 |             4.535418 |                  6.677749 |    1008.464582 |   1.290215 | 8.583333 |       0.0 |   6.828335 |              8.721431 |                      0.0 |                  8.721431 |                   1.632082 |                 13.878543 |                     0.0 |                31.460569 |    -5.157112 |             0.0 |       -22.739138 |        -12.246461 |             30.429381 |                 100.0 |                82.404044 |               97.879101 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.092017 |     0.06 | 0.010692 |  0.002913 |      0.0 |    -2.504219 |     0.893205 |      0.0 |      0.0 | 0.000027 | 0.000027 |     0.0 |  10.484764 |      0.0 | -26.393086 |      0.0 |    6.830815 | 0.926162 | 0.257333 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.130082 | 0.036134 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |        0.0 | 0.022046 |  548.045191 | 93.187684 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.100753 | 0.029329 |   0.0 |   0.0 |  0.0 |   0.0 | 0.036134 |
    | 2010-12-12 13:00:00 |  0.0 | -0.4 |              0.0 |                      1.0 |       77.078958 |             69.7 |       6.3 |              1014.0 | 0.0 | 0.0 |  3.254167 |                84.125 |                   0.1 |                      7.393446 |            15.334225 |   0.0 | -0.4 |  3.254167 |    5.050025 |         5.424101 |           5.050025 |          6.3 |                 5.931876 |                      7.718916 |                         0.0 |                          0.0 |                      0.434155 |                           0.547693 |             4.134518 |                  6.493538 |    1009.865482 |   1.293105 | 8.583333 |       0.0 |        2.4 |             69.371062 |                      0.0 |                 69.371062 |                  13.800803 |                 14.023359 |                     0.0 |                      0.0 |    55.347703 |             0.0 |              0.0 |         -0.222556 |             33.810423 |                 100.0 |                  82.4181 |               97.881113 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.131541 |      0.0 | 0.046731 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.925328 |  0.25687 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.124564 | 0.034601 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |        0.0 |      nan |  554.228525 | 93.140954 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |  0.09815 | 0.026414 |   0.0 |   0.0 |  0.0 |   0.0 | 0.034601 |
    | 2010-12-12 14:00:00 |  0.0 | -0.6 |              0.2 |                      1.0 |       60.254042 |             66.9 |       7.3 |              1015.0 | 0.0 | 0.0 |  2.958333 |             83.483333 |                   0.3 |                      7.393446 |            16.085314 |   0.0 | -0.6 |  2.958333 |    5.851616 |         5.347282 |           5.851616 |          7.3 |                 5.845604 |                       7.55838 |                         0.0 |                          0.0 |                      0.428574 |                           0.537642 |             3.910709 |                  6.309988 |    1011.089291 |    1.29544 | 8.583333 |       0.0 |        3.6 |             54.228638 |                      0.0 |                 54.228638 |                  14.476783 |                 15.622576 |                     0.0 |                      0.0 |    38.606061 |             0.0 |              0.0 |         -1.145794 |             29.178858 |                 100.0 |                82.479539 |               97.889901 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.146499 |      0.0 | 0.047228 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.924657 | 0.256497 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.119404 | 0.033168 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |        0.0 |      nan |  559.211858 | 93.093726 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.095615 | 0.023789 |   0.0 |   0.0 |  0.0 |   0.0 | 0.033168 |
    | 2010-12-12 15:00:00 |  0.0 | -0.8 |              0.2 |                      1.0 |         14.8373 |             64.4 |       7.5 |              1015.0 | 0.0 | 0.0 |    2.6625 |             82.729167 |                   0.4 |                      7.393446 |            16.177955 |   0.0 | -0.8 |    2.6625 |    6.011934 |         5.293842 |           6.011934 |          7.5 |                 5.760442 |                      7.400795 |                         0.0 |                          0.0 |                      0.423055 |                           0.527748 |             3.709725 |                  6.122616 |    1011.290275 |   1.296488 | 8.583333 |       0.0 |        4.8 |              13.35357 |                      0.0 |                  13.35357 |                   14.56016 |                 16.526635 |                     0.0 |                      0.0 |    -3.173065 |             0.0 |              0.0 |         -1.966476 |             28.400756 |                 100.0 |                82.541638 |               97.898771 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.135015 |      0.0 | 0.042618 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.924147 | 0.256214 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |  0.11457 | 0.031825 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |        0.0 |      nan |  562.995191 | 93.051109 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.093145 | 0.021424 |   0.0 |   0.0 |  0.0 |   0.0 | 0.031825 |
    | 2010-12-12 16:00:00 |  0.0 | -1.3 |              0.2 |                 0.046405 |        0.168856 |             65.4 |       7.0 |              1017.0 | 0.0 | 0.0 |    2.3625 |             82.029167 |                   0.6 |                      7.386143 |            16.184078 |   0.0 | -1.3 |    2.3625 |    5.611139 |         5.237063 |           5.611139 |          7.0 |                 5.552313 |                      7.243956 |                         0.0 |                          0.0 |                      0.409522 |                           0.517875 |             3.631213 |                  5.942156 |    1013.368787 |   1.301474 | 8.583333 |       0.0 |        7.8 |               0.15197 |                      0.0 |                   0.15197 |                   14.56567 |                 18.187261 |                     0.0 |                      0.0 |   -18.035291 |             0.0 |              0.0 |         -3.621591 |             30.429381 |                 100.0 |                82.597682 |               98.771746 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.11254 |      0.0 | 0.036657 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.924042 | 0.256156 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.110034 | 0.030565 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |        0.0 |      nan |  563.778525 | 93.014451 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.090739 | 0.019295 |   0.0 |   0.0 |  0.0 |   0.0 | 0.030565 |
    | 2010-12-12 17:00:00 |  0.0 | -1.6 |              0.0 |                      0.0 |             0.0 |             67.6 |       6.0 |              1017.0 | 0.0 | 0.0 |  2.066667 |             81.658333 |                   0.6 |                      7.386143 |            16.184078 |   0.0 | -1.6 |  2.066667 |    4.809547 |         5.140204 |           4.809547 |          6.0 |                 5.430651 |                      7.092172 |                         0.0 |                          0.0 |                      0.401583 |                           0.508295 |              3.67112 |                  5.791349 |     1013.32888 |   1.302892 | 8.583333 |       0.0 |        9.6 |                   0.0 |                      0.0 |                       0.0 |                   14.56567 |                 18.333316 |                     0.0 |                      0.0 |   -18.333316 |             0.0 |              0.0 |         -3.767646 |             35.500944 |                 100.0 |                82.645894 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |   0.08847 |      0.0 |  0.03163 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.924179 | 0.256232 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.105773 | 0.029381 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |        0.0 |      nan |  562.761858 | 92.982822 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.088395 | 0.017377 |   0.0 |   0.0 |  0.0 |   0.0 | 0.029381 |
    | 2010-12-12 18:00:00 |  0.0 | -2.0 |              0.0 |                      0.0 |             0.0 |             70.2 |       4.8 |              1018.0 | 0.0 | 0.0 |  1.770833 |             81.179167 |                   0.6 |                      7.386143 |            16.184078 |   0.0 | -2.0 |  1.770833 |    3.847638 |         5.036665 |           3.847638 |          4.8 |                 5.272101 |                      6.943199 |                         0.0 |                          0.0 |                      0.391203 |                           0.498867 |             3.701015 |                  5.636431 |    1014.298985 |   1.306085 | 8.583333 |       0.0 |       12.0 |                   0.0 |                      0.0 |                       0.0 |                   14.56567 |                 18.489582 |                     0.0 |                      0.0 |   -18.489582 |             0.0 |              0.0 |         -3.923912 |             44.376181 |                 100.0 |                82.687497 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.063209 |      0.0 | 0.025658 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.924639 | 0.256487 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.101762 | 0.028267 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |        0.0 |      nan |  559.345191 | 92.957163 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.086112 |  0.01565 |   0.0 |   0.0 |  0.0 |   0.0 | 0.028267 |
    | 2010-12-12 19:00:00 |  0.0 | -2.1 |              0.0 |                      0.0 |             0.0 |             70.4 |       4.2 |              1019.0 | 0.0 | 0.0 |  1.483333 |               80.6375 |                   0.6 |                      7.386143 |            16.184078 |   0.0 | -2.1 |  1.483333 |    3.366683 |         4.939806 |           3.366683 |          4.2 |                 5.233109 |                      6.801075 |                         0.0 |                          0.0 |                      0.388644 |                           0.489848 |             3.684109 |                  5.484217 |    1015.315891 |    1.30786 | 8.583333 |       0.0 |       12.6 |                   0.0 |                      0.0 |                       0.0 |                   14.56567 |                 18.648085 |                     0.0 |                      0.0 |   -18.648085 |             0.0 |              0.0 |         -4.082415 |             50.715635 |                 100.0 |                82.721247 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.054415 |      0.0 |  0.02375 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 |  0.92518 | 0.256788 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.097982 | 0.027217 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |        0.0 |      nan |  555.328525 | 92.933413 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.083888 | 0.014095 |   0.0 |   0.0 |  0.0 |   0.0 | 0.027217 |
    | 2010-12-12 20:00:00 |  0.0 | -2.2 |              0.0 |                      0.0 |             0.0 |             69.5 |       4.9 |              1019.0 | 0.0 | 0.0 |    1.1875 |             80.083333 |                   0.6 |                      7.386143 |            16.184078 |   0.0 | -2.2 |    1.1875 |    3.927797 |         4.889707 |           3.927797 |          4.9 |                 5.194372 |                      6.657516 |                         0.0 |                          0.0 |                        0.3861 |                           0.480714 |             3.610088 |                  5.331561 |    1015.389912 |   1.308379 | 8.583333 |       0.0 |       13.2 |                   0.0 |                      0.0 |                       0.0 |                   14.56567 |                 18.808978 |                     0.0 |                      0.0 |   -18.808978 |             0.0 |              0.0 |         -4.243308 |             43.470544 |                 100.0 |                82.752491 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.065452 |      0.0 |  0.02619 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.925802 | 0.257133 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.094415 | 0.026226 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |        0.0 |      nan |  550.711858 | 92.907223 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.081721 | 0.012694 |   0.0 |   0.0 |  0.0 |   0.0 | 0.026226 |
    | 2010-12-12 21:00:00 |  0.0 | -2.6 |              0.0 |                      0.0 |             0.0 |             71.6 |       4.0 |              1020.0 | 0.0 | 0.0 |     0.875 |             79.491667 |                   0.6 |                      7.386143 |            16.184078 |   0.0 | -2.6 |     0.875 |    3.206365 |         4.782828 |           3.206365 |          4.0 |                 5.041946 |                       6.50878 |                         0.0 |                          0.0 |                      0.376066 |                           0.471225 |             3.610033 |                  5.173937 |    1016.389967 |   1.311601 | 8.583333 |       0.0 |       15.6 |                   0.0 |                      0.0 |                       0.0 |                   14.56567 |                 18.977604 |                     0.0 |                      0.0 |   -18.977604 |             0.0 |              0.0 |         -4.411934 |             53.251417 |                 100.0 |                82.786946 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.048111 |      0.0 | 0.021424 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.926748 | 0.257658 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.091042 | 0.025289 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |        0.0 |      nan |  543.695191 | 92.885799 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |  0.07961 | 0.011432 |   0.0 |   0.0 |  0.0 |   0.0 | 0.025289 |
    | 2010-12-12 22:00:00 |  0.0 | -3.1 |              0.0 |                      0.0 |             0.0 |             74.5 |       3.4 |              1020.0 | 0.0 | 0.0 |      0.55 |             78.945833 |                   0.6 |                      7.386143 |            16.184078 |   0.0 | -3.1 |      0.55 |     2.72541 |         4.662589 |            2.72541 |          3.4 |                 4.856984 |                      6.357212 |                         0.0 |                          0.0 |                       0.36384 |                           0.461527 |             3.618453 |                  5.018754 |    1016.381547 |   1.314025 | 8.583333 |       0.0 |       18.6 |                   0.0 |                      0.0 |                       0.0 |                   14.56567 |                 19.143113 |                     0.0 |                      0.0 |   -19.143113 |             0.0 |              0.0 |         -4.577443 |             62.648725 |                 100.0 |                82.815133 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.034966 |      0.0 | 0.016822 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.928097 | 0.258409 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 |  0.08785 | 0.024403 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |        0.0 |      nan |  533.678525 | 92.868976 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 | 0.077553 | 0.010296 |   0.0 |   0.0 |  0.0 |   0.0 | 0.024403 |
    | 2010-12-12 23:00:00 |  0.0 | -3.9 |              0.0 |                      0.0 |             0.0 |             80.2 |       3.4 |              1020.0 | 0.0 | 0.0 |  0.229167 |             78.720833 |                   0.6 |                      7.386143 |            16.184078 |   0.0 | -3.9 |  0.229167 |     2.72541 |          4.54903 |            2.72541 |          3.4 |                  4.57351 |                      6.210652 |                         0.0 |                          0.0 |                      0.344989 |                           0.452122 |             3.667955 |                  4.889077 |    1016.332045 |   1.317905 | 8.583333 |       0.0 |       23.4 |                   0.0 |                      0.0 |                       0.0 |                   14.56567 |                 19.269853 |                     0.0 |                      0.0 |   -19.269853 |             0.0 |              0.0 |         -4.704183 |             62.648725 |                 100.0 |                82.837267 |               98.814229 |      0.0 |     0.0 |        0.0 |         0.0 |     0.0 |     0.0 |      0.0 |             nan |     0.0 |        0.0 |     0.0 |        0.0 |     0.0 |    0.0 |  0.024921 |      0.0 |  0.01189 |       0.0 |      0.0 |          nan |          0.1 |      0.0 |      0.0 |      0.0 |      0.0 |     0.0 |        0.0 |      0.0 |        0.0 |      0.0 |         0.0 | 0.930093 | 0.259522 |      0.0 |      0.0 |  0.0 |  0.0 | 0.0 |  0.0 |      0.0 | 0.084823 | 0.023562 |      0.0 |   0.0 |  0.0 |      0.0 |   nan |       0.0 |       0.0 |        0.0 |      nan |  518.861858 | 92.857086 |      0.0 |      0.0 |   0.0 |   0.0 |  0.0 |  0.07555 | 0.009273 |   0.0 |   0.0 |  0.0 |   0.0 | 0.023562 |

>>> round_(model.check_waterbalance(conditions_acker_winter_ktschnee))
0.0
"""
# import...
# from standard-library
from typing import *

# ...from HydPy
from hydpy.exe.modelimports import *
from hydpy.core import masktools
from hydpy.core import modeltools
from hydpy.core.typingtools import *

# ...from lland
from hydpy.models.lland import lland_model
from hydpy.models.lland import lland_masks
from hydpy.models.lland.lland_constants import *


class Model(modeltools.AdHocModel):
    """Penman-Monteith and Knauf with snow interception version of HydPy-L-Land
    (|lland_v4|)."""

    INLET_METHODS = (lland_model.Pick_QZ_V1,)
    RECEIVER_METHODS = ()
    ADD_METHODS = (
        lland_model.Return_AdjustedWindSpeed_V1,
        lland_model.Return_ActualVapourPressure_V1,
        lland_model.Return_NetLongwaveRadiationInz_V1,
        lland_model.Return_NetLongwaveRadiationSnow_V1,
        lland_model.Return_Penman_V1,
        lland_model.Return_PenmanMonteith_V1,
        lland_model.Return_EnergyGainSnowSurface_V1,
        lland_model.Return_SaturationVapourPressure_V1,
        lland_model.Return_SaturationVapourPressureSlope_V1,
        lland_model.Return_NetRadiation_V1,
        lland_model.Return_WSensInz_V1,
        lland_model.Return_WSensSnow_V1,
        lland_model.Return_WLatInz_V1,
        lland_model.Return_WLatSnow_V1,
        lland_model.Return_WSurfInz_V1,
        lland_model.Return_WSurf_V1,
        lland_model.Return_TempS_V1,
        lland_model.Return_TempSInz_V1,
        lland_model.Return_WG_V1,
        lland_model.Return_BackwardEulerErrorInz_V1,
        lland_model.Return_BackwardEulerError_V1,
        lland_model.Return_ESnowInz_V1,
        lland_model.Return_ESnow_V1,
        lland_model.Return_TempSSurface_V1,
    )
    RUN_METHODS = (
        lland_model.Calc_QZH_V1,
        lland_model.Update_LoggedTemL_V1,
        lland_model.Calc_TemLTag_V1,
        lland_model.Update_LoggedRelativeHumidity_V1,
        lland_model.Calc_DailyRelativeHumidity_V1,
        lland_model.Update_LoggedSunshineDuration_V1,
        lland_model.Calc_DailySunshineDuration_V1,
        lland_model.Update_LoggedPossibleSunshineDuration_V1,
        lland_model.Calc_DailyPossibleSunshineDuration_V1,
        lland_model.Update_LoggedGlobalRadiation_V1,
        lland_model.Calc_DailyGlobalRadiation_V1,
        lland_model.Calc_NKor_V1,
        lland_model.Calc_TKor_V1,
        lland_model.Calc_TKorTag_V1,
        lland_model.Calc_WindSpeed2m_V1,
        lland_model.Calc_ReducedWindSpeed2m_V1,
        lland_model.Update_LoggedWindSpeed2m_V1,
        lland_model.Calc_DailyWindSpeed2m_V1,
        lland_model.Calc_WindSpeed10m_V1,
        lland_model.Calc_SaturationVapourPressure_V1,
        lland_model.Calc_DailySaturationVapourPressure_V1,
        lland_model.Calc_SaturationVapourPressureSlope_V1,
        lland_model.Calc_DailySaturationVapourPressureSlope_V1,
        lland_model.Calc_ActualVapourPressure_V1,
        lland_model.Calc_DailyActualVapourPressure_V1,
        lland_model.Calc_NBes_Inzp_V1,
        lland_model.Calc_SNRatio_V1,
        lland_model.Calc_SBes_V1,
        lland_model.Calc_SnowIntMax_V1,
        lland_model.Calc_SnowIntRate_V1,
        lland_model.Calc_NBesInz_V1,
        lland_model.Calc_SBesInz_V1,
        lland_model.Calc_STInz_V1,
        lland_model.Calc_WaDaInz_SInz_V1,
        lland_model.Calc_WNiedInz_ESnowInz_V1,
        lland_model.Calc_TempSInz_V1,
        lland_model.Update_ASInz_V1,
        lland_model.Calc_ActualAlbedoInz_V1,
        lland_model.Calc_NetShortwaveRadiationInz_V1,
        lland_model.Calc_RLAtm_V1,
        lland_model.Update_ESnowInz_V1,
        lland_model.Calc_SchmPotInz_V1,
        lland_model.Calc_SchmInz_STInz_V1,
        lland_model.Calc_GefrPotInz_V1,
        lland_model.Calc_GefrInz_STInz_V1,
        lland_model.Calc_EvSInz_SInz_STInz_V1,
        lland_model.Update_WaDaInz_SInz_V1,
        lland_model.Update_ESnowInz_V2,
        lland_model.Calc_WATS_V2,
        lland_model.Calc_WaDa_WAeS_V2,
        lland_model.Calc_WNied_ESnow_V1,
        lland_model.Calc_TempS_V1,
        lland_model.Update_TauS_V1,
        lland_model.Calc_ActualAlbedo_V1,
        lland_model.Calc_NetShortwaveRadiationSnow_V1,
        lland_model.Calc_TZ_V1,
        lland_model.Calc_WG_V1,
        lland_model.Update_ESnow_V1,
        lland_model.Calc_SchmPot_V2,
        lland_model.Calc_Schm_WATS_V1,
        lland_model.Calc_GefrPot_V1,
        lland_model.Calc_Gefr_WATS_V1,
        lland_model.Calc_EvS_WAeS_WATS_V1,
        lland_model.Update_WaDa_WAeS_V1,
        lland_model.Update_ESnow_V2,
        lland_model.Calc_NetShortwaveRadiation_V1,
        lland_model.Calc_DailyNetShortwaveRadiation_V1,
        lland_model.Calc_DailyNetLongwaveRadiation_V1,
        lland_model.Calc_NetRadiation_V1,
        lland_model.Calc_DailyNetRadiation_V1,
        lland_model.Calc_DryAirPressure_V1,
        lland_model.Calc_DensityAir_V1,
        lland_model.Calc_AerodynamicResistance_V1,
        lland_model.Calc_SoilSurfaceResistance_V1,
        lland_model.Calc_LanduseSurfaceResistance_V1,
        lland_model.Calc_ActualSurfaceResistance_V1,
        lland_model.Calc_G_V2,
        lland_model.Calc_EvPo_V2,
        lland_model.Calc_EvI_Inzp_V3,
        lland_model.Calc_EvB_V2,
        lland_model.Update_EBdn_V1,
        lland_model.Calc_SFF_V1,
        lland_model.Calc_FVG_V1,
        lland_model.Calc_QKap_V1,
        lland_model.Calc_QBB_V1,
        lland_model.Calc_QIB1_V1,
        lland_model.Calc_QIB2_V1,
        lland_model.Calc_QDB_V1,
        lland_model.Update_QDB_V1,
        lland_model.Calc_BoWa_V1,
        lland_model.Calc_QBGZ_V1,
        lland_model.Calc_QIGZ1_V1,
        lland_model.Calc_QIGZ2_V1,
        lland_model.Calc_QDGZ_V1,
        lland_model.Calc_QBGA_V1,
        lland_model.Update_QDGZ_QBGZ_QBGA_V1,
        lland_model.Update_QDGZ_QBGZ_QBGA_V2,
        lland_model.Calc_QIGA1_V1,
        lland_model.Calc_QIGA2_V1,
        lland_model.Calc_QDGZ1_QDGZ2_V1,
        lland_model.Calc_QDGA1_V1,
        lland_model.Calc_QDGA2_V1,
        lland_model.Calc_QAH_V1,
        lland_model.Calc_QA_V1,
    )
    OUTLET_METHODS = (lland_model.Pass_QA_V1,)
    SENDER_METHODS = ()
    SUBMODELS = (
        lland_model.PegasusESnowInz,
        lland_model.PegasusESnow,
        lland_model.PegasusTempSSurface,
    )
    idx_hru = modeltools.Idx_HRU()

    def check_waterbalance(
        self,
        initial_conditions: Dict[str, Dict[str, ArrayFloat]],
    ) -> float:
        r"""Determine the water balance error of the previous simulation run in mm.

        Method |Model.check_waterbalance| calculates the balance error as follows:

          .. math::
            \sum_{t=t0}^{t1} \Bigg(
            NKor_t
            + QZH_t
            - \bigg( \sum_{k=1}^{nhru} fhru^k \cdot \Big( EvI_t^k + EvSInz_t^k + EvS_t^k + EvB_t^k \Big) \bigg)
            - QAH_t
            \Bigg)
            + \sum_{k=1}^{nhru} fhru^k \cdot \bigg(
            \Big( Inzp_{t0}^k - Inzp_{t1}^k \Big)
            + \Big( SInz_{t0}^k - SInz_{t1}^k \Big)
            + \Big( WAeS_{t0}^{k} - WAeS_{t1}^{k} \Big)
            + \Big( BoWa_{t0}^{k} - BoWa_{t1}^{k} \Big)
            \bigg) -
            \sum_{t=t0}^{t1} \bigg(
            \Big( QDGZ1_t - QDGA1_t \Big)
            + \Big( QDGZ2_t - QDGA2_t \Big)
            + \Big( QIGZ1_t - QIGA1_t \Big)
            + \Big( QIGZ2_t - QIGA2_t \Big)
            + \Big( QBGZ_t - QBGA_t \Big)
            \bigg)

        The returned error should usually be in scale with numerical precision so that
        it does not affect the simulation results in any relevant manner.  The only
        exception we are aware of is the "generation" of additional water when the
        base flow storage cannot meet the water demand required for the calculated
        capillary rise (see :ref:`lland_v1_acker_qkap_negq-false`).

        Pick the required initial conditions before starting the simulation run via
        property |Sequences.conditions|.  See the integration tests of the application
        model |lland_v4| for some examples.
        """
        control = self.parameters.control
        fluxes = self.sequences.fluxes
        last = self.sequences.states
        first = initial_conditions["states"]
        idxs_water = numpy.isin(control.lnk, [WASSER, FLUSS, SEE])
        idxs_land = numpy.invert(idxs_water)
        idxs_soil = numpy.invert(numpy.isin(control.lnk, [VERS, WASSER, FLUSS, SEE]))
        idxs_forest = numpy.isin(control.lnk, [LAUBW, MISCHW, NADELW])
        return (
            numpy.sum(fluxes.nkor.series * control.fhru)
            + numpy.sum(fluxes.qzh.series)
            - numpy.sum(fluxes.evi.series * control.fhru)
            - numpy.sum((fluxes.evsinz.series * control.fhru)[:, idxs_forest])
            - numpy.sum((fluxes.evs.series * control.fhru)[:, idxs_land])
            - numpy.sum((fluxes.evb.series * control.fhru)[:, idxs_soil])
            - numpy.sum(fluxes.qah.series)
            - numpy.sum(((last.inzp - first["inzp"]) * control.fhru)[idxs_land])
            - numpy.sum(((last.sinz - first["sinz"]) * control.fhru)[idxs_forest])
            - numpy.sum(((last.waes - first["waes"]) * control.fhru)[idxs_land])
            - numpy.sum(((last.bowa - first["bowa"]) * control.fhru)[idxs_soil])
            - numpy.sum(last.qdgz1.series - last.qdga1.series)
            - numpy.sum(last.qdgz2.series - last.qdga2.series)
            - numpy.sum(last.qigz1.series - last.qiga1.series)
            - numpy.sum(last.qigz2.series - last.qiga2.series)
            - numpy.sum(last.qbgz.series - last.qbga.series)
        )


class Masks(masktools.Masks):
    """Masks applicable to |lland_v4|."""

    CLASSES = lland_masks.Masks.CLASSES


tester = Tester()
cythonizer = Cythonizer()
